#!/usr/local/bin/perl
# $Id: autogeoloc.pl,v 1.2 2006/11/03 11:19:46 ruben Exp $
#
# All Rights Reserved
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose and without fee is hereby granted,
# provided that the above copyright notice appear in all copies and that
# both that copyright notice and this permission notice appear in
# supporting documentation, and that the name of the author not be
# used in advertising or publicity pertaining to distribution of the
# software without specific, written prior permission.
#
# THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
# ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS; IN NO EVENT SHALL
# AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY
# DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
# AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

#------------------------------------------------------------------------------
# Module Header
# Filename          : autogeoloc.pl
# Purpose           : reads GPS coordinates from GENE files and geoloc ttreg entry. Then it takes one of the following actions:
#                     1 - update geoloc ttreg entry when geoloc value is null(only for newly installed testboxes)
#                     2 - print out "New GPS coordinates found on GENE file" if coordinates in GENE files and ttreg geolog field differ
#                     3 - print out "No Position line on GENE file" if GPS coordinates aren't present in the GENE files
# Author            : Maximo Alves
# Date              : 27 June  2002
# Description       :

# Language Version  : Perl 5
# OSs Tested        : Linux
# Command Line      : autogeoloc.pl -> search GENE files on /ncc/ttrpro/data/raw/year/month/day/
#                     autogeoloc.pl -d /directory/ -> search GENE files on /directory/
#                     autogeoloc.pl -h -> prints help message
# Input Files       :
# Output Files      :
# External Programs : ttregconfig, readttreg
#
# Problems          :
#
# To Do             :
#
# Comments          :
#
#------------------------------------------------------------------------------
use strict;
use POSIX;

# --------------- Variable declaration ------------------------------
my $ttroot      = "/ncc/ttpro";
my $RAWDATA     = "${ttroot}/data/raw/";
my $VERBOSE     = "1";
my $datafolder  = "";
my $READTTREG   = "${ttroot}/bin/readttreg.pl";
my $TTREGCONFIG = "${ttroot}/bin/ttregconfig.pl";
my $ttconfig_cmd = strftime("${ttroot}/bin/ttconfig -d '%Y-%m-%d 00:00:00'", gmtime(time()));
# --------------------- Sub routines --------------------------------

sub determine_current_day_folder {
    my $year  = "";
    my $month = "";
    my $day   = "";
    my $time  = time;

    my @aux = localtime( $time - 86400 );    #actually yesterday
    $year       = $aux[5] + 1900;
    $month      = $aux[4] + 1;
    $day        = $aux[3];
    $month      = "0" . $month if ( $month < 10 );
    $day        = "0" . $day if ( $day < 10 );
    $datafolder = $RAWDATA . $year . "/" . $month . "/" . $day . "/";

}    #sub  determine_current_day_folder ends here !!!

sub get_GPS_values {

#Thu Jun 27 00:00:02 2002:Position from 20020626 230000 to 235959: 35.0968 +/- 0.0000, 33.2135 +/- 0.0000

    my $line        = $_[0];
    my @aux         = split( /:/, $line );
    my @aux1        = split( /,/, $aux[4] );
    my $reclatitude = $aux1[0];
    $reclatitude =~ s/^\s//;
    my $reclongitude = $aux1[1];
    $reclongitude =~ s/^\s//;

    chomp($reclongitude);

    @aux = split( /\+\/\-/, $aux1[0] );
    my $latitude = $aux[0];
    @aux = split( /\+\/\-/, $aux1[1] );
    my $longitude = $aux[0];

    #extract hour and minutes from hour minutes seconds
    my ( $hour, $min ) = $latitude =~ /(\d+)\.(\d\d)/;
    $latitude = $hour . "." . $min;
    my ( $hour, $min ) = $longitude =~ /(\d+)\.(\d\d)/;
    $longitude = $hour . "." . $min;

    return ( $latitude, $longitude, $reclatitude, $reclongitude );
}    #sub get_GPS_values ends here !!!

sub get_GPS_altitude {

#Mon Dec  1 23:00:12 2003:Altitude from 20031201 220000 to 230000: 93.5429 +/- 1.5717
    my $line        = $_[0];
    my @aux         = split( /:/, $line );
    my @aux1        = split( /,/, $aux[4] );
    my $recaltitude = $aux1[0];
    $recaltitude =~ s/^\s//;
    @aux = split( /\+\/\-/, $aux1[0] );
    my $altitude = $aux[0];

    return ( $altitude, $recaltitude );
}    #sub get_GPS_altitude ends here !!!

sub get_ttreg_geoloc {

    #GENE.tt01.ripe.net.20020626
    my ( $tb, $aux, @aux, @aux1, $latitude, $longitude, $altitude );

    @aux = split( /\./, $_[0] );
    $tb  = $aux[1];
    $tb  = "tt04" if ( $tb =~ /jordaan/ );
    $tb  = "tt02" if ( $tb =~ /osdorp/ );
    $aux = `$READTTREG $tb geoloc`;
    if ( !($aux) ) {
        return ( $tb, 0, 0, 0 );
    }
    @aux  = split( /,/,      $aux );
    @aux1 = split( /\+\/\-/, $aux[0] );
    $latitude  = $aux1[0];
    @aux1      = split( /\+\/\-/, $aux[1] );
    $longitude = $aux1[0];
    if ( defined( $aux[2] ) ) {
        @aux1 = split( /\+\/\-/, $aux[2] );
        $altitude = $aux1[0];
    }
    else {
        $altitude = "0.0000";
    }

    #extract hour and minutes from hour minutes seconds
    $_ = $latitude;
    my ( $hour, $min ) = $latitude =~ /(\d+)\.(\d\d)/;
    $latitude = $hour . "." . $min;
    $_        = $longitude;
    my ( $hour, $min ) = $longitude =~ /(\d+)\.(\d\d)/;
    $longitude = $hour . "." . $min;

    return ( $tb, $latitude, $longitude, $altitude );

}    #sub get_ttreg_geoloc emds here !!!

# --------------- Start Point -------------------------------------

#parse command line

my $parameter = "";
my @cdmaboxes = qx($ttconfig_cmd -v CDMALIST);
chomp(@cdmaboxes);

$parameter = shift(@ARGV);
if ($parameter) {
    if ( $parameter eq "-d" ) {
        $datafolder = shift(@ARGV);
    }
    elsif ( $parameter eq "-h" ) {
        print "*** autogeoloc.pl Help *** \n";
        print "Command line: \n";
        print
"autogeoloc.pl -> search GENE files on /ncc/ttrpro/data/raw/year/month/day/ \n";
        print
          "autogeoloc.pl -d /directory/ -> search GENE files on /directory/ \n";
        print "autogeoloc.pl -h -> prints help message \n";
        print "Purpose: \n";
        print
"It reads GPS coordinates from GENE files and geoloc ttreg entry. Then it takes one of the following actions: \n";
        print
" 1 - update geoloc ttreg entry when geoloc value is null(only for newly installed testboxes)\n";
        print
" 2 - print out 'New GPS coordinates found on GENE file' if coordinates in GENE files and ttreg geolog field differ\n";
        print
" 3 - print out 'No Position line on GENE file' if GPS coordinates aren't present in the GENE files \n";
    }
}
else {
    &determine_current_day_folder;
}
opendir( NEW, $datafolder ) || die "Cannot open $datafolder \n";
while ( my $filename = readdir(NEW) ) {
    if ( $filename =~ /^GENE/ ) {
        if ( open( FP, $datafolder . $filename ) ) {
            my $firstlocline = "";
            my $firstaltline = "";
            my $Positionflag = 1;
            my $Altitudeflag = 1;
            while (<FP>) {
				chomp();
                if (/Position/) {
                    $Positionflag = 0;
                    if ( !(/0.0000 \+\/\- 0.0000,  0.0000 \+\/\- 0.0000/) ) {
                        $firstlocline = $_;
                    }
                }
                if (/Altitude/) {
                    $Altitudeflag = 0;
                    if ( !(/0.0000 \+\/\- 0.0000/) ) {
                        $firstaltline = $_;
                    }
                }
            }
            my ( $tb, $ttreglatitude, $ttreglongitude, $ttregaltitude ) =
              &get_ttreg_geoloc($filename);
            next if grep { /^${tb}$/ } @cdmaboxes;
            if ($Positionflag) {

                #no position line on GENE file, go for next file
                print
"WARNING: $filename has no GPS coordinates(e.g. no Position line) ! \n";
                next;
            }
            if ($Altitudeflag) {

                #no altitude line on GENE file, go for next file
                print
"WARNING: $filename has no GPS coordinates(e.g. no Altitude line) ! \n";
                next;
            }

            my (
                $currentlatitude, $currentlongitude,
                $reclatitude,     $reclongitude
              )
              = &get_GPS_values($firstlocline);
            my ( $currentaltitude, $recaltitude ) =
              &get_GPS_altitude($firstaltline);
            if (   ( !($ttreglatitude) )
                && ( !($ttreglongitude) )
                && ( !($ttregaltitude) ) )
            {    #new test box
                my $auxrec = "$reclatitude, $reclongitude";
                print
"WARNING: $tb does not have geoloc position values in ttreg  \n";
                my $aux = "$TTREGCONFIG MOD geoloc $tb '$auxrec'";

                system($aux);
                print "$auxrec geoloc value stored in ttreg \n";
            }
            elsif (( $ttreglatitude ne $currentlatitude )
                || ( $ttreglongitude ne $currentlongitude )
                 )
            {

                #changes on existing box
                my $auxrec = "$reclatitude, $reclongitude, $recaltitude";
                print "WARNING: New GPS coordinates found on GENE $tb file. The coordinates have changed from $ttreglatitude, $ttreglongitude, $ttregaltitude to $auxrec\n";
		if (open(GEOLOC2MAPQUEST, "echo '$auxrec' | /ncc/ttpro/bin/ttreg2mapquest |"))
		{
			my (@output) = (<GEOLOC2MAPQUEST>);
			print "Look at the following URL if the change makes any sense:\n\n", (split(/\s+/,$output[$#output]))[1],"\n\n";
			close(GEOLOC2MAPQUEST);
		}
                print "Please change ttreg geoloc entry !!! \n";
            }
            elsif ( ( $ttregaltitude  ne $currentaltitude ) && $currentaltitude)
            {
                #changes on existing box
                my $auxrec = "$reclatitude, $reclongitude, $recaltitude";
                my $aux = "$TTREGCONFIG MOD geoloc $tb '$auxrec' > /dev/null 2>&1";
                # print "INFO: New GPS altitude found on GENE $tb file. The height has changed from $ttregaltitude to $recaltitude.  \n";
                system($aux);
                # print "(changes silently accepted)\n";
            }

        }
        else {
            print "ERROR: couldn't open $filename ($!)\n";
        }
    }
    else {
        next;
    }
}    # while(my $filename
