/*
Copyright (c) 2000                      RIPE NCC

$Id: FillTree.cxx,v 1.7 2006/11/04 14:40:49 ruben Exp $

All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of the author not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS; IN NO EVENT SHALL
AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY
DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

/*
-------------------------------------------------------------------------------
Module Header
Filename          : FillTree.C
Author            : Henk Uiterwaal, Rene Wilhelm
Date              : 18-DEC-1998
Revised           : 26-MAY-2000, specify output file on command line
Revised           : 17-DEC-2002, allow PCKB files to be in .gz format
Description       : Create one TTM delay ROOT file from specified PCKB file(s)
Language Version  : C++
Command Line	  : FillTree -o <filename.root>  <file> ...
-------------------------------------------------------------------------------
*/

#include "TROOT.h"
#include "TError.h"
#include "TTree.h"
#include "TFile.h"
#include "TBranch.h"

#include "Delay.h"

#include "stdio.h"
#include "stdlib.h"
#include "unistd.h"

/* TRUE and FALSE */
#ifndef TRUE
  #define TRUE (1)
  #define FALSE (0)
#endif


/*
 *   global variables
 */


char *progname;

/*
 * Subroutine header
 *
 * Purpose:       print message how to use the program
 * Side effect:	  program is terminated
 *
 */

void usage () {
   fprintf(stderr, "Usage: %s -d -o <outputfile> <inputfile> ...\n", progname );
   exit(1);
}

/*
 * Subroutine header
 *
 * Purpose:       open data file
 * Description:   opens a data file in specified mode
 *                if filename to be read ends in .gz, read via gzcat
 *                if filename to be read does not exist, check if .gz exists
 *                return NULL in case of error
 */

FILE *openfile (char *filename, char *mode, int *gzip) {

	char *temp;
	FILE *fp;
	int found;

	found = FALSE;
	*gzip = FALSE;

	if (strchr(mode, 'r') == NULL) {
		/* open for writing */
		fp = fopen (filename, mode);
	}
	else {
		/* open for reading */
		temp = filename + strlen(filename) - 3; 
		if (strcmp(temp, ".gz") == 0 ) {
		        *gzip = TRUE;
			temp = (char *) malloc (strlen(filename) + 7);
        		temp = strcpy (temp, "gzcat "); 
			temp = strcat (temp, filename);
			fp = popen(temp, mode);
			free(temp);
		}
		else if ((fp = fopen(filename, mode)) == NULL) {
			/* open failed, check if .gz exists */

			temp = (char *) malloc (strlen(filename) + 10);
        		temp = strcpy (temp, filename); 
			temp = strcat (temp, ".gz");

			if ((fp = fopen (temp, "r")) != NULL) {
	         		/* yes it does! */
				fclose(fp);
		        	*gzip = TRUE;
				temp = strcpy (temp, "gzcat "); 
				temp = strcat (temp, filename);
				temp = strcat (temp, ".gz");
				fp = popen(temp, mode);
			}
			free(temp);
		}
	}
	return (fp);
}

/*
 * Subroutine header
 *
 * Purpose:       the main program
 *
 */
	
int main (int argc, char **argv) {
 
	extern char *optarg;
	extern int optind;
	int ch, oflag, verbose, debug;
	
	gErrorIgnoreLevel=1;

	TROOT simple ("Simple", "Example");

	Int_t split = 1;
	Int_t bsize = 64000;
	Delay *delay = 0;

	FILE  *fp;
	TFile *hfile;
	int    readpipe;	// true when reading from a popen stream

	struct packet *newpacket;
	int events, i; 

	oflag = 0;
	verbose = 0;
	progname=argv[0];
	while ((ch = getopt(argc, argv, "dvo:")) != -1) {
		switch(ch) {
		case 'o':
			if (oflag == 0) {
				oflag = 1;    
				hfile = new TFile (optarg, "RECREATE");
			}
			else {
				usage();
			}
			break;
		case 'v':
			verbose = 1;    
		 	break;
		case 'd':
			gErrorIgnoreLevel=0;
			debug = 1;    
		 	break;
		default:
			usage();
		}
	}

	if (oflag == 0) {
		usage();
	}

	// move argc/argv forward to start of <filenames>

	argc -= optind;	
	argv += optind;

 	newpacket = (struct packet *) malloc (sizeof(struct packet));
	TTree *tree  = new TTree ("tree", "Example");

	delay = new Delay();
	tree->Branch ("delay", "Delay", &delay, bsize, split);

	for (i=0; i<argc; i++) {

		// Open data file 
		if (verbose == 1) {
			printf("opening %s\n", argv[i]);
		}
		fp = openfile (argv[i], "r", &readpipe);
		if (fp == NULL) {
			fprintf (stderr, "%s: cannot open %s\n", progname, argv[i]);
			exit(1);
		} 

		// Read file
		events = 0;
		while (fread (newpacket, sizeof(struct packet), 1, fp) != 0) {

			events++;
			delay = new Delay(newpacket);

			if (verbose && (events % 100) == 0) {
				printf("packet %d:\t target: %d , delay %f\n",
				   events, delay->GetTargetId(), delay->GetPacketDelay()); 
			}

			tree->Fill ();
			delete delay;	// can't be reused by ROOT

		}
		if (events == 0) {
			fprintf(stderr, "%s: no data found on file %s !?\n",
				progname, argv[i]);
		}
		else if (verbose == 1) {
			fprintf (stderr, "Read %d Delay Packets from file %s\n",
				events, argv[i]);
		} 
		if (readpipe) {
			fclose (fp);
		}
		else {
			pclose (fp);
		}
	}

	hfile->Write ();
	hfile->Close ();

	return 0;
} 

static char const rcsid[] = "$Id: FillTree.cxx,v 1.7 2006/11/04 14:40:49 ruben Exp $";
static char const *rcsid_p = rcsid;   // Prevent g++ from optimizing out rcsid
