#!/usr/bin/perl -w
#
#
# Copyright (c) 2000, 2001  RIPE NCC
#
# $Id: plotasnhops.pl,v 1.13 2007/01/26 15:06:27 ruben Exp $
#
#
# All Rights Reserved
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose and without fee is hereby granted,
# provided that the above copyright notice appear in all copies and that
# both that copyright notice and this permission notice appear in
# supporting documentation, and that the name of the author not be
# used in advertising or publicity pertaining to distribution of the
# software without specific, written prior permission.
#
# THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
# ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS; IN NO EVENT SHALL
# AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY
# DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
# AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
#------------------------------------------------------------------------------
# Module Header
# Filename          :   plotasnhops.pl
# Purpose           :   wrapper for astracestat
# Authors           :   Maximo Alves RIPE NCC
# Date              :   11-Sep-2002
# Description       :
# Language Version  :   perl5
# TODO              :
#------------------------------------------------------------------------
=pod

=head1 NAME

B<plotasnhops> -- Plot as number hops distribution, generate ROOT files for as number hops distribution

=head1 SYNOPSIS:

plotasnhops [-h] | [-n] [-v] [-d] [-p | -t | -r <ISO8601 period> ] 
    [-R <ROOT data dir> ] [-T <temporary data dir>] [-o <plots output dir>] 
    [-M <mysql server list>]
    start

=head1 OPTIONS

=over

=item B<-h> Short help message

=item B<-n> dry run, with proper Log::Log4perl debug level will show what was to be done

=item B<-v> Be verbosive

=item B<-d> Enable debug output

=item B<-p I<ISO8601 period>> Specify ISO8601 period. An ISO8601 period consist out of an ISO8601 time seperated by a duration specifier (P([0-9]+[DHMS]){1,})

=item B<-t I<ISO8601 period>> B<DEPRECIATED> Specify ISO8601 period. An ISO8601 period consist out of an ISO8601 time seperated by a duration specifier (P([0-9]+[DHMS]){1,})

=item B<-r I<ISO8601 period>> B<DEPRECIATED> Specify ISO8601 period. No plots will be generated wich is the default periods spanning more than one day. An ISO8601 period consist out of an ISO8601 time seperated by a duration specifier (P([0-9]+[DHMS]){1,})

=item B<-R I<ROOTdata_basedir>> specify an alternate base directory for ROOT data files.  YYYY/MM/DD is appended to it. It's used for input and output

=item B<-T I<temporary_dir>>  Where to store temporary files

=item B<-o I<plot_dir>>  Where to store generated plots. Plots are not synced to the webserver

=item B<-M I<mysql_server_list>> list of mysql servers to be used in merge_packets they are selected in round robin order

=back

By using B<start> you will automaticly select the last 1 day period leading up till today

=head1 NOTES

Previous versions of the nhop tools only produced histograms for the number of
hops (ip and as wise) seen on exact the same time the day before. That approach
was wrong and the current tools use ISO8601 period specification to produce
more reliable results and take the whole day into account.

The command line switches of the previous versions are thus not backwards
compatible.


=head1 DEBUGGING

plotasnhops uses Log::Log4perl and registers itself under the domain C<ttpro.plotasnhops>
Signalling merge_data with C<USR2> will force it to reread ${HOME}/.log4perl.conf

=cut

use strict;
use Time::Local;    # support functions for time/date calculations
use POSIX;
use Getopt::Std;
use Time::HiRes qw(gettimeofday tv_interval);
use LockFile::Simple;
use Log::Log4perl;
use DBI;

# --------------- Variable declaration ------------------------------

$ENV{PATH} = "/ncc/ttpro/bin:/usr/local/bin:/usr/ucb:/bin:/usr/bin";

$Getopt::Std::STANDARD_HELP_VERSION = 1 if $Getopt::Std::VERSION >= 1.05;

Log::Log4perl::init_and_watch( "$ENV{'HOME'}/.log4perl.conf", 'USR2' );

our $VERSION = sprintf "%d.%03d", q$Revision: 1.13 $ =~ /(\d+)/g;

my $Dsn    = "DBI:mysql:database=TTM;host=%s";
my $DBUser = "ttmuser";
my $DBPass = "";

my @MYSQL_Host = qw(localhost);

my %options;
my ( $t_start, $t_end );

my $TMPDIR          = "/var/tmp";
my $TTPRO           = "/ncc/ttpro";
my $PLOTDIR         = "${TTPRO}/data/plots/nhop/";
my $ROOTDIR         = "${TTPRO}/ttpro/data/root";
my $astracestat_bin = "/ncc/ttpro/bin/astracestat";
my $DEBUG           = 0;
my $VERBOSE         = 0;
my $DOMAIN          = ".ripe.net";
my $OUTFILE         = "plotasnhops.out";
my $ASQUERY         = "asquery";
my $ASQUERY_SQL = "SELECT DISTINCT Records.src, Records.dst, ASpaths.len, ASpaths.path FROM Records, ASpaths WHERE Records.src=%u AND Records.dst IN (%s) AND Records.aspathid=ASpaths.id AND ((Records.tstart BETWEEN %u AND %u ) OR (%u BETWEEN Records.tstart AND Records.tend+3600)) ORDER BY Records.dst,Records.tstart DESC";
my $SQLSERVER = "marula.ripe.net";

my $noplots      = 0;
my $dry_run      = 0;
my $TEMPROOTFILE = "as.temp.root";
my $TTCONFIG     = "/ncc/ttpro/bin/ttconfig ";

#set env variables

# Configure logger
my $logger = Log::Log4perl->get_logger('ttpro.plotasnhops');

my $t0 = [gettimeofday];
&main;
$logger->info("finished delta=", tv_interval($t0));

# --------------------- Sub routines --------------------------------

sub usage() {
	print <<"EOT";
plotasnhops [-h] | [-n] [-v] [-d] [-p | -t | -r <ISO8601 period> ] 
    [-R <ROOT data dir> ] [-T <temporary data dir>] [-o <plots output dir>] 
    [-M <mysql server list>]
    start

EOT

}    #sub usage ends here !!!

sub mkdirtree($)
{
	my $dirpath = $_[0];
	if (! -d $dirpath) {
		# create dirpath
		my $dirpath_base;
		foreach my $dirpart (split(/\//,$dirpath))
		{
			next unless defined($dirpart) && length($dirpart);
			$dirpath_base .= "/$dirpart";

			if(! -d $dirpath_base)
			{
				mkdir($dirpath_base) || $logger->logdie("$dirpath: mkdir $dirpath (at $dirpath_base) failed: $!");
			}
		}
	}
}

sub parse_commandline() {

	my $default_duration = strftime("%Y%m%d/P1D",(gmtime(time - 24 * 3600)));

	$logger->debug("process command line ...");

	getopts("hnp:t:r:vdo:T:R:M:",\%options);

	if ($options{'h'} ) {
		&usage;
		exit;
	}
	if ($options{'n'}) {
		$dry_run= "1";
	}
	if ($options{'d'}) {
		$DEBUG = "1";
	}
	if ($options{'v'}) {
		$VERBOSE = "1";
	}

	if ($options{'R'}) {
		$ROOTDIR = $options{'R'};
		if (-d $ROOTDIR)
		{
			$logger->debug("Setting ROOT directory to $ROOTDIR");
		}
		else
		{
			$logger->logdie("ROOT directory $ROOTDIR nonexistant!");
		}
	}
	if ($options{'T'}) {
		$TMPDIR = $options{'T'};
		if (-d $TMPDIR)
		{
			$logger->debug("Setting TMP directory to $TMPDIR");
		}
		else
		{
			$logger->logdie("TMP directory $TMPDIR nonexistant!");
		}
	}
	if ($options{'o'}) {
		$PLOTDIR = $options{'o'};
		if (-d $PLOTDIR)
		{
			$logger->debug("Setting PLOT directory to $PLOTDIR");
		}
		else
		{
			$logger->logdie("PLOT directory $PLOTDIR nonexistant!");
		}
	}
	if ($options{'r'}) {
		$logger->debug("Generate no plots");
		$noplots = 1;
		$options{'p'} = $options{'r'};
	}
	if ($options{'t'}) {
		$logger->warn("-t depreciated, please use -p <ISO8601 period>");
		$options{'p'} = $options{'t'};
	}
	if ($options{'p'} || grep { /^start$/} @ARGV ) {
		my $range = $options{'p'} ? $options{'p'} : $default_duration;
		if (($t_start,$t_end) = DateTime::Format::ISO8601::Range::range2time($range))
		{
			$logger->debug("Range is: ", scalar gmtime($t_start), " - " , scalar gmtime($t_end) );
		}
		else
		{
			$logger->logdie("Invalid ISO8601 time range \"$range\"");
		}
	}
	if (exists($options{'M'}))
	{
		@MYSQL_Host = split(/,/,$options{'M'});
		foreach my $MYSQL_Host (@MYSQL_Host)
		{
			if ( ! gethostbyname($MYSQL_Host)) {
				$logger->fatal("\"$MYSQL_Host\" unresolvable");
				exit 1;
			}
		}
		$logger->debug("changing list of available MySQL servers to (@MYSQL_Host)...");
	}
}

sub generatepath($) {
	my ($ref)  = $_[0];
	return sprintf("%s/%s",$ROOTDIR, strftime("%Y/%m/%d",gmtime($ref)));
}    # sub generatepath() ends here !!!

sub gettbid() {
	my $ret = 0;
	my $tb = ( split( /\./, $_[0] ) )[0];
	if ( $tb =~ /osdorp/ ) {
		$ret = 2;
	}
	elsif ( $tb =~ /jordaan/ ) {
		$ret = 4;
	}
	else {
		$_ = $tb;
		s/[a-zA-Z]+//;
		$ret = $_;
	}
	return $ret;
}

sub exec_astracestat() {
	my $flag = $_[0];
	my $t1 = [gettimeofday];

	my $execstr = "";
	if ($flag) {
		$execstr = "${astracestat_bin} -H -c 'notitle ' -f $TMPDIR/$OUTFILE  -s routestat.as.root -e $PLOTDIR  -b $TMPDIR/$TEMPROOTFILE";
	}
	else {
		$execstr = "${astracestat_bin} -c 'AS plots for TT' -f $TMPDIR/$OUTFILE  -s routestat.as.root -e $PLOTDIR  -b $TMPDIR/$TEMPROOTFILE";
	}
	if ($DEBUG) {
		$execstr = "$execstr -d";
	}
	if ($dry_run)
	{
		$logger->info("Dry run: $execstr");
	} else {
		$logger->debug("Running $execstr");
		my $resp = qx($execstr);

		$logger->debug("Finished astracestat delta=", tv_interval($t1));

		if ( $resp =~ /ERROR/ ) {
			$logger->fatal($resp);
		}
		elsif ( $resp =~ /WARNING/ ) {
			$logger->warn($resp);
		}
		else {
			$logger->debug($resp);
		}
		unlink("$TMPDIR/$TEMPROOTFILE") unless ($DEBUG);
	}
	unlink("$TMPDIR/$OUTFILE") unless ($DEBUG);
}

sub compress_aspath($)
{
	my ($aspath) = $_[0];
	my (@compressed_aspath,$prevhop, $lastseen_hop);

	$prevhop = -1;
	$lastseen_hop = -1;

	foreach my $hop (split(/\s+/, $aspath))
	{
		if ($hop ne $prevhop)
		{
			if ($prevhop ne "0" ) 
			{
				if ($hop ne "0" )
				{
					push(@compressed_aspath,$hop);
					$lastseen_hop = $hop;
				}
			}
			else
			{
				if ($hop ne $lastseen_hop)
				{
					push(@compressed_aspath,0,$hop);
					$lastseen_hop = $hop;
				}
			}
			$prevhop = $hop;
		}
	}
	if ($prevhop eq "0")
	{
		push(@compressed_aspath,0);
	}
	return @compressed_aspath;
}

# --------------- Start Point -------------------------------------

sub main {

	my ($Dbh,$MYSQL_Host);

	parse_commandline();

	#first create filehandle to write one sequence os AShops per TB
	unless ( open( FP, "> $TMPDIR/$OUTFILE" ) ) {
		$logger->logdie("ERROR: could not open $TMPDIR/$OUTFILE ($!)");
	}

	# Try configured mysql servers
	do {
		$MYSQL_Host = shift(@MYSQL_Host);
		$logger->debug("Trying to connect to: ", sprintf($Dsn,$MYSQL_Host));
		$Dbh = DBI->connect(sprintf($Dsn,$MYSQL_Host), $DBUser, $DBPass);
	} while (!$Dbh && $#MYSQL_Host > -1);

	if( $Dbh ) {
		for ( my $timestamp = $t_start ; $timestamp < $t_end ; $timestamp += 86400 ) {
			my (@tbsrc,@tbdst);

			$logger->info("creating AS level nhop plot for ", strftime("%Y-%m-%d %H:%M:%S",gmtime($timestamp)));

			my $path = generatepath($timestamp);

			#make timestamp for ttconfig
			my $ttconfig_cmd = sprintf("$TTCONFIG -d '%s' -v TTLIST ACTIVE",strftime("%Y-%m-%d %H:%M:%S",gmtime($timestamp)));
			my @tblist  = map {chomp;$_ } qx($ttconfig_cmd);
			@tbsrc = map {"${_}$DOMAIN"} @tblist;

			mkdirtree($path);

			#after creating list with TB names must go into a loop to query TTM DB
			# Use a direct database connection for performance

			print FP "\@ $path/:$timestamp \n";
			$logger->debug("\@ $path:$timestamp");
			@tbdst = @tbsrc;
			foreach my $tbsrc (@tbsrc) {
				print FP &gettbid($tbsrc) . ": ";
				my $t1 = [gettimeofday];
				my $query = sprintf($ASQUERY_SQL,&gettbid($tbsrc), join(",", map { &gettbid($_) } grep { !/^${tbsrc}$/ } @tbdst ) ,$timestamp,$timestamp + 24 * 3600, $timestamp);
				my $sth = $Dbh->prepare($query) || $logger->logdie("Couldn't prepare \"$query\": ",$Dbh->errstr);
				$logger->debug("Prepared query: $query");
				if ($sth->execute > 0 ) 
				{
					my $curdst=-1;
					my @seen_aspaths;
					while (my($src, $dst, $aspathlen, $aspath) = $sth->fetchrow_array )
					{
						# Skip any earlier in time records when starttime is endtime
						# Otherwise include all aspaths seen
						if ( $t_start != $t_end || $curdst != $dst)
						{
							my @aspath = compress_aspath($aspath);
							if ( !grep {/^@aspath$/} @seen_aspaths) {
								my $nashop = scalar @aspath;
								push(@seen_aspaths,"@aspath");
								$curdst = $dst;
							
								$logger->debug("row: ($src, $dst, $aspathlen, @aspath)");
								if ( defined($nashop) && $nashop =~ /\d+/ ) {
									print FP $nashop . " ";
								}
							}
							else
							{
								$logger->debug("Skipping duplicate compressed aspath $aspath");
							}
						}
					}
					$logger->debug("Finished query and processing delta=", tv_interval($t1));
				}
				else
				{
					$logger->warn("\"$query\" failed: ", $Dbh->errstr) if ($Dbh->errstr);
				}
				$sth->finish if defined $sth;
				print FP "\n";
			}    #foreach my $tbsrc
		}

		close(FP);
		$Dbh->disconnect;
	}
	else {
		$logger->logdie("Couldn't connect to ", sprintf($Dsn,$MYSQL_Host), " as $DBUser: ", DBI->errstr);
	}
	#Ok, now there's a file listing the information we need for the histograms
	# must call astracestat to do the job

	# Disable plotting of longer than one day timespans

	&exec_astracestat(($t_end - $t_start <= 24 * 3600)  ? 0 : $noplots);
}    # sub main ends here !!!

sub HELP_MESSAGE() {
        &usage;
}

package DateTime::Format::ISO8601::Range;
use DateTime::Format::ISO8601;
use DateTime::Duration;


sub duration_spec2dt($)
{
	my $duration_spec = 'P(\d+)([DHMS])((\d+)([HMS]))?((\d+)([MS]))?((\d+)(S))?';
	my ($dt,$duration,@duration);
	if ((@duration) = $_[0] =~ /^$duration_spec$/) {
		my ($days,$hours,$minutes,$seconds) = (0,0,0,0);
		while (@duration) { 
			my ($duration_time,$duration_unit);
			($duration_time,$duration_unit,@duration) = @duration;
			pop(@duration) if $#duration > -1;
			if (defined($duration_unit)) {
				SWITCH : {
					$duration_unit eq 'D' && do {
						$days = $duration_time;
						last SWITCH;
						};
					$duration_unit eq 'H' && do {
						$hours = $duration_time;
						last SWITCH;
						};
					$duration_unit eq 'M' && do {
						$minutes = $duration_time;
						last SWITCH;
						};
					$duration_unit eq 'S' && do {
						$seconds = $duration_time;
						last SWITCH;
						};
				}
			}
		}
		$dt = DateTime::Duration->new( 
                                days    => $days,
                                hours   => $hours,
                                minutes => $minutes,
                                seconds => $seconds );
	}
	return $dt;
}

sub range2time($) {
	my ($range) = @_;
	my($start,$end,$duration,@result);
	my $iso8601 = DateTime::Format::ISO8601->new(base_datetime => DateTime->from_epoch( epoch => 0)) ;

	($start,$end) = split(/\//, $range);

	if (defined($start) && defined($end))
	{
		if ($duration = duration_spec2dt($start)) {
			$logger->debug("Start time is duration specifier");
			$end = $iso8601->parse_datetime($end);
			$start = $end - $duration;
		} elsif ($duration = duration_spec2dt($end)) {
			$logger->debug("End time is duration specifier");	
			$start = $iso8601->parse_datetime($start);
			$end = $start + $duration;
		}
		else
		{
			$start = $iso8601->parse_datetime($start);
			$end = $iso8601->parse_datetime($end);
		}
		@result = ($start->epoch,$end->epoch);
	}
	return @result;
}
