#!/usr/bin/perl -w
#
#
# Copyright (c) 2000, 2001  RIPE NCC
#
# $Id: plotxtimenhop.pl,v 1.12 2007/01/16 10:46:14 ruben Exp $
#
#
# All Rights Reserved
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose and without fee is hereby granted,
# provided that the above copyright notice appear in all copies and that
# both that copyright notice and this permission notice appear in
# supporting documentation, and that the name of the author not be
# used in advertising or publicity pertaining to distribution of the
# software without specific, written prior permission.
#
# THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
# ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS; IN NO EVENT SHALL
# AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY
# DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
# AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
#------------------------------------------------------------------------------
# Module Header
# Filename          :   plotipnhops.pl
# Purpose           :   wrapper for xtimeplot
# Authors           :   Maximo AlvesRIPE NCC
# Date              :   04-Dec-2002
# Description       :
# Language Version  :   perl5
# TODO              :
#------------------------------------------------------------------------
=pod

=head1 NAME

B<plotxtimenhop> -- Plot long term ip/as number hops distribution trends

=head1 SYNOPSIS:

plotxtimenhop [-h] | [-n] [-v] [-d] [-p <ISO8601 period> ] [-t ip|as]
    [-R <ROOT data dir> ] [-T <temporary data dir>] [-o <plots output dir>] 
	start

=head1 OPTIONS

=over

=item B<-h> Short help message

=item B<-n> dry run, with proper Log::Log4perl debug level will show what was to be done

=item B<-v> Be verbosive

=item B<-d> Enable debug output

=item B<-p I<ISO8601 period>> Specify ISO8601 period. An ISO8601 period consist out of an ISO8601 time seperated by a duration specifier (P([0-9]+[DHMS]){1,})

=item B<-t I<ip|as>> Either select IP or AS operation mode

=item B<-R I<ROOTdata_basedir>> specify an alternate base directory for ROOT data files.  YYYY/MM/DD is appended to it. It's used for input and output

=item B<-T I<temporary_dir>>  Where to store temporary files

=item B<-o I<plot_dir>>  Where to store generated plots. Plots are not synced to the webserver

=back

By using B<start> you will automaticly select the last 180 day period leading up till today

=head1 NOTES

Previous versions of the nhop tools only produced histograms for the number of
hops (ip and as wise) seen on exact the same time the day before. That approach
was wrong and the current tools use ISO8601 period specification to produce
more reliable results and take the whole day into account.

The command line switches of the previous versions are thus not backwards
compatible.


=head1 DEBUGGING

plotasnhops uses Log::Log4perl and registers itself under the domain C<ttpro.plotasnhops>
Signalling merge_data with C<USR2> will force it to reread ${HOME}/.log4perl.conf

=cut


use strict;
use Time::Local;    # support functions for time/date calculations
use POSIX;
use Getopt::Std;
use Time::HiRes qw(gettimeofday tv_interval);
use LockFile::Simple;
use Log::Log4perl;

# --------------- Variable declaration ------------------------------

$ENV{PATH} = "/ncc/ttpro/bin:/usr/local/bin:/usr/ucb:/bin:/usr/bin";

$Getopt::Std::STANDARD_HELP_VERSION = 1 if $Getopt::Std::VERSION >= 1.05;

Log::Log4perl::init_and_watch( "$ENV{'HOME'}/.log4perl.conf", 'USR2' );

our $VERSION = sprintf "%d.%03d", q$Revision: 1.12 $ =~ /(\d+)/g;

my %options;
my ( $t_start, $t_end );

my $TMPDIR          = "/var/tmp";

my $PLOTDIR  = "/ncc/ttpro/data/plots/nhop/";
my $ROOTDIR  = "/ncc/ttpro/data/root/";
my $FILELIST = "filelist.nhop";
my $TTPRO    = "/ncc/ttpro/";
my $xtimeplot_bin = "/ncc/ttpro/bin/xtimeplot";
my $measurementtime = 0;
my @pathlist        = ();
my $DEBUG           = 0;
my $VERBOSE         = 0;
my @timelist        = ();                                 #list with parameters
my $TTCONFIG        = "/ncc/ttpro/bin/ttconfig";
my $PLOTTITLE       = "xtime plots for";
my $type            = "";
my $dry_run      = 0;

# Configure logger
my $logger = Log::Log4perl->get_logger('ttpro.plotxtimenhops');

my $t0 = [gettimeofday];
&main;
$logger->info("finished delta=", tv_interval($t0));

# --------------------- Sub routines --------------------------------

sub usage(){
	print <<"EOT";
plotxtimenhop [-h] | [-n] [-v] [-d] [-p <ISO8601 period> ] [-t ip|as]
    [-R <ROOT data dir> ] [-T <temporary data dir>] [-o <plots output dir>] 
	start

EOT
}#sub usage ends here !!!

sub mkdirtree($)
{
	my $dirpath = $_[0];
	if (! -d $dirpath) {
		# create dirpath
		my $dirpath_base;
		foreach my $dirpart (split(/\//,$dirpath))
		{
			next unless defined($dirpart) && length($dirpart);
			$dirpath_base .= "/$dirpart";

			if(! -d $dirpath_base)
			{
				mkdir($dirpath_base) || $logger->logdie("$dirpath: mkdir $dirpath (at $dirpath_base) failed: $!");
			}
		}
	}
}

sub parse_commandline() {

	my $default_duration = strftime("P180D/%Y%m%d",(gmtime(time - 24 * 3600)));

	$logger->debug("process command line ...");

	getopts("hnp:t:vdo:T:R:",\%options);

	if ($options{'h'} ) {
		&usage;
		exit;
	}
	if ($options{'n'}) {
		$dry_run= "1";
	}
	if ($options{'d'}) {
		$DEBUG = "1";
	}
	if ($options{'v'}) {
		$VERBOSE = "1";
	}

	if ($options{'R'}) {
		$ROOTDIR = $options{'R'};
		if (-d $ROOTDIR)
		{
			$logger->debug("Setting ROOT directory to $ROOTDIR");
		}
		else
		{
			$logger->logdie("ROOT directory $ROOTDIR nonexistant!");
		}
	}
	if ($options{'T'}) {
		$TMPDIR = $options{'T'};
		if (-d $TMPDIR)
		{
			$logger->debug("Setting TMP directory to $TMPDIR");
		}
		else
		{
			$logger->logdie("TMP directory $TMPDIR nonexistant!");
		}
	}
	if ($options{'o'}) {
		$PLOTDIR = $options{'o'};
		if (-d $PLOTDIR)
		{
			$logger->debug("Setting PLOT directory to $PLOTDIR");
		}
		else
		{
			$logger->logdie("PLOT directory $PLOTDIR nonexistant!");
		}
	}
	if ($options{'t'}) {
		if (!(($type) = $options{'t'} =~ /^(ip|as)$/)) {
			$logger->logdie("use either ip or as for -t");
		}
	}
	if ($options{'p'} || grep { /^start$/} @ARGV ) {
		my $range = $options{'p'} ? $options{'p'} : $default_duration;
		if (($t_start,$t_end) = DateTime::Format::ISO8601::Range::range2time($range))
		{
			$logger->debug("Range is: ", scalar gmtime($t_start), " - " , scalar gmtime($t_end) );
		}
		else
		{
			$logger->logdie("Invalid ISO8601 time range \"$range\"");
		}
	}
	if (!$type) {
		$logger->logdie("please use ip or as for -t");
	}
}

sub generatepath($) {
	my ($ref)  = $_[0];
	return sprintf("%s/%s",$ROOTDIR, strftime("%Y/%m/%d",gmtime($ref)));
}    # sub generatepath() ends here !!!


sub gettbid() {
    my $ret = 0;
    my $tb = ( split( /\./, $_[0] ) )[0];
    if ( $tb =~ /osdorp/ ) {
        $ret = 2;
    }
    elsif ( $tb =~ /jordaan/ ) {
        $ret = 4;
    }
    else {
        $_ = $tb;
        s/[a-zA-Z]+//;
        $ret = $_;
    }
    return $ret;
}

# --------------- Start Point -------------------------------------

sub main {

    parse_commandline();
	my $t1;

    # create file listing all routestart.root files for the given period
    open( FP, "> $TMPDIR/$FILELIST") || $logger->logdie("can't open $TMPDIR/$FILELIST ($!)");
	for ( my $timestamp = $t_start ; $timestamp < $t_end ; $timestamp += 86400 ) {

		my $path = generatepath($timestamp);
		mkdirtree($path);

        if ( opendir( NEW, $path ) ) {
            while ( my $filename = readdir(NEW) ) {
                if ( $filename =~ /routestat(.*)$type(.*)root$/ ) {
                    print FP "$path/$filename\n";
                }
            }     #while
        }
    }
    close(FP);

	#make timestamp for ttconfig
	my $ttconfig_cmd = sprintf("$TTCONFIG -d '%s' -v TTLIST ONLINE",strftime("%Y-%m-%d %H:%M:%S",gmtime($t_end)));
    #get list of test boxes
	my @tblist  = map {chomp;$_ } qx($ttconfig_cmd);
    foreach my $tb (@tblist) {

        my ($tbid)     = $tb =~ /^tt(.*)$/;
        my $execstr = "$xtimeplot_bin -f $TMPDIR/$FILELIST -i $tbid  -a $t_start -e $t_end -t $type -o $PLOTDIR -b '$PLOTTITLE $tb'";

        if ($DEBUG) {
            $execstr = "$execstr -d";
        }
		if ($dry_run) {
			$logger->debug("Dry run: $execstr");
		} else {
			$t1 = [gettimeofday];
			$logger->debug("Running $execstr");
			my $resp = qx($execstr);

			$logger->debug("Finished xtimeplot delta=", tv_interval($t1));

			if ( $resp =~ /ERROR/ ) {
				$logger->fatal($resp);
			}
			elsif ( $resp =~ /WARNING/ ) {
				$logger->warn($resp);
			}
			else {
				$logger->debug($resp);
			}
		}
    }

    #now create summary histogram
    my $execstr  = "$xtimeplot_bin -f $TMPDIR/$FILELIST -i 0  -a $t_start -e $t_end -t $type -o $PLOTDIR -b '$PLOTTITLE summary'";
    if ($DEBUG) {
        $execstr .= " -d";
    }

	if ($dry_run) {
		$logger->debug("Dry run: $execstr");
	} else {
		$t1 = [gettimeofday];
		$logger->debug("Running $execstr");
		my $resp = qx($execstr);

		$logger->debug("Finished summary xtimeplot delta=", tv_interval($t1));

		if ( $resp =~ /ERROR/ ) {
			$logger->fatal($resp);
		}
		elsif ( $resp =~ /WARNING/ ) {
			$logger->warn($resp);
		}
		else {
			$logger->debug($resp);
		}
	}
	unlink("$TMPDIR/$FILELIST") unless ($DEBUG);
}    # sub main ends here !!!

sub HELP_MESSAGE() {
        &usage;
}

package DateTime::Format::ISO8601::Range;
use DateTime::Format::ISO8601;
use DateTime::Duration;


sub duration_spec2dt($)
{
	my $duration_spec = 'P(\d+)([DHMS])((\d+)([HMS]))?((\d+)([MS]))?((\d+)(S))?';
	my ($dt,$duration,@duration);
	if ((@duration) = $_[0] =~ /^$duration_spec$/) {
		my ($days,$hours,$minutes,$seconds) = (0,0,0,0);
		while (@duration) { 
			my ($duration_time,$duration_unit);
			($duration_time,$duration_unit,@duration) = @duration;
			pop(@duration) if $#duration > -1;
			if (defined($duration_unit)) {
				SWITCH : {
					$duration_unit eq 'D' && do {
						$days = $duration_time;
						last SWITCH;
						};
					$duration_unit eq 'H' && do {
						$hours = $duration_time;
						last SWITCH;
						};
					$duration_unit eq 'M' && do {
						$minutes = $duration_time;
						last SWITCH;
						};
					$duration_unit eq 'S' && do {
						$seconds = $duration_time;
						last SWITCH;
						};
				}
			}
		}
		$dt = DateTime::Duration->new( 
                                days    => $days,
                                hours   => $hours,
                                minutes => $minutes,
                                seconds => $seconds );
	}
	return $dt;
}

sub range2time($) {
	my ($range) = @_;
	my($start,$end,$duration,@result);
	my $iso8601 = DateTime::Format::ISO8601->new(base_datetime => DateTime->from_epoch( epoch => 0)) ;

	($start,$end) = split(/\//, $range);

	if (defined($start) && defined($end))
	{
		if ($duration = duration_spec2dt($start)) {
			$logger->debug("Start time is duration specifier");
			$end = $iso8601->parse_datetime($end);
			$start = $end - $duration;
		} elsif ($duration = duration_spec2dt($end)) {
			$logger->debug("End time is duration specifier");	
			$start = $iso8601->parse_datetime($start);
			$end = $start + $duration;
		}
		else
		{
			$start = $iso8601->parse_datetime($start);
			$end = $iso8601->parse_datetime($end);
		}
		@result = ($start->epoch,$end->epoch);
	}
	return @result;
}
