#!/usr/local/bin/perl5_latest -w

# Copyright (c) 1998                            RIPE NCC
#
# All Rights Reserved
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose and without fee is hereby granted,
# provided that the above copyright notice appear in all copies and that
# both that copyright notice and this permission notice appear in
# supporting documentation, and that the name of the author not be
# used in advertising or publicity pertaining to distribution of the
# software without specific, written prior permission.
#
# THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
# ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS; IN NO EVENT SHALL
# AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY
# DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
# AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

#------------------------------------------------------------------------------
# Module Header
# Filename          : cgiquery.pl 
# Purpose           : Front-end to CGI query from TT DB
# Author            : Manuel Valente <manuel@ripe.net>
# Date              : 20000525
# Description       : The program will open a socket on ginkgo, send a query 
#                     to the DB, and format the results nicely.
# Language Version  : Perl 5.0005
# OSs Tested        : BSDI 3.1
# Command Line      : N/A
# Input Files       : Web form, TT Database
# Output Files      : Dynamic Web page
# External Programs : 
#------------------------------------------------------------------------------

use strict;
use CGI;
use IO::Socket;
use Time::Local;

$! = 1;

my $progname            = $ENV{'SCRIPT_NAME'};
$progname               =~ s|/+|/|g;            #remove double slashes

my $testboxes = '/ncc/WWW2/www_site/pub/test-traffic/Plots/LIST_OF_TESTBOXES';

# Main function
{
	my (%list,@BoxList);
		
	open (F, $testboxes) or die ("Cannot open $testboxes: $!");
	
	while (<F>)
	{
		/^\d+\s+tt(\d+)\.ripe\.net\s+\S+\s*$/ or next;
		$list{sprintf ("tt%02d.ripe.net",$1)} = 1;
	};	
	close (F);
	
	@BoxList = sort keys %list;
	
	my $q = new CGI;
	
	# Print HTML header for browser
	print "Content-type: text/html\n\n";
	
	my ($header,$footer) = get_headers();	
	
	my $title = 'Query Routing Vector Database';

	# Check if it's the first time the script is called
	if (defined ($q->param('SRC')))
	{
		$q->param('time_start') or $q->param(-name=>'time_start',-value=>'00:00:00');
		$q->param('time_end')   or $q->param(-name=>'time_end',  -value=>'00:00:00');	
	
		$title .=" \[".$q->param('SRC')." - ".$q->param('DST')."\]";		
	};
	
	$header =~ s/(\$\w+)/$1/gee;
	$footer =~ s/(\$\w+)/$1/gee;

	print $header;
	
	print get_form($q,@BoxList);
	
	if (defined ($q->param('SRC')))
	{
		get_result($q,@BoxList);
	};	
		
	print $footer;
	
};

# Purpose           : Send query, print results nicely formatted
# Comments          : Prints data to STDOUT
# Input             : List of parameters to pass the DB:
#                     SRC box, DST box, time start, time end
# Output            : NULL
sub get_result
{
	my ($q,@boxlist) = @_;

	my (@line,$c,$ll);
	
	my $trimmed_result = 0;

	my $socket = IO::Socket::INET->new(PeerAddr => 'ginkgo.ripe.net',
		PeerPort => '4200',
		Proto    => "tcp",
		Type     => SOCK_STREAM)
	or die "Couldn't connect to ginkgo : $@\n";

	my $src = $q->param('SRC');
	my $dst = $q->param('DST');
	
	$src eq $dst and do
	{
		print ("<P>Error: Source and Destination IP are identical - $src.</P>");
		return;
	};
	
	my $tstart = get_timestamp ($q->param('date_start'),$q->param('time_start'));
	
	$tstart == -1 and return();
	
	my $tend   = get_timestamp ($q->param('date_end'),$q->param('time_end'));
			
	$tend == -1 and return();
	
	# Build the command to run
	my $command = "$src $dst $tstart $tend\n";
	
	#print "COMMAND:$command\n";

	print $socket $command;

	my ($line,$result);
	
	while ($line = <$socket>)
	{
		$line eq ".\n" and last;
		$line =~ /^\.(\d+)$/ and do
		{
			$trimmed_result = $1;
			last;
		};
		$result .= $line;
	};
	
	close($socket);
		
	# For each line in the result
	foreach (split "\n",$result)
	{
		@line = split;

		print "<P><TABLE BORDER=1 WIDTH=800><TR><TH>Dates</TH><TH>Route-ID</TH><TH>Records</TH><TH>Element</TH><TH>IP Address</TH><TH>Hostname</TH></TR><TR><TD ALIGN=RIGHT>From ";
	
		print scalar(gmtime($line[2]));
		print "<BR>To ";
		print scalar(gmtime($line[3]));
		print "</TD>";
	
		print "<TD ALIGN=RIGHT>$line[4]</TD>";
		print "<TD ALIGN=RIGHT>$line[6]</TD>";
	
		$c = 0;
	
		for my $k (7..$#line)
		{
			$c++;
			$k == 7 or print"<TR><TD COLSPAN=3></TD>";
			print "<TD ALIGN=RIGHT>$c</TD>";
			print ("<TD ALIGN=RIGHT>".$line[$k]."</TD>");
		
			$ll = gethostbyaddr(inet_aton($line[$k]),AF_INET) || '<EM>Unknown address</EM>';
			
			print ("<TD ALIGN=RIGHT>$ll</TD>");
			print "</TR>";
		};
		print "</TABLE><BR></P>";
	};
	
	$trimmed_result and do
	{
		print "<P>Warning: Too many results were returned. The query only shows the first $trimmed_result results.</P>";
	};
	
};

# Purpose           : Get a timestamp[ from a date and time
# Comments          : 
# Input             : date and time
# Output            : timestamp (-1 if error)
sub get_timestamp
{
	my ($date,$time) = @_;
	
	my ($stamp);
	
	my ($realdate,$realtime) = ($date,$time);
	
	$date =~ s/\D//g;	
	$time =~ s/\D//g;
	
	$date =~ /^(\d{4})(\d{2})(\d{2})$/ or do
	{
		print ("<P>Invalid date format $realdate</P>");
		return (-1);
	};
	
	my ($y,$m,$d) = ($1,$2,$3);
	
	($y>=1 and $m>=1 and $d>=1 and $m<=12 and $d<=31) or do
	{
		print ("<P>Invalid date format $realdate</P>");
		return (-1);
	};	
	
	$time .= '0'x(6 - length($time));
	
	$time =~ /^(\d{2})(\d{2})(\d{2})$/ or do
	{
		print  ("<P>Invalid time format $realtime</P>");
		return (-1);
	};
	
	my ($h,$mn,$s) = ($1,$2,$3);
	
	($h>=0 and $h<=23 and $mn>=0 and $mn <=59 and $s>=0 and $s<=59) or do
	{
		print  ("<P>Invalid time format $realtime</P>");
		return (-1);
	};
		
	my $t = timegm($s,$mn,$h,$d,$m-1,$y);
	
	$t == -1 and print  ("<P>Invalid date format $realdate - $realtime</P>");
	
	return ($t);
};


# Purpose           : Print the FORM for entering data
# Comments          : 
# Input             : CGI object, List of boxes
# Output            : content (html data)
sub get_form
{
	my ($q,@BoxList) = @_;

	my $content = '';
	
	# Set the variables to empty string if they're not defined.
	my $src = $q->param('SRC')||'';
	my $dst = $q->param('DST')||'';
	my $d_s = $q->param('date_start')||'';
	my $t_s = $q->param('time_start')||'';
	my $d_e = $q->param('date_end')||'';
	my $t_e = $q->param('time_end')||'';
	
	# FORM
	$content = "<P><HR></P>";
	
	$content.="<P>Date Format: yyyymmdd<br>Time Format: hh:mm:ss</P>";
	
	$content.="<P><TABLE><TR><TD><FORM ACTION =\"$progname\" METHOD=POST>
	SRC: </TD><TD><SELECT NAME=SRC>";
	
	# Print all testboxes for SRC combo
	foreach (@BoxList)
	{
		# If a test box was SELECTED, select it again
		if ($_ eq $src)
		{
			$content .= "<OPTION SELECTED>$_</OPTION>";		
		}
		else
		{
			$content .= "<OPTION>$_</OPTION>";
		};
	};
	
	$content .= "</SELECT></TD>";	
	
	# Print SRC Input fields
	$content .= "<TD>Start Date</TD><TD><INPUT TYPE=TEXT NAME=date_start VALUE=$d_s></TD><TD> Start Time</TD><TD><INPUT TYPE=TEXT NAME=time_start VALUE=$t_s></TD></TR>";

	$content .= "<TR><TD>DST: </TD><TD><SELECT NAME=DST>";
	
	# Print all testboxes for DST combo
	foreach (@BoxList)
	{
		if ($_ eq $dst)
		{
			$content .= "<OPTION SELECTED>$_</OPTION>";	
		}
		else
		{
			$content .= "<OPTION>$_</OPTION>";
		};
	};
	
	$content .= "</SELECT></TD>";

	# Print DST Input fields	
	$content .= "<TD>End Date</TD><TD><INPUT TYPE=TEXT NAME=date_end VALUE=$d_e></TD><TD> End Time</TD><TD><INPUT TYPE=TEXT NAME=time_end VALUE=$t_e></TD></TR>";

	# Print submit button
	$content .="<TR><TD COLSPAN=6 ALIGN=RIGHT><INPUT TYPE=SUBMIT NAME=SUBMIT VALUE=\"Submit\"></TD></TR>";
	
	$content .= "</FORM></TABLE></P><P><HR></P>";
		
	return $content;
};

sub get_headers
{

	my $header = '
<HTML>
<HEAD>

<TITLE>$title</TITLE>

<LINK REL="STYLESHEET" HREF="/templates/nccstyle1.css" TYPE="text/css"> 
<STYLE TYPE="text/css">
<!--
  A, ADDRESS, B, BODY, P, CENTER, TH, DD, DL, DT, H1, H2, H3, LI, UL, H4, H5, OL, TABLE ,TD
  {
  font-family: Arial, sans-serif, Helvetica;
  }
-->
</STYLE>
</HEAD>
<BODY BGCOLOR="#ffffff" TEXT="#003399" LINK="#990099" VLINK="#003399" ALINK="CC0000">
<A name="top of page"></a>

<TABLE WIDTH="645">
<TR>
<TD><A HREF="/ripencc/"><IMG SRC="/ripencc/pics/ncclogo-small.gif"
ALT="ncc logo" WIDTH="130" HEIGHT="84" BORDER="0"></A></TD>
<TD><IMG SRC="/ripencc/pics/topbarncc.gif"
ALT="" WIDTH="516" HEIGHT="26" BORDER="0"
USEMAP="#map2"><MAP NAME="map2">
<AREA SHAPE="RECT" COORDS="140,-1,214,30" HREF="/index.html" ALT="home"> 
<AREA SHAPE="RECT" COORDS="215,0,291,25" HREF="/ripencc/whatsnew.html" ALT="what\'s new">

<AREA SHAPE="RECT" COORDS="292,1,355,28"
 HREF="/ripencc/pub-services/db/whois.html" ALT="Whoisdb"> 
<AREA SHAPE="RECT" COORDS="355,0,403,27" HREF="/ripencc/search.html" ALT="Search "> 
<AREA SHAPE="RECT" COORDS="404,0,466,27" HREF="/ripencc/sitemap.html" ALT="Site Map"> 
<AREA SHAPE="RECT" COORDS="465,0,515,24" HREF="/ripencc/faq/index.html"
 ALT="Frequently Asked Questions"> 
</MAP>
</TD>
</TR>
</TABLE></a>
<P><H1>$title</H1></P>

';

	my $footer = '
<TABLE WIDTH="645">
<TR>
<TD COLSPAN="6"><IMG SRC="/ripencc/pics/ttm-bar.gif" ALT="bottom bar"
WIDTH="645" HEIGHT="20" BORDER="0" USEMAP="#map1"><MAP NAME="map1">
<AREA SHAPE="RECT" COORDS="0,0,133,19" HREF="mailto:tt-ops@ripe.net"
 ALT="Mail the ttm operators"> 
<AREA SHAPE="RECT" COORDS="524,0,649,24" HREF="mailto:ncc@ripe.net"
 ALT="Mail the RIPE NCC"> 
</MAP>
</TD>
</TR>
<TR>
<TD COLSPAN="6"><CENTER>
<BR>
<A HREF="#top of page"><FONT SIZE="-1">back to the top</FONT></A><BR><br>
</CENTER>
</TD>
</TR>
</TABLE>
</BODY>
</HTML>';

	return ($header,$footer);
};


