#!/usr/local/bin/perl5 -w

# Copyright (c) 2000,2002                            RIPE NCC
#
# All Rights Reserved
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose and without fee is hereby granted,
# provided that the above copyright notice appear in all copies and that
# both that copyright notice and this permission notice appear in
# supporting documentation, and that the name of the author not be
# used in advertising or publicity pertaining to distribution of the
# software without specific, written prior permission.
#
# THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
# ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS; IN NO EVENT SHALL
# AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY
# DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
# AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

#------------------------------------------------------------------------------
# Module Header
# Filename          : cgiquery.pl 
# Purpose           : Front-end to CGI query from TT DB
# Author            : Manuel Valente <manuel@ripe.net>
# Date              : 20000525
# Revised	    : 20020428, fill in defaults for start/end dates
# Revised	    : 20020604, add links to plots-on-demand
# Description       : The program will open a socket on ginkgo, send a query 
#                     to the DB, and format the results nicely.
# Language Version  : Perl 5.0005
# OSs Tested        : BSDI 3.1
# Command Line      : N/A
# Input Files       : Web form, TT Database
# Output Files      : Dynamic Web page
# External Programs : 
#
# $Id: gridquery-kauri.pl,v 1.2 2002/10/18 22:32:33 wilhelm Exp $
#------------------------------------------------------------------------------

use strict;
use CGI;
use IO::Socket;
use Time::Local;

$! = 1;

my $progname            = $ENV{'SCRIPT_NAME'};
$progname               =~ s|/+|/|g;            #remove double slashes

my $pod_url = "/cgi-bin/gttm/pod";
my $domain = ".ripe.net";
my $testboxes = '/var/www/pub/ripencc/mem-services/ttm/Plots/LIST_OF_GRIDBOXES';
my @GPSINFO = `cat /ncc/www/publish/pub/ripencc/mem-services/ttm/Plots/GPSINFO.grid`;

my ($status, $f, %TTvals, %TTname, %host);

# Main function
{
	my (%list,@BoxList);
		
	open (F, $testboxes) or die ("Cannot open $testboxes: $!");
	
	while (<F>)
	{
		/^\d+\s+tt(\d+)\.ripe\.net\s+\S+\s*$/ or next;
		$list{sprintf ("tt%02d.ripe.net",$1)} = 1;
	};	
	close (F);
	
	chomp @GPSINFO;
	foreach $f (@GPSINFO) {

    		# is this a status line?
		if (!($f =~ /^status:\t(\w+)$/)) {
			# is this a testbox line?
       			my ($name, $lat, $long, $place) = split ('\t', $f);
       			push @BoxList, $name;
       			$TTvals{$name} = "$name: $place";
			if (! ($name =~ /\./) ) {
				# unqualified hostname, add default domain
				$TTname{$name} =  $name . $domain ;
			}
			else {
				$TTname{$name} = $name;
			}
    		}
	}

	@BoxList = sort keys %TTvals;

	my $q = new CGI;
	
	# Print HTML header for browser
	print "Content-type: text/html\n\n";
	
	my ($header,$footer) = get_headers();	
	
	my $subtitle = 'Query Traceroute Database';

	# Check if it's the first time the script is called
	if (defined ($q->param('SRC')))
	{
		$q->param('time_start') or $q->param(-name=>'time_start',-value=>'00:00');
		$q->param('time_end')   or $q->param(-name=>'time_end',  -value=>'00:00');	
	
		$subtitle .=" \[".$q->param('SRC')." - ".$q->param('DST')."\]";		
	};
	
	$header =~ s/(\$\w+)/$1/gee;
	$footer =~ s/(\$\w+)/$1/gee;

	print $header;
	
	print get_form($q,@BoxList);
	
	if (defined ($q->param('SRC')))
	{
		get_result($q,@BoxList);
	};	
		
	print $footer;
	
};

# Purpose           : Send query, print results nicely formatted
# Comments          : Prints data to STDOUT
# Input             : List of parameters to pass the DB:
#                     SRC box, DST box, time start, time end
# Output            : NULL
sub get_result
{
	my ($q,@boxlist) = @_;

	my (@line,$c,$ll);
	
	my $trimmed_result = 0;

	my $socket = IO::Socket::INET->new(PeerAddr => 'kauri.ripe.net',
		PeerPort => '4200',
		Proto    => "tcp",
		Type     => SOCK_STREAM)
	or die "Couldn't connect to ginkgo : $@\n";

	my $src = $q->param('SRC');
	my $dst = $q->param('DST');
	
	$src eq $dst and do
	{
		print ("<P>Error: Source and Destination IP are identical - $src.</P>");
		return;
	};
	
	my $tstart = get_timestamp ($q->param('date_start'),$q->param('time_start'));
	
	$tstart == -1 and return();
	
	my $tend   = get_timestamp ($q->param('date_end'),$q->param('time_end'));
			
	$tend == -1 and return();
	
	my $resolve = $q->param('resolve');

	# Build the command to run
	my $command = "$TTname{$src} $TTname{$dst} $tstart $tend\n";
	
	#print "COMMAND:$command\n";

	print $socket $command;

	my ($line,$result);
	
	while ($line = <$socket>)
	{
		$line eq ".\n" and last;
		$line =~ /^\.(\d+)$/ and do
		{
			$trimmed_result = $1;
			last;
		};
		$result .= $line;
	};
	
	close($socket);
		
	# For each line in the result
	foreach (split "\n",$result)
	{
		@line = split;

		print "<P><TABLE BORDER=1 WIDTH=800><TR><TH>Dates</TH><TH>Route-ID</TH><TH>Records</TH><TH>Element</TH><TH>IP Address</TH>";

		my $nhops = $line[5];

		if ($resolve ne "no") {
			print "<TH>Hostname</TH>";
		}

		if ($line[7+$nhops+1]) {
			# non-zero hops in AS path
			print "<TH>AS&nbsp;Num(s)</TH>";
		}

		print "</TR>\n<TR><TD ALIGN=RIGHT>From ";
	
		print scalar(gmtime($line[2]));
		print "<BR>To ";
		print scalar(gmtime($line[3]));
		print "</TD>";
	
		print "<TD ALIGN=RIGHT>$line[4]</TD>";
		print "<TD ALIGN=RIGHT>$line[6]</TD>";
	
		$c = 0;
	
		for my $k (7..$nhops+6)
		{
			$c++;
			$k == 7 or print"<TR><TD COLSPAN=3></TD>";
			print "<TD ALIGN=RIGHT>$c</TD>";
#			print ("<TD ALIGN=RIGHT>".$line[$k]."</TD>");
		
			if (($line[$k] ne "255.255.255.255") &&
			    ($line[$k] ne "0.0.0.0")) {
				print ("<TD ALIGN=RIGHT>",
				       "<A HREF=/perl/whois?-ar\%20$line[$k]>",
				       $line[$k], "</A></TD>");
				if ($resolve ne "no") { 
					if (! $host{$line[$k]}) {
						eval {
                                        	        local $SIG{ALRM} = sub {die "alarm\n"}; 
                                        	        alarm 1;  # timeout after 1sec
                                        	        $ll = gethostbyaddr(inet_aton($line[$k])
							      ,AF_INET) || '<EM>Unknown address</EM>';
                                        	        alarm 0;
							die "OK";
                                        	}; 
                                        	if ($@ eq "alarm\n") {
                                        	        $ll = '<EM>Unknown address</EM>';
                                        	}
						$host{$line[$k]} = $ll;
					}

					print ("<TD ALIGN=RIGHT>$host{$line[$k]}</TD>");
				}

				# Add AS number if non-zero
				if ($line[$nhops+2+$k]) {
					print ("<TD ALIGN=RIGHT>$line[$nhops+2+$k]</TD>");
				}
			}
			else {
				print ("<TD ALIGN=RIGHT>* * *</TD>");
				if ($resolve ne "no") {
					print ("<TD>&nbsp;</TD>\n");
				}
			}
			print "</TR>";
		};
		print "</TABLE><BR></P>";
	};
	
	$trimmed_result and do
	{
		print "<P>Warning: Too many results were returned. The query only shows the first $trimmed_result results.</P>";
	};
	
};

# Purpose           : Get a timestamp[ from a date and time
# Comments          : 
# Input             : date and time
# Output            : timestamp (-1 if error)
sub get_timestamp
{
	my ($date,$time) = @_;
	
	my ($stamp);
	
	my ($realdate,$realtime) = ($date,$time);
	
	$date =~ s/\D//g;	
	$time =~ s/\D//g;
	
	$date =~ /^(\d{4})(\d{2})(\d{2})$/ or do
	{
		print ("<P>Invalid date format $realdate</P>");
		return (-1);
	};
	
	my ($y,$m,$d) = ($1,$2,$3);
	
	($y>=1 and $m>=1 and $d>=1 and $m<=12 and $d<=31) or do
	{
		print ("<P>Invalid date format $realdate</P>");
		return (-1);
	};	
	
	$time .= '0'x(6 - length($time));
	
	$time =~ /^(\d{2})(\d{2})(\d{2})$/ or do
	{
		print  ("<P>Invalid time format $realtime</P>");
		return (-1);
	};
	
	my ($h,$mn,$s) = ($1,$2,$3);
	
	($h>=0 and $h<=23 and $mn>=0 and $mn <=59 and $s>=0 and $s<=59) or do
	{
		print  ("<P>Invalid time format $realtime</P>");
		return (-1);
	};
		
	my $t = timegm($s,$mn,$h,$d,$m-1,$y);
	
	$t == -1 and print  ("<P>Invalid date format $realdate - $realtime</P>");
	
	return ($t);
};


# Purpose           : Print the FORM for entering data
# Comments          : 
# Input             : CGI object, List of boxes
# Output            : content (html data)
sub get_form
{
	my ($q,@BoxList) = @_;

	my $content = '';
	
        # Set defaults
        
        my ($hour, $day, $mon, $year);
        my ($startday, $endday);

        ($a, $a, $hour, $day, $mon, $year, $a, $a, $a) = gmtime(time());

        # yesterday's traceroute date usually processed by 0500 GMT

        if ($hour > 5) {
                $year += 1900;
                $endday = $year .  substr("0".++$mon, length($mon )-1, 2) .
                                substr("0".$day,length($day)-1, 2);

                ($a, $a, $hour, $day, $mon, $year, $a, $a, $a) =
                        gmtime(time()- 86400);
                $year += 1900;
                $startday = $year .  substr("0".++$mon, length($mon )-1, 2) .
                                substr("0".$day,length($day)-1, 2);
        }
        else {
                # last data until the end of day before yesterday
                ($a, $a, $hour, $day, $mon, $year, $a, $a, $a) =
                        gmtime(time()- 86400);
                $year += 1900;
                $endday = $year .  substr("0".++$mon, length($mon )-1, 2) .
                                substr("0".$day,length($day)-1, 2);

                ($a, $a, $hour, $day, $mon, $year, $a, $a, $a) =
                        gmtime(time()- 2 * 86400);
                $year += 1900;
                $startday = $year .  substr("0".++$mon, length($mon )-1, 2) .
                                substr("0".$day,length($day)-1, 2);
        }


	# Set the variables to empty string if they're not defined.
	my $src = $q->param('SRC')||'';
	my $dst = $q->param('DST')||'';
	my $d_s = $q->param('date_start')||$startday;
	my $t_s = $q->param('time_start')||"00:00";
	my $d_e = $q->param('date_end')|| $endday;
	my $t_e = $q->param('time_end')|| "00:00";
	my $resolve = $q->param('resolve')|| "yes";
	
	# FORM
	$content = "<P><HR></P>";
	
	$content.="<P>Date Format: yyyymmdd<br>Time Format: hh:mm:ss</P>";
	
	$content.="<P><TABLE CELLSPACING=5><TR><TD><FORM ACTION =\"$progname\" METHOD=GET>
	SRC: </TD><TD COLSPAN=4><SELECT NAME=SRC>";
	
	# Print all testboxes for SRC combo
	foreach (@BoxList)
	{
		# If a test box was SELECTED, select it again
		if ($_ eq $src)
		{
			$content .= "<OPTION SELECTED VALUE=$_>$TTvals{$_}</OPTION>";		
		}
		else
		{
			$content .= "<OPTION VALUE=$_>$TTvals{$_}</OPTION>";
		};
	};
	
	$content .= "</SELECT></TD></TR>";	
	
	$content .= "<TR><TD>DST: </TD><TD COLSPAN=4><SELECT NAME=DST>";
	
	# Print all testboxes for DST combo
	foreach (@BoxList)
	{
		if ($_ eq $dst)
		{
			$content .= "<OPTION SELECTED VALUE=$_>$TTvals{$_}</OPTION>";		
		}
		else
		{
			$content .= "<OPTION VALUE=$_>$TTvals{$_}</OPTION>";
		};
	};
	
	$content .= "</SELECT></TD></TR><TR>";

	# Print SRC Input fields
	$content .= "<TD>Start Date</TD><TD>&nbsp;&nbsp;&nbsp;<INPUT TYPE=TEXT NAME=date_start SIZE=10 VALUE=$d_s></TD><TD> Start Time</TD><TD ALIGN=LEFT><INPUT TYPE=TEXT SIZE=5 NAME=time_start VALUE=$t_s></TD>";

 
	$content .= "<TD ALIGN=RIGHT>Resolve IPs &nbsp;&nbsp;<SELECT NAME=resolve>";
        foreach ("yes", "no")
        {
                if ($_ eq $resolve)
                {
                        $content .= "<OPTION SELECTED>$_</OPTION>";
                }
                else
                {
                        $content .= "<OPTION>$_</OPTION>";
                };
        }

	$content .= "</SELECT></TD></TR>";


	# Print DST Input fields	
	$content .= "<TD>End Date</TD><TD>&nbsp;&nbsp;&nbsp;<INPUT TYPE=TEXT NAME=date_end SIZE=10 VALUE=$d_e></TD><TD> End Time</TD><TD ALIGN=LEFT><INPUT TYPE=TEXT SIZE=5 NAME=time_end VALUE=$t_e></TD>";

	# Print submit button
	$content .="<TD ALIGN=RIGHT><INPUT TYPE=SUBMIT NAME=SUBMIT VALUE=\"Submit\"></TD></TR>";
	
	$content .= "</FORM></TABLE></P><P><HR></P>";
		
        $content .= "The DataGrid part of RIPE NCC's Test Traffic Measurements";
        $content .= " is coordinated by NIKHEF/UvA. <br>Contact person: ";
	$content .= '<A HREF=mailto:antony@nikhef.nl> Antony Antony </A><P>';

	if ($src ne $dst) {
		$content .= "<A HREF=\"${pod_url}?SRC=$src&DST=$dst&date_start=$d_s&time_start=$t_s&date_end=$d_e&time_end=$t_e&plottype=delay\">";
		$content .= "delay</A> or ";
		$content .= "<A HREF=\"${pod_url}?SRC=$src&DST=$dst&date_start=$d_s&time_start=$t_s&date_end=$d_e&time_end=$t_e&plottype=jitter\">";
		$content .= "jitter</A> plots for this time period</P><HR>";
	}

	return $content;
};

sub get_headers
{

	my $header = '
<HTML>
<HEAD>

<TITLE>$subtitle</TITLE>

<LINK REL="STYLESHEET" HREF="/templates/nccstyle1.css" TYPE="text/css"> 
<STYLE TYPE="text/css">
<!--
  A, ADDRESS, B, BODY, P, CENTER, TH, DD, DL, DT, H1, H2, H3, LI, UL, H4, H5, OL, TABLE ,TD
  {
  font-family: Arial, sans-serif, Helvetica;
  }
-->
</STYLE>
</HEAD>
<BODY BGCOLOR="#ffffff" TEXT="#003399" LINK="#990099" VLINK="#003399" ALINK="CC0000">
<A name="top of page"></a>

<TABLE WIDTH="645">
<TR>
<TD><A HREF="/ripencc/"><IMG SRC="/ripencc/pics/ncclogo-small.gif"
ALT="ncc logo" WIDTH="130" HEIGHT="84" BORDER="0"></A></TD>
<TD>
<A HREF="http://eu-datagrid.web.cern.ch/"><IMG SRC="http://eu-datagrid.web.cern.ch/eu-datagrid/images/DataGridLogo.gif" ALT="data grid logo" BORDER="0"></A></TD>
</TD>
<TD>
<A HREF="http://www.nikhef.nl/"><IMG SRC="http://www.nikhef.nl/logo.gif" ALT="NIKHEF logo" BORDER="0"></A></TD>
</TD>
</TR>
</TABLE></a>

<H1><A HREF=/ttm/>Test Traffic Measurements</A></H1>
<H2>$subtitle</H2>

';

	my $footer = '
<TABLE WIDTH="645">
<TR>
<TD COLSPAN="6"><IMG SRC="/ripencc/pics/ttm-bar.gif" ALT="bottom bar"
WIDTH="645" HEIGHT="20" BORDER="0" USEMAP="#map1"><MAP NAME="map1">
<AREA SHAPE="RECT" COORDS="0,0,133,19" HREF="mailto:tt-ops@ripe.net"
 ALT="Mail the ttm operators"> 
<AREA SHAPE="RECT" COORDS="524,0,649,24" HREF="mailto:ncc@ripe.net"
 ALT="Mail the RIPE NCC"> 
</MAP>
</TD>
</TR>
<TR>
<TD COLSPAN="6"><CENTER>
<BR>
<A HREF="#top of page"><FONT SIZE="-1">back to the top</FONT></A><BR><br>
</CENTER>
</TD>
</TR>
</TABLE>
</BODY>
</HTML>';

	return ($header,$footer);
};


