/*
 * Copyright (c) 2000                      RIPE NCC
 *
 *
 * All Rights Reserved
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of the author not be
 * used in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.
 *
 * THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS; IN NO EVENT SHALL
 * AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY
 * DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * -------------------------------------------------------------------------------
 * Module Header
 * Filename          : updatedb.c
 * Purpose           : Update The Test Traffic DBs with RVEC(6) data
 * Author            : Manuel Valente <manuel@ripe.net>
 * Date              : 20000719
 * Revised		  : 20020814 Rene Wilhelm: added AS path information
 * Revised           : 20021210 Florian Frotzler: added IPv6 compatibility
 * Description       :
 * Language Version  : gcc version 2.95.2
 * OSs Tested        : Solaris 2.6, Solaris 8 and (partly) Linux
 * Command Line      : updatedb [-4] [-6] path
 * Input Files       : TTM RVEC and/or RVEC6 (traceroute measurements) files
 * Output Files      :
 * External Programs : mysqld
 * Problems          : Should make sure that it scales up to 100 boxes
 * To Do             :
 * Comments          : requires glib library (hash tables, sorting etc)
 * -------------------------------------------------------------------------------
*/

static char const rcsid[] = "$Id: updatedb.c,v 1.20 2009/12/16 16:01:34 eromijn Exp $";

#define PROGNAME "updatedb"

#include "as.h"
#include "common.h"
#include "read_line.h"
#include "digest.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <arpa/inet.h>

/* IPv6 address in binary form */
#define INET6_BINSTRLEN 16

#define REPORT_FORMAT	"file=%s vectors=%u delta=%u vectors_sec=%f usertimes=%lu systimes=%lu\n"

struct cachetables
{
	GHashTable     *ranges,
	               *routes,
	               *asnums,
	               *aspaths;
};

struct jobtracker_set_fdset_param
{
	fd_set         *read_fds;
	int             max_fd;
	int 			vectors_read;
};

struct sliceinfo
{
	time_t          t_start;
	int             slice_start,
	                slice_end;
	FILE           *read_fd;
};

int             get_pathid(char *path, int npoints, GHashTable * hash, const char *table, const char *field, int src, MYSQL * mysql);
int             get_pathid_ip6(char *path, int npoints, GHashTable * hash, const char *table, const char *field, int src, MYSQL * mysql);
int             sort_records(gconstpointer a, gconstpointer b);
void            process_records(gpointer data, gpointer user_data);
void            remove_records(gpointer data, gpointer user_data);
int             process_file(MYSQL * mysql, char *filename, struct cachetables *tables, int ip6flag);
int             process_slice(const char *mysql_host, int slice_start, int slice_end, char **files, struct cachetables *tables, int ip6flag);
int             spawn_slices(const char *mysql_host, unsigned int n, char **files, unsigned int jobs, struct cachetables *tables, int ip6flag);
int             waitforjob(int signal);
void            jobtracker_set_fdset(gpointer key, gpointer val, gpointer userdata);
void            jobtracker_read_fdset(gpointer key, gpointer val, gpointer userdata);
int 			last_path_id(MYSQL *mysql,const char *table);

static unsigned int      debug = 0,
	verbose = 0,
	forceupdate = 1,	/* XXX For now, always update. Needs to be fixed */
	dryrun = 0,
	jobs_running = 0;

static GHashTable *jobtracker = NULL;

FILE	*commspipe=NULL;

/*! 
 * \brief get id of (ip or AS) path in the SQL table
 *
 * \param path text representation of an IP or AS path
 *
 * \param npoints number of elements in the path specified
 *
 * \param table operate on this SQL tables. Either "ASpaths" or "Routes"
 *
 * \param field name of the field of the SQL table. "path for the "ASpaths" table, "route" for the "Routes" table
 *
 * \param mysql mysql connection for either the IPv4 or IPv6 database
 *
 * \return int numeric id of the entry in the table
 */

int get_pathid(char *path, int npoints, GHashTable * hash, const char *table, const char *field, int src, MYSQL * mysql)
{

	struct routestruct *rsp;
	char            query[1024];
	char            crc[32],
	               *p_crc;
	int             id;

	/*
	 * Get crc value (MD5) for route 
	 */
	crc_calc(path, crc);

	/*
	 * Check if route exists 
	 */
	if ((rsp = g_hash_table_lookup(hash, &crc)))
	{
		if (debug > 2)
			fprintf(commspipe, "Found crc %s for %s\n", crc, path);

		id = rsp->id;
	}
	else
	{

		/*
		 * Insert route in DB 
		 */
		id = last_path_id(mysql,table)+1;
		sprintf(query, "INSERT INTO %s (id,crc,len,src,%s) VALUES (%u,\"%s\",%u,%u,\"%s\")", table, field, id, crc, npoints, src, path);
		if (dryrun)
		{
			if (debug)
				fprintf(commspipe, "%s\n", query);
		}
		else if (_mysql_query_with_retry(mysql, query, _MYSQL_FAILED_QUERY_RETRY, _MYSQL_FAILED_QUERY_DELAY, _MYSQL_FAILED_QUERY_BACKOFF))
		{
			fprintf(stderr, "Error: query %s: %s\n", query, mysql_error(mysql));
			exit(1);
		}

		if ((rsp = malloc(sizeof (struct routestruct))) == NULL)
		{
			perror("get_pathid");
			abort();
		}
		rsp->id = id;
		rsp->len = npoints;
		strcpy(rsp->crc, crc);

		/*
		 * Add route to hash of routes 
		 */
		if ((p_crc = strdup(crc)) == NULL)
		{
			perror("get_pathid");
			abort();
		}
		g_hash_table_insert(hash, p_crc, rsp);
	}
	return (id);
}

/*! 
 * \brief get id of IPv6  path in the SQL table. IPv6 routes are stored binary instead of textual
 *
 * \param path text representation of an v6 IP path
 *
 * \param npoints number of elements in the path specified
 *
 * \param table operate on this SQL tables. Should be "Routes"
 *
 * \param field name of the field of the SQL table. Should be "route" for the "Routes" table
 *
 * \param mysql mysql connection for the IPv6 database
 *
 * \return int numeric id of the entry in the table
 */

int get_pathid_ip6(char *path, int npoints, GHashTable * hash, const char *table, const char *field, int src, MYSQL * mysql)
{
	struct routestruct *rsp;
	char            query[4096];
	char            crc[32],
	               *p_crc;
	int             id,
	                i;
	char          **ap,
	               *arg[32];
	char           *buff,
	               *res;
	char            addr[INET6_BINSTRLEN];
	char            number[4];	/* ("0" - "255") + \0 */

	buff = path;

	/*
	 * split the hops in path and store them into arg[] 
	 */

#if defined(USE_STRTOK)			/* POSIX compliant */

	ap = arg;
	*ap = strtok(buff, " ");
	while (*ap != NULL)
	{
		if (**ap != '\0')
		{
			++ap;
		}
		*ap = strtok(NULL, " \t");
	}

#else							/* BSD specific */

	for (ap = arg; (*ap = strsep(&buff, " ")) != NULL;)
	{
		if (**ap != '\0')
		{
			++ap;
		}
	}
#endif

	/*
	 * the length of the binary string in the database  is proportional
	 * to number of hops
	 */
	if ((res = (char *) malloc(INET6_BINSTRLEN * npoints)) == NULL)
	{
		perror("get_pathid_ip6");
		abort();
	}
	memset(res, 0, INET6_BINSTRLEN * npoints);

	/*
	 * convert the hops from ASCII representation (arg[])
	 * to binary and fill binary string (res)
	 */
	for (i = 0; i < npoints; i++)
	{
		if (strcmp(arg[i], "0::0") == 0)
		{						/* do nothing */
		}
		else
		{
			inet_pton(AF_INET6, arg[i], (void *) addr);
			memcpy(res + INET6_BINSTRLEN * i, addr, INET6_BINSTRLEN);
		}
	}

	/*
	 * hex:         0x20 0x01 0x06 0x40 ...
	 * dez:           32   01   06   64 ...
	 * mysql:    CHAR(32,  01,  06,  64, ... )
	 *
	 * build mysql sub query string
	 */
	if ((buff = (char *) malloc(20 + npoints * INET6_BINSTRLEN * 4)) == NULL)
	{
		perror("get_pathid_ip6");
		abort();
	}
	memset(buff, 0, 20 + npoints * INET6_BINSTRLEN * 4);
	strcpy(buff, "CHAR(");
	for (i = 0; i < (npoints * INET6_BINSTRLEN); i++)
	{
		sprintf(number, "%d", res[i] & 0xFF);
		strcat(buff, number);
		if (i != (npoints * INET6_BINSTRLEN - 1))
			strcat(buff, ",");
	}
	strcat(buff, "))");

	/*
	 * Get crc value (MD5) for route 
	 */
	crc_calc(buff, crc);

	free(res);

	/*
	 * Check if route exists 
	 */
	if ((rsp = g_hash_table_lookup(hash, &crc)))
	{
		if (debug > 2)
			fprintf(commspipe, "Found crc %s for %s\n", crc, path);
		id = rsp->id;
	}
	else
	{

		/*
		 * Insert route in DB 
		 * XXX insert in alternate "fast" access table too ?
		 */
		id = last_path_id(mysql,table)+1;
		sprintf(query, "INSERT INTO %s (id,crc,len,src,%s) VALUES (%u,\"%s\",%u,%u,", table, field, id, crc, npoints, src);
		strcat(query, buff);

		if (dryrun)
		{
			if (debug)
				fprintf(commspipe, "%s\n", query);
		}
		else if (_mysql_query_with_retry(mysql, query, _MYSQL_FAILED_QUERY_RETRY, _MYSQL_FAILED_QUERY_DELAY, _MYSQL_FAILED_QUERY_BACKOFF))
		{
			fprintf(stderr, "Error: query %s: %s\n", query, mysql_error(mysql));
			exit(1);
		}

		rsp = malloc(sizeof (struct routestruct));
		rsp->id = id;
		rsp->len = npoints;
		strcpy(rsp->crc, crc);

		/*
		 * Add route to hash of routes 
		 */
		if ((p_crc = strdup(crc)) == NULL)
		{
			perror("get_pathid_ip6");
			abort();
		}
		g_hash_table_insert(hash, p_crc, rsp);
	}
	free(buff);
	return (id);
}

/*
-------------------------------------------------------------------------------
Purpose           : sort records hash
Params            : 1st element, 2nd element
Returns           : 0
Comments          : used by g_slist_sort
*/
int sort_records(gconstpointer a, gconstpointer b)
{
	/*
	 * Store the two elements in pointers to recstruct 
	 */
	const struct recstruct *pa = (const struct recstruct *) a;
	const struct recstruct *pb = (const struct recstruct *) b;

	/*
	 * First sort by src 
	 */
	if (pa->src < pb->src)
	{
		return (-1);
	}
	if (pa->src > pb->src)
	{
		return (1);
	}

	/*
	 * Then sort by dst 
	 */
	if (pa->dst < pb->dst)
	{
		return (-1);
	}
	if (pa->dst > pb->dst)
	{
		return (1);
	}

	/*
	 * Finally, sort by timestamp 
	 */
	if (pa->timestamp < pb->timestamp)
	{
		return (-1);
	}
	if (pa->timestamp > pb->timestamp)
	{
		return (1);
	}

	return (0);
}

/*
-------------------------------------------------------------------------------
Purpose           : Insert data from the records hash into the DB
Params            : pointer to a record, pointer to full record
Returns           : NULL
Comments          : used by g_slist_foreach after list has been sorted
*/
void process_records(gpointer data, gpointer user_data)
{
	/*
	 * rec contains one row of uncompressed data: rangeid, routeid,
	 * aspathid, timestamp
	 * 
	 * fullrec contains one row of compressed data: rangeid, routeid,
	 * aspathid, tstart, tend, numrec
	 */

	struct recstruct *rec;
	struct recordstruct *fullrec;
	struct procstruct *proc;
	char            query[1024];
	MYSQL_RES      *result;
	MYSQL_ROW       row;
	MYSQL          *mysql;
	unsigned int    id;
	unsigned int    tstart;
	unsigned int    numrec;
	unsigned int    routeid;
	unsigned int    aspathid;
	unsigned int    flag;

	rec = (struct recstruct *) data;
	proc = (struct procstruct *) user_data;

	mysql = proc->mysql;
	fullrec = (struct recordstruct *) proc->rec;

	id = tstart = numrec = aspathid = routeid = flag = 0;

	/*
	 * Normal case: same rangeid and same routeid 
	 */
	if ((fullrec->src == rec->src) && (fullrec->dst == rec->dst) && (fullrec->routeid == rec->routeid) && (fullrec->aspathid == rec->aspathid))
	{
		fullrec->tend = rec->timestamp;
		fullrec->numrec++;
		return;
	}

	if (fullrec->src != 0)
	{
		/*
		 * something has changed, insert current fullrec in the DB 
		 */

		if(proc->lastupdate < fullrec->tstart)
		{

			/* XXX queue up for extended insert ? */
			sprintf(query,
					"INSERT INTO Records (src,dst,routeid,aspathid,tstart,tend,numrec) VALUES (%u,%u,%u,%u,%u,%u,%u)",
					fullrec->src, fullrec->dst, fullrec->routeid, fullrec->aspathid, fullrec->tstart, fullrec->tend, fullrec->numrec);

			if (dryrun)
			{
				if (debug > 1)
					fprintf(commspipe, "%s\n", query);
			}
			else if (_mysql_query_with_retry(mysql, query, _MYSQL_FAILED_QUERY_RETRY, _MYSQL_FAILED_QUERY_DELAY, _MYSQL_FAILED_QUERY_BACKOFF))
			{
				fprintf(stderr, "Error: query %s: %s\n", query, mysql_error(mysql));
				return;
			}
		}
		else if (debug > 1)
			fprintf(commspipe, "Skipping update as timestamp of current record is less than last seen timestamp\n");
	}

	if ((fullrec->src != rec->src) || (fullrec->dst != rec->dst))
	{
		/*
		 * source or destination changed; since the list is ordered
		 * we will get here only once for each src/dst combination
		 */

		/*
		 * get last stored record between this src and dst 
		 * XXX cache ???
		 */
		
		sprintf(query, "SELECT id,routeid,aspathid,tstart,tend,numrec FROM Records WHERE src=%u AND dst=%u AND tstart=(SELECT MAX(tstart) FROM Records WHERE src=%u AND dst=%u)", rec->src, rec->dst, rec->src, rec->dst);

		if (_mysql_query_with_retry(mysql, query, _MYSQL_FAILED_QUERY_RETRY, _MYSQL_FAILED_QUERY_DELAY, _MYSQL_FAILED_QUERY_BACKOFF))
		{
			fprintf(stderr, "Error: mysql query %s: %s\n", query, mysql_error(mysql));
			return;
		}

		if (!(result = mysql_use_result(mysql)))
		{
			fprintf(stderr, "Error: mysql_use_result: %s\n", mysql_error(mysql));
			mysql_free_result(result);
			return;
		}

		if ((row = mysql_fetch_row(result)))
		{
			id = atoi(row[0]);
			routeid = atoi(row[1]);
			aspathid = atoi(row[2]);
			tstart = atoi(row[3]);
			numrec = atoi(row[5]);

			if (abs(rec->timestamp - atoi(row[4])) <= TIME_LIMIT)
			{
				flag = 1;
			}
		}

		mysql_free_result(result);
	}

	if (flag && (routeid == rec->routeid) && (aspathid == rec->aspathid))
	{
		/*
		 * the new record is a continuation from last one stored.
		 * delete old record from the DB, it will be replaced by a
		 * new one when 'fullrec' is flushed
		 */

		sprintf(query, "DELETE FROM Records WHERE id=%u", id);

		if (dryrun)
		{
			if (debug)
				fprintf(commspipe, "%s\n", query);
		}
		else if (_mysql_query_with_retry(mysql, query, _MYSQL_FAILED_QUERY_RETRY, _MYSQL_FAILED_QUERY_DELAY, _MYSQL_FAILED_QUERY_BACKOFF))
		{
			fprintf(stderr, "Error: query %s: %s\n", query, mysql_error(mysql));
			return;
		}

		/*
		 * Set fullrec to record fetched from DB 
		 */
		fullrec->src = rec->src;
		fullrec->dst = rec->dst;
		fullrec->routeid = rec->routeid;
		fullrec->aspathid = rec->aspathid;
		fullrec->tstart = tstart;
		fullrec->tend = rec->timestamp;
		fullrec->numrec = numrec + 1;
	}
	else
	{
		/*
		 * routeid changed or old record is no longer valid; set
		 * fullrec to rec
		 */

		fullrec->src = rec->src;
		fullrec->dst = rec->dst;
		fullrec->routeid = rec->routeid;
		fullrec->aspathid = rec->aspathid;
		fullrec->tstart = rec->timestamp;
		fullrec->tend = rec->timestamp;
		fullrec->numrec = 1;
	}
	return;
}

/*
-------------------------------------------------------------------------------
Purpose           : Delete records from Records hash
Params            : pointer to record
Returns           : NULL
Comments          : used by g_slist_foreach
*/
void remove_records(gpointer data, gpointer user_data)
{
	free(data);
}

/*
-------------------------------------------------------------------------------
Purpose           : The important function: process data from an RVEC or RVEC6 file,
                    store it in hashes
Params            : filename, ranges hash, routes hash, aspaths & asnums hashes,
		    flag that marks the IP family
Returns           : Number of lines processed
Comments          :
*/
int process_file(MYSQL * mysql, char *filename, struct cachetables *tables, int ip6flag)
{
	GSList         *records;

	gzFile          fdz;

	unsigned int    src,
	                dst,
	                error_count,
					ioerror=0,
	                tt,
	                npoints,
	                gz,
	                k;

	int             t,
	                dstreached,
				   *testbox_id=NULL,
					retry=3,
	                prev_testbox_id = -1;

	z_off_t			fpos_z=0;

	char           *src_ip6,
				   *dst_ip6,
					buf[4096],
					vector[(MAX_IP6_SIZE + 1) * MAX_HOPS],
					path[(MAX_AS_SIZE + 1) * MAX_HOPS],
					query[1024],
				   *pp,
				   *filenamelink;

	struct recstruct *rec;
	struct recordstruct *fullrec;

	struct procstruct *proc=NULL;


	records = NULL;

	/*
	 * Allocate memory to store fullrec, empty it 
	 */
	if ((fullrec = malloc(sizeof (struct recordstruct))) == NULL)
	{
		perror("process_file");
		abort();
	}
	memset(fullrec, 0, sizeof (struct recordstruct));

	gz = 0;

	k = 0;

	pp = NULL;

	fdz = NULL;

	/*
	 * Check if filename starts with RVEC[6] 
	 */
	filenamelink = strrchr(filename, '/');
	if (filenamelink == NULL)
		return (0);
	if (strncmp(filenamelink, "/RVEC", 5) != 0)
		return (0);

	/* always use gzopen */
	if ((fdz = gzopen(filename, "rb")) == NULL)
	{
		fprintf(commspipe, "Cannot open %s: %s\n", filename, strerror(errno));
		return (0);
	}
	/* TODO Get last pos, seek to last pos */

	error_count = 0;

	if (verbose)
		fprintf(commspipe, "Processing for %s\n", filename);
	setproctitle("Processing for %s", filename);

	/*
	 * read file. Always use zlib io as it is transparent.
	 */
	while (!ioerror && !gzeof(fdz))
	{
		if (gzgets(fdz, buf, sizeof (buf)) == Z_NULL)
		{
			if (!gzeof(fdz))
			{
				if (retry--)
				{
					gzclose(fdz);
					if (((fdz = gzopen(filename, "rb")) == NULL) || (gzseek(fdz,fpos_z,SEEK_SET) != fpos_z))
					{
						int gzerr;
						fprintf(commspipe, "Cannot reopen and seek to last known position (%lu) for %s: %s (%s)\n", fpos_z, filename, gzerror(fdz,&gzerr),gzerr == Z_ERRNO ? strerror(errno): "no errno");
						ioerror=1;
					}
					else if (debug)
						fprintf(commspipe, "Reopening and seek to last known position (%lu) for %s succeeded (%d retries to go)\n", fpos_z, filename, retry);
				}
				else
				{
					int gzerr;
					fprintf(commspipe,"Error while reading file %s: %s (%s)\n",filename,gzerror(fdz,&gzerr),gzerr == Z_ERRNO ? strerror(errno): "no errno");
					ioerror=1;
				}
			}
			continue;
		}
		else
			fpos_z = gztell(fdz);

		/* XXX unchecked FIXME! could continue here with corrupt buffers! */

		/* TODO store lastpos for this file */

		/*
		 * Empty vector 
		 */
		memset(vector, 0, sizeof (vector));

		/*
		 * Parse the line 
		 */
		if (!ip6flag)
		{
			t = read_line(buf, &src, &dst, &tt, &npoints, vector, path, tables->asnums);
		}
		else
		{
			t = read_line_ip6(buf, &src_ip6, &dst_ip6, &tt, &npoints, vector, path, tables->asnums);
		}

		/*
		 * Check if function returned 0 
		 */
		if (t > 0)
		{
			error_count++;
			continue;
		}
		else if (t == 0)
		{
			dstreached = TRUE;
		}
		else
		{						/* t < 0 */
			dstreached = FALSE;
		}

		k++;

		if (debug > 1 && k % 1500 == 0)
			fprintf(commspipe, "file=%s errors=%d lines=%d\n", filename, error_count, k);

		if (npoints == 0)
		{
			/*
			 * Make vector non-empty, for MySQL ? or CRC ? 
			 */
			strcpy(vector, " ");
		}

		/*
		 * process info, insert record 
		 */

		if ((rec = malloc(sizeof (struct recstruct))) == NULL)
		{
			perror("process_file");
			abort();
		}

		if ((testbox_id = (int *) g_hash_table_lookup(tables->ranges, !ip6flag ? (gconstpointer) & src : (gconstpointer) src_ip6)) == NULL)
		{

			fprintf(commspipe, "IP %s not found in ranges hash (file: %s)\n", !ip6flag ? rv_inttoip(src, vector) : src_ip6, filename);

			/*
			 * source IP not found, no point in
			 * processing this file
			 */
			free(rec);
			free(fullrec);
			return (0);
		}

		if (g_hash_table_size(tables->routes) == 0 || g_hash_table_size(tables->aspaths) == 0 || prev_testbox_id != *testbox_id)
		{
			if (verbose)
				fprintf(commspipe, "Loading routes and aspaths for %d\n", *testbox_id);
			setproctitle("Loading routes and aspaths for %d", *testbox_id);
			close(RRsock);
			if (debug)
			{
				fprintf(commspipe, "before removing: routes entries=%d\n", g_hash_table_size(tables->routes));
				fprintf(commspipe, "before removing: aspaths entries=%d\n", g_hash_table_size(tables->aspaths));
			}
			g_hash_table_remove_all(tables->routes);
			g_hash_table_remove_all(tables->aspaths);
			if (debug)
			{
				fprintf(commspipe, "after removing: routes entries=%d\n", g_hash_table_size(tables->routes));
				fprintf(commspipe, "after removing: aspaths entries=%d\n", g_hash_table_size(tables->aspaths));
			}
			get_route_table(mysql, tables->routes, *testbox_id);
			get_aspaths_table(mysql, tables->aspaths, *testbox_id);
			if (debug)
			{
				fprintf(commspipe, "after loading routes entries=%d\n", g_hash_table_size(tables->routes));
				fprintf(commspipe, "after loading aspaths entries=%d\n", g_hash_table_size(tables->aspaths));
			}
			prev_testbox_id = *testbox_id;
			RRopen(RR_SERVER, RR_SERVICE);
			RRconnect();
			if (verbose)
				fprintf(commspipe, "Processing for %s\n", filename);
			setproctitle("Processing for %s", filename);
		}
		rec->src = *testbox_id;

		if ((testbox_id = (int *) g_hash_table_lookup(tables->ranges, !ip6flag ? (gconstpointer) & dst : (gconstpointer) dst_ip6)) == NULL)
		{
			if (verbose)
				fprintf(commspipe, "IP %s not found in ranges hash (file: %s)\n", !ip6flag ? rv_inttoip(dst, vector) : dst_ip6, filename);

			continue;
		}

		rec->dst = *testbox_id;

		rec->routeid = ip6flag ? get_pathid_ip6(vector, npoints, tables->routes, "Routes", "route", *testbox_id, mysql) : 
									get_pathid(vector, npoints, tables->routes, "Routes", "route", *testbox_id, mysql);
		rec->aspathid = get_pathid(path, npoints, tables->aspaths, "ASpaths", "path", *testbox_id, mysql);
		rec->timestamp = tt;

		/*
		 * Add rec to list 
		 */
		records = g_slist_append(records, rec);
	}

	if(!ioerror)
	{
		/* Sort list, process all records */

		records = g_slist_sort(records, sort_records);

		if ((proc = malloc(sizeof (struct procstruct))) == NULL)
		{
			perror("process_file");
			abort();
		}

		proc->mysql = mysql;
		proc->rec = fullrec;
		proc->lastupdate = 0;

		if (!forceupdate)
		{
			/* 
			 * FIXME 
			 * Although selecting the max timestamp from a source to update the
			 * last record seen, it occasionally seems to insert records though.
			 * Especially the last full records case might be a bit odd in this
			 * case and needs further investigation before actually enabling
			 * this feature. RVS 20070206
			 */

			sprintf(query, "SELECT MAX(tstart) FROM Records WHERE src=%u", *testbox_id);
			if (_mysql_query_with_retry(mysql, query, _MYSQL_FAILED_QUERY_RETRY, _MYSQL_FAILED_QUERY_DELAY, _MYSQL_FAILED_QUERY_BACKOFF))
			{
				fprintf(stderr, "Error: mysql query %s: %s\n", query, mysql_error(mysql));
			}
			else
			{
				MYSQL_RES      *result;
				MYSQL_ROW       row;

				if ((result = mysql_use_result(mysql)) && ((row = mysql_fetch_row(result))))
				{
					proc->lastupdate = atoi(row[0]);
					if(debug)
						fprintf(commspipe,"Last update for %d seen at %d\n",*testbox_id,proc->lastupdate);
					mysql_free_result(result);
				}
				else
				{
					fprintf(stderr, "Error: mysql_use_result: %s\n", mysql_error(mysql));
				}
			}
		}

		if (verbose)
			fprintf(commspipe, "Updating database for %s\n", filename);
		setproctitle("Updating database for %s", filename);
		g_slist_foreach(records, process_records, proc);

		if(proc->lastupdate < fullrec->tstart)
		{

			/* flush last fullrec into the DB */
			/* XXX queue up for extended insert ? */
			sprintf(query,
					"INSERT INTO Records (src,dst,routeid,aspathid,tstart,tend,numrec) VALUES (%u,%u,%u,%u,%u,%u,%u)",
					fullrec->src, fullrec->dst, fullrec->routeid, fullrec->aspathid, fullrec->tstart, fullrec->tend, fullrec->numrec);

			if (dryrun)
			{
				if (debug)
					fprintf(commspipe, "(last fullrec) %s\n", query);
			}
			else if (_mysql_query_with_retry(mysql, query, _MYSQL_FAILED_QUERY_RETRY, _MYSQL_FAILED_QUERY_DELAY, _MYSQL_FAILED_QUERY_BACKOFF))
			{
				fprintf(stderr, "Error: query %s: %s\n", query, mysql_error(mysql));
				exit(1);
			}
		}
		else if (debug)
			fprintf(commspipe, "Skipping update of last fullrec as timestamp of current record is less than last seen timestamp\n");

		/*
		 * Empty list 
		 */
		free(proc);
	}
	g_slist_foreach(records, remove_records, NULL);

	g_slist_free(records);

	records = NULL;

	free(fullrec);

	/* Close file, report errors */
	
	gzclose(fdz);

	if (error_count > 0 && verbose)
	{
		fprintf(commspipe, "%s has %d errors.\n", filename, error_count);
	}

	setproctitle(NULL);
	return (k);
}

int process_slice(const char *mysql_host, int slice_start, int slice_end, char **files, struct cachetables *tables, int ip6flag)
{
	/*
	 * For each IPv4/6 file 
	 */
	int             i,
	                k = 0,
		timesum;
	double		z;

	struct tms      tmsdata;
	clock_t         usertimes,
	                systimes;

	time_t          start,
	                end;

	MYSQL           mysql;

	char            filename[FILENAME_MAX];

	systimes = 0;
	usertimes = 0;

	if (debug)
	{
		fprintf(commspipe,"Processing from %d to %d:\n", slice_start, slice_end);
		if (debug)
		{
			for (i = slice_start; i < slice_end; i++)
				fprintf(commspipe,"\t%s\n", files[i]);
		}
	}

	if (!ip6flag)
		open_admin_connection(&mysql, mysql_host);
	else
		open_admin_connection_ip6(&mysql, mysql_host);

	for (i = slice_start; i < slice_end; i++)
	{
		struct stat f_stat;
		start = time(NULL);

		strncpy(filename, files[i], FILENAME_MAX - 1);

		/* Doesn't look to be there, maybe it is compressed... */
		if (stat(filename, &f_stat) != 0 && errno == ENOENT)
			strncat(filename,COMPRESSED_FILE_EXTENTION, FILENAME_MAX - strlen(filename) - 1 );

		/* Still isn't there ?  then make a note of it and skip. */
		if ((stat(filename, &f_stat) != 0) && (errno = ENOENT))
		{
			fprintf(stderr, "Error: file %s listed but not found %s(skipping)\n", files[i], strerror(errno));
			continue;
		}

		RRopen(RR_SERVER, RR_SERVICE);
		RRconnect();

		/*
		 * Process this file 
		 */
		if ((k = process_file(&mysql, filename, tables, ip6flag)) == 0)
		{
			if (verbose)
				fprintf(commspipe, "No results from %s\n", filename);
			continue;
		}
		close(RRsock);

		end = time(NULL);
		timesum += (end - start);

		z = (end - start) ? k / (end - start) : k;

		times(&tmsdata);
		usertimes = tmsdata.tms_utime - usertimes;
		systimes = tmsdata.tms_stime - systimes;

		fprintf(commspipe, REPORT_FORMAT, filename, k, (end - start), z, usertimes, systimes);

		usertimes = tmsdata.tms_utime;
		systimes = tmsdata.tms_stime;

	}
	mysql_close(&mysql);
	return k;
}

int spawn_slices(const char *mysql_host, unsigned int n, char **files, unsigned int jobs, struct cachetables *tables, int ip6flag)
{

	unsigned int             real_jobs = jobs > n ? n : jobs;
	unsigned int             n_start, vectors_read = 0;

	for (n_start = 0; n_start < n; n_start += (n / real_jobs) + 1)
	{
		int             n_end = n_start + (n / real_jobs) < n ? n_start + 1 + (n / real_jobs) : n;

		if (jobs_running == real_jobs)
			vectors_read += waitforjob(0);
		else
		{
			int             pdes[2];
			pid_t          *pid = malloc(sizeof (pid_t));

			/* redirect stdin ? */
			if (pipe(pdes) == -1)
			{
				perror("spawn_slices");
				abort();
			}

			if ((*pid = fork()) == -1)
			{
				perror("spawn_slices");
				abort();
			}
			else if (*pid)
			{
				jobs_running++;
				if (verbose)
					printf("Forked pid %d, %d running...\n", *pid, jobs_running);

				struct sliceinfo *tracker = malloc(sizeof (struct sliceinfo));

				tracker->t_start = time(NULL);
				tracker->slice_start = n_start;
				tracker->slice_end = n_end;
				tracker->read_fd = fdopen(pdes[0], "r");

				setlinebuf(tracker->read_fd);
				close(pdes[1]);

				g_hash_table_insert(jobtracker, pid, tracker);
			}
			else
			{
				if ((commspipe = fdopen(pdes[1],"w")) == NULL)
				{
					perror("Couldn't set up communications to parent");
					abort();
				}
				setlinebuf(commspipe);
				close(pdes[0]);

				process_slice(mysql_host, n_start, n_end, files, tables, ip6flag);
				g_hash_table_destroy(tables->ranges);
				g_hash_table_destroy(tables->routes);
				g_hash_table_destroy(tables->aspaths);
				g_hash_table_destroy(tables->asnums);

				free_filenames(n, files);
				fflush(commspipe);
				fclose(commspipe);
				exit(0);
			}
		}
	}
	while (jobs_running)
		vectors_read += waitforjob(0);

	if (tables->ranges)
		g_hash_table_destroy(tables->ranges);
	if (tables->routes)
		g_hash_table_destroy(tables->routes);
	if (tables->aspaths)
		g_hash_table_destroy(tables->aspaths);
	if (tables->asnums)
		g_hash_table_destroy(tables->asnums);

	free_filenames(n, files);

	return vectors_read;
}

void jobtracker_set_fdset(gpointer key, gpointer val, gpointer userdata)
{
	int             pid = *(int *) key;
	struct sliceinfo *tracker = (struct sliceinfo *) val;
	struct jobtracker_set_fdset_param *param = (struct jobtracker_set_fdset_param *) userdata;

	FD_SET(fileno(tracker->read_fd), param->read_fds);
	if (param->max_fd < fileno(tracker->read_fd))
		param->max_fd = fileno(tracker->read_fd);
}

void jobtracker_read_fdset(gpointer key, gpointer val, gpointer userdata)
{
	int             pid = *(int *) key, vectors_read=0;
	char            message[1024];
	struct sliceinfo *tracker = (struct sliceinfo *) val;
	struct jobtracker_set_fdset_param *param = (struct jobtracker_set_fdset_param *) userdata;

	if (FD_ISSET(fileno(tracker->read_fd), param->read_fds))
	{
		if(fgets(message, sizeof (message), tracker->read_fd))

		{
			int delta;
			long unsigned int usertimes,systimes;
			float vectors_sec;
			char filename[1024];
	
			if (sscanf(message, REPORT_FORMAT, filename, &vectors_read,&delta,&vectors_sec,&usertimes,&systimes))
				param->vectors_read += vectors_read;
			if (verbose)
				printf("%s[%d] child fd=%d: %s", PROGNAME, pid, fileno(tracker->read_fd) , message);
		}
	}
}

int waitforjob(int signal)
{
	pid_t           pid;
	fd_set          read_fds;
	int             status = 0, vectors_read=0;
	struct sliceinfo *tracker;
	struct jobtracker_set_fdset_param param = { &read_fds, 0 ,0 };
	struct timeval  timeout = { 1, 500 };

	/* select on read_fd's from sub processes, which have their stdout mapped to the pipe  / read */

	FD_ZERO(&read_fds);
	g_hash_table_foreach(jobtracker, jobtracker_set_fdset, &param);

	if (param.max_fd)
		if ((status = select(param.max_fd + 1, &read_fds, NULL, NULL, NULL)) == -1)
		{
			perror("waitforjob");
			abort();
		}

	if (status)
		g_hash_table_foreach(jobtracker, jobtracker_read_fdset, &param);

	vectors_read = param.vectors_read;

	if (debug)
		printf("Waiting for process... %d running\n", jobs_running);
	if ((pid = waitpid(-1, &status, WNOHANG)) != -1)
		if (pid)
		{
			jobs_running--;
			if ((tracker = g_hash_table_lookup(jobtracker, &pid)))
			{
				if (debug)
				{
					printf("process %d returned... delta=%d\n", pid, time(NULL) - tracker->t_start);
				}
				fclose(tracker->read_fd);
				g_hash_table_remove(jobtracker, &pid);
			}
			else
				fprintf(stderr, "WARNING: untracked process %d returned, status=%d\n", pid, status);
			if (status)
			{
				if (WTERMSIG(status))
					fprintf(stderr, "process %d exited due to signal %d %s\n", pid, WTERMSIG(status), WCOREDUMP(status) ? "(core dumped)" : "");
			}
		}
	return vectors_read;
}

int last_path_id(MYSQL *mysql,const char *table) {
	char            query[1024];
	int last_id=-1;
	sprintf(query, "SELECT MAX(id) from %s", table);
	if (_mysql_query_with_retry(mysql, query, _MYSQL_FAILED_QUERY_RETRY, _MYSQL_FAILED_QUERY_DELAY, _MYSQL_FAILED_QUERY_BACKOFF))
	{
		fprintf(stderr, "Error: mysql query %s: %s\n", query, mysql_error(mysql));
	}
	else
	{
		MYSQL_RES      *result;
		MYSQL_ROW       row;

		if ((result = mysql_use_result(mysql)) && ((row = mysql_fetch_row(result))))
		{
			last_id = atoi(row[0]);
			if(debug)
				fprintf(commspipe,"Last id for %s: %u\n",table,last_id);
			mysql_free_result(result);
		}
		else
		{
			fprintf(stderr, "Error: mysql_use_result: %s\n", mysql_error(mysql));
		}
	}
	return last_id;
}

/*
-------------------------------------------------------------------------------
Purpose           : Main function - Initialize and process files
Params            : filename containing the RVEC files
Returns           : 0
Comments          :
*/
int main(int argc, char *argv[])
{
	unsigned int    n,
	                m,
	                k,
	                tot,
	                tot_ip6,
	                timesum;
	const char           *path, *mysql_host=select_mysql_host();

	struct cachetables tables_ip4,
	                tables_ip6;

	int             ip4flag = 0;	/* "1" means do ONLY IPv4 data */
	int             ip6flag = 0;	/* "1" means do ONLY IPv6 data */

	char           *files[4096];
	char           *files_ip6[4096];
	char            path_boxes[50];

	char            ch;

	unsigned int    jobs = 1;

	time_t          start = time(NULL),
		end;

	memset(&tables_ip4, 0, sizeof (struct cachetables));
	memset(&tables_ip6, 0, sizeof (struct cachetables));

	k = tot = tot_ip6 = timesum = 0;

	strcpy(path_boxes, PATHBOXES);

	/*
	 * Check arguments 
	 */
	if (argc < 2)
	{
		fprintf(stderr, "Usage: %s [-n] [-h mysqlserver ] [-v] [-j 1..N ] [-d] [-4] [-6] <filename>\n", argv[0]);
		exit(1);
	}
	while ((ch = getopt(argc, argv, "fnh:vd46H:j:")) != -1)
	{
		switch (ch)
		{
		case 'f':
			forceupdate = 1;
			break;
		case 'n':
			dryrun = 1;
			break;
		case 'v':
			verbose++;
			break;
		case 'd':
			debug++;
			break;
		case '4':
			ip4flag = 1;
			if (verbose)
				printf("Processing only IPv4 data...\n");
			break;
		case '6':
			ip6flag = 1;
			if (verbose)
				printf("Processing only IPv6 data...\n");
			break;
		case 'j':
			jobs = atoi(optarg);
			if (jobs < 1)
			{
				printf("\"%s\"needs to be a number in the range 1..N\n", optarg);
				exit(1);
			}
			if (verbose)
				printf("Configuring for %d parallel jobs...\n", jobs);
			break;
		case 'h':
			mysql_host=optarg;
			break;
		default:
			fprintf(stderr, "Usage: %s [-n] [-h] [-v] [-d] [-4] [-6] <filename>\n", argv[0]);
			break;
		}
	}
	argc -= optind;
	argv += optind;

	path = argv[0];

	if (ip4flag && ip6flag)
	{
		fprintf(stderr, "IP switch type mismatch, please check your syntax!\n");
		exit(1);
	}
	if (path == NULL)
	{
		fprintf(stderr, "Usage: %s [-n] [-h] [-v] [-d] [-4] [-6] <filename>\n", argv[0]);
		exit(1);
	}
	if (jobs > 1 && debug > 1)
	{
		jobs = 1;
		printf("debugging has been requested, limiting to single parallel job only\n");
	}


	n = m = 0;

	/*
	 * IPv4 and IPv6 currently use their own SQL handles which connect to
	 * the v4 or v6  SQL database. This was easiest for Florian to
	 * implement. But if we don't need the flexibility of hosting these
	 * DBs on different server hosts, we can probably streamline the code
	 * to use one MySQL handle and switch DBs as appropriate with the API
	 */

	if (!ip6flag)
	{

		/* IPv4 specific  */

		/* Initialization: open connection, read routes and ranges  */

		tables_ip4.ranges = g_hash_table_new_full(g_int_hash, g_int_equal, free, free);
		tables_ip4.asnums = g_hash_table_new_full(g_int_hash, g_int_equal, free, free);

		tables_ip4.routes = g_hash_table_new_full(g_str_hash, g_str_equal, free, free);
		tables_ip4.aspaths = g_hash_table_new_full(g_str_hash, g_str_equal, free, free);

		get_ranges(tables_ip4.ranges, path_boxes, 0);

		n = get_filenames(path, files, 0);
		if (debug)
			printf("%s: Found %d IPv4 files\n", path, n);
	}
	if (!ip4flag)
	{

		/* IPv6 specific */

		/* Initialization: open connection, read routes and ranges  */

		tables_ip6.ranges = g_hash_table_new_full(g_str_hash, g_str_equal, free, free);
		tables_ip6.asnums = g_hash_table_new_full(g_str_hash, g_str_equal, free, free);

		tables_ip6.routes = g_hash_table_new_full(g_str_hash, g_str_equal, free, free);
		tables_ip6.aspaths = g_hash_table_new_full(g_str_hash, g_str_equal, free, free);

		get_ranges(tables_ip6.ranges, path_boxes, 1);

		m = get_filenames(path, files_ip6, 1);
		if (debug)
			printf("%s: Found %d IPv6 files\n", path, m);
	}

	if ((jobtracker = g_hash_table_new_full(g_int_hash, g_int_equal, free, free)))
	{
		setproctitle("%s (controller)",PROGNAME);
		tot = spawn_slices(mysql_host, n, files, jobs, &tables_ip4, 0);
		tot_ip6 = spawn_slices(mysql_host, m, files_ip6, jobs, &tables_ip6, 1);

		g_hash_table_destroy(jobtracker);
	}

	end = time(NULL);

	if (!ip6flag && (tot == 0))
	{
		printf("No IPv4 files could be processed.\nPlease check that you have used full path filenames.\n");

	}
	if (!ip4flag && (tot_ip6 == 0))
	{
		printf("No IPv6 files could be processed.\nPlease check that you have used full path filenames.\n");

	}
	printf("%u IPv4 files (%u lines) and %u IPv6 files (%u lines) processed in %u seconds\n", n, tot, m, tot_ip6, end - start);

	exit(0);
}
