/*
Copyright (c) 2000,2001		     RIPE NCC


All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of the author not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS; IN NO EVENT SHALL
AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY
DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

/*
-------------------------------------------------------------------------------
Module Header
Filename	  : TTMDelayData.C
Author	    	  : Rene Wilhelm
Date	      	  : 13-OCT-2000
Revised		  : 20-JUN-2001, better object model, new methods
Revised		  : 13-AUG-2001, TestBox has become a class
Description       : Implementation of TTMDelayData class (create chain of
		    ROOT TTree files for specific measurement period)
Language Version  : C++
OSs Tested	  : Solaris 2.6
$Id: TTMDelayData.C,v 1.15 2006/11/13 15:21:27 ruben Exp $
-------------------------------------------------------------------------------
*/

#include <iostream>
#include <algorithm>
#include <cstring>
#include <cerrno>
#include <sys/stat.h>

#include <TFile.h>
#include <TTree.h>

#include "TTMDelayData.h"
#include "TTMStatistics.h"
#include "TTMDelayHistograms.h"
#include "CSVdump.h"
#include "Delay.h"

#include "utils.h"


RCSID("$Id: TTMDelayData.C,v 1.15 2006/11/13 15:21:27 ruben Exp $");



/*
-------------------------------------------------------------------------------
Subroutine Header
Purpose	   	:     TTMDelayData constructor 
Side Effects    :     initialize private data members
Comments	:
-------------------------------------------------------------------------------
*/

TTMDelayData::TTMDelayData(bool _checkdata) : checkdata(_checkdata) {

	chain = NULL;
	sumTree = NULL;

	delayStats = NULL;
	csv = NULL;

	dataProcessed = 0;
	maxBoxId = 0;
	numIntervals = 0;
	endTime = 0;
	timePeriod = 0;
}

/*
-------------------------------------------------------------------------------
Subroutine Header
Purpose	  	:   TTMDelayData destructor
Side Effects    :   deletes ROOT TChain + dynamically created datapath string
Comments	:
-------------------------------------------------------------------------------
*/

TTMDelayData::~TTMDelayData() {

	if (chain != NULL)	 delete chain;

	if (csv != NULL)	 delete[] csv;
	if (delayStats != NULL)  delete[] delayStats;
	if (!delayHistos.empty())
		for_each(delayHistos.begin(),delayHistos.end(),delete_pointer<TTMHistograms*>());
}

/*
-------------------------------------------------------------------------------
Subroutine Header
Purpose	   :   Set TTM Configuration for this data chain
Input parameters  :   config - TestBoxConfig object, a list of all
		      TestBoxes which were actively participating in the 
		      measurement network at the end time of the data chain
Comments	  :   also initializes internal bookkeeping parameters
-------------------------------------------------------------------------------
*/

void TTMDelayData::SetConfig(TTMOptions& opt, TestBoxConfig& conf, int numint) {
	
   
	tbConfig = &conf;
	progOptions = &opt;

	
	int oldNumBoxes = numBoxes;

	int oldNumIntervals = numIntervals;
	numIntervals = numint;


	// initialize data members from program options
	
	dataPath    = progOptions->InputPath();
	endTime     = progOptions->EndTime();
	outputDir   = progOptions->OutputPath();
	tmpDir      = progOptions->TmpDir();
	plotFormat  = progOptions->PlotFormat();
	storageType = progOptions->StorageType();
	timePeriod  = progOptions->TimePeriod();
	ipv6Chain   = (progOptions->IPversion() == 6);

	outputFile   = progOptions->OutputPath();

	struct stat statbuf;
	if (stat(outputFile.c_str(), &statbuf) == 0) {
		// it exists

		if (S_ISDIR(statbuf.st_mode)) {
			// it is a directory, append default output file name
			outputFile = outputDir + "/summary.root";
		}
	}
	else if (progOptions->Verbose()) {
		perror(outputFile.c_str());
	}

	// Determine the corresponding Id to TestBox translation table
	// since array indices run 0..arraySize-1, we have to allocate one
	// more element to allow direct indexing with testbox ids

	maxBoxId = tbConfig->GetMaxBoxId();

	nSources = tbConfig->GetNumEntries();

	// fill the slots which correspond to active boxes and their targets
	// (essential when only one source box present in running config)
	
	// This whole mapping could probably be moved to the TestBox class
	// where a GetTargetIndex(targetId) method would return the 
	// position of targetId in targetList array.
	// It would mean that delayStats and delayHistos arrays
	// in TTMDelayData class would be created and dimensioned
	// anew for each source box. For now we stick with this 
	// less ideal "global" id2index mapping.  [RW 20030113]


	// id2index is now actually a STL map, which maps a testbox id to the offset in the 
	// delayHistos/delayStats/csv arrays.
	// refactoring these array's to testbox id indexed maps is not that feasible, as it would introduce
	// out of bounds checking and STL iterators for each of them, where as it is now handled by the id2index
	// map, acting as a general data sentry.
	// (refactoring these array's to STL vectors might be still interesting though...)
	// [RVS 20031110]

	// XXX same code as in TTMSummaryChain.C, consolidate plse

	sourceList = tbConfig->GetTBList();
	int index = 0;
	for (map<uint,TestBox*>::iterator k=sourceList.begin() ; k != sourceList.end() ; k++) {
		uint srcid = PAIR_KEY(k);
		if (srcid > tbConfig->GetMaxBoxId()) {
			cerr << progOptions->ProgramName()
			     <<" ERROR: source id "
			     << srcid << " out of range; expected 1-"
			     << tbConfig->GetMaxBoxId() << endl;
		}
		else {
			if (id2index.find(srcid) == id2index.end()) {
				// not yet defined
				id2index[srcid] = index++;
			}
		}

		// now check this box' targets

		map<uint,TestBox*> &target = PAIR_VAL(k)->GetTargets();
		for (map<uint,TestBox*>::iterator j=target.begin() ; j != target.end() ; j++) {
			uint tgtid = PAIR_KEY(j);
			if (tgtid > tbConfig->GetMaxBoxId()) {
				cerr << progOptions->ProgramName()
				     <<" ERROR: target id "
				     << tgtid << " out of range; expected 1-"
				     << tbConfig->GetMaxBoxId() << endl;
			}
			else {
				if (id2index.find(tgtid) == id2index.end()) {
					// not yet defined
					id2index[tgtid] = index++;
				}
			}
		}
		
	}

	numBoxes = index;

	if (progOptions->Verbose()) {
		cout << "Found " << numBoxes
		     << " distinct boxes in src/tgt config" << endl;
	}


	if (((oldNumBoxes != numBoxes) && (oldNumBoxes != 0))  ||
	    ((oldNumIntervals != numIntervals) && (oldNumIntervals != 0))) { 
		// have to redo all analysis with different interval settings
		// delete current array with TTMStatistics

		if (delayStats != NULL) {
			delete[] delayStats;
			delayStats = NULL;
		}
	}
}

/*
-------------------------------------------------------------------------------
Subroutine Header
Purpose	   	  :   create (new) ROOT TTree chain 
Comments	  :   one chain is created of ALL specified sources
		  :   existing TChain of this object will be deleted
Future		  :   add code to Reset/Clear histograms for Delay Plots
-------------------------------------------------------------------------------
*/

void TTMDelayData::Chain(TestBox* source) {
	

#define BUFLENGTH 15

	char charbuf[BUFLENGTH];  // small buffer for temporary storage
	struct tm *timestruct;
	time_t starttime, timestamp;	

	int    nfiles;		  // number of files to chain


// for all targets ...

	// initalize data members

	// dispose of old ROOT chain if it still exists
	if (chain != NULL) {
		delete chain;
	}

	// create ROOT chain
	chain = new TChain("tree");


	starttime = endTime - timePeriod;

	if (endTime % SECONDSPERDAY == 0) {
		// endtime on day boundary, push it forward a little
		// to get the correct data file
		timestamp = starttime + 5;  
	}
	else {
		timestamp = starttime;
	}

	// since endtime is in time_t (seconds since 1/1/1970) format
	// we can determine day numbers (since 1/1/1970) by simple integer
	// division; number of files to chain is difference between
	// end and start day.  Subtract and add 5 seconds respectively
	// to avoid rounding errors

	int endday = (endTime-5) / SECONDSPERDAY;
	int startday = (endTime-timePeriod+5) / SECONDSPERDAY;
	nfiles = (endday - startday) + 1;

	for (int i = nfiles; i>0; i--)
	{

		// copy path
		string fullpath = dataPath + "/";

		timestruct = gmtime(&timestamp);
		if (storageType == 'H') {
			// append appropriate <year>/<month>/<day>
			strftime(charbuf, BUFLENGTH, "%Y/%m/%d/", timestruct);
			fullpath += charbuf;
		}

		if (ipv6Chain) {
			fullpath += "IPv6.";
		}
			
		// construct full filename from source name and timestruct
		strftime(charbuf, BUFLENGTH, ".%Y%m%d.root", timestruct);
		fullpath += source->GetName();
		fullpath += charbuf;

		// check if file exists before adding

		struct stat statbuf;
		if (stat(fullpath.c_str(), &statbuf) < 0) {
			perror(fullpath.c_str());
		}
		else {
			chain->Add(fullpath.c_str());
			if (progOptions->Verbose()) {
				cerr << "Adding " << fullpath.c_str() << endl;
			}
		}
		timestamp += SECONDSPERDAY; // increase timestamp by a day
	}
}

/*
-------------------------------------------------------------------------------
Subroutine Header
Purpose	   	  :   Process packets in TTM Data chain
		      Update TTM Statistics object(s)
Side Effects      :   
Comments	  :   Target list has been set by SetTargets() method
TODO		  :   Also fill histograms for later plotting

-------------------------------------------------------------------------------
*/
int TTMDelayData::ProcessPackets(TestBox* source, ttmdata type) {
	
	int index;
	float source_ntp_avg=0;
	int clock_ok_seen=0;
	time_t starttime = endTime - timePeriod;

	if (chain == NULL) {
		cerr << "TTMDelayData::ProcessPackets : " << 
			"No data chain defined, please call " <<
			"Chain() method first" << endl;
		return(0);
	}

	InitStorage(source, type);
		
	// loop over all data

	Delay *delay = new Delay();
	chain->SetBranchAddress("delay", &delay);
	Int_t nMeasurements = (Int_t) chain->GetEntries();

	nProcessed = 0;
	nArrived = 0;
	for (Int_t eventNo=0; eventNo < nMeasurements; eventNo++) {
		if (chain->GetEvent(eventNo) < 0) {
			// problem reading this "event"?
			continue;
		}

		Int_t targetId = delay->GetTargetId();
		if ( (targetId <= (int)maxBoxId) && (targetId > 0)) {
			// valid targetId
			if (id2index.find(targetId) == id2index.end()) {
				continue;  // out of range, no target, skip
			}
			else
				index = id2index[targetId];
	


	       		// part of requested targets list
			if (delay->GetSourceId() == (Int_t) source->GetId()) {
				// packet from requested source
				if (progOptions->Debug())
					cout << "INFO: " << delay->GetSourceId() << ": delayHistos[" << index << "] for target box " << targetId << endl;
				

				if (type == delaystats) {
					delayStats[index].Update(delay, starttime);
				}
				else if (type == delayhist) {
					if (delayHistos[index])
						delayHistos[index]->Fill(delay);
					else if (progOptions->Verbose())
						cout << "ERROR: " << delay->GetSourceId() << ": No TTMHistograms object in delayHistos[" << index << "] for target box " << targetId << endl;
				}
				else if (type == csvdump) {
					if (!checkdata)
						csv[index].Dump(delay);
					else
					{
						if (delay->SourceClockOK())
							clock_ok_seen++;

						source_ntp_avg += delay->GetSourceNtp();
					}
				}
				// XXX try to fudge ipdv in here ?

				if (delay->Status() != PacketLost)
					nArrived++;
				nProcessed++;
			}
		}
	}

    if (progOptions->Verbose()) {
		cout << "Source Box " << source->GetId() << ": " << nProcessed
		     << " measurements processed" << endl;
	}

	if(checkdata)
	{
		cout << "name=" << source->GetName() << " packets=" << nProcessed << 
								" arrived=" << nArrived << 
								" srcclockok=" << clock_ok_seen << 
								" source_ntp_avg=" << source_ntp_avg/nArrived << endl;
	}

	dataProcessed = 1;

	return(nProcessed);
}


/*
-------------------------------------------------------------------------------
Subroutine Header
Purpose	   	  :   Initialize storage for data processing
		      Create Arrays, Histograms etc.
Side Effects      :   

-------------------------------------------------------------------------------
*/
void TTMDelayData::InitStorage(TestBox* source, ttmdata type) {

 	// Create Statistics, Histogram or CSV objects for all boxes
 	// in TB config  (the superset of all selected sources
 	// and their targets);

	if (type == delaystats ) {

		if (delayStats == NULL) {
			// create new objects

			delayStats = new TTMStatistics[numBoxes];
			for (int i=0; i<numBoxes; i++) {
				delayStats[i].SetNumIntervals(numIntervals);
			}
		}
		else {
			// current allocation is fine,
			// just reset statistics counters

			for (int i=0; i<numBoxes; i++) {
				delayStats[i].Reset();
			}
		}
	}
	
	// Histograms

	else if (type == delayhist) {
	
		map<uint, TestBox*> &target = source->GetTargets();

		if (!delayHistos.empty()) {
			// delete old Histograms
			for_each(delayHistos.begin(),delayHistos.end(),delete_pointer<TTMHistograms*>());

			delayHistos.clear();
		}

		// We need to reserve the amount of histograms needed in the vector
		// because we can get a const reference to a default value when we try to insert it immediatly
		// using the [] operator
		//
		// delayHistos = new TTMDelayHistograms[numBoxes] is not going to work anymore because 
		// sizeof(TTMDelayHistograms) != sizeof(TTMHistograms) and thus messes up our virtual function table
		// TTMHistograms is a pure virtual base class providing a histogram configure, fill and plot interface.
		// [RVS 20031111]

		// XXX Do we really need reserve() ?
		delayHistos.reserve(numBoxes);

		// resize() is needed to properly report delayHistos.size()
		// reserve() doesn't do that for us.
		delayHistos.resize(numBoxes);

		for (map<uint,TestBox*>::iterator i=target.begin(); i != target.end(); i++) {
			int k = id2index[PAIR_KEY(i)];
			delayHistos[k] = new TTMDelayHistograms(source, PAIR_KEY(i), progOptions);	// I'm the mighty polymorphic object pointer
		}
	}

	else if (type == csvdump) {

		// CSV Output

		map<uint,TestBox*> &target = source->GetTargets();

		if (csv != NULL) {
			// delete old Histograms
			delete[] csv;
		}

		csv = new CSVdump[numBoxes];

		for (map<uint,TestBox*>::iterator i=target.begin(); i != target.end(); i++) {
			int k = id2index[PAIR_KEY(i)];

			string prefix = "";
			if (ipv6Chain) {
				prefix = "IPv6.";
			}

			// Don't output files on sanity report!
			if (!checkdata)
				csv[k].Open(outputDir, source->GetShortName(),
					PAIR_VAL(i)->GetShortName(), prefix, delays);
		}
	}
	// TODO try to fudge ipdv in here ?

}


/*
-------------------------------------------------------------------------------
Subroutine Header
Purpose	  	 :   Create summary ROOT file for this DelayData set
Descpription	 :   For each source box chain the data files for 
		     configured dates and output the statistics to
		     the summary TTree file

-------------------------------------------------------------------------------
*/

void TTMDelayData::Summarize() {
	

	// Create ROOT TTree

	Int_t split = 1;
	Int_t bsize = 64000;

	TFile *sumfile = new TFile (outputFile.c_str(), "RECREATE");
	sumTree  = new TTree ("TTMSummary", "TTM DelaySummary");

       	if (progOptions->Verbose()) {
		cout << "Opened " << outputFile << " for Summary tree" << endl;
	}
		
	summary = new DelaySummary();
   	sumTree->Branch ("summary", "DelaySummary", &summary, bsize, split);


	for(map<uint,TestBox*>::iterator i = sourceList.begin(); i != sourceList.end(); i++) {

		Chain(PAIR_VAL(i));
		if (ProcessPackets(PAIR_VAL(i), delaystats) != 0) {
			// some delay data found
			FillSummaryTree(PAIR_VAL(i), sumfile);
		}
	}

	sumfile->Close();
}

/*
-------------------------------------------------------------------------------
Subroutine Header
Purpose	  	 :   Create delay Plots for this DelayData set
Descpription	 :   For each source box chain the data files for 
		     configured dates and create the plots in
		     the output directory

-------------------------------------------------------------------------------
*/

void TTMDelayData::Plot() {
	

	for(map<uint,TestBox*>::iterator i = sourceList.begin(); i != sourceList.end(); i++) {

		Chain(PAIR_VAL(i));
		if (ProcessPackets(PAIR_VAL(i), delayhist) != 0) {
			// some delay data found
			PlotHistos(PAIR_VAL(i));
		}
	}
}

/*
-------------------------------------------------------------------------------
Subroutine Header
Purpose	  	 :   Output data in CVS format
Descpription	 :   For each source box chain the data files for 
		     configured dates and dump data to a file

-------------------------------------------------------------------------------
*/

void TTMDelayData::OutputCSV() {
	
	for(map<uint,TestBox*>::iterator i = sourceList.begin(); i != sourceList.end(); i++) {
		Chain(PAIR_VAL(i));
		if (ProcessPackets(PAIR_VAL(i), csvdump) && checkdata) {
			// Output per box sanity statistics
		}
	}
}

		
/*
-------------------------------------------------------------------------------
Subroutine Header
Purpose	  	 :   Fill summary Tree with collected statistics
Input		 :   source - Source Testbox
		     hfile  - Pointer to ROOT TTree file

For each target box output the percentile levels for five intervals:

	HRS   0 -  6
	HRS   6 - 12
	HRS  12 - 18
	HRS  18 - 24
	HRS   0 - 24

If the chain has multiple days worth of data, the intervals accumulate
data for the whole period. I.e. the median in the 1st interval represents
the median of all measurements that took place between 0 and 6 UTC in *any*
of the days of the chained period!

Also output the loss rates for the specific intervals 

-------------------------------------------------------------------------------
*/

void TTMDelayData::FillSummaryTree(TestBox* source, TFile* hfile) {
		

	map<uint,TestBox*> &target = source->GetTargets();

	int nfills = 0;

	for (map<uint,TestBox*>::iterator i=target.begin(); i != target.end(); i++) {

		uint targetId = PAIR_KEY(i);
		int k = id2index[targetId];

		for (int j=0; j<=numIntervals;j++) {

			summary = new DelaySummary();

			summary->SetSourceId (source->GetId());
			summary->SetTargetId (targetId);
			summary->SetBinNo(j);

			summary->SetNumDays  ((timePeriod+5)/SECONDSPERDAY);

			Double_t binsize, timelow;

			if (j==0) {
				// interval 0 summarizes whole period
				binsize = timePeriod;
				timelow = endTime-timePeriod;
			}
			else {
				// other intervals cover specific part of day
				// It's hard to find proper value of binsize
				// because each bin <j> is in fact a collection
				// of <numdays> smaller bins.
				// <timelow> for such a bin is even harder
				// to define. For now we stick with these.

				binsize = SECONDSPERDAY/numIntervals;
				timelow = endTime - timePeriod + (j - 1) * binsize;
			}

			summary->SetBinSize (binsize);
			summary->SetTimeLow (timelow);
			summary->SetNumEntries(delayStats[k].GetNumPacketsValid(j)); 
			summary->SetLevel025(delayStats[k].GetDelayPerc(j, 2.5));
			summary->SetMedian  (delayStats[k].GetDelayPerc(j, 50));
			summary->SetLevel975(delayStats[k].GetDelayPerc(j, 97.5));
			summary->SetLoss    (delayStats[k].GetLossRate(j));

			sumTree->Fill(); // put the above object in the tree
			nfills++;

			delete summary;  // can't be reused by ROOT
		}
	}
		

	hfile->Write ();

/*
	cout << "Source Box : " << source->GetId() <<endl<<endl;

	for (int i=0; i<nTargets; i++) {
		uint targetId = target[id]->GetId();
		int k = id2index[targetId];
		cout <<"Target: "  << target[k].GetId() <<endl;
		cout <<"Valid Packets: " << delayStats[k].GetNumPacketsValid(0); 
		for (int j=1; j<=numIntervals;j++) {
			cout << " " << delayStats[k].GetNumPacketsValid(j);
		}
		cout <<endl;
		cout <<"Loss Rate: " << delayStats[k].GetLossRate(0); 
		for (int j=1; j<=numIntervals;j++) {
			cout << " " << delayStats[k].GetLossRate(j);
		}
		cout <<endl;
		cout <<"Median: " << delayStats[k].GetDelayPerc(0, 50); 
		for (int j=1; j<=numIntervals;j++) {
			cout << " " << delayStats[k].GetDelayPerc(j, 50);
		}
		cout <<endl;
	}
	
*/
	return;
}

/*
-------------------------------------------------------------------------------
Subroutine Header
Purpose	  	 :   Plot the delay histograms objects
Input		 :   source - Source Testbox


-------------------------------------------------------------------------------
*/

void TTMDelayData::PlotHistos(TestBox* source) {

	map<uint, TestBox*> &target = source->GetTargets();

	for (map<uint,TestBox*>::iterator i=target.begin(); i != target.end(); i++) {

		int k = id2index[PAIR_KEY(i)];
		delayHistos[k]->Plot();
	}
}
