/*
Copyright (c) 2000                      RIPE NCC


All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of the author not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS; IN NO EVENT SHALL
AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY
DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

/*
-------------------------------------------------------------------------------
Module Header
Filename          : TTMDelayHistograms.C
Author            : Rene Wilhelm
Date              : 13-OCT-2000
Description       : Implementation of TTM TTMDelayHistograms class
Language Version  : C++
OSs Tested        : Solaris 2.6
External Programs : ps2gif (which in turn needs ghostscript and imaging tools)
$Id: TTMDelayHistograms.C,v 1.13 2008/04/09 18:52:16 ruben Exp $
-------------------------------------------------------------------------------
*/

#define PATH "/ncc/ttpro/data/root"   // path to TTM ROOT data storage
#define TYPE 'H'		      // hierarchical (year/month/day) storage

#include <cstdlib>
#include <ctime>
#include <iostream>
#include <string>
#include <TCanvas.h>
#include <TGaxis.h>
#include <TLine.h>
#include <TPaveLabel.h>
#include <TPostScript.h>
#include <TText.h>

#include "Delay.h"
#include "TTMDelayHistograms.h"
#include "utils.h"

// Spme color definitions
#define BG_CANVAS 169
#define BG_TITLE 119
#define BG_PAD 19
#define GRAPH_STD_LINE 1
#define GRAPH_PACKET_SENT 7
#define GRAPH_BAD_BLOCKS 2
#define GRAPH_INVALID 38
#define GRAPH_BAD_TARGET_CLOCK 5
#define GRAPH_BAD_SOURCE_CLOCK 6
#define GRAPH_CLOCKS_OK 3 // green
#define GRAPH_ARRIVED 3 // green
#define GRAPH_ROUTES_N_HOPS 2
#define GRAPH_DELAY1D_LINE 5
#define TEXT_LABEL 1


RCSID("$Id: TTMDelayHistograms.C,v 1.13 2008/04/09 18:52:16 ruben Exp $");


/*
-------------------------------------------------------------------------------
Subroutine Header
Description	  : Add new datapoint to Histograms
Input		  : pointer to Delay object with result from TTM measurement
Side Effects	  : various ROOT histograms updated
Language Version  : C++
OSs Tested        : Linux
-------------------------------------------------------------------------------
*/

void TTMDelayHistograms::Fill (Delay* packet) {

	// process one delay measurement: fill histograms and percentiles
	// calling program already selected right TTMHistogram object
	
	int nhops, routeid;
	
	
	Double_t packettime = packet->GetPacketTime(starttime);
	if ((packettime < starttime) || (packettime > endtime)) {
		// outside selected time interval, silently skip
		return;
	}

	// transform UNIX timestamp into ROOT timestamp
	// (i.e. seconds since 1/1/1995 instead of seconds since 1/1/1970)

	packettime -= ROOT_TIMEZERO;

	// Fill histograms

	PacketsSent->Fill(packettime);

	// number of hops
	if ((routeid=packet->GetRouteId()) > 0) {
 		nhops=packet->GetNhops();

#define MAXHOPS 30

		Float_t value = (nhops >= MAXHOPS ? -0.02 :  0 ) + (nhops * (maxdelay-mindelay) / MAXHOPS + mindelay);
		RoutesNHops->Fill(packettime, value);
		RoutesInfo->Update(routeid, nhops) ;
	}


	PacketStatus status = packet->Status();
	
	if (status == ClockValid) {
		Float_t packetdelay = packet->GetPacketDelay();
		Delay2D->Fill(packettime, packetdelay);
		Delay1D->Fill(packetdelay);
		ClocksOK->Fill(packettime);

		// in future perhaps: Add(value, interval);
		// where  interval=(packettime mod delta)

		DelayPercentile->Add(packetdelay);

	}
	else if (status == PacketLost) {
		PacketsLost->Fill(packettime);
	}
	else {
		if (allpackets) {
			Float_t packetdelay = packet->GetPacketDelay();
			Delay2D->Fill(packettime, packetdelay);
			Delay1D->Fill(packetdelay);
		}

		if (status == ClockSrcValid) {
			// source ok, target wrong
			BadTargetClock->Fill(packettime);
		}
		else if (status == ClockTrgValid) {
			// target ok, source wrong
			BadSourceClock->Fill(packettime);
		}
		else if (status == ClockInvalid) {
			// both clocks wrong
			BadClocks->Fill(packettime);
		}
		else {
			cerr << "ERROR: unknown packet status " << status
			     << "Source Id "  << packet->GetSourceId()
			     << " Target Id " << packet->GetTargetId() << endl;
		}
	}

}

/*
-------------------------------------------------------------------------------
Subroutine Header
Description	  : 
Input		  :
-------------------------------------------------------------------------------
*/

void TTMDelayHistograms::Plot() {

	// plot the (filled) histograms 
	// files are named <outputDir>/tt<src>tt<dst>.<type>
	// <src> = source id, <dst> = target_id, 
	// <type> = ps, eps, gif

	map<uint,TestBox*> &target = source->GetTargets();

	string pathname, buffer, basename;
	string sourcename = source->GetShortName();
	string targetname = target[targetNum]->GetShortName();

	// filename =  <outputDir>/<sourcename>.<targetname>.<extension>
	// possibly preceded by "IPv6."  

	TPostScript *psout;

	
	gStyle->SetLineScalePS(0.2); // force thin lines in PS output
	gStyle->SetOptStat(0);       // turn off statistics box


	TCanvas *canvas;
   if (outputOption.compare("all") == 0) // Draw old variant of the plot
	   canvas = new TCanvas("canvas", "Plots", 0, 0, 1270, 990);
   else
	   canvas = new TCanvas("canvas", "Plots", 0, 0, 800, 600);

   canvas->SetFillColor(BG_CANVAS);


	basename = "";

	basename += (ipVersion == 6) ?  "IPv6." : "";
	basename += sourcename + "." + targetname;

	if (plotFormat == eps) {
    buffer = basename + ".eps";
		pathname = tmpDir + "/" + buffer;

		// open eps file - type 113 in ROOT
		psout = new TPostScript(pathname.c_str(), 113);
	}
	else {
    buffer = basename + ".ps";
		pathname = tmpDir + "/" + buffer;
		// open ps file  - type 112 = landscape
		psout = new TPostScript(pathname.c_str(), 112);
		psout->Range(25.5, 18.8);   // centre on A4 size paper
	}
	PlotHistos(sourcename, targetname);

  	// now update canvas (needed to write to PS file?)
  	canvas->Update();

	psout->Close();
	delete psout;
	delete canvas;

	if (plotFormat == gif) {
		// now convert them to gif; create ps2gif command line which
		// specificies all .ps files that need converting.
		//
		// BoundingBox parameters are taken from a manual run of the
		// bbfig tool; will be fine as long as ROOT PostScript prolog
		// does not change. 
		//
		// rotate to get proper orientation
		// scale up a bit to make text more legible
		//

		buffer = string(PS2GIF) + " -o " + outputDir + "/" + basename + ".gif" + " " + pathname;

		if (verbose) {
			cerr << "executing " << buffer << endl;
		}

		if (system(buffer.c_str()) != 0) {
			cerr << "ERROR converting to gif: " << pathname << endl;
		}
		else {
			// conversion OK -> cleanup ps (they can be real large)
			if (verbose) {
				cerr << "removing " << pathname << endl;
			}

			if (unlink(pathname.c_str()) != 0) {
				cerr << "ERROR removing ps file? " <<
				pathname <<endl;
			}
		}
	}
	else if (plotFormat == pdf) {
		// now convert them to pdf; create ROOTps2pdf command line which
	 	// specificies all .ps files that need converting.
	 	//
	 	// BoundingBox parameters are taken from a manual run of the
	 	// bbfig tool; will be fine as long as ROOT PostScript prolog
	 	// does not change. 
	 	//
	 	// rotate to get proper orientation
		// scale up a bit to make text more legible
		//

		buffer = string(PS2PDF) + " -o " + outputDir + " " + pathname;
 
		if (verbose) {
			cerr << "executing " << buffer << endl;
		}
 
		if (system(buffer.c_str()) != 0) {
			cerr << "ERROR converting to pdf: " << pathname << endl;
		}
		else {
			// conversion OK -> cleanup ps (they can be real large)
			if (verbose) {
				cerr << "removing " << pathname << endl;
			}
 
			if (unlink(pathname.c_str()) != 0) {
				cerr << "ERROR removing ps file? " <<
				pathname <<endl;
			}
		}
	} 
  else 
  {
		string b = "mv " + pathname + " " + outputDir + "/" + buffer;
		system(b.c_str());
  }
}


/*
-------------------------------------------------------------------------------
Subroutine Header
Purpose           :   Plot the Histograms
Input parameters  :   sourcename - char* for source box
	      targetname - char* for target box
-------------------------------------------------------------------------------
*/

		
void TTMDelayHistograms::PlotHistos(const string &sourcename, const string &targetname) {
	
	char buf[175];
	Int_t bin;
   TPad *pad1;
   TPad *pad2;
   TPad *pad3;
   TPad *pad4;
   TPad *pad5;

   if (outputOption.compare("xml") == 0) {
      CreateXMLSummary(sourcename, targetname);
      return;
   }


   // create pads
   if (outputOption.compare("all") == 0) { // Draw old variant of the plot
      pad1 = new TPad("pad1", "This is pad 1",0.01,0.49,0.41,0.95,BG_PAD);
      pad2 = new TPad("pad2", "This is pad 2",0.42,0.49,0.84,0.95,BG_PAD);
      pad3 = new TPad("pad3", "This is pad 3",0.01,0.02,0.41,0.48,BG_PAD);
      pad4 = new TPad("pad4", "This is pad 4",0.42,0.02,0.84,0.48,BG_PAD);
      pad5 = new TPad("pad5", "This is pad 5",0.85,0.02,0.99,0.95,BG_PAD);
      // force them to be visible on the canvas
      pad1->Draw();
      pad2->Draw();
      pad3->Draw();
      pad4->Draw();
      pad5->Draw();
   } else if (outputOption.compare("1") == 0) {
      pad1 = new TPad("pad1", "This is pad 1",0,0,1,1,BG_PAD);
      pad1->Draw();
   } else if (outputOption.compare("2") == 0) {
      pad2 = new TPad("pad2", "This is pad 2",0,0,1,1,BG_PAD);
      pad2->Draw();
   } else if (outputOption.compare("3") == 0) {
      pad3 = new TPad("pad3", "This is pad 3",0,0,1,1,BG_PAD);
      pad3->Draw();
   } else if (outputOption.compare("4") == 0) {
      pad4 = new TPad("pad4", "This is pad 4",0,0,1,1,BG_PAD);
      pad4->Draw();
   }



   // Top Centre Label
   if (outputOption.compare("all") == 0) { // Draw old variant of the plot
      sprintf(buf, "               Delays from %s to %s.  Start: %s  End: %s UTC",
            sourcename.c_str(), targetname.c_str(), cstarttime1, cendtime1);
      // Added arc to make it look nicer
      TPaveLabel *pl = new TPaveLabel(0.2,0.956,0.8,0.998,buf,"brarc");
      pl->SetFillColor(BG_TITLE);
      pl->SetTextSize(0.45);
      pl->Draw();

      sprintf(buf, "IPv%d", ipVersion);
      //
      //	Only label as IPv4 if box is v6 enabled; requires more info  
      //	from TestBoxConfig --> future extension
      //
      //	TPaveLabel *ip = new TPaveLabel(0.01,0.956,0.1,0.998,buf,"br");
      if (ipVersion == 6) {
         // Added an arc to make it look nicer
         TPaveLabel *ip = new TPaveLabel(0.01,0.956,0.1,0.998,buf,"brarc");
         ip->SetTextColor(9);
         ip->SetTextSize(0.8);
         ip->Draw();
      }
   }

   // Top left
   if (outputOption.compare("all") == 0 || outputOption.compare("1") == 0) {
      pad1->cd();
      pad1->SetGridx();
      pad1->SetGridy();
      if (logscale) {
         pad1->SetLogy();
      }

      // first draw hopcount, to prevent painting over measurement points
      if (!logscale) {
          RoutesNHops->Draw();
          Delay2D->Draw("SAME");
      } else {
          Delay2D->Draw();
      }

      time_t endtime_root = endtime - ROOT_TIMEZERO;
      TGaxis *hopAxis = new TGaxis(endtime_root, mindelay, endtime_root, 
            maxdelay, 0.0, 30.0, 506,"+L");

      hopAxis->SetName("hopAxis");
      hopAxis->SetTitle("Number of hops");
      hopAxis->SetLineColor(2);
      //	hopAxis->SetLabelColor(2);
      hopAxis->SetLabelOffset(0.01);
      if (!logscale) 
          hopAxis->Draw();
   }

   // Top Right
   if (outputOption.compare("all") == 0 || outputOption.compare("2") == 0) {
      pad2->cd();
      pad2->SetGridx();
      pad2->SetGridy();
      if (logscale) {
         pad2->SetLogx();
      }
      Delay1D->Draw();
   }

   // capture satistics *before* adding up Histograms!
   int validpackets = (int) ClocksOK->GetEntries();
   int sourceok = (int) BadTargetClock->GetEntries();
   int targetok = (int) BadSourceClock->GetEntries();
   int badclocks = (int) BadClocks->GetEntries();
   int lostpackets = (int) PacketsLost->GetEntries();

   // Bottom Left
   if (outputOption.compare("all") == 0 || outputOption.compare("3") == 0) {
      pad3->cd();
      pad3->SetGridx();
      pad3->SetGridy();

      // add histograms to create stacked plots (1 bin = multiple data items)

      BadSourceClock->Add(BadSourceClock, ClocksOK, 1, 1);
      BadTargetClock->Add(BadTargetClock, BadSourceClock, 1, 1);
      BadClocks->Add(BadClocks, BadTargetClock, 1, 1);

      // PacketsSent histo is the background, everything not obscured 
      // by valid/invalid clock histograms indicates packet loss.

      // ROOT 5 Autoclips to PacketsSent->GetMinimum()
      PacketsSent->SetMinimum(0);

      PacketsSent->Draw();
      BadClocks->Draw("SAME");
      BadTargetClock->Draw("SAME");
      BadSourceClock->Draw("SAME");
      ClocksOK->Draw("SAME");
   }

   // Bottom Right
   if (outputOption.compare("all") == 0 || outputOption.compare("4") == 0) {
      pad4->cd();
      pad4->SetGridx();
      pad4->SetGridy();

      // XXX FIXME Apparently ROOT 5 doesn't like the way we treat our histograms
      // Usiing
      //Arrived->Add(PacketsSent, PacketsLost, 1, -1);
      //Arrived->Divide(Arrived, PacketsSent, 1, 1);
      // Will double the amount of bins in Arrived. This is a mathematical equal work around

      Arrived->Reset();
      ArrivedInvalid->Reset();
      for(bin=1; bin<=binsX; bin++) {
         if( float(PacketsSent->GetBinContent(bin)) == 0 ) { // We don't have enough data for that bin, so show it in a different color.
            ArrivedInvalid->SetBinContent(bin, 1 );
            Arrived->SetBinContent(bin, 0);
         } else {
            ArrivedInvalid->SetBinContent(bin, 0);
            Arrived->SetBinContent(bin, float(1*PacketsSent->GetBinContent(bin) + -1*PacketsLost->GetBinContent(bin)) / float(PacketsSent->GetBinContent(bin)) );
         }
      }
      // ROOT 5 Autoclips to Arrived->GetMinimum()
      Arrived->SetMinimum(0);
      ArrivedInvalid->SetMinimum(0);
      Arrived->Draw();
      ArrivedInvalid->Draw("SAME");
   }

   // Statistics
   if (outputOption.compare("all") == 0) { // Only print for old plots
      pad5->cd();

      // 

      // percentiles for packet loss (using fraction arrived per bin)
      LossPercentile->Reset();
      for(bin=1; bin<=binsX; bin++) {
         LossPercentile->Add(float(1-Arrived->GetBinContent(bin)));
      }

      // write output with TLatex-class in new root-version  
      strcpy(buf, "STATISTICS:");
      Float_t x = 0.5;       // Starting point
      Float_t y = 0.95;
      Int_t align = 22;      // horizontally and vertically centered
      Font_t font = TATTTEXT_HELVETICA_BOLD_R_NORMAL;
      write_stats(pad5, buf, x, y, align, font);

      y -=0.03;
      //	write_stats(pad5, "Histogram", x, y, align, font);
      //	y -=0.02;
      write_stats(pad5, "Delay @& Hops:", x, y, align, font);

      font = TATTTEXT_HELVETICA_MEDIUM_R_NORMAL;	
      y -=0.025;
      sprintf(buf,     "Entries: %g", Delay1D->GetEntries());
      write_stats(pad5, buf, x, y, align, font);
      y -=0.02;
      sprintf(buf,     "Overflow: %g", Delay1D->GetBinContent(binsY+1) );
      write_stats(pad5, buf, x, y, align, font);
      y -=0.02;
      sprintf(buf,     "Underflow: %g", Delay1D->GetBinContent(0) );
      write_stats(pad5, buf, x, y, align, font);

      align = 32;
      Float_t x1 = 0.52;
      Float_t x2 = 0.95;

      y -=0.025;
      write_stats(pad5, "2.5 Perc:", x1, y, align, font);
      sprintf(buf,     "%8.1fms", DelayPercentile->GetLevel(2.5));
      write_stats(pad5, buf, x2, y, align, font);

      y -=0.02;
      write_stats(pad5, "Median:", x1, y, align, font);
      sprintf(buf,     "%8.1fms", DelayPercentile->GetLevel(50.0));
      write_stats(pad5, buf, x2, y, align, font);

      y -=0.02;
      write_stats(pad5, "97.5 Perc:", x1, y, align, font);
      sprintf(buf,     "%8.1fms", DelayPercentile->GetLevel(97.5));
      write_stats(pad5, buf, x2, y, align, font);

      y -=0.02;
      write_stats(pad5,"Mean:", x1, y, align, font);
      sprintf(buf,     "%8.1fms", Delay1D->GetMean());
      write_stats(pad5, buf, x2, y, align, font);
      y -=0.02;
      write_stats(pad5,"RMS:", x1, y, align, font);
      sprintf(buf,     "%8.1fms", Delay1D->GetRMS());
      write_stats(pad5, buf, x2, y, align, font);

      align=22;
      y -=0.025;
      sprintf(buf,     "Min. hops: %d", RoutesInfo->GetMinHops());
      write_stats(pad5, buf, x, y, align, font);
      y -=0.02;
      sprintf(buf,     "Max. hops: %d", RoutesInfo->GetMaxHops());
      write_stats(pad5, buf, x, y, align, font);

      y -=0.03;
      TLine line1(0.0,0.0,0.0,0.0);
      line1.SetLineWidth(2);
      line1.SetLineColor(4);
      line1.DrawLine(0.05, y, 0.95, y); 

      y -=0.04;
      font = TATTTEXT_HELVETICA_BOLD_R_NORMAL;     
      write_stats(pad5, "Packets Sent/Valid:", x, y, align, font);
      y -=0.025;

      font = TATTTEXT_HELVETICA_MEDIUM_R_NORMAL;	
      int totalsent = (int) PacketsSent->GetEntries();
      sprintf( buf,     "Total: %i", totalsent);
      write_stats(pad5, buf, x, y, align, font);
      y -=0.02;

      float percent =  validpackets * 100.0 / totalsent;
      sprintf( buf,     "Valid: %i = %.3g %%", validpackets, percent);
      write_stats(pad5, buf, x, y, align, font);

      y -=0.02;
      percent = targetok * 100.0 / totalsent;
      sprintf( buf,     "Bad Sent: %i = %.2g %%", targetok, percent);
      write_stats(pad5, buf, x, y, align, font);

      y -=0.02;
      percent = sourceok * 100.0 / totalsent;
      sprintf( buf,     "Bad Recv: %i = %.2g %%", sourceok, percent);
      write_stats(pad5, buf, x, y, align, font);

      y -=0.02;
      percent = badclocks * 100.0 / totalsent;
      sprintf( buf,     "2 Clocks bad: %i = %.2g %%", badclocks, percent);
      write_stats(pad5, buf, x, y, align, font);

      y -=0.02;
      percent = lostpackets * 100.0 / totalsent;
      sprintf( buf,     "Lost: %i = %.2g %%", lostpackets, percent);
      write_stats(pad5, buf, x, y, align, font);

      y -=0.03;
      line1.SetLineColor(2);
      line1.DrawLine(0.05, y, 0.95, y);  
      y -=0.04;

      font = TATTTEXT_HELVETICA_BOLD_R_NORMAL;     
      write_stats(pad5,"Packets lost:", x, y, align, font);

      font = TATTTEXT_HELVETICA_MEDIUM_R_NORMAL;	
      align = 32;	// right justified
      x1 += 0.03;
      x2 -= 0.03;

      y -=0.025;
      write_stats(pad5, "2.5 Perc:", x1, y, align, font);
      sprintf(buf,     "%5.1f%%", 100 * LossPercentile->GetLevel(2.5));
      write_stats(pad5, buf, x2, y, align, font);
      y -=0.02;
      write_stats(pad5, "Median:", x1, y, align, font);
      sprintf(buf,     "%5.1f%%", 100 * LossPercentile->GetLevel(50.0));
      write_stats(pad5, buf, x2, y, align, font);
      y -=0.02;
      write_stats(pad5, "97.5 Perc:", x1, y, align, font);
      sprintf(buf,     "%5.1f%%", 100 * LossPercentile->GetLevel(97.5));
      write_stats(pad5, buf, x2, y, align, font);


      // estimate of sending process uptime
      int count = 0;

      for (bin = 1; bin <= binsX; bin++) {
         if (PacketsSent->GetBinContent(bin) > 0) {
            count++;
         }
      }
      float uptime = count*100/(float)binsX;

      y -=0.025;
      align = 22;	// centered
      sprintf(buf,     "Uptime: %.3g %%", uptime);
      write_stats(pad5, buf, x, y, align, font);

      y -=0.03;
      line1.SetLineColor(3);
      line1.DrawLine(0.05, y, 0.95, y); 

      y -=0.04;
      font = TATTTEXT_HELVETICA_BOLD_R_NORMAL; 
      write_stats(pad5, "Over-all statistic:", x, y, align, font);

      font = TATTTEXT_HELVETICA_MEDIUM_R_NORMAL;
      y -=0.01;  
      //	y -=0.02;  
      //	write_stats(pad5, "Number of measurements", x, y, align, font);
      //	y -=0.02;
      //	sprintf( buf,     "in period: %i", nMeasurements);
      //	write_stats(pad5, buf, x, y, align, font);

      y -=0.02;
      Float_t days = ((endtime-starttime)/86400);
      if (days > 1) {
         sprintf( buf,     "Time period: %.2g days", days);
      }
      else {
         sprintf( buf,     "Time period: %.2g day",  days);
      }
      write_stats(pad5, buf, x, y, align, font);
      y -=0.035;
      write_stats(pad5, "Number of routing", x, y, align, font);
      y -=0.02;
      sprintf( buf,     "vVectors: %i", RoutesInfo->GetEntries());
      write_stats(pad5, buf, x, y, align, font);
      y -=0.02;
      sprintf( buf,     "Flaps: %i", RoutesInfo->GetNumFlaps());
      write_stats(pad5, buf, x, y, align, font);
      y -=0.035;
      sprintf( buf,     "Number of bins: %d", binsX );
      write_stats(pad5, buf, x, y, align, font);
      y -=0.02;
      sprintf( buf,     "Minutes/bin: %.3g", ((endtime-starttime)/binsX/60.0));
      write_stats(pad5, buf, x, y, align, font);
   }
}


void TTMDelayHistograms::CreateXMLSummary(const string &sourcename, const string &targetname) {
   Int_t bin;

   // percentiles for packet loss (using fraction arrived per bin)
   LossPercentile->Reset();
   for(bin=1; bin<=binsX; bin++) {
      LossPercentile->Add(float(1-Arrived->GetBinContent(bin)));
   }

   // capture satistics *before* adding up Histograms!
   int validpackets = (int) ClocksOK->GetEntries();
   int sourceok = (int) BadTargetClock->GetEntries();
   int targetok = (int) BadSourceClock->GetEntries();
   int badclocks = (int) BadClocks->GetEntries();
   int lostpackets = (int) PacketsLost->GetEntries();



   // write output with TLatex-class in new root-version  
   cout << "<statistics source=\"" << sourcename << "\" target=\"" << targetname << "\">";

   cout << "<delayHops>";
   cout << "<entries>" << Delay1D->GetEntries() << "</entries>";
   cout << "<overflow>" << Delay1D->GetBinContent(binsY+1) << "</overflow>";
   cout << "<underflow>" << Delay1D->GetBinContent(0) << "</underflow>";
   cout << "<percentile type=\"delay\" level=\"2.5\">" << DelayPercentile->GetLevel(2.5) << "</percentile>";
   cout << "<percentile type=\"delay\" level=\"50\">" << DelayPercentile->GetLevel(50.0) << "</percentile>";
   cout << "<percentile type=\"delay\" level=\"97.5\">" << DelayPercentile->GetLevel(97.5) << "</percentile>";
   cout << "<mean>" << Delay1D->GetMean() << "</mean>";
   cout << "<rms>" << Delay1D->GetRMS() << "</rms>";
   cout << "<minHops>" << RoutesInfo->GetMinHops() << "</minHops>";
   cout << "<maxHops>" << RoutesInfo->GetMaxHops() << "</maxHops>";
   cout << "</delayHops>";

   int totalsent = (int) PacketsSent->GetEntries();
   cout << "<sentValid>";
   cout << "<total>" << totalsent << "</total>";
   cout << "<valid>" << validpackets * 100.0 / totalsent << "</valid>";
   cout << "<badSent>" << targetok * 100.0 / totalsent << "</badSent>";
   cout << "<badRecv>" << sourceok * 100.0 / totalsent << "</badRecv>";
   cout << "<badClocks>" << badclocks * 100.0 / totalsent << "</badClocks>";
   cout << "<lost>" << lostpackets * 100.0 / totalsent << "</lost>";
   cout << "</sentValid>";

   cout << "<packetsLost>";
   cout << "<percentile type=\"loss\" level=\"2.5\">" << LossPercentile->GetLevel(2.5) << "</percentile>";
   cout << "<percentile type=\"loss\" level=\"50\">" << LossPercentile->GetLevel(50) << "</percentile>";
   cout << "<percentile type=\"loss\" level=\"97.5\">" << LossPercentile->GetLevel(97.5) << "</percentile>";
   // estimate of sending process uptime
   int count = 0;

   for (bin = 1; bin <= binsX; bin++) {
      if (PacketsSent->GetBinContent(bin) > 0) {
         count++;
      }
   }
   float uptime = count*100/(float)binsX;
   cout << "<uptime>" << uptime << "</uptime>";
   cout << "</packetsLost>";

   cout << "<general>";
   cout << "<period>" << (endtime-starttime)/86400 << "</period>";
   cout << "<vectors>" << RoutesInfo->GetEntries() << "</vectors>";
   cout << "<flaps>" << RoutesInfo->GetNumFlaps() << "</flaps>";
   cout << "</general>";
}

/*
-------------------------------------------------------------------------------
Subroutine Header
Description	  : 
Input		  :
-------------------------------------------------------------------------------
*/

void TTMDelayHistograms::CreateHistos() {
	char	buf[100];     // small buffer for character strings
	char    xtitle[100];  // buffer for x-axis title
	string	timeformat;  // indicates how time on axis is formatted



	
	binsY = 250; 

	// Some initialisation based on time period

	int ndiv = 0;
	int ndays = (endtime - starttime) / 3600 / 24;
	if (ndays>3)  {
		timeformat = "%b %d";      // <month> <day> 
		strcpy(xtitle, "Time [month day]");
		if (ndays>7) {
			ndiv = -506;	  // 6 major divisions
		}
	}
	else {
		timeformat = "%H:%M";	 //  <hour>:<minute>
		sprintf(xtitle, "      %s%26cTime [hour:minute]%26c%s", cstarttime2,
			' ', ' ', cendtime2);
	}

	// transform UNIX timestamps into ROOT timestamps
	// (seconds since 1/1/1995 instead of seconds since 1/1/1970)

	time_t endtime_root = endtime - ROOT_TIMEZERO;
	time_t starttime_root = starttime - ROOT_TIMEZERO;


	RoutesInfo      = new RoutingInfo;
	DelayPercentile = new Percentiles;


	sprintf(buf, "Delay2D[%d]", targetNum);
	Delay2D = new TH2F(buf, "Delay vs. Time", binsX*2, starttime_root, 			endtime_root, binsY*2, mindelay, maxdelay);

	// Histogram of delay values
	// Note the size binsY is also used in Plot() method!
	sprintf(buf, "Delay1D[%d]", targetNum);
	Delay1D = new TH1F(buf, "Packet Delay", binsY, mindelay, maxdelay);
   Delay1D->SetLabelColor(TEXT_LABEL);

	sprintf(buf, "PacketsSent[%d]", targetNum);
 	PacketsSent = new TH1F(buf, "Packets Sent/Valid", binsX, 
		starttime_root, endtime_root);
   PacketsSent->SetLabelColor(TEXT_LABEL);

	sprintf(buf, "ClocksOK[%d]", targetNum);
 	ClocksOK = new TH1F(buf, "Clocks OK", binsX,
		starttime_root, endtime_root);
   ClocksOK->SetLabelColor(TEXT_LABEL);

	sprintf(buf, "BadSourceClock[%d]", targetNum);
	BadSourceClock = new TH1F(buf, "Bad Source Clock", binsX,
		starttime_root, endtime_root);
   BadSourceClock->SetLabelColor(TEXT_LABEL);

	sprintf(buf, "BadTargetClock[%d]", targetNum);
	BadTargetClock = new TH1F(buf, "Bad Target Clock", binsX,
		starttime_root, endtime_root);
   BadTargetClock->SetLabelColor(TEXT_LABEL);

	sprintf(buf, "BadClocks[%d]", targetNum);
	BadClocks = new TH1F(buf, "Bad Clocks", binsX,
		starttime_root, endtime_root);
   BadClocks->SetLabelColor(TEXT_LABEL);

	// hops
	sprintf(buf, "RoutesNHops[%d]", targetNum);
  	RoutesNHops = new TH2F(buf, "Packet Delay / Hopcount",
			binsX*2, starttime_root, endtime_root,
			binsY*2, mindelay, maxdelay);
   RoutesNHops->SetLabelColor(TEXT_LABEL);

	// lost
	sprintf(buf, "PacketsLost[%d]", targetNum);
  	PacketsLost = new TH1F(buf, "Packets Arrived/Lost", binsX,
			starttime_root, endtime_root);
   PacketsLost->SetLabelColor(TEXT_LABEL);


	// Now Set various histograms attributes (persistent over reset)

	Delay1D->SetFillColor(GRAPH_DELAY1D_LINE); // yellow
	Delay1D->SetXTitle("Delay [msec]");
	Delay1D->SetYTitle("# Packets [Entries/Bin]");
   Delay1D->SetTitleOffset(1.4, "Y");
   Delay1D->SetTitleSize(0.035, "Y");
   Delay1D->SetLabelSize(0.035, "Y");

	// Force division of "delay" scale in 1D and 2D histograms
	// ROOT 3.0 default behaviour is not good

	Delay1D->SetNdivisions(-505, "x");
	RoutesNHops->SetNdivisions(-505, "y");

	RoutesNHops->SetMarkerSize(2);
	RoutesNHops->SetMarkerColor(GRAPH_ROUTES_N_HOPS);
	RoutesNHops->SetXTitle(xtitle);
	RoutesNHops->SetYTitle("Delay [msec]");
   RoutesNHops->SetTitleOffset(1.2, "Y");
	if (ndiv != 0) {
			RoutesNHops->SetNdivisions(ndiv,"x"); 
	}
	RoutesNHops->GetXaxis()->SetTimeDisplay(1);
	RoutesNHops->GetXaxis()->SetTimeFormat(timeformat.c_str());

	PacketsSent->SetFillColor(GRAPH_PACKET_SENT); // cyan
	PacketsSent->SetLineColor(GRAPH_STD_LINE); // black
	PacketsSent->SetXTitle(xtitle);
	PacketsSent->SetYTitle("# Packets [Entries/Bin]");
   PacketsSent->SetTitleOffset(1.1, "Y");
	if (ndiv != 0) {
			PacketsSent->SetNdivisions(ndiv,"x"); 
	}
	PacketsSent->GetXaxis()->SetTimeDisplay(1);
	PacketsSent->GetXaxis()->SetTimeFormat(timeformat.c_str());

	BadClocks->SetFillColor(GRAPH_BAD_BLOCKS); // red
	BadClocks->SetLineColor(GRAPH_STD_LINE); // black

	BadTargetClock->SetFillColor(GRAPH_BAD_TARGET_CLOCK); // yellow
	BadTargetClock->SetLineColor(GRAPH_STD_LINE); // black

	BadSourceClock->SetFillColor(GRAPH_BAD_SOURCE_CLOCK);  // magenta
	BadSourceClock->SetLineColor(GRAPH_STD_LINE);  // black

	ClocksOK->SetFillColor(GRAPH_CLOCKS_OK); // green
	ClocksOK->SetLineColor(GRAPH_STD_LINE); // black


	// finally two auxilary histograms, one for arrived/lost ratio and one to visualise bins without sufficient data
	sprintf(buf, "Arrived[%d]", targetNum);
	Arrived = new TH1F(buf, "Packets Arrived/Lost", binsX,
			starttime_root, endtime_root);   

	Arrived->SetXTitle(xtitle);
	Arrived->SetYTitle("Fraction Arrived");
   Arrived->SetTitleOffset(1.1, "Y");
	Arrived->SetFillColor(GRAPH_ARRIVED);      // green
	if (ndiv != 0) {
		Arrived->SetNdivisions(ndiv,"x"); 
	}
	Arrived->GetXaxis()->SetTimeDisplay(1);
	Arrived->GetXaxis()->SetTimeFormat(timeformat.c_str());

	sprintf(buf, "InvalidBins[%d]", targetNum);
	ArrivedInvalid = new TH1F(buf, "InvalidBins", binsX,
			starttime_root, endtime_root);   

	ArrivedInvalid->SetFillColor(GRAPH_INVALID);
	ArrivedInvalid->SetFillColor(GRAPH_INVALID);
}

/*
-------------------------------------------------------------------------------
Subroutine Header
Description	  : 
Input		  :
-------------------------------------------------------------------------------
*/

void TTMDelayHistograms::ClearHistos () {
	// Reset all histograms and percentiles

	Delay2D->Reset();
	Delay1D->Reset();
	PacketsSent->Reset();
	ClocksOK->Reset();
	BadSourceClock->Reset();
	BadTargetClock->Reset();
	BadClocks->Reset();
	RoutesNHops->Reset();
	PacketsLost->Reset();

	RoutesInfo->Reset();
	DelayPercentile->Reset();	
}

/*
-------------------------------------------------------------------------------
Subroutine Header
Description	  : 
Input		  :
-------------------------------------------------------------------------------
*/

void TTMDelayHistograms::DeleteHistos () {
	// delete allocated histograms, reset pointers to NULL, scalars to 0

	if (source != NULL) {
		// if source==NULL, no histograms exist
	}
	source = NULL;
}

/*
-------------------------------------------------------------------------------
Subroutine Header
Description	  : Constructor for TTMHistogram objects
Input		  :
-------------------------------------------------------------------------------
*/

TTMDelayHistograms::TTMDelayHistograms(TestBox *src, int tgtnum, TTMOptions* opt) : TTMHistograms(src,tgtnum,opt) {

	// default construtor
	// initialize everything to zero or NULL

	Delay2D		= NULL;
	Delay1D		= NULL;
	PacketsSent	= NULL;
	ClocksOK	= NULL;
	BadSourceClock	= NULL;
	BadTargetClock	= NULL;
	BadClocks	= NULL;
	RoutesNHops	= NULL;
	PacketsLost	= NULL;
	Arrived		= NULL;

	DelayPercentile = NULL;

	// defaults values for delay histogram parameters
	maxdelay = 250;
	binsX = 168; 	 // week plot: 1 bin per 60 minutes

	LossPercentile = new Percentiles;

	mindelay   = opt->MinDelay();
	maxdelay   = opt->MaxDelay();
	logscale   = opt->Logscale();
	allpackets = opt->AllPackets();
   outputOption = opt->OutputOption();

   // For small timeperiodes make the number of bins smaller as well.
   binsX = opt->TimePeriod() / 150;
   if( binsX > 168 )
      binsX = 168;
   if( binsX < 1 )
      binsX = 1;

   CreateHistos();
}

/*
-------------------------------------------------------------------------------
Subroutine Header
Description	  : Destructor for TTMHistogram objects
		    delete all dynamically created data
-------------------------------------------------------------------------------
*/

TTMDelayHistograms::~TTMDelayHistograms() {

	if (verbose && (source != NULL)) {
		cout << "destructing TTMDelayHistograms object " << endl;
	}
	delete Delay2D;
	delete Delay1D;
	delete PacketsSent;
	delete ClocksOK;
	delete BadSourceClock;
	delete BadTargetClock;
	delete BadClocks;
	delete RoutesNHops;
	delete PacketsLost;
	
	delete RoutesInfo;
	delete DelayPercentile;	
	delete Arrived;
	delete ArrivedInvalid;
	delete LossPercentile;
}
