/*
Copyright (c) 2001		     RIPE NCC


All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of the author not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS; IN NO EVENT SHALL
AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY
DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

/*
-------------------------------------------------------------------------------
Module Header
Filename	  : TTMOptions.C
Author	    	  : Rene Wilhelm
Date	      	  : 31-OCT-2001
Description       : TTMOptions class implemenation
Language Version  : C++
OSs Tested        : Solaris 2.6, Solaris 8, Debian Linux 2.2
TODO		  : Implement destructor (nice to clean up)
		    Implement dynamic help function

$Id: TTMOptions.C,v 1.17 2008/04/09 13:06:55 ruben Exp $
-------------------------------------------------------------------------------
*/

#include <cmath>
#include <iostream>
#include <sstream>
#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <ctime>

#include <unistd.h>
#include <sysexits.h>

using namespace std;

#include "TTMOptions.h"
#include "utils.h"


RCSID("$Id: TTMOptions.C,v 1.17 2008/04/09 13:06:55 ruben Exp $");


/*
-------------------------------------------------------------------------------
Subroutine Header
Purpose	   	  :   Initializing constructor
Input 		  :   argument vector, option string (getopt(3) syntax)
Comments	  :   This constructor assigns default values to all
		      parameters, and then parses the argument vector,
		      setting optional parameters to the provided values
-------------------------------------------------------------------------------
*/

TTMOptions::TTMOptions(int argc, char** argv, const char* optstring) 
{

	extern char        *optarg;
	char                buffer[50];
	char                ch;

	sourceId = 0;
	targetId = 0;
	endTime = 0;
	timePeriod = 0;
	ipVersion = 4;
	inputPath = "/ncc/ttpro/data/root";
	outputPath = "/ncc/ttpro/data/plots";
   outputOption = "all";
	tmpDir = "/var/tmp";
	minDelay = 0;
	maxIPDV = maxDelay = 250;

	debug = verbose = checkdata = logScale = allPackets = dateInFilename = enableErrorPlot = savePercentiles = createHTML = customMaxIPDV = customMinDelay =
	customMaxDelay = false;

	storageType = 'H';	// hierarchical
	sumPeriod = day;
	optionString = optstring;

	plotFormat = gif;

	progName = argv[0];

	// ensure GMT/UTC for time operations
	// the const_cast is to remove a compiler warning
	putenv (const_cast< char* >("TZ=GMT+0"));	

	bool               cflag = false;
	bool               fflag = false;
	bool               mflag = false;
	bool               oflag = false;
	bool               Oflag = false;
	bool               Rflag = false;
	bool               pflag = false;
	bool               sflag = false;
	bool               tflag = false;
	bool               uflag = false;
	bool               v4flag = false;
	bool               v6flag = false;
	bool               zflag = false;

	char               *endpointer;

	while ((ch = getopt (argc, argv, optstring)) != -1)

		switch (ch)
		{
		case '4':
			if (v6flag)
			{
				usage ();	/* only either one of -4 or -6
						 * allowed */
			}
			ipVersion = 4;
			v4flag = true;
			break;
		case '6':
			if (v4flag)
			{
				usage ();	/* only either one of -4 or -6
						 * allowed */
			}
			ipVersion = 6;
			v6flag = true;
			break;
		case 'a':
			allPackets = true;
			break;
		case 'c':
			checkdata = true;
			break;
		case 'D':
			dateInFilename = true;
			break;
		case 'e':
			enableErrorPlot = true;
			break;
		case 'f':
			if (fflag)
			{
				usage ();	/* only one -f allowed */
			}

			if (strcmp (optarg, "csv") == 0)
			{
				plotFormat = csv;
			}
			else if (strcmp (optarg, "gif") == 0)
			{
				plotFormat = gif;
			}
			else if (strcmp (optarg, "ps") == 0)
			{
				plotFormat = ps;
			}
			else if (strcmp (optarg, "eps") == 0)
			{
				plotFormat = eps;
			}
			else if (strcmp (optarg, "pdf") == 0)
			{
				plotFormat = pdf;
			}
			else
			{
				usage ();
			}
			fflag = true;
			break;
		case 'H':
			createHTML = true;
			break;
		case 'l':
			logScale = true;
			break;
		case 'm':
			if (!mflag)
			{
				customMaxDelay = mflag = true;
				maxDelay = strtod (optarg, &endpointer);
				if (*endpointer != '\0')
				{
					usage ();
				}
				if (maxDelay < 0)
				{
					cerr << progName << ": -m positive value expected" << endl;
					exit (EX_USAGE);
				}
				else if ((maxDelay == HUGE) || (maxDelay == HUGE_VAL))
				{
					cerr << progName << ": -m value out of range" << endl;
					exit (EX_USAGE);
				}
			}
			else
			{
				/* only one -m allowed */
				usage ();
			}
			break;
		case 'o':
			if (!oflag)
			{
				outputPath = optarg;
				oflag = true;

				/* truncate trailing / */
				while (*(outputPath.rbegin()) == '/' && outputPath.size() > 1)
					outputPath.resize(outputPath.size()-1);
				if (outputPath.size() == 0)
					outputPath = ".";
			}
			else
			{
				/* only one -o allowed */
				usage ();
			}
			break;
      case 'O':
         if (!Oflag)
         {
            outputOption = optarg;
            Oflag = true;
         }
         else
         {
				/* only one -O allowed */
            usage ();
         }
         break;
		case 'p':
			if (!pflag)
			{
				pflag = true;
				decodePeriod (optarg, endTime, timePeriod);
			}
			else
			{
				/* only one -p allowed */
				usage ();
			}
			break;
		case 'r':
			savePercentiles = true;
			break;
		case 'R':
			if (!Rflag)
			{
				inputPath = optarg;
				Rflag = true;

				/* truncate trailing / */
				while (*(inputPath.rbegin()) == '/' && inputPath.size() > 1)
					inputPath.resize(inputPath.size()-1);
				if (inputPath.size() == 0)
					inputPath = ".";
			}
			else
			{
				/* only one -R allowed */
				usage ();
			}
			break;
		case 's':
			if (!sflag)
			{
				sflag = true;
				if (sscanf (optarg, "%[0-9]", buffer) != 1)
				{
					usage ();
				}
				sourceId = atoi (buffer);

			}
			else
			{
				/* only one -s allowed */
				usage ();
			}
			break;
		case 't':
			if (!tflag)
			{
				tflag = true;
				if (sscanf (optarg, "%[0-9]", buffer) != 1)
				{
					usage ();
				}
				targetId = atoi (buffer);
			}
			else
			{
				/* only one -t allowed */
				usage ();
			}
			break;
		case 'u':
			if (!uflag)
			{
				customMaxIPDV = uflag = true;
				maxIPDV = strtod (optarg, &endpointer);
				if (*endpointer != '\0')
				{
					usage ();
				}
				if (maxIPDV < 0)
				{
					cerr << progName << ": -u positive value expected" << endl;
					exit (EX_USAGE);
				}
				else if ((maxIPDV == HUGE) || (maxIPDV == HUGE_VAL))
				{
					cerr << progName << ": -u value out of range" << endl;
					exit (EX_USAGE);
				}
			}
			else
			{
				/* only one -u allowed */
				usage ();
			}
			break;
		case 'd':
			debug = true;
			break;
		case 'v':
			verbose = true;
			break;
		case 'w':
			sumPeriod = week;
			break;
		case 'z':
			if (!zflag)
			{
				customMinDelay = zflag = true;
				minDelay = strtod (optarg, &endpointer);
				if (*endpointer != '\0')
				{
					usage ();
				}
				if (minDelay < 0 && !getenv("TTM_NEGATIVE_DELAYS") )
				{
					cerr << progName << ": -z positive value expected" << endl;
					exit (EX_USAGE);
				}
				else if ((minDelay == HUGE) || (minDelay == HUGE_VAL))
				{
					cerr << progName << ": -z value out of range" << endl;
					exit (EX_USAGE);
				}
			}
			else
			{
				/* only one -z allowed */
				usage ();
			}
			break;
		case 'h':
		default:
			usage ();
		}

	if (!pflag || (tflag && !sflag))
	{
		/* time period is only required argument */
		/* specific target requires source to be specified as well */
		usage ();
	}

	if (!cflag)
	{
		// construct default config command:
		// "TTCONFIG_COMMAND -d <enddate> TTCONFIG_COMMAND_ARGS"

		char strftime_buf[26];

		configCmd = TTCONFIG_COMMAND;

		if (strftime (strftime_buf, 26, "'%Y-%m-%d %H:%M:%S'", localtime (&endTime)) == 0)
		{
			cerr << progName << ": strftime(): ERROR converting time " << (int) endTime << endl;
		}
		else
			configCmd += string(" -d ") + strftime_buf + string(" ");

		if (ipVersion == 6)
		{
			configCmd += TTCONFIG_COMMAND_ARGS6;
		}
		else
		{
			configCmd += TTCONFIG_COMMAND_ARGS;
		}
	}
}

/*
-------------------------------------------------------------------------------
Subroutine Header
Purpose	  	:   TTMOptions destructor
Comments	:
-------------------------------------------------------------------------------
*/

/*
-------------------------------------------------------------------------------
Subroutine header
Purpose         :   Error handler
Input           :   none
Output          :   error message
Side effects    :   terminates program
-------------------------------------------------------------------------------
 */

void TTMOptions::invaliddate() {
	cerr << progName << ": invalid date/time specification" << endl;
	exit(EX_USAGE);
}

/*
-------------------------------------------------------------------------------
Subroutine header

Purpose          :   decode ascii ISO8601 time period string into an
		     ending time in time_t format (seconds since 1970)
	  	     and a duration
Input            :   char  *datestring  string to be decoded
Output           :   time_t& endtime     system timestamp of end time
	     time_t& duration    duration in seconds

-------------------------------------------------------------------------------
*/

void TTMOptions::decodePeriod (string datestring, time_t& endtime, time_t& duration) {
	string::size_type part1, part2;
	char flag1, flag2;
	time_t time1, time2;

	/* datestring =  ISO8601 time period string */
	/* two parts seperated by /  */


	part1 = 0;
	if ((part2 = datestring.find('/')) == string::npos ) {
		invaliddate();
	}

	parseISO8601 (datestring.substr(part1,part2), time1, flag1);
	parseISO8601 (datestring.substr(part2+1), time2, flag2);

	if ((flag1 == 'T') && (flag2 == 'D')) {
		/* format: starttime/duration */
		duration = time2;
		endtime = time1 + duration;
	}
	else if ((flag1 == 'D') && (flag2 == 'T')) {
		/* format: duration/endtime */
		duration = time1;
		endtime = time2;
	}
	else if ((flag1 == 'T') && (flag2 == 'T')) {
		/* format: starttime/endtime */
		if (time1 >= time2) {
			/* negative or zero time interval */
			invaliddate();
		}
		else {
			endtime = time2;
			duration = time2 - time1;
		}
	}
	else {
		invaliddate();
	}

}

/*
-------------------------------------------------------------------------------
Subroutine header

Purpose	  	:  decode ascii time string into duration of time or
		   fixed point in time; the string should be in the
		   formats described by ISO8601 standard

NOTE		:  to not complicate the routine too much, points in time
		   MUST be complete representations of either a date (CCYYMMDD)
		   or a date+time (CCYYMMDDTHHMMSS).
	           - and : are allowed as separators.

		   similarly, durations are only supported when specified
		   in days, hours, minutes and seconds. Length of months
		   and years are content dependant.

Input		:  char*   text  string to be decoded
Output		:  time_t& time  system timestamp or duration  (seconds)
	   char&   flag  indicates if returned time is timestamp
		   or duration

-------------------------------------------------------------------------------
*/

#define DAYFLAG    0x001000
#define DHMSFLAG   0x001111
#define HOURFLAG   0x000100
#define HMSFLAG    0x000111
#define MINFLAG    0x000010
#define MSFLAG     0x000011
#define SECFLAG    0x000001

void TTMOptions::parseISO8601 (string datestring, time_t& isotime, char& flag) {
	string::iterator c;
	int num;

	struct tm  tmstruct;

	int year     = 0;
	int month    = 0;
	int seconds  = 0;
	int minutes  = 0;
	int hours    = 0;
	int days     = 0;

	int dateflags = 0;   /* flag which date component we've seen */

	memset(&tmstruct,0,sizeof(tmstruct));

	c = datestring.begin();
	isotime = 0;

	if (*c++ == 'P') {
		/* duration */
		flag = 'D';
		while (c != datestring.end()) {
			num = 0;
			while (*c >= '0' && *c <= '9') {
				/* assumes ASCII sequence! */
				num = 10 * num + *c++ - '0';
			}

			switch (*c++) {
			case 'D':
				if (dateflags & DHMSFLAG) {
					/* day, hour, min or sec already set */
					invaliddate();
				}
				else {
					dateflags |= DAYFLAG;
					days = num;
				}
				break;
			case 'H':
				if (dateflags & HMSFLAG) {
					/* hour, min or sec already set */
					invaliddate();
				}
				else {
					dateflags |= DAYFLAG;
					hours = num;
				}
				break;
			case 'M':
			  	if (dateflags & MSFLAG) {
					/* min or sec already set */
					invaliddate();
				}
				else {
					dateflags |= MINFLAG;
					minutes = num;
			  	}
				break;
			case 'S':
				if (dateflags & SECFLAG) {
					/* sec already set */
					invaliddate();
				}
				else {
					dateflags |= SECFLAG;
					seconds = num;
				}
				break;
			default:
				invaliddate();
			}
		}
		isotime = seconds + 60 * minutes + 3600 * hours +
			   SECONDSPERDAY * days;
	}
	else {
		/* point in time, must be one of
			CCYYMMDD
			CCYY-MM-DD
			CCYYMMDDTHHMM
			CCYY-MM-DDTHH:MM
			CCYYMMDDTHHMMSS
			CCYY-MM-DDTHH:MM:SS
		*/
		c = datestring.begin();
		flag = 'T';

		/* NOTE: we have to check for the extended format first,
		   because otherwise the separting '-' will be interpreted
		   by sscanf as signs of a 1 digit integer .... :-(  */

		if (sscanf(datestring.c_str(), "%4u-%2u-%2u", &year, &month, &days) == 3) {
			c += 10;
		}
		else if (sscanf(datestring.c_str(), "%4u%2u%2u", &year, &month, &days) == 3) {
			c += 8;
		}
		else {
			invaliddate();
		}

		tmstruct.tm_year = year - 1900;
		tmstruct.tm_mon  = month - 1;
		tmstruct.tm_mday = days;

		if (*c == '\0') {
			tmstruct.tm_hour = 0;
			tmstruct.tm_sec = 0;
			tmstruct.tm_min = 0;
			isotime = mktime (&tmstruct);
		}
		else if (*c == 'T') {
			/* time of day part */
			c++;
			
			if (sscanf(string(c,c+4 > datestring.end() ?  datestring.end() : c+4).c_str(), "%2d%2d", &hours, &minutes)  == 2) {
				c += 4;
			}
			else if (sscanf(string(c,c+5 > datestring.end() ?  datestring.end() : c+5).c_str(), "%2d:%2d", &hours, &minutes) == 2) {
				c += 5;
			}
			else {
				invaliddate();
			}

			if (*c == ':') {
				c++;
			}

			if (c < datestring.end()) {
			   	if (sscanf(string(c,c+2 > datestring.end() ?  datestring.end() : c+2).c_str(), "%2d", &seconds)  == 1) {
					c += 2;
			   	}
			   	else {
					invaliddate();
			   	}
				if (*c != '\0') {      /* something left? */
					invaliddate();
				}
			}
			tmstruct.tm_hour = hours;
			tmstruct.tm_min = minutes;
			tmstruct.tm_sec = seconds;
			isotime = mktime (&tmstruct);
		}

		else {
			invaliddate();
		}
	}

}

TTMOptions::~TTMOptions() {

}

void TTMOptions::usage() {

	const char *cp = optionString;
	ostringstream verbosemessage;

	cerr << "usage: " << progName;

	while (*cp != '\0') {
		switch (*cp) {
		case '4':
			// IPv4 instead of IPv6 data, optional
			cerr << " [-4]";
			verbosemessage << "	-4 select IPv4 data" << endl;
			break;
		case '6':
			// IPv6 instead of IPv4 data, optional
			cerr << " [-6]";
			verbosemessage << "	-6 select IPv6 data" << endl;
			break;
		case 'a':
			// all packets, optional
			cerr << " [-a]";
			verbosemessage << "	-a include all available packets , optional" << endl;
			break;
		case 'c':
			// check contents of root file
			cerr << " [-c]";
			verbosemessage << "	-c check packets in ROOT file (no output), optional" << endl;
			break;
		case 'd':
			// enable debugging output, optional
			cerr << " [-d]";
			verbosemessage << "	-d enable too verbose debugging output, optional" << endl;
			break;
		case 'D':
			// date in filename, optional
			cerr << " [-D]";
			verbosemessage << "	-D date in filename, optional" << endl;
			break;
		case 'e':
			// enable error plot, optional
			cerr << " [-e]";
			verbosemessage << "	-e enable error plot, optional" << endl;
			break;
		case 'f':
			// plot format, optional
			cp++;
			cerr << " [-f fmt]";
			verbosemessage << "	-f fmt plot format, optional (gif,ps,pdf)" << endl;
			break;
		case 'H':
			// generate html, optional
			cerr << " [-H]";
			verbosemessage << "	-H generate html, optional" << endl;
			break;
		case 'm':
			// max delay, optional
			cp++;
			cerr << " [-m max]";
			verbosemessage << "	-m max delay, optional ( > 0 )" << endl;
			break;
		case 'o':
			// output dir, optional
			cp++;
			cerr << " [-o outdir]";
			verbosemessage << "	-o select output directory" << endl;
			break;
      case 'O':
         // output option, optional
         cp++;
         cerr << " [-O output option]";
         verbosemessage << "-O output option, optional (all: normal output, xml: XML stats, 1..4: only show pan nr)" << endl;
         break;
		case 'p':
			// time period, required
			cp++;
			cerr << " -p period";
			verbosemessage << "	-p select ISO 8601 time period" << endl;
			verbosemessage << "	   CCYY-?MM-?DD[THH[MM[SS]]]/P[0-9]+(D|H|M|S)" << endl;
			break;
		case 'r':
			// save percentiles, optional
			cerr << " [-r]";
			verbosemessage << "	-r save percentiles, optional" << endl;
			break;
		case 'R':
			// output dir, optional
			cp++;
			cerr << " [-R ROOT data dir]";
			verbosemessage << "	-R select ROOT data input directory" << endl;
			break;
		case 's':
			// source id, optional
			cp++;
			cerr << " [-s src]";
			verbosemessage << "	-s src source id, optional (0 .. maxtestboxid)" << endl;
			break;
		case 't':
			// target id, optional
			cp++;
			cerr << " [-t tgt]";
			verbosemessage << "	-t tgt target id, optional (0 .. maxtestboxid)" << endl;
			break;
		case 'u':
			// max delay, optional
			cp++;
			cerr << " [-u max]";
			verbosemessage << "	-u max ip delay variation, optional ( > 0 )" << endl;
			break;
		case 'v':
			// verbose, optional
			cerr << " [-v]";
			verbosemessage << "	-v verbose, optional" << endl;
			break;
		case 'w':
			// week summaries, optional
			cerr << " [-w]";
			verbosemessage << "	-w week summaries, optional" << endl;
			break;
		case 'z':
			// min delay, optional
			cp++;
			cerr << " [-z min]";
			verbosemessage << "	-z min delay, optional ( > 0 )" << endl;
			break;
		default:
			cerr << "	" << *cp;
		}
		cp++;
	}
	cerr << endl;

	if (verbose)
		cerr << verbosemessage.str() << endl;

	exit(EX_USAGE);
}
