/*
Copyright (c) 2000,2001		  RIPE NCC


All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of the author not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS; IN NO EVENT SHALL
AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY
DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

/*
-------------------------------------------------------------------------------
Module Header
Filename	  : TestBoxConfig.C
Author	    : Rene Wilhelm
Date	      : 17-OCT-2000
Revised for Linux : 15-JUN-2001
Revised		  : 10-AUG-2001, TestBox target lists initialised from constructor
Revised		  : 31-OCT-2001, Create config based on program options
Description       : Implementation of TestBoxConfig class
Language Version  : C++
OSs Tested	: Solaris 2.6, Solaris 8, Debian Linux 2.2 
Todo		  : implement GetBoxByName method

This class provides an interface between analysis software and the
configuration of the TTM network (which boxes are on-line, what are
their names and ids, what are the targets for measurements from a
given box etc.). 

When requested via program options, the list of boxes in a
TestBoxConfig object will be limited to a selected source or
subset of sources.

$Id: TestBoxConfig.C,v 1.9 2006/11/13 15:21:28 ruben Exp $
-------------------------------------------------------------------------------
*/

#include <iostream>
#include <algorithm>
#include <string>

#include <cstdio>
#include <cerrno>
#include <cstring>

#include "TestBoxConfig.h"
#include "TTMOptions.h"
#include "utils.h"

RCSID("$Id: TestBoxConfig.C,v 1.9 2006/11/13 15:21:28 ruben Exp $");

// ------------------------------------
// TestBoxConfig constructor: 
//
// Input: 
//	options   -  TTMOptions: defaults or explicitly passed on command line  
// 

#define MAXWORDS 20
#define MAXCHARS 200

TestBoxConfig::TestBoxConfig(TTMOptions &options) { 

	char    line[MAXCHARS];
	char   *point = &line[0];
	char  **ap;
	char   *arg[MAXWORDS];
	FILE   *fp;
	int number_of_boxes=0;
	
	maxBoxId = 0;
	list_size = 0;

	// the command which will return TestBox Configuration
	// as a minimum, lines should have box id in first and box name
	// in second field; other fields as well as commented lines (#)
	// are ignored.

	// TODO Parse ttconfig -v TARGETS instead to set the target list, using elements from tbtracker or something like that

	const string &command = options.ConfigCommand();

	
	/* Read configuration data from pipe */

	if (options.Verbose()) {
		cout << "TestBox Config command: " << command << endl;
	}

	if ((fp = popen (command.c_str(), "r")) == NULL) {
		cerr << "ERROR: TestBoxConfig: Cannot open pipe " << command << endl;	 
		exit(1);
	}


tryagain:

	/* Read data */
	while ( (point=fgets(line, MAXCHARS, fp)) != NULL ) {

		/* # sign indicates comment */
		if (line[0] == '#') { continue; }

		/* Split line into pieces */
		ap = arg;
		*ap = strtok (point, " \t");
		while (*ap != NULL) {
			if (**ap != '\0') {
				++ap;
			}
			*ap = strtok (NULL, " \t");
		}
	
		uint id = atoi(arg[0]);
		
		// check if we've seen this one before
		// yes, skip it: first name found is real name
		// and we do not yet care about IP addresses

		if (tblist.find(id) != tblist.end())
			continue;	

		int length = strlen(arg[1]);
		char *last = arg[1] + length - 1;

		if (*last == '\n') {
			*last = '\0';
			length--;
		}

		// new entry

		number_of_boxes++;
		tblist[id] = new TestBox(id, arg[1]);

		// Keep track of allocation here
		tbtracker.push_back(tblist[id]);

		if (id > maxBoxId) {
			maxBoxId = id;
		}
	}

	if (number_of_boxes == 0) {
#ifdef DEBUG
		perror("fgets");
#endif
		if (ferror(fp)) {       // A system error occurred 
			 clearerr(fp);
			 if (errno == EINTR) {
#ifdef DEBUG
				// Retry
				cout << "fgets intterupted! retrying ... \n"; 
#endif
				goto tryagain;
			 }
		}
	}

	pclose (fp);

	// now initialize the target lists
	SetTargets();

	// We have the full config, now restrict to specific src/tgt
	// if requested so by program options.

	if (options.SourceId() != 0) {
		TestBox *src = GetBoxById(options.SourceId());
		if (src != NULL) {

			tblist.clear();
			tblist[options.SourceId()] = src;
		}
		else {
			cerr << options.ProgramName() << ": ABORT: Box " <<
			options.SourceId() << " not found in testbox config" <<
			endl;
			exit (1);
		}
			
		if (options.TargetId() != 0) {
			TestBox *tgt = src->FindTargetById(options.TargetId());
			if (tgt != NULL) {

				map<uint,TestBox*> singletarget;
				singletarget[options.TargetId()] = tgt;
				src->SetTargets(singletarget);
				maxBoxId = max(options.SourceId(), options.TargetId());
			}
			else {
				cerr << options.ProgramName() <<
				   ": ABORT: Box " << options.TargetId() <<
				   " not configured as target of box "<<
				   options.SourceId() << endl;
				exit (1);
			}
		}
	}
}


// ------------------------------------
// SetTargets	- find active targets for all boxes in this Config
//
//
// Currently we have a full mesh, every active box sending to and
// receiving from every other active box. However, scaling problems
// could prevent us from continuing this in future.
//
// In anticipation of such a change, all analysis software should 
// be flexible and get the 'target' information from TestBox.targetlist[]
// instead of implicitly assuming a full measurement mesh.

void TestBoxConfig::SetTargets() {

	for (map<uint,TestBox*>::iterator j=tblist.begin() ; j != tblist.end() ; j++) {

		TestBox *source = PAIR_VAL(j);

		for (map<uint,TestBox*>::iterator i=tblist.begin() ; i != tblist.end() ; i++) {

			// Don't add myself as a target
			if (PAIR_KEY(i) != source->GetId()) {
				source->AddTarget(PAIR_VAL(i));
			}
		}
	}
}
			
// ------------------------------------
//
// GetBoxById	-  Get TestBox struct associated with given numeric id

TestBox *TestBoxConfig::GetBoxById(uint id) { 

	// simple linear search; replace by hash in future?	

	map<uint,TestBox*>::iterator i,j;

	for (i = tblist.begin(); i != tblist.end() && PAIR_VAL(i); i++)
	{
		if (PAIR_VAL(i)->GetId() != id)
		{
			map<uint,TestBox*> &tgtlst = PAIR_VAL(i)->GetTargets();
			if ((j = tgtlst.find(id)) != tgtlst.end())
				return PAIR_VAL(j);
		}
		else
		{
			return PAIR_VAL(i);
		}
	}
	return(NULL); // not found ??
}

TestBox            *TestBoxConfig::GetBoxByName (const string & name, bool shortname)
{
	// create iterators for browsing the map.
	// iterators can be seen as "pointers the C++ way"

	map < uint, TestBox * >::iterator i, j;

	for (i = tblist.begin (); i != tblist.end () && PAIR_VAL (i); i++)
	{
		if (shortname ? (PAIR_VAL (i)->GetShortName () != name) : (PAIR_VAL (i)->GetName () != name))
		{
			map < uint, TestBox * >&tgtlst = PAIR_VAL (i)->GetTargets ();
			for (j = tgtlst.begin (); j != tgtlst.end () && PAIR_VAL (j); j++)
				if (shortname ? (PAIR_VAL (j)->GetShortName () != name) : (PAIR_VAL (j)->GetName () != name))
					return PAIR_VAL (j);
		}
		else
		{
			return PAIR_VAL (i);
		}
	}
	return (NULL);		// not found ??
}

TestBoxConfig::~TestBoxConfig()
{
	// Cleanup allocated memory of testboxes
	// Do this by walking the tbtracker vector, and clear all pointers

	for_each(tbtracker.begin(),tbtracker.end(),delete_pointer<TestBox*>());
}
