/*
Copyright (c) 2001                      RIPE NCC


All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of the author not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS; IN NO EVENT SHALL
AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY
DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

/*
-------------------------------------------------------------------------------
Module Header
Filename          : ipdv.C
Author            : Reinhard Sojka
Date              : 05-NOV-2001
Description       : Program for creating TTM IPDV ("jitter") plots
Language Version  : C++
OSs Tested        : Solaris 2.6
Command line options:

    --help  get an overview of the options
    -6      IP version 6 data
    -s 	    source box no.
    -p	    start time and end time, format is YYYYMMDD/YYYYMMDD,
      	    any ISO 8601 format is ok, like YYYYMMDDThh:mm:ss or similar

    -t	    target box no., -1 for all connected boxes, optional, default = -1

    -D	    date in file name, 'on' or 'off', optional,
	    default = 'off' (ipdv.tt_A.tt_B.start.end.*)
    -e	    enable error plot, 'on' or 'off', optional
 	    (default = 'off')
    -f	    output file format, 'gif',  'ps', 'pdf' or 'eps', optional
	    (default = 'gif')
    -m      maximum delay to be displayed on histograms, optional
	    (default = 800 ms)
    -z      minimum delay to be displayed on histograms, optional
	    (default = 0 ms)
    -u      maximum IPDV to be displayed on histograms, optional
	    (default = 400 ms)
    -o	    output directory for plots, optional
	    (default is current directory)
    -h	    create a HTML file for the webpage, 'on' or 'off', optional
	    (default = 'off')
    -r	    write percentiles to a file, 'on' or 'off', optional
	    (default = 'off')
    -v      verbose, 'on' or 'off', optional, default = 'off'

Notes/Description:

	program to read the data of a time interval betweeen some minutes and
	several days in different lists, filtered by testboxnumber

	sort the data by arrival time and calculata IPDV for the (existing)
	testboxes in file; calculates also the experimental error and draws
	an error plot>

	plots will be stored as *.eps file in current folder
	normalized delay histograms, added statistic pad
	autozoom for delay distribution and IPDV distribution
	added percentiles for statistics pad
	output as gif or eps, options/switches made conform with delay
	calculation

	changed time input, changed handling of time formats and creation
	of file names now with the correct time in the histograms
	(Root has its own time stamp, starting 1995)
	percentiles can be writtten to a file
	added the possibility to customize the plots

external programs:   ps2gif to convert eps into gif format


$Id: ipdv.C,v 1.20 2006/11/13 15:21:28 ruben Exp $
-------------------------------------------------------------------------------
*/


#define TTCONFIG_COMMAND  "/ncc/ttpro/bin/ttconfig"
#define TTCONFIG_COMMAND_ARGS "-v active"
//#define DEFAULT_OUTPUTDIR "/ncc/ttpro/spool/plots"

#include <iostream>
#include <sstream>
#include <stdexcept>
#include <fstream>
#include <vector>
#include <map>

#include <cstdio>
#include <cstring>
#include <ctime>
#include <cstdlib>
#include <cerrno>

#include <unistd.h>
#include <sys/stat.h>
#include <sysexits.h>
#include <stdio.h>

#ifndef __CINT__
#include <TROOT.h>
#include <TApplication.h>
#include <TCanvas.h>
#include <TChain.h>
#include <TTree.h>
#include <TFile.h>
#include <TH1.h>
#include <TH2.h>
#include <TF1.h>
#include <TGraphErrors.h>
#include <TPaveLabel.h>
#include <TText.h>
#include <TLine.h>
#include <TStyle.h>
#include <TPostScript.h>
#include <TError.h>
#endif

#include <Delay.h>

using namespace std;

#include "TTMOptions.h"
#include "TestBoxConfig.h"
#include "Percentiles.h"	// Rene's class to calculate the percentiles
#include "utils.h"

RCSID("$Id: ipdv.C,v 1.20 2006/11/13 15:21:28 ruben Exp $");

// forward declaration
struct ttmData;

// * * *  declaration of subroutines  * * *
void                CreateAxisStyle (char *xTitle, char *xTimeFormat, Double_t minXax, Double_t maxXax, plotformat outputFormat);
void                CreatePlotTitle (string &titleOfWholePlot, Int_t boxNo, string &SourceBox, Double_t minXax, Double_t maxXax);
void                CreateChar (string &SourceBox, string &startYYYYMMDD, string &endYYYYMMDD, Int_t start_Box, Double_t minXax,
				Double_t maxXax);
time_t              CreateRootFileName (char *sourceFile, time_t act_TimeStamp, Int_t start_Box);
void                CreatePlotFileNameDate (string &plotFileName, string &baseFileName, Int_t boxNo, const string &SourceBox, const string &startYYYYMMDD,
					    const string &endYYYYMMDD, plotformat outputFormat);
void                CreatePlotFileNameStd (string &plotFileName, string &baseFileName, Int_t boxNo, const string &SourceBox, plotformat outputFormat);
void                WriteStatisticsPad (TPad * pad, Int_t allPackets, Int_t validPackets, Int_t noOfFlaps,
					Percentiles * ipdvPercentile, Float_t rmsIPDV, Float_t meanDelay, Float_t rmsDelay);
void                write_stats (TPad * pad, const char *ptr_text, const Float_t x, const Float_t y, const Short_t align,
				 Font_t font);
void                PutDataInElement (Delay * delay, ttmData * element);
void                CalculateAndFillHistograms (vector<ttmData> &access, TH1F * h1, TH2F * h2, TH2F * h3, TH2F * h4, TH1F * h5,
						TGraphErrors * ipdvErr, Percentiles * ipdvPercentile, Int_t enableErrorPlot);
void                SetHistogramProperties (TH1F * h1, TH2F * h2, TH2F * h3, TH2F * h4, TH1F * h5, TH2F * dummy, char *xTitle,
					    char *xTimeFormat);
void                DrawHistograms (TH1F * h1, TH2F * h2, TH2F * h3, TH2F * h4, TH1F * h5, TPad * pad1, TPad * pad2, TPad * pad3,
				    TPad * pad4, Int_t customIPDVPlot, Int_t customDelayPlot);
void                DrawErrorPlot (TH2F * dummy, TGraphErrors * ipdvErr, TPad * errorPad);
void                SavePercentiles (Percentiles * ipdvPercentile, const char *SourceBox, Int_t boxNo, ofstream & IntoFile,
				     Int_t firstLine);
void                CreateHTMLFile (Percentiles * ipdvPercentile, const char *anyString, Int_t boxNo, ofstream & IntoFile,
				    Int_t paragraphID);
float               CheckPlotParameter (float &maxIPDV, float &maxCustomDelay, float &minCustomDelay);
const               string & boxid2name (uint id);


// * * *  declaration of classes * * *

struct ttmData
{
	UInt_t              PacketId;	// Identifier of the packet
	Int_t               SourceId;	// ID# of the sending testbox
	Int_t               TargetId;	// ID# of the receiving testbox
	Double_t            ArrivalTime;	// -1.0 if undefined
	Double_t            PacketDelay;	// Delay in ms, -1.0 if undefined
	Int_t               Nhops;	// Number of hops, -1 if unknown
	Float_t             SourceNtp;	// Sending clock NTP estimated error
	Float_t             TargetNtp;	// Receiving clock NTP estimated error

	ttmData(): PacketId(0),SourceId(0),TargetId(0),ArrivalTime(-1.0),PacketDelay(-1.0),Nhops(-1),SourceNtp(0),TargetNtp(0) { /* XXX EMPTY */ }
	ttmData(const Delay &delay) {Delay *theDelay=const_cast<Delay *>(&delay);PutDataInElement(theDelay,this);}
	ttmData& operator=(const Delay &);
	void PutDataInElement (Delay * delay, ttmData * element);
};

inline bool operator< (const ttmData & lhs, const ttmData & rhs)
{
	return (lhs.ArrivalTime < rhs.ArrivalTime);
}

// * * * G L O B A L  V A R I A B L E S * * *
const Int_t         MAX = 125;	// number of testboxes

// path to root files
// 4096 == FILENAME_MAX
char                fn_Path[4096] = "/ncc/ttpro/data/root";




// Root has its own time, starting 1995 - needed for histograms
const time_t        convToRootTime = 788918400;



char               *progname;
TStyle             *rootstyle;	// to get access to gStyle

int                 ipVersion = 4;	// default IPv4 data

TestBoxConfig	*tbConfig = NULL;	

// * * * M A I N * * *
int main (int argc, char **argv)
{
// fetch options and set defaults

	TTMOptions options(argc, argv, "46rDdeHvf:m:o:p:R:s:t:u:z:");

	// Ignore ROOT messages
	gErrorIgnoreLevel=1;

	if (options.Debug())
		gErrorIgnoreLevel=0;
	// assigns the name, which is used to start the program, to argv[0]
	progname = argv[0];

	// ROOT initialisation
	TROOT               simple ("TTM", "Plots");

	gROOT->SetBatch ();	// no graphics windows
	TApplication       *app = new TApplication ("delayPlots", NULL, NULL, NULL, 0);

	// ensure correct timezone for date<->time_t conversions
	putenv (const_cast<char*>("TZ=GMT+0"));

	// set default values and get the input data

	// make the IPDV plots with specified parameters, default autozoom
	Int_t               customIPDVPlot = 0;

	// make the delay plots with specified parameters, default is autozoom
	Int_t               customDelayPlot = 0;

	Float_t             maxIPDV = 400,
	                    maxCustomDelay = 800,
	                    minCustomDelay = 0;
	


	TestBoxConfig config(options);

	strncpy (fn_Path,options.InputPath().c_str(),sizeof(fn_Path));	// fn_Path is global
	

	// make sure that the parameter for custom plots make sense

	ipVersion = options.IPversion();

	if (options.CustomMaxDelay() || options.CustomMinDelay())
	{
		if (options.CustomMaxDelay()) 
			maxCustomDelay = options.MaxDelay();
		if (options.CustomMinDelay()) 
			minCustomDelay = options.MinDelay();
		customDelayPlot = 1;
	}
	if (options.CustomMaxIPDV())
	{
		maxIPDV = options.MaxIPDV();
		customIPDVPlot = 1;
	}

	CheckPlotParameter (maxIPDV, maxCustomDelay, minCustomDelay);

	Int_t               boxNo;	// contains the actual testbox

	// create pointers - lists are organized in an array

	map<uint, vector<ttmData> > packetmap;

	// create arrays
	map<uint,int>		allPackets;	// for statistics
	map<uint,int>		validPackets;

	// will store the actual routing vector
	map<uint,double>	 routingVectorId;

	// Int_t noOfVectors[MAX]; //how many vectors per List

	// how often changed the routing vector
	map<uint,int>		noOfFlaps;
	map<uint,Percentiles>	ipdvPercentile;

	tbConfig =  &config;
	map<uint,TestBox*> &tblist = config.GetTBList(); 

	boxNo=PAIR_VAL(tblist.begin())->GetId();	// XXX prevent compiler warning

	// initialize pointers and arrays for this source
	for (map<uint,TestBox*>::iterator i=tblist[options.SourceId()]->GetTargets().begin() ; i != tblist[options.SourceId()]->GetTargets().end(); i++)
	{
		boxNo = PAIR_VAL(i)->GetId();

		allPackets[boxNo] = 0;
		validPackets[boxNo] = 0;

		// routing vector is assigned(min. 1 vector)
		routingVectorId[boxNo] = 0.0;

		ipdvPercentile[boxNo].Reset();	// initialize content
	}

	// create a chain of files and read first file
	if (options.Verbose() > 0)	// info - if you want to see it
		// just for info
		cerr << endl << "INFO: loading file(s)..." << endl;

	// file name, will be added to chain
	char                sourceFile[250];

	TChain              chain ("tree");

	// add one or more files to chain, if necessary
	time_t              act_TimeStamp = options.EndTime() - options.TimePeriod();

	do
	{
		act_TimeStamp = CreateRootFileName (sourceFile, act_TimeStamp, options.SourceId());

		struct stat         statbuf;

		if (stat (sourceFile, &statbuf) < 0)
		{
			if (options.Verbose() > 0)
				perror (sourceFile);
		}
		else
		{
			chain.Add (sourceFile);
		}

		if (options.Verbose() > 0)	// info - if you want to see it
			cerr << "INFO:\t" << sourceFile << endl;
	}
	while (act_TimeStamp <= (options.EndTime()));

	// XXX should round endtime to end of day

	// create a delay object to store data per event (a whole set of data)
	Delay              *delay = new Delay ();

	chain.SetBranchAddress ("delay", &delay);

	// get number of lines(=events) in source file
	Int_t               nTotalEvents = (Int_t) chain.GetEntries () - 1;

	// to get range of x axis for histograms - calculate it from start
	// parameters in a subroutine
	Double_t            minXax = (Double_t) options.EndTime() - options.TimePeriod();
	Double_t            maxXax = (Double_t) options.EndTime();

	Int_t               readPackets = 0;	// debug

	// find selected packets and store them in the list
	// read in starts from the end to get a list which starts with the
	// first line

	if (options.Verbose() > 0)	// info - if you want to see it
		cerr << endl << "INFO: putting data in list(s)... (no# of events: " << nTotalEvents << ")" << endl;	// just
														//
	// for info

	if (nTotalEvents == -1)
	{
		if (options.Verbose() > 0)
			cerr << "WARNING: No data to process for tuple (" << options.SourceId() << " <=> " << options.TargetId() << ")" << endl;
		exit (EX_NOINPUT);
	}

	for (Int_t eventNo = nTotalEvents; eventNo > (-1); eventNo--)
	{
		chain.GetEvent (eventNo);	// read a set of data

		// select packets
		if ((delay->GetArrivalTime () < maxXax) && (delay->GetArrivalTime () > minXax))
		{

			allPackets[delay->GetTargetId ()]++;	// increment
			// counter

			// only valid packets
			if (delay->Status () == ClockValid)	
			{
				// select packets, create element and put it in
				// (correct) list

				// extract data from all target boxes
				// in chain

				// Skip non matching targetid's when we are working on a single target

				if ((options.TargetId() > 0) && ( (uint)delay->GetTargetId () != options.TargetId() ))
					continue;

				ttmData            element = *delay;

				// put element in list

				packetmap[element.TargetId].push_back(element);

				// counter - debug
				readPackets++;	


				// increase counter for valid packets
				validPackets[element.TargetId]++;

				// find out how often the route was switched -
				// maybe a bit uneffective here

				if ((delay->GetRouteId() > 0) && (routingVectorId[delay->GetTargetId ()] != delay->GetRouteId ()))
				{
					routingVectorId[delay->GetTargetId ()] = delay->GetRouteId ();

					if (noOfFlaps.find(delay->GetTargetId ()) != noOfFlaps.end())
						noOfFlaps[delay->GetTargetId ()]++;
					else
						noOfFlaps[delay->GetTargetId ()] = 0;
				}

			}	// endif ClocksOK()
		}
	}			// end for-loop

	// sorting all lists by PacketId
	// XXX iterate on packetmap instead...
	for (map<uint,TestBox*>::iterator i=tblist[options.SourceId()]->GetTargets().begin() ; i != tblist[options.SourceId()]->GetTargets().end(); i++)
	{
		boxNo = PAIR_VAL(i)->GetId();

		// sorting only lists containing data
		if (packetmap[boxNo].size() > 2 )
		{
			if (options.Verbose() > 0)	// info - if you want to see it
				cerr << "INFO: Sorting for box " << boxNo << ", no# of elements " << packetmap[boxNo].size() << endl;

			sort(packetmap[boxNo].begin(), packetmap[boxNo].end());
		}
		else
		{
			if (options.Verbose() > 0)	// info - if you want to see it
				cerr << "WARNING: no data for box " << boxNo << endl;
		}
	}

	// force thin lines in PS output to avoid cluttering histograms
	gStyle->SetLineScalePS (0.2);

	// create canvas
	TCanvas            *c1 = new TCanvas ("c1", "IPDV plots", 0, 0, 1270, 990);

	// create pads to display histograms and statistics
	TPad               *pad1 = new TPad ("pad1", "This is pad 1", 0.01, 0.49, 0.43, 0.95);

	// 1-4 will display histograms
	TPad               *pad2 = new TPad ("pad2", "This is pad 2", 0.44, 0.49, 0.86, 0.95);
	TPad               *pad3 = new TPad ("pad3", "This is pad 3", 0.01, 0.02, 0.43, 0.48);
	TPad               *pad4 = new TPad ("pad4", "This is pad 4", 0.44, 0.02, 0.86, 0.48);
	TPad               *pad5 = new TPad ("pad5", "This is pad 5", 0.87, 0.02, 0.99, 0.95);

	// will display statistics
	// draw grid

	pad1->SetGridx ();
	pad1->SetGridy ();
	pad2->SetGridx ();
	pad2->SetGridy ();
	pad3->SetGridx ();
	pad3->SetGridy ();
	pad4->SetGridx ();
	pad4->SetGridy ();

	// force them to be visible on the canvas - histograms don't appear
	// without this procedure

	pad1->Draw ();
	pad2->Draw ();
	pad3->Draw ();
	pad4->Draw ();
	pad5->Draw ();

	// prepare title of the canvas c1
	TPaveLabel         *pl1 = new TPaveLabel (0.175, 0.956, 0.825, 0.998, "foo", "br");

	// prepare, text will be filled later
	pl1->SetTextSize (0.45);
	pl1->Draw ();

	if (ipVersion == 6)
	{
		TPaveLabel         *ip = new TPaveLabel (0.01, 0.956, 0.1, 0.998, "IPv6", "br");

		ip->SetTextColor (2);
		ip->SetTextSize (0.8);
		ip->Draw ();
	}


	// and once again for the next canvas with a pad to display the error
	// plot
	TCanvas            *c2 = new TCanvas ("c2", "more IPDV plots", 0, 0, 1270, 990);
	TPad               *errorPad = new TPad ("errorPad", "to display errors", 0.02, 0.04, 0.98, 0.92);	// will
														//
	//
	// display error graph
	//
	errorPad->SetGridx ();
	errorPad->SetGridy ();
	errorPad->Draw ();

	// prepare title of the canvas c2
	TPaveLabel         *pl2 = new TPaveLabel (0.175, 0.956, 0.825, 0.998, "foo", "br");

	// prepare, text will be filled later
	pl2->SetTextSize (0.45);
	pl2->Draw ();

	// create a 1dim-histogram
	TH1F               *h1 = new TH1F ("h1", "IPDV", 800, ((-1) * maxIPDV), maxIPDV);

	TH1F               *h5 = new TH1F ("h5", "PacketDelay", 800, minCustomDelay, maxCustomDelay);

	// create 2dim histograms with dynamic borders according to selected time frame
	TH2F               *h2 = new TH2F ("h2", "IPDV vs time", 500, (minXax - convToRootTime), (maxXax - convToRootTime), 400, ((-1) * maxIPDV), maxIPDV);

	//delay
	TH2F               *h3 = new TH2F ("h3", "delay (black) and hops*10 (red)", 500, (minXax - convToRootTime), (maxXax - convToRootTime), 400, minCustomDelay, maxCustomDelay);	// delay
																							//
	//nhops
	TH2F               *h4 = new TH2F ("h4", "delay (black) and hops*10 (red)", 500, (minXax - convToRootTime), (maxXax - convToRootTime), 400, minCustomDelay, maxCustomDelay);	// nhops
																							//
	TH2F               *dummy = new TH2F ("dummy", "IPDV error", 500, (minXax - convToRootTime), (maxXax - convToRootTime), 400, -400, 400);	// just
																			//
	// a placeholder to get the axes for the error diagram

	// create title and style of X axis
	char                xTitle[100];
	char                xTimeFormat[6];

	CreateAxisStyle (xTitle, xTimeFormat, minXax, maxXax, options.PlotFormat());

	// call subroutine for calculation

	// set properties of the histograms
	SetHistogramProperties (h1, h2, h3, h4, h5, dummy, xTitle, xTimeFormat);

	// call subroutine

	// convert variables(numbers) in strings, so they can be used in file
	// names
	string                SourceBox,	// number of source box
	                startYYYYMMDD,
	                endYYYYMMDD;

	CreateChar (SourceBox, startYYYYMMDD, endYYYYMMDD, options.SourceId(), minXax, maxXax);

	// write percentiles to a file, if selceted at data input
	ofstream            IntoFile;	// file handler - declaration here,
	ofstream            theHTML_File;	// file handler - declaration


	// otherwise compiler complains
	if (options.SavePercentiles() > 0)
	{
		string percFileName;	// selected target Box

		percFileName = options.OutputPath() + "/ipdv_percentiles." + boxid2name(options.SourceId()) + ".txt";

		// open file and write data
		IntoFile.open (percFileName.c_str());	// open (or create) file

		// call subroutine to write the first lines
		if(!IntoFile.fail())
			SavePercentiles (&(ipdvPercentile[options.SourceId()]), boxid2name(options.SourceId()).c_str(), options.SourceId(), IntoFile, 1);
		else
			cerr << "WARNING: cannot open " << percFileName << " (" << strerror(errno) << ")" << endl;
	}

	if (options.CreateHTML() > 0)
	{
		string htmlFileName;	// selected target Box

		htmlFileName = options.OutputPath() + "/ipdv_LinkPage." + boxid2name(options.SourceId()) + ".html";

		// open file and write data
		theHTML_File.open (htmlFileName.c_str());	// open (or create)
		// file

		if(!theHTML_File.fail())
			CreateHTMLFile (&(ipdvPercentile[options.SourceId()]), boxid2name(options.SourceId()).c_str(), options.SourceId(), theHTML_File, 0);
		else
			cerr << "WARNING: cannot open " << htmlFileName << " (" << strerror(errno) << ")" << endl;
		// call subroutine to write the first lines
	}

	// format key for plot file - see class TPostScript
	Int_t               formatKey;

	// standard output format is gif
	if (options.PlotFormat() == eps)
		formatKey = 113;	// 113=eps
	else
		// 112=ps - gif will be created from a *.ps file (by an
		// external Program)
		formatKey = 112;


	// other strings for filenames and titles
	string              plotFileName,	// file name of plot
						baseFileName, // base file name without extension
	                    errorPlotFileName,	// file name of plot
	                    titleOfWholePlot;	// title of the plot,

	// displayed on top of the canvas c1
	char                buffer[250];

	// calculate IPDVs and draw+save plots - loop over all data

	if (options.Verbose() > 0)	// info - if you want to see it
		cerr << "INFO: calculating and drawing for " << options.SourceId() << "..." << endl;	// just for
	// info

	// XXX iterate on packetmap instead
	for (map<uint,TestBox*>::iterator i=tblist[options.SourceId()]->GetTargets().begin() ; i != tblist[options.SourceId()]->GetTargets().end(); i++)
	{
		boxNo = PAIR_VAL(i)->GetId();
		if (options.Verbose())
			cerr << "INFO: At boxid " << boxNo << endl;

		// calcualting only lists containing data
		if (packetmap[boxNo].size() > 2)
		{
			// clear histograms
			// XXX option "ICE" doesn't reset the number of entries,
			// maybe fixed in a new version of ROOT
			// h1->Reset("ICE");
			h1->Reset ();
			h2->Reset ("ICE");
			h3->Reset ("ICE");
			h4->Reset ("ICE");
			h5->Reset ();
			// XXX option "ICE" doesn't reset the number of
			// entries, maybe fixed in a new version of ROOT
			// h5->Reset("ICE");

			// create a TGraphErrors - has to be done inside loop
			TGraphErrors       *ipdvErr = new TGraphErrors (validPackets[boxNo]);

			// constructor will need the number of entries

			// do the calculation - call subroutine
			
			CalculateAndFillHistograms (packetmap[boxNo], h1, h2, h3, h4, h5, ipdvErr, &(ipdvPercentile[boxNo]),
						    options.EnableErrorPlot());

			// create the file name and prepare to save the image,
			// plus adding of the output directory
			if (options.DateInFilename() > 0)
				CreatePlotFileNameDate (plotFileName, baseFileName, boxNo, boxid2name(options.SourceId()).c_str(), startYYYYMMDD.c_str(), endYYYYMMDD.c_str(), options.PlotFormat());
			else
				CreatePlotFileNameStd (plotFileName, baseFileName, boxNo, boxid2name(options.SourceId()).c_str(), options.PlotFormat());
			// create whole path and file name
			strcpy (buffer, options.TmpDir().c_str());

			// plotFileName may be used later for the name of the
			// error plot

			// prepare to write to file
			TPostScript        *epsHandle = new TPostScript (string(options.TmpDir() + "/" + plotFileName).c_str(), formatKey);
			if (epsHandle == NULL)
			{
				cerr << "ERROR: No output possible to " << buffer << endl;
				exit(EX_UNAVAILABLE);
			}

			// not for *.eps because of unexpected results
			if (formatKey != 113)
				// center of A4 page (according to delays program)
				epsHandle->Range (25.5, 18.8);

			// draw the histograms on canvas c1
			c1->cd ();
			DrawHistograms (h1, h2, h3, h4, h5, pad1, pad2, pad3, pad4, customIPDVPlot, customDelayPlot);

			// create title of the plot to be displayed on top of
			// the canvas
			CreatePlotTitle (titleOfWholePlot, boxNo, SourceBox, minXax, maxXax);
			c1->cd ();
			pl1->SetLabel (titleOfWholePlot.c_str());
			pl1->Draw ();

			// write data to statistics pad
			pad5->cd ();
			WriteStatisticsPad (pad5, allPackets[boxNo], validPackets[boxNo], noOfFlaps[boxNo], &(ipdvPercentile[boxNo]), (h1->GetRMS()), (h5->GetMean()), (h5->GetRMS()));
			// write canvas to disk and delete handler
			c1->cd ();
			c1->Update ();
			epsHandle->Close ();
			delete              epsHandle;

			// convert *.ps to *.gif - if it has been selected at
			// data input
			if (options.PlotFormat() == gif)
			{
				// call system function and delete the eps file

				sprintf (buffer, "ps2gif -b 41 28 574 750.75 -s 1.2 -r -o %s/%s.gif %s/%s ", options.OutputPath().c_str(), baseFileName.c_str(), options.TmpDir().c_str(), plotFileName.c_str());
				system (buffer);

				sprintf (buffer, "%s/%s", options.TmpDir().c_str(), plotFileName.c_str());
				unlink (buffer);
			}
			else if (options.PlotFormat() == pdf)
			{
				// convert to pdf

				// call system function and delete the eps file
				sprintf (buffer, "ROOTps2pdf -o %s %s/%s", options.OutputPath().c_str(), options.TmpDir().c_str(), plotFileName.c_str());
				system (buffer);

				sprintf (buffer, "%s/%s", options.TmpDir().c_str(), plotFileName.c_str());
				unlink (buffer);
			}
			else if ((options.PlotFormat() == ps) || (options.PlotFormat() == eps))
			{
				char dst[500]; 
				sprintf (dst, "mv %s/%s %s/%s", options.TmpDir().c_str(), plotFileName.c_str(), options.OutputPath().c_str(), plotFileName.c_str());
				system(dst);
			}

			// draw error plot - if it has been enabled at data
			// input
			if (options.EnableErrorPlot() > 0)
			{
				// create file name for error graph
				errorPlotFileName = options.TmpDir() + "/error_" + plotFileName;

				// draw plot
				TPostScript        *epsErrorHandle = new TPostScript (errorPlotFileName.c_str(), formatKey);

				c2->cd ();
				DrawErrorPlot (dummy, ipdvErr, errorPad);

				c2->cd ();
				pl2->SetLabel (titleOfWholePlot.c_str());
				pl2->Draw ();

				// save to disk
				c2->cd ();
				c2->Update ();
				epsErrorHandle->Close ();
				delete              epsErrorHandle;

				// and convert *.ps to *.gif - if it has been
				// selected at data input
				if (options.PlotFormat() == gif)
				{
					// call the system function and delete
					// the eps file

					sprintf (buffer, "ps2gif -b 41 28 574 750.75 -s 1.2 -r -o %s/error_%s.gif %s", options.OutputPath().c_str(), baseFileName.c_str(), errorPlotFileName.c_str());
					system (buffer);

					unlink(errorPlotFileName.c_str());
				}

			}

			// write percentiles to a file or write HTML file, if
			// selceted at data input
			if (options.SavePercentiles() > 0)
				SavePercentiles (&(ipdvPercentile[boxNo]), boxid2name(boxNo).c_str(), boxNo, IntoFile, 0);


			if (options.CreateHTML() > 0)
				CreateHTMLFile (&(ipdvPercentile[boxNo]), plotFileName.c_str(), boxNo, theHTML_File, 1);

			// delete error plot
			delete              ipdvErr;

		}		// endif
	}			// end for loop

	// write the end of the file - if file creation has been selected as
	// data input
	if (options.CreateHTML() > 0)
		CreateHTMLFile (&(ipdvPercentile[boxNo]), boxid2name(boxNo).c_str(), boxNo, theHTML_File, 2);

	// clean up and delete list
	Int_t               delPackets = 0;

	if (options.Verbose() > 0)	// info - if you want to see it
	{
		// XXX iterate on packetmap instead
		for (map<uint,TestBox*>::iterator i=tblist[options.SourceId()]->GetTargets().begin() ; i != tblist[options.SourceId()]->GetTargets().end(); i++)
		{
			boxNo = PAIR_VAL(i)->GetId();
			delPackets += packetmap[boxNo].size();
		}			// end for loop
	}

	delete              delay;
	delete              h1;	// histograms
	delete              h2;
	delete              h3;
	delete              h4;
	delete              h5;
	delete              dummy;

	// delete f_gauss;
	delete              pad1;
	delete              pad2;
	delete              pad3;
	delete              pad4;
	delete              pad5;
	delete              errorPad;
	delete              pl1;
	delete              pl2;	// labels
	delete              c1;
	delete              c2;
	delete              app;

	// close files
	IntoFile.close ();	// close file
	theHTML_File.close ();


	if (options.Verbose() > 0)	// info - if you want to see it
	{
		cerr << "INFO: " << readPackets << " packets filtered" << endl;
		cerr << "INFO: " << delPackets << " packets deleted" << endl;
	}
	return (1);
}				// * * * end main * * *

// * * * S U B R O U T I N E S * * *

//generates the axis title and the (time) format string for the x axis
void CreateAxisStyle (char *xTitle, char *xTimeFormat, Double_t minXax, Double_t maxXax, plotformat outputFormat)
{
	// generates the axis title, depending on the time frame
	// converts unix timeformat into "readable" strings
	char                minText[20];
	char                maxText[8];

	// type long is needed for conversion
	const time_t		t_minXax = (long) minXax;
	const time_t		t_maxXax = (long) maxXax;
	tm                 *timestruct = gmtime (&t_minXax);

	// convert the first parameter into a time structure

	Int_t               ndays = (Int_t) (maxXax - minXax) / 3600 / 24;

	// from Unix time format to days
	if (ndays > 3)
	{
		// extract only year
		strftime (minText, 20, "%Y", timestruct);

		sprintf (xTitle, "      %s%26cTime [month day]%35c", minText, ' ', ' ');
		strcpy (xTimeFormat, "%b %d");	// <month> <day> - axis style
	}
	else
	{
		strftime (minText, 20, "%Y %b %d", timestruct);
		timestruct = gmtime (&t_maxXax);
		strftime (maxText, 8, "%b %d", timestruct);
		sprintf (xTitle, "      %s%20cTime [hour:minute]%21c%s", minText, ' ', ' ', maxText);
		if (outputFormat == eps )
			// different look for eps (=2) output - don't know why
			sprintf (xTitle, "      %s%13cTime [hour:minute]%14c%s", minText, ' ', ' ', maxText);
		// <hour> <minute> - axis style
		strcpy (xTimeFormat, "%H:%M");

	}

}


//generates the title for the canvas c1
void CreatePlotTitle (string &titleOfWholePlot, Int_t boxNo, string &SourceBox, Double_t minXax, Double_t maxXax)
{
	ostringstream titleOfWholePlot_buf;

	string TargetBox = boxid2name(boxNo);	// selected target Box

	char                timeToText[20];
	const time_t		t_minXax = (long) minXax;

	// type long is needed for conversion
	const time_t		t_maxXax = (long) maxXax;

	// convert the first parameter into a time structure
	tm                 *timestruct;

	titleOfWholePlot_buf << "Delay Variations from " << SourceBox << " to " << TargetBox;

	// extract date and time
	timestruct = gmtime (&t_minXax);
	strftime (timeToText, 20, "%Y-%b-%d,  %H:%M", timestruct);

	titleOfWholePlot_buf << "  Start: " << timeToText;


	// extract date and time
	timestruct = gmtime (&t_maxXax);
	strftime (timeToText, 20, "%Y-%b-%d, %H:%M", timestruct);

	titleOfWholePlot_buf << "  End: " << timeToText << " UTC";

	titleOfWholePlot = titleOfWholePlot_buf.str();
}

//converts numbers into strings, needed for filenames
void CreateChar (string &SourceBox, string &startYYYYMMDD, string &endYYYYMMDD, Int_t start_Box, Double_t minXax, Double_t maxXax)
{
	// generates the strings needed for generating the strings for plot
	// file name (version with date in name)
	// converts unix timeformat into "readable" strings

	// type long is needed for conversion
	const time_t		t_minXax = (long) minXax;
	const time_t		t_maxXax = (long) maxXax;
	char strfbuf[9];

	// convert the first parameter into a time structure
	tm                 *timestruct = gmtime (&t_minXax);

	// create strings
	strftime (strfbuf, 9, "%Y%m%d", timestruct);
	startYYYYMMDD = strfbuf;

	// convert the first parameter into a time structure
	timestruct = gmtime (&t_maxXax);
	strftime (strfbuf, 9, "%Y%m%d", timestruct);
	endYYYYMMDD = strfbuf;

	SourceBox = boxid2name(start_Box);	// selected target Box
}

//create the root file, to add this filt to the tree
time_t CreateRootFileName (char *sourceFile, time_t act_TimeStamp, Int_t start_Box)
{
	string              srcBox;	// source box

	// XXX special handling of tt02 ("osdorp") and tt04 ("jordaan")
	switch (start_Box)
	{
	case 2:
		srcBox = "osdorp";
		break;
	case 4:
		srcBox = "jordaan";
		break;
	default:
		srcBox = boxid2name (start_Box);
		break;
	}


	char                buffer[50];

	// XXX to be sure to get all data, maybe not really necesary
	time_t              calcTime = act_TimeStamp - 15;

	// convert parameter into a time structure
	tm                 *timestruct = gmtime (&calcTime);

	// create file name (incl. path)
	strcpy (sourceFile, fn_Path);	// fn_Path is global
	strftime (buffer, 45, "/%Y/%m/%d/", timestruct);	// make the file name
	strcat (sourceFile, buffer);

	if (ipVersion == 6)
	{
		strcat (sourceFile, "IPv6.");
	}

	strcat (sourceFile, srcBox.c_str());
	strftime (buffer, 45, ".ripe.net.%Y%m%d.root", timestruct);
	strcat (sourceFile, buffer);

	// 86400 = 60*60*24 = seconds per day
	act_TimeStamp = act_TimeStamp + 86400;

	return act_TimeStamp;
}

//create the file name of the plot - default
void CreatePlotFileNameStd (string &plotFileName, string &baseFileName, Int_t boxNo, const string &SourceBox, plotformat outputFormat)
{
	ostringstream plotFileName_buf;
	string TargetBox = boxid2name(boxNo);	// selected target Box

	if (ipVersion == 6)
	{
		plotFileName_buf << "IPv6.ipdv.";
	}
	else
	{
		plotFileName_buf << "ipdv.";
	}

	plotFileName_buf << SourceBox << "." << TargetBox;

	baseFileName = plotFileName_buf.str();

	// specify output Format - gif=0, ps=1, eps=2
	if (outputFormat == eps )
		plotFileName_buf << ".eps";
	else
		plotFileName_buf << ".ps";
	// gif will be generated from an external program

	plotFileName = plotFileName_buf.str();
}

//create the file name of the plot file  -  with date, coded in file name
void CreatePlotFileNameDate (string &plotFileName, string &baseFileName, Int_t boxNo, const string &SourceBox, const string &startYYYYMMDD, const string &endYYYYMMDD,
			     plotformat outputFormat)
{
	ostringstream plotFileName_buf;

	string TargetBox = boxid2name(boxNo);	// selected target Box

	if (ipVersion == 6)
	{
		plotFileName_buf << "IPv6.ipdv.plot.";
	}
	else
	{
		plotFileName_buf << "ipdv.plot.";
	}
	plotFileName_buf << SourceBox << "." << TargetBox << "." << startYYYYMMDD << "." << endYYYYMMDD;

	baseFileName = plotFileName_buf.str();

	// specify output Format - gif=0, ps=1, eps=2
	if (outputFormat == eps )
		plotFileName_buf << ".eps";
	else
		plotFileName_buf << ".ps";
	// gif will be generated from an external program

	plotFileName = plotFileName_buf.str();
}

//write (and calculate) data in the statistics pad
void WriteStatisticsPad (TPad * pad, Int_t allPackets, Int_t validPackets, Int_t noOfFlaps, Percentiles * ipdvPercentile,
			 Float_t rmsIPDV, Float_t meanDelay, Float_t rmsDelay)
{
	Double_t            x,
	                    y;	// coordinates for printing
	char                buf[30];
	Int_t               align;	// align of the text
	Font_t              font;	// font of the text

	// clear pad
	pad->Clear ();

	// write the title
	x = 0.5;
	y = 0.95;		// starting point
	align = 22;		// horizontally and vertically centered
	font = TATTTEXT_HELVETICA_BOLD_R_NORMAL;		// helvetiva-bold-r - see TAttText class
	write_stats (pad, "STATISTICS", x, y, align, font);	// call
	// subroutine

	// write valid and invalid packets
	y = y - 0.03;		// draw a line to seperate data on pad
	TLine               line1 (0.0, 0.0, 0.0, 0.0);

	line1.SetLineWidth (2);
	line1.SetLineColor (4);
	line1.DrawLine (0.05, y, 0.95, y);

	y -= 0.04;
	write_stats (pad, "Packets sent/valid", x, y, align, font);
	y = y - 0.025;
	sprintf (buf, "Total: %d=100.00%%", allPackets);
	font = TATTTEXT_HELVETICA_MEDIUM_R_NORMAL;		// helvetica-medium-r - see TAttText class
	write_stats (pad, buf, x, y, align, font);

	y = y - 0.025;
	Double_t            percent = validPackets * 100.0 / allPackets;

	sprintf (buf, "Valid: %d=%.2f%%", validPackets, percent);
	write_stats (pad, buf, x, y, align, font);

	y = y - 0.05;
	strcpy (buf, "Number of Routing");
	write_stats (pad, buf, x, y, align, font);

	y = y - 0.025;
	sprintf (buf, "flaps: %d", noOfFlaps);
	write_stats (pad, buf, x, y, align, font);


	// percentiles
	y = y - 0.03;		// draw a line to seperate data on pad
	line1.DrawLine (0.05, y, 0.95, y);

	y -= 0.04;
	font = TATTTEXT_HELVETICA_BOLD_R_NORMAL;		// helvetiva-bold-r - see TAttText class
	write_stats (pad, "IPDV Percentiles", x, y, align, font);

	y = y - 0.03;
	font = TATTTEXT_HELVETICA_MEDIUM_R_NORMAL;		// helvetica-medium-r - see TAttText class
	sprintf (buf, "Mean: %7.2fms", ipdvPercentile->GetLevel (50.0));
	write_stats (pad, buf, x, y, align, font);

	y = y - 0.025;
	sprintf (buf, "RMS:  %7.2fms", rmsIPDV);
	write_stats (pad, buf, x, y, align, font);

	y = y - 0.025;
	sprintf (buf, "50%% > %5.1fms", ipdvPercentile->GetLevel (25.0));
	write_stats (pad, buf, x, y, align, font);

	y = y - 0.025;
	sprintf (buf, "50%% < %5.1fms", ipdvPercentile->GetLevel (75.0));
	write_stats (pad, buf, x, y, align, font);

	y = y - 0.025;
	sprintf (buf, "85.0%% > %5.1fms", ipdvPercentile->GetLevel (7.5));
	write_stats (pad, buf, x, y, align, font);

	y = y - 0.025;
	sprintf (buf, "85.0%% < %5.1fms", ipdvPercentile->GetLevel (92.5));
	write_stats (pad, buf, x, y, align, font);

	y = y - 0.025;
	sprintf (buf, "97.5%% > %5.1fms", ipdvPercentile->GetLevel (1.25));
	write_stats (pad, buf, x, y, align, font);

	y = y - 0.025;
	sprintf (buf, "97.5%% < %5.1fms", ipdvPercentile->GetLevel (98.75));
	write_stats (pad, buf, x, y, align, font);


	// mean and RMS of delay distribution
	y = y - 0.03;		// draw a line to seperate data on pad
	line1.DrawLine (0.05, y, 0.95, y);

	y -= 0.04;
	font = TATTTEXT_HELVETICA_BOLD_R_NORMAL;		// helvetiva-bold-r - see TAttText class
	write_stats (pad, "Delay Distribution", x, y, align, font);
	y = y - 0.025;
	sprintf (buf, "Mean: %6.2fms", meanDelay);
	font = TATTTEXT_HELVETICA_MEDIUM_R_NORMAL;		// helvetica-medium-r - see TAttText class
	write_stats (pad, buf, x, y, align, font);

	y = y - 0.025;
	sprintf (buf, "RMS: %6.2fms", rmsDelay);
	write_stats (pad, buf, x, y, align, font);

}


//********************************************************/
//  function name: write_stats
//  arguments:     *pad        -> pointer to current pad
//                 *ptr_text   -> pointer to string
//                 x           -> x-coordinate NTC
//                 y           -> y-coordinate NTC
//                 align       -> alignment (see ROOT TattText class)
//   font       -> font id   (see ROOT TattText class)
//  purpose:       writes 'ptr_text' on 'pad' as TText on
//                 coordinates 'x' and 'y' with alignment
//                 'align'. Use TLatex-class for a better
//                 correspondence between screen and print-out
//                 in newer versions of ROOT (above 2.21/08)
//  Author:   Johann Gutauer

/**************************************************************/
void write_stats (TPad * pad, const char *ptr_text, const Float_t x, const Float_t y, const Short_t align, Font_t font)
{
	pad->cd ();
	TText              *xlabel = new TText (x, y, ptr_text);

	xlabel->SetTextFont (font);
	xlabel->SetTextColor (1);
	xlabel->SetTextSize (0.11);
	xlabel->SetTextAlign (align);
	xlabel->Draw ();

}


ttmData& ttmData::operator=(const Delay &delay)
{
	Delay *theDelay = const_cast<Delay *>(&delay);
	PutDataInElement(theDelay,this);
	return *this;
}

//put a set of data from a root file into an a ttmData element
void ttmData::PutDataInElement (Delay * delay, ttmData * element)
{
	element->PacketId = delay->GetPacketId ();
	element->SourceId = delay->GetSourceId ();
	// element->SourcePort = delay->GetSourcePort();
	element->TargetId = delay->GetTargetId ();
	// element->TargetPort = delay->GetTargetPort();
	// element->PacketSize = delay->GetPacketSize();
	element->ArrivalTime = delay->GetArrivalTime ();
	element->PacketDelay = delay->GetPacketDelay ();
	// element->SourceClock = delay->GetSourceClock();
	// element->TargetClock = delay->GetTargetClock();
	element->Nhops = delay->GetNhops ();
	// element->RouteId = delay->GetRouteId();
	element->SourceNtp = delay->GetSourceNtp ();
	element->TargetNtp = delay->GetTargetNtp ();


}


//calculates delay variation and errors, put it in histograms, fill percentiles
void CalculateAndFillHistograms (vector<ttmData> &access, TH1F * h1, TH2F * h2, TH2F * h3, TH2F * h4, TH1F * h5, TGraphErrors * ipdvErr,
				 Percentiles * ipdvPercentile, Int_t enableErrorPlot)
{
	// calculate experimantal error of clocks - no square root is
	// calcualted
	Double_t            actError_2,
	                    nextError_2,
	                    ipdvError;

	// error of whole IPDV calculation (per pair)

	Double_t            ipdVariation;	// the IPDV
	Int_t               packetNo = 0;	// needed to fill error plot

	vector<ttmData>::iterator i,j;

	// Set to start of list
	j=access.begin();

	// j should be ahead one element of i
	// and that is what we need
	i=j++;	

	do
	{
		// Dereference, but as references to avoid assignment operator
		// involvement
		ttmData &helper=*i, &nexthelper=*j;

		// IPDV(i) = D(i+1) - D(i) - according to
		// <draft-ietf-ippm-ipdv-07.txt> page 5,
		// see page 8, where it is defined the other way
		// around-informed the authors
		ipdVariation = (nexthelper.PacketDelay) - (helper.PacketDelay);

		// fill the histograms
		h1->Fill (ipdVariation);
		h2->Fill (((helper.ArrivalTime) - convToRootTime), ipdVariation);
		h3->Fill (((helper.ArrivalTime) - convToRootTime), helper.PacketDelay);
		h4->Fill (((helper.ArrivalTime) - convToRootTime), (helper.Nhops) * 10);
		h5->Fill (helper.PacketDelay);

		// add (or fill) percentiles
		ipdvPercentile->Add (ipdVariation);

		// calculate and fill error diagram - if it has been enabled at
		// data input
		if (enableErrorPlot > 0)
		{
			// calcualte errors - faster version
			// expError(packet) = sqrt(SourceNtp^2 + TargetNtp^2);
			// ipdvError = sqrt(actError^2 + nextError^2)
			// no square root on expError, so no "^2" in the
			// ipdvError

			actError_2 = (((helper.SourceNtp) * (helper.SourceNtp)) + ((helper.TargetNtp) * (helper.TargetNtp)));
			nextError_2 =
				(((nexthelper.SourceNtp) * (nexthelper.SourceNtp)) +
				 ((nexthelper.TargetNtp) * (nexthelper.TargetNtp)));
			// ipdvError = sqrt(actError^2 + nextError^2)
			ipdvError = sqrt (actError_2 + nextError_2);

			// fill diagram
			ipdvErr->SetPoint (packetNo, ((helper.ArrivalTime) - convToRootTime), ipdVariation);	// point
			ipdvErr->SetPointError (packetNo, 0.0, ipdvError);
			// error for this point
		}
		i++;j++;
		packetNo++;
	}
	while (j != access.end());
}

// set the histogram title, time format, axis title , ....
void SetHistogramProperties (TH1F * h1, TH2F * h2, TH2F * h3, TH2F * h4, TH1F * h5, TH2F * dummy, char *xTitle, char *xTimeFormat)
{
	// set properties of the histograms
	// histogram for IPDV distribution
	h1->SetMarkerColor (1);
	// h1->SetFillColor(7);
	h1->SetFillColor (4);
	h1->SetStats (kFALSE);	// no statisics displayed on histogram
	h1->GetXaxis ()->SetTitle ("[ms]");
	h1->GetYaxis ()->SetTitle ("[%]    ");

	// histogram for IPDV vs time
	h2->SetMarkerColor (4);
	h2->GetXaxis ()->SetTimeDisplay (1);
	h2->GetXaxis ()->SetTimeFormat (xTimeFormat);
	h2->GetXaxis ()->SetTitle (xTitle);
	h2->GetYaxis ()->SetTitle ("[ms]   ");
	h2->SetStats (kFALSE);

	// histograms for hops and delay vs time
	h3->SetMarkerColor (1);	// delay
	h3->GetXaxis ()->SetTimeDisplay (1);
	h3->GetXaxis ()->SetTimeFormat (xTimeFormat);
	h3->GetXaxis ()->SetTitle (xTitle);
	h3->GetYaxis ()->SetTitle ("[ms]   ");
	h3->SetStats (kFALSE);
	h4->SetMarkerColor (2);	// nhops
	h4->SetStats (kFALSE);

	// histogram for delay ditribution
	h5->SetMarkerColor (2);
	h5->SetFillColor (5);
	h5->SetStats (kFALSE);
	h5->GetXaxis ()->SetTitle ("[ms]");
	h5->GetYaxis ()->SetTitle ("[%]    ");


	// set properties of the error diagram axis ("dummy" is a place holder
	// - TGraphErrors will be created in loop)
	dummy->GetXaxis ()->SetTimeDisplay (1);
	dummy->GetXaxis ()->SetTimeFormat (xTimeFormat);
	dummy->GetXaxis ()->SetTitle (xTitle);
	dummy->GetYaxis ()->SetTitle ("[ms]   ");
	dummy->SetStats (kFALSE);


}

//draw the histograms on canvas c1
void DrawHistograms (TH1F * h1, TH2F * h2, TH2F * h3, TH2F * h4, TH1F * h5, TPad * pad1, TPad * pad2, TPad * pad3, TPad * pad4,
		     Int_t customIPDVPlot, Int_t customDelayPlot)
{
	// first, declaration of some variables
	Int_t               binNo = 0;	// for loops, will contain the number

	// of the bon for zoom operation
	Double_t            limit = 0.0;	// exit condition for loop
	Double_t            sumBins = 0.0;	// to add up the bins in a loop
	Double_t            scale = 0.0;	// factor to normalize the

	// histograms

	if (customIPDVPlot == 0)	// by default - with auto zooming
	{
		// dymanic zoom for the IPDV distribution - displays approx.
		// 99% of the contents
		h1->SetAxisRange (-400, 400);	// needs initalisation, but
		// don't know why
		limit = 0.005 * h1->GetEntries ();

		// only right half will be analyzed, distribution is symetric
		sumBins = 0.0;

		// 801 to get the overflow bin - histogram range is from -400
		// ms to 400 ms
		for (binNo = 801; binNo > 419; binNo--)
		{
			sumBins = sumBins + h1->GetBinContent (binNo);
			if ((sumBins > limit))
				break;
		}		// end finding the dynamic zoom range

		// normalize the histogram with the IPDV distribution
		pad1->cd ();
		// normalize the histogram to percent
		scale = 100 / h1->Integral ();
		h1->Scale (scale);

		// zoom in - dynamic
		h1->SetAxisRange ((400 - binNo), (binNo - 400));

		h1->Draw ();

		// draw the histogram with IPDVs vs time
		pad2->cd ();

		// initialisation - in bin numbers
		h2->GetYaxis ()->SetRange (0, 400);
		if (binNo < 540)
			// zoom in on y-axis - numbers are bin numbers
			h2->GetYaxis ()->SetRange (125, 275);

		h2->Draw ();
	}
	else			// draw customzed IPDV plots
	{
		// normalize the histogram with the IPDV distribution
		pad1->cd ();
		// normalize the histogram to percent
		scale = 100 / h1->Integral ();
		h1->Scale (scale);
		h1->Draw ();

		// draw the histogram with IPDVs vs time
		pad2->cd ();
		h2->Draw ();

	}

	if (customDelayPlot == 0)	// by default - with auto zooming
	{
		// normalize and draw the delay distribution
		pad3->cd ();

		// dynamic zoom, more than 99% of the values (in the bins) will
		//
		// be displayed - don't know if it is really usefull
		// needs initalisation, but don't know why
		h5->SetAxisRange (0, 800);
		limit = 0.01 * h5->GetEntries ();
		sumBins = 0.0;

		// 801 to get the overflow bin
		for (binNo = 801; binNo > 29; binNo--)
		{
			sumBins = sumBins + h5->GetBinContent (binNo);
			if ((sumBins > limit))
				break;
		}		// end finding the dynamic zoom range

		// normalize the histogram to percents
		scale = 100 / h5->Integral ();
		h5->Scale (scale);
		// zoom in - dynamic
		if (binNo < 750)
			h5->SetAxisRange (0, (binNo + 10));
		h5->Draw ();

		// draw the histograms with the delays(h3) and hops(h4)
		pad4->cd ();

		// initialisation - numbers are bin numbers
		h3->GetYaxis ()->SetRange (0, 400);

		// zoom in - dynamic
		if (binNo < 250)
			// zoom in on y-axis - bin numbers
			h3->GetYaxis ()->SetRange (0, 125);
		else
			// zoom in on y-axis - bin numbers
			h3->GetYaxis ()->SetRange (0, 200);

		// just in case
		if (binNo > 390)
			// numbers are bin numbers
			h3->GetYaxis ()->SetRange (0, 400);

		h3->Draw ();
		h4->Draw ("sameP");
	}
	else			// customized delay plots
	{

		// normalize and draw the delay distribution
		pad3->cd ();

		// normalize the histogram to percents
		scale = 100 / h5->Integral ();
		h5->Scale (scale);
		h5->Draw ();

		// draw the histograms with the delays(h3) and hops(h4)
		pad4->cd ();

		h3->Draw ();
		h4->Draw ("sameP");

	}

}


//draw the error plot on canvas c2
void DrawErrorPlot (TH2F * dummy, TGraphErrors * ipdvErr, TPad * errorPad)
{
	errorPad->cd ();

	dummy->Draw ();		// just to get the axes
	ipdvErr->SetMarkerStyle (25);
	ipdvErr->SetMarkerSize (0.25);
	ipdvErr->SetMarkerColor (3);
	ipdvErr->Draw ("P");	// drawn without axes


}

//create a HTML file to link the plots from the web site - if it has been selected at data input
void CreateHTMLFile (Percentiles * ipdvPercentile, const char *anyString, Int_t boxNo, ofstream & IntoFile, Int_t paragraphID)
{
	// paragraphID = 0 - write the first line of the file
	if (paragraphID == 0)
	{
		IntoFile << "<html>" << endl << "<head>" << endl;
		IntoFile << "<title>Delay Variations - Testbox " << anyString << "</title>" << endl;
		IntoFile << "</head>" << endl << "<body>" << endl;
		IntoFile << "<br><table width=795 border=0><tr align=center><td><h2>from Testbox " << anyString <<
			" to </h2></td></tr></table>" << endl;
		// IntoFile << "<table width=795 border=2 cellpadding=5>" <<
		// endl; //Opera doesn't like this style
		IntoFile << "<table border=2 cellpadding=5>" << endl;
		IntoFile <<
			"<colgroup> <col width=55> <col width=105> <col width=105> <col width=105> <col width=105> <col width=105> <col width=105> <col width=105> </colgroup>"
			<< endl;

		// write table headline
		IntoFile << "<tr align=center>";
		IntoFile << "<td>tt</td>";	// box number
		IntoFile << "<td>mean</td>";	// mean
		IntoFile << "<td>50%<br>lower edge</td>";	// 50% lower
								// edge
		IntoFile << "<td>50%<br>upper edge</td>";	// 50% upper
								// edge
		IntoFile << "<td>85%<br>lower edge</td>";	// 85% lower
								// edge
		IntoFile << "<td>85%<br>upper edge</td>";	// 85% upper
								// edge
		IntoFile << "<td>97.5%<br>lower edge</td>";	// 97.5% lower
								// edge
		IntoFile << "<td>97.5%<br> upper edge</td>";	// 97.5% upper
								// edge
		IntoFile << "</tr>" << endl;

	}

	// paragraphID = 1 - add a line of data to the file
	if (paragraphID == 1)
	{
		// manipulate file name from *.eps to *.gif
		char                buffer[250] = "";

		// cut "ps" from filename
		strncat (buffer, anyString, (strlen (anyString) - 2));
		strcat (buffer, "gif");	// and add "gif"

		// create a line for the table on the website
		IntoFile << "<tr align=center>";
		IntoFile << "<td><a href=" << buffer << ">" << boxNo << "</a></td>";	// box
											// number
		IntoFile << "<td>" << ipdvPercentile->GetLevel (50.0) << " ms</td>";	// mean
		IntoFile << "<td>" << ipdvPercentile->GetLevel (25.0) << " ms</td>";	// 50%
											// lower
											// edge
		IntoFile << "<td>" << ipdvPercentile->GetLevel (75.0) << " ms</td>";	// 50%
											// upper
											// edge
		IntoFile << "<td>" << ipdvPercentile->GetLevel (7.5) << " ms</td>";	// 85%
											// lower
											// edge
		IntoFile << "<td>" << ipdvPercentile->GetLevel (92.5) << " ms</td>";	// 85%
											// upper
											// edge
		IntoFile << "<td>" << ipdvPercentile->GetLevel (1.25) << " ms</td>";	// 97.5%
											// lower
											// edge
		IntoFile << "<td>" << ipdvPercentile->GetLevel (98.75) << " ms</td>";	// 97.5%
											// upper
											// edge
		IntoFile << "</tr>" << endl;

	}

	// paragraphID = 2 - add last line to the file
	if (paragraphID == 2)
	{
		IntoFile << "</table>" << endl << "</body>" << endl << "</html>" << endl;
	}

}


//write percentiles to file - if it has been selected at data input
void SavePercentiles (Percentiles * ipdvPercentile, const char *SourceBox, Int_t boxNo, ofstream & IntoFile, Int_t firstLine)
{
	// firstLine = 1 - write the first line of the file
	if (firstLine > 0)
	{
		IntoFile << "# percentiles of all boxes connected to testbox " << SourceBox << endl << endl;
		IntoFile << "#testbox    mean  50%_lower 50%_upper 85%_lower 85%_upper 97.5%_lower 97.5%_upper" << endl << endl;

	}
	// firstLine = 0 - add a line of data to the file
	else
	{
		IntoFile.width (6);
		IntoFile << boxNo;	// box number
		IntoFile.width (10);
		IntoFile << ipdvPercentile->GetLevel (50.0);	
		IntoFile.width (10);
		// 50% lower edge
		IntoFile << ipdvPercentile->GetLevel (25.0);	
		IntoFile.width (10);
		// 50% upper edge
		IntoFile << ipdvPercentile->GetLevel (75.0);	
		IntoFile.width (10);
		// 85% lower edge
		IntoFile << ipdvPercentile->GetLevel (7.5);	
		IntoFile.width (10);
		// 85% upper edge
		IntoFile << ipdvPercentile->GetLevel (92.5);	
		IntoFile.width (10);
		// 97.5% lower edge
		IntoFile << ipdvPercentile->GetLevel (1.25);	
		IntoFile.width (10);
		// 97.5% upper edge
		IntoFile << ipdvPercentile->GetLevel (98.75);	
		IntoFile << endl;
	}
}

//check if the parameters for the custom plots make sense (and the program doesn't crash)
float CheckPlotParameter (float &maxIPDV, float &maxCustomDelay, float &minCustomDelay)
{

/*	*maxIPDV > 0
 *	*minCustomDelay >= 0
 *	*minCustomDelay < *maxCustomDelay
 */
	// program should run with these values (hope so)
	if ((maxIPDV > 0) && (minCustomDelay >= 0) && (minCustomDelay < maxCustomDelay))
		return (1);
	else			
	{
		if (maxIPDV < 0)
			maxIPDV = 2;
		if (minCustomDelay < 0)
			minCustomDelay = 0;
		if (maxCustomDelay < minCustomDelay)
			maxCustomDelay = minCustomDelay + 10;	// no better
		// ideas
	}
	return (1);
}

const string& boxid2name(uint id)
{
	TestBox *tbentry = tbConfig->GetBoxById(id);
	if(tbentry == NULL)
	{
		cerr << "ERROR: No such testbox id " << id << endl;
		throw range_error("Testbox ID out of range");
	}
	return tbentry->GetShortName();
}
