#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include "packet.h"
#include "testbox.h"
#include "querydb.h"
#include "assign_vectors.h"
#include "merge_packets.h"

static char const rcsid[] = "$Id: assign_vectors.c,v 1.7 2007/02/06 11:59:54 ruben Exp $";

/* 
 * Subroutine header
 *
 * Input: debug flag, source box and UNIX time_t timestamp at start of day 
 *        ipflag which is either IPv4 (FALSE) or IPv6 (TRUE)
 *
 * Output: list_of_packets[] array with updated route information
 *         from SQL tracerouteDB
 */

void assign_vectors (const char *mysql_host, int debug, int srcid, time_t starttime, int ip6flag) {

	int i;
	int match;
	int matched=0;
	int matched6=0;
	int numpack;
	time_t packettime;

	int source, target;
        GHashTable* traceroute_data;

	MYSQL mysql;
	MYSQL mysql_ip6;

	/* sanity check */
	if (srcid <= 0) {
		fprintf(stderr,"%d: invalid testbox id\n", srcid);
		return;
	}

	source = srcid;

	if (!ip6flag) {

		/* Initialize connection to MySQL traceroute DB */
		open_user_connection(&mysql, mysql_host);

		/* Cache all traceroute info originating at this source on this day */
        	if (!(traceroute_data = build_data (&mysql,source,starttime)))
        	{
                	fprintf (stderr,"[IPv4] Could not build Data Hash for box %d at time_t %d ?\n", srcid, starttime);
                	fprintf (stderr,"   skipping traceroute assignments ....\n");
                	return;
        	}

		numpack = number_of_packets;

	} else {
	
		open_user_connection_ip6(&mysql_ip6, mysql_host);	

		if (!(traceroute_data = build_data (&mysql_ip6,source,starttime)))
		{
                	fprintf (stderr,"[IPv6] Could not build Data Hash for box %d at time_t %d ?\n", srcid, starttime);
                	fprintf (stderr,"   skipping traceroute assignments ....\n");
                	return;
        	}                

		numpack = number_of_packets6;
	}	

	/* Loop over all packets */
	for (i=0; i<numpack ; i++) {

		if (debug == TRUE && ((i % (numpack / 5)) == 0 ))
			setproctitle("Assigning Routing vectors for tt%02d [%3d %%]",srcid, (i * 100 / numpack));

		/* Valid packet ? */
		if (!ip6flag && (list_of_packets[i] != NULL)) {

			source = list_of_packets[i]->source_id ;
			if (source != srcid) {
				fprintf(stderr, "WARNING source id %d found while processing IPv4 data for box %d\n", source, srcid);
				continue; /* next packet */
			}

			target = list_of_packets[i]->target_id ;

			/* get packet's timestamp from the packet id
 			   Note that this will produce wrong results
			   if (for testing) serial number are *not* 
			   set to the time packet was send out */
             
			packettime = id2time(list_of_packets[i]->packet_id,starttime);  

			match = match_vector(traceroute_data, target, packettime,
				&list_of_packets[i]->routeid,  &list_of_packets[i]->nhops);
			if (match) matched++;
		}
		if (ip6flag && (list_of_packets6[i] != NULL)) {

                        source = list_of_packets6[i]->source_id ;
                        if (source != srcid) {
                                fprintf(stderr, "WARNING source id %d found while processing IPv6 data for box %d\n", source, srcid);
                                continue; /* next packet */
                        }

                        target = list_of_packets6[i]->target_id ;

                        /* get packet's timestamp from the packet id
                           Note that this will produce wrong results
                           if (for testing) serial number are *not*
                           set to the time packet was send out */

                        packettime = id2time(list_of_packets6[i]->packet_id,starttime);

                        match = match_vector(traceroute_data, target, packettime,
                                &list_of_packets6[i]->routeid,  &list_of_packets6[i]->nhops);
                        if (match) matched6++;
                }
	}

	if (debug==TRUE) {
		setproctitle(NULL);
		printf ("Assigned routing vectors to %d IPv4 and %d IPv6 packets\n", matched, matched6);
	}

	if (!ip6flag) mysql_close(&mysql);
		else mysql_close(&mysql_ip6);

	return;

}

// Function id2time()
//
// Purpose: convert packetid to Unix time stamp
// Comment: upto 27 June 2002 packet id equaled Unix time stamp of scheduled 
//          delay measurement. After that, to support higher frequencies
//          than 1 per second, packet id counts ticks of a 100 Hz clock,
//          starting at 0 on 1/1/2002 and cycles after 1577880000 ticks
//          (about six months)

time_t id2time(unsigned int packetid, time_t starttime) {

#define LASTUNIXTIME 1025189400    // Thu Jun 27 14:50:00 2002 GMT
#define TIMEZERO     1009843200    // Tue Jan  1 00:00:00 2002 GMT
#define CYCLEPERIOD  20000000      // cycle period in seconds
#define FREQUENCY    100.0         // force floating point calculation

        time_t packettime;
        int nCycles;
        unsigned int starttimeId;     // packetid at starttime

        if (starttime <= LASTUNIXTIME) {
                // no two cycles in delay plot
                if (packetid > LASTUNIXTIME) {
                        packettime = packetid / FREQUENCY + TIMEZERO;
                }
                else {
                        packettime = packetid;
                }
        }
        else {
                // all ids in new packetid schema
                nCycles = (starttime - TIMEZERO) / CYCLEPERIOD;
                starttimeId = (starttime - TIMEZERO - nCycles * CYCLEPERIOD) * FREQUENCY ;
                if (packetid < starttimeId) {
                        // must have completed a cycle 
                        nCycles++;
                }
                packettime = packetid / FREQUENCY + nCycles * CYCLEPERIOD + TIMEZERO;
        }
        return(packettime);
}


