#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "testbox.h"
#include "packet.h"
#include "merge_packets.h"
#include "assign_vectors.h"

static char const rcsid[] = "$Id: merge_packets.c,v 1.12 2007/02/06 12:04:52 ruben Exp $";

/*
 * Subroutine header
 *
 * Purpose: calculate UNIX time_t at start of given day
 *
 */

time_t          date_to_timestamp(char *date);

time_t date_to_timestamp(char *date)
{

	int             year,
	                month,
	                days;
	struct tm       tmstruct;

	if ((strlen(date) == 8) && (sscanf(date, "%4u%2u%2u", &year, &month, &days) == 3))
	{

		tmstruct.tm_year = year - 1900;
		tmstruct.tm_mon = month - 1;
		tmstruct.tm_mday = days;

		tmstruct.tm_hour = 0;
		tmstruct.tm_sec = 0;
		tmstruct.tm_min = 0;
		return (mktime(&tmstruct));
	}
	else
	{
		fprintf(stderr, "ERROR: %s: invalid date\n", date);
		exit(1);
	}
}

/*
 * ------------------------------------------------------------------
 * Subroutine header
 * 
 * Main program for merge_packets
 * 
 * ------------------------------------------------------------------
 */

int main(int argc, char *argv[])
{

	FILE           *fileout;
	int             i,
	                j;
	int             iarg = 1;
	char            ch;
	int             date = 0;
	time_t          timestamp = time(NULL);
	int             verbose = FALSE;
	const char     *sourcetb = "tt01.ripe.net";
	const char     *TB_file = "./LIST_OF_TESTBOXES";
	const char     *PCKB_area = "../data";
	const char     *SNDP_area = "../data";
	const char     *MYSQL_host = "localhost";

	int             ipflag = 0;

	/* just to make sure date/time calculations use the right timezone */
	putenv("TZ=GMT+0");

	/* Set to yesterday */
	timestamp -= ((timestamp % 86400) + 86400);

	/* Command line arguments */
	if (argc <= 1)
	{
		printf("Usage: program [flags] files to process\n");
		printf(" -d            Date to process (YYYYMMDD)\n");
		printf(" -s            Source tb (ttXX.ripe.net)\n");
		printf(" -v            Verbose output for debugging\n");
		printf(" -L filename   name of file with test-box data, default %s\n", TB_file);
		printf(" -P directory  name of directory for PCKB files, default %s\n", PCKB_area);
		printf(" -M filename   name of mysql tracerouteDB server, default %s\n", MYSQL_host);
		printf(" -S directory  name of directory with SNDP files, default %s\n", SNDP_area);
		return FALSE;
	}
	while ((ch = getopt(argc, argv, "L:M:P:S:d:s:v")) != -1)
	{
		switch (ch)
		{
		case 'd':
			date = atoi(optarg);
			timestamp = date_to_timestamp(optarg);
			break;
		case 'L':
			TB_file = optarg;
			break;
		case 'M':
			MYSQL_host = optarg;
			break;
		case 'P':
			PCKB_area = optarg;
			break;
		case 's':
			sourcetb = optarg;
			break;
		case 'S':
			SNDP_area = optarg;
			break;
		case 'v':
			verbose = TRUE;
			break;
		default:
			printf("Unknown option %c\n", ch);
			return FALSE;
			break;
		}
		if (optind > iarg)
		{
			iarg = optind;
		}
	}

	if (date == 0)
	{
		printf("Date should be > 0\n");
		return FALSE;
	}
	/* Read test-box data */
	if (read_test_box_config(ASCII, TB_file, verbose) == FALSE)
	{
		return FALSE;
	}
	else
	{
		if (verbose == TRUE)
		{
			printf("Test-box data read from %s\n", TB_file);
		}
	}

	number_of_packets = 0;
	number_of_packets6 = 0;

	/* Read existing PCKN file for this day */
	read_existing_packets(PCKB_area, sourcetb, date, BINARY, verbose);

	/* Read new packets for this day */
	read_new_packets(SNDP_area, sourcetb, date, verbose);

	if (number_of_packets == 0)
	{
		fprintf(stderr, "%s: no IPv4 packets send on date %d ?\n", sourcetb, date);
	}
	if ((verbose == TRUE) && (number_of_packets6 == 0))
	{
		fprintf(stderr, "%s: no IPv6 packets send on date %d ?\n", sourcetb, date);
	}
	if (number_of_packets || number_of_packets6)
	{
		/* Read data for received packets */
		for (i = iarg; i < argc; i++)
		{
			if (!ipflag && strstr(argv[i], "RCDP6"))
				ipflag = 1;
			setproctitle("Reading %s data for %s [%3d %%]", ipflag ? "RCDP6" : "RCDP", sourcetb, (i * 100) / argc);
			read_received_packets(argv[i], verbose);
		}
	}
	if (!ipflag)
	{

		/* Count # packets with a valid time-stamp */
		j = 0;
		for (i = 0; i < number_of_packets; i++)
		{
			if (list_of_packets[i]->arrival_time < 0.0)
				j++;
		}
		if (verbose == TRUE)
		{
			printf("[IPv4] Arrival time < 0 for %d out of %d\n", j, number_of_packets);
		}
		/* Write packets */
		fileout = open_packet_file(BINARY, PCKB_area, sourcetb, date, 0);

		/* Merge packets and vectors */
		setproctitle("Assigning IPv4 routing vectors for %s from %d", sourcetb, date);
		assign_vectors(MYSQL_host, verbose, test_box_id(sourcetb), timestamp, 0);

		setproctitle("Writing PCKB for %s from %d", sourcetb, date);
		for (i = 0; i < number_of_packets; i++)
		{
			if (list_of_packets[i] != NULL)
			{
				write_packet_data(fileout, list_of_packets[i]);
				if (verbose == TRUE)
				{
					print_packet(list_of_packets[i]);
				}
				free(list_of_packets[i]);
				list_of_packets[i] = NULL;
			}
		}
	}
	else
	{

		/* Count # packets with a valid time-stamp */
		j = 0;
		for (i = 0; i < number_of_packets6; i++)
		{
			if (list_of_packets6[i]->arrival_time < 0.0)
				j++;
		}
		if (verbose == TRUE)
		{
			printf("[IPv6] Arrival time < 0 for %d out of %d\n", j, number_of_packets6);
		}
		/* Write packets */
		fileout = open_packet_file(BINARY, PCKB_area, sourcetb, date, 1);

		/* Merge packets and vectors */
		setproctitle("Assigning IPv6 routing vectors for %s from %d", sourcetb, date);
		assign_vectors(MYSQL_host, verbose, test_box_id(sourcetb), timestamp, 1);

		setproctitle("Writing PCKB6 for %s from %d", sourcetb, date);
		for (i = 0; i < number_of_packets6; i++)
		{
			if (list_of_packets6[i] != NULL)
			{
				write_packet_data(fileout, list_of_packets6[i]);
				if (verbose == TRUE)
				{
					print_packet(list_of_packets6[i]);
				}
				free(list_of_packets6[i]);
				list_of_packets6[i] = NULL;
			}
		}
	}

	close_new_packet_file(fileout);
	setproctitle(NULL);

	return (0);
}
