// Ascii4ans - retrieve data for the ANS <-> RIPE NCC comparison
//
// chain all files for given date
// retrieve data for all connections between tt01, tt02, tt23, tt31 and tt28
// (RIPE, ANS, CERN, SLAC)
//
// output in ascii:
//
//   srcId trgtId timestamp  delay (micro secs)  nhops  routing-vector#
//     1     23   93000000     10000.0            3        25    (Valid packet)
//     1     23   93000001        -2.0            0         0    (Lost packet)
//
// Author: Rene Wilhelm


#include "TROOT.h"
#include "TChain.h"
#include "TFile.h"

#include "Delay.h"


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <errno.h>
#include <sys/types.h>

#define RIPE1  1
#define RIPE2  2
#define ANS   23
#define SLAC  28
#define CERN  31
// #define NBOXES 5 
#define NBOXES 3

#define CONFIGFILE "/ncc/ttpro/config/LIST_OF_TESTBOXES"
#define ROOTDATADIR  "/ncc/ttpro/root_data"
#define OUTPUTDIR  "/ncc/ttpro/spool"
#define MAXCHARS   2048


char *progname;
int errno;
int verbose;

// forward declaration
void writePacket (FILE *outfile, Delay *packet);
int matchHosts(const int nbox, Int_t boxlist[], Int_t sourceId, Int_t targetId,
	Int_t &srcfound, Int_t &trgtfound);

// Usage
void usage () {
   fprintf(stderr, "Usage: %s <date (yyyymmdd)>\n", progname);
   exit(1);
}
	
//--------------
// MAIN PROGRAM 
//--------------
main (int argc, char **argv) {
 
	extern char *optarg;
	extern int optind;

//	Int_t boxlist[NBOXES] = { RIPE1, RIPE2, ANS, CERN, SLAC };
	Int_t boxlist[NBOXES] = { RIPE1, RIPE2, ANS };
	FILE *outfile[NBOXES][NBOXES];
	int ch, i, j;

	int day, month, year;
	char *datestring, *hostname, *filename;
	char *suffix, pathname[MAXCHARS];

	struct dirent *direntry;
	Int_t sourceId, targetId, eventNo;

	TROOT simple ("Simple", "Example");

	Int_t split = 1;
	Int_t bsize = 64000;
	Delay *delay = 0;


	// For the input file.
	FILE  *fp;
	DIR   *dirp;
	TFile *hfile;
	int    eventsfound, files, accept;

	verbose = 0;
	progname=argv[0];

	while ((ch = getopt(argc, argv, "v")) != -1) {
		switch(ch) {
	   		case 'v':
	         		verbose = 1;    
	    			break;
	   		default:
	      			usage();
		}
	}
	argc -= optind;
	argv += optind;

	// only one argument may be left to process now: date!

	if (argc != 1) {
		usage();
	}

	// decode date

	datestring=argv[0];
	if (strlen(datestring) != 8 ||
	    sscanf(datestring, "%4d%2d%2d", &year, &month, &day) != 3) {
		fprintf (stderr,"%s: ERROR decoding datestring %s\n", progname, argv[1]);
	 	usage();
	}


	// open directory with per host ROOT files (info on data _send_ by host)

	sprintf(pathname, "%s/%4.4d/%2.2d/%2.2d", ROOTDATADIR, year, month, day);
	if ((dirp = opendir(pathname)) == NULL) {
		fprintf(stderr, "%s: opening directory %s: %s\n", progname,
		pathname, strerror(errno));
		exit(1);
	}
	
	// open ASCII output files for all permutations of sender & receiver

	for (i=0; i<NBOXES; i++) {
		for (j=0; j<NBOXES; j++) {
			if (i == j) {
				continue;
			}

			sprintf(pathname, "%s/tt%2.2d.tt%2.2d.%s.txt",
				OUTPUTDIR, boxlist[i], boxlist[j], datestring);

			if ((outfile[i][j] = fopen (pathname, "w")) == NULL) {
				fprintf(stderr, "%s: output file %s: %s\n",
					progname, pathname, strerror(errno));
				exit(1);
			}
		}
	}

	// chain all .root files in directory 

	TChain chain("tree");
	while ((direntry = readdir(dirp)) != NULL) {
		filename=direntry->d_name;
		if (((suffix=strrchr(filename, '.')) != NULL) && 
		    (strcmp(suffix, ".root") == 0)) {
			sprintf(pathname, "%s/%4.4d/%2.2d/%2.2d/%s",
			    ROOTDATADIR, year, month, day, filename);
			chain.Add(pathname);
		}
	}
	closedir(dirp);

        delay = new Delay();
	chain.SetBranchAddress("delay", &delay);
	Int_t nTotalEvents = chain.GetEntries();
	if (verbose == 1) {
		printf("Total #entries in %s chain: %d\n", datestring, nTotalEvents);
	}

	// loop over all events, write output files

        eventsfound=0;
	for (eventNo=0; eventNo < nTotalEvents; eventNo++) {

	        // read next delay into memory 

		if ( (chain.GetEvent(eventNo) < 0) ) {
			// problem reading this event ?
            		continue;
        	}

		// if source or target matches: add to output 

		sourceId = delay->GetSourceId();
		targetId = delay->GetTargetId();
		if (matchHosts(NBOXES, boxlist, sourceId, targetId, i, j)) {
			writePacket(outfile[i][j], delay);
        		eventsfound++;
		}
	}
	if (eventsfound == 0) {
		fprintf(stderr, "%s: no events found !?\n", progname);
	}
	else if (verbose == 1) {
		printf("%d events found for %s\n", eventsfound, datestring);
	}

	return 0;
} 


/*
-------------------------------------------------------------------------------
Subroutine Header
Purpose           :     write Packet delay info to file (ascii format)
Side Effects      :
Comments          :
*/

void writePacket (FILE *outfile, Delay *packet) {
	
	// write packet data to output file

	if (packet->ClocksOK()) {
		// Valid packet
		if (verbose) {
			fprintf(stderr,"Valid packet\n");
		}
		fprintf(outfile, "%3d%3d%10d\t%9.1f\t%2d\t%9d\n",
			packet->GetSourceId(), packet->GetTargetId(),
			packet->GetPacketId(), packet->GetPacketDelay()*1000,
			packet->GetNhops(), packet->GetRouteId());
	}
	else if ((packet->GetArrivalTime() < 0.0)) {
		// Negative Arrvialtime == packet lost
		if (verbose) {
			fprintf(stderr,"LOST packet\n");
		}
		fprintf(outfile, "%3d%3d%10d\t%9.1f\t%2d\t%9d\n",
			packet->GetSourceId(), packet->GetTargetId(),
			packet->GetPacketId(), -2.0, 0, 0);
	}
}

/*
-------------------------------------------------------------------------------
Subroutine Header
Purpose           :     check if both source and target ID are present
		  :	in the list of boxes to create output for.
Side Effects      :
Comments          :	returns the positions in boxlist[] of the
		  : 	sourceId and TargetId
*/

int  matchHosts(const int nboxes, Int_t boxlist[], Int_t sourceId,
	Int_t targetId, Int_t &srcfound, Int_t &trgtfound) {
	
	
	srcfound=-1;
	trgtfound=-1;
	for (Int_t i=0; i<nboxes; i++) {
		if (boxlist[i] == sourceId) {
			srcfound=i;
		}
		if (boxlist[i] == targetId) {
			trgtfound=i;
		}
	}

	return ((srcfound != -1) && (trgtfound != -1));
}
static char const   rcsid[] = "$Id: ascii4ans.cxx,v 1.2 2006/11/04 14:40:50 ruben Exp $";
static char const  *rcsid_p = rcsid;    // Prevent g++ from optimizing out rcsid
