#------------------------------------------------------------------------------
# Module Header
# Filename          : impmincateg
# Purpose           : Imports 'minimum billing categories' from file
# Author            : Roderik
# Date              : 8 December 1997
# Description       : Reads minimum billing categories for a bunch of
#                     registries from a specified file, and inserts them into
#                     'min-categ' fields, with the specified year added.
#                     The format of the file should be as output by rchange(1)
#                     (columns separated by \s+, first = registry, 4th = cat.)
# Language Version  : Perl (Tested with v5.00401)
# OSs Tested        : BSD/OS 2.1
# Command Line      : reg -x impmincateg year filename
# Input Files       : (file specified on command line)
# Output Files      : $REGDIR/* (usually /ncc/registries/*)
# External Programs : perl module regwrite.pm
# Problems          :
# To Do             :
# Comments          : This is a 'reg x-script' which can be called with
#                     'reg -x <scriptname>. However, Why not just 'use'
#                     the needed modules in here too? Now it can also be 
#                     called directly if we want...
#                     (Note: 'use strict' is necessary here because the strict
#                     directive apparently doesn't survive a 'do <filename>'.
# The following line will be read by 'reg -x':
# reg-x-description: import minimum billing categories (from 'rchange' file)
#------------------------------------------------------------------------------

## Many comments added here; this script is used as an example template
## for making a (not complicated) change to the registry data.

## About accessing and changing registry data:
## The data is in a hash, with the attribute names as keys and the
## registry data as values.
## You get passed a pointer to the hash with registry data.
## You access hash-values via a pointer to the hash in the following way:
## $pointer->{hashkey}     (i.e. in our case that's $regp->{'attribute'})
##
## Newline has been stripped off the end of the data.
## Multiple attributes with the same name are concatenated, separated by
## newline(\n). Within a multi-line attribute, lines are separated by
## formfeed (\n).
## So for example you'd get all reg-logs in an array with
##                                   @tmp = split(/\n/,$regp->{'reg-log'})
## Now $tmp[0] could still contain multiple lines; if you wanted to print it,
## you'd stilll need to do           $tmp[0] =~ s/\f/\n/g 
## before a                          print "$tmp[0]\n"        .


#-- initialisation

## use regread.pm or regwrite.pm (depending on whether you only want to read
## the data, or read & write it. regwrite uses regread itself.)
use strict;
use regwrite;

my $year = shift @ARGV or die "usage: reg -x impmincateg year filename\n";
my $file = shift @ARGV or die "usage: reg -x impmincateg year filename\n";
die "usage: reg -x impmincateg year filename\n" if @ARGV;
die "impmincateg: year should be > 1900\n" if $year < 1901;
die "impmincateg: $file is not a file\n" if !-f $file;
die "impmincateg: cannot read file $file\n" if !-r $file;


#-- real work

## Changing registry data is done best through changereg(), which does 
## the low-level work (RCS handling, handling errors when you can't
## read/write something) for you.
## You need to supply changereg() with the changes that have to be made:
## you need to define a subroutine doing the changes, and pass that to
## changereg().


##  Step 1:
## Define a subroutine (or, actually, a pointer to an anonymous subroutine)
## which will actually do the changes in the registry data.
## This will be called from changereg() with as arguments:
## - a pointer to the hash containing registry data
## - any extra arguments that you have passed to changereg().
my $changeit = sub {
    my ($regp,$year,$mincateg) = @_;
    
    # See if any line of the 'min-categ' attribute starts with the specified 
    # year.
    if ($regp->{'min-categ'} =~ /^$year/m) {

	## Error situation.
	## We can choose to do 2 things:
	## - pass a string containing an error message back to changereg().
	##   A message cointaining your error will be printed, and your
	##   main routine will get this same message passed back.
	## - pass 'undef' back to changereg(). Nothing will be printed, and
	##   your main routine will get a numeric error code passed back.
	##   (For more info, see /usr/local/src/regwrite.pl, at changereg().)
	
	# Here, I do the second thing & print the error message myself.
        print STDERR "$regp->{'regid'}: a value for $year is already present\n";
	return undef;
    }

    # Add attribute (separate from existing one by newline, if applicable)
    $regp->{'min-categ'} .= "\n" if $regp->{'min-categ'};
    $regp->{'min-categ'} .= "$year $mincateg";

    ## You should return the same pointer to the (changed) registry data
    ## to changereg().
    return $regp;
};
# Don't forget this semicolon here... :)


##  Step 2:
## The main routine. You should call changereg() from here, with as arguments:
## - registry ID
## - 0 (why? That's why. If you're really interested, see
##      /usr/local/src/regwrite.pl, at changereg().)
## - The pointer to your subroutine which actually changes the reg-data
## - any additional arguments that you want to pass to your subroutine

# In this case (and in many cases), there's a while{} loop around the call to
# changereg(), because we want to change many registries.

my $flag = 0;
open(F,$file) or die "open $file: $!\n";
while (<F>) {
    next if /^Registry/;
    last if /^\s+$/;

    my @tmp = split(/\s+/);
    # $tmp[0] = registry ID
    # $tmp[3] = size
    if ($tmp[3] !~ /^(ENTERPRISE|SMALL|MEDIUM|LARGE)$/) {
	warn "$tmp[0]: Cannot deal with new size '$tmp[3]'\n";
    }
    
    ## Here we go...
    ## changereg() returns an error code, or 0 for success.
    ##  (it might also return a text-string, if you returned
    ##  a text-string from your routine.)
    # We have 2 additional arguments here; the year & minimum category.

    unless (changereg($tmp[0],0,$changeit,$year,$tmp[3])) {
	print "$tmp[0] - $tmp[3]\n";
    }
}
