#------------------------------------------------------------------------------
# Module Header
# Filename          : maill
# Purpose           : Re-generates registry mailinglists
# Author            : Roderik/Daniel
# Date              : 14 November 1997 (& earlier)
# Description       : Reads mailinglist attributes from all registries in
#                     hashes (one hash for each list; address = key,
#                     associated registry = value) and writes them to files
# Language Version  : Perl (Tested with v5.00401)
# OSs Tested        : BSD/OS 2.1
# Command Line      : reg -x maill
# Input Files       : info extracted from $REGDIR/* (usually /ncc/registries/*)
# Output Files      : $LISTDIR/{local-ir}ncc-co|ip-provs}.reg
# External Programs : perl module regread.pm
# Problems          :
# To Do             :
# Comments          : This is a 'reg x-script' which can be called with
#                     'reg -x <scriptname>. However, Why not just 'use'
#                     the needed modules in here too? Now it can also be 
#                     called directly if we want...
#                     (Note: 'use strict' is necessary here because the strict
#                     directive apparently doesn't survive a 'do <filename>'.
# The following line will be read by 'reg -x':
# reg-x-description: fully re-generate mailinglists
#------------------------------------------------------------------------------

use strict;
use regread;

my ($LSTDIR) = 'LSTDIR-CONF';

my ($regp);		# a registry
my (@f);		# registry names
my ($f,$i);
my (%localir);
my (%contrib);
my (%provs);

@f = regfiles(0);
while ($f = shift(@f)) {
    $regp = readreg($f);

    next if $regp->{'type'} !~ /^(PROVIDER|ENTERPRISE|FYI)$/ && $regp->{'bill-hack'} ne 'NORMAL';
    foreach $i (split(/\n/, $regp->{'lst-localir'})) {
	$i = &regwrite::normmadr($i);
	$localir{$i} .= "$f ";
    }
    foreach $i (split(/\n/, $regp->{'lst-contrib'})) {
	$i = &regwrite::normmadr($i);
	$contrib{$i} .= "$f ";
    }

    # slightly different rules for ip-provs list...
    next unless  $regp->{'type'} eq 'PROVIDER';
    ($i,) = servicelevel($regp);
    next unless $i eq 'NORMAL';
    foreach $i (split(/\n/, $regp->{'lst-provs'})) {
	$i = &regwrite::normmadr($i);
	$provs{$i} .= "$f ";
    }
    
}
&maill_write('local-ir.reg', \%localir);
&maill_write('ncc-co.reg', \%contrib);
&maill_write('ip-provs.reg', \%provs);


sub maill_write {	# write out a mailing list
    my ($f, $listp) = @_;

    # Schema for preventing 2 simultaneous updates: first create link
    # from [list].updating to [list]. This will fail/wait/try again as
    # long as [list].updating exists (meaning another update process
    # is running).
    # When you've created this link, you're safe that another reg
    # process won't interrupt you. Unlink [list], build new one, and
    # remove [list].updating.

    until (link "$LSTDIR/$f", "$LSTDIR/$f.updating") {
	if (-f "$LSTDIR/$f.updating") {
	    print STDERR "Temporary file $LSTDIR/$f.updating exists.\nIt shouldn't, or only for a very short period,\nwhile another reg process is updating mailinglists.\nPausing 2 seconds; I hope it has disappeared then...\n";
	    sleep 2;
	} else {
	    warn "open $LSTDIR/$f.updating: $!\n";
	    return 1;
	}
    }
    unlink("$LSTDIR/$f") or do {warn "unlink $LSTDIR/$f: $!\nWatch out; still existing hard link $f.updating -> $f will cause trouble!\n"; return 1};
    open(F, ">$LSTDIR/$f") or do {warn "open $LSTDIR/$f: $!\nWatch out; still existing file $f.updating will cause trouble!\n"; return 1};


    foreach $i (sort bydomain (keys(%$listp))) {
	chop($listp->{$i});
	print F "$i ($listp->{$i})\n";
    }
    close(F);

    chmod(0664, "$LSTDIR/$f") or warn "chmod 0644 $LSTDIR/$f: $!\n";
    unlink("$LSTDIR/$f.updating") or do {warn "unlink $LSTDIR/$f.updating: $!\nWatch out; still existing file $f.updating will cause trouble!\n"; return 1};
}

sub bydomain {	# sort routine for mail addresses
	my(@a, @b);

	@a = split(/@/, $a);
	@b = split(/@/, $b);

	@a = reverse(split(/\./, pop(@a)));
	@b = reverse(split(/\./, pop(@b)));

	return (join('.', @a) cmp join('.', @b));
}
