# Copyright (c) 1993, 1994, 1995, 1996, 1997  The TERENA Association
#
# All Rights Reserved
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose and without fee is hereby granted,
# provided that the above copyright notice appear in all copies and that
# both that copyright notice and this permission notice appear in
# supporting documentation, and that the name of the author not be
# used in advertising or publicity pertaining to distribution of the
# software without specific, written prior permission.
#
# THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
# ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS; IN NO EVENT SHALL
# AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY
# DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
# AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

# $Id: regread.pm,v 1.21 2004/10/21 08:35:29 ttraffic Exp $
#------------------------------------------------------------------------------
# Module Header
# Filename          : regread.pm
# Purpose           : Provide read functions for internal registry database
# Author            : Roderik Muit
# Date              : 5 November 1997
# Description       :
# Language Version  : Perl (Tested with v5.00401)
# OSs Tested        : BSD/OS 2.1
# Command Line      : -
# Input Files       : files read are in $REGDIR/* (usually /ncc/registries/*)
# Output Files      : -
# External Programs :
# Problems          :
# To Do             : replace 2-digit year by 4-digit
# Comments          : Rewrite (perl5, use strict, split up in modules)
#                     from original reg program written by Daniel F. Karrenberg
#                     & maintained by me later on. The whole idea and half
#                     the code is his.
#                     Old source code can be found in reg.sh
# Comments          : There's one not-exported & not-referenced-in-this-file
#                     subroutine in here, that's because it looks so much like
#                     the previous one (easier to maintain?)  
#------------------------------------------------------------------------------

# Comment on data structure: registry data is one big associated array
# (per registry), with attribute-names as keys and attribute-values as values.
# Multiple attribute-values for the same attribute are separated by \n
# in the value. Multiple lines within an attribute-value are separated by \f
# (so first split on \n, then on \f)
# Properties & syntax of attribute-names are defined in the hash %tags.
# Data is stored in plain textfile.
#------------------------------------------------------------------------------

require 5.003;

package regread;
use Exporter ();
use strict;
use vars       qw(@ISA @EXPORT $REGDIR @tags %tags %money);

@ISA       = qw(Exporter);
@EXPORT    = qw($REGDIR @tags %tags %money &regfiles &findreg &readreg &servicelevel &getcontacts);

# Def. variables that are global to this package (but not exported)
#
my ($WHOISC) = '/usr/local/bin/whois	'; # WHOIS binary path (&ngetcontacts)

# Def. variables to be exported
#
$REGDIR = '/ncc/ttpro/ttreg/data';

@tags = (  # this array  determines the ordering in forms, output etc
'error',
'id',
'ttid',
'org',
'address',
'city',
'country',
'admin-c',
'ship-addr',
'phone',
'fax-no',
'e-mail',
'remark', 
'nagios-prnt',
'nagios-grp',
'audit-trail',
'full-upd',
'lst-hosts',
'lst-members',
'-I',
'chain',
'ipaddr',
'ipaddrx1',
'ipaddrx2',
'ipaddrx3',
'ipaddrx4',
'ifnamex1',
'ifnamex2',
'ifnamex3',
'ifnamex4',
'old-ipaddr',
'old1ipaddr',
'old2ipaddr',
'old3ipaddr',
'old4ipaddr',
'gateway',
'netmask',
'dnsserv',
'monitor',
'monitor6',
'syslog',
'alert-run',
'alert-on',
'alert6-on',
'alert-ign',
'alert-lvl',
'alert6ign',
'alert6lvl',
'ipaddr6',
'ipaddr6x1',
'ipaddr6x2',
'ipaddr6x3',
'ipaddr6x4',
'ifname6x1',
'ifname6x2',
'ifname6x3',
'ifname6x4',
'old-ipaddr6',
'old1ipaddr6',
'old2ipaddr6',
'old3ipaddr6',
'old4ipaddr6',
'gateway6',
'prefixlen6',
'dnsserv6',
'mailalrm',
'dnsalias',
'group',
'sw-version',
'clockhw',
'geoloc',
'status',
'status-old',
'state-log',
'ttops-log', 
'ttops-alert',
'next-hop',
'next6-hop',
'o-delay-t',
'o6delay-t',
'u-delay-t',
'u6delay-t',
'r-delay-t',
'r6delay-t',
'o-tracer-t',
'o6tracer-t',
'u-tracer-t',
'u6tracer-t',
'r-tracer-t',
'r6tracer-t',
'o-route-t',
'o6route-t',
'u-route-t',
'u6route-t',
'r-route-t',
'r6route-t',
'port-number',
'auxmr',
'rsync',
'-B',
'bill-addr',
'bill-mail',
'bill-ref', 
'bill-vatno',
'bill-debno',
'bill-proto',
'bill-scheme',
'bill-remark', 
'bill-remind',
'bill-signup',
'bill-hack',
'bill-first',
'bill-status',
'invoice', 
'payment',
'hard-sold',
'serv-sold',
'bill-log',
'bill-stat',  
'-E',
);

# %tags describes the attributes
# The value is to be split at tabs! into $opt and $regex
#
# options:  
#
#	P = public attribute (published in registry list) (basically implies F)
#	F = in registry form (-f)
#           (Note: there's 3 situations: P+F = published on FTP site,
#            F= not on FTP site but in form that's sent to registry
#            neither = NCC private)
#	M = mandatory attribute
#	S = single attibute per object
#	T = multi-line attributes possible (tab)
#	D = needs date (yymmdd)
#	W = needs initials (who)
#            W only set when D also set. (otherwise, &printreg needs change)

#	B = billing data (-B)
#	I = ttreg data (-I)
#	L = mailing list data (-L)
#	R = registry data (-R)

# regex: value has to match for syntax check

%tags = (
'id',    'PFMIRBS	^[0-9][0-9][0-9][0-9]$',
'ttid',    'PFMSIRB	^tt[0-9][0-9]+$',
'chain',    'PFI	^\S+$',
'ipaddr',    'PFMI	^(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})$',
'ipaddrx1',    'PFI	^(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}\/\d{1,2})$',
'ipaddrx2',    'PFI	^(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}\/\d{1,2})$',
'ipaddrx3',    'PFI	^(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}\/\d{1,2})$',
'ipaddrx4',    'PFI	^(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}\/\d{1,2})$',
'ifnamex1',    'PFI',
'ifnamex2',    'PFI',
'ifnamex3',    'PFI',
'ifnamex4',    'PFI',
'old-ipaddr',    'PFI	^(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})$',
'old1ipaddr',    'PFI	^(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}\/\d{1,2})$',
'old2ipaddr',    'PFI	^(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}\/\d{1,2})$',
'old3ipaddr',    'PFI	^(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}\/\d{1,2})$',
'old4ipaddr',    'PFI	^(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}\/\d{1,2})$',
'gateway',    'PFMI	^(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})$',
'netmask',    'PFMI	^(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})$',
'dnsserv',    'PFI	^(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})$',
'monitor',    'PFI	^(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}\/\d{1,2})$',
'monitor6',    'PFI	^([0-9A-Fa-f]{0,4}:([0-9A-Fa-f]{0,4}:){1,6}[0-9A-Fa-f]{1,4}\/\d{1,3})$',
'syslog',     'PFI	^((\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}|[a-zA-Z]+[\w\-\.]+):\d+\/(kern|user|mail|daemon|auth|syslog|lpr|news|uucp|cron|authpriv|ftp|local0|local1|local2|local3|local4|local5|local6)\.(emerg|alert|crit|err|warning|notice|info|debug))$',
'alert-run',   'PFI',
'alert-on',   'PFI',
'alert6-on',   'PFI',
'alert-ign',   'PFI',
'alert-lvl',   'PFI',
'alert6ign',   'PFI',
'alert6lvl',   'PFI',
'ipaddr6',    'PFI	^([0-9A-Fa-f]{0,4}:([0-9A-Fa-f]{0,4}:){1,6}[0-9A-Fa-f]{1,4})$',
'ipaddr6x1',    'PFI	^([0-9A-Fa-f]{0,4}:([0-9A-Fa-f]{0,4}:){1,6}[0-9A-Fa-f]{1,4}\/\d{1,3})$',
'ipaddr6x2',    'PFI	^([0-9A-Fa-f]{0,4}:([0-9A-Fa-f]{0,4}:){1,6}[0-9A-Fa-f]{1,4}\/\d{1,3})$',
'ipaddr6x3',    'PFI	^([0-9A-Fa-f]{0,4}:([0-9A-Fa-f]{0,4}:){1,6}[0-9A-Fa-f]{1,4}\/\d{1,3})$',
'ipaddr6x4',    'PFI	^([0-9A-Fa-f]{0,4}:([0-9A-Fa-f]{0,4}:){1,6}[0-9A-Fa-f]{1,4}\/\d{1,3})$',
'ifname6x1',    'P',
'ifname6x2',    'P',
'ifname6x3',    'P',
'ifname6x4',    'P',
'old-ipaddr6',    'PFI	^([0-9A-Fa-f]{0,4}:([0-9A-Fa-f]{0,4}:){1,6}[0-9A-Fa-f]{1,4})$',
'old1ipaddr6',    'PFI	^([0-9A-Fa-f]{0,4}:([0-9A-Fa-f]{0,4}:){1,6}[0-9A-Fa-f]{1,4}\/\d{1,3})$',
'old2ipaddr6',    'PFI	^([0-9A-Fa-f]{0,4}:([0-9A-Fa-f]{0,4}:){1,6}[0-9A-Fa-f]{1,4}\/\d{1,3})$',
'old3ipaddr6',    'PFI	^([0-9A-Fa-f]{0,4}:([0-9A-Fa-f]{0,4}:){1,6}[0-9A-Fa-f]{1,4}\/\d{1,3})$',
'old4ipaddr6',    'PFI	^([0-9A-Fa-f]{0,4}:([0-9A-Fa-f]{0,4}:){1,6}[0-9A-Fa-f]{1,4}\/\d{1,3})$',
'gateway6',    'PFI	^([0-9A-Fa-f]{0,4}:([0-9A-Fa-f]{0,4}:){1,6}[0-9A-Fa-f]{1,4})$',
'prefixlen6',	'PFI	^\d+$',
'dnsserv6',    'PFI	^(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})$',
'mailalrm',    'FI	^\S+@\S+$',
'dnsalias',    'I	^\S+$',
'group',	'MS	^group[A-Z]$',
'sw-version',	'MS	^\d+$',
'clockhw',	'MS	^(gps|cdma)$',
'geoloc',    'I ^-?\d{1,3}\.\d{4}\s+\+\/-\s+\d\.\d{4},\s+-?\d{1,3}\.\d{4}\s+\+\/-\s+\d\.\d{4}$',
'status',    'MIS	^(SETUP|OFF|WATCH|ON|HISTORIC|RESERVED)$',
'status-old',    'IS	^(SETUP|OFF|WATCH|ON|HISTORIC|RESERVED)$',
'org',    'PFMSI',
'address',    'PFMIT',
'ship-addr',    'PFIT',
'city',       'PFISM',
'country',    'PFMS	^[A-Z]{2}(\s+[A-Z]{2})*$',
'admin-c',	'PFM',
'phone',    'PFM	\+\d',
'fax-no',    'PF	\+\d',
'e-mail',    'PFM	^(\S+@\S+|< none >)$',
'remark',    'PFT',
'nagios-prnt', 'P',
'nagios-grp', 'P',
'audit-trail', 'P',
'full-upd',	'PFT',
'lst-hosts',	'PFT',
'lst-members',	'PFT',
'state-log',    'TD	^(9\d{5}|\d{8})\s+',
'ttops-log',    'TDW	^(9\d{5}|\d{8})\s+',
'ttops-alert',    'TDW	^(9\d{5}|\d{8})\s+',
'next-hop',		'P',
'next6-hop',	'P',
'o-delay-t',	'P',
'o6delay-t',    'P',
'u-delay-t',    'P',
'u6delay-t',    'P',
'r-delay-t',    'P',
'r6delay-t',    'P',
'o-tracer-t',   'P',
'o6tracer-t',   'P',
'u-tracer-t',   'P',
'u6tracer-t',   'P',
'r-tracer-t',   'P',
'r6tracer-t',   'P',
'o-route-t',   'P',
'o6route-t',   'P',
'u-route-t',   'P',
'u6route-t',   'P',
'r-route-t',   'P',
'r6route-t',   'P',
'port-number',  'P',
'auxmr',      'P',
'rsync', 	'PFMI	^(no|yes)$',
'bill-proto',    'BFS	^(E-MAIL( ASCII)?|SNAIL-MAIL)$',
'bill-scheme',    'BFS	^(YEARLY|HALF-YEARLY|QUARTERLY)( LATE)?( USD)?( SPECIAL) ?$',
'bill-vatno',    'BFS',
'bill-ref',    'BFS',
'bill-debno',    'BS	^\d{6}$',
'bill-signup',    'BS	^\d{8}\s+\d+$',
'bill-hack',    'BS	^(NORMAL\s+\S+.*|NONE\s+\S+.*)$',
'bill-addr',    'BFT',
'bill-mail',    'BF	^\S+@\S+$',
'bill-remark',    'BFTW',
'bill-remind',	'TDW	^\d{8}\s+',
'bill-first', 'BFS	^\TT\d+',
'bill-status', 'BFS	^(YES|NO)(\w*\/.+)?',
'invoice',    'BD	^\d{8}\s+\d+([,\.]\d{2})?\s+(TT\d{4}-\d{3}|tt\d{6})(\s+\w+.*)?$',
'payment',    'BD	^\d{8}\s+\d+([,\.]\d{2})?\s+(TT\d{4}-\d{3}|tt\d{6})(\s+\w+.*)?$',
'hard-sold',  'BF       ^\d+\s+\d+\s+\d+\s+\d+\s+(YES|NO)$',
'serv-sold',  'B        ^\d+\s+\d{8}\s+\d+$',
'bill-log',    'BTDW	^\d{8}\s+',
'bill-stat',    'BSD	^\d{8}\s+START\s+\S+',
'error',  'T'
);

%money = (
    '95 SMALL',      2000,
    '95 MEDIUM',       6000,
    '95 LARGE',     12000,
    '95 ENTERPRISE',   1000,
    '95 NCCBUDGET',  529000,
    '95 SIGNUPFEE',    2000,
    '95 BILLCHARGE',     65,
    '96 SMALL',      1500,
    '96 MEDIUM',       4500,
    '96 LARGE',      8500,
    '96 ENTERPRISE',   1000,
    '96 NCCBUDGET',  916000,
    '96 SIGNUPFEE',    2000,
    '96 BILLCHARGE',     65,
    '97 SMALL',      2200,
    '97 MEDIUM',       3000,
    '97 LARGE',        4000,
    '97 ENTERPRISE',   2200,
    '97 NCCBUDGET', 1984000,
    '97 SIGNUPFEE',    1300,
    '97 BILLCHARGE',     50,
    '98 SMALL',        2450,
    '98 MEDIUM',       3400,
    '98 LARGE',        4500,
    '98 ENTERPRISE',   2450,
    '98 NCCBUDGET', 3457000,
    '98 SIGNUPFEE',    2000,
    '98 BILLCHARGE',     50,
);

#
# exported functions
#

#------------------------------------------------------------------------------
# Subroutine Header
# Purpose           : gives back list of all registry IDs (or selection)
# Side Effects      :
# Comments          :
#
# in: flag
#   if flag is false, return all
#   if flag==1, then return all enterprises, providers and other NORMALs
#   (This is really slow, so hardly used. Only for maillist generation.)
# out: list containing reg-ids. Empty list on error.
#
sub regfiles ($) {

    my($flag) = @_;
    my($f);
    my(@f);

    opendir(REGDIR, $REGDIR) or  do {warn "opendir $REGDIR:$!\n"; return ()};
    
    if (!$flag) {
	@f = sort(grep(/^tt[0-9][0-9]+$/, readdir(REGDIR)));
    } else {
	foreach $f (sort(grep(/^tt[0-9][0-9]+$/, readdir(REGDIR)))) {
	    open(F, "$REGDIR/$f") or do {warn "open $f: $!\n"; next};
	    while (<F>) {
		if (/^type:\s+(.*)$/ and $1 =~ /^(ENTERPRISE|PROVIDER|FYI)$/) {
		    push(@f,$f);
		    last;
		}
	    }
	    close(F);
	}
    }
    closedir(REGDIR);
    return @f;
};

#------------------------------------------------------------------------------
# Subroutine Header
# Purpose           : Finds registry ID based on possibly non-exact match
# Side Effects      : may print on STDERR & exit if no/multiple matching
#                     regIDs.
# Comments          : exiting is weird for a function in a module - 
#                     but I don't see any real objections so am leaving it...
#
# in: string
# out: Same string, if that string exactly matches a registry ID
#      The registry ID, if that string matches exactly one registry ID
#      No return if string matches 0 or >1.
#
sub findreg ($) {
    my($str) = @_;
    my(@f);
    my($f);

#	@f = grep(/$str/, &regfiles(1));
#	@f = grep(/$str/, &regfiles(0)) if @f == 0;
# this is too slow

    @f = grep(/$str/, regfiles(0));
    # exact match: return that one.
    if (grep(/^$str$/, @f)) {
	return($str);
    } elsif (@f == 0) {
	print STDERR "$str: no match\n";
	exit 1;
    } elsif (@f == 1) {
	return($f[0]);
    }
		
    while ($f = shift(@f)) {
	open(F, "$REGDIR/$f") or do {warn "open $f: $!\n"; next};
	printf STDERR "%-16.16s - ", $f;
	while (<F>) {
	    if (/org:\s+(.*)$/) {
		print STDERR $1;
	    }
	}
	print "\n";
	close F;
    }
    exit 2;
}

#------------------------------------------------------------------------------
# Subroutine Header
# Purpose           : read registry data from file & check/fill error
# Side Effects      : 'error' attribute may be manipulated
# Comments          :
#  A note about the 'error' attribute:
#  There are 2 kinds of errors:
#  - the ones that ignore a certain line. They should not be deleted 
#    automatically because the ignored line is nowhere else in the data anymore
#    These errors will start with '** ignored: '
#  - other ones that report syntax errors. They will be deleted on reading,
#    and possibly regenerated while checking. (They might not be there anymore,
#    though; for instance after editing a registry with errors.)
#
# Comments          : Input value can be 2 things: regID (which will be
#                     expanded to a filename) or already fully qualified
#                     filename. (Will assume second if '/' in input value)
#
# in: reg-id or filename
# out: pointer to hash with data; or undef if something went wrong
#
sub readreg ($) {
    my($f) = @_;
    my($t, $v, $regp, $opt, $regex,$tmp);

    $tmp = ($f=~/\// ? $f : $REGDIR.'/'.$f);
    open(F, $tmp) or do {warn "open $f: $!\n"; return undef;};

    # read registry data
    while (<F>) {
	next if /^#/;
	next if /^\s*$/;

	if (/^(\S+):\s+(.*)$/) {
	    ($t = $1) =~ tr/A-Z/a-z/;
	    $v = $2;
	    $v =~ s/\s+$//;
	    next if ($tags{$t}=~/^\S*D/ && $v =~ /^\d{8}\s*$/);
	    next if ($tags{$t}=~/^\S*D/ && $tags{$t}=~/^\S*W/ && $v =~ /^\d{8}\s+\S+\s*$/);
	    $regp->{$t} .= "\n" if $regp->{$t};
	    $regp->{$t} .= $v;
	} elsif ($tags{$t}=~/^\S*T/ && /^\s+(.*)$/) {
	    $regp->{$t} .= "\f" if $regp->{$t};
	    $regp->{$t} .= $1;
	} elsif (/^\s*(.*)\s*$/) {
	    $regp->{'error'} .= "\n" if $regp->{'error'};
	    $regp->{'error'} .= "** ignored: encountered strange line";
	    $regp->{'error'} .= " after \"$t:\"" if $t;
	    $regp->{'error'} .= ":\f\"$1\"\f(DELETE THIS ERROR AFTER FIXING!)";
	}
    }
    close F;

    # delete all non-retraceable errors
    $t = $regp->{'error'};
    $regp->{'error'} = '';
    foreach $tmp ( split(/\n/,$t)) {
	$regp->{'error'} .= $tmp."\n" if $tmp =~ /^\*\* ignored:/;
    }
    # remark: From here on until the end of readreg, $regp->{'error'} will
    # (be the only attribute to) have \n at its end.

    # check all existing attributes (former checkreg from here)
    foreach $t (@tags) {
	next if $t =~ '^-';
	($opt, $regex) = split ("\t", $tags{$t});
	
	if ($opt =~ /M/ && !$regp->{$t}) {
	    $regp->{'error'} .= "mandatory attribute '$t' missing.\n";
	}
	if ($opt =~ /S/ && $regp->{$t} =~ /\n/) {
	    $regp->{'error'} .= "attribute '$t' is single line.\n";
	}
	if ($regex && $regp->{$t}) {
	    foreach $v (split("\n", $regp->{$t})) {
		if ($v !~ /$regex/) {
		    $regp->{'error'} .= "'$t: $v' syntax error, syntax is /$regex/\n";
		}
	    }
	}
    }

    # checks per attribute
    # regid should be equal to filename (only if no complete filename given)
    if ($f !~ /\//) {
	$regp->{'error'} .= "registry ID is not equal to filename\n" if $f ne $regp->{'ttid'};
    }
    # old-regid & new-regid should exist
    if ($regp->{'old-regid'}) {
	$regp->{'error'} .= "Old registry '$regp->{'old-regid'}' does not exist\n" if !-f "$REGDIR/$regp->{'old-regid'}";
    }
    if ($regp->{'new-regid'} =~ /\d+\s+(\S+)/) {
	$regp->{'error'} .= "New registry '$1' does not exist\n" if !-f "$REGDIR/$1";
    }

    # heuristic changes per attribute
    $regp->{'country'} =~ tr/a-z/A-Z/;

    # check for non-existing attributes
    while (($t, $v) = each %$regp) {
	next if defined $tags{$t};
	$regp->{'error'} .= "** ignored: unknown attribute name\f\"$t: $v\"\f(DELETE THIS ERROR AFTER FIXING!)\n";
	delete $regp->{$t};
    }
    
    chomp $regp->{'error'};
    return ($regp);
}

#------------------------------------------------------------------------------
# Subroutine Header
# Purpose           : return servicelevel
# Side Effects      :
# Comments          : output servicelevel can be 'NONE' or 'NORMAL'
#
# in: pointer to registry data
# out: servicelevel, comments
# warning: assumes working directory == registry database!!
#
sub servicelevel ($);
sub servicelevel ($) {
	
    my ($regp) = @_;
    my ($lvl, $com);

    if ($regp->{'bill-hack'} =~ /^NORMAL\s+(.*)/) {
	$lvl = 'NORMAL';
	$com = $1;
    } elsif ($regp->{'bill-hack'} =~ /^NONE\s+(.*)/) {
	$lvl = 'NONE';
	$com = $1;
    } elsif ($regp->{'bill-parent'}) {
	# this does not check against infinite 'bill-parent loops'...
	if (-r "$REGDIR/$regp->{'bill-parent'}") {
	    ($lvl, $com) = servicelevel(readreg($regp->{'bill-parent'}));
	    $com .= "parent is $regp->{'bill-parent'}";
	} else {
	    $lvl = 'NONE';
	    $com = "orphaned by $regp->{'bill-parent'}";
	}
    } elsif ($regp->{'type'} eq 'PENDING') {
	$lvl = 'NONE';
	$com = 'PENDING';
    } elsif ($regp->{'type'} eq 'CLOSED') {
	$lvl = 'NONE';
	$com = 'CLOSED';
    } elsif ($regp->{'type'} eq 'FYI') {
	$lvl = 'NONE';
	$com = 'FYI';
    } elsif ($regp->{'commit'}) {
	$lvl = 'NORMAL';
	$com = '';
    } else {
	$lvl = 'NONE';
	$com = 'cannot determine any status';
    }
    return($lvl, $com);
}

# pubservicelevel is called from &regwrite::printreg only - but left here
# so it's close to &servicelevel
# The only difference is, the output strings are slightly different.
#
# (Watch out; there's another servicelevel-like routine. That's in the
# 'stats x-script'.)
sub pubservicelevel ($) { 
	
    my ($regp) = @_;
    my ($lvl, $com);

    if ($regp->{'bill-hack'} =~ /^NORMAL\s+(.*)/) {
	$lvl = 'NORMAL';
	$com = $1;
    } elsif ($regp->{'bill-hack'} =~ /^NONE\s+(.*)/) {
	$lvl = 'NO SERVICE!';
	$com = 'Please contact <billing@ripe.net> for details.';
    } elsif ($regp->{'bill-parent'}) {
	# this does not check against infinite 'bill-parent loops'...
	if (-r "$REGDIR/$regp->{'bill-parent'}") {
	    ($lvl, $com) = &pubservicelevel(readreg($regp->{'bill-parent'}));
	    $com .= '';
	} else {
	    $lvl = 'NONE';
	    $com = 'Please contact <billing@ripe.net> for details.';
	}
    } elsif ($regp->{'type'} eq 'PENDING') {
	$lvl = 'PENDING';
	$com = 'Please contact <ncc@ripe.net> for details.';
    } elsif ($regp->{'type'} eq 'FYI') {
	$lvl = 'not served by the NCC';
	$com = '';
    } elsif ($regp->{'commit'}) {
	$lvl = 'NORMAL';
	$com = "";
    } else {
	$lvl = 'NO SERVICE!';
	$com = 'Please contact <billing@ripe.net> for details.';
    }
    return($lvl, $com);
}

#------------------------------------------------------------------------------
# Subroutine Header
# Purpose           : Get person templates of contacts from RIPE Database
# Side Effects      :
# Comments          :
#
# called from: showreg, writetemp, writepub
# in: array containing admin-c tech-c
# out: 1 scalar containing all lines with whois data
#
sub getcontacts {

    my (@contacts)=@_;
    my (%contseen);
    my ($contact,$key,$ret);

    foreach $contact (@contacts) {
	next if $contact =~ /^\s*$/;
	if ($contact =~ /\(\s*(\S+)\s*\)/) {
	    $key = $1;
	}
	else {
	    $key = $contact;
	}
	$contact =~ y/A-Za-z0-9-//;
	next if $contseen{$key};
	open(F, "$WHOISC -T person $key|") or do {warn "whois: $!\n"; return ''};
	$ret .= "$contact\n\n";
	while (<F>) {
	    if (/No entries found/) {
		$ret .= "No information found.\n";
		last;
	    }
	    $ret .= $_;
	}
	$ret .= "\n";
	close(F);
	$contseen{$key} = 1;
    }
    return($ret);
}
#EOF
