# Copyright (c) 1993, 1994, 1995, 1996, 1997  The TERENA Association
#
# All Rights Reserved
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose and without fee is hereby granted,
# provided that the above copyright notice appear in all copies and that
# both that copyright notice and this permission notice appear in
# supporting documentation, and that the name of the author not be
# used in advertising or publicity pertaining to distribution of the
# software without specific, written prior permission.
#
# THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
# ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS; IN NO EVENT SHALL
# AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY
# DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
# AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

# $Revision: 1.1 $ %D%
#------------------------------------------------------------------------------
# Module Header
# Filename          : regwrite.pm
# Purpose           : Provide write functions for internal registry database
# Author            : Roderik Muit
# Date              : 5 November 1997
# Description       :
# Language Version  : Perl (Tested with v5.00401)
# OSs Tested        : BSD/OS 2.1
# Command Line      : -
# Input Files       : info read from $REGDIR/* (usually /ncc/registries/*)
# Output Files      : same files.
# External Programs : uses regread.pm
# Problems          :
# To Do             :
# Comments          : Rewrite (perl5, use strict, split up in modules)
#                     from original reg program written by Daniel F. Karrenberg
#                     & maintained by me later on. The whole idea and half
#                     the code is his.
#                     Old source code can be found in reg.sh
#------------------------------------------------------------------------------

# Comment on data structure: registry data is one big associated array
# (per registry), with attribute-names as keys and attribute-values as values.
# Multiple attribute-values for the same attribute are separated by \n
# in the value. Multiple lines within an attribute-value are separated by \f
# (so first split on \n, then on \f)
# Properties & syntax of attribute-names are defined in the hash %tags.
#------------------------------------------------------------------------------
# Comment on programming style: I just made up this convention:
#   Exported Functions imported from separate modules I don't prepend with '&'
#   Subroutines in this filename I do.
#   For strange reasons there are sometimes subroutines that I call in
#    separate modules that are not in the modules' @EXPORT list; for this
#    I use the fully specified subroutine name (prepended with '&')
#------------------------------------------------------------------------------

require 5.003;

package regwrite;
use Exporter ();
use strict;
use regread;
use vars qw(@ISA @EXPORT);

@ISA       = qw(Exporter);
@EXPORT    = qw(&changereg &editreg &newreg &printreg &rev_checkin &rev_checkout);

# Def. variables that are global to this package (but not exported)
#
my ($TMPDIR) = '/tmp';
my ($RCSDIR) = 'RCSDIR-CONF';                    # RCS path
my ($PUBDIR) = 'PUBDIR-CONF'; # public FTP data
my ($LSTDIR) = 'LSTDIR-CONF';   # mailing lists
my ($LIBDIR) = 'LIBDIR-CONF';       # mailing lists
my ($NEWREGSTATS) = '/ncc/docs/office/new-registries';

#
# exported functions
#

#------------------------------------------------------------------------------
# Subroutine Header - changereg
# Purpose           : gives back list of all registry IDs (or selection)
# Side Effects      :
# Comments          :
#
#
# You should call this subroutine if you want to change something in reg.
# It does the necessary pre- and post-checks for you.
# In between reading/checking and writing reg-data, it calls
# your-subroutine with as arguments:
# - a pointer to the hash containing registry data
# - optionally, the 4th-last argument to changereg
#
# Your-subroutine should change something in the hash and return the pointer 
# to it, or something else if something went wrong.
# If you return the undefined value, &changereg will return error code 5
# and be silent. If you return anything else (but a pointer), a warning
# message containing whatever you just returned will be generated on STDERR
# and &changereg will return that value.
#
# in: - filename
#     - continue-flag
#     - (pointer to) your-subroutine
#     - any arguments to pass to subroutine
# returnvalue: 0=OK
#              1=cannot read registry file
#              2=cannot write registry file
#              3=error on checkout
#              4=registry data had errors, not changing (if !continue-flag)
#              5=something went wrong when changing registry data
# Note on continue-flag:
#  If continue-flag == 0, then changereg will return without
#  changing anything if &checkreg revealed errors in the registry data.
#  (This basically is: if $regp->{'error'} is defined after &checkreg.)
#  If continue-flag == 1, then changereg will continue, but still print out
#  the errors on STDERR.
#  Otherwise, changereg will continue silently.
#  So normally, you would set this flag to 0 unless you're sure what
#  you're doing. (Like in &editreg you set the flag to 2, because you just
#  want to display the "error" attributes in the users' editor and let them
#  decide what to do with it.)

sub changereg ($$$@) {

    my($f,$contflag,$subroutine,@subargs) = @_;
    my(%old);
    my($regp, $tmp);

    rev_checkout("$REGDIR/$f") and return 3;
    if (!-w "$REGDIR/$f") {
	warn "cannot write $f\n";
	rev_checkin("$REGDIR/$f");
	return 2;
    }

    $regp = readreg($f);
    if (!defined($regp)) {
	rev_checkin("$REGDIR/$f");
	return 1;
    }
    if ($regp->{'error'}) {
	if ($contflag < 2) {
	    $tmp = $regp->{'error'};
	    $tmp =~ s/\f/\n   /g;
	    warn "Registry $f has errors:\n$tmp\n\n";
	}
	if ($contflag == 0) {
	    warn "Aborting.\n";
	    rev_checkin("$REGDIR/$f");
	    return 4; 
	}
    }

    %old = %$regp;
	
    $regp = &$subroutine($regp,@subargs);
    if (!defined($regp)) {
	rev_checkin("$REGDIR/$f");
	return 5;
    }
    unless (ref($regp)) {
	warn "Error in routine changing registry data (it returned '$regp')\n";
	rev_checkin("$REGDIR/$f");
	return $regp;
    }

    $tmp = printreg("$REGDIR/$f",$regp,'e');
    rev_checkin("$REGDIR/$f");
    return 4 if $tmp;

    &checkafterchanging($f,$regp,\%old);
    return 0;
}
sub checkafterchanging ($$$) {
    # This is not an exported function, but I just put it here because it's
    # a continuation of changereg. In a separate routine so newreg can also
    # use this.
    my ($f,$regp,$oldp) = @_;
    my ($tmp, $t, $localir, $contrib, $provs, $nlocalir, $ncontrib, $nprovs);
    my ($oldlevel, $level);
    ($oldlevel) = servicelevel($oldp);
    ($level) =  servicelevel($regp);
    print STDERR "==> Service Level Changed from $oldlevel to $level.\n" if $oldlevel ne $level;

    return if $Getopt::Std::opt_r;


    # If this looks like a newly set up registry:
    # Write to statistics file, and prompt to make invoice.
    # (Actually I'm not too happy about having this in the changereg routine.
    #  But I don't want to spend time reprogramming.
    #  Just crossing my fingers and hoping that nothing except 'reg -e'
    #  will ever invoke this piece of code.)
    if ($level eq 'NORMAL' and $oldlevel eq 'NONE') {
	my $tmp;
	if ($oldp->{'type'} eq 'PENDING' || !$oldp->{'commit'}) {
	    print STDERR "\nFor the statistics:\n";
	    do { $tmp = &prompt("Is this a newly set up registry (y/n)?" ,'y');
	     } until $tmp =~ /^[yn]/i;
	    if ($tmp =~ /^y/i) {
		if (!open (F,">>$NEWREGSTATS")) {
		    warn "open $NEWREGSTATS for append: $!\n";
		} else {
		    print F &thisdate," $regp->{'regid'}\n";
		    close F;
		}
	    }
	}
	do { $tmp = &prompt("Generate invoice (y/n)?" ,'y');
	 } until $tmp =~ /^[yn]/i;
	# If yes, execute genbill and never return.
	# We don't have to do anything here anymore anyway.
	exec '/usr/local/bin/genbill',$regp->{'regid'} if $tmp =~ /^y/i;
    }


    # Write public entry (judgment call: as this consumes some time, I first
    # introduce some extra overhead to see whether something's actually
    # changed.)
    foreach $t (@tags) {
	if ($tags{$t} =~ /^\S*P/ and $regp->{$t} ne $oldp->{$t}) {
	    $tmp = 1;
	    last;
	}
    }
    if ($tmp) {
	print STDERR "updating $f on FTP site...\n";
	writepub($f, $regp);
    }

    # write maillists (first check if something's changed, because this 
    # definately takes too long otherwise, and is not necessary in most cases)
    if ($oldp->{'type'} =~ /^(PROVIDER|ENTERPRISE|FYI)$/ or $oldp->{'bill-hack'} eq 'NORMAL') {
	$localir = $oldp->{'lst-localir'};
	$contrib = $oldp->{'lst-contrib'};
    }
    $provs = $oldp->{'lst-provs'} if $oldp->{'type'} eq 'PROVIDER';
    
    if ($regp->{'type'} =~ /^(PROVIDER|ENTERPRISE|FYI)$/ or $regp->{'bill-hack'} eq 'NORMAL') {
	$nlocalir = $regp->{'lst-localir'};
	$ncontrib = $regp->{'lst-contrib'};
    }
    $nprovs = $regp->{'lst-provs'} if $regp->{'type'} eq 'PROVIDER';
    
    &maill_upd($f,'local-ir.reg',$nlocalir) if ($localir || $nlocalir) && $localir ne $nlocalir;
    &maill_upd($f,'ncc-co.reg',$ncontrib) if ($contrib || $ncontrib) && $contrib ne $ncontrib;
    &maill_upd($f,'ip-provs.reg',$nprovs) if ($provs || $nprovs) && $provs ne $nprovs;
    return;
}

#------------------------------------------------------------------------------
# Subroutine Header
# Purpose           : edit registry file
# Side Effects      :
# Comments          : This is basically a call to (the generic) changereg,
#                     with as subroutine something that lets you interactively
#                     edit the registry data (i.e. pops up your editor)
# in: filename, option (can be 'v' or '')
# out: see changereg

sub editreg ($$) {

    my ($f,$options) = @_;

    my $mysub = sub {
	my ($regp,$opts) = @_;
	my ($edit) = 'y';
	my ($tmpf) = "/tmp/reg.e.$$";
	my ($e) = $ENV{'EDITOR'} || 'vi';
	my ($tmp);

	while ($edit =~ /^\s*[ey]/) {
	    printreg($tmpf, $regp,'E#x'.$opts);
	    system($e, $tmpf); 
	    $regp = readreg($tmpf);
	    unlink($tmpf);
	    return undef unless defined $regp;
	    last unless $regp->{'error'} ne '';
	    $tmp = $regp->{'error'};
	    $tmp =~ s/\f/\n   /g;
	    print STDERR "\nErrors Detected:\n$tmp\n";
	    $edit = &prompt("\nEdit again (y/n)?" ,"y");
	}
	return $regp;
    };

    $f = changereg($f,2,$mysub,$options);

    return $f;
}

#------------------------------------------------------------------------------
# Subroutine Header
# Purpose           : create new registry 
# Side Effects      :
# Comments:
# While similar to &changereg/&editreg, cannot really be incorporated in them.
# So I put this routine here (even though nothing except 'reg' itself
# will probably use it) so making changes to &changereg & &newreg at the
# same time is easier
# (BTW if I had put &newreg in 'reg', I'd have had to export &maill_upd, so
# I would always export a routine that nobody'll use... And now I can also
# put &rev_create next to &rev_checkin)
#
sub newreg ($$) {

    my ($f,$opts) = @_;
    my ($tmpf) = "$TMPDIR/reg.e.$$";
    my ($edit) = "y";
    my ($e) = $ENV{"EDITOR"};
    my ($regp, $t, $v, $newval);
    my (%old);

    $e = "vi" unless $e;
    die "\"$f\": not a regid" if $f !~ /^tt[0-9]+$/;
    die "$f exists.\n" if -f "$REGDIR/$f";
    umask(0);
    open(NEW, ">$REGDIR/$f") or die "create $f: $!\n";
    close(NEW);
    $regp = readreg("$LIBDIR/template") if -r "$LIBDIR/template";
    $regp->{'ttid'} = $f;
    $regp->{'error'} = '';

    # add today's date and blanks if they are missing
    foreach $t (@tags) {
	if ($tags{$t} =~ /^\S*D/) {
	    $newval = '';
	    foreach $v (split("\n", $regp->{$t})) {
		if ($v !~ /^9\d{5}/) {
		    $v = &thisdate . ' ' . $v;
		}
		$newval .= "$v\n";
	    }
	    $regp->{$t} = $newval;
	}
    }

    while ($edit =~ /^\s*[ey]/) {
	printreg($tmpf, $regp,'E#x'.$opts);
	system($e, $tmpf); 
	$regp = readreg($tmpf);
	unlink($tmpf);
	exit 1 unless defined $regp;
	last unless $regp->{'error'} ne '';
	$t = $regp->{'error'};
	$t =~ s/\f/\n   /g;
	print STDERR "\nErrors Detected:\n$t\n";
	$edit = &prompt("\nEdit again (y/n)?" ,"y");
    }

    undef $edit;
    while ($edit !~ /^(yes|no)$/) {
	$edit = &prompt("\nCreate registry $f (yes/no)?" ,"yes");
    }
    if ($edit eq 'no') {
	unlink("$REGDIR/$f") or warn "unlink $f: $!\n";
	print "\n";
	exit 1;
    }
    
    printreg("$REGDIR/$f",$regp,'e');
    &rev_create("$REGDIR/$f");
    
    &checkafterchanging($f,$regp,\%old); # %old's just empty
    return 0;
}

#------------------------------------------------------------------------------
# Subroutine Header
# Purpose           : print registry data (to a filehandle)
# Side Effects      :
# Comments          : The flags kind of sprang into existence according to 
#                     what different subroutines needed (which I merged into 1)
#
# in: filename, pointer to reg data, flag
#     Note: filename is not registry ID! Should already have full pathname.
#      for STDOUT, you'd use filename '-'
#
#     flag contains 'e': print error messages which cannnot be re-generated
#                        from registry data (these start with '** ignored:')
#      (This flag should always be set when writing data to the database!!!)
#     flag contains 'E': print all error messages
#                           (see readreg header for more on error messages)
#     flag contains '#': print extra '#' in between 'kinds of data'
#     flag contains 'v': verbose output (descriptions)
#     flag contains 'x': show extra empty line for multi-line or empty
#                        attributes (used by editreg, newreg, showform)
#
#     flag contains 'o': output only fields of type mentioned in flag
#                        (+ first part, which is always displayed)
#                        (flag contains string of uppercase letters)
#     flag contains 'f': output only fields with 'F' option (used by showform)
#     flag contains 'p': output only fields with 'P' option (used by writepub)
# out: 0 on success, 1 on error
#
# --below information may useful for reference - but may be outdated--
# called by: editreg (with E#x(/v)), changereg (with e)
#            newreg (with E#x(/v), and e), showform (with fx(/v)),
#            showreg (with E#o(/v)), writepub (with p)
# So far I think this routine will only be used by reg itself.
# Exporting it only for showform and showreg.
#
sub printreg ($$$) {

    my ($f, $regp, $flag) = @_;
    my ($t, $v, $opt, $regex, $showit);
    my (%verbname, %verbdesc);

    open(F, ">$f") or  do {warn "cannot write $f:$!\n"; return 1};

### As this is now the only routine where verbose descriptions are
# being used, I've inserted &getverbose here...
    # get verbose descriptions for attributes
    if ($flag=~/v/) {
	my ($t,$desc,$name);
	open(G, "$LIBDIR/descriptions") or do {warn "open $LIBDIR/descriptions: $!\n"; return 1};

	while (<G>) {
	    next if /^#/;
	    if (/^-(\S+)\s+(.*)\s*$/) {
		if ($t) {
		    $verbdesc{$t} = $desc;
		    $verbname{$t} = $name;
		}
		$t = $1;
		last if $t eq "END";
		$name = $2;
		$desc = "";
	    }
	    else {
		$desc .= $_;
	    }
	}
	close (G);
    }

    ### OK, now do your job.

    $showit = 1; # always show first part of data
    foreach $t (@tags) {
	next if $t eq 'error' && !$regp->{'error'};
	($opt, $regex) = split ("\t", $tags{$t});

	# If start of next 'kind of reg data': print extra '#'
        # unless you're not showing this kind of data.
	if ($t =~ /^-(.)/) {
	    eval "\$showit = \$Getopt::Std::opt_$1";
	    print F "#\n" if $flag=~/\#/ && !($flag=~/o/ && !$showit);
	    next;
	}
	# only print if you're showing this kind of data
	next if ($flag=~/f/ && $opt!~/F/) || ($flag=~/p/ && $opt!~/P/) || ($flag=~/o/ && !$showit);

	# print description
	if ($flag=~/v/ && $t ne 'error') {
	    print F "#\n# $verbname{$t}\n#\n";
	    foreach $v (split("\n", $verbdesc{$t})) {
		print F "# $v\n";
	    }
	    print F "#\n# ";
	    print F $opt=~/M/ ? "Mandatory, " : "Optional, ";
	    print F $opt=~/S/ ? "Single Attribute per Object, " : "Multiple Attributes per Object possible, ";
	    print F "Multi-Line Attribute, " if $opt =~ /T/;
	    if ($opt=~/P/) {
		print F "Published in Registry List.";
	    } else {
		print F $opt =~ /F/ ? "Not Published." : "NCC PRIVATE!";
	    }
	    print F "\n";
	    print F "# Syntax: /$regex/\n" if ($regex);
	    print F "#\n";
	}

	# print data
	if ($t eq 'error') {
	    if ($flag =~ /e/i) {
		foreach $v (split("\n", $regp->{$t})) {
		    next unless $flag =~ /E/ || $v =~ /^\*\* ignored:/;
		    &writeatt(\*F, 'ERROR', $v, $opt);
		}
		print F "\n";
            # remark: funny little coincidence here - you can still
            #  recognise that a registry has only retraceable errors, even
	    #  when flag 'e' is set (not 'E') - because the first
	    #  line will be a newline...
	    # I think I will leave it like this.
	    }
	} else {
	    &writeatt(\*F, $t, $regp->{$t}, $opt) if $regp->{$t}; 

	    # write extra empty line if:
	    # - x-flag set, attribute has option D(W) and is not 
            #   single-line & already filled. (print incl. this date (& user))
	    # - x-flag set, attribute has no option D and is empty; or,
            #   v-flag set, attribute is empty (print only attribute-name)
	    if ($flag =~ /x/ and $opt =~ /D/ and !($opt=~/S/ and $regp->{$t})) {
		if ($opt =~ /W/) {
		    printf F "%-12.12s %s %s \n", "$t:", &thisdate, &thisuser;
		} else {
		    printf F "%-12.12s %s \n", "$t:", &thisdate;
		}
	    } elsif (!$regp->{$t} and ($flag =~ /v/ or ($flag =~ /x/ and $opt !~ /D/))) {
		printf F "%-12.12s \n", "$t:";
	    }
	}

	print F "\n" if $flag=~/v/;
    }
    
    # print servicelevel
    # This is funny: Print servicelevel only when output is restricted to
    # certain fields; not when it is not restricted (for newreg and editreg)
    # Also don't print in form.
    if ($flag=~/o/ && $Getopt::Std::opt_S) {
	($t,$v) = servicelevel($regp);
	print F "#\nService Level: $t ($v)\n";
    } elsif ($flag=~/p/) {
	# Dirty hack: When billing parent sspecified, usually there is no
	# billing-categ specified - so we still need to do this.
	if ($regp->{'bill-parent'}) {
	    printf F "%-12.12s %s\n", 'bill-categ:',"Part of 'supernational' $regp->{'bill-parent'}";
	}
	($t,$v) = &regread::pubservicelevel($regp);
	print F "NCC-service: $t $v\n";
    }
	
    # print persons
    # This is funny: Print persons only when output is restricted to certain
    # fields; not when it is not restricted (for newreg and editreg)
    if ($flag=~/[p]/ or $flag=~/o/ && $Getopt::Std::opt_P) {
	$t = getcontacts(split("\n", $regp->{'admin-c'}."\n".$regp->{'tech-c'}));	
	if ($flag=~/p/) {
	    $v = `date`;
	    chomp $v;
	    print F "\n\nThe contact information below has been taken from the\n";
	    print F "RIPE database at $v.\n";
	    print F "Check whois.ripe.net for any changes.\n\n";
	    print F $t;
	} else {
	    print F "\n\n###### Referenced Contacts, taken from RIPE Database\n#\n";
	    foreach $v (split("\n", $t)) {
		print F "# $v\n";
	    }
	    print F "#\n";
	}
    }

    close(F);
    return 0;
}

# Note: I deleted all the SCCS stuff.
# Reason: I got rid of the "chdir()" at the start of reg. It now does
# not (need to) change working directory to $REGDIR in order for its
# routines to work. (Which is kind of a requirement for libs...)
# However, SCCS does not work then. (And we don't use it anyway):
# - RCS checks in/out $REGDIR/RCS/file,v <-> $REGDIR/file
# - SCCS checks in/out $REGDIR/SCCS/s.file <-> $PWD/file

# These routines now need complete filename (incl. path), not only registry ID!

#------------------------------------------------------------------------------
# Subroutine Header
# Purpose           : checks in revision of a registry file
# Side Effects      :
# Comments          : Preferrably don't use this yourself, use changereg
#             No error messages on failed system calls; RCS prints its own.
sub rev_checkin {
    my($f) = @_; # full filename
    my ($dir);

    ($dir) = $f=~/(.*)\//; # $1 is everything from start till last '/'

    if (-d "$dir/RCS") {
	system("$RCSDIR/ci", '-u', '-q', "-m$ENV{'LOGNAME'}", "$f") and return 1;
    } else {
	warn "No directory $dir/RCS\n";
	return 1;
    }
    return 0;
}	

#------------------------------------------------------------------------------
# Subroutine Header
# Purpose           : checks out revision of a registry file
# Side Effects      :
# Comments          : Preferrably don't use this yourself, use changereg
#             No error messages on failed system calls; RCS prints its own.
#
sub rev_checkout {
    my($f) = @_;
    my ($dir);

    ($dir) = $f=~/(.*)\//; # $1 is everything from start till last '/'

    if (-d "$dir/RCS") {
	system("$RCSDIR/co", '-l', '-q', "$f") and return 1;
    } else {
	warn "No directory $dir/RCS\n";
	return 1;
    }
    return 0;
}

# not exported, but put close to other revision routines
sub rev_create {	# create a file - called by newreg
    my($f) = @_;
    my ($dir);

    ($dir) = $f=~/(.*)\//; # $1 is everything from start till last '/'

    if (-d "$dir/RCS") {
	system("$RCSDIR/ci", '-u', '-t-New Registry', "$f") and return 1;
    } else {
	warn "No directory $dir/RCS\n";
	return 1;
    }
    return 0;
}

#
# end of exported subroutines
#

# writeatt: write an attribute to filehandle.
#
# in: - pointer to filehandle (e.g. \*STDOUT)
#     - attribute name (e.g. 'regid', 'bill-addr', 'allocated'; == hash-key)
#     - value of attribute (associated value of hash-key)
#     - 'attributes of attribute' ($tags{attribute})
# out: nothing
# used by printreg
#
sub writeatt ($$$$) {

    my($f, $t, $val, $opt) = @_;
    my($v);
    my($d);
    my($first);
    
    return() if !$val;
    
    if ($t eq 'assigned' || $t eq 'allocated') {
	foreach $d (sort bynet split("\n", $val)) {
	    $first = 1;
	    foreach $v (split("\f", $d)) {
		printf $f "%-12.12s %s\n", "$t:", $v if $first;
		printf $f "%-12.12s %s\n", '', $v unless $first;
		$first = 0;
	    }
	}
    } elsif ($opt =~ /D/) {
	foreach $d (sort bydate split("\n", $val)) {
	    $first = 1;
	    foreach $v (split("\f", $d)) {
		printf $f "%-12.12s %s\n", "$t:", $v if $first;
		printf $f "%-12.12s %s\n", '', $v unless $first;
		$first = 0;
	    }
	}
    } else {
	foreach $d (split("\n", $val)) {
	    $first = 1;
	    foreach $v (split("\f", $d)) {
		printf $f "%-12.12s %s\n", "$t:", $v if $first;
		printf $f "%-12.12s %s\n", '', $v unless $first;
		$first = 0;
	    }
	}
    }
    return();
}

# writepub: write public part of registry data to file on FTP site
#
sub writepub ($$) {

    my ($f,$regp) = @_; # regid, pointer to hash with data

    if ($regp->{'type'} !~ /^(PROVIDER|ENTERPRISE|FYI)$/ and $regp->{'bill-hack'} ne 'NORMAL') {
	unlink("$PUBDIR/$f")|| warn "unlink $PUBDIR/$f: $!\n" if -f "$PUBDIR/$f";
    } else {
	printreg("$PUBDIR/$f",$regp,'p');
    }
}

# maill_upd: update (not re-make) mailinglists
# in: registry, list==filename, entries for this registry (separated by \n)
# called from: newreg editreg
# uses: normmadr
#
# Note: if you change anything: there's s very similar routine in the 'maill'
# x-script.
#
sub maill_upd ($$$) {

    my($regid,$f,$entries) = @_;
    my(@regaddrs);
    my(@tmp);
    my($tmp,$addr,$rest);

    foreach $tmp (split(/\n/,$entries)) {
	push (@regaddrs,&normmadr($tmp));
    }

    # Schema for preventing 2 simultaneous updates: first create link
    # from [list].updating to [list]. This will fail/wait/try again as
    # long as [list].updating exists (meaning another update process
    # is running).
    # When you've created this link, you're safe that another reg
    # process won't interrupt you. Unlink [list], build new one, and
    # remove [list].updating.

    until (link "$LSTDIR/$f", "$LSTDIR/$f.updating") {
	if (-f "$LSTDIR/$f.updating") {
	    print STDERR "Temporary file $LSTDIR/$f.updating exists.\nIt shouldn't, or only for a very short period,\nwhile another reg process is updating mailinglists.\nPausing 2 seconds; I hope it has disappeared then...\n";
	    sleep 2;
	} else {
	    warn "open $LSTDIR/$f.updating: $!\n";
	    return 1;
	}
    }

    unlink("$LSTDIR/$f") or do {warn "unlink $LSTDIR/$f: $!\nWatch out; still existing hard link $f.updating -> $f will cause trouble!\n"; return 1};
    open(OUTF, ">$LSTDIR/$f") or do {warn "open $LSTDIR/$f: $!\nWatch out; still existing file $f.updating will cause trouble!\n"; return 1};

    print STDERR "updating mailing-list $LSTDIR/$f...\n";
    open(F, "$LSTDIR/$f.updating") or do {warn "open $LSTDIR/$f.updating: $!\nWatch out; still existing file $f.updating will cause trouble!\n"; close F; return 1};
    while (<F>) {
	($addr,$rest) = split(/\s+/,$_,2);
	chop ($rest);
	$rest =~ s/[\(\)]/ /g;
	if (grep(/^$addr$/,@regaddrs)) {
	    @regaddrs = grep(!/^$addr$/,@regaddrs);
	    $rest .= "$regid " if $rest !~ / $regid /;
	    substr($rest,0,1) = '(';
	    substr($rest,-1,1) = ')';
	    print OUTF "$addr $rest\n";
	} else {
	    $rest =~ s/ $regid / /;
	    if ($rest ne ' ') {
		substr($rest,0,1) = '(';
		substr($rest,-1,1) = ')';
		print OUTF "$addr $rest\n";
	    }
	}
    }
    close (F);
    while ($tmp = pop(@regaddrs)) {
	print OUTF "$tmp ($regid)\n" if $tmp ne '-';
    }
    close(OUTF);
    
    chmod(0664, "$LSTDIR/$f") or warn "chmod 0644 $LSTDIR/$f: $!\n";
    unlink("$LSTDIR/$f.updating") or do {warn "unlink $LSTDIR/$f.updating: $!\nWatch out; still existing file $f.updating will cause trouble!\n"; return 1};
}

sub normmadr {	# normalise mail address - called from maill_upd
    my($adr) = @_;
    my($localpart);
    my($domain);
    if ($adr =~ /^([A-Za-z0-9._\/=-]+)@([A-Za-z0-9.-]+)$/) {
	($localpart = $1) =~ tr/A-Z/a-z/;
	($domain = $2) =~ tr/A-Z/a-z/;
    }
    else {
	warn "strange mail address: '$adr'\n";
	return($adr);
    }
    return($localpart.'@'.$domain);
}

sub thisdate {
    my($sec, $min, $hour, $day, $month, $year) = localtime(time);
    $month++;
    $hour = sprintf("%02d", $hour);
    $min = sprintf("%02d", $min);
    $month = sprintf("%02d", $month);
    $day = sprintf("%02d", $day);
    $year += 1900;
    return("$year$month$day");
}

sub thisuser {  # find the initials for this user
	return($ENV{'USERNAME'}) if $ENV{'USERNAME'};
	return($ENV{'LOGNAME'}) if $ENV{'LOGNAME'};
	return('???');
}

sub bydate {	# sort routine for attribute dates
    my($adate, $bdate);

    if ($a =~ /^(\d+)\s+/) {
	$adate =$1;
    }
    if ($b =~ /^(\d+)\s+/) {
	$bdate =$1;
    }
    return ( $adate <=> $bdate );
}

sub bynet {	# sort routine for network numbers
    my($anum, $bnum);
    my(@anums);
    my(@bnums);
    my($i);

    if ($a =~ /^\d+\s+\S+\s+([\d.]+)/) {
	$anum =$1;
    }
    if ($b =~ /^\d+\s+\S+\s+([\d.]+)/) {
	$bnum =$1;
    }

    @anums = split(/\./, $anum);
    @bnums = split(/\./, $bnum);

    foreach $i (0 .. 3) {
	return(1) if $anums[$i]>$bnums[$i];
	return(-1) if $anums[$i]<$bnums[$i];
    }
    return(0);
}

sub prompt {
    my($p, $d) = @_;
    my($in);
	
    $|=1;
    print "$p [$d]: ";
    $in = <STDIN>;
    chomp($in);
    $|=0;
    return ($in eq '' ? $d : $in);
}
#EOF
