#------------------------------------------------------------------------------
#
#$Id: stats.pl,v 1.2 2006/11/04 14:40:50 ruben Exp $
#
# Module Header
# Filename          : stats
# Purpose           : Divides the registries in different categs and gives a
#                     count of each
# Author            : Roderik/Daniel
# Date              : 14 November 1997 (& earlier)
# Description       : Uses hashes with types, billing categories & 'sorts'
#                      (==self-defined 'finer grained' categories) as hashkeys
#                      and counter as value. Also, if -v option specified,
#                      lists registries per sort.
# Language Version  : Perl (Tested with v5.00401)
# OSs Tested        : BSD/OS 2.1
# Command Line      : reg -x stats [-v]
# Input Files       : info extracted from $REGDIR/*
# Output Files      : -
# External Programs : perl module regread.pm
# Problems          :
# To Do             :
# Comments          : This is a 'reg x-script' which can be called with
#                     'reg -x <scriptname>. However, Why not just 'use'
#                     the needed modules in here too? Now it can also be 
#                     called directly if we want...
#                     (Note: 'use strict' is necessary here because the strict
#                     directive apparently doesn't survive a 'do <filename>'.
# The following line will be read by 'reg -x':
# reg-x-description: general registry statistics
#------------------------------------------------------------------------------

use strict;
use Getopt::Std;
use regread;


my(@f);	# registry names
my($f,$regp,$i,$count);
my($sort,$comment);

#	The stats
my($entr);	# total number of database entries
my($oreg);      # total number of registries (old count)
my($nreg);      # total number of real registries
my(%type);	# registry type counters
my(%categ);	# registry (billing) category counters
my(%sortlist); 
my(%sort);     # my own (finer) division; list of registries (+ optional: comments)


die "stats: Wrong argument specified (only -v is possible)\n" if !getopts('v');

# gather the stats
@f = regfiles(0);
$entr = @f;
	
while ($f = shift(@f)) {
    $regp = readreg($f);

    # increase type-count
    $type{$regp->{'type'}}++;

    # if it is a 'regular' registry: increase size-count
    $categ{$regp->{'bill-categ'}}++ if $regp->{'type'} eq 'PROVIDER' || $regp->{'type'} eq 'ENTERPRISE';

    # increase sort-count, add to sort-list
    ($sort,$comment) = &expslevel($regp);
    $sort{$sort}++;

    $sortlist{$sort} .= $f;
    if ($comment) {
	$i = 25 - length($f);
	$i = 1 if $i<1;
	$sortlist{$sort} .= ' ' x $i . $comment;
    }
    $sortlist{$sort} .= "\n";
}

# print the stats
print "Entries: $entr\n\n";

$oreg = $type{'ENTERPRISE'} + $type{'PROVIDER'};
$nreg = $sort{'normal normal'} + $sort{'forced normal'} + $sort{'forced none'};

print "Registries: $oreg (old count)  $nreg (new count)\n";
print "  (FYI: old == enterprise + provider; new = normal + forced (normal/none)\n";
print "\nTypes:\n";
foreach $i (sort(keys %type)) {
    printf "%-12.12s %d\n", $i, $type{$i};
}

print "\nSizes (only for types PROVIDER+ENTERPRISE):\n";
foreach $i (sort(keys %categ)) {
    printf "%-15.15s %d\n", $i, $categ{$i};
}

print "\nService level:\n\n";

print"NORMAL:\n";
$count = 0;
foreach $i ('normal normal','forced normal') {
    $count += $sort{$i};
    printf(" - %-29s: %s\n",$i,$sort{$i});
    delete $sort{$i};
}
printf(" - %-29s: %s\n",'total',$count);

print"NONE:\n";
$count = 0;
foreach $i ('forced none','forced none (old last-resort)','pending','no agreement','non-existing bill-parent') {
    $count += $sort{$i};
    printf(" - %-29s: %s\n",$i,$sort{$i});
    delete $sort{$i};
}
printf(" - %-29s: %s\n",'total',$count);

# others (which are closed and fyi, but this way also traps other unforseen)
print"OTHERS:\n";
$count = 0;
foreach $i (sort keys %sort) {
    $count += $sort{$i};
    printf(" - %-29s: %s\n",$i,$sort{$i});
}
printf(" - %-29s: %s\n",'total',$count);

if ($Getopt::Std::opt_v) {
    print "\n\n\nAPPENDIX: sorts of registries\n";
    foreach $i (sort keys %sortlist) {
	print "\n * $i\n";
	print $sortlist{$i};
    }
}

#
# expslevel: returns 'sort of registry' & (optional) comments
#

sub expslevel {

    my ($regp) = @_;
    my ($sort,$comment);

    if ($regp->{'bill-hack'} =~ /^NORMAL\s+(.*)/) {
	return ('forced normal');
    } elsif ($regp->{'bill-hack'} =~ /^NONE\s+(.*)/ || $regp->{'bill-hack'} =~ /^TIME-P\s+(.*)/) {
	if ($regp->{'type'} eq 'LAST-RESORT') {
	    return ('forced none (old last-resort)');
	} else{
	    return ('forced none',$1);
	}
    } elsif ($regp->{'bill-parent'}) {
	if (-r "$REGDIR/$regp->{'bill-parent'}") {
	    ($sort,$comment) = &expslevel(readreg($regp->{'bill-parent'}));
	    $comment .= " (bill-parent $regp->{'bill-parent'})";
	    return ($sort,$comment);
	} else {
	    return ('non-existing bill-parent',"orphaned by $regp->{'bill-parent'}");
	}
    } elsif ($regp->{'type'} eq 'CLOSED') {
	return ('closed');
    } elsif ($regp->{'type'} eq 'FYI') {
	return ('fyi');
    } elsif ($regp->{'type'} eq 'PENDING') {
	return ('pending');
    } elsif ($regp->{'commit'}) {
	return ('normal normal');
    } else {
	return ('no agreement');
    }
}
