#!/usr/local/bin/perl
#
# Copyright (c) 2002  RIPE NCC
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose and without fee is hereby granted,
# provided that the above copyright notice appear in all copies and that
# both that copyright notice and this permission notice appear in
# supporting documentation, and that the name of the author not be
# used in advertising or publicity pertaining to distribution of the
# software without specific, written prior permission.
#
# THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
# ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS; IN NO EVENT SHALL
# AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY
# DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
# AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
# 
#------------------------------------------------------------------------------
# Module Header
# Filename          :   map_gen.pl
# Purpose           :   Generation of interactive testbox maps with GMT tools.
# Author	    :	Andrew Marcus <andrew@ripe.net>, RIPE NCC intern
#                   :                 <awm22@cornell.edu>
# Date	            :   12-FEB-2002
# Description       :   For TTM project website, easy navigation
# Input             :   $TTROOT/bin/ttconfig -v GPSINFO
#                   :   maps.config
# Output            :   various maps, maps.include (with image map)
# Language Version  :   perl5
# External Programs :   GMT package
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
#  define global data structures for all functions
#------------------------------------------------------------------------------

# global hashes
our %maps;                  # stores all map parameters, loaded from maps.config
                           #   - format: hash (map name, hash (parameter name, parameter value))
our %ttpixels;              # stores the pixel locations (x, y) if each test box on each map
                           #   - format: hash (map name, hash (testbox, array (x, y)))
our %clusters;              # stores the locations of clusters of boxes on each map
                           #   - format: hash (map name, array (hash (parameter name, parameter value))
our %statii;                # stores formating information for testboxes in several statuses

# defaults (formatting) hashes
our %map_defaults;          # stores css fields and other info to be used on all maps
our %cluster_defaults;      # stores css fields and other info to be used for creating cluster layers
our %status_defaults;       # stores css fields and other info for displaying all statuses

# global indices
our @maps_index;            # stores the list of all maps to print, in their breadth-first order
my @create_maps = ();      # stores which maps we'd like to generate with GMT tools
our @cluster_index;         # stores the list of all clusters of boxes

# GPS info hashes
my %ttInfo;                # stores info about each box (lat, long, place)
my %ttList;                # stores indices of boxes for each status
my %ttStatus;              # stores status of each testbox

#------------------------------------------------------------------------------
# parse command-line arguments
#------------------------------------------------------------------------------

my $verbose = 0;	# silent by default
my $confdir = ".";

while ($#ARGV >= 0) {
	if ($ARGV[0] eq '-m' || $ARGV[0] eq '--maps') {         # create listed maps only
	    for ($i = 1; $i <= $#ARGV; $i++) {
		$create_maps[($i-1)] = $ARGV[i];
	    	shift @ARGV;
    	    }
    
	}
	elsif ($ARGV[0] eq '-c' || $ARGV[0] eq '--confdir') {         
	        shift @ARGV;
		$confdir = $ARGV[0];
	}
	elsif ($ARGV[0] eq '-v' || $ARGV[0] eq '--verbose') {         
	    $verbose = 1;
	}
	else {                                                # other, output help menu
	    print 
	"This program creates the include file for map_index.cgi,\n",
	"and all maps associated with it, reading specfications.\n",
	"--------------------------------------------------------\n",
	"Input:  maps.config\n",
	"Output: $CGIdir/maps.include,\n",
	"        $WWWdir/map_<map_name>.gif\n",  
	"Usage:  map_gen [-c | --confdir <dir>] [ (-m|--maps) map1 map2...]\n",
	"        default          : create all maps\n",
	"        (-c | --confdir) : read config data from specified directory\n",
	"        (-m | --maps)    : create only the following maps\n",
	"	 (-v | --verbose) : verbose output\n\n";
#	"Defined maps:\n";
#	    foreach $m (@maps_index) {  print "\t$m"; }
#	    print "\nTo define more maps, edit maps.config\n";
	    die   "\n";
	}
        shift @ARGV;
}

# import config data

print "reading $confdir/maps.locations ...\n" if ($verbose) ;
do "$confdir/maps.locations";       # fetch global folder location defaults

print "reading $confdir/color_convert.pl ...\n" if ($verbose) ;
do "$confdir/color_convert.pl";     # fetch color conversion functions (dec <=> hex)

#------------------------------------------------------------------------------
# import and process data in maps.config
#------------------------------------------------------------------------------

&process_maps();

if ( $#create_maps < 0 ) {
	if ( $#maps_index >=0 ) {
		# default, create all maps
        	@create_maps = @maps_index; 
	}
	else {
		print "WARNING: No maps to create ???\n";
	}
}


&output_maps();
&output_html();
&prepare_files();

###############################################################################
# generate GPS info hashes
###############################################################################

sub get_gpsinfo {

    foreach my $status (@{$status_defaults{'status-order'}}) {

	my @GPSinfo = `$GENgps $status`;
	chomp @GPSinfo;
	
	foreach $f (@GPSinfo) {
	    my ($ttbox, $lat, $long, $place)  = split ("\t", $f);

	    # skip boxes which haven't been initialized (i.e., they've never taken data)
	    next if ($lat == 0 && $long == 0);

	    $ttInfo{$ttbox}{'lat'}   = $lat;
	    $ttInfo{$ttbox}{'long'}  = $long;
	    $ttInfo{$ttbox}{'place'} = $place;
	    
	    push @{$ttList{$status}}, ($ttbox);  # add to ttbox index
	    $ttStatus{$ttbox} = $status;         # add to statuses hash, so we can determine the status later
	}
    }
}

###############################################################################
##  MAIN: Function to process map parameters and build map tree
###############################################################################

sub process_maps {
    
    print "Processing $confdir/maps.config ...\n" if ($verbose);
    do "$confdir/maps.config";

    # retrieve GPS information
    &get_gpsinfo();

    # don't include inactive maps
    foreach my $m (keys %maps) {
	if (! $maps{$m}{'active'}) {
	    print "map $m INACTIVE!\n" if ($verbose);
	    delete $maps{$m};        # delete the key for this map in the hash
	}
    }

    # compute children of each map
    foreach my $m (keys %maps) {
	my $parent = $maps{$m}{'parent'};
	if ($parent && $parent ne "") {
	    push ( @{$maps{$parent}{'children'}}, ($m) );
	    print "map $m parent $parent ...\n" if ($verbose);
	}
    }

    # build the map tree, (breadth-first pushing onto a queue, starting at the root)
    @maps_index = ($root);
    $maps{$root}{'z-index'} = 0;

    for (my $i = 0; $i <= $#maps_index; $i++) {  # the length of @maps_index increases
	push ( @maps_index, @{ $maps{$maps_index[$i]}{'children'} } );

	my $parent = $maps{$maps_index[$i]}{'parent'};
	$maps{$maps_index[$i]}{'z-index'} = $maps{$parent}{'z-index'} + 1;
    }

    # get color attributes of each map from defaults, if not specified
    foreach my $m (@maps_index) {
	foreach $field (qw(water-color land-color outline)) {
	    if (! exists $maps{$m}{$field} && exists $map_defaults{$field}) {
		$maps{$m}{$field} = $map_defaults{$field};
	    }
	    # make sure it's in dec format (rrr/ggg/bbb)
	    $maps{$m}{$field} = color_hex2dec($maps{$m}{$field});
	}
	# setup GMT prefixes, if not already set
	$maps{$m}{'water-color'} =~ s/^(\d)/-S$1/;
	$maps{$m}{'land-color'}  =~ s/^(\d)/-G$1/;
	$maps{$m}{'outline'}     =~ s/^(\d)/-W$1/;
    }

    # get style-sheet attributes for each status from defaults, if not specified
    foreach my $status (@{$status_defaults{'status-order'}}) {
	foreach my $field (keys %{$status_defaults{'css'}}) {
	    if (! exists $statii{$status}{'css'}{$field}) {
		$statii{$status}{'css'}{$field} = $status_defaults{'css'}{$field};
	    }
	}
    }

    # setup dot options, for testboxes in each status we want to plot
    foreach my $status (@{$status_defaults{'status-order'}}) {
	
	# make sure colors are in appropriate format
	$statii{$status}{'dot-color'}    = color_hex2dec ($statii{$status}{'dot-color'});
	$statii{$status}{'css'}{'color'} = color_dec2hex ($statii{$status}{'css'}{'color'}) 
	    if (exists $statii{$status}{'css'}{'color'});

	# setup dot colors for GMT, if not already set
	$statii{$status}{'dot-color'} =~ s/^(\d)/-G$1/;

	# compute dot width (how wide to make imagemap spots)
	# dot-shape is of the format "-Sc5p", where 5 is the width (in points = pixels)
	$statii{$status}{'dot-shape'} =~ m/-\w+(\d+)p/;   # capture the width part into variable $1
	$statii{$status}{'dot-width'} = $1;
	if (! defined $1) {
	    print "Error: please specify dot-shape for status $status in points (1/72 inch)!\n";
	}
    }

    # parse GMT options, separate multiple parameters with a space
    foreach my $m (@maps_index) {
	foreach $field (keys %{$maps{$m}}) {
	    $maps{$m}{$field} =~ s/(\d+)-([A-Z])/$1 -$2/g;  # convert .../000-I2/...
	}                                                # to      .../000 -I2/...
    }

    # compute width and height from ps2gif dimensions
    foreach my $m (@maps_index) {
	my ($h1,$w1,$h2,$w2, $r) = split (" ", $maps{$m}{'pixels'});
	$maps{$m}{'width'}  = $w2 - $w1;
	$maps{$m}{'height'} = $h2 - $h1;
    }

    # compute layer options
    foreach my $m (@maps_index) {
	$maps{$m}{'visibility'} = ($m eq $root) ? "visible"  : "hidden";
	$maps{$m}{'position'}   = ($m eq $root) ? "relative" : "absolute";
	$maps{$m}{'left'}       =  $maps{$m}{'x-pos'};
	$maps{$m}{'top'}        =  $maps{$m}{'y-pos'};
    }

    # calculate pixel locations of other maps, testboxes, and clusters of testboxes
    &get_map_locations();
    &get_ttbox_locations();
    &get_cluster_locations();

    # compute cluster options
    foreach my $m (@maps_index) {
	for (my $i = 0; $i <= $#{ $clusters{$m} }; $i++) {
	    my @boxes            = split (",", $clusters{$m}[$i]{'boxes'});
	    my ($x1,$y1,$x2,$y2) = split (",", $clusters{$m}[$i]{'hotspot'});

	    # set parameters for this layer
	    $clusters{$m}[$i]{'position'}       = 'absolute';
	    $clusters{$m}[$i]{'name'}           = 'cluster_' . $m . '_' . $i;
	    $clusters{$m}[$i]{'visibility'}     = 'hidden';
	    
	    # add this layer to the end of the cluster_index list
	    push (@cluster_index, ($clusters{$m}[$i]{'name'}));

	    # compute z-index: 1 greater than the z-index of the parent layer
	    $clusters{$m}[$i]{'z-index'}        = ($maps{$m}{'z-index'} + 1);   
	    
	    # define default width and height for all cluster layers
	    my $width  = $cluster_defaults{'layer-width'};
	    my $height = 5;
	    foreach my $box (@boxes) {
		if (exists $statii{$ttStatus{$box}}{'css'}{'font-size'}) {
		    $height += $statii{$ttStatus{$box}}{'css'}{'font-size'} *
			       $cluster_defaults{'css'}{'line-height'};    
		} else {
		    $height += $cluster_defaults{'css'}{'font-size'} * 
			       $cluster_defaults{'css'}{'line-height'};
		}
	    }
	    
	    $clusters{$m}[$i]{'width'}          = $width;
	    $clusters{$m}[$i]{'height'}         = $height;
	    $clusters{$m}[$i]{'left'}           = $x1 - int ($width  / 2);
	    $clusters{$m}[$i]{'top'}            = $y1 - int ($height / 2);
	}
    }
    print "defined maps " , join(" ", @maps_index) , "\n"  if ($verbose);
}

#######################################################################
##  PROCESS: Function to compute locations of all maps on their parent map
#######################################################################

sub get_map_locations {

    # compute hotspots (where the imagemap on one layer should open another)
    # calculate the projection of the long/lat of each map onto its parent map
    foreach $m (@maps_index) {
	if ($m ne $root) {
	    my $GMTproject = "";
	    my $parent = $maps{$m}{'parent'};
	    foreach $field (qw(projection region)) {
		$GMTproject .= " " . $maps{$parent}{$field};
	    }
	    # get dimensions of the region field (format: -Rlong0/long1/lat0/lat1)
	    my ($w1,$w2,$h1,$h2) = split ("/", substr ($maps{$m}{'region'}, 2));
	    
	    # use GMT mapproject to retrieve correct map locations on parent map
	    my $hotspot = `echo "$w1\t$h1\n$w2\t$h2\n" | $GMTBIN/mapproject $GMTproject -Dp`;
	    my ($px1,$py1,$px2,$py2) = split (m/\t|\n/s, $hotspot);
	    
	    # adjust pixels, inverting from 0 at lower-left to 0 at upper-left
	    $px1 = int (                           $px1) + $maps{$parent}{'x-adjust'};
	    $px2 = int (                           $px2) + $maps{$parent}{'x-adjust'};
	    $py1 = int ($maps{$parent}{'height'} - $py1) + $maps{$parent}{'y-adjust'};
	    $py2 = int ($maps{$parent}{'height'} - $py2) + $maps{$parent}{'y-adjust'};
	    $maps{$m}{'hotspot'} = "$px1,$py2,$px2,$py1";
	}
    }
}

#######################################################################
##  PROCESS: Function to compute locations (in points, then pixels) of all testboxes
#######################################################################

sub get_ttbox_locations {
    
    foreach $m (@maps_index) {

	my $GMTproject = "";
	foreach my $field (qw(projection region)) {
	    $GMTproject .= " " . $maps{$m}{$field};
	}

	foreach my $status (@{$status_defaults{'status-order'}}) {

	    my $echostring = "";
	    foreach my $ttbox (@{$ttList{$status}}) {
		$echostring .= "$ttInfo{$ttbox}{'long'}\t$ttInfo{$ttbox}{'lat'}\n";
	    }

	    # compute imagemap areas for each testbox point
	    my @TTlocate = `echo "$echostring" | $GMTBIN/mapproject $GMTproject -Dp`;
	    chomp @TTlocate;

	    my $t = 0;
	    foreach $f (@TTlocate) {
		my ($x, $y) = split ('\t', $f);
		my $px = int (                      $x + $maps{$m}{'x-adjust'});   # linear transformations to
		my $py = int ($maps{$m}{'height'} - $y + $maps{$m}{'y-adjust'});   # align imagemap with image
		my $ttbox = $ttList{$status}[$t++];
	    
		# only include testboxes which appear within the range of the map
		if ( $px > 0 && $px < $maps{$m}{'width'} && $py > 0 && $py < $maps{$m}{'height'} ) {
		    
		    # only include testboxes which aren't overlapped by a map hotspot
		    my $overlapped = 0;
		    foreach $child (@{ $maps{$m}{'children'} }) {
			my ($x1, $y1, $x2, $y2) = split (",", $maps{$child}{'hotspot'});
			
			if ( $px > $x1 && $px < $x2 && $py > $y1 && $py < $y2 ) {
			    $overlapped = 1;
			    last;
			}
		    }
		    # finally, include this testbox's coordinates on this map
		    if (!$overlapped) {
			$ttpixels{$m}{$ttbox} = [$px, $py];
		    }
		}
	    }
	}
    }
}

#######################################################################
##  PROCESS: Function to compute locations of all overlapping testboxes
#######################################################################

sub get_cluster_locations {

    # implementation details: 
    #
    #   * (N^2)/2 traversal through all displayed testboxes on this particular map
    #     - calculate the distance between each pair of boxes
    #   * Build a tree:
    #     - if this distance is less than the width of one spot, put it into a hash table
    #     - each pair of nodes goes into the hash table once, but both keys are created
    #     - traverse, breadth-first, the tree formed by the arrays for each of the hash keys
    #     - as each node is traversed, remove it from future traversal
    #   * Find boundaries for a rectangular imagemap around this cluster of points
    
    foreach $m (@maps_index) {

	#------------------------------------------------------------------
	#  Calculate distances
	#------------------------------------------------------------------
    
	my %close_boxes = {};  # local,  hash (map, hash (ttboxes, array(ttboxes close to this one)))
	my %radius = {};       # local,  hash (ttbox, radius)
	my $maxdistance = $cluster_defaults{'distance'};

	my @map_boxes = sort keys %{ $ttpixels{$m} };

	# get radius of each box
	foreach my $box (@map_boxes) {
	    $radius{$box} = int (($statii{$ttStatus{$box}}{'dot-width'} + 1) / 2);
	}

	while (@map_boxes) {
	    $tt1 = shift @map_boxes;     # symmetric, so we need to only look at each pair once
	    
	    foreach $tt2 (@map_boxes) {  # starts tt2 at the element *after* tt1, (n*(n-1))/2
		
		my ($x1, $y1) = @{ $ttpixels{$m}{$tt1} };
		my ($x2, $y2) = @{ $ttpixels{$m}{$tt2} };
		my $distance = sqrt( ($x1 - $x2)**2 + ($y1 - $y2)**2 );    # distance between dots

		$maxdistance = $radius{$tt1} if ($radius{$tt1} > $maxdistance);
		$maxdistance = $radius{$tt2} if ($radius{$tt2} > $maxdistance);
		
		if ($distance <= $maxdistance) {
		    push ( @{ $close_boxes{$m}{$tt1} }, ($tt2) );          # put pair into hash table
		    push ( @{ $close_boxes{$m}{$tt2} }, () );              # define other key of pair
		}
	    }
	}
	
	#------------------------------------------------------------------
	# traverse through tree of close boxes, build map hotspots for clusters of them
	#------------------------------------------------------------------
	
	# get clusters, $box is the testbox in a cluster with the lowest number
	foreach my $box (sort keys %{ $close_boxes{$m} } ) {
	    # if this key doesn't exist, we've already dealt with it
	    next if (! exists($close_boxes{$m}{$box}));
	    
	    my @nodes = ($box);         # a queue of nodes, which gets appended to
	    my $boxes_in_cluster = "";  # a comma-separated string of all the boxes in this cluster 
	    my ($xmin, $ymin) = @{ $ttpixels{$m}{$box} };   # top-left     corner of hotspot
	    my ($xmax, $ymax) = @{ $ttpixels{$m}{$box} };   # bottom-right corner of hotspot
	    $xmin -= $radius{$box};
	    $ymin -= $radius{$box};
	    $xmax += $radius{$box};
	    $ymax += $radius{$box};
	    
	    # traverse clusters, breadth-first
	    while (@nodes) {
		$node = shift (@nodes);                           # dequeue first node

		# if the node does not exist as a key, we've already dealt with it
		next if (! exists($close_boxes{$m}{$node}));

		push (@nodes, @{ $close_boxes{$m}{$node} });  # enqueue children nodes
		delete $close_boxes{$m}{$node};               # don't search this node again
		
		# store node in a string.   This string will be sorted, and each item appear once,
		# because of the nature of the tree it derives from
		$boxes_in_cluster .= $node . ",";
		
		my ($x, $y) = @{ $ttpixels{$m}{$node} };
		
		# calculate max and min x and y coordinates of the hotspot
		$xmin = $x - $radius{$node} if ($x - $radius{$node} < $xmin);
		$xmax = $x + $radius{$node} if ($x + $radius{$node} > $xmax);
		$ymin = $y - $radius{$node} if ($y - $radius{$node} < $ymin);
		$ymax = $y + $radius{$node} if ($y + $radius{$node} > $ymax);
		
		# remove this testbox from this map, since it's included under clusters instead
		delete $ttpixels{$m}{$node};
	    }
	    my $hotspot = $xmin . "," . $ymin . "," . $xmax . "," . $ymax;

	    push (@{ $clusters{$m} }, ({'boxes'   => $boxes_in_cluster, # separated by ","
					'hotspot' => $hotspot}));       # separated by ","
	}
    }
}

###############################################################################
##  MAIN: Function to create all map plots
###############################################################################

sub output_maps {

    # create only desired maps (see command line section above)
    foreach $m (@create_maps) {

	# is this a valid map?
	if (! exists($maps{$m})) {
	    print "A map for $m is not defined, sorry!\n";
	    next;
	}
	
	my ($GMTcoast, $GMTpsxy, $GMTbasemap) = "";
	print "Creating map of $m...\n" if ($verbose);
	
	# Plot coastlines
	foreach $field (qw(projection region res features water-color land-color outline)) {
	    if (exists $maps{$m}{$field}) {
		$GMTcoast   .= " " . $maps{$m}{$field};
	    }
	}
	
	# Plot basic map layout
	foreach $field (qw(projection region border)) {
	    $GMTbasemap .= " " . $maps{$m}{$field};
	}

	# Plot testboxes
	foreach $field (qw(projection region)) {
	    $GMTpsxy    .= " " . $maps{$m}{$field};
	}

	my $ps_file = "$WWWSYNCbase/$WWWsync/maps.${m}.ps";
	
	`$GMTBIN/pscoast   $GMTcoast   -K    >  $ps_file`;
	
	# Get testboxes for each status, to plot on maps
	# Plot statuses in reverse order, so that more important statuses get placed on top of less important ones
	foreach my $status (reverse @{$status_defaults{'status-order'}}) {
	    my ($GPSlatlong, $GMTpsxy2) = "";

	    foreach my $ttbox (@{$ttList{$status}}) {
		my $long = $ttInfo{$ttbox}{'long'};
		my $lat  = $ttInfo{$ttbox}{'lat'};
		$GPSlatlong .= $long . "\t" . $lat . "\n";
	    }
	    $GMTpsxy2  = $statii{$status}{'dot-color'} . " "; 
	    $GMTpsxy2 .= $statii{$status}{'dot-shape'};

	    `echo "$GPSlatlong" | $GMTBIN/psxy $GMTpsxy $GMTpsxy2 -K -O >> $ps_file`;
	}
	`$GMTBIN/psbasemap $GMTbasemap -O >> $ps_file`;

        #----------------------------------------------------------------------
        #  Create gif files
	#----------------------------------------------------------------------

	print "Creating gif for map of $m...\n" if ($verbose);
	`$PS2GIF -b $maps{$m}{'pixels'} $ps_file`;
	unlink ($ps_file);
    }
}


###############################################################################
##  MAIN: Function to print the html imagemap
###############################################################################

sub output_html {
    print "Creating HTML...\n" if ($verbose);

    #--------------------------------------------------------------------------
    #  begin HTML output (executable, so calling cgi can substitute variables)
    #--------------------------------------------------------------------------
    open (HTML, ">$WWWSYNCbase/$CGIsync/maps.include");

    &output_javascript;
    &output_styles;
    &output_imagemaps;
    &output_header;
    &output_layers($root);  # recursive function, we want to start at the root

    close HTML;
}

###############################################################################
##  HTML: Function to produce javascript for use with the maps
###############################################################################
sub output_javascript {
    
    print HTML "<script language='Javascript' src='$WWWdir/$JAVAFUNCS'></script>\n";
    print HTML "<script language='Javascript'>\n\n",
	       "    // set the name for this window, so that all links can return to it\n",
	       "    window.name = 'base_map';\n\n";

    # output children array
    print HTML "    var children = new Array();\n";
    foreach $m (@maps_index) {
	print HTML "    children['$m']\t= new Array(";
	my $comma = 0;
	foreach $child (@{$maps{$m}{'children'}}) {
	    if ($comma) { print HTML ","; } else { $comma = 1; }
	    print HTML "'$child'";
	}
	foreach $cluster (@{$clusters{$m}}) {
	    if ($comma) {print HTML ","; } else { $comma = 1; }
	    print HTML "'", $cluster->{'name'}, "'";
	}
	print HTML ");\n";
    }
    print HTML "\n";

    # output parent array
    print HTML "    var parent = new Array();\n";
    foreach $m (@maps_index) {
	if ($maps{$m}{'parent'}) {
	    print HTML "    parent['$m']\t= '" . $maps{$m}{'parent'} . "';\n";
	}
    }
    foreach $c (@cluster_index) {
	$c =~ /^cluster_(\w+)_(\d+)/;
	print HTML "    parent['$c']\t= '$1';\n";
    }
    
    print HTML "\n</script>\n";
}

#######################################################################
#  HTML: Function to create style elements
#######################################################################

sub output_styles {

    print HTML "<style type=\"text/css\"><!--\n";

    # styles for layers (maps, clusters)
    foreach $m (@maps_index, @cluster_index) {

	my $param_hash;    # reference to hash with parameters for this layer
	my $defaults_hash; # reference to hash with default parameters for this type of layer

	# determine what type of layer this is, set params accordingly
	if (exists $maps{$m}) {
	    $param_hash    = \%{ $maps{$m} };                # reference to appropriate %maps hash
	    $defaults_hash = \%{ $map_defaults{'css'} };     # reference to appropriate defaults hash

	} else {
	    # parse cluster map and number out of layer name
	    $m =~ m/^cluster_(\w+)_(\d+)$/;                  # format: cluster_<map>_<num>
	    $param_hash    = \%{ $clusters{$1}[$2] };        # reference to appropriate %clusters hash
	    $defaults_hash = \%{ $cluster_defaults{'css'} }; # reference to appropriate defaults hash
	}
	
	my $style = "";
	foreach $field (qw(position visibility z-index)) {
	    if (exists $param_hash->{$field}) {
		$style .= $field . ":" . $param_hash->{$field} . "; ";
	    }
	}
	foreach $field (qw(font-size left top width height)) {
	    if (exists $param_hash->{$field}) {
		$style .= $field . ":" . $param_hash->{$field} . "px; ";
	    }
	}
	foreach $field (keys %{ $defaults_hash }) {
	    $style .= $field . ":" . $defaults_hash->{$field} . "; ";
	}
	
	print HTML "#$m\t{$style}\n";
    }
    
    # styles for statuses
    foreach my $status (@{$status_defaults{'status-order'}}) {
	my $tag   = $status_defaults{'style-tag'};
	my $style = "";
	foreach $field (keys %{$statii{$status}{'css'}}) {
	    $style .= $field . ":" . $statii{$status}{'css'}{$field} . "; ";
	}
	my $class = lc $status;
	print HTML ".$class\t{$style}\n";
    }
    print HTML "--></style>\n";
}

#######################################################################
#  HTML: Function to create image maps
#######################################################################

sub output_imagemaps {

    foreach $m (@maps_index) {

	# start imagemap
	print HTML "<map name='map_$m'>\n";
	    
	# create imagemap areas for each of the layers above this one (if any exist)
	foreach $child ( @{ $maps{$m}{'children'} } ) {
	    my $hotspot = $maps{$child}{'hotspot'};
	    my $width   = $maps{$child}{'width'}  + 50;
	    my $height  = $maps{$child}{'height'} + 150;
	    print HTML 
		"  <area shape='rect' coords='$hotspot' href='#map#?$child'",
		"    onMouseOver=\"javascript:map_display('$child'); display('Map of $child')\"",
		"     onMouseOut=\"javascript:display('');\"",
		"        onClick=\"javascript:return open_window(this.href,'map_window_$child',$width,$height);\"",
		">\n";
	}
	
	# create imagemap areas for each cluster of testboxes
	for ($i = 0; $i <= $#{ $clusters{$m} }; $i++) {
	    my $hotspot = $clusters{$m}[$i]{'hotspot'};
	    my $boxes   = $clusters{$m}[$i]{'boxes'};
	    $boxes =~ s/,$//;   # desired format: "box1, box2, ..., boxN"
	    $boxes =~ s/,/, /g; # remove comma at the end, and add whitespace
	    print HTML 
		"  <area shape='rect' coords='$hotspot' href='#cluster#?cluster_${m}_${i}'",
		"    onMouseOver=\"javascript:map_display('cluster_${m}_${i}'); display('$boxes');\"",
		"     onMouseOut=\"javascript:display('');\"",
		"        onClick=\"javascript:return open_window(this.href,'cluster_window',600,300);\"",
		">\n";
	}
	
	# create imagemap areas for each testbox point on this map
	foreach $ttbox (sort keys %{ $ttpixels{$m} }) {
	    my ($px, $py) = @{ $ttpixels{$m}{$ttbox} };
	    
	    # get radius of box, rounding up
	    my $radius = int (($statii{$ttStatus{$ttbox}}{'dot-width'} + 1) / 2);
	    
	    # an executable link, with values to be filled out by the calling cgi function...
	    my $link = "#link#?$ttbox";
	    
	    print HTML 
		"  <area shape='circle' coords='$px,$py,$radius' href=\"$link\" alt='$ttbox'", 
		"    onMouseOver=\"javascript:display('$ttbox: $ttInfo{$ttbox}{'place'}  **$ttStatus{$ttbox}**');\"",
		"     onMouseOut=\"javascript:display('')\"",
		"        onClick=\"javascript:return open_basepage(this.href)\"",
		">\n";
	}
	
	# create an imagemap area covering the whole map, for closing above layers
	my $height = $maps{$m}{'height'} + 100;
	my $width  = $maps{$m}{'width'}  + 100;
	print HTML
	    "  <script language='javascript'>\n",
	    "    if (ns4 || ie4) {    // Netscape 6 uses image onMouseOver property instead\n",
	    "        document.write(\"<area shape='rect' coords='0,0,$width,$height'",
	                             " href='javascript:void();'",
	                             " onMouseOver=\\\"javascript:map_hide_children('$m','map');\\\"",
	                             ">\");\n",
	    "    }\n",
	    "  </script>\n";

	print HTML "</map>\n",
    }
}

#######################################################################
#  HTML: Function to create testbox info display box and status legend
#######################################################################

sub output_header {
    print HTML
	"<table width=$maps{$root}{'width'}>\n",
	"  <tr>\n",
	"    <td>\n",
	"      <form name=\"displayform\">\n",
	"        <input type=\"textbox\" name=\"displaybox\" size=90>\n",
        "      </form>\n",
	"    </td>\n",
	"    <td align='right'>\n",
	"      (<a href=\"map_help.html\">HELP</a>)\n",
	"    </td>\n",
	"  </tr>\n",
	"</table>\n",
	"<table width=$maps{$root}{'width'}>\n",
	"  <tr>\n";
    foreach my $status (@{$status_defaults{'status-order'}}) {
	my $dot_color = color_dec2hex($statii{$status}{'dot-color'});
	my $tag = $status_defaults{'style-tag'};
	my $class = lc $status;
	print HTML
	    "    <td>\n",
	    "      <font size='+1' color='$dot_color'><b>*</b></font>\n",
	    "      Boxes with <$tag class=$class>$status</$tag> status\n",
	    "    </td>\n";
    }
    print HTML
	"  </tr>\n",
	"</table>\n";
}

#######################################################################
#  HTML: Function to create layers for maps and clusters
#######################################################################

sub output_layers {

    # INPUTS
    my $m  = $_[0];  # node (starts at root)
    my $ws = $_[1];  # whitespace (optional)
    if (!$ws) { $ws = ""; }

    # BEGIN LAYER
    print HTML "\n", $ws, "<div id=\"$m\">\n";

    # CREATE MAP CONTENT
    if (exists $maps{$m}) {
	print HTML 
	    $ws, "  <img id=${m}_IMG src='$WWWdir/maps.$m.gif'",
	            " width=$maps{$m}{'width'} height=$maps{$m}{'height'}",
	            " usemap='#map_$m' border=0",
	            " onMouseOver=\"javascript:map_hide_children('$m','img');\">\n";
    
    # CREATE CLUSTER CONTENT
    } else {
       	$m =~ /^cluster_(\w+)_(\d+)$/;
	my $map = $1;
	my $num = $2;

        # put links to each box on this layer
	my @boxes = split (",", $clusters{$map}[$num]{'boxes'});
	foreach $ttbox (@boxes) {
	    my $tag   = $status_defaults{'style-tag'};
	    my $class = lc $ttStatus{$ttbox};
	    print HTML 
		$ws, 
		"  <a href=\"#link#?$ttbox\"",
		"   onMouseOver=\"javascript:display('$ttbox: $ttInfo{$ttbox}{'place'}  **$ttStatus{$ttbox}**');\"",
		"    onMouseOut=\"javascript:display('');\"",
		"       onClick=\"javascript:return open_basepage(this.href);\"",
		"  ><$tag class=$class>$ttbox</$tag></a><br>\n";
	}
    }

    # CREATE SUB-LAYERS RECURSIVELY
    # maps
    foreach $child (@{$maps{$m}{'children'}}) {
	&output_layers ($child, $ws . "  ");
    }
    # clusters
    foreach $i (@{$clusters{$m}}) {
	&output_layers ($i->{'name'}, $ws . "  ");
    }

    # END LAYER
    print HTML $ws, "</div>\n";
}

#######################################################################
#  MAIN:  generate and move all additional files for CGI script
#######################################################################

sub prepare_files {
    print "Preparing files for the web site ...\n" if ($verbose);

    # create GPSINFO in CGI directory
    open (GPS, ">$WWWSYNCbase/$CGIsync/GPSINFO");
    foreach my $status (@{$status_defaults{'status-order'}}) {           
                                                      # Format:
	print GPS "status:\t$status\n";               # status: ACTIVE
	foreach my $ttbox (@{$ttList{$status}}) {     # tt01    1.2345  34.5678 RIPE NCC, Amsterdam, NL 
	    print GPS                                 # . . .
		$ttbox, "\t",                         # tt75    23.4567 30.7395 GlobalSoft.com Ltd, Nicosia, CY
		$ttInfo{$ttbox}{'lat'}, "\t",         # status: OFF
		$ttInfo{$ttbox}{'long'}, "\t",        # tt11    10.2345 32.1098 ECRC Network Services, Munich, DE    
		$ttInfo{$ttbox}{'place'}, "\n";       # . . . etc
	}
    }
    close GPS;
}
