# $Id: TTM.pm,v 1.13 2008/01/08 10:45:52 ruben Exp $
# vi:ts=4:sw=4
# vim:fdm=marker

package RIPE::TTM;

=head1 NAME

RIPE::TTM

=head1 SYNOPSIS

	
use RIPE::TTM;

=head1 DESCRIPTION


RIPE::TTM is a collection of recurring functions and data in Test Traffic CGI's

=over 4

=cut

use strict;

=item B<Datamembers>

I<Various datastructures holding testbox relation information>:

=item C<%testboxes> - testboxname indexed array of addresses. 

The array is a reference.

=item C<@testboxes6> - shorthand list with IPv6 enabled testboxes

=item C<@testboxes4> - shorthand list with IPv4 enabled testboxes

=item C<%gpsinfo> - parsed GPSINFO file information

The structure has four self explaining members: C<place>, C<status>, C<lat> and C<long>

=item C<%chainmap> - per chain filtered information

=item C<$chainmapfile> - Location of the chainmap

=item C<$gpsinfofile> - Location of the GPSINFO file

=item C<$testboxesfile> - Location of the LIST_OF_TESTBOXES file

=item C<$dbmgroupfile> - Location of the dbmgroup file

The format of the dbmgroup is host dependant

=item C<$base_url> - Base URL of all pregenerated plot/mini plots

Not to be mistaken as C<$overview_url>!

=item C<$pod_url> - URL of plots on demand

=item C<$routing_url> - URL of the routing vector query program

=item C<$overview_url> - URL of the general plot index

=item C<$detail_url> - URL of the detailed plot index

=item C<$summary_url> - URL of the summaries

=item C<$images_url> - Where to find mockup plots

Used by plots on demand to provide the user with an image while the server
computes a plot

=item C<$hotlist_url> - URL of the hotlist management CGI

=item C<$use_basic_auth> - Use HTTP Basic Authentification

=item C<$use_chain_auth> - Use the private mesh chain authentification

=item C<$use_ttplots_http_proxy> - Use firewall bypass code

=item C<$remote_user> - Store the name of the HTTP Basic Authentificated user

=item C<@usergroups> - Parsed C<$dbmgroupfile> information for the current C<$remote_user>

=item C<$nohotlist> - Boolean to globally disable hotlists in the module.

Is used as a pragma in C<use RIPE::TTM qw(-nohotlist)>

=item C<$nolicensecheck> - Boolean to globally disable the one time license check in the module.

Is used as a pragma in C<use RIPE::TTM qw(-nolicensecheck)>

=item C<$ttplots_server> - Name of the plot server

=item C<@ttplots_server> - Names of the plot server, to be used round-robin

=item C<$ttplots_port> - plot server port

=item C<$ttquery_host> - Name of the routing vector server

=item C<@ttquery_host> - Names of the routing vector server, to be used round-robin

=item C<$ttquery_port> - routing vector server port

=cut


use vars qw(%testboxes @testboxes6 @testboxes4 %chainmap %gpsinfo @TTLIST @hotlist
			$chainmapfile $gpsinfofile $testboxesfile $dbmgroupfile
			$base_url $pod_url $routing_url $overview_url $detail_url $summary_url $images_url $hotlist_url $ttm_js_url $ttm_styles_url
			$use_basic_auth $use_chain_auth $use_ttplots_http_proxy $remote_user @usergroups $nohotlist $nolicensecheck
			$ttplots_server @ttplots_server $ttplots_port $ttquery_host @ttquery_host $ttquery_port);

use Sys::Hostname;

BEGIN { # {{{
	use Exporter	();
	use CGI			2.86 qw/-newstyle_urls -nosticky :standard :imagemap/;
	use CGI::Carp;
	use POSIX;
	use Time::Local;
	use vars		qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS);

	$VERSION		= do { my @r = (q$Revision: 1.13 $ =~ /\d+/g); sprintf "%d."."%02d" x $#r, @r };
	@ISA			= qw(Exporter);
	@EXPORT			= ( );
	@EXPORT_OK		= qw(&_is_included 
							&maketooltip 
							&tboxsort 
							&print_header 
							&print_footer 
							&get_headers
							&cleanup_query_string 
							&createzoommap
							&get_timestamp
						);
	%EXPORT_TAGS	= ( all => [qw(&_is_included 
									&maketooltip 
									&tboxsort 
									&print_header 
									&print_footer 
									&get_headers
									&cleanup_query_string 
									&createzoommap
									&get_timestamp
								)]);

	$remote_user = exists ($ENV{'REMOTE_USER'}) ? $ENV{'REMOTE_USER'} : 'ttraffic';

	# fill in host depending locations

	SWITCH: for (hostname()) {
		/^(tipu|degu)$/	&& do {

				# File locations
				$chainmapfile	= '/home/ttraffic/ttpro/www/data/plots/chainmap';
				$gpsinfofile	= '/home/ttraffic/ttpro/www/data/plots/GPSINFO';
				$testboxesfile	= '/home/ttraffic/ttpro/www/data/plots/LIST_OF_TESTBOXES';
				$dbmgroupfile	= '/nonexistant/dbmgroup';

				# Url locations
				$base_url		= '/plots';
				$pod_url		= '/plots/pod.cgi';
				$routing_url	= '/plots/routing.cgi';
				$overview_url	= '/plots/';
				$detail_url		= '/plots/showplots.cgi';
				$summary_url	= '/summaries/';
				$images_url	= '/images';
				$hotlist_url	= '/plots/hotlist.cgi';
				$ttm_styles_url = 'http://www.ripe.net/ttm/ttmstyles.css';
				$ttm_js_url	= 'http://www.ripe.net/ttm/ttm.js';

				# We can use basic auth through http here
				$use_basic_auth	= 0; 

				# We can make use of the $ENV{'TTMCHAIN'} enviroment to restrict access
				$use_chain_auth	= 1; 

				# ttplots service is firewalled and we use LWP::UserAgent to
				# retrieve our data from http://www.ripe.net/ instead
				$use_ttplots_http_proxy = 1; 

				# ttplots/ttquery locations
				$ttplots_server	= 'cedar.ripe.net';
				$ttplots_port	= 4250;
				$ttquery_host	= 'rimu.ripe.net';
				$ttquery_port	= '4200';

				last SWITCH;
			};

		/^(aquila|falcon|kite|hawk|x27)$/	&& do {	   # new website

				# File locations
				$chainmapfile	= '/nonexistant/chainmap';
				$gpsinfofile	= '/srv/www/pub-dyn/projects/ttm/Plots/GPSINFO';
				$testboxesfile	=  '/srv/www/pub-dyn/projects/ttm/Plots/LIST_OF_TESTBOXES';

				# Url locations
				$base_url		= '/ttm/Plots';
				$pod_url		= '/cgi-bin/ttm/plots-on-demand';
				$routing_url	= '/cgi-bin/ttm/routing';
				$overview_url	= '/ttm/Plots/';
				$detail_url		= '/ttm/Plots/plots.cgi';
				$summary_url	= '/ttm/Plots/summary.cgi';
				$images_url	= '/ttm/Plots/images';
				$hotlist_url	= '/ttm/Plots/hotlist.cgi';
				$ttm_styles_url = '/ttm/ttmstyles.css';
				$ttm_js_url		= '/ttm/ttm.js';

				# We can use basic auth through http here
				$use_basic_auth	= 0; 

				# We can make use of the $ENV{'TTMCHAIN'} enviroment to restrict access
				$use_chain_auth	= 0; 

				# ttplots service is firewalled and we use LWP::UserAgent to
				# retrieve our data from http://www.ripe.net/ instead
				$use_ttplots_http_proxy = 0; 

				# ttplots/ttquery locations
				$ttplots_server	= 'cedar.ripe.net';
				@ttplots_server = qw(cedar.ripe.net rimu.ripe.net);
				$ttplots_port	= 4250;
				$ttquery_host	= 'rimu.ripe.net';
				@ttquery_host	= qw(cedar.ripe.net rimu.ripe.net);
				$ttquery_port	= '4200';

				last SWITCH;
			};
		die "running on \"$_\" is not supported";
	}

	sub import {

		my (@pragmas) = grep { /^-\w+/ } @_;
		@_ = grep {!/^-\w+/} @_;

		$nohotlist=0;

		if (@pragmas)
		{
			for (@pragmas) {
				SWITCH: {
					/^-nohotlist$/ && do {
							$nohotlist=1;
							last SWITCH;
						};
					/^-nolicensecheck$/ && do {
							$nolicensecheck=1;
							last SWITCH;
						};
					push(@_,$_);
				}
			}
		}

		RIPE::TTM->export_to_level(1, @_);

		# Continue with initialisation of globals
		RIPE::TTM->init();
	}

} # }}}

# Javascript helpers {{{
# XXX FIXME Should be reworked to automaticly detect and handle checkboxes/droplists etc. 
# Not all browsers support this.value as a means to access the current setting or set it to a value

my $license_validate = <<'EOF';

/* Copyright http://www.perlscriptsjavascripts.com 
   Free and commercial Perl and JavaScripts     */
   
// page to go to if cookie does not exists
go_to = "/plots/license.cgi";

// number of days cookie lives for
num_days = 1200;
function ged(noDays){
    var today = new Date();
    var expr = new Date(today.getTime() + noDays*24*60*60*1000);
    return  expr.toGMTString();
}

function readCookie(cookieName){
    var start = document.cookie.indexOf(cookieName);
    if (start == -1){ 
        window.location = go_to;
    }
}

function checkLicense(){
	readCookie("ttmlicense");
}

window.onload=checkLicense();
EOF

my $validate_range = <<'EOF';
function validate_range(obj, lowval, hival, defval, message) {
if ((obj.value < lowval) || (obj.value > hival))
        {
                obj.value=defval;
                alert(message);
                return false;
        }
        return true;
}
EOF

my $validate_regex = <<'EOF';
function validate_regex(obj, regex, defval, message) {
        var re = new RegExp(regex);
        if (!re.test(obj.value))
        {
                obj.value=defval;
                alert(message);
                return false;
        }
        return true;
}
EOF

my $validate_equality = <<'EOF';
function validate_equality(thelvalue, thevalue, theoperator, defval, message) {
		var code="'" + String(thelvalue) + "' " + String(theoperator) + " '" + String(thevalue) + "'";
        if (!eval(code))
        {
                thelvalue=defval;
                alert(message);
                return false;
        }
        return true;
}
EOF

# }}}

my %plotlegend = ( # Legend for the plots, suitable for maketooltip() {{{ 
	delaytopleft => 
		li(b("Delay & Number-of-Hops as a function of time")) . 
		li("The black points show the delays as function of time.") .   li("The red points depict the number of hops (times 10) detected in the closest preceding traceroute measurement; unless the number of hops is changing rapidly, these points will merge into line segments. Obviously, the number of hops only flags drastic changes in the routing; subtle differences like load balancing over multiple communication lines do not show up. For more detailed information on routing, the TTM traceroute database should be consulted."),
	delaytopright => 
		li(b("Distribution of delays")) . 
		li("This histogram shows the distribution of the delays over the time interval of the previous plot, e.g. it is the projection along the y-axis of the previous plot. The number of entries, over- and underflows are the same as for the previous plot."),
	delaybottomleft => 
		li(b("Packets sent")) . 
		li("The " . span( { -style => "color: green;" },"green area") . " shows the number of packets sent per bin of time with valid NTP time-stamps on both the sending and receiving side.") .
		li("The " . span( { -style => "color: yellow;" },"yellow area") . " shows the number of packets sent per bin of time with a valid NTP time-stamp on the sending but not on the receiving side") .
		li("The " . span( { -style => "color: purple;" },"purple area") . " shows the number of packets sent per bin of time with a valid NTP time-stamp on the receiving but not on the sending side.") .
		li("The " . span( { -style => "color: red;" },"red area") . " shows the number of packets sent per bin of time without valid NTP time-stamps on both the sending and receiving side.") .
		li("The " . span( { -style => "color: cyan;" },"cyan area") . " shows the number of packets per bin of time that were lost. These packets left the sender but were not registered by the receiver, either because they were lost in the network or because the receiver was off-line."),
	delaybottomright => 
		li(b("Packet arrival as a function of time")) . 
		li("For each bin of the bottom left plot, the bottom right histogram shows which fraction of the sent packets arrived at the destination box (regardless of clock status). When the experiment is not running, i.e. no packets sent, the \"fraction arrived\" bins are set to 0. To distinguish between a situation of 100% packet loss, one will have to check the \"packets sent/valid\" bottom left plot."),
	ipdvtopleft => 
		li(b("Distribution of delay variations")) . 
		li("This histogram shows the distribution of the delay variations (i.e. the difference between consecutive one-way delay measurements) for the specified time period."),
	ipdvtopright => 
		li(b("Delay variations vs. time")) . 
		li("This shows the delay variations as seen at the various times in the measurement period. The top left plot is the 1-D projection of this."),
	ipdvbottomleft => 
		li(b("One-way delay measurements: Packet delay distribution")) . 
		li("For easy reference, the results of one-way delay measurements used in the delay variation calculations are shown in the bottom plots. the bottom left plot shows the distribution"),
	ipdvbottomright => 
		li(b("One-way delay measurements: Packet delay vs. time")) . 
		li("For easy reference, the results of one-way delay measurements used in the delay variation calculations are shown in the bottom plots. the bottom right plot shows the delay measurments vs. time. In addition, red dots indicate the number of hops (times 10) seen in the nearest traceroute measurement."),
	trendstop=> 
		li(b("One-way Delay Trend Plot")) . 
		li("trends of 2.5 percentile (minimum) delay"),
	trendsmiddle=> 
		li(b("One-way Delay Trend Plot")) . 
		li("trends in median (average) delay"),
	trendsbottom=> 
		li(b("One-way Delay Trend Plot")) . 
		li("trends of 97.5 percentile (maximum) delay"),
);
# }}}

=item init

Initialize data structures

=cut

sub init() # {{{
{
	@hotlist = retrieve_hotlist();

	listoftestboxes($testboxesfile);

	gpsinfo($gpsinfofile);

	if ($use_chain_auth)
	{
	####	use NDBM_File;
		my (%Chainlist);
		die "TTMCHAIN not in environment, bailing out" unless exists $ENV{'TTMCHAIN'};
		tie (%chainmap, 'NDBM_File', $chainmapfile,O_RDONLY,0640) or _abort("couldn't tie chainmap: $!");

		# Fill in the chain map
		foreach my $relation (grep { /^$ENV{'TTMCHAIN'}\+/ } keys %chainmap)
		{
			my ($src,$dst);

			if ( ($src,$dst) = $relation=~ /^\w+\+(tt\d+[a-z]?)\+(tt\d+[a-z]?)$/) 
			{
				# XXX too aggressive check ? this kills HISTORIC box support!
				push (@{$Chainlist{$src}}, $dst) if (exists($gpsinfo{$src}) && exists($gpsinfo{$dst}));
			}
		}
		untie(%chainmap);
		%chainmap = %Chainlist;

		# Now filter gpsinfo and listoftestboxes again

		foreach my $key (keys %gpsinfo)
		{
			delete $gpsinfo{$key} unless exists($chainmap{$key});
		}

		foreach my $key (keys %testboxes)
		{
			delete $testboxes{$key} unless exists($chainmap{$key});
		}

		@testboxes4 = grep {/^(@{[join('|',keys %testboxes)]})$/} @testboxes4;
		@testboxes6 = grep {/^(@{[join('|',keys %testboxes)]})$/} @testboxes6;
	}

	if ($use_basic_auth)
	{
	####	use DB_File;
		my (%dbmgroup);
		tie (%dbmgroup, 'DB_File', $dbmgroupfile) or _abort("couldn't tie ACL database: $!");
		@usergroups = (split (/,/, $dbmgroup{$remote_user}));
		@TTLIST = sort tboxsort grep { /^(@{[join('|',@usergroups)]})$/ } keys %gpsinfo;
		untie(%dbmgroup);
	}
	else
	{
		@TTLIST = sort tboxsort keys %gpsinfo;
	}


} # }}}

=item _is_included()

To check whether the cgi is directly called or indirectly through an server side include:
C< E<lt>!--#include virtual="program.cgi" --E<gt> >

=cut

sub _is_included() # {{{
{
	return (exists($ENV{'SERVER_PROTOCOL'}) && $ENV{'SERVER_PROTOCOL'} eq "INCLUDED" ) ? 1 : 0;
} # }}}

=item maketooltip(@)

Pregenerates tooltip javascript code for use 

=cut

sub maketooltip(@) # {{{
{
	my ($list) = ul(map {li($_)} @_ );
	return "this.T_STICKY=true;this.T_DELAY=1000;return escape('$list')"
} # }}}


=item tboxsort($$)

sort testboxes by testbox id.

B<TODO:> Implement geographicaly and hopwise sorting

=cut

sub tboxsort($$) # {{{
{
	# XXX Need to use 'my($x,$y)' instead of the package globals $a, $b as when used in
	# EXPORT'ed mode otherwise the wrong $a and $b get filled.
	# unfortunately, parametrized sort routines are slower than void ones.

	my ($x,$y) = @_;
	($x =~ /tt(\d+)/)[0] <=> ($y =~ /tt(\d+)/)[0]
							||
	($x =~ /tt\d+([a-z]?)/)[0] cmp ($y =~ /tt\d+([a-z]?)/)[0]

} # }}}

=item print_header($$)

print HTML and standard NCC header

Input  : main title (heading1) and subtitle (heading2)
Output : HTML code on stdout

B<TODO:> print_header should be renamed to get_header and needs an extra
argument to provide extra stuff in the C<E<lt>headE<gt>> section of the
document.

=cut

sub print_header($$) # {{{
{
	my ($maintitle,$subtitle) = @_;


	my $header = join("",
                "<!--#set var=\"title\" value=\"$maintitle: $subtitle\"-->",
                "<!--#include virtual=\"/includes/ripencc/ttmhead1.part1.inc\"-->",
                "<link rel=\"stylesheet\" type=\"text/css\" href=\"$ttm_styles_url\" />",
                "<style type=\"text/css\">",
                "<!--/* <![CDATA[ */",
                "B, BODY, P, CENTER, TH, DD, DL, DT, H1, H2, H3, LI, UL, H4, H5, OL, TABLE ,TD { font-family: Arial, sans-serif, Helvetica; }" .
		"TABLE[rules=rows] TR { border-top: solid }\n" .
		"TABLE[rules=rows]    { border-collapse: collapse; border-top: hidden }",
                "/* ]]> */--> </style>",

#
#  We cannot use ttmhead1.inc as-is when called by license.cgi 
#  because the "onload="checkLicense()" in the <body> directive
#  causes infinite recursion. For now, print_header filters
#  always puts a <body> without "onload" in the html, but it
#  should be adapted to do this only when called by license.cgi
#
		"</head><body>",
		`cat /var/www/pub/includes/ripencc/ttmhead1.part2.inc|egrep -i -v '</head>|<body'`,
#                "<!--#include virtual=\"/includes/ripencc/ttmhead1.part2.inc\"-->",
                a({ -name => "top"}, ""), h1($maintitle), h2($subtitle));




	return _is_included() ? "&nbsp;" : "$header";
} # }}}

=item print_footer()

print footer

Input  : none
Output : HTML code on stdout

=cut

sub print_footer() # {{{
{
	return script({language=>"JavaScript", type=>"text/javascript", src=>"http://www.ripe.net/ttm/wz_tooltip.js"}, ""), 
	_is_included() ? "&nbsp;" : (
		blockquote("Data from the RIPE NCC TTM service, &copy;RIPE NCC, 2003, all rights reserved.", br, 
			"Data may not be redistributed without prior approval of the RIPE NCC.",br, 
			"Please see ", a({ -name=>"aup", -href=>"http://www.ripe.net/ripe/docs/ripe-300.html"}, "http://www.ripe.net/ripe/docs/ripe-300.html"), "for details."
		), "<!--#include virtual=\"/includes/ripencc/ttmfoot1.inc\"-->"
	);
} # }}}

=item get_headers($$)

Backward compatibility with scripts that need it. 

Use print_header and print_footer instead.

=cut

sub get_headers($$) # {{{
{
	# XXX Should work with templates ala SSI

	my ($header) = join("", print_header($_[0],$_[1]));
	my ($footer) = join("", print_footer());
	return ($header,$footer);
} # }}}

=item cleanup_query_string($$$)

Cleanup the query string, replacing an existing parameter with a new
value. Order is important because of the in-document anchors.
Changing the order here will cause a fetch from the server and
not having the browser jump around in its cached copy.

=cut

sub cleanup_query_string($$$) # {{{
{
	my ($paramname,$newparamvalue,$cgi) = @_;
	my $cleanedup_cgi = new CGI($cgi);

	if (defined($newparamvalue))
	{
		$cleanedup_cgi->param($paramname,$newparamvalue);
	}
	else
	{
		# Remove parameter if undefined, the "clear hotlist" depends on it for instance...
		$cleanedup_cgi->delete($paramname);
	}
	return $cleanedup_cgi->query_string();
} # }}}

=item createzoommap($$$$$$;$)

Create a html image map for the delay, jitter and trends plot. It also
calculates zoomable regions for use in plots on demand software

The html structure is returned

B<Caveat:> The regions for the various plot types are hardcoded!

=cut

sub createzoommap($$$$$$;$) # {{{
{
	my ($URL, $timestamp_start, $timestamp_end, $slices, $type, $anchor, $id) = @_;
	my (%time_box, %delay_box, @areas);

	# id is an optional argument
	$id = $type unless defined($id);

	# Declare boxes to "superimpose"
	# Get these coordinates from e.g. with Gimp's web imagemap filter


	%time_box = (
		"delay" => [ 
						{ x => 46, y => 62, w => 304, h => 246 },	# top left
						{ x => 46, y => 376, w => 304, h => 246 },	# bottom left
						{ x => 434, y => 376, w => 304, h => 246 },	# bottom right
					],
		"jitter" => [
						{ x => 434, y => 63, w => 304, h => 246 },	# top right
						{ x => 434, y => 376, w => 304, h => 246 },	# bottom right
					], 
		"trends" => [
						{ x => 84, y => 53, w => 616, h => 166 },	# top
						{ x => 84, y => 266, w => 616, h => 166 },	# middle
						{ x => 84, y => 479, w => 616, h => 166 },	# bottom
					],
	);
	%delay_box = (
		"delay" => [
						{ x => 10, y => 62, w => 36, h => 246 },	# delay axis top left
						#{ x => 434, y => 310, w => 304, h => 25 },	# delay axis top right
						{ x => 434, y => 63, w => 304, h => 246 },	# top right
					],
		"jitter" => [
						# { x => 44, y => 624, w => 304, h => 25 },
						{ x => 46, y => 376, w => 304, h => 246 },	# bottom left
						{ x => 399, y => 376, w => 35, h => 246 },	# bottom right
					], 
		"trends" => [
						{ x => 20, y => 53, w => 64, h => 164 },
						{ x => 20, y => 266, w => 64, h => 164 },
						{ x => 20, y => 479, w => 64, h => 164 },
					],
	);
	
	my %__type_map = ( "" => "delay", ipdv => "jitter");

	@areas = ();

	$type = $__type_map{$type} if (exists($__type_map{$type}));

	# Create imagemaps for interactive zooming
	# Just use binary search here...

	if (exists $time_box{$type})
	{
		if($timestamp_end > 0)
		{
			my ($slice, $timestamp_delta);

			$timestamp_delta = $timestamp_end - $timestamp_start;

			for($slice = 0 ; $slice < $slices; $slice++)
			{
				# Clone parameter list for this slice

				my $URLslice = new CGI($URL);
				my $URLlimitdelay = new CGI($URL);

				my $delayrange = $URL->param('maxdelay') - $URL->param('mindelay');

				$URLslice->param( -name => 'embedded', -value => 'true');
				$URLslice->param( -name => 'plottype', -value => $type);

				# start of slice
				$URLslice->param( -name => 'date_start', -value => strftime("%Y%m%d", gmtime($timestamp_start + ($slice * ( $timestamp_delta / $slices)) )));
				$URLslice->param( -name => 'time_start', -value => strftime("%H:%M",   gmtime($timestamp_start + ($slice * ( $timestamp_delta / $slices)) )));

				# end of slice
				# XXX '-1' is under consideration as I do not for sure if this means "<" or "<=" for the ttplots server...

				$URLslice->param( -name => 'date_end', -value => strftime("%Y%m%d",   gmtime($timestamp_start + (($slice + 1) * ( $timestamp_delta / $slices)) -1 )));
				$URLslice->param( -name => 'time_end', -value => strftime("%H:%M",    gmtime($timestamp_start + (($slice + 1) * ( $timestamp_delta / $slices)) -1 )));

				# Limit delays here

				$URLlimitdelay->param( -name => 'embedded', -value => 'true');
				$URLlimitdelay->param( -name => 'mindelay', -value => ($URL->param('mindelay') + ( $slice * ( $delayrange / $ slices ) ) ));
				$URLlimitdelay->param( -name => 'maxdelay', -value => ($URL->param('mindelay') + ( ($slice + 1) * ( $delayrange / $slices ) ) ));

				# Create areas for each defined rectangle for the given plottype.
				# We assume the timescale is the same for all these rectangles...
					
				foreach my $time_box (@{$time_box{$type}})
				{
					my $message = "Click on image to show plot on demand ranging from " .
						strftime("%Y%m%d %H:%M", gmtime($timestamp_start + ($slice * ( $timestamp_delta / $slices)) )) .
						" till " .
						strftime("%Y%m%d %H:%M", gmtime($timestamp_start + (($slice + 1) * ( $timestamp_delta / $slices)) -1 ));

					push (@areas, Area({ -shape => "rect", 
							-coords => join(",",$time_box->{'x'} + ($slice * ceil($time_box->{'w'} / $slices)) , $time_box->{'y'}, $time_box->{'x'} + (($slice + 1) * ceil($time_box->{'w'} / $slices)) - 1, $time_box->{'y'} + $time_box->{'h'} ),
							-alt => $message, 
							-onmouseover=>maketooltip($message), 
							-href=>"${pod_url}?" . $URLslice->query_string() . "#$anchor"}));
				}

				# Create horizontal/vertical zoom for the min and max delay parameters. probably javascript only...

				foreach my $delay_box (@{$delay_box{$type}})
				{
					my $message = "Click here on this axis to limit the delay from " .
						$URLlimitdelay->param('mindelay') .
						" ms till " .
						$URLlimitdelay->param('maxdelay') .
						" ms";

					if ($delay_box->{'w'} < $delay_box->{'h'})
					{
						# Probably vertical axis, cope with it
						push (@areas, Area({ -shape => "rect", 
								-coords => join(",",$delay_box->{'x'}, ($delay_box->{'y'} + $delay_box->{'h'} ) - ($slice * ceil($delay_box->{'h'} / $slices)) , $delay_box->{'x'} + $delay_box->{'w'} , ( $delay_box->{'y'} + $delay_box->{'h'} ) - (($slice + 1) * ceil($delay_box->{'h'} / $slices)) - 1 ),
								-alt => $message, 
								-onmouseover=>maketooltip($message), 
								-href=>"${pod_url}?" . $URLlimitdelay->query_string() . "#$anchor"}));
					}
					else
					{
						# Horizontal axis
						push (@areas, Area({ -shape => "rect", 
								-coords => join(",",$delay_box->{'x'} + ($slice * ceil($delay_box->{'w'} / $slices)) , $delay_box->{'y'}, $delay_box->{'x'} + (($slice + 1) * ceil($delay_box->{'w'} / $slices)) - 1, $delay_box->{'y'} + $delay_box->{'h'} ),
								-alt => $message, 
								-onmouseover=>maketooltip($message), 
								-href=>"${pod_url}?" . $URLlimitdelay->query_string() . "#$anchor"}));
					}
				}

			}
		}
		if ( $type eq "delay" )
		{	
			push(@areas,Area({ -shape => "rect", -coords => "11,33,387,339", -alt => "Delay & Number-of-Hops as a function of time", -onmouseover=>maketooltip(ul($plotlegend{"delaytopleft"})), -target=>"legend", -href=>"http://www.ripe.net/ttm/Help/delayplot.html#delaytopleft"}),
				Area({ -shape => "rect", -coords => "397,32,773,339", -alt => "Distribution of delays", -onmouseover=>maketooltip(ul($plotlegend{"delaytopright"})), -target=>"legend", -href=>"http://www.ripe.net/ttm/Help/delayplot.html#delaytopright"}),
				Area({ -shape => "rect", -coords => "9,345,388,652", -alt => "Packets sent", -onmouseover=>maketooltip(ul($plotlegend{"delaybottomleft"})), -target=>"legend", -href=>"http://www.ripe.net/ttm/Help/delayplot.html#delaybottomleft"}),
				Area({ -shape => "rect", -coords => "397,345,773,650", -alt => "Packet arrival as a function of time", -onmouseover=>maketooltip(ul($plotlegend{"delaybottomright"})), -target=>"legend", -href=>"http://www.ripe.net/ttm/Help/delayplot.html#delaybottomright"}),
			);
		}
		elsif ( $type eq "jitter" )
		{
			push(@areas,Area({ -shape => "rect", -coords => "10,34,387,339", -alt => "Distribution of delay variations", -onmouseover=>maketooltip(ul($plotlegend{"ipdvtopleft"})), -target=>"legend", -href=>"http://www.ripe.net/ttm/Help/ipdvplot.html#ipdvtopleft"}),
				Area({ -shape => "rect", -coords => "399,33,776,339", -alt => "Delay variations vs. time", -onmouseover=>maketooltip(ul($plotlegend{"ipdvtopright"})), -target=>"legend", -href=>"http://www.ripe.net/ttm/Help/ipdvplot.html#ipdvtopright"}),
				Area({ -shape => "rect", -coords => "9,344,388,652", -alt => "One-way delay measurements: Packet delay distribution", -onmouseover=>maketooltip(ul($plotlegend{"ipdvbottomleft"})), -target=>"legend", -href=>"http://www.ripe.net/ttm/Help/ipdvplot.html#ipdvbottomleft"}),
				Area({ -shape => "rect", -coords => "398,344,776,651", -alt => "One-way delay measurements: Packet delay vs. time", -onmouseover=>maketooltip(ul($plotlegend{"ipdvbottomright"})), -target=>"legend", -href=>"http://www.ripe.net/ttm/Help/ipdvplot.html#ipdvbottomright"}),
			);
		}
		elsif ( $type eq "trends" )
		{
			push(@areas,Area({ -shape => "rect", -coords => "9,34,776,238", -alt => "One-way Delay Trend Plot", -onmouseover=>maketooltip(ul($plotlegend{"trendstop"})), -target=>"legend", -href=>"http://www.ripe.net/ttm/Help/trendplot.html"}),
				Area({ -shape => "rect", -coords => "9,246,776,454", -alt => "One-way Delay Trend Plot", -onmouseover=>maketooltip(ul($plotlegend{"trendsmiddle"})), -target=>"legend", -href=>"http://www.ripe.net/ttm/Help/trendplot.html"}),
				Area({ -shape => "rect", -coords => "9,459,776,660", -alt => "One-way Delay Trend Plot", -onmouseover=>maketooltip(ul($plotlegend{"trendsbottom"})), -target=>"legend", -href=>"http://www.ripe.net/ttm/Help/trendplot.html"}),
				Area({ -shape => "rect", -coords => "8,34,892,659", -alt => "One-way Delay Trend Plot", -target=>"legend", -href=>"http://www.ripe.net/ttm/Help/trendplot.html"}));
		}

		return Map( { -name => $id, -id => $id }, @areas, Area({ -shape => "default", -nohref => "nohref"}));
	}
	return ();
} # }}}

=item get_timestamp($$)

Convert a YYYY-MM-DD (I<1st parameter>) and a HH:MM:SS (I<second parameter>) time
back to seconds sinds the unix epoch

=cut

sub get_timestamp($$) # {{{
{
	my ($date,$time) = @_;

	my ($stamp);

	my ($realdate,$realtime) = ($date,$time);

	$date =~ s/\D//g;
	$time =~ s/\D//g;

	$date =~ /^(\d{4})(\d{2})(\d{2})$/ or do
	{
		die("Invalid date format \"$realdate\"\n");
	};

	my ($y,$m,$d) = ($1,$2,$3);

	($y>=1 and $m>=1 and $d>=1 and $m<=12 and $d<=31) or do
	{
		die("Invalid date format \"$realdate\"\n");
	};

	$time .= '0'x(6 - length($time));

	$time =~ /^(\d{2})(\d{2})(\d{2})$/ or do
	{
		die("Invalid time format \"$realtime\"\n");
	};

	my ($h,$mn,$s) = ($1,$2,$3);

	($h>=0 and $h<=23 and $mn>=0 and $mn <=59 and $s>=0 and $s<=59) or do
	{
		die("Invalid time format \"$realtime\"\n");
	};

	my $t = timegm($s,$mn,$h,$d,$m-1,$y);

	$t == -1 and die("Invalid date format $realdate - $realtime\n");

	return ($t);
} # }}}

=item CGIerror

Pretty print error messages when cgi's are calling die and such

=cut

sub CGIerror # {{{
{
	my $msg = shift;	
		print print_header("Test Traffic Measurement","Runtime error"),
			p({ -class=>"error"}, b("ERROR: "),"$msg"),
		print_footer
} # }}}

=item listoftestboxes($)

Fill in the following hashes and lists in this module:
C<%testboxes, @testboxes6, @testboxes4>

The contents can be filtered by an active hotlist

=cut

sub listoftestboxes($) # {{{
{
	my ($testboxes) = $_[0];

	# Fills in: %testboxes, @testboxes6, @testboxes4
	# Possible hotlist filtering

	open (LOT, $testboxes) or die("Cannot open $testboxes: $!");
	while (defined( my $input = <LOT>))
	{
		my ($id, $name, $address) = $input =~ /^(\d+)\s+(tt\d+[a-z]?)\.ripe\.net\s+(\S+)\s*$/ or next;

		if (exists($testboxes{$name}))
		{
			push(@{$testboxes{$name}},$address);
		}
		else
		{
			$testboxes{$name} = [ $address ];
		}

		push (@testboxes6, $name) if $address =~ /^([0-9A-Fa-f]{0,4}:([0-9A-Fa-f]{0,4}:){1,6}[0-9A-Fa-f]{1,4})$/;
		push (@testboxes4, $name) if $address =~ /^(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})$/;
	}
	close (LOT);
	# Hotlist support 

	if ($#hotlist > -1)
	{
		# only include boxes defined in the hotlist, if the user has one

		@testboxes4 = grep {/^(@{[join('|',@hotlist)]})$/} @testboxes4;
		@testboxes6 = grep {/^(@{[join('|',@hotlist)]})$/} @testboxes6;
	}
} # }}}

=item gpsinfo($)

Fills in C<%gpsinfo> in this module, it might be filtered by an active
hotlist and for sure is filtered by a chainmap if C<$use_chain_auth> is set to
an non false value. (but not within this routine)

=cut

sub gpsinfo($) # {{{
{
	# Fills in: %gpsinfo
	# Requires: %chainmap (specific chain filtered chainmap)
	#
	# Filtered by %chainmap
	# Possible hotlist filtering

	my ($gpsinfo) = $_[0];
	my ($status);

	open(GPSINFO, $gpsinfo) || die "Couldn't open $gpsinfo: $!";
	while (defined(my $input = <GPSINFO>))
	{
		chomp $input;
		if ($input =~ /^status:\t(\w+)$/) {
			# is this a status line
			$status = $1;
		} else {
			# is this a testbox line?
			my ($name, $lat, $long, $place) = split ('\t', $input);

			# TODO Hotlist support goes in here!
			if ($#hotlist > -1)
			{
				next unless grep { /^${name}$/ } @hotlist;
			}

			$gpsinfo{$name}{'place'}  = $place;
			$gpsinfo{$name}{'status'} = $status;
			$gpsinfo{$name}{'lat'} = $lat;
			$gpsinfo{$name}{'long'} = $long;
			$gpsinfo{$name}{'longdesc'} = "$place ($status)";
			$gpsinfo{$name}{'fqdn'} = "${name}.ripe.net";

			($gpsinfo{$name}{'org'}, $gpsinfo{$name}{'city'}, $gpsinfo{$name}{'country'}) = split (/,/, $place);
			foreach my $twonicid ("a" .. "z")
			{
				$gpsinfo{"${name}${twonicid}"} = $gpsinfo{$name} if exists($testboxes{"${name}${twonicid}"});
			}
		}
	}
	close(GPSINFO);

} # }}}

=item retrieve_hotlist()

Tries to get an active hotlist from a cookie depending on current IP protocol
selected. Returns an @array

=cut

sub retrieve_hotlist() # {{{
{
	my ($cgi,$ipv,@hotlist);

	$cgi = exists($ENV{'REDIRECT_QUERY_STRING'}) ?
			new CGI($ENV{'REDIRECT_QUERY_STRING'}) :
			new CGI;

	# user want to surpress hotlist usage
	return () if defined($cgi->param('nohotlist')) || $nohotlist;

	$ipv = ( defined($cgi->param('ipv')) && $cgi->param('ipv') =~ /^[46]$/ ) ? $cgi->param('ipv') : 4;

	if ($cgi->cookie("hotlist$ipv"))
	{
		@hotlist = $cgi->cookie("hotlist$ipv");
	}
	return (@hotlist);
} # }}}

=item check_license()

Checks wether the end user has accepted the acceptable use policy for the TTM
data

=cut

sub check_license() # {{{
{
	my ($cgi,$ipv,@hotlist);

	$cgi = exists($ENV{'REDIRECT_QUERY_STRING'}) ?
			new CGI($ENV{'REDIRECT_QUERY_STRING'}) :
			new CGI;

	if (!($cgi->cookie("ttmlicense") && $cgi->cookie("ttmlicense") eq "YES"))
	{
		# user didn't accepted the license yet, redirect to license CGI
	}
} # }}}

1;
