#!/usr/local/bin/perl -wT

# $Id: pod.cgi,v 1.7 2008/04/09 14:11:06 ruben Exp $

# Copyright (c) 2000,2001                            RIPE NCC
#
# All Rights Reserved
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose and without fee is hereby granted,
# provided that the above copyright notice appear in all copies and that
# both that copyright notice and this permission notice appear in
# supporting documentation, and that the name of the author not be
# used in advertising or publicity pertaining to distribution of the
# software without specific, written prior permission.
#
# THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
# ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS; IN NO EVENT SHALL
# AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY
# DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
# AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

#------------------------------------------------------------------------------
# Module Header
# Filename          : plots.pl 
# Purpose           : CGI Front-end to TTM plots-on-demand 
# Author            : Ruben van Staveren, Rene Wilhelm
# Date              : 20030429
# Description       : This program creates a HTML page with an embedded
#		      TTM delay plot (reference to plots-on-demand) and
#		      a pointer to the TTM traceroute query for the
#		      corresponding time period.
# Language Version  : Perl 5.6
# OSs Tested        : Linux, Solaris 8
# Command Line      : N/A
# Input Files       : Web form, TTM configuration file (list_of_testboxes)
# Output Files      : Dynamic Web page
# Credits	    : main structure copied from tracerouteDB CGI by Manuel Valente
# External Programs : txt2gif
#------------------------------------------------------------------------------

use 5.006;

use strict;
use CGI 2.86 qw/-newstyle_urls -nosticky :standard :imagemap/;
use CGI::Carp;
use IO::Socket;
use Time::Local;
use POSIX;
use LWP::UserAgent;
use HTTP::Request::Common;

# FIXME makemap(1) on solaris uses ndbm
# use NDBM_File;

sub _abort($);
sub _is_included();
sub get_result();
sub get_timestamp($$);
sub print_remark();
sub print_footer();
sub print_form();
sub print_header($$);
sub tboxsort();
sub createzoommap($$$$$$);
sub maketooltip($);

# Basic authentification 
my %dbmgroup;
my $dbmgroupfile	= "/var/www/dynamic/tt/dbmgroup"; 
my $user			= $ENV{REMOTE_USER};
my @usergroups;

# Chainmap authentification
my %chainmap;
my $chainmapfile	= "/home/ttraffic/ttpro/www/data/plots/chainmap";
# my $chainmapfile	= "/nonexistant/chainmap";

my $ttplots_server	= "cedar.ripe.net";
my $ttplots_port	= 4250;
# my $traceroutecgi	= "/plots/routing.cgi";
my $traceroutecgi	= "/cgi-bin/ttm/routing";

my $use_basic_auth	= 0; # We can use basic auth through http here
my $use_chain_auth  = 0; # We can make use of the $ENV{'TTMCHAIN'} enviroment to restrict access
my $use_ttplots_http_proxy = 0; # ttplots service is firewalled and we use LWP::UserAgent to retrieve our data from http://www.ripe.net/ instead

$ENV{'PATH'} 		= "/bin/:/usr/bin:/ncc/ttpro/bin:/var/www/bin/np/";

					# List of Test Boxes - will need to be updated

my $testboxes		= '/var/www/pub-dyn/projects/ttm/Plots/LIST_OF_TESTBOXES';
my $gpsinfo		= '/var/www/pub-dyn/projects/ttm/Plots/GPSINFO';

my $maintitle		= 'Test Traffic Measurements';
my $subtitle		= exists $ENV{'TTMCHAIN'} ? "Plots on Demand for $ENV{'TTMCHAIN'}" : "Plots on Demand";

my $maxdays			= 31;	# maximum timeperiod

my %plotlegend = (
	delaytopleft => 
		li(b("Delay & Number-of-Hops as a function of time")) . 
		li("The black points show the delays as function of time.") .   li("The red points depict the number of hops (times 10) detected in the closest preceding traceroute measurement; unless the number of hops is changing rapidly, these points will merge into line segments. Obviously, the number of hops only flags drastic changes in the routing; subtle differences like load balancing over multiple communication lines do not show up. For more detailed information on routing, the TTM traceroute database should be consulted."),
	delaytopright => 
		li(b("Distribution of delays")) . 
		li("This histogram shows the distribution of the delays over the time interval of the previous plot, e.g. it is the projection along the y-axis of the previous plot. The number of entries, over- and underflows are the same as for the previous plot."),
	delaybottomleft => 
		li(b("Packets sent")) . 
		li("The " . font({ -color=>"green" },"green area") . " shows the number of packets sent per bin of time with valid NTP time-stamps on both the sending and receiving side.") .
		li("The " . font({ -color=>"yellow" },"yellow area") . " shows the number of packets sent per bin of time with a valid NTP time-stamp on the sending but not on the receiving side") .
		li("The " . font({ -color=>"purple" },"purple area") . " shows the number of packets sent per bin of time with a valid NTP time-stamp on the receiving but not on the sending side.") .
		li("The " . font({ -color=>"red" },"red area") . " shows the number of packets sent per bin of time without valid NTP time-stamps on both the sending and receiving side.") .
		li("The " . font({ -color=>"cyan" },"cyan area") . " shows the number of packets per bin of time that were lost. These packets left the sender but were not registered by the receiver, either because they were lost in the network or because the receiver was off-line."),
	delaybottomright => 
		li(b("Packet arrival as a function of time")) . 
		li("For each bin of the bottom left plot, the bottom right histogram shows which fraction of the sent packets arrived at the destination box (regardless of clock status). When the experiment is not running, i.e. no packets sent, the \"fraction arrived\" bins are set to 0. To distinguish between a situation of 100% packet loss, one will have to check the \"packets sent/valid\" bottom left plot."),
	ipdvtopleft => 
		li(b("Distribution of delay variations")) . 
		li("This histogram shows the distribution of the delay variations (i.e. the difference between consecutive one-way delay measurements) for the specified time period."),
	ipdvtopright => 
		li(b("Delay variations vs. time")) . 
		li("This shows the delay variations as seen at the various times in the measurement period. The top left plot is the 1-D projection of this."),
	ipdvbottomleft => 
		li(b("One-way delay measurements: Packet delay distribution")) . 
		li("For easy reference, the results of one-way delay measurements used in the delay variation calculations are shown in the bottom plots. the bottom left plot shows the distribution"),
	ipdvbottomright => 
		li(b("One-way delay measurements: Packet delay vs. time")) . 
		li("For easy reference, the results of one-way delay measurements used in the delay variation calculations are shown in the bottom plots. the bottom right plot shows the delay measurments vs. time. In addition, red dots indicate the number of hops (times 10) seen in the nearest traceroute measurement."),
);

# Make usable through SSI and standalone operation
my $cgi = exists($ENV{'REDIRECT_QUERY_STRING'}) ?
		new CGI($ENV{'REDIRECT_QUERY_STRING'}) :
		new CGI;

# FIXME unused...
my $self_url = exists($ENV{'REDIRECT_SCRIPT_URL'}) ? 
				$ENV{'REDIRECT_SCRIPT_URL'} :
				$cgi->url(-relative=>1);

my $ua = LWP::UserAgent->new(env_proxy => 1,
                             keep_alive => 1,
                             timeout => 120,
                            );

my %BoxList;		# list of ALL boxes, including historic ones
my @GPSINFO;		# list of current boxes (SETUP, ON or OFF) plus location

my ($status, $f, %TTvals, @TTLIST, %ttInfo);

# Javascript helpers
# XXX FIXME Should be reworked to automaticly detect and handle checkboxes/droplists etc. 
# Not all browsers support this.value as a means to access the current setting or set it to a value

my $validate_range = <<'EOF';
function validate_range(obj, lowval, hival, defval, message) {
if ((obj.value < lowval) || (obj.value > hival))
        {
                obj.value=defval;
                alert(message);
                return false;
        }
        return true;
}
EOF

my $validate_regex = <<'EOF';
function validate_regex(obj, regex, defval, message) {
        var re = new RegExp(regex);
        if (!re.test(obj.value))
        {
                obj.value=defval;
                alert(message);
                return false;
        }
        return true;
}
EOF

my $validate_equality = <<'EOF';
function validate_equality(thelvalue, thevalue, theoperator, defval, message) {
		var code="'" + String(thelvalue) + "' " + String(theoperator) + " '" + String(thevalue) + "'";
        if (!eval(code))
        {
                thelvalue=defval;
                alert(message);
                return false;
        }
        return true;
}
EOF


# Main function
MAIN:
{
 	my (%Chainlist, %list);
 	
	if ($use_chain_auth)
	{
		die "TTMCHAIN not in environment, bailing out" unless exists $ENV{'TTMCHAIN'};
		tie (%chainmap, 'NDBM_File', $chainmapfile,O_RDONLY,0640) or _abort("couldn't tie chainmap: $!");
	}

	if ($use_basic_auth)
	{
		tie (%dbmgroup, 'DB_File', $dbmgroupfile) or _abort("couldn't tie ACL database: $!");
		@usergroups = (split (/,/, $dbmgroup{$user}));
	}
		
	open (F, $testboxes) or _abort ("Cannot open $testboxes: $!");
	
	while (<F>)
	{
		/^(\d+)\s+(tt\d+[a-z]?\.ripe\.net)\s+\S+\s*$/ or next;
		$BoxList{$2} = $1;
	};      
	close (F);
	
	if ($use_chain_auth)
	{
		%Chainlist = map { s/^$ENV{'TTMCHAIN'}\+(tt\d+[a-z]?)\+tt\d+[a-z]?/$1/; $_ => 1} (grep { /^$ENV{'TTMCHAIN'}/} keys %chainmap);
	}

	@GPSINFO = `cat $gpsinfo`;
	chomp @GPSINFO;
	foreach $f (@GPSINFO) {
		if ($f =~ /^status:\t(\w+)$/) {
			# is this a status line?
			$status = $1;
		} else {
			# is this a testbox line?
			my ($name, $lat, $long, $place) = split ('\t', $f);

			if ($use_chain_auth)
			{
				next unless exists $Chainlist{$name};
			}

			$ttInfo{$name}{'place'}  = $place;
			$ttInfo{$name}{'status'} = $status;

			$TTvals{$name} = "$name: $place";

			foreach my $twonicid ("a" .. "z")
			{
				if(exists($BoxList{"${name}${twonicid}.ripe.net"}))
				{
					$TTvals{"${name}${twonicid}"} = "${name}${twonicid}: $place";
					$ttInfo{"${name}${twonicid}"}{'place'}  = $place;
					$ttInfo{"${name}${twonicid}"}{'status'} = $status;
				}
			}
		}
	}

	if ($use_basic_auth)
	{
		# We can do this because a usergroup can bear the same name as a testbox
		# The grep here thus wields out possible selections this user has no access to

		@TTLIST = sort tboxsort grep { /^(@{[join('|',@usergroups)]})$/ } keys %TTvals;
	}
	else
	{
		@TTLIST = sort tboxsort keys %TTvals;
	}
	
	# Check if we have an image to return or just should print the form

	if (defined ($cgi->param('SRC')) && ($cgi->param('SUBMIT') ne "false"))
	{
		# Makes identification of multiple open browser tabs/windows easier
		$subtitle .=" \[".$cgi->param('SRC')." - ".$cgi->param('DST')."\]";		

		get_result();
	}
	else
	{
		# Print HTML header for browser
		print_header($maintitle,$subtitle);
		print print_form, print_remark;
		print_footer;
	};
};

# Purpose           : Send query, return results nicely formatted
# Comments          : 
# Input             : List of parameters to pass the POD server:
#                     SRC box, DST box, time start, time end, format
# Output            : formatted HTML or Image data on stdout

sub get_result()
{
	my (@line,$c,$ll);
	
	my $content = '';
	
	# Build the command to run

	my $src = $cgi->param('SRC');
	my $dst = $cgi->param('DST');
	my $maxdelay = $cgi->param('maxdelay') || 250 ;
	my $mindelay = $cgi->param('mindelay') || 0 ;
	my $zoomlevel = $cgi->param('zoomlevel') || 6 ;
	my $format = $cgi->param('format') || "gif" ;
	my $plottype = $cgi->param('plottype')|| "delay";
	my $ipv = $cgi->param('ipv')|| "4";
	my $logscale = $cgi->param('logscale') || "0";
	
	$format = "ps"  if ($format eq "PostScript") ;
	$format = "csv" if ($format eq "ascii") ;
	$format = "pdf"  if ($format eq "PDF") ;

	if ($src eq $dst) {
		&_abort ("Source and Destination are identical - $src.");
	}
	
	$src =~ s/(tt\d+[a-z]?)\.ripe\.net/$1/;
	$dst =~ s/(tt\d+[a-z]?)\.ripe\.net/$1/;

	# Check src and destination are in valid range
	# (not listed in the form, but historic boxes are silently accepted)

	my $boxlist = " " .  join(" ", keys %BoxList) . " " ;
	if ($boxlist !~  m/\s${src}[\.\s]/) {
		&_abort ("Invalid source box - $src.");
	}

	if ($boxlist !~  m/\s${dst}[\.\s]/) {
		&_abort ("Invalid destination - $dst.");
	}
	
	if ($use_chain_auth)
	{
		_abort("$src -> $dst: access denied. soure and destination must be in this chain.")	unless defined($chainmap{"$ENV{'TTMCHAIN'}+${src}+${dst}"}); 
	}


	if ($use_basic_auth)
	{
		# Now check if source or destination matches access rights

		my $grouplist = " " . join(" ", @usergroups) . " ";
		$grouplist =~ s/ \S+_admin//g;

		# XXX if the "tipu" user also has access to the "hidden" groups, the '$user ne "tipu"' may be removed

		if ( $user ne "tipu" && ( ($grouplist !~  m/\s${src}[\.\s]/) &&
			 ($grouplist !~  m/\s${dst}[\.\s]/) )) {
			$grouplist =~ s/(\S+) (\S+) $/$1 or $2/;
			&_abort ("$src -> $dst: access denied. soure or destination must be $grouplist.");
		}
	}

	# Check other parameters

	if (! (($zoomlevel =~ /^\s*\d+\s*$/) || 
	       ($zoomlevel =~ /^\s*\d+\.\d+\s*$/)) ) {
		&_abort ("Min Delay field ($zoomlevel) must be numeric");
	}
	
	if (! (($mindelay =~ /^\s*\d+\s*$/) || 
	       ($mindelay =~ /^\s*\d+\.\d+\s*$/)) ) {
		&_abort ("Min Delay field ($mindelay) must be numeric");
	}

	if (! (($maxdelay =~ /^\s*\d+\s*$/) || 
	       ($maxdelay =~ /^\s*\d+\.\d+\s*$/)) ) {
		&_abort ("Max Delay field ($maxdelay) must be numeric");
	}

	if ($mindelay > $maxdelay) {
		&_abort ("Min Delay <em>larger</em> than Max Delay");
	}

	# FIXME Hardcoded constants should go in top of code
	if ($zoomlevel > 8 || $zoomlevel < 2 ) {
		&_abort ("Zoom level <em>out of range</em>");
	}

	my $tstart = get_timestamp ($cgi->param('date_start'),$cgi->param('time_start'));
	
	$tstart == -1 and return();
	
	my $tend   = get_timestamp ($cgi->param('date_end'),$cgi->param('time_end'));
			
	$tend == -1 and return();

	my $timelimit = 86400 * $maxdays;
	if (($tend - $tstart) < 0) {
		&_abort("End date/time <em>before</em> Start date/time");
	}
	elsif (($tend - $tstart) > $timelimit) {
		
		if ($plottype eq "delay") {
			# backward compatibility: long period -> trends
			$plottype = "trends";
		}
		elsif ($plottype eq "trends") {
			# nothing to do	
		}
		else {
			&_abort ("Time period too large; maximum is $maxdays days");
		}
	}
	
	$mindelay =~ s/\s+//g;
	$maxdelay =~ s/\s+//g;

	my $dstart = $cgi->param('date_start');
	my $dend   = $cgi->param('date_end');
	$tstart    = $cgi->param('time_start');
	$tend      = $cgi->param('time_end');

	my $af = "INET";
	if ($ipv eq "6") {
		$af = "INET6";
	}
		
	my $embedded = $cgi->param('embedded') || "true";

	if (($format eq "csv") && ($plottype eq	"jitter")) {
		_abort("$plottype: ASCII format not (yet) implemented\n");
	}

	if ($embedded eq "false" || ($format eq "ps") || ($format eq "pdf") || ($format eq "csv")) {

		if (!$use_ttplots_http_proxy)
		{
			# connect to ttplots server and fetch plot

			my $delta_t = get_timestamp($dend,$tend) - get_timestamp($dstart,$tstart);

			$tend = $dend;
			$tstart = $dstart;
			$tstart .= "T" . $cgi->param('time_start') if ($cgi->param('time_start'));
			$tend .= "T" . $cgi->param('time_end') if ($cgi->param('time_end'));
			$logscale = $logscale eq 'yes' ? 'logscale' : '';

			my $command = sprintf("%02d %02d",$BoxList{"${src}.ripe.net"}, $BoxList{"${dst}.ripe.net"}) . " $tstart $tend $mindelay $maxdelay $format $plottype $af all $logscale\n";

			my $socket = IO::Socket::INET->new(PeerAddr => $ttplots_server,
				PeerPort => $ttplots_port,
				Proto    => "tcp",
				Type     => SOCK_STREAM)
			or &_abort("Couldn't connect to plot server");


			print $socket $command;

			my $line = <$socket>;   # get first line

			if ($format eq "csv") {
				print header(-type=>'text/plain', -attachment=>"${src}-${dst}-${plottype}-${dstart}-${delta_t}.txt");
				do {print $line;} while ($line = <$socket>);
			}
			elsif (substr($line, 0, 5) eq "%!PS-") {
				print header(-type=>'application/postscript', -attachment=>"${src}-${dst}-${plottype}-${dstart}-${delta_t}.ps");
					do {print $line;} while ($line = <$socket>);
			}
			elsif (substr($line, 0, 5) eq "GIF87")  { 
				print header(-type=>'image/gif', -attachment=>"${src}-${dst}-${plottype}-${dstart}-${delta_t}.gif");
				do {print $line;} while ($line = <$socket>);
			}
			elsif (substr($line, 0, 5) eq '%PDF-') {
				print header(-type=>'application/pdf', -attachment=>"${src}-${dst}-${plottype}-${dstart}-${delta_t}.pdf");
					do {print $line;} while ($line = <$socket>);
			}
			else {
					do {$content .= $line;} while ($line = <$socket>);
					&_abort($content);
			}
			close($socket);
		}
		else
		{
			# connect to http://www.ripe.net/cgi-bin/ttm/plots-on-demand and fetch plot if ttplots is firewalled

			my $delta_t = get_timestamp($dend,$tend) - get_timestamp($dstart,$tstart);
			my $URL = "http://www.ripe.net/cgi-bin/ttm/plots-on-demand?ipv=$ipv&SRC=$src&DST=$dst&date_start=$dstart&date_end=$dend&time_start=$tstart&time_end=$tend&mindelay=$mindelay&maxdelay=$maxdelay&format=$format&plottype=$plottype&logscale=$logscale&embedded=false";
			my $_header_printed=0;

			$ua->credentials("www.ripe.net:80", "Test Traffic", "tipu", "4T1poeOnl1");
			$ua->request(GET($URL), sub {
					my($data, $response, $protocol) = @_;
					if (!$_header_printed)
					{
						# This adjusts the filename in the save-as dialog box
						my %extensionmap = qw(text/plain .txt application/postscript .ps image/gif .gif image/png .png application/pdf .pdf);

						print header(-type=>scalar($response->content_type), -attachment=>"tt${src}-tt${dst}-${plottype}-${dstart}-${delta_t}$extensionmap{$response->content_type}");
						$_header_printed=1;
					}
					print $data;
				}
			);
		}
	}
	else {
		my ($URL, $URLswap, $URLexpanddelayscale, $URLexpandtimescale, $QUERYtraceroutes, $traceroutes, $timestamp_start, $timestamp_end, $timestamp_delta);

		$URL = new CGI("ipv=$ipv&SRC=$src&DST=$dst&date_start=$dstart&date_end=$dend&time_start=$tstart&time_end=$tend&mindelay=$mindelay&maxdelay=$maxdelay&format=$format&plottype=$plottype&logscale=$logscale&embedded=false");
		$QUERYtraceroutes = new CGI("ipv=$ipv&SRC=$src&DST=$dst&date_start=$dstart&date_end=$dend&time_start=$tstart&time_end=$tend");

		$URLswap = new CGI($URL);
		$URLexpanddelayscale = new CGI($URL);
		$URLexpandtimescale = new CGI($URL);

		$traceroutes = "${traceroutecgi}?" . $QUERYtraceroutes->query_string();

		$timestamp_start = get_timestamp($dstart,$tstart);
		$timestamp_end = get_timestamp($dend,$tend);

		$timestamp_delta = ($timestamp_end - $timestamp_start) * $zoomlevel;

		# FIXME Prevent usage of not existing hours as it won't plot

		$URLexpandtimescale->param( -name => 'date_end', -value => strftime("%Y%m%d",	gmtime(($timestamp_end + $timestamp_delta) > ( time() - 86400 ) ? ( time() - 86400 ) : ($timestamp_end + $timestamp_delta))));
		$URLexpandtimescale->param( -name => 'time_end', -value => strftime("%H:%M",	gmtime(($timestamp_end + $timestamp_delta) > ( time() - 86400 ) ? ( time() - 86400 ) : ($timestamp_end + $timestamp_delta))));
		$URLexpandtimescale->param( -name => 'date_start', -value => strftime("%Y%m%d",	gmtime(($timestamp_start - $timestamp_delta) < 0 ? 0 : ($timestamp_start - $timestamp_delta))));
		$URLexpandtimescale->param( -name => 'time_start', -value => strftime("%H:%M",	gmtime(($timestamp_start - $timestamp_delta) < 0 ? 0 : ($timestamp_start - $timestamp_delta))));

		$URLexpanddelayscale->param( -name => 'mindelay', -value => ($mindelay - (($maxdelay - $mindelay) * $zoomlevel)) < 0 ? 0 : ($mindelay - (($maxdelay - $mindelay) * $zoomlevel)) );
		$URLexpanddelayscale->param( -name => 'maxdelay', -value => ($maxdelay + (($maxdelay - $mindelay) * $zoomlevel)) );

		$URLswap->param(-name => "SRC", -value => $dst);
		$URLswap->param(-name => "DST", -value => $src);

		$URLswap->param(-name => "embedded", -value => "true");
		$URLexpanddelayscale->param(-name => "embedded", -value => "true");
		$URLexpandtimescale->param(-name => "embedded", -value => "true");

		print_header($maintitle, $subtitle);
		print print_form,
			p("If your browser is javascript enabled, it will use ", a({ -href => "http://www.walterzorn.com/tooltip/tooltip_e.htm"}, "tooltips"),  "to show zoom options and legends when you hover above the image"),
			p("Click on the time related plots to zoom in on another interval, Click on the delay distribution plots to limit the visible delay boundaries"),
			CGI::div( { -align => "center" },
				table( Tr(
				td({ -align => "center" },"[", a({ -href => $traceroutes, -target=> "traceroutes"}, "collected traceroutes for this period"), "]"),
				td({ -align => "center" },"[", a({ -href => $URLswap->url(-absolute=>1, -query=>1) . "#plot" }, "swap source and destination"), "]"),
				td({ -align => "center" },"[", a({ -href => $URLexpanddelayscale->url(-absolute=>1, -query=>1) . "#plot"}, "expand delay boundaries"), "]"),
				td({ -align => "center" },"[", a({ -href => $URLexpandtimescale->url(-absolute=>1, -query=>1) . "#plot" }, "expand timeline"), "]"),
				))
			),
			p(font({ -size=>"-1" }, "If image fails to load first time, try again with reload/refresh button of your browser")),
			p(a({ -name => "plot"}, ""),img({-border=>0, -usemap=>"#$plottype$timestamp_start$timestamp_end", -alt=>"Please wait while image is computed", -width=>903, -height=>666, -src=>$URL->url(-absolute=>1, -query=>1)})),
			p(a({-href=>$URL->url(-absolute=>1, -query=>1)}, i("Download image..."))),
			createzoommap($URL, $timestamp_start, $timestamp_end, $zoomlevel, $plottype, "plot"),	
			print_remark;
       		print_footer;
	}
};

sub get_timestamp($$)
{
	my ($date,$time) = @_;
	
	my ($stamp);
	
	my ($realdate,$realtime) = ($date,$time);
	
	$date =~ s/\D//g;	
	$time =~ s/\D//g;
	
	$date =~ /^(\d{4})(\d{2})(\d{2})$/ or do
	{
		&_abort ("Invalid date format $realdate");
	};
	
	my ($y,$m,$d) = ($1,$2,$3);
	
	($y>=1 and $m>=1 and $d>=1 and $m<=12 and $d<=31) or do
	{
		&_abort ("Invalid date format $realdate");
	};	
	
	$time .= '0'x(6 - length($time));
	
	$time =~ /^(\d{2})(\d{2})(\d{2})$/ or do
	{
		&_abort  ("Invalid time format $realtime");
	};
	
	my ($h,$mn,$s) = ($1,$2,$3);
	
	($h>=0 and $h<=23 and $mn>=0 and $mn <=59 and $s>=0 and $s<=59) or do
	{
		&_abort  ("Invalid time format $realtime");
	};
		
	my $t = timegm($s,$mn,$h,$d,$m-1,$y);
	
	$t == -1 and &_abort("Invalid date format $realdate - $realtime");
	
	return ($t);
}



# Purpose           : Print the FORM for entering data
# Comments          : 
# Input             : CGI object, List of boxes
# Output            : content (html data)

sub print_form()
{
	my ($hour, $day, $mon, $year);
	my ($startday, $endday);

	# Set defaults

	my $content = '';

	($a, $a, $hour, $day, $mon, $year, $a, $a, $a) = gmtime(time());

	# yesterday's ROOT files should be there by 0800 GMT

	if ($hour > 7) {
		$endday = strftime("%Y%m%d",gmtime());
		$startday = strftime("%Y%m%d",gmtime(time()-86400));
	}
	else {
		# last data until the end of day before yesterday
		$endday = strftime("%Y%m%d",gmtime(time()-86400));
		$startday = strftime("%Y%m%d",gmtime(time()-(2*86400)));
	}

	my $src = $cgi->param('SRC')||$TTLIST[0];
	my $dst = $cgi->param('DST')||$TTLIST[1];
	my $d_s = $cgi->param('date_start')|| $startday;
	my $t_s = $cgi->param('time_start')||'00:00';
	my $d_e = $cgi->param('date_end')|| $endday;
	my $t_e = $cgi->param('time_end')||'00:00';
	my $zoomlevel = $cgi->param('zoomlevel')||"6";
	my $mindelay = $cgi->param('mindelay')||"0";
	my $maxdelay = $cgi->param('maxdelay')||"250";
	my $format = $cgi->param('format')|| "gif";
	my $plottype = $cgi->param('plottype')|| "delay";
	my $ipv = $cgi->param('ipv')|| "4";

	# FORM
	
	return start_form(-method=> "GET",
				-name=>"theform",
				-action=>$cgi->url(-relative=>1) . "#plot",
				),
	table({ -border=>0, -cellspacing=>5 },
		Tr(td({ -colspan=>6 }, "<HR>")),

		# Print all testboxes for SRC combo
		Tr(
		td("SRC:"), 
		td( {-colspan=>5, -align =>"right" }, 
			popup_menu(-name=> "SRC",
						-values=> \@TTLIST,
						-labels => \%TTvals,
						-default => $src,
						-onChange => "return validate_equality(this.options[this.selectedIndex].value, this.form.DST.options[this.form.DST.selectedIndex].value, '!=', '$src', 'Source cannot be the same as the destination')"
						)
			)
		),

		# Print all testboxes for DST combo
		Tr(
		td("DST:"), 
		td( {-colspan=>5, -align =>"right" }, 
			popup_menu(-name=> "DST",
						-values=> \@TTLIST,
						-labels => \%TTvals,
						-default => $dst,
						-onChange => "return validate_equality(this.options[this.selectedIndex].value, this.form.SRC.options[this.form.SRC.selectedIndex].value, '!=', '$dst', 'Destination cannot be the same as the source')"
						)
			)
		),

		Tr(
			td("Min Delay (ms):"),td(textfield( -name=>"mindelay", -size=>4, -onChange=> "return validate_regex(this,'^[0-9]+(\.[0-9]+)?\$','$mindelay','Please enter a number') && validate_range(this,0,Number(this.form.maxdelay.value),'$mindelay','Min Delay needs to be below Max Delay')", -default=>$mindelay)),
			td("Start Date"),td(textfield( -name=>"date_start", -size=>10, -onChange=> "return validate_regex(this,'^[0-9]{8}\$','$d_s','Please enter a date in the format YYYYMMDD')", -default=>$d_s)),
			td("Start Time"),td(textfield( -name=>"time_start", -size=>5, -onChange=> "return validate_regex(this,'^[0-2][0-9]:[0-5][0-9]\$','$t_s','Please enter a time in the format HH:MM')", -default=>$t_s),
				a({ -href=>"http://www.worldtimezone.com/time-world.htm"}, "UTC")
				),
		),
		Tr(
			td("Max Delay (ms):"),td(textfield( -name=>"maxdelay", -size=>4, -onChange=> "return validate_regex(this,'^[0-9]+(\.[0-9]+)?\$','$maxdelay','Please enter a number') && validate_range(this,Number(this.form.mindelay.value),4294967295,'$maxdelay','Max Delay needs to be above Min Delay')", -default=>$maxdelay)),
			td("End Date"),td(textfield( -name=>"date_end", -size=>10, -onChange=> "return validate_regex(this,'^[0-9]{8}\$','$d_e','Please enter a date in the format YYYYMMDD')", -default=>$d_e)),
			td("End Time"),td(textfield( -name=>"time_end", -size=>5, -onChange=> "return validate_regex(this,'^[0-2][0-9]:[0-5][0-9]\$','$t_e','Please enter a time in the format HH:MM')", -default=>$t_e),
				a({ -href=>"http://www.worldtimezone.com/time-world.htm"}, "UTC")
				),
		),
		Tr(
			td("Output format:"),
			td( { -colspan => 2},
			radio_group(-name => "format",
						-values => [ "gif", "PostScript", "PDF", "ascii" ],
						-default => $format)),
			td("&nbsp;"),
			td( { -colspan => 2},
				radio_group(-name => "ipv",
							-values => [ "4", "6" ],
							-labels => { "4" => "IPv4" , "6" => "IPv6" },
							-default => $ipv),
				"&nbsp;",
				checkbox(-name => "logscale",
							-value => 'yes',
							-label => 'logscale',	
				),
			),
		),
		Tr(
			td("Plot type:"),
			td( { -colspan => 2},
			radio_group(-name => "plottype",
						-values => [ "delay", "trends", "jitter" ],
						-default => $plottype)),
			td(  { -colspan => 3},

			# FIXME Hardcoded constants should go in top of code

			"When clicking on plot, ",
			popup_menu(-name=> "zoomlevel",
						-values => [ (2..8) ],
						-labels => { map { $_ => "Zoom ${_}x" } (2..8) },
						-default => $zoomlevel,
						-onChange => "this.form.submit();",
						), " times"
				)
			),
		Tr(
			# Print submit button
			td({ -colspan=>6, -align=>"right"}, submit(-name=>"SUBMIT", -value=>"Submit"))
			),
		Tr(
			td({ -colspan=>6 }, hr)
			),
	),
	end_form;
}

# Subroutine Header
#
# Purpose           : _abort
# Comments          : 
# Input             : error message
# Output            : HTML page

sub _abort($)
{
	my ($message) = @_;

	my $embedded = $cgi->param('embedded')|| "true";

	if ("$embedded" eq "true") {
		print_header($maintitle, $subtitle);
		print p(font({ -color=>"red"}, b("ERROR: "),"$message")),
			print_form,
			print_remark;
		print_footer;
	}
	else {
		# return error message as GIF

		# have to make STDOUT unbuffered, to make the
		# 'Content-type' line appear before 'txt2gif' output
		select STDOUT;
		$| = 1;
		print "Content-type: image/gif\n\n";

		# FIXME , we should use GD instead...

		open(TXT2GIF, "|txt2gif");
		print TXT2GIF "ERROR: $message\n";
		warn "ERROR: $message";
		close(TXT2GIF);
	}
	exit(1);
}

# Subroutine Header
#
# Purpose           : print HTML and standard NCC header
# Comments          : 
# Input             : main title (heading1) and subtitle (heading2)
# Output            : HTML code on stdout

sub print_header($$)
{
	my ($maintitle,$subtitle) = @_;
	print header;
	if (_is_included()) {
		print "&nbsp;";
	}
	else {
		print "<!--#set var=\"title\" value=\"$maintitle: $subtitle\"-->",
		      "<!--#include virtual=\"/includes/ripencc/ttmhead1.inc\"-->",
#print "<HTML><HEAD><TITLE>AAP</TITLE></HEAD><Body>",
		      script({language=>"JavaScript1.2", type=>"text/javascript"}, "$validate_regex"),
		      script({language=>"JavaScript1.2", type=>"text/javascript"}, "$validate_range"),
		      script({language=>"JavaScript1.2", type=>"text/javascript"}, "$validate_equality"),
		      h1($maintitle), h2($subtitle);
	}
}

# Subroutine Header
#
# Purpose           : print remarks
# Comments          : 
# Input             : none
# Output            : HTML code on stdout

sub print_remark() 
{
	# closing remarks on this script

	return p(b("NOTE"),": plots are generated in real time;
		depending on the requested period, it can take 10 to 90 seconds
		to create the image. The daily generated (last day/week/month)
		plots are available from the", a({ -href=>"/ttm/Plots/"}, "main TTM results page"),
		"We also have a", a({ -href=> "/ttm/Plots/locations.cgi"},"world map"), "with locations of the boxes",
		),
	p("Data are available from 19990801 onwards, though not all
			present boxes had already been commissoned at that time.
			Today's data are collected the coming night and will be available
			for plots-on-demand starting approximately 08:00 UTC tomorrow morning"),
	p("To view PostScript format you need an appropriate
			application (e.g. ", a({ -href=>"http://www.cs.wisc.edu/~ghost/index.html"}, "ghostview"), 
			") installed on your computer and configured in your browser"),
	p("When selecting ASCII,
		you will create a table with data instead of a plot.
		This table can be imported into other applications.  The data consists of
		11 colomns, separated by commas.  The various fields are:"), blockquote(
		table(
			Tr(td("1 "),td(b("Src")),td("Identifier for the sending test-box")),
			Tr(td("2 "),td(b("Dst")),td("Identifier for the destination test-box")),
			Tr(td("3 "),td(b("Size")),td("Number of bytes in the packet")),
			Tr(td("4 "),td(b("Arrival")),td("Arrival time (unix time stamp).")),
			Tr(td("5 "),td(b("Delay")),td("Delay (ms)")),
			Tr(td("6 "),td(b("SrcClk")),td("Status of the sending clock")),
			Tr(td("7 "),td(b("DstClk")),td("Status of the destination clock")),
			Tr(td("8 "),td(b("Nhops")),td("Number of hops")),
			Tr(td("9 "),td(b("Route")),td("Identifier for the route the packet took")),
			Tr(td("10"),td(b("SrcErr")),td("Estimated error on the sending clock")),
			Tr(td("11"),td(b("DstErr")),td("Estimated error on the receiving clock")),
		)
	);
}

# Subroutine Header
#
# Purpose           : print footer
# Comments          : 
# Input             : none
# Output            : HTML code on stdout

sub print_footer() {
	print script({language=>"JavaScript", type=>"text/javascript", src=>"/projects/ttm/wz_tooltip.js"}, "");

	if (_is_included()) {
		print "&nbsp", end_html;
	}
	else {
		print blockquote("Data from the RIPE NCC TTM service, &copy;RIPE NCC, 2003, all rights reserved.", br, 
			"Data may not be redistributed without prior approval of the RIPE NCC.",br, 
			"Please see ", a({ -name=>"aup", -href=>"http://www.ripe.net/ripe/docs/ripe-300.html"},
			"http://www.ripe.net/ripe/docs/ripe-300.html"), "for details."),
		        "<!--#include virtual=\"/includes/ripencc/ttmfoot1.inc\"-->";
	}
}

sub _is_included()
{
	return (exists($ENV{'SERVER_PROTOCOL'}) && $ENV{'SERVER_PROTOCOL'} eq "INCLUDED" ) ? 1 : 0;
}

sub tboxsort()
{
	# Other sort options possible here which global parameter
	# Think like hopwise of geographicaly wise sorting

	($a =~ /tt(\d+)/)[0] <=> ($b =~ /tt(\d+)/)[0]
						||
	($a =~ /tt\d+([a-z]?)/)[0] cmp ($b =~ /tt\d+([a-z]?)/)[0]
}

sub createzoommap($$$$$$)
{
	my ($URL, $timestamp_start, $timestamp_end, $slices, $type, $anchor) = @_;
	my (%time_box, %delay_box, @areas);

	# Declare boxes to "superimpose"
	# Get these coordinates from e.g. with Gimp's web imagemap filter

	%time_box = (
		"delay" => [ 
						{ x => 46, y => 62, w => 304, h => 246 },	# top left
						{ x => 46, y => 376, w => 304, h => 246 },	# bottom left
						{ x => 434, y => 376, w => 304, h => 246 },	# bottom right
					],
		"jitter" => [
						{ x => 434, y => 63, w => 304, h => 246 },	# top right
						{ x => 434, y => 376, w => 304, h => 246 },	# bottom right
					], 
		"trends" => [
						{ x => 84, y => 53, w => 616, h => 166 },	# top
						{ x => 84, y => 266, w => 616, h => 166 },	# middle
						{ x => 84, y => 479, w => 616, h => 166 },	# bottom
					],
	);
	%delay_box = (
		"delay" => [
						{ x => 10, y => 62, w => 36, h => 246 },	# delay axis top left
						#{ x => 434, y => 310, w => 304, h => 25 },	# delay axis top right
						{ x => 434, y => 63, w => 304, h => 246 },	# top right
					],
		"jitter" => [
						# { x => 44, y => 624, w => 304, h => 25 },
						{ x => 46, y => 376, w => 304, h => 246 },	# bottom left
						{ x => 399, y => 376, w => 35, h => 246 },	# bottom right
					], 
		"trends" => [
						{ x => 20, y => 53, w => 64, h => 164 },
						{ x => 20, y => 266, w => 64, h => 164 },
						{ x => 20, y => 479, w => 64, h => 164 },
					],
	);

	@areas = ();

	# Create imagemaps for interactive zooming
	# Just use binary search here...

	if (exists $time_box{$type})
	{
		my ($slice, $timestamp_delta);

		$timestamp_delta = $timestamp_end - $timestamp_start;

		for($slice = 0 ; $slice < $slices; $slice++)
		{
			# Clone parameter list for this slice

			my $URLslice = new CGI($URL);
			my $URLlimitdelay = new CGI($URL);

			my $delayrange = $URL->param('maxdelay') - $URL->param('mindelay');

			$URLslice->param( -name => 'embedded', -value => 'true');

			# start of slice
			$URLslice->param( -name => 'date_start', -value => strftime("%Y%m%d", gmtime($timestamp_start + ($slice * ( $timestamp_delta / $slices)) )));
			$URLslice->param( -name => 'time_start', -value => strftime("%H:%M",   gmtime($timestamp_start + ($slice * ( $timestamp_delta / $slices)) )));

			# end of slice
			# XXX '-1' is under consideration as I do not for sure if this means "<" or "<=" for the ttplots server...

			$URLslice->param( -name => 'date_end', -value => strftime("%Y%m%d",   gmtime($timestamp_start + (($slice + 1) * ( $timestamp_delta / $slices)) -1 )));
			$URLslice->param( -name => 'time_end', -value => strftime("%H:%M",    gmtime($timestamp_start + (($slice + 1) * ( $timestamp_delta / $slices)) -1 )));

			# Limit delays here

			$URLlimitdelay->param( -name => 'embedded', -value => 'true');
			$URLlimitdelay->param( -name => 'mindelay', -value => sprintf("%0.3f", $URL->param('mindelay') + ( $slice * ( $delayrange / $ slices ) ) ));
			$URLlimitdelay->param( -name => 'maxdelay', -value => sprintf("%0.3f", $URL->param('mindelay') + ( ($slice + 1) * ( $delayrange / $slices ) ) ));

			# Create areas for each defined rectangle for the given plottype.
			# We assume the timescale is the same for all these rectangles...
				
			foreach my $time_box (@{$time_box{$type}})
			{
				my $message = "Click on image to show plot on demand ranging from " .
					strftime("%Y%m%d %H:%M", gmtime($timestamp_start + ($slice * ( $timestamp_delta / $slices)) )) .
					" till " .
					strftime("%Y%m%d %H:%M", gmtime($timestamp_start + (($slice + 1) * ( $timestamp_delta / $slices)) -1 ));

				push (@areas, Area({ -shape => "rect", 
						-coords => join(",",$time_box->{'x'} + ($slice * ceil($time_box->{'w'} / $slices)) , $time_box->{'y'}, $time_box->{'x'} + (($slice + 1) * ceil($time_box->{'w'} / $slices)) - 1, $time_box->{'y'} + $time_box->{'h'} ),
						-alt => $message, 
						-onmouseover=>maketooltip($message), 
						-href=>$URLslice->url(-absolute=>1, -query=>1 ) . "#$anchor" }));
			}

			# Create horizontal/vertical zoom for the min and max delay parameters

			foreach my $delay_box (@{$delay_box{$type}})
			{
				my $message = "Click here on this axis to limit the delay from " .
					$URLlimitdelay->param('mindelay') .
					" ms till " .
					$URLlimitdelay->param('maxdelay') .
					" ms";

				if ($delay_box->{'w'} < $delay_box->{'h'})
				{
					# Probably vertical axis, cope with it
					push (@areas, Area({ -shape => "rect", 
							-coords => join(",",$delay_box->{'x'}, ($delay_box->{'y'} + $delay_box->{'h'} ) - ($slice * ceil($delay_box->{'h'} / $slices)) , $delay_box->{'x'} + $delay_box->{'w'} , ( $delay_box->{'y'} + $delay_box->{'h'} ) - (($slice + 1) * ceil($delay_box->{'h'} / $slices)) - 1 ),
							-alt => $message, 
							-onmouseover=>maketooltip($message), 
							-href=>$URLlimitdelay->url(-absolute=>1, -query=>1 ) . "#$anchor" }));
				}
				else
				{
					# Horizontal axis
					push (@areas, Area({ -shape => "rect", 
							-coords => join(",",$delay_box->{'x'} + ($slice * ceil($delay_box->{'w'} / $slices)) , $delay_box->{'y'}, $delay_box->{'x'} + (($slice + 1) * ceil($delay_box->{'w'} / $slices)) - 1, $delay_box->{'y'} + $delay_box->{'h'} ),
							-alt => $message, 
							-onmouseover=>maketooltip($message), 
							-href=>$URLlimitdelay->url(-absolute=>1, -query=>1 ) . "#$anchor" }));
				}
			}

		}
		if ( $type eq "delay" )
		{	
			push(@areas,Area({ -shape => "rect", -coords => "11,33,387,339", -alt => "Delay & Number-of-Hops as a function of time", -onmouseover=>maketooltip($plotlegend{"delaytopleft"}), -target=>"legend", -href=>"http://www.ripe.net/ttm/General/tt_matrix.html#delaytopleft"}),
				Area({ -shape => "rect", -coords => "397,32,773,339", -alt => "Distribution of delays", -onmouseover=>maketooltip($plotlegend{"delaytopright"}), -target=>"legend", -href=>"http://www.ripe.net/ttm/General/tt_matrix.html#delaytopright"}),
				Area({ -shape => "rect", -coords => "9,345,388,652", -alt => "Packets sent", -onmouseover=>maketooltip($plotlegend{"delaybottomleft"}), -target=>"legend", -href=>"http://www.ripe.net/ttm/General/tt_matrix.html#delaybottomleft"}),
				Area({ -shape => "rect", -coords => "397,345,773,650", -alt => "Packet arrival as a function of time", -onmouseover=>maketooltip($plotlegend{"delaybottomright"}), -target=>"legend", -href=>"http://www.ripe.net/ttm/General/tt_matrix.html#delaybottomright"}),
			);
		}
		elsif ( $type eq "jitter" || $type eq "ipdv" )
		{
			push(@areas,Area({ -shape => "rect", -coords => "10,34,387,339", -alt => "Distribution of delay variations", -onmouseover=>maketooltip($plotlegend{"ipdvtopleft"}), -target=>"legend", -href=>"http://www.ripe.net/ttm/General/tt_matrix.html#ipdvtopleft"}),
				Area({ -shape => "rect", -coords => "399,33,776,339", -alt => "Delay variations vs. time", -onmouseover=>maketooltip($plotlegend{"ipdvtopright"}), -target=>"legend", -href=>"http://www.ripe.net/ttm/General/tt_matrix.html#ipdvtopright"}),
				Area({ -shape => "rect", -coords => "9,344,388,652", -alt => "One-way delay measurements: Packet delay distribution", -onmouseover=>maketooltip($plotlegend{"ipdvbottomleft"}), -target=>"legend", -href=>"http://www.ripe.net/ttm/General/tt_matrix.html#ipdvbottomleft"}),
				Area({ -shape => "rect", -coords => "398,344,776,651", -alt => "One-way delay measurements: Packet delay vs. time", -onmouseover=>maketooltip($plotlegend{"ipdvbottomright"}), -target=>"legend", -href=>"http://www.ripe.net/ttm/General/tt_matrix.html#ipdvbottomright"}),
			);
		}
		# FIXME Needs trendplot explanation
		#
		# elsif ( $type eq "trends" )
		# {
		# }

		return Map( { -name => "$type$timestamp_start$timestamp_end", -id => $type }, @areas, Area({ -shape => "default", -nohref => "nohref"}));
	}
	return ();
}

sub maketooltip($)
{
	return "this.T_STICKY=true;this.T_DELAY=1000;return escape('<ul>$_[0]<li>To <a href=\"#top\">top</a></li></ul>')"
}
