#!/usr/local/bin/perl -wT

# Copyright (c) 2000,2002                            RIPE NCC
#
# All Rights Reserved
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose and without fee is hereby granted,
# provided that the above copyright notice appear in all copies and that
# both that copyright notice and this permission notice appear in
# supporting documentation, and that the name of the author not be
# used in advertising or publicity pertaining to distribution of the
# software without specific, written prior permission.
#
# THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
# ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS; IN NO EVENT SHALL
# AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY
# DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
# AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

#------------------------------------------------------------------------------
# Module Header
# Filename          : cgiquery.pl
# Purpose           : Front-end to CGI query from TT DB
# Authors           : Manuel Valente, Rene Wilhelm  RIPE NCC
# Date              : 20000525
# Revised	    : 20020428, fill in defaults for start/end dates
# Revised	    : 20020604, add links to plots-on-demand
# Revised	    : 20030124, add IPv6 support, AS level traces
# Description       : CGI front end to query TTM traceroute Database
# Language Version  : Perl 5
# OSs Tested        : BSDI 3.1, Solaris8, Linux
# Command Line      : N/A
# Input Files       : Web form, TTM Traceroute Database, GPSINFO file
# Output Files      : Dynamic Web page
# External Programs : ttqueryd on a compute server
#
# $Id: routing.cgi,v 1.5 2004/02/20 15:28:27 ruben Exp $
#------------------------------------------------------------------------------

use 5.006;

use lib qw(/home/ttraffic/lib/perl5.6.0);

use strict;
use CGI 2.86 qw/-newstyle_urls -nosticky :standard :imagemap/;
use CGI::Carp;
use IO::Socket;
use POSIX;
use LWP::UserAgent;
use HTTP::Request::Common;
use Algorithm::Diff qw(diff sdiff LCS traverse_sequences traverse_balanced);
use Statistics::Descriptive;
use RIPE::TTM qw(:all);

$ENV{'PATH'} 		= "/bin/:/usr/bin:/ncc/ttpro/bin:/home/ttraffic/bin";

sub get_form($@);
sub get_result($@);
sub get_result_routingdiffs_by_time($$$$$$);		# XXX good candidate to get an object as argument vector
sub get_result_routingdiffs_by_frequency($$$$$$);
sub resolve_hop($$$$);
sub getname($$$);
sub ttquery_to_datastructures($);

my %getname_lookupcache = ();

# XXX ??? set errno to 1 ?
$! = 1;

my $client	        = $ENV{REMOTE_ADDR};
my $progname            = $ENV{'SCRIPT_NAME'};
$progname               =~ s|/+|/|g;            #remove double slashes

my $DBsources = "RIPE,ARIN,APNIC,RADB,CW,VERIO"; #search order for whois

my $domain = ".ripe.net";

my $getasname = "/usr/local/bin/whois -T aut-num -ar -s $DBsources AS\%s |grep descr";

my %sortmode = (
	date => "Date",
	occurances => "Occurrances",
	daterev => "Date (reverse)",
	occurancesrev => "Occurrances (reverse)",
);

# defeat random order %hash sorting when using this data for the popup_menu formfield

my @sortmode = qw(date occurances daterev occurancesrev);

my (%TTvals);

# Main function
MAIN: {
	my (%list,@BoxList);

	# Fill hashes with TestBox configuration data

	@BoxList = sort keys %RIPE::TTM::testboxes;
	%TTvals = map { $_ => $_ . ": " . $RIPE::TTM::gpsinfo{$_}{'longdesc'} } keys %RIPE::TTM::gpsinfo;

	#
	# Create queury, output form, and (when present) results
	#

	# Make usable through SSI and standalone operation
	my $cgi = exists($ENV{'REDIRECT_QUERY_STRING'}) ?
		new CGI($ENV{'REDIRECT_QUERY_STRING'}) :
		new CGI;

	if (!defined($cgi->param('ipv')))
	{
		$cgi->param('ipv',4);
		print redirect $cgi->url(-absolute=>1, -query=>1);
		exit(0);
	}

	my $subtitle = 'Query IPv4 Traceroute Database';
	if ($cgi->param('ipv') == 6) {
		$subtitle = 'Query IPv6 Traceroute Database';
	}

	$subtitle = "Hotlist $subtitle" if $#RIPE::TTM::hotlist > 0;

	# Check if it's the first time the script is called
	if (defined ($cgi->param('SRC')) && defined ($cgi->param('DST')))
	{
		$cgi->param('time_start') or $cgi->param(-name=>'time_start',-value=>'00:00');
		$cgi->param('time_end')   or $cgi->param(-name=>'time_end',  -value=>'00:00');

		$subtitle .=" \[".$cgi->param('SRC')." - ".$cgi->param('DST')."\]";
	};

	my ($header,$footer) = get_headers("Test Traffic Measurement","$subtitle");

	print header,$header,
			get_form($cgi,@BoxList),
			(defined ($cgi->param('SRC')) && defined ($cgi->param('DST')) ) ?
				get_result($cgi,@BoxList) :
				p("Click&nbsp;", a({-href=>"http://www.ripe.net/ttm/Help/traceroutedb.html"},"here"), "&nbsp;for an explanation regarding the Query Traceroute Database CGI."),
			$footer;
};

# Subroutine Header
# Purpose           : Send query, print results nicely formatted
# Comments          : Prints data to STDOUT
# Input             : List of parameters to pass the DB:
#                     SRC box, DST box, time start, time end
# Output            : NULL

sub get_result($@)
{
	my ($cgi,@boxlist) = @_;
	my ($URLpodjitter,$URLpoddelay, $QUERYboxoverview, $QUERYoverview, $QUERYsummary);
	my $URLswap = new CGI($cgi);

	my ($line,$result,@formatted_result);

	my ($routing_timestamps,$routing_vector,$aspath_vector, @prev_rvec_hops);

	my $trimmed_result = 0;

	my $ipv = $cgi->param('ipv')  || "4" ;	# IP version
	my $type = $cgi->param('type') || "IP";

	my $self_url = exists($ENV{'REDIRECT_SCRIPT_URL'}) ? 
					$ENV{'REDIRECT_SCRIPT_URL'} :
					$cgi->url(-relative=>1);

	$result = "";

	if  (($type eq "AS") && ($ipv == 6)) {
		return p( { -class => "error" }, "Traces in AS space not yet implemented for IPv6");
	}

	my $src = $cgi->param('SRC');
	my $dst = $cgi->param('DST');

	my $af = ($ipv == 6) ? "INET6" : "INET" ;

	$src eq $dst and do
	{
		return p({ -class => "error" }, "Error: Source and Destination IP are identical - $src.");
	};

	my $tstart = get_timestamp ($cgi->param('date_start'),$cgi->param('time_start'));

	$tstart == -1 and return();

	my $tend   = get_timestamp ($cgi->param('date_end'),$cgi->param('time_end'));

	$tend == -1 and return();

	my $resolve = $cgi->param('resolve');

	my $sortmode = $cgi->param('sortmode') || "date";

	if (!(grep {/$sortmode/} @sortmode))
	{
		return p({ -class => "error" }, "Error: unsupported sorting mode \"$sortmode\".");
	}

	my $sdiff = $cgi->param('sdiff') || "yes";

	if (!(grep {/$sdiff/} qw(yes no)))
	{
		return p({ -class => "error" }, "Error: Please use yes or no, not \"$sdiff\".");
	}

	# XXX strip of domain to match with allowed user groups (still needed ?)

	$src =~ s/(tt\d+[a-z]?)\.ripe\.net/$1/;
	$dst =~ s/(tt\d+[a-z]?)\.ripe\.net/$1/;

	if ($RIPE::TTM::use_chain_auth)
	{
		return p({ -class => "error" }, "$src -> $dst: access denied. source and destination must be in this chain.")	unless ( exists($RIPE::TTM::chainmap{$src}) && grep {/^${dst}$/} @{$RIPE::TTM::chainmap{$src}});
	}

	if ($RIPE::TTM::use_basic_auth)
	{
		# check access rights

		my $grouplist = " " . join(" ", @RIPE::TTM::usergroups) . " ";
		$grouplist =~ s/ \S+_admin//g;

		if ( $RIPE::TTM::remote_user ne "tipu" && ( ($grouplist !~  m/\s${src}[\.\s]/) &&
			 ($grouplist !~  m/\s${dst}[\.\s]/) )) {
				$grouplist =~ s/(\S+) (\S+) $/$1 or $2/;
				return p( { -class => "error" }, "$src -> $dst: access denied. soure or destination must be $grouplist.");
		}
	}

	if (!$RIPE::TTM::use_ttplots_http_proxy)
	{
		my $socket = IO::Socket::INET->new(PeerAddr => $RIPE::TTM::ttquery_host,
			PeerPort => $RIPE::TTM::ttquery_port,
			Proto    => "tcp",
			Type     => SOCK_STREAM)
		or die "Couldn't connect to $RIPE::TTM::ttquery_host: $@\n";

		# Build the command to run
		my $command = "$RIPE::TTM::gpsinfo{$src}{'fqdn'} $RIPE::TTM::gpsinfo{$dst}{'fqdn'} $tstart $tend $type $af $client\n";

		print $socket $command;

		while ($line = <$socket>)
		{
			$line eq ".\n" and last;
			$line =~ /^\.(\d+)$/ and do
			{
				$trimmed_result = $1;
				last;
			};
			$result .= $line;
		};

		close($socket);

		# XXX Check if we are used in passthrough mode and return immediatly if so
		# Yes this is an ugly hack, but we can do that because we haven't printed anything yet
		# We are in the middle of building up our output which is actually a list which is being printed out
		#

		if ($RIPE::TTM::remote_user eq "tipu")
		{
			# XXX print $trimmed_result

			select STDOUT;
			$| = 1;

			print header(-type=>'text/plain', -attachment=>"route.txt");
			print $result;
			exit(0);
		}
	}
	else
	{
		# connect to http://www.ripe.net/cgi-bin/ttm/router and request passthrough mode
		my $ua = LWP::UserAgent->new(env_proxy => 1,
									 keep_alive => 1,
									 timeout => 120,
									);

		my $URL = "http://www.ripe.net/cgi-bin/ttm/routing?" . $cgi->query_string();

		$ua->credentials("www.ripe.net:80", "Test Traffic", "tipu", "4T1poeOnl1");
		$ua->request(GET($URL), sub {
				my($data, $response, $protocol) = @_;

				$data =~ s/\r//gms;
				$result .= $data;
			}
		);
	}

	# The $result scalar holds all output from the backend ttquery program
	# It consists of a sequence of lines where each line has following
	# structure:
	#
	# <src host> <dst host> <start> <end> <id> <nhops> <nrecords> <hop> ...
	#
	# If the query is for IP traceroutes and AS path information is
	# available the line continues with <asid> <nAShops> <AShop> ...
	#

	if ($result eq "") {
		return p({ -class => "error" }, "No data found for these boxes and dates!",
				($ipv == 6) ? p("Current IPv6 configured boxes are listed ", a({ -href => "/plots/?ipv=6" }, "here")) : "&nbsp;")
	}


	# we got some data, prepare output:

	SWITCH: {

		$sortmode eq "date" && do
		{
			@formatted_result = get_result_routingdiffs_by_time($af, $type, $resolve, 0, $sdiff, $result);
			last SWITCH;
		};
		$sortmode eq "daterev" && do 
		{
			@formatted_result = get_result_routingdiffs_by_time($af, $type, $resolve, 1, $sdiff, $result);
			last SWITCH;
		};
		$sortmode eq "occurances" && do 
		{
			@formatted_result = get_result_routingdiffs_by_frequency($af, $type, $resolve, 0, $sdiff, $result);
			last SWITCH;
		};
		$sortmode eq "occurancesrev" && do 
		{
			@formatted_result = get_result_routingdiffs_by_frequency($af, $type, $resolve, 1, $sdiff, $result);
			last SWITCH;
		};
		die "unsupported sortmode $sortmode";
	}

	my $link = ($type eq "AS") ? "AS number " : "IP address " ;
	my $d_s = $cgi->param('date_start');
	my $t_s = $cgi->param('time_start');
	my $d_e = $cgi->param('date_end');
	my $t_e = $cgi->param('time_end');

	$URLpoddelay  = new	CGI("SRC=$src;DST=$dst;ipv=$ipv;date_start=$d_s;time_start=$t_s;date_end=$d_e;time_end=$t_e;plottype=delay");
	$URLpodjitter = new CGI("SRC=$src;DST=$dst;ipv=$ipv;date_start=$d_s;time_start=$t_s;date_end=$d_e;time_end=$t_e;plottype=jitter");
	$QUERYboxoverview = new CGI("ipv=$ipv;src=$src;dst=$dst");
	$QUERYoverview = new CGI("ipv=$ipv;node=$src;view=delay");
	$QUERYsummary = new CGI("ipv=$ipv;boxname=$src;period=day");

	$URLswap->param(-name => "SRC", -value => $dst);
	$URLswap->param(-name => "DST", -value => $src);

	return 	hr,table( { -width=>"100%"},
				Tr(td( { -valign=>"top", -align=>"center"},
					a({ -name => "result"}, ""),
					p({ -style => "font-size: smaller;" }, "Click&nbsp;", a({-href=>"http://www.ripe.net/ttm/Help/traceroutedb.html"},"here"), "&nbsp;for an explanation regarding the Query Traceroute Database CGI, or go to the&nbsp;",a({ -href => "#legend"},"legend"),"&nbsp;instead."),
					table( { -width=>"100%"}, Tr(
							td( { -align => "center" }, "[" . a({ -href => $URLswap->url(-absolute=>1, -query=>1) . "#result" }, "Swap Source and Destination") . "]" ),
							td( { -align => "center" }, "[" . join('/',map  { $ipv == $_ ? u("IPv$_") : a({ -href => "$self_url?" . cleanup_query_string("ipv", $_, $cgi) . "#result" }, "IPv$_" ) } (4,6)) . "]"),
					))
				),
				td( { -align => "right" },
					CGI::div( { -style=> "font-size: x-small;" },
						table( Tr(
							Tr(td({ -align => "right" }, a({ -href => "#top" }, "Adjust parameters..."))),
							Tr(td({ -align => "right" }, a({ -href => "${RIPE::TTM::overview_url}?" . $QUERYoverview->query_string() . "#$dst" }, "See delay plot overview for $src..."))),
							Tr(td({ -align => "right" }, a({ -href => "${RIPE::TTM::detail_url}?" . $QUERYboxoverview->query_string() }, "See all plots for $src and $dst..."))),
							Tr(td({ -align => "right" }, a({ -href => "${RIPE::TTM::summary_url}?" . $QUERYsummary->query_string() . "#$dst", -target=> "traceroutes"}, "See the summaries for $src..."))),
							Tr(td({ -align => "right" }, a({ -href => "${RIPE::TTM::pod_url}?" . $URLpoddelay->query_string() . "#plot" },"See delay plots on demand for this period..."))),
							Tr(td({ -align => "right" }, a({ -href => "${RIPE::TTM::pod_url}?" . $URLpodjitter->query_string() . "#plot" }, "See jitter plots on demand for this period..."))),
						))
					)
				))
			),
			@formatted_result, hr, p("Click on $link to consult whois database"), 
			p("When sorting on (reverse) occurance count, the total / mean number will give you a figure whether a route was stable in that timeframe: a high occurance count with low mean indicate many routeflaps. the most preferred route has the most time used."),
			$sdiff eq "yes" ?  CGI::div({ -style => "font-size: x-small" }, a({-name =>"legend"},"&nbsp;"), table( { -cellpadding => 5, -cellspacing => 0, -width => "40%"},
			Tr(th({ -class => "datatable" },"Legend for the colorized diff output")),
			Tr(td( { -class => "diffchanged" }, "This color combination shows these hops have changed since the previous entry")),
			Tr(td( { -class => "diffadded" }, "This color combination shows these hops where added since the previous entry")),
			Tr(td( { -class => "diffremoved" }, "This color combination shows these hops where removed since the previous entry. No hop number is printed but they are to be guaranteed in the same order as they where originally")),
			)) : "" ,br
	($ipv == 6) ? p(b("Note:")," ipv6 hostname lookups can take some time") : "",
		
	$trimmed_result  ?
		p({ -class => "warning" } , "Warning: Too many results were returned. The query only shows the first $trimmed_result results.") :
		"";
};

# Purpose           : Print the FORM for entering data
# Comments          :
# Input             : CGI object, List of boxes
# Output            : content (html data)
sub get_form($@)
{
	my ($cgi,@BoxList) = @_;

	my $content = '';

	# Set defaults

	my ($hour, $day, $mon, $year);
	my ($startday, $endday);
	my ($URLpodjitter,$URLpoddelay);
	my $URLswap = new CGI($cgi);

	($a, $a, $hour, $day, $mon, $year, $a, $a, $a) = gmtime(time());

	# yesterday's traceroute date usually processed by 0500 GMT

	if ($hour > 5) {
		$endday = strftime("%Y%m%d",gmtime());
		$startday = strftime("%Y%m%d",gmtime(time()-86400));
	}
	else {
		# last data until the end of day before yesterday
		$endday = strftime("%Y%m%d",gmtime(time()-86400));
		$startday = strftime("%Y%m%d",gmtime(time()-(2*86400)));
	}


	# Set the variables to empty string if they're not defined.
	my $src = $cgi->param('SRC')||$RIPE::TTM::TTLIST[0];
	my $dst = $cgi->param('DST')||  (grep {!/^${src}$/} @RIPE::TTM::TTLIST)[0];
	my $d_s = $cgi->param('date_start')||$startday;
	my $t_s = $cgi->param('time_start')||"00:00";
	my $d_e = $cgi->param('date_end')|| $endday;
	my $t_e = $cgi->param('time_end')|| "00:00";
	my $resolve = $cgi->param('resolve')|| "yes";
	my $type = $cgi->param('type')|| "IP";
	my $ipv = $cgi->param('ipv')|| "4";
	my $sortmode = $cgi->param('sortmode')|| "date";
	my $sdiff = $cgi->param('sdiff')|| "yes";

	# FORM

	$URLpoddelay  = new	CGI("SRC=$src&DST=$dst&ipv=$ipv&date_start=$d_s&time_start=$t_s&date_end=$d_e&time_end=$t_e&plottype=delay");
	$URLpodjitter = new CGI("SRC=$src&DST=$dst&ipv=$ipv&date_start=$d_s&time_start=$t_s&date_end=$d_e&time_end=$t_e&plottype=jitter");

	$URLswap->param(-name => "SRC", -value => $dst);
	$URLswap->param(-name => "DST", -value => $src);


	# XXX Interesting additions to consider:
	# DONE - Sorting by Route ID and Occurances
	# DONE - Show unique Route ID's only 
	# - Select routing vector for side by side diff 
	# - Color AS, but that bites with the diff I think...

	return start_form(-method=> "GET",
				-name=>"theform",
				-action=>$cgi->url(-relative=>1) . "#result",
				),hr,
		defined($cgi->param('nohotlist')) ? hidden(-name => 'nohotlist', -value => 'true') : "",
		table({ -border=>0, -cellspacing=>5 },
			# Tr(td({ -colspan=>6 }, hr)),

			# Print all testboxes for SRC combo
			Tr(
			td("SRC:"),
			td( {-colspan=>5, -align =>"right" },
				popup_menu(-name=> "SRC",
							-values=> \@RIPE::TTM::TTLIST,
							-labels => \%TTvals,
							-default => $src,
							-onChange => "return validate_equality(this.options[this.selectedIndex].value, this.form.DST.options[this.form.DST.selectedIndex].value, '!=', '$src', 'Source cannot be the same as the destination')"
							)
				)
			),

			# Print all testboxes for DST combo
			Tr(
			td("DST:"),
			td( {-colspan=>5, -align =>"right" },
				popup_menu(-name=> "DST",
							-values=> \@RIPE::TTM::TTLIST,
							-labels => \%TTvals,
							-default => $dst,
							-onChange => "return validate_equality(this.options[this.selectedIndex].value, this.form.SRC.options[this.form.SRC.selectedIndex].value, '!=', '$dst', 'Destination cannot be the same as the source')"
							)
				)
			),

			# Print SRC Input fields

			Tr(
				td("Start Date"),td(textfield( -name=>"date_start", -size=>10, -onChange=> "return validate_regex(this,'^[0-9]{8}\$','$d_s','Please enter a date in the format YYYYMMDD')", -default=>$d_s)),
				td("Start Time"),td(textfield( -name=>"time_start", -size=>5, -onChange=> "return validate_regex(this,'^[0-2][0-9]:[0-5][0-9](:[0-5][0-9])?\$','$t_s','Please enter a time in the format HH:MM:SS')", -default=>$t_s),
					a({ -href=>"http://www.worldtimezone.com/time-world.htm"}, "UTC")
					),
				td({ -colspan => 2},
					"Resolve IPs &nbsp;&nbsp;",
					radio_group(-name => "resolve",
								-values => [ "yes", "no" ],
								-default => $resolve),
					)
				),

			# Print DST Input fields

			Tr(
				td("End Date"),td(textfield( -name=>"date_end", -size=>10, -onChange=> "return validate_regex(this,'^[0-9]{8}\$','$d_e','Please enter a date in the format YYYYMMDD')", -default=>$d_e)),
				td("End Time"),td(textfield( -name=>"time_end", -size=>5, -onChange=> "return validate_regex(this,'^[0-2][0-9]:[0-5][0-9](:[0-5][0-9])?\$','$t_e','Please enter a time in the format HH:MM')", -default=>$t_e),
					a({ -href=>"http://www.worldtimezone.com/time-world.htm"}, "UTC")
					),
				td({-colspan=>2, -align =>"right" }, "Sort&nbsp;on&nbsp;",
					popup_menu(-name=> "sortmode",
								-values=> \@sortmode,
								-labels => \%sortmode,
								-default => $sortmode,
								)
				)
			),
			Tr(
				td( { -colspan => 2},
					"Traces in&nbsp;&nbsp;",
					radio_group(-name => "type",
								-values => [ "IP", "AS" ],
								-default => $type),
					),
				td( { -colspan => 2},
					"For&nbsp;",
					radio_group(-name => "ipv",
								-values => [ "4", "6" ],
								-labels => { "4" => "IPv4" , "6" => "IPv6" },
								-default => $ipv),
								"&nbsp;test traffic"
					),
				td( { -colspan => 2},
					"Show&nbsp;diffs&nbsp;",
					radio_group(-name => "sdiff",
								-values => [ "yes", "no" ],
								-default => $sdiff)
					),
				# Print submit button
				),
			Tr(
				td({ -colspan=>6, -align=>"right"}, submit(-name=>"SUBMIT", -value=>"Submit"))
				
			)
		),
	end_form,
		$#RIPE::TTM::hotlist > 0 ? 
				p(a({ -href => $cgi->url(-absolute=>1, -query=>1) . ";nohotlist"  }, "turn off"), " your hotlist temporarily or ", 
				a({ -href => "${RIPE::TTM::hotlist_url}?ipv=$ipv" }, "edit"), " the one that is currently active") 
			: defined($cgi->param('nohotlist')) ? 
				p(a({ -href => $cgi->url(-relative=>1) . "?" . cleanup_query_string("nohotlist", undef, $cgi) }, "turn on"), " temporarily disabled hotlist...") 
				: p(a({ -href => "${RIPE::TTM::hotlist_url}?ipv=$ipv" }, "create a hotlist"), " to discard plots you are not interested in...");
};

# Subroutine header
#
# Purpose: output HTML code for the present IP/AS hop
#

sub resolve_hop($$$$) {
	my ($hop, $type, $af, $resolve) = @_;

	my (@nums) = split('/', $hop);	# we only expect this for ASnums
	my (@resolve) = ();
	my $prefix = $type;

	$prefix = "" if ($type eq "IP");

	if ($resolve ne "no") {
		@resolve = (
			span({ -style => "font-size: smaller" },
				join("/",(map {
					getname($_, $type, $af)
		} @nums))));
	}

	return (
		join("/",(map {
			a( { -target => "resolve", -href => /^3ffe:/ ? "http://whois.6bone.net/cgi-bin/whois?$_" : "http://www.ripe.net/perl/whois?-ar\%20$prefix$_" },
				span({ -style => "font-size: smaller" },$_)
				)
		} @nums))
	), @resolve;
}

# Subroutine header
# Purpose: lookup host or ASname

sub getname($$$) {
	my ($num, $type, $af) = @_;

	my $ll;

	confess "cannot resolve empty/undefined value" unless (defined($num) && $num);

	# Untaint
	($num) = $num =~ /^([\d\.:a-fA-F]+)$/;

	if (exists($getname_lookupcache{$num}))
	{
		# We cache the value here. %getname_lookupcache is globally defined

		$ll = $getname_lookupcache{$num};
	}
	else
	{
		eval {
				local $SIG{ALRM} = sub {die "alarm\n"};
			if ($af eq "INET6") {
				alarm 12;
			}
			else {
					alarm 1;  # timeout after 1sec
			}
			if ($type eq "AS") {
				my $cmd = sprintf($getasname, $num);
				open(GETASNAME, "$cmd |") || die "Couldn't do AS lookup: $!";
				while (defined( my $line = <GETASNAME>)) {
					$line =~ /^descr:\s+(\w.*)$/;
					$ll = $1;
					last if ($ll);
				}
				close(GETASNAME);
			}
			else {
				if ($af eq "INET") {
						$ll = gethostbyaddr(inet_aton($num), AF_INET) ||
					 em("Hostname not found");
				}
				elsif ($af eq "INET6") {
					$ll = `/home/ttraffic/bin/aaaa2name $num 2>&1`;
					$ll = em("Hostname not found") if ($? != 0) ;
				}
			}
				alarm 0;
			die "OK";
			};
		if ($@ eq "alarm\n") {
			$ll = em("Host lookup timed out");
		}
		$getname_lookupcache{$num} = $ll;
	}
	return($ll);
}

sub ttquery_to_datastructures($)
{
	my ($result) = @_;


	# XXX We could tap into the plots-on-demand ascii output to obtain (average) delays during that
	# timeframe

	my (%routing_timestamps, %routing_vector, %aspath_vector);

	foreach my $rvec  (split("\n",$result))
	{
		my @line = split(/\s+/,$rvec);
		my $nhops = $line[5];

		last if $#line == 0; # . aka non truncated end of output seen

		$routing_timestamps{$line[2]}{'duration'} = $line[3] - $line[2];
		$routing_timestamps{$line[2]}{'routeid'} = $line[4];
		$routing_timestamps{$line[2]}{'nrecords'} = $line[6];
		$routing_timestamps{$line[2]}{'aspathid'} = $line[7+$nhops];

		if(!exists($routing_vector{$line[4]}))
		{
			$routing_vector{$line[4]}{'routeid'} = $line[4];
			$routing_vector{$line[4]}{'aspathid'} = $line[7+$nhops];

			$routing_vector{$line[4]}{'route'} = [ @line[7 .. (7+$nhops-1)] ];

			$routing_vector{$line[4]}{'dates'} = [ ( $line[2] ) ];
		}
		else
		{
			push(@{$routing_vector{$line[4]}{'dates'}},$line[2]);
		}

		if ($#line > 7+$nhops)
		{
			if(!exists($aspath_vector{$line[7+$nhops]}))
			{
				$aspath_vector{$line[7+$nhops]}{'routeid'} = $line[7+$nhops];
				$aspath_vector{$line[7+$nhops]}{'aspathid'} = $line[7+$nhops];

				$aspath_vector{$line[7+$nhops]}{'aspath'} = [ @line[9+$nhops .. $#line ] ];

				$aspath_vector{$line[7+$nhops]}{'dates'} = [ ( $line[2] ) ];

			}
			else
			{
				push(@{$aspath_vector{$line[7+$nhops]}{'dates'}},$line[2]);
			}
		}
	}
	return wantarray ? ( \%routing_timestamps, \%routing_vector, \%aspath_vector) :
					{
						"timestamps" => \%routing_timestamps,
						"routes"	=> \%routing_vector,
						"aspaths"	=> \%aspath_vector
					};
}

sub get_result_routingdiffs_by_time($$$$$$)
{
	my ($af, $type, $resolve, $order_reverse,$sdiff, $result) = @_;
	my ($routing_timestamps,$routing_vector,$aspath_vector) = ttquery_to_datastructures($result);
	my ($c, @prev_rvec_hops, @displayorder);

	@displayorder = sort keys %{$routing_timestamps};
			
	@displayorder = reverse @displayorder if ($order_reverse);

	return table({ -border => 0, -cellpadding => 5, -cellspacing => 0, -width => 800},
		( map {

			# Create some shortcuts from the datastructures, save some strain from the eyes
			my (@rvec_hops, @asvec_hops, %ipas_map,%ipresolve_map, %asresolve_map, @theRows,@diffs);
			my ($timestamp_start,$timestamp_stop) = ($_, $_+$routing_timestamps->{$_}{'duration'});

			# Can also contain AS path information if $type eq AS
			@rvec_hops  = @{$routing_vector->{$routing_timestamps->{$timestamp_start}{'routeid'}}{'route'}};

			if ($type ne "AS")
			{
				@asvec_hops = @{$aspath_vector->{$routing_timestamps->{$timestamp_start}{'aspathid'}}{'aspath'}};
				%ipas_map = map { $rvec_hops[$_] => $asvec_hops[$_] } (0 .. $#rvec_hops);
			}

			my $nhops = $#rvec_hops + 1;

			my @target_period = (
				td({ -align => "right" },
					table(
						Tr(
							td({ -align => "right", -style => "white-space: nowrap;"},
								span({ -style => "font-size: smaller" },"From&nbsp;", strftime("%Y/%m/%d %H:%M:%S", gmtime($timestamp_start)))
								)
						),
						Tr(
							td({ -align => "right", -style => "white-space: nowrap;"},
								span({ -style => "font-size: smaller" },"To&nbsp;", strftime("%Y/%m/%d %H:%M:%S", gmtime($timestamp_stop)))
								)
						)
					)
				),
				td({ -align => "right" },span({ -style => "font-size: smaller" },$routing_timestamps->{$timestamp_start}{'routeid'})),
				td({ -align => "right" },span({ -style => "font-size: smaller" },$routing_timestamps->{$timestamp_start}{'nrecords'}))
			);

			$c = -1;

			# XXX fill @prev_rvec_hops with rvec_hops the first time so we have an zero diff
			# This is for uniformity in the code

			@prev_rvec_hops = @rvec_hops unless $#prev_rvec_hops > 0;
			@diffs = sdiff(\@prev_rvec_hops,\@rvec_hops);

			@theRows = map
				{
					my ($currenthop);
					my ($css_class) = "diffnone";

					$_->[0] =~ /[uc+]/ && $c++; # Only advance on [uc+] states in diff

					$css_class = "diffchanged" if $_->[0] eq 'c' ;
					$css_class = "diffadded" if $_->[0] eq '+' ;
					$css_class = "diffremoved" if $_->[0] eq '-' ;
					$css_class = "diffnone" if $sdiff eq "no";

					$currenthop = $rvec_hops[$c];

					if ($css_class eq "diffremoved")
					{
						$currenthop = $_->[1];
					}

					# XXX Clean output is: skip table row on $css_class eq "diffremoved", and set $css_class to "normal"

					($sdiff eq "no" && $css_class eq "diffremoved") ? 
					"" :
					Tr(	{-class => $css_class},
						($c == 0 && $css_class ne "diffremoved") ? @target_period : td({ -class => $css_class,  -colspan => 3 }, "&nbsp;"),
						td({ -class => $css_class,  -align => "right" },span({ -style => "font-size: smaller" },($css_class eq "diffremoved") ? "(Removed)" : $c+1)),
							(($currenthop ne "255.255.255.255") &&
								($currenthop ne "0::0") &&
								($currenthop ne "0.0.0.0") &&
								($currenthop ne "0")) ? (
									(map { td({ -class => $css_class,  -align => "right" },$_) } resolve_hop($currenthop, $type, $af, $resolve)),
									($type eq "IP") ?
										exists($ipas_map{$currenthop}) ? td({ -class => $css_class,  -align => "right" },$ipas_map{$currenthop}) : td({ -class => $css_class, -align => "right" },"&nbsp;-&nbsp;") :
										""
								) : (
									td({ -class => $css_class,  -align => "right" },em("unknown")),
									($resolve ne "no" && ($type eq "IP")) ?
										(td({ -class => $css_class,  -align => "right" },em("&nbsp;no response")), td("&nbsp;")) :
										td("&nbsp;")
						)
					)
				} @diffs;
			@prev_rvec_hops = @rvec_hops;

			Tr(
				th({ -class => "datatable" },span({ -style => "font-size: smaller" },"Dates")),
				($type eq "AS") ? (
					th({ -class => "datatable" },span({ -style => "font-size: smaller" },"ASpath Id")),
					th({ -class => "datatable" },span({ -style => "font-size: smaller" },"Occurrances")),
					th({ -class => "datatable" },span({ -style => "font-size: smaller" },"Hop")),
					th({ -class => "datatable" },"AS-num(s)"),
					($resolve ne "no") ? th({ -class => "datatable" },span({ -style => "font-size: smaller" },"AS-name")) : "",
				) : (
					th({ -class => "datatable" },span({ -style => "font-size: smaller" },"Route Id")),
					th({ -class => "datatable" },span({ -style => "font-size: smaller" },"Occurrances")),
					th({ -class => "datatable" },span({ -style => "font-size: smaller" },"Hop")),
					th({ -class => "datatable" },span({ -style => "font-size: smaller" },"IP Address")),
					($resolve ne "no") ? th({ -class => "datatable" },span({ -style => "font-size: smaller" },"Host name")) : "",
				),
				# non-zero hops in AS path
				($#asvec_hops > -1) ? th({ -class => "datatable" },span({ -style => "font-size: smaller" },"AS&nbsp;Num(s)")) : ""
			), @theRows

		} ( @displayorder ) )
	),
}

sub get_result_routingdiffs_by_frequency($$$$$$)
{
	my ($af, $type, $resolve, $order_reverse, $sdiff, $result) = @_;
	my ($routing_timestamps,$routing_vector,$aspath_vector) = ttquery_to_datastructures($result);
	my ($c, @prev_rvec_hops, %pathfreq, $freqmode, @displayorder);

	# XXX tunable
	# $freqmode = 'duration';
	$freqmode = 'nrecords';

	foreach my $timestamp (keys %{$routing_timestamps})
	{
		if (!exists($pathfreq{$routing_timestamps->{$timestamp}{'routeid'}}))
		{
			$pathfreq{$routing_timestamps->{$timestamp}{'routeid'}} = [$routing_timestamps->{$timestamp}{$freqmode}];
		}
		else
		{
			push(@{$pathfreq{$routing_timestamps->{$timestamp}{'routeid'}}}, $routing_timestamps->{$timestamp}{$freqmode});
		}
	}

	# We can now do a distribution analysis on %pathfreq on a per route/ASid basis

	# XXX finetuning here please

	@displayorder = sort 
	{
		my ($stat_a,$stat_b);
		$stat_a = Statistics::Descriptive::Full->new();
		$stat_b = Statistics::Descriptive::Full->new();

		$stat_a->add_data(@{$pathfreq{$a}});
		$stat_b->add_data(@{$pathfreq{$b}});

		$stat_b->sum() <=> $stat_a->sum()			# total occurances of this path for this period
						||
		$stat_a->count() <=> $stat_b->count()	# total count periods. many countings indicate instability

	} keys %pathfreq;

	@displayorder = reverse @displayorder if ($order_reverse);

	return table({ -border => 0, -cellpadding => 5, -cellspacing => 0, -width => 800},
		( map {

			# Create some shortcuts from the datastructures, save some strain from the eyes
			my (@rvec_hops, @asvec_hops, %ipas_map,%ipresolve_map, %asresolve_map, @theRows,@diffs, $duration, @target_period);

			# Can also contain AS path information if $type eq AS
			@rvec_hops  = @{$routing_vector->{$_}{'route'}};

			if ($type ne "AS")
			{
				@asvec_hops = @{$aspath_vector->{$routing_vector->{$_}{'aspathid'}}{'aspath'}};
				%ipas_map = map { $rvec_hops[$_] => $asvec_hops[$_] } (0 .. $#rvec_hops);
			}

			my $nhops = $#rvec_hops + 1;
			my $stat = Statistics::Descriptive::Full->new();
			$stat->add_data(@{$pathfreq{$_}});

			$c = -1;
			$duration = 0;

			foreach my $timestamp_start (@{$routing_vector->{$_}{'dates'}})
			{
				foreach my $timestamp_duration ($routing_timestamps->{$timestamp_start}{'duration'})
				{
					$duration+=$timestamp_duration;
				}
			}
			$duration = sprintf("%d&nbsp;wk(s), %d&nbsp;days, %d&nbsp;hr(s), %d&nbsp;min(s), %d&nbsp;sec(s)",
				int ($duration / (7 * (24 * 3600))),
				int ( ($duration % (7 * (24 * 3600))) / (24 * 3600)  ),
				int ( ($duration % (24 * 3600)) / 3600),
				int ( ($duration % 3600) / 60),
				int (  $duration % 60 )
			);
		
			@target_period = (
				td({ -align => "center" },span({ -style => "font-size: smaller" },$duration)),
				td({ -align => "center" },span({ -style => "font-size: smaller" },$_)),
				td({ -align => "center" },span({ -style => "font-size: smaller" }, join('&nbsp;/&nbsp;',map {sprintf ("%.2f",$_) } ( $stat->sum(),$stat->mean() ))))
			);

			# XXX fill @prev_rvec_hops with rvec_hops the first time so we have an zero diff
			# This is for uniformity in the code

			@prev_rvec_hops = @rvec_hops unless $#prev_rvec_hops > 0;
			@diffs = sdiff(\@prev_rvec_hops,\@rvec_hops);

			@theRows = map
				{
					my ($currenthop);
					my ($css_class) = "diffnone";

					$_->[0] =~ /[uc+]/ && $c++; # Only advance on [uc+] states in diff

					$css_class = "diffchanged" if $_->[0] eq 'c' ;
					$css_class = "diffadded" if $_->[0] eq '+' ;
					$css_class = "diffremoved" if $_->[0] eq '-' ;
					$css_class = "diffnone" if $sdiff eq "no";

					$currenthop = $rvec_hops[$c];

					if ($css_class eq "diffremoved")
					{
						$currenthop = $_->[1];
					}

					($sdiff eq "no" && $css_class eq "diffremoved") ? 
					"" :
					Tr(	{-class => $css_class},
						($c == 0 && $css_class ne "diffremoved") ?
									@target_period : 
									td({ -class => $css_class,  -colspan => 3 }, "&nbsp;"),
						td({ -class => $css_class,  -align => "right" },span({ -style => "font-size: smaller" },($css_class eq "diffremoved") ? "(Removed)" : $c+1)),
							(($currenthop ne "255.255.255.255") &&
								($currenthop ne "0::0") &&
								($currenthop ne "0.0.0.0") &&
								($currenthop ne "0")) ? (
									(map { td({ -class => $css_class,  -align => "right" },$_) } resolve_hop($currenthop, $type, $af, $resolve)),
									($type eq "IP") ?
										exists($ipas_map{$currenthop}) ? td({ -class => $css_class,  -align => "right" },$ipas_map{$currenthop}) : td({ -class => $css_class, -align => "right" },"&nbsp;-&nbsp;") :
										""
								) : (
									td({ -class => $css_class,  -align => "right" },em("unknown")),
									($resolve ne "no" && ($type eq "IP")) ?
										(td({ -class => $css_class,  -align => "right" },em("&nbsp;no response")), td("&nbsp;")) :
										td("&nbsp;")
						)
					)
				} @diffs;

			@prev_rvec_hops = @rvec_hops;

			Tr(
				th({ -class => "datatable" },span({ -style => "font-size: smaller" },"Total Time used")),
				($type eq "AS") ? (
					th({ -class => "datatable" },span({ -style => "font-size: smaller" },"ASpath Id")),
					th({ -class => "datatable" },span({ -style => "font-size: smaller" },"Occurrances (&nbsp;Total&nbsp;/&nbsp;Mean&nbsp;)")),
					th({ -class => "datatable" },span({ -style => "font-size: smaller" },"Hop")),
					th({ -class => "datatable" },"AS-num(s)"),
					($resolve ne "no") ? th({ -class => "datatable" },span({ -style => "font-size: smaller" },"AS-name")) : "",
				) : (
					th({ -class => "datatable" },span({ -style => "font-size: smaller" },"Route Id")),
					th({ -class => "datatable" },span({ -style => "font-size: smaller" },"Occurrances (&nbsp;Total&nbsp;/&nbsp;Mean&nbsp;)")),
					th({ -class => "datatable" },span({ -style => "font-size: smaller" },"Hop")),
					th({ -class => "datatable" },span({ -style => "font-size: smaller" },"IP Address")),
					($resolve ne "no") ? th({ -class => "datatable" },span({ -style => "font-size: smaller" },"Host name")) : "",
				),
				# non-zero hops in AS path
				($#asvec_hops > -1) ? th({ -class => "datatable" },span({ -style => "font-size: smaller" },"AS&nbsp;Num(s)")) : ""
			), @theRows

		} ( @displayorder ) )
	),
}
