/*
 * process_dag.c
 *
 * Reads a dagsnap file and calculates the offset between the timestamp in the payload
 * and the DAG timestamp.
 *
 * Supported formats: 
 * *) The sender tool included in this package
 * *) The TTM send_packet format
 *
 * Author: Thomas Wana <tom@ripe.net>
 *
 * $Id: process_dag.c,v 1.15 2005/03/17 11:51:49 tom Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/time.h>
#include <sys/timex.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <net/ethernet.h>
#include <netinet/in_systm.h>
#include <sys/param.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include "dagapi.h"

#ifndef _POSIX_MAPPED_FILES
#error Unable to mmap files; exiting
#endif

#define ADDRARGS(addr) (addr)&0xFF,((addr)>>8)&0xFF,((addr)>>16)&0xFF,((addr)>>24)&0xFF

// for FORMAT_SENDER
struct outpkt
{
        /*
         * The structure returned by ntp_gettime
         */
        struct ntptimeval ntv;

        /*
         * A flag that determines if the fields in ntv have micro- or nanosecond resolution
         */
        struct timex tx;
} __attribute__ ((packed));

struct outpkt_both
{
	/*
	 * Two ntptimeval structures to hold both the userland and
	 * the kernel timestamp
	 */
	struct ntptimeval ntv_user;
	struct ntptimeval ntv_kernel;

	/* 
	 * The structure returned by ntp_adjtime
	 */
	struct timex tx;
} __attribute__ ((packed));

// for FORMAT_SEND_PACKET
struct test_packet_header
{
	long                packet_id;  /* needed for later coupling */
	ushort              offset_to_data; /* that is where the data goes * offset
					     * relative to the end of the * packet
					     * header */

};
/* stuct for syscal getppsval. For interaction with time and pps related paramenters. */
struct pps_val
{
	struct timeval      current_time;   /* Current timeofday as reported by
					     * the (NTP driven kernel clock) */
	struct timeval      last_pps;   /* time of the last PPS as set by *
					 * hardpps */
	int                 pps_status; /* clock status bits (as defined in
					 * timex.h) Most imortant is that 0x0100 is
					 * set. PPSSIGNAL pps signal present. */
	long                pll_offset; /* NTPs idea of the offset to the real
					 * time. */
	long                maxerror;   /* PLL error estimates */
	long                esterror;

};
struct test_packet_data
{
	/* data byte is in network order... */
	ushort              version;
	long                fingerprint;

	/* Note that fingerprint is not NULL terminated */
	struct pps_val      timepars;   /* time parameters see getpps.h */
	long                serial_number;
};

enum { FORMAT_SENDER, FORMAT_SEND_PACKET, FORMAT_SENDER_KERNEL, FORMAT_SENDER_BOTH } format;

/*
 * Converts the unsigned long long timestamp from the DAG card to seconds
 * and nanoseconds.
 * From the DAG Extensible Record Format doc, Page 10, "Timestamps" (adapted
 * to support nanoseconds)
 */
void ts_to_s_ns(unsigned long long ts, unsigned long *s, unsigned long *ns)
{
	*s=(ts>>32)&0xFFFFFFFFULL;
	ts=((ts&0xFFFFFFFFULL)*1000*1000*1000);
	ts+=(ts & 0x80000000ULL)<<1;
	*ns=ts>>32;
	if(*ns >= 1000000000)
	{
		*ns-=1000000000;
		*s+=1;
	}
}

/*
 * Calculates tdiff = t1 - t2.
 * The second part is always positive, the returned integer determines wether the
 * complete result is positive (0) or negative (1). This is because tv_sec can't
 * be negative when it's 0.
 */
int timespec_diff(struct timespec *tdiff, struct timespec *t1, struct timespec *t2)
{
	int negative=0;
	tdiff->tv_sec=t1->tv_sec-t2->tv_sec;
	tdiff->tv_nsec=t1->tv_nsec-t2->tv_nsec;

	if(tdiff->tv_nsec > 1000000000)
	{
		tdiff->tv_nsec-=1000000000;
		tdiff->tv_sec++;
		negative=0;
	}
	if(tdiff->tv_nsec < 0)
	{
		if(tdiff->tv_sec>0)
		{
			tdiff->tv_nsec+=1000000000;
			tdiff->tv_sec--;
			negative=0;
		}
		else if(tdiff->tv_sec==0)
		{
			tdiff->tv_nsec*=-1;
			negative=1;
		}
		else
		{
			tdiff->tv_nsec*=-1;
			tdiff->tv_sec*=-1;
			negative=1;
		}
	}
	return negative;
}

/*
 * Converts the unsigned long long timestamp from the DAG card to seconds
 * and microseconds.
 * From the DAG Extensible Record Format doc, Page 10, "Timestamps" 
 */
void ts_to_s_us(unsigned long long ts, unsigned long *s, unsigned long *us)
{
	*s=(ts>>32)&0xFFFFFFFFULL;
	ts=((ts&0xFFFFFFFFULL)*1000*1000);
	ts+=(ts & 0x80000000ULL)<<1;
	*us=ts>>32;
	if(*us >= 1000000)
	{
		*us-=1000000;
		*s+=1;
	}
}

void dumppacket(dag_record_t *pkt)
{
	const char *types[]={"Legacy", "PoS w/HDLC framing", "Ethernet", "ATM Cell", 
			     "reassembled AAL5 frame", "Multi-channel HDLC frame",
			     "Multi-channel Raw link data", "Multi-channel ATM Cell"};
	unsigned long s, ns;
	struct ether_header *ehdr;
	struct ip *iph;

	ts_to_s_ns(pkt->ts, &s, &ns);

	fprintf(stderr, "INFO: new packet record\n");
	fprintf(stderr, "INFO: time:           %ld s, %ld ns\n", s, ns);
	fprintf(stderr, "INFO: packet len:     %d\n", ntohs(pkt->wlen));
	fprintf(stderr, "INFO: type:           %s\n", types[pkt->type]);
	fprintf(stderr, "INFO: flags.iface:    %d\n", pkt->flags.iface);
	fprintf(stderr, "INFO: flags.vlen:     %d\n", pkt->flags.vlen);
	fprintf(stderr, "INFO: flags.trunc:    %d\n", pkt->flags.trunc);
	fprintf(stderr, "INFO: flags.rxerror:  %d\n", pkt->flags.rxerror);
	fprintf(stderr, "INFO: flags.dserror:  %d\n", pkt->flags.dserror);

	ehdr=(struct ether_header *)((char *)pkt+18);
	fprintf(stderr, "INFO:    ether.dhost: %02x:%02x:%02x:%02x:%02x:%02x\n",
			ehdr->ether_dhost[0],
			ehdr->ether_dhost[1],
			ehdr->ether_dhost[2],
			ehdr->ether_dhost[3],
			ehdr->ether_dhost[4],
			ehdr->ether_dhost[5]);
	fprintf(stderr, "INFO:    ether.shost: %02x:%02x:%02x:%02x:%02x:%02x\n",
			ehdr->ether_shost[0],
			ehdr->ether_shost[1],
			ehdr->ether_shost[2],
			ehdr->ether_shost[3],
			ehdr->ether_shost[4],
			ehdr->ether_shost[5]);
	fprintf(stderr, "INFO:    ether.type:  %d\n", ntohs(ehdr->ether_type));
	if(ntohs(ehdr->ether_type)!=ETHERTYPE_IP)
	{
		fprintf(stderr, "WARNING: no IP packet, ignoring\n");
		fprintf(stderr, "INFO: --------------------\n");
		return;
	}
	iph=(struct ip *)((char *)ehdr+sizeof(struct ether_header));
	fprintf(stderr, "INFO:    ip.proto:    %d\n", iph->ip_p);
	fprintf(stderr, "INFO:    ip.saddr:    %d.%d.%d.%d\n", ADDRARGS(iph->ip_src.s_addr));
	fprintf(stderr, "INFO:    ip.daddr:    %d.%d.%d.%d\n", ADDRARGS(iph->ip_dst.s_addr));

	fprintf(stderr, "INFO: --------------------\n");
}

int main(int argc, char **argv)
{
	dag_record_t *pkt;
	int fd;
	void *mmapptr, *ptr;
	struct stat st;
	long length;

	if(argc!=3)
	{
		fprintf(stderr, "Usage: %s format filename\n\n", argv[0]);
		fprintf(stderr, "format can be: sender | sender_kernel | send_packet | sender_both (see process_dag.c for details)\n\n");
		return -1;
	}

	if(!strcmp(argv[1], "sender"))
	{
		format=FORMAT_SENDER;
	}
	else if(!strcmp(argv[1], "send_packet"))
	{
		format=FORMAT_SEND_PACKET;
	}
	else if(!strcmp(argv[1], "sender_kernel"))
	{
		format=FORMAT_SENDER_KERNEL;
	}
	else if(!strcmp(argv[1], "sender_both"))
	{
		format=FORMAT_SENDER_BOTH;
	}
	else
	{
		fprintf(stderr, "Unsupported format '%s'\n", argv[1]);
		return -1;
	}

	if(stat(argv[2], &st)<0)
	{
		perror("stat");
		return -2;
	}
	length=(long)st.st_size;
	if((fd=open(argv[2], 0))<0)
	{
		perror("open");
		return -2;
	}
	if((mmapptr=mmap(0, st.st_size, PROT_READ, MAP_SHARED, fd, 0))==MAP_FAILED)
	{
		perror("mmap");
		return -2;
	}

	for(ptr=mmapptr;length>0;length-=ntohs(pkt->rlen), ptr+=ntohs(pkt->rlen))
	{
		struct ether_header *ehdr;
		struct ip *iph;
		struct udphdr *udphdr;
		struct timespec tsdag, tsdata, tsoffset;
		int neg;
		pkt=(dag_record_t *)ptr;
		// dumppacket(pkt);

		/*
		 * Check the packet for sanity and get a pointer to the data
		 */
		if(pkt->flags.rxerror || pkt->flags.dserror)
		{
			fprintf(stderr, "ERROR: receiver or DAG error in packet, ignoring\n");
			continue;
		}
		if(pkt->type!=2)
		{
			fprintf(stderr, "ERROR: unsupported packet type %d, ignoring\n", pkt->type);
			continue;
		}
		ehdr=(struct ether_header *)((char *)pkt+18);
		if(ntohs(ehdr->ether_type)!=ETHERTYPE_IP)
		{
			fprintf(stderr, "ERROR: no IP packet, ignoring (ether_type = 0x%x)\n",ntohs(ehdr->ether_type));
			continue;
		}
		iph=(struct ip *)((char *)ehdr+sizeof(struct ether_header));
		if(iph->ip_p!=IPPROTO_UDP)
		{
			// fprintf(stderr, "WARNING: no UDP packet, ignoring (protocol = %d)\n", iph->ip_p);
			continue;
		}
		udphdr=(struct udphdr *)((char *)iph+iph->ip_hl*4);
		printf("INFO: UDP packet from %d.%d.%d.%d (%02x:%02x:%02x:%02x:%02x:%02x) port %u -> "
				"%d.%d.%d.%d (%02x:%02x:%02x:%02x:%02x:%02x) port %u, ",
				ADDRARGS(iph->ip_src.s_addr), 
				ehdr->ether_shost[0],
				ehdr->ether_shost[1],
				ehdr->ether_shost[2],
				ehdr->ether_shost[3],
				ehdr->ether_shost[4],
				ehdr->ether_shost[5],
				ntohs(udphdr->uh_sport),
				ADDRARGS(iph->ip_dst.s_addr),
				ehdr->ether_dhost[0],
				ehdr->ether_dhost[1],
				ehdr->ether_dhost[2],
				ehdr->ether_dhost[3],
				ehdr->ether_dhost[4],
				ehdr->ether_dhost[5],
				ntohs(udphdr->uh_dport));

		/*
		 * Extract the two timestamps and compare them
		 */
		ts_to_s_ns(pkt->ts, (unsigned long *)&tsdag.tv_sec, &tsdag.tv_nsec);
		if(format==FORMAT_SENDER || format==FORMAT_SENDER_KERNEL)
		{
			struct outpkt *data=(struct outpkt *)((char *)udphdr+sizeof(struct udphdr));
			tsdata=data->ntv.time;
			if(!(data->tx.status & STA_NANO))
			{
				/*
				 * When the timestamp in the payload is in microseconds, 
				 * also round the DAG timestamp to full microseconds.
				 */
				ts_to_s_us(pkt->ts, (unsigned long *)&tsdag.tv_sec, &tsdag.tv_nsec);
				tsdag.tv_nsec*=1000;
				tsdata.tv_nsec*=1000;
			}
			neg=timespec_diff(&tsoffset, &tsdag, &tsdata);
			printf("accuracy = %s, tsdata = %ld.%09ld, tsdag = %ld.%09ld, offset = %s%ld.%09ld\n", 
					data->tx.status & STA_NANO ? "nano" : "micro", tsdata.tv_sec,
					tsdata.tv_nsec, tsdag.tv_sec, tsdag.tv_nsec, 
					neg==1 ? "-" : "", tsoffset.tv_sec, tsoffset.tv_nsec);
		}
		else if(format==FORMAT_SENDER_BOTH)
		{
			struct outpkt_both *data=(struct outpkt_both *)((char *)udphdr+sizeof(struct udphdr));
			struct timespec tsdata_user, tsdata_kernel, tsoffset_user, tsoffset_kernel, tsoffset_between;
			int neg_user, neg_kernel, neg_between;
			tsdata_user=data->ntv_user.time;
			tsdata_kernel=data->ntv_kernel.time;
			if(!(data->tx.status & STA_NANO))
			{
				/*
				 * When the timestamp in the payload is in microseconds, 
				 * also round the DAG timestamp to full microseconds.
				 */
				ts_to_s_us(pkt->ts, (unsigned long *)&tsdag.tv_sec, &tsdag.tv_nsec);
				tsdag.tv_nsec*=1000;
				tsdata_user.tv_nsec*=1000;
				tsdata_kernel.tv_nsec*=1000;
			}
			neg_user=timespec_diff(&tsoffset_user, &tsdag, &tsdata_user);
			neg_kernel=timespec_diff(&tsoffset_kernel, &tsdag, &tsdata_kernel);
			neg_between=timespec_diff(&tsoffset_between, &tsdata_kernel, &tsdata_user);
			printf("accuracy = %s, tsdata_user = %ld.%09ld, tsdata_kernel = %ld.%09ld, "
				"tsdag = %ld.%09ld, offset_user = %s%ld.%09ld, offset_kernel = %s%ld.%09ld, "
				"offset_between = %s%ld.%09ld\n", 
					data->tx.status & STA_NANO ? "nano" : "micro", 
					tsdata_user.tv_sec, tsdata_user.tv_nsec, 
					tsdata_kernel.tv_sec, tsdata_kernel.tv_nsec, 
					tsdag.tv_sec, tsdag.tv_nsec, 
					(neg_user==1 ? "-" : ""), tsoffset_user.tv_sec, tsoffset_user.tv_nsec,
					(neg_kernel==1 ? "-" : ""), tsoffset_kernel.tv_sec, tsoffset_kernel.tv_nsec,
					(neg_between==1 ? "-" : ""), tsoffset_between.tv_sec, tsoffset_between.tv_nsec);
		}
		else if(format==FORMAT_SEND_PACKET)
		{
			struct test_packet_header *hdr=
				(struct test_packet_header *)((char *)udphdr+sizeof(struct udphdr));
			struct test_packet_data *data=
				(struct test_packet_data *)((char *)hdr+ntohs(hdr->offset_to_data)+
							    sizeof(struct test_packet_header));
			printf("packet_id = %ld, ", ntohl(hdr->packet_id));
			printf("version = %hd, ", ntohs(data->version));
			printf("fingerprint = 0x%X, ", ntohl(data->fingerprint));
			printf("serial number = %ld, ", ntohl(data->serial_number));
			printf("pps status = 0x%X, ", data->timepars.pps_status);
			printf("pll offset = %ld, ", data->timepars.pll_offset);
			printf("maxerror = %ld, ", data->timepars.maxerror);
			printf("esterror = %ld, ", data->timepars.esterror);
			printf("current time = %ld:%ld, ", data->timepars.current_time.tv_sec,
					data->timepars.current_time.tv_usec);
			tsdata.tv_sec=data->timepars.current_time.tv_sec;
			tsdata.tv_nsec=data->timepars.current_time.tv_usec*1000;
			neg=timespec_diff(&tsoffset, &tsdag, &tsdata);
			printf("offset = %s%ld.%09ld\n", neg==1 ? "-" : "", tsoffset.tv_sec, tsoffset.tv_nsec);
		}
	}

	fprintf(stderr, "INFO: program ended with restlen = %ld\n",length);
	
	munmap(mmapptr, st.st_size);
	close(fd);
	
	return 0;
}
