/*
 * receiver.c
 *
 * Author: Thomas Wana <tom@ripe.net>
 *
 * $Id: receiver.c,v 1.2 2005/03/23 14:56:01 tom Exp $
 * 
 */

#include <sys/time.h>
#include <sys/timex.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

/*
 * RIPE Kernel level timestamping. These are setsockopt options that are not
 * defined in the libc headers.
 */
#define SO_RIPETSENABLE		0x2000
#define SO_RIPETSOFFSET		0x100F
#define SO_RIPETSFORMAT		0x100E

#define offsetof(TYPE, MEMBER) ((size_t) &((TYPE *)0)->MEMBER)

struct outpkt 
{
	/*
	 * The structure returned by ntp_gettime
	 */
	struct ntptimeval ntv;

	/* 
	 * The structure returned by ntp_adjtime
	 */
	struct timex tx;
} __attribute__ ((packed));

struct outpkt_both
{
	/*
	 * Two ntptimeval structures to hold both the userland and
	 * the kernel timestamp
	 */
	struct ntptimeval ntv_user;
	struct ntptimeval ntv_kernel;

	/* 
	 * The structure returned by ntp_adjtime
	 */
	struct timex tx;
} __attribute__ ((packed));

enum { TS_USER, TS_KERNEL, TS_BOTH } tsmode;
enum { RCV_USER, RCV_KERNEL } recvmode;

/*
 * Calculates tdiff = t1 - t2.
 * The second part is always positive, the returned integer determines wether the
 * complete result is positive (0) or negative (1). This is because tv_sec can't
 * be negative when it's 0.
 */
int timespec_diff(struct timespec *tdiff, struct timespec *t1, struct timespec *t2)
{
	int negative=0;
	tdiff->tv_sec=t1->tv_sec-t2->tv_sec;
	tdiff->tv_nsec=t1->tv_nsec-t2->tv_nsec;

	if(tdiff->tv_nsec > 1000000000)
	{
		tdiff->tv_nsec-=1000000000;
		tdiff->tv_sec++;
		negative=0;
	}
	if(tdiff->tv_nsec < 0)
	{
		if(tdiff->tv_sec>0)
		{
			tdiff->tv_nsec+=1000000000;
			tdiff->tv_sec--;
			negative=0;
		}
		else if(tdiff->tv_sec==0)
		{
			tdiff->tv_nsec*=-1;
			negative=1;
		}
		else
		{
			tdiff->tv_nsec*=-1;
			tdiff->tv_sec*=-1;
			negative=1;
		}
	}
	return negative;
}

int main(int argc, char **argv)
{
	struct sockaddr_in sin;
	int sock;
	struct outpkt pkt;
	struct outpkt_both pkt_both;

	if(argc!=4)
	{
		fprintf(stderr, "Usage: %s [user|kernel|both] [user|kernel] port\n\n", argv[0]);
		return -1;
	}

	if(!strcmp(argv[1], "user"))
	{
		tsmode=TS_USER;
	}
	else if(!strcmp(argv[1], "kernel"))
	{
		tsmode=TS_KERNEL;
	}
	else if(!strcmp(argv[1], "both"))
	{
		tsmode=TS_BOTH;
	}
	else
	{
		fprintf(stderr, "unsupported tsmode '%s'\n", argv[1]);
		return -2;
	}
	if(!strcmp(argv[2], "user"))
	{
		recvmode=RCV_USER;
	}
	else if(!strcmp(argv[2], "kernel"))
	{
		recvmode=RCV_KERNEL;
	}
	else
	{
		fprintf(stderr, "unsupported receiving mode '%s'\n", argv[1]);
		return -2;
	}

	if(tsmode!=TS_BOTH)
	{
		// lets get rid of this quickly ...
		printf("sorry, no other mode than TS_BOTH currently supported.\n");
		return -1;
	}

	/*
	 * Create and setup the socket
	 */
	if((sock=socket(AF_INET, SOCK_DGRAM, 0))<0)
	{
		perror("socket");
		return -2;
	}
	sin.sin_family=PF_INET;
	sin.sin_port=ntohs(strtol(argv[3], NULL, 10));
	sin.sin_addr.s_addr=htonl(INADDR_ANY);
	if(bind(sock, (struct sockaddr *)&sin, sizeof(sin))<0)
	{
		perror("bind");
		return -2;
	}

	if(recvmode==RCV_KERNEL)
	{
		/* 
		 * Setup the kernel timestamping
		 */
		int offset;
		int dummy=1;

		offset=offsetof(struct outpkt_both, ntv_user)+offsetof(struct ntptimeval, time);
		
		/* enable the timestamping */
		if (setsockopt (sock, SOL_SOCKET, SO_RIPETSENABLE, (void *)&dummy, sizeof(dummy)) < 0)
		{
			fprintf (stderr, "setsockopt SO_RIPETSENABLE failed: %s", strerror(errno));
			return -1;
		}
		/* Tell the kernel at which offset to put the timestamp */
		if (setsockopt (sock, SOL_SOCKET, SO_RIPETSOFFSET, (void *)&offset, sizeof(offset)) < 0)
		{
			fprintf (stderr, "setsockopt SO_RIPETSOFFSET failed: %s", strerror(errno));
			return -5;
		}
		/* Enable the highest precision mode (send struct timevals with nanoseconds) */
		dummy=1;
		if (setsockopt (sock, SOL_SOCKET, SO_RIPETSFORMAT, (void *)&dummy, sizeof(dummy)) < 0)
		{
			fprintf (stderr, "setsockopt SO_RIPETSFORMAT failed: %s", strerror(errno));
			return -5;
		}
	}

	/*
	 * Main loop
	 */
	while(1)
	{
		int n;
		struct timex tx;
		struct ntptimeval ntv;
		struct timespec tsrecv, tsdata_user, tsdata_kernel;
		struct timespec tsoffset_user, tsoffset_kernel, tsoffset_between;
		int neg_user, neg_kernel, neg_between;

		memset(&pkt, 0, sizeof(pkt));
		memset(&pkt_both, 0, sizeof(pkt_both));
		memset(&tx, 0, sizeof(tx));
		memset(&ntv, 0, sizeof(ntv));
		if(tsmode==TS_BOTH)
		{
			/*
			 * Start time critical
			 */
			n=recvfrom(sock, (void *)&pkt_both, sizeof(pkt_both), 0, NULL, NULL);
			if(n<0)
			{
				perror("recvfrom");
				close(sock);
				return -123;
			}
			if(ntp_gettime(&ntv)<0)
			{
				perror("ntp_gettime");
				return -1;
			}
			/*
			 * end time critical
			 */
			if(n!=sizeof(pkt_both))
			{
				fprintf(stderr, "short packet: %d\n", n);
				continue;
			}
			if(ntp_adjtime(&tx)<0)
			{
				perror("ntp_adjtime");
				return -1;
			}

			if(recvmode==RCV_USER)
			{
				tsrecv=ntv.time;
			}
			else if(recvmode==RCV_KERNEL)
			{
				tsrecv=pkt_both.ntv_user.time;
			}
			tsdata_user=pkt_both.ntv_user.time;
			tsdata_kernel=pkt_both.ntv_kernel.time;

			neg_user=timespec_diff(&tsoffset_user, &tsrecv, &tsdata_user);
			neg_kernel=timespec_diff(&tsoffset_kernel, &tsrecv, &tsdata_kernel);
			neg_between=timespec_diff(&tsoffset_between, &tsdata_kernel, &tsdata_user);
			printf("new packet: local accuracy = %s, tsdata_user = %ld.%09ld, "
				"tsdata_kernel = %ld.%09ld, recv time = %ld.%09ld, "
				"offset_user = %s%ld.%09ld, offset_kernel = %s%ld.%09ld, "
				"offset_between = %s%ld.%09ld"
				"\n",
				((tx.status & STA_NANO) ? "nano" : "micro"), 
				tsdata_user.tv_sec, tsdata_user.tv_nsec,
				tsdata_kernel.tv_sec, tsdata_kernel.tv_nsec,
				tsrecv.tv_sec, tsrecv.tv_nsec,
				(neg_user==1 ? "-" : ""), tsoffset_user.tv_sec, tsoffset_user.tv_nsec,
				(neg_kernel==1 ? "-" : ""), tsoffset_kernel.tv_sec, tsoffset_kernel.tv_nsec,
				(neg_between==1 ? "-" : ""), tsoffset_between.tv_sec, tsoffset_between.tv_nsec);
		}
		fflush(stdout);
	}

	return 0;
}
