/*
 * sender.c
 *
 * Author: Thomas Wana <tom@ripe.net>
 *
 * $Id: sender.c,v 1.10 2005/03/17 11:51:49 tom Exp $
 * 
 */

#include <sys/time.h>
#include <sys/timex.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

/*
 * RIPE Kernel level timestamping. These are setsockopt options that are not
 * defined in the libc headers.
 */
#define SO_RIPETSENABLE		0x2000
#define SO_RIPETSOFFSET		0x100F
#define SO_RIPETSFORMAT		0x100E

#define offsetof(TYPE, MEMBER) ((size_t) &((TYPE *)0)->MEMBER)

struct outpkt 
{
	/*
	 * The structure returned by ntp_gettime
	 */
	struct ntptimeval ntv;

	/* 
	 * The structure returned by ntp_adjtime
	 */
	struct timex tx;
} __attribute__ ((packed));

struct outpkt_both
{
	/*
	 * Two ntptimeval structures to hold both the userland and
	 * the kernel timestamp
	 */
	struct ntptimeval ntv_user;
	struct ntptimeval ntv_kernel;

	/* 
	 * The structure returned by ntp_adjtime
	 */
	struct timex tx;
} __attribute__ ((packed));

enum { TS_USER, TS_KERNEL, TS_BOTH } tsmode;

int main(int argc, char **argv)
{
	struct sockaddr_in sin;
	int sock;
	struct hostent *he;
	struct outpkt pkt;
	struct outpkt_both pkt_both;

	if(argc!=4)
	{
		fprintf(stderr, "Usage: %s [user|kernel|both] host port\n\n", argv[0]);
		return -1;
	}

	if(!strcmp(argv[1], "user"))
	{
		tsmode=TS_USER;
	}
	else if(!strcmp(argv[1], "kernel"))
	{
		tsmode=TS_KERNEL;
	}
	else if(!strcmp(argv[1], "both"))
	{
		tsmode=TS_BOTH;
	}
	else
	{
		fprintf(stderr, "unsupported tsmode '%s'\n", argv[1]);
		return -2;
	}

	/*
	 * Create and setup the socket
	 */
	if((sock=socket(AF_INET, SOCK_DGRAM, 0))<0)
	{
		perror("socket");
		return -2;
	}
	if(tsmode==TS_KERNEL || tsmode==TS_BOTH)
	{
		/*
		 * enable kernel-timestamping on the socket
		 */
		int offset;
		int dummy = 1;
		
		if(tsmode==TS_KERNEL)
		{
			offset=offsetof(struct outpkt, ntv)+offsetof(struct ntptimeval, time);
		}
		else if(tsmode==TS_BOTH)
		{
			offset=offsetof(struct outpkt_both, ntv_kernel)+offsetof(struct ntptimeval, time);
		}
		if (setsockopt (sock, SOL_SOCKET, SO_RIPETSENABLE, (void *)&dummy, sizeof(dummy)) < 0)
		{
			fprintf (stderr, "setsockopt SO_RIPETSENABLE failed: %s", strerror(errno));
			return -1;
		}
		/* Tell the kernel at which offset to put the timestamp */
		if (setsockopt (sock, SOL_SOCKET, SO_RIPETSOFFSET, (void *)&offset, sizeof(offset)) < 0)
		{
			fprintf (stderr, "setsockopt SO_RIPETSOFFSET failed: %s", strerror(errno));
			return -5;
		}
		/* Enable the highest precision mode (send struct timevals with nanoseconds) */
		dummy=1;
		if (setsockopt (sock, SOL_SOCKET, SO_RIPETSFORMAT, (void *)&dummy, sizeof(dummy)) < 0)
		{
			fprintf (stderr, "setsockopt SO_RIPETSFORMAT failed: %s", strerror(errno));
			return -5;
		}
	}
	sin.sin_family=PF_INET;
	sin.sin_port=ntohs(strtol(argv[3], NULL, 10));
	if((he=gethostbyname(argv[2]))==NULL)
	{
		herror("gethostbyname");
		return -3;
	}
	sin.sin_addr.s_addr=*(in_addr_t *)he->h_addr_list[0];

	/*
	 * Main loop
	 */
	memset(&pkt, 0, sizeof(pkt));
	memset(&pkt_both, 0, sizeof(pkt_both));
	while(1)
	{
		/*
		 * Time critical section starts here
		 * ntp_gettime should be last
		 */
		if(tsmode==TS_USER || tsmode==TS_KERNEL)
		{
			if(ntp_adjtime(&pkt.tx)<0)
			{
				perror("ntp_adjtime");
				return -1;
			}
			if(ntp_gettime(&pkt.ntv)<0)
			{
				perror("ntp_gettime");
				return -1;
			}
			if(sendto(sock, (const void *)&pkt, sizeof(pkt), 0, (const struct sockaddr *)&sin, sizeof(sin))<0)
			{
				perror("sendto");
				return -1;
			}
		}
		else /* TS_BOTH */
		{
			if(ntp_adjtime(&pkt_both.tx)<0)
			{
				perror("ntp_adjtime");
				return -1;
			}
			if(ntp_gettime(&pkt_both.ntv_user)<0)
			{
				perror("ntp_gettime");
				return -1;
			}
			if(sendto(sock, (const void *)&pkt_both, sizeof(pkt_both), 0, (const struct sockaddr *)&sin, sizeof(sin))<0)
			{
				perror("sendto");
				return -1;
			}
		}
		/*
		 * Time critical section done
		 */

		printf("packet sent\n");
		memset(&pkt, 0, sizeof(pkt));
		memset(&pkt_both, 0, sizeof(pkt_both));
		sleep(1);
	}

	return 0;
}
