/*
 * changeboot.c : dump or set the boot devices on an AMI-BIOS
 *
 * In the NVRAM, two bytes set the boot devices:
 *   0x53, low four bits: first devices
 *   0x53, high four bits: second devices
 *   0x54, low four bits: third devices
 *
 * There is also an checksum which must be corrected in 0x3e and 0x3f.
 *
 * The nvram is accessed through address 0x70 and 0x71. 0x70 is 
 * the address latch, and then the byte can be read or written
 * using 0x71.
 *
 * -DLINUX for Linux and -DFREEBSD for freebsd.
 * 
 * Since this is linked against getopt, it has to be GPL'ed.
 *
 * Copyright (C) 2001, Ripe NCC
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Sean Young <sean@ripe.net>
 */

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <getopt.h>
#include <stdlib.h>

#ifdef FREEBSD
#include <machine/cpufunc.h>
#define IO_DEV  	"/dev/io"

#endif

#ifdef LINUX
#define inline __inline__	/* just ansi compatibility */
#include <asm/io.h>
#include <sys/io.h>

/* 
 * for some reason outb has its parameters the wrong way 
 * round on linux. I'm not sure if the following is ansi c.
 */
#define outb(port, data)	outb(data, port)
#endif

#define NVRAM_SIZE	128
#define LAST_TYPE	(0xb)

static const char *boot_types[] = {
        "disabled",     /* 0 */
        "ide0",         /* 1 */
        "ide1",         /* 2 */
        "ide2",         /* 3 */
        "ide3",         /* 4 */
        "floppy a:",    /* 5 */
        "floppy b:",    /* 6 */
        "zip a: ls120", /* 7 */
        "atapi zip c:", /* 8 */
        "cdrom",        /* 9 */
        "scsi",         /* a */
        "network",      /* b */
	NULL,		/* c */
	NULL,		/* d */
	NULL,		/* e */
	NULL,		/* f */
};

static void print_usage () {
    int i;

    printf ("changeboot 1.0, changes boot devices on AMI BIOS\n"
	    "Usage: changeboot [OPTION]\n\n"
	    " -h, --help\t\tPrint this help.\n"
	    " -d, --dump\t\tDump current setup.\n"
	    "     --boot0=DEV\tSet first boot device to DEV\n"
	    "     --boot1=DEV\tSet second boot device to DEV\n"
	    "     --boot2=DEV\tSet third boot device to DEV\n"
	    "\n"
	    "DEV is one the following devices:\n");

   for (i=0;i<=LAST_TYPE;i++) {
	printf ("\t%d\t\t%s\n", i, boot_types[i]);
   }
}

static int read_nvram (unsigned char *nvram);
static int write_nvram (unsigned char *nvram);

int main (int argc, char *argv[]) {
    static struct option long_options[] = {
	{ "dump", 0, 0, 'd' },
	{ "help", 0, 0, 'h' }, 
 	{ "boot0", 1, 0, '0' },
	{ "boot1", 1, 0, '1' },
	{ "boot2", 1, 0, '2' },
   };
   int	c, b0, b1, b2, b0_val, b1_val, b2_val, dump;
   int  option_index;
   unsigned char nvram[NVRAM_SIZE];

   option_index = dump = b0 = b1 = b2 = 0;

   /* to prevent compiler warnings */
   b0_val = b1_val = b2_val = 0;

   while (1) {
	c = getopt_long (argc, argv, "dh", long_options, &option_index);

	if (c == -1) break;

	switch (c) {
	case 'h':
	    print_usage ();
	    return 0;
	case 'd':
	    dump = 1;
	    break;
	case '0':
	    b0 = 1;
	    if ((1 != sscanf (optarg, "%d", &b0_val) ) ||
	     	(b0_val < 0) || (b0_val > LAST_TYPE)) {
		fprintf (stderr, "%s: invalid argument for --boot0 -- %s\n", 
		    argv[0], optarg);
		return 2;
	    }
	    break;
	case '1':
	    b1 = 1;
	    if ((1 != sscanf (optarg, "%d", &b1_val) ) ||
		(b1_val < 0) || (b1_val > LAST_TYPE)) {
		fprintf (stderr, "%s: invalid argument for --boot1 -- %s\n", 
		    argv[0], optarg);
		return 2;
	    }
	    break;
	case '2':
	    b2 = 1;
	    if ((1 != sscanf (optarg, "%d", &b2_val) ) ||
	    	(b2_val < 0) || (b2_val > LAST_TYPE)) {
		fprintf (stderr, "%s: invalid argument for --boot2 -- %s\n", 
		    argv[0], optarg);
		return 2;
	    }
	    break;
	case '?':
	    break;
	}
    }

    if (optind < argc) {
	fprintf (stderr, "%s: invalid argument -- %s\n", argv[0], argv[optind]);
	return 2;
    }

    if (dump && (b0 || b1 || b2) ) {
	fprintf (stderr, "%s: --dump and --boot cannot be combined\n", argv[0]);
	return 2;
    }

    if (dump) {
	if (read_nvram (nvram) ) return 1;

	printf ("boot0: %s\n", boot_types[(nvram[0x53] & 0x0f)]);
	printf ("boot1: %s\n", boot_types[(nvram[0x53] >> 4)]);
	printf ("boot2: %s\n", boot_types[(nvram[0x54] & 0x0f)]);

    } else if  (b0 || b1 || b2) {
	int checksum, new53, new54;

	if (read_nvram (nvram) ) return 1;

	/* 
  	 * I don't know how the checksum is calculated, however what I do know is that
   	 * it's a set of bytes added up, included the two bytes that we're modifying.
         * So by calculating the difference we know what the difference of the checksum
	 * is. It's a hack, but it is known to work. :-)
 	 */

	new53 = nvram[0x53];
	new54 = nvram[0x54];

	if (b0) new53 = (new53 & 0xf0) | b0_val;
	if (b1) new53 = (new53 & 0x0f) | b1_val << 4;
	if (b2) new54 = (new54 & 0xf0) | b2_val;
	
	checksum = (nvram[0x3e] << 8) | nvram[0x3f];
	checksum += new53 - nvram[0x53];
	checksum += new54 - nvram[0x54];
	nvram[0x3e] = checksum >> 8;
	nvram[0x3f] = checksum & 0xff;
	nvram[0x53] = new53;
	nvram[0x54] = new54;

	if (write_nvram (nvram) ) return 1;

    } else {
	fprintf (stderr, "%s: missing argument\n\nTry `changeboot --help'\n", argv[0]);
	return 2;
    }

    return 0;
}

static int write_nvram (unsigned char *nvram)  {
#ifdef FREEBSD
    int fd;

    /* get permission for direct I/O */
    if (-1 == (fd = open (IO_DEV, O_RDONLY))) {
        perror (IO_DEV);
        return 1;
    }
#endif

    /* write the register + checksum */
    outb (0x70, 0x53); outb (0x71, nvram[0x53]);
    outb (0x70, 0x54); outb (0x71, nvram[0x54]);
    outb (0x70, 0x3e); outb (0x71, nvram[0x3e]);
    outb (0x70, 0x3f); outb (0x71, nvram[0x3f]);

#ifdef FREEBSD
    close (fd);
#endif
    return 0;
}

static int read_nvram (unsigned char *nvram) {
    int i;

#ifdef FREEBSD
    int fd;

    /* get permission for direct I/O */
    if (-1 == (fd = open (IO_DEV, O_RDONLY))) {
        perror (IO_DEV);
        return 1;
    }
#endif

#ifdef LINUX
    if (ioperm (0x70, 2, 1) ) {
	perror ("ioperm");
     }
#endif

    for (i=0;i<NVRAM_SIZE;i++) {
        outb (0x70, i);
        nvram[i] = inb (0x71);
    }

#ifdef FREEBSD
    close (fd);
#endif
    
    return 0;
}


