/*
 * process_dag.c
 *
 * Reads two dagsnap files, finds matching packets (UDP source and dest port,
 * payload memcmp) and calculates a time offset between the two DAG timestamps.
 *
 * Author: Thomas Wana <tom@ripe.net>
 *
 * $Id: process_dag.c,v 1.4 2005/04/06 08:20:22 tom Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/time.h>
#include <sys/timex.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <net/ethernet.h>
#include <netinet/in_systm.h>
#include <sys/param.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include "dagapi.h"

#ifndef _POSIX_MAPPED_FILES
#error Unable to mmap files; exiting
#endif

#define ADDRARGS(addr) (addr)&0xFF,((addr)>>8)&0xFF,((addr)>>16)&0xFF,((addr)>>24)&0xFF

/*
 * Converts the unsigned long long timestamp from the DAG card to seconds
 * and nanoseconds.
 * From the DAG Extensible Record Format doc, Page 10, "Timestamps" (adapted
 * to support nanoseconds)
 */
void ts_to_s_ns(unsigned long long ts, unsigned long *s, unsigned long *ns)
{
	*s=(ts>>32)&0xFFFFFFFFULL;
	ts=((ts&0xFFFFFFFFULL)*1000*1000*1000);
	ts+=(ts & 0x80000000ULL)<<1;
	*ns=ts>>32;
	if(*ns >= 1000000000)
	{
		*ns-=1000000000;
		*s+=1;
	}
}

/*
 * Calculates tdiff = t1 - t2.
 * The second part is always positive, the returned integer determines wether the
 * complete result is positive (0) or negative (1). This is because tv_sec can't
 * be negative when it's 0 (i.e. there is no concept of +0 or -0).
 */
int timespec_diff(struct timespec *tdiff, struct timespec *t1, struct timespec *t2)
{
	int negative=0;
	tdiff->tv_sec=t1->tv_sec-t2->tv_sec;
	tdiff->tv_nsec=t1->tv_nsec-t2->tv_nsec;

	if(tdiff->tv_nsec > 1000000000)
	{
		tdiff->tv_nsec-=1000000000;
		tdiff->tv_sec++;
		negative=0;
	}
	if(tdiff->tv_nsec < 0)
	{
		if(tdiff->tv_sec>0)
		{
			tdiff->tv_nsec+=1000000000;
			tdiff->tv_sec--;
			negative=0;
		}
		else if(tdiff->tv_sec==0)
		{
			tdiff->tv_nsec*=-1;
			negative=1;
		}
		else
		{
			tdiff->tv_nsec*=-1;
			tdiff->tv_sec*=-1;
			negative=1;
		}
	}
	return negative;
}

/*
 * Converts the unsigned long long timestamp from the DAG card to seconds
 * and microseconds.
 * From the DAG Extensible Record Format doc, Page 10, "Timestamps" 
 */
void ts_to_s_us(unsigned long long ts, unsigned long *s, unsigned long *us)
{
	*s=(ts>>32)&0xFFFFFFFFULL;
	ts=((ts&0xFFFFFFFFULL)*1000*1000);
	ts+=(ts & 0x80000000ULL)<<1;
	*us=ts>>32;
	if(*us >= 1000000)
	{
		*us-=1000000;
		*s+=1;
	}
}

void dumppacket(dag_record_t *pkt)
{
	const char *types[]={"Legacy", "PoS w/HDLC framing", "Ethernet", "ATM Cell", 
			     "reassembled AAL5 frame", "Multi-channel HDLC frame",
			     "Multi-channel Raw link data", "Multi-channel ATM Cell"};
	unsigned long s, ns;
	struct ether_header *ehdr;
	struct ip *iph;

	ts_to_s_ns(pkt->ts, &s, &ns);

	fprintf(stderr, "INFO: new packet record\n");
	fprintf(stderr, "INFO: time:           %ld s, %ld ns\n", s, ns);
	fprintf(stderr, "INFO: packet len:     %d\n", ntohs(pkt->wlen));
	fprintf(stderr, "INFO: type:           %s\n", types[pkt->type]);
	fprintf(stderr, "INFO: flags.iface:    %d\n", pkt->flags.iface);
	fprintf(stderr, "INFO: flags.vlen:     %d\n", pkt->flags.vlen);
	fprintf(stderr, "INFO: flags.trunc:    %d\n", pkt->flags.trunc);
	fprintf(stderr, "INFO: flags.rxerror:  %d\n", pkt->flags.rxerror);
	fprintf(stderr, "INFO: flags.dserror:  %d\n", pkt->flags.dserror);

	ehdr=(struct ether_header *)((char *)pkt+18);
	fprintf(stderr, "INFO:    ether.dhost: %02x:%02x:%02x:%02x:%02x:%02x\n",
			ehdr->ether_dhost[0],
			ehdr->ether_dhost[1],
			ehdr->ether_dhost[2],
			ehdr->ether_dhost[3],
			ehdr->ether_dhost[4],
			ehdr->ether_dhost[5]);
	fprintf(stderr, "INFO:    ether.shost: %02x:%02x:%02x:%02x:%02x:%02x\n",
			ehdr->ether_shost[0],
			ehdr->ether_shost[1],
			ehdr->ether_shost[2],
			ehdr->ether_shost[3],
			ehdr->ether_shost[4],
			ehdr->ether_shost[5]);
	fprintf(stderr, "INFO:    ether.type:  %d\n", ntohs(ehdr->ether_type));
	if(ntohs(ehdr->ether_type)!=ETHERTYPE_IP)
	{
		fprintf(stderr, "WARNING: no IP packet, ignoring\n");
		fprintf(stderr, "INFO: --------------------\n");
		return;
	}
	iph=(struct ip *)((char *)ehdr+sizeof(struct ether_header));
	fprintf(stderr, "INFO:    ip.proto:    %d\n", iph->ip_p);
	fprintf(stderr, "INFO:    ip.saddr:    %d.%d.%d.%d\n", ADDRARGS(iph->ip_src.s_addr));
	fprintf(stderr, "INFO:    ip.daddr:    %d.%d.%d.%d\n", ADDRARGS(iph->ip_dst.s_addr));

	fprintf(stderr, "INFO: --------------------\n");
}

int main(int argc, char **argv)
{
	dag_record_t *pkt1, *pkt2;
	int fd1, fd2;
	void *mmapptr1, *ptr1, *mmapptr2, *ptr2;
	struct stat st1, st2;
	long length, length2;

	if(argc!=3)
	{
		fprintf(stderr, "Usage: %s filename1 filename2\n\n", argv[0]);
		return -1;
	}

	/*
	 * mmap the first file
	 */
	if(stat(argv[1], &st1)<0)
	{
		perror("stat");
		return -2;
	}
	length=(long)st1.st_size;
	if((fd1=open(argv[1], 0))<0)
	{
		perror("open");
		return -2;
	}
	if((mmapptr1=mmap(0, st1.st_size, PROT_READ, MAP_SHARED, fd1, 0))==MAP_FAILED)
	{
		perror("mmap");
		return -2;
	}
	/*
	 * mmap the second file
	 */
	if(stat(argv[2], &st2)<0)
	{
		perror("stat");
		return -2;
	}
	length2=(long)st2.st_size;
	if((fd2=open(argv[2], 0))<0)
	{
		perror("open");
		return -2;
	}
	if((mmapptr2=mmap(0, st2.st_size, PROT_READ, MAP_SHARED, fd2, 0))==MAP_FAILED)
	{
		perror("mmap");
		return -2;
	}

	for(ptr1=mmapptr1;length>0;length-=ntohs(pkt1->rlen), ptr1+=ntohs(pkt1->rlen))
	{
		struct ether_header *ehdr;
		struct ip *iph;
		struct udphdr *udphdr1, *udphdr2;
		struct timespec tsdag1, tsdag2, tsdiff;
		struct ntptimeval *data1, *data2;
		int neg;
		int match;
		pkt1=(dag_record_t *)ptr1;
		pkt2=NULL;

		/*
		 * Check the packet for sanity and get a pointer to the data
		 */
		if(pkt1->flags.rxerror || pkt1->flags.dserror)
		{
			fprintf(stderr, "ERROR: receiver or DAG error in packet, ignoring\n");
			continue;
		}
		if(pkt1->type!=2)
		{
			fprintf(stderr, "ERROR: unsupported packet type %d, ignoring\n", pkt1->type);
			continue;
		}
		ehdr=(struct ether_header *)((char *)pkt1+18);
		if(ntohs(ehdr->ether_type)!=ETHERTYPE_IP)
		{
			fprintf(stderr, "ERROR: no IP packet, ignoring (ether_type = 0x%x)\n",ntohs(ehdr->ether_type));
			continue;
		}
		iph=(struct ip *)((char *)ehdr+sizeof(struct ether_header));
		if(iph->ip_p!=IPPROTO_UDP)
		{
			// fprintf(stderr, "WARNING: no UDP packet, ignoring (protocol = %d)\n", iph->ip_p);
			continue;
		}
		udphdr1=(struct udphdr *)((char *)iph+iph->ip_hl*4);
		printf("INFO: UDP packet from %d.%d.%d.%d (%02x:%02x:%02x:%02x:%02x:%02x) port %u -> "
				"%d.%d.%d.%d (%02x:%02x:%02x:%02x:%02x:%02x) port %u, ",
				ADDRARGS(iph->ip_src.s_addr), 
				ehdr->ether_shost[0], ehdr->ether_shost[1],
				ehdr->ether_shost[2], ehdr->ether_shost[3],
				ehdr->ether_shost[4], ehdr->ether_shost[5],
				ntohs(udphdr1->uh_sport),
				ADDRARGS(iph->ip_dst.s_addr),
				ehdr->ether_dhost[0], ehdr->ether_dhost[1],
				ehdr->ether_dhost[2], ehdr->ether_dhost[3],
				ehdr->ether_dhost[4], ehdr->ether_dhost[5],
				ntohs(udphdr1->uh_dport));
		data1=(struct ntptimeval *)((char *)udphdr1+sizeof(struct udphdr));

		/*
		 * Find the matching packet. I know the algorithm sucks, there is room
		 * for optimization.
		 */
		match=0;
		length2=st2.st_size;
		for(ptr2=mmapptr2; length2>0 && match==0; length2-=ntohs(pkt2->rlen), ptr2+=ntohs(pkt2->rlen))
		{
			struct ether_header *ehdr;
			struct ip *iph;

			pkt2=(dag_record_t *)ptr2;
			/*
			 * Check the packet for sanity and get a pointer to the data
			 */
			if(pkt2->flags.rxerror || pkt2->flags.dserror)
			{
				fprintf(stderr, "   ERROR: receiver or DAG error in packet, ignoring "
						"(packet length=%d)\n", pkt2->rlen);
				continue;
			}
			if(pkt2->type!=2)
			{
				fprintf(stderr, "   ERROR: unsupported packet type %d, ignoring (packet length=%d)\n", 
						pkt2->type, pkt2->rlen);
				continue;
			}
			ehdr=(struct ether_header *)((char *)pkt2+18);
			if(ntohs(ehdr->ether_type)!=ETHERTYPE_IP)
			{
				// fprintf(stderr, "   ERROR: no IP packet, ignoring (ether_type = 0x%x)\n",ntohs(ehdr->ether_type));
				continue;
			}
			iph=(struct ip *)((char *)ehdr+sizeof(struct ether_header));
			if(iph->ip_p!=IPPROTO_UDP)
			{
				// fprintf(stderr, "WARNING: no UDP packet, ignoring (protocol = %d)\n", iph->ip_p);
				continue;
			}
			udphdr2=(struct udphdr *)((char *)iph+iph->ip_hl*4);
			data2=(struct ntptimeval *)((char *)udphdr2+sizeof(struct udphdr));

			if(udphdr1->uh_sport==udphdr2->uh_sport &&
			   udphdr1->uh_dport==udphdr2->uh_dport &&
			   memcmp(data1, data2, sizeof(*data1))==0)
			{
				match=1;
			}
		}

		if(match==0)
		{
			fprintf(stderr, "   ERROR: couldn't find matching packet, ignoring\n");
			printf("nomatch\n");
			continue;
		}

		/*
		 * pkt1 and pkt2, udphdr1 and udphdr2, data1 and data2 now point to the right packets
		 */

		/*
		 * Extract the two timestamps and compare them
		 */
		ts_to_s_ns(pkt1->ts, (unsigned long *)&tsdag1.tv_sec, &tsdag1.tv_nsec);
		ts_to_s_ns(pkt2->ts, (unsigned long *)&tsdag2.tv_sec, &tsdag2.tv_nsec);
		neg=timespec_diff(&tsdiff, &tsdag2, &tsdag1);
		printf("offset=%s%ld.%09ld\n",(neg==0 ? "" : "-"), tsdiff.tv_sec, tsdiff.tv_nsec);
	}

	fprintf(stderr, "INFO: program ended with restlen = %ld\n",length);
	
	munmap(mmapptr1, st1.st_size);
	munmap(mmapptr2, st2.st_size);
	close(fd1);
	close(fd2);
	
	return 0;
}
