/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 *
 * $Id: dcf77.c,v 1.3 2005/01/26 15:28:53 tom Exp $
 *
 * This receiver use the reverse principle of all other DCF77 receivers I have
 * seen implemented, instead of relying on the S/N to be good enough for the
 * correct minute-synchronization to trivially stand out (ie: by the absense
 * of the 59th second pulse), I instead start out with 60 possible hypothesis
 * for where the minute starts, and disprove them as I collect bits.
 *
 * There are many features of the DCF77 timegram which can be checked even
 * though we do not know what time/date it actually is.  For instance a minute
 * reading of 71 would obviously be wrong, as would illegal BCD digits, wrong
 * parity or the two timezone bits being identical.
 *
 * It is possible to also check that the first 14 bits follow a particular
 * pattern, they have been zero every single time I have looked but the
 * PTB reserves the right to play with them.
 *
 * The principle behind this is based on the observation that noise almost
 * always consist of missing pulses, and that '0' and '1' bit pulses very
 * rarely gets confused.
 *
 * This method allows us to tighten the frontend filters much more than
 * other receivers because we can tolerate a high ratio of missing pulses.
 *
 * It is interesting to note that this method does not even require a full
 * minute worth of pulses before the top of the minute is uniquely determined:
 * even though we only have one or two bits actually supporting it, all other
 * possibilities have two or more bits rejecting them as possibilities.
 *
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <err.h>
#include <errno.h>
#include <math.h>
#include <stdlib.h>

#include <ntpns.h>
#include "dcf77.h"

struct logh *l_debug, *l_filter, *l_msg;

/*
 * These are our default tweak parameters
 */
static struct dcf77 defaults = {
	.max_gap = 10,
	.max_jitter = .020,
	.low_zero = .070,
	.high_zero = .140,
	.low_one = .165,
	.high_one = .250,
	.ts_zero = .145,
	.ts_one = .255,
	.disproof = 2,
	.squelch = .33,
	.timeout = 600,
	.lockbits = 20,
	.untrack = 2,
	.delay = 48e-3,
	.preamble = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
};

static TAILQ_HEAD(, dcf77)	dcf77_head =
    TAILQ_HEAD_INITIALIZER(dcf77_head);

/**********************************************************************/

/*
 * This is slightly complicated, but there is a good reason:  In order to 
 * minimize the effects of noise which mostly take the form of spurious extra
 * pulses, we want to poll the PPS-API as soon as possible after the likely
 * end of the pulse.
 *
 * Once we have a pulse of valid length, we poll again 1.135 second after
 * the start time of that pulse, to look for a following short pulse.
 *
 * If we find a valid pulse at that time, we start over from above, if no 
 * valid pulse is found, we look for a long pulse .1 second later.
 *
 * If we find a valid pulse at that time, we start over from above, if none
 * is found we look for a short pulse .9 second later.
 *
 * A slight complication here is that our pulses are timestamped on
 * CLOCK_REALTIME, but eventlib may run on CLOCK_MONOTONIC and the fact
 * that we can only fiddle the interval setting on recurring timers.
 *
 * All the magic numbers are tweakable.
 *
 *
 *     A     B  C  D  E                                 a     b  c  d  e
 *     v     v  v  v  v                                 v     v  v  v  v
 *
 * ----       ----- ------------------------------------       ----- -------
 *     |     |     |                                    |     |     |
 *     |     |     |                                    |     |     |
 *      ----- -----                                      ----- -----
 *
 *
 */

static void
dcf77_timer(evContext ctx, void *uap, struct timespec due, struct timespec inter __unused)
{
	struct dcf77 *ip;
	double dt, dw, delta;
	struct timespec ts1, ts2;
	int i, j, class;
	time_t t;

	ip = uap;

	if (ip->clock_step != clock_step) {
		dcf77_change_state(ip, DCF_INIT);
		ip->clock_step = clock_step;
		return;
	}
	time(&t);
	t -= ip->t0;
	if (t < 0 || t > ip->timeout) {
		Log(l_msg, "%d No luck, try again t %d\n", ip->t0, t);
		dcf77_change_state(ip, DCF_INIT);
		return;
	}

	if (ip->lasti == -1) {
		ts2 = evConsTime(0,  90000000);
	} else if (ip->timer_state == SHORT) {
		/*
		 * We are at 'C' and set up things for the case where we find
		 * no pulse, so we aim the timer at 'E' to look for a long
		 * pulse.  In theory we could exploit the PPS-API to tell us
		 * if the first flank of the long pulse happened, but there
		 * seems to be little actual benefit.
		 */
		ip->timer_state = LONG;
		ts2 = evSubTime(ip->sample_one, ip->sample_zero);
	} else {
		/*
		 * Here we are at 'E' and set up things for the case were we
		 * find no pulse, so we aim for 'c'.
		 */
		ip->timer_state = SHORT;
		ts2 = ip->sample_zero;
		ts2.tv_sec++;
		ts2 = evSubTime(ts2, ip->sample_one);
	}
	j = 0;
	i = PpsPickup(&ip->pps, ip->ts, 2);
	if (i != 2) {
		/* Too bad, no pulse, aim timer, done. */
		i = evResetTimer(ctx, ip->tid, dcf77_timer, ip, due, ts2);
		if (i < 0)
			err(1, "evResetTimer() failed %s:%d",
			     __FILE__, __LINE__);
		return;
	}
	record("DCF77PPS", ip->unit, ip->ts, sizeof ip->ts);

	/* So we had a pulse, when and how big ? */

	dw = TsDiff(&ip->ts[1], &ip->ts[0]);

	/* Update pulse-width histogram */
	i = dw / NDWBUCKET;
	if (i >= NDWHIST)
		i = NDWHIST - 1;
	ip->dwhist[i]++;

	if (dw < ip->low_zero)
		class = -2;		/* too short */
	else if (dw < ip->high_zero)
		class = -1;		/* Short pulse, ie: A-B */
	else if (dw < ip->low_one)
		class = 0;		/* Not short, not long, throw away */
	else if (dw < ip->high_one)
		class = 1;		/* Long pulse, ie: A-D */
	else
		class = 2;		/* Too long */

	if (ip->lasti == -1) {
		dt = 1.0;
	} else {
		dt = TsDiff(&ip->ts[0], &ip->ts[2]);
	}

	/* Update pulse-spacing histogram */
	i = (int)(dt + .5);
	if (i < NDTHIST)
		ip->dthist[i]++;
	else
		ip->dthist[NDTHIST - 1]++;

	if (dt > 59 || dt < 0) {
		/*
		 * Loss of signal or timetravel: eject, EJECT!
		 *
		 * '59' is not entirely random.  The blameallocation code
		 * depends on knowing the previos minute and it would be
		 * confused by a missing minute.  It doesn't pay off to
		 * fix this confusion because an empty minute does not
		 * assign any blame anyway.
		 */
		Log(l_msg, "%d loss of signal dt=%f\n", ip->t0, dt);
		dcf77_change_state(ip, DCF_INIT);
		return;
	}

	if ((fabs(dt / i - 1.0) > ip->max_jitter && dt < ip->max_gap) ||
	    (class != -1 && class != 1)) {
		/*
		 * dt fell out side the wedge-shaped window or dw fell
		 * outside the two permissible ranges: reject pulse.
		 */
		char buf0[30], buf1[30], buf2[30];
		TsFormat(&ip->ts[0], buf0);
		TsFormat(&ip->ts[1], buf1);
		TsFormat(&ip->ts[2], buf2);
		Log(l_filter,
		   "Reject (%d) i %d dw %f dt %f last %s this %s ... %s\n",
		   class, i, dw, dt, buf2, buf0, buf1);
		ip->noise++;
	} else {
		/* Valid pulse, aim timer at 'c' relative to 'A' */

		/* ts1 = 'A' (start of pulse) [UTC sec] */
		TsTimespec(&ip->ts[0], &ts1);

		/* ts2 = 'C' or 'E' (time now) [UTC sec] */
		clock_gettime(CLOCK_REALTIME, &ts2);

		/* ts2 = how long since pulse started [sec] */
		ts2 = evSubTime(ts2, ts1);

		/* ts2 = when did pulse start [eventlib sec] */
		ts2 = evSubTime(evNowTime(), ts2);

		/* ts2 = when should we look again [eventlib sec] */
		ts2 = evAddTime(ip->sample_zero, ts2);
		ts2.tv_sec++;

		/* ts2 = how long from our due is that [sec] */
		ts2 = evSubTime(ts2, due);

		/* Look for short pulse next */
		ip->timer_state = SHORT;

		ip->ts[2] = ip->ts[0];

		TsNormalize(&ip->ts[0]);
		delta = (int64_t)ip->ts[0].ts.bt.frac;
		delta /= 18446744073709551616.0; /* 2^64 in case you wondered */
		delta -= ip->delay;
		if (ip->state == DCF_LOCK) {
			SourceDelta(ip->sid, &ip->ts[0],
			    delta - ip->delta, SRC_UTC);
		} else {
			SourceDelta(ip->sid, &ip->ts[0],
			    delta, SRC_FRACTION);
		}

		/* Round second based on fraction */
		i = ip->ts[0].ts.bt.sec;
		if ((int64_t)ip->ts[0].ts.bt.frac < 0)
			i++;
		i %= 60;

		while (ip->lasti != -1) {
			/* Add missing pulses */
			j = (ip->lasti + 1) % 60;
			if (j == i)
				break;
			dcf77_addpulse(ctx, ip, j, 0);
			ip->missing++;
		}
		dcf77_addpulse(ctx, ip, i, class);
	}
	i = evResetTimer(ctx, ip->tid, dcf77_timer, ip, due, ts2);
	if (i < 0)
		LogErr(Lerror, "evResetTimer(dcf77) failed %s:%d",
		    __FILE__, __LINE__);
}

/**********************************************************************/

static void
dcf77_showparam(FILE *f, struct dcf77 *ip)
{
	int i;

	fprintf(f, "%-20.20s %.0f seconds\n",
	    "propagation_delay:", ip->propagation_delay);
	fprintf(f, "%-20.20s %.0f seconds\n", "max_gap:", ip->max_gap);
	fprintf(f, "%-20.20s %.3f seconds\n", "max_jitter:", ip->max_jitter);
	fprintf(f, "%-20.20s %.3f seconds\n", "low_zero:", ip->low_zero);
	fprintf(f, "%-20.20s %.3f seconds\n", "high_zero:", ip->high_zero);
	fprintf(f, "%-20.20s T+%.3f seconds\n", "ts_zero:", ip->ts_zero);
	fprintf(f, "%-20.20s %.3f seconds\n", "low_one:", ip->low_one);
	fprintf(f, "%-20.20s %.3f seconds\n", "high_one:", ip->high_one);
	fprintf(f, "%-20.20s T+%.3f seconds\n", "ts_one:", ip->ts_one);
	fprintf(f, "%-20.20s %.0f claims\n", "disproof:", ip->disproof);
	fprintf(f, "%-20.20s %.0f seconds\n", "timeout:", ip->timeout);
	fprintf(f, "%-20.20s %.0f %% noise\n", "squelch:", ip->squelch * 100);
	fprintf(f, "%-20.20s %.0f bits\n", "lockbits:", ip->lockbits);
	fprintf(f, "%-20.20s %.0f bits\n", "untrack:", ip->untrack);
	fprintf(f, "%-20.20s %.6f seconds\n", "delay:", ip->delay);
	fprintf(f, "%-20.20s ", "preamble:");
	for (i = 0; i < NPREAMBLE; i++) {
		if (ip->preamble[i] == -1)
			fprintf(f, "0");
		else if (ip->preamble[i] == 1)
			fprintf(f, "1");
		else 
			fprintf(f, "X");
	}
	fprintf(f, "\n");
}

static int
dcf77_tweak(evContext ctx __unused, FILE *f, void *priv, char **tok, int flag)
{
	char *p;
	int i;
	struct dcf77 *ip;
	double a;

	ip = priv;
	if ((flag & CFG_SHOW)) {
		dcf77_showparam(f, ip);
		return (0);
	}
	if (tok[0] == NULL || (flag & CFG_HELP)) {
		ConfigHelp(f, "dcf77 $instance tweak [preamble $string] [max_jitter $n] [low_zero $n] [high_zero $n] [low_one $n] [high_one $n] [disproof $n] [timeout $n] [squelch $n] [ts_zero $n] [ts_one $n] [lockbits $n] [untrack $n] [delay $n]");
		return (0);
	}
	if (tok[1] == NULL && !(flag & CFG_NO)) {
		fprintf(f, "Specify value for parameter\n");
		return (-1);
	}
	if (!strcmp(tok[0], "preamble")) {
		if (flag & CFG_NO) {
			memcpy(ip->preamble, defaults.preamble, sizeof ip->preamble);
			return (0);
		}
		p = tok[1];
		for (i = 0; i < NPREAMBLE && *p; i++, p++) {
			if (*p == '0')
				ip->preamble[i] = -1;
			else if (*p == '1')
				ip->preamble[i] = 1;
			else if (*p == 'x' || *p == 'X')
				ip->preamble[i] = 0;
			else {
				fprintf(f,
				    "Unknown character in preamble ignored\n");
			}
		}
		return (0);
	}
	errno = 0;
	if (!(flag & CFG_NO)) {
		a = strtod(tok[1], &p);
		if (errno || *p != '\0' || a < 0.0) {
			fprintf(f, "Invalid numerical argument\n");
			return (-1);
		}
	} else {
		a = 0;
	}
#define SET_FOO(foo) do { \
	ip->foo = (flag & CFG_NO) ? defaults.foo : a; \
	} while (0)
	if (!strcmp(tok[0], "max_gap")) {
		SET_FOO(max_gap);
	} else if (!strcmp(tok[0], "propagation_delay")) {
		SET_FOO(propagation_delay);
	} else if (!strcmp(tok[0], "max_jitter"))	{ SET_FOO(max_jitter);
	} else if (!strcmp(tok[0], "low_zero"))		{ SET_FOO(low_zero);
	} else if (!strcmp(tok[0], "high_zero"))	{ SET_FOO(high_zero);
	} else if (!strcmp(tok[0], "low_one"))		{ SET_FOO(low_one);
	} else if (!strcmp(tok[0], "high_one"))		{ SET_FOO(high_one);
	} else if (!strcmp(tok[0], "disproof"))		{ SET_FOO(disproof);
	} else if (!strcmp(tok[0], "timeout"))		{ SET_FOO(timeout);
	} else if (!strcmp(tok[0], "squelch"))		{ SET_FOO(squelch);
	} else if (!strcmp(tok[0], "ts_zero"))		{ SET_FOO(ts_zero);
	} else if (!strcmp(tok[0], "ts_one"))		{ SET_FOO(ts_one);
	} else if (!strcmp(tok[0], "lockbits"))		{ SET_FOO(lockbits);
	} else if (!strcmp(tok[0], "untrack"))		{ SET_FOO(untrack);
	} else if (!strcmp(tok[0], "delay"))		{
		SET_FOO(delay);
		dcf77_change_state(ip, DCF_INIT);
	} else {
		fprintf(f, "Unknown parameter\n");
		return (-1);
	}
#undef SET_FOO
	return (0);
}

static int
dcf77_hist(evContext ctx __unused, FILE *f, void *priv, char **tok __unused, int flag __unused)
{
	struct dcf77 *ip;
	int i, j, k, l, b[56];

	ip = priv;

	if (!(flag & (CFG_SHOW | CFG_NO | CFG_HELP))) {
		if (tok[0] != NULL && !strcmp(tok[0], "zero")) {
			memset(ip->dwhist, 0, sizeof ip->dwhist);
			memset(ip->dthist, 0, sizeof ip->dthist);
			return (0);
		}
	}

	if (tok[0] == NULL || !strcmp(tok[0], "dw")) {
		b[0] = 0;
		b[1] = ip->low_zero / NDWBUCKET;
		b[2] = ip->high_zero / NDWBUCKET;
		b[3] = ip->low_one / NDWBUCKET;
		b[4] = ip->high_one / NDWBUCKET;
		b[5] = NDWHIST;
		fprintf(f, "dW [s] histogram\n");
		fprintf(f, "|%-14s|%-14s|%-14s|%-14s|%-14s|\n",
		    " too short",
		    " zero bit",
		    " in between",
		    " one bit",
		    " too long");
		for (i = 0; i < NDWHIST; i++) {
			j = 0;
			for (k = 0; k < 5; k++) {
				l = b[k] + i;
				if (l >= b[k + 1]) {
					fprintf(f, "|              ");
				} else {
					fprintf(f, "|%.3f %8u",
					     l * NDWBUCKET, ip->dwhist[l]);
					j++;
				}
			}
			fprintf(f, "|\n");
			if (!j)
				break;
		}
		fprintf(f, "\n");
	}
	if (tok[0] == NULL || !strcmp(tok[0], "dt")) {
		fprintf(f, "dT [s] histogram\n");
		j = NDTHIST / 5;
		for (i = 0; i < j; i++) {
			for (k = 0; k < 5; k++) {
				l = i + j * k;
				if (l < NDTHIST)
					fprintf(f, "|%2d %9u",
					    l, ip->dthist[l]);
			}
			fprintf(f, "|\n");
		}
		fprintf(f, "\n");
	}
	return (0);
}


static int
dcf77_pps(evContext ctx, FILE *f, void *priv, char **tok, int flag)
{
	struct dcf77 *ip;

	if (tok[0] == NULL || (flag & CFG_HELP)) {
		ConfigHelp(f, "dcf77 $instance pps {pps_params}");
		return (0);
	}
	ip = priv;
	return (PpsConfig(ctx, f, &ip->pps, tok, flag));
}


static struct dcf77 *
dcf77_instance(evContext ctx, u_int unit, int create)
{
	struct dcf77 *ip;
	char *p;
	int i;

	TAILQ_FOREACH(ip, &dcf77_head, next) {
		if (ip->unit == unit)
			return (ip);
	}
	if (create == 0)
		return (NULL);

	ip = calloc(sizeof *ip, 1);
	if (ip == NULL)
		return (NULL);
	*ip = defaults;
	ip->unit = unit;
	TAILQ_INSERT_TAIL(&dcf77_head, ip, next);
	ip->sid = SourceAdd("dcf77_%lu", ip->unit);
	SourceRefid(ip->sid, "DCF\0");
	SourceParam(ip->sid, SRC_UPDATE_RATE, 1.0);
	SourceParam(ip->sid, SRC_LOS_TIMEOUT, 15.0);
	SourceParam(ip->sid, SRC_VOTES, 1.0);
	PpsCreate(&ip->pps, 2);
	ip->sample_zero = evConsTime(0, ip->ts_zero * 1e9);
	ip->sample_one = evConsTime(0, ip->ts_one * 1e9);
	ip->minute = 0;
	ip->target = -1;
	ip->lasti = -1;
	ip->ts[2].ts.ts = evUTCTime();
	ip->ts[2].fmt = TS_TIMESPEC;
	i = evSetTimer(ctx, dcf77_timer, ip,
	     evConsTime(0, 0), evConsTime(1, 0), &ip->tid);
	if (i < 0)
		err(1, "evSetTimer() failed %s:%d", __FILE__, __LINE__);
	i = evConfigTimer(ctx, ip->tid, "rate", 0);
	if (i < 0)
		err(1, "evConfigTimer() failed %s:%d", __FILE__, __LINE__);
	p = getenv("TZ");
	if (p == NULL || strcmp(p, "Europe/Berlin")) {
		fprintf(stderr, "setting TZ for dcf77\n");
		putenv("TZ=Europe/Berlin");
		tzset();
	}

	dcf77_change_state(ip, DCF_INIT);
	return (ip);
}


static struct configlist *dcf77_cfl;

static int
dcf77_dcf77(evContext ctx, FILE *f, void *priv __unused, char **tok, int flag)
{
	int u;
	char *p;
	struct dcf77 *ip;

	if (tok[0] == NULL) {
		fprintf(f, "Specify DCF77 unit number\n");
		return (-1);
	}
	u = strtoul(tok[0], &p, 0);
	if (*p != '\0') {
		fprintf(f, "DCF77 unit number not numeric\n");
		return (0);
	}
	if (flag & CFG_NO) {
		ip = dcf77_instance(ctx, u, 0);
		if (ip != NULL)
			fprintf(f, "XXX: no dcf77\n");
		return (0);
	}
	ip = dcf77_instance(ctx, u, 1);
	if (ip == NULL)
		return (ENOMEM);
	return (ConfigCall(dcf77_cfl, ctx, f, ip, tok + 1, flag));
}

ref_init_t dcf77_init;

const char *
dcf77_init(evContext ctx __unused)
{
	static int once = 0;


	if (once) { 
		warnx("dcf77 module already loaded");
		return (NULL);
	} else {
		once = 1;
	}
	l_debug = LogNewHandle(Ldebug, "dcf77");
	l_filter = LogNewHandle(l_debug, "filter");
	l_msg = LogNewHandle(l_debug, "msg");
	ConfigAdd(NULL, "dcf77", dcf77_dcf77, "refclock: DCF77 pulses");
	dcf77_cfl = ConfigCreate();
	ConfigAdd(dcf77_cfl, "pps", dcf77_pps, "PPS-API configuration");
	ConfigAdd(dcf77_cfl, "tweak", dcf77_tweak, "Tweak parameters");
	ConfigAdd(dcf77_cfl, "hist", dcf77_hist, "Pulse Histograms");
	return (MAGIC);
}
