/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: config.c,v 1.3 2005/01/26 15:28:53 tom Exp $
 *
 * Parse the configuration and get things going.
 *
 */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <err.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/queue.h>

#include <ntpns.h>
#include "internal.h"

struct config {
	const char		*keyword;
	const char		*help;
	conffunc_t		*func;
	LIST_ENTRY(config)	list;
};

struct configlist {
	LIST_HEAD(, config)	head;
};

static struct configlist *gcfl;

void
ConfigHelp(FILE *f, const char *s)
{

	fprintf(f, "%s\n", s);
}

struct configlist *
ConfigCreate()
{
	struct configlist *cfl;

	cfl = calloc(sizeof *cfl, 1);
	if (cfl == NULL)
		return (cfl);
	LIST_INIT(&cfl->head);
	return (cfl);
}

void
ConfigDestroy(struct configlist *cfl)
{
	struct config *cfg;

	while (!LIST_EMPTY(&cfl->head)) {
		cfg = LIST_FIRST(&cfl->head);
		LIST_REMOVE(cfg, list);
		free(cfg);
	}
	free(cfl);
}

int
ConfigAdd(struct configlist *cfl, const char *kw, conffunc_t *func, const char *help)
{
	struct config *cfg;

	if (cfl == NULL)
		cfl = gcfl;
	cfg = calloc(sizeof *cfg, 1);
	if (cfg == NULL)
		return (-1);
	cfg->keyword = kw;
	cfg->func = func;
	cfg->help = help;
	LIST_INSERT_HEAD(&cfl->head, cfg, list);
	return (0);
}

void
ConfigDel(struct configlist *cfl, const char *kw)
{
	struct config *cfg;

	if (cfl == NULL)
		cfl = gcfl;
	LIST_FOREACH(cfg, &cfl->head, list) {
		if (strcmp(kw, cfg->keyword))
			continue;
		LIST_REMOVE(cfg, list);
		free(cfg);
		return;
	}
	warnx("ConfigDel(%p, %s) found nothing", cfl, kw);
}

int
ConfigCall(struct configlist *cfl, evContext ctx, FILE *f, void *priv, char **tok, int flag)
{
	struct config *cfg;

	if (cfl == NULL)
		cfl = gcfl;
	if (tok[0] == NULL) {
		if (flag & CFG_HELP) {
			LIST_FOREACH(cfg, &cfl->head, list)
				fprintf(f, "%-20.20s %s\n", cfg->keyword, cfg->help);
			return (0);
		}
		fprintf(f, "need more to make complete command\n");
		return (-1);
	}
	LIST_FOREACH(cfg, &cfl->head, list) {
		if (strcmp(tok[0], cfg->keyword))
			continue;
		return(cfg->func(ctx, f, priv, tok + 1, flag));
	}
	fprintf(f, "Failed to find a match for \"%s\"\n", tok[0]);
	return (-1);
}

/*
 * Then a couple of pseudo-magic global keywords
 */

static int
conf_show(evContext ctx, FILE *f, void *priv, char **tok, int flag)
{

	return(ConfigCall(gcfl, ctx, f, priv, tok, flag | CFG_SHOW));
}

static int
conf_no(evContext ctx, FILE *f, void *priv, char **tok, int flag)
{

	return(ConfigCall(gcfl, ctx, f, priv, tok, flag | CFG_NO));
}

static int
conf_help(evContext ctx, FILE *f, void *priv, char **tok, int flag)
{

	return(ConfigCall(gcfl, ctx, f, priv, tok, flag | CFG_HELP));
}

/*
 * Stuff to read a filedescriptor and configure from it
 */

struct cfgsrc;

static LIST_HEAD(, cfgsrc) cfgsrc_head =
	LIST_HEAD_INITIALIZER(cfgsrc_head);

/*
 * This structure contains the status for each
 * of the sources of config info currently in use.
 */
struct cfgsrc {
	LIST_ENTRY(cfgsrc)	next;
	evContext		ctx;
	evFileID		fid;
	FILE			*fout;
	int			flags;
	int			fd;
	char			*buf;
	char			**tok;
	int			ntok;
	int			closing;

	void			*priv;
};

static void
Configure(struct cfgsrc *cf, char *line)
{
	char *p;
	int i, flag;

	flag = 0;
	/* get rid of whitespace, control characters */
	while (*line && !isgraph(*line))
		line++;
	p = strchr(line, '\0');
	if (p != NULL)
		for (p--; p >= line && *p && !isgraph(*p); p--)
			*p = '\0';

	/* ignore empty lines and comments */
	if (!*line)
		return;
	if (*line == '#')
		return;

	/* tokenize */
	if (cf->tok == NULL) {
		cf->ntok = 8;
		cf->tok = malloc(sizeof *cf->tok * cf->ntok);
	}
	for (i = 0; ; i++) {
		if (i == cf->ntok - 1) {
			cf->ntok *= 2;
			cf->tok = realloc(cf->tok, sizeof *cf->tok * cf->ntok);
		}
		while(line != NULL && (*line == ' ' || *line == '\t'))
			line++;
		cf->tok[i] = strsep(&line, " \t");
		if (cf->tok[i] == NULL)
			break;
	}

	/* If the last token is a question mark, treat this as a help request */
	if (!strcmp(cf->tok[i - 1], "?")) {
		cf->tok[i - 1] = NULL;
		flag |= CFG_HELP;
	}

	cf->priv = NULL;
	i = ConfigCall(NULL, cf->ctx, cf->fout, &cf->priv, cf->tok, flag);
	if (i > 0)
		fprintf(cf->fout, "Error: %s\n", strerror(i));
	return;
}

static void
Linify(struct cfgsrc *cf, char *buf)
{
	char *p, *s, *q;

	p = buf;
	if (cf->buf != NULL) {
		/* We had leftovers */
		p = realloc(cf->buf, strlen(cf->buf) + strlen(buf) + 1);
		if (p == NULL) {
			cf->buf = NULL;
			free(buf);
			return;
		}
		strcat(p, buf);
		free(buf);
		cf->buf = NULL;
	}
	q = p;
	for(;;) {
		s = strchr(q, '\n');
		if (s != NULL) {
			*s++ = '\0';
			Configure(cf, q);
			if (cf->flags & CONFIG_PROMPT)
				fprintf(cf->fout, "%d> ", cf->fd);
			fflush(cf->fout);
		} else {
			break;
		}
		q = s;
	}
	if (strlen(q)) {
		/* leave leftovers */
		cf->buf = strdup(q);
	}
	free(p);
}


static void
EvConfig(evContext ctx, void *uap, int fd, int eventmask __unused)
{
	char *buf;
	int i;
	struct cfgsrc *cp;

	cp = uap;

	buf = malloc(BUFSIZ + 1);
	if (buf != NULL) {
		i = read(fd, buf, BUFSIZ);
		if (i > 0) {
			buf[i] = '\0';
			Linify(cp, buf);
			buf = NULL;
			if (!cp->closing)
				return;
		}
	}
	i = evDeselectFD(ctx, cp->fid);
	if (i)
		err(1, "evDeselectFD(stdin)");
	debug(1, "Closing fd %d\n", fd);
	log_close(fd);
	if (fileno(cp->fout) == cp->fd)
		fclose(cp->fout);
	else
		close(fd);
	if (cp != NULL) {
		LIST_REMOVE(cp, next);
		if (cp->buf != NULL)
			free(cp->buf);
		if (cp->tok != NULL)
			free(cp->tok);
		free(cp);
	}
	if (buf != NULL)
		free(buf);
#if 0
	/* XXX: this doesn't work because we (may) already have threads */
	if (fd == 0)
		daemon(1, 1);
#endif
}

void
ConfigSource(evContext ctx, FILE *fo, FILE *fi, int fd, int flags)
{
	struct cfgsrc *cp;
	int i;

	if (fd < 0) {
		if (fi == NULL)
			err(1, "must specify something to ConfigSource");
		fd = fileno(fi);
	}
	if (fi == NULL && fo == NULL)
		fo = fi = fdopen(fd, "r+");
	else if (fi == NULL)
		fi = fdopen(fd, "r");
	if (fd < 0 || fi == NULL)
		err(1, "ConfigSource() giving up");
	
	cp = calloc(sizeof *cp, 1);
	if (cp == NULL) {
		nomem();
		fclose(fi);
		return;
	}
	cp->fd = fd;
	cp->fout = fo;
	cp->flags = flags;
	cp->ctx = ctx;
	LIST_INSERT_HEAD(&cfgsrc_head, cp, next);
	i = evSelectFD(ctx, fd, EV_READ | EV_EXCEPT, EvConfig, cp, &cp->fid);
	if (i) {
		fclose(fi);
		err(1, "evSelectFD()");
		return;
	}
	if (cp->flags & CONFIG_PROMPT) {
		fprintf(cp->fout, "%d> ", cp->fd);
		fflush(cp->fout);
	}
}

static int
conf_file(evContext ctx, FILE *f, void *priv __unused, char **tok, int flag)
{
	FILE *fi;

	if (tok[0] == NULL || (flag & CFG_HELP)) {
		ConfigHelp(f, "file $filename");
		return (0);
	}
	if (flag & CFG_NO)
		return (0);
	if (flag & CFG_SHOW)
		return (0);
	fi = fopen(tok[0], "r");
	if (fi == NULL) {
		fprintf(f, "open(%s): %s\n", tok[0], strerror(errno));
		return (-1);
	}
	ConfigSource(ctx, f, fi, -1, 0);
	return (0);
}

static int
conf_close(evContext ctx __unused, FILE *f, void *priv __unused, char **tok __unused, int flag __unused)
{
	struct cfgsrc *cp;

	if (f == stderr)
		return (1);
	LIST_FOREACH(cp, &cfgsrc_head, next)
		if (cp->fout == f)
			break;
	if (cp != NULL)
		cp->closing = 1;

	return (0);
}

void
init_config(evContext ctx __unused)
{

	gcfl = ConfigCreate();
	ConfigAdd(gcfl, "show", conf_show, "display information");
	ConfigAdd(gcfl, "list", conf_show, "display information");
	ConfigAdd(gcfl, "no", conf_no, "negate command");
	ConfigAdd(gcfl, "help", conf_help, "display help information");
	ConfigAdd(gcfl, "file", conf_file, "read config from file");
	ConfigAdd(gcfl, "close", conf_close, "close interactive session");
}

