/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: log.c,v 1.5 2005/01/26 15:28:53 tom Exp $
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <err.h>
#include <errno.h>
#include <syslog.h>
#include <stdarg.h>
#include <string.h>
#include <assert.h>

#include "ntpns.h"
#include "internal.h"

struct logh *Lerror, *Lwarn, *Linfo, *Ldebug;

static int term_fd = -1;

struct logh {
	LIST_ENTRY(logh)	next;
	LIST_HEAD(, logh)	list;
	const char		*name;
	struct logh		*parent;
	u_int			flags;
};

/* Bottom three bits are for syslog priority */
#define LF_ENABLE		(1<<3)
#define LF_SYSLOG		(1<<4)
#define LF_STDOUT		(1<<5)
#define LF_STDERR		(1<<6)
#define LF_TERM			(1<<7)
#define LF_CLOCKSTATS		(1<<8)

static struct logh		top;

#if 0
struct log {
	STAILQ_ENTRY(log)	next;
	u_int			flags;
	char			*text;
};

static STAILQ_HEAD(, log) loglist = STAILQ_HEAD_INITIALIZER(loglist);
#endif

static FILE *fpclockstats = NULL;
#define MJD_1970          40587   /* MJD for 1 Jan 1970 */

struct logh *
LogNewHandle(struct logh *pp, const char *name)
{
	struct logh *hp;

	if (pp == NULL)
		pp = &top;
	hp = calloc(sizeof *hp, 1);
	if (hp != NULL) {
		hp->name = name;
		LIST_INSERT_HEAD(&pp->list, hp, next);
		LIST_INIT(&hp->list);
		hp->parent = pp;
		hp->flags = pp->flags;
	}
	return (hp);
}

void
LogDeleteHandle(struct logh *hp)
{
	struct logh *ep;

	while (!LIST_EMPTY(&hp->list)) {
		ep = LIST_FIRST(&hp->list);
		LIST_REMOVE(ep, next);
		LogDeleteHandle(ep);
	} 
	free(hp);
}


void
LogErr(struct logh *lh, const char *fmt, ...)
{
	va_list ap;
	char *p, *e;

	e = strerror(errno);

	if (!(lh->flags & LF_ENABLE))
		return;
	va_start(ap, fmt);
	vasprintf(&p, fmt, ap);
	va_end(ap);
	if (p == NULL)
		return;		/* XXX: count one missing */
	if (lh->flags & LF_SYSLOG)
		syslog(lh->flags & LOG_PRIMASK, "%s: %s", p, e);
	if (lh->flags & LF_STDOUT) {
		fprintf(stdout, "%s: %s\n", p, e);
		fflush(stdout);
	}
	if (lh->flags & LF_STDERR) {
		fprintf(stderr, "%s: %s\n", p, e);
		fflush(stderr);
	}
	if (lh->flags & LF_CLOCKSTATS) {
		if(fpclockstats) {
			fprintf(fpclockstats, "%s: %s\n", p, e);
		       	fflush(fpclockstats);
		}
	}
	free(p);
}

int
LogActive(struct logh *lh)
{

	return (lh->flags & LF_ENABLE);
}

static void LogDo(struct logh *lh, int do_newline, const char *fmt, va_list args)
{
	char *p;

	if (!(lh->flags & LF_ENABLE))
		return;
	vasprintf(&p, fmt, args);
	if (p == NULL)
		return;		/* XXX: count one missing */
	if (lh->flags & LF_SYSLOG)
		syslog(lh->flags & LOG_PRIMASK, p);
	if (lh->flags & LF_STDOUT) {
		fprintf(stdout, "%s", p);
		if(do_newline)
			fprintf(stdout, "\n");
		fflush(stdout);
	}
	if (lh->flags & LF_STDERR) {
		fprintf(stderr, "%s", p);
		if(do_newline)
			fprintf(stderr, "\n");
		fflush(stderr);
	}
	if (lh->flags & LF_CLOCKSTATS) {
		if(fpclockstats) {
			if(do_newline)
			{
				// get time, log line header
				struct timespec ts;
				clock_gettime(CLOCK_REALTIME, &ts);
				fprintf(fpclockstats, "%u %.3f 127.127.0.1 ", 
						ts.tv_sec/86400+MJD_1970,
						ts.tv_sec%86400+
						ts.tv_nsec/1000.0/1000.0/1000.0);
			}
			fprintf(fpclockstats, "%s", p);
			if(do_newline)
				fprintf(fpclockstats, "\n");
			fflush(fpclockstats);
		}
	}
	if ((lh->flags & LF_TERM) && term_fd > -1) {
		write(term_fd, p, strlen(p));
		if(do_newline)
			write(term_fd, "\n", 1);
	}
	free(p);
}

void Log(struct logh *lh, const char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
   LogDo(lh, 1, fmt, ap);
	va_end(ap);
}

void LogNoNewline(struct logh *lh, const char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
   LogDo(lh, 0, fmt, ap);
	va_end(ap);
}

void
LogAssert(const char *file, const char *func, int line, const char *fmt, ...)
{
	va_list ap;
	char *p, *q, *e;
	int i;

	i = errno;
	e = strerror(i);
	va_start(ap, fmt);
	vasprintf(&p, fmt, ap);
	va_end(ap);

	asprintf(&q, "Assertion failed\n%s:%d %s(): %s\nerrno = %d (%s)",
	     file, line, func, p, i, e);
	syslog(LOG_EMERG, "%s", q);
	fprintf(stderr, "%s", q);
	fflush(stderr);
	fprintf(stdout, "%s", q);
	fflush(stdout);
	abort();
}

static void
log_bits_recurse(struct logh *hp, u_int reset, u_int set)
{
	struct logh *ep;

	LIST_FOREACH(ep, &hp->list, next)
		log_bits_recurse(ep, reset, set);
	hp->flags &= ~reset;
	hp->flags |= set;
}

static void
log_show_recurse(struct logh *hp, FILE *f, char *prefix)
{
	struct logh *ep;
	char *p;

	if (hp == NULL)
		return;
	p = prefix + strlen(prefix);
	strcat(p, hp->name);
	strcat(p, " ");
	fprintf(f, "%-30.30s", prefix);
	if (hp->flags & LF_ENABLE)
		fprintf(f, " enable");
	if (hp->flags & LF_STDERR)
		fprintf(f, " stderr");
	if (hp->flags & LF_STDOUT)
		fprintf(f, " stdout");
	if (hp->flags & LF_SYSLOG)
		fprintf(f, " syslog(%d)", hp->flags & LOG_PRIMASK);
	if (hp->flags & LF_TERM)
		fprintf(f, " term");
	if (hp->flags & LF_CLOCKSTATS)
		fprintf(f, " clockstats");
	fprintf(f, "\n");
	LIST_FOREACH(ep, &hp->list, next)
		log_show_recurse(ep, f, prefix);
	*p = '\0';
}

static struct logh *
log_find_recurse(struct logh *hp, char **tok)
{
	struct logh *ep;

	if (tok[0] == NULL)
		return (hp);
	LIST_FOREACH(ep, &hp->list, next)
		if (!strcmp(ep->name, tok[0]))
			return (log_find_recurse(ep, tok + 1));
	return (NULL);
}


static int
log_msg(evContext ctx __unused, FILE *f, void *priv __unused, char **tok, int flag)
{
	char buf[1024];
	struct logh *hp;
	u_int flag1, flag2;

	if (tok[0] == NULL || (flag & CFG_HELP)) {
		ConfigHelp(f, "msg [enable] [term] [stdout] [stderr] [syslog] [clockstats] $msg_tree");
		return (0);
	}
	if ((flag & CFG_SHOW)) {
		hp = log_find_recurse(&top, tok);
		if (hp == NULL) {
			fprintf(f, "Not found\n");
		} else {
			*buf = '\0';
			log_show_recurse(hp, f, buf);
		}
		return (0);
	}
	flag1 = flag2 = 0;
	for (; tok[0] != NULL; tok++) {
		if (!strcmp(tok[0], "enable")) {
			flag1 |= LF_ENABLE;
			continue;
		}
		if (!strcmp(tok[0], "stdout")) {
			flag1 |= LF_STDOUT;
			continue;
		}
		if (!strcmp(tok[0], "stderr")) {
			flag1 |= LF_STDERR;
			continue;
		}
		if (!strcmp(tok[0], "syslog")) {
			flag1 |= LF_SYSLOG;
			continue;
		}
		if (!strcmp(tok[0], "term")) {
			flag1 |= LF_TERM;
			term_fd = fileno(f);
			continue;
		}
		if (!strcmp(tok[0], "clockstats")) {
			flag1 |= LF_CLOCKSTATS;
			continue;
		}
		break;
	}
	hp = log_find_recurse(&top, tok);
	if (hp == NULL) {
		fprintf(f, "Not found\n");
		return (0);
	}
	if ((flag & CFG_NO) && flag1 == 0)
		flag1 = LF_ENABLE;
	if (flag & CFG_NO)
		log_bits_recurse(hp, flag1, flag2);
	else
		log_bits_recurse(hp, flag2, flag1);
	return (0);
}

void
log_close(int fd)
{

	if (fd == term_fd) {
		term_fd = -1;
		log_bits_recurse(&top, LF_TERM, 0);
	}
}


void
init_log(evContext ctx __unused)
{

	LIST_INIT(&top.list);
	top.name = "";

	openlog("NTPns", LOG_NDELAY, LOG_DAEMON);
	Lerror = LogNewHandle(NULL, "error");
	Lerror->flags |= LF_ENABLE | LF_SYSLOG | LF_STDERR | LOG_ERR;
	Lwarn = LogNewHandle(NULL, "warning");
	Lwarn->flags |= LF_ENABLE | LF_SYSLOG | LF_STDERR | LOG_WARNING;
	Linfo = LogNewHandle(NULL, "info");
	Linfo->flags |= LF_ENABLE | LF_SYSLOG | LF_STDOUT | LOG_INFO;
	Ldebug = LogNewHandle(NULL, "debug");
	Ldebug->flags |= LF_SYSLOG | LF_STDOUT | LOG_DEBUG;

	fpclockstats=fopen("/var/log/xntp/clockstats", "a");
	if(fpclockstats==NULL)
	{
		fprintf(stderr, "warning: can't open /var/log/xntp/clockstats: %s\n",strerror(errno));
	}

	ConfigAdd(NULL, "msg", log_msg, "message control functions");
}
