/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: pll.c,v 1.5 2005/01/26 15:28:53 tom Exp $
 *
 * PLL handling.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <err.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/queue.h>
#include <sys/timex.h>
#include <math.h>

#include "ntpns.h"
#include "pll.h"
#include "internal.h"

struct logh *L_pll;

int clock_step;
int pll_locked;

static void
dump_timex(struct timex *tx)
{

	Log(L_pll, 
	 "modes %x status %x offset %ld freq %.3e maxerror %ld esterror %ld constant %ld prec %ld tol %ld",
	    tx->modes, tx->status, tx->offset, tx->freq / (65536.0 * 1e6),
	    tx->maxerror, tx->esterror, tx->constant, tx->precision,
	    tx->tolerance);
}

void
kern_freq(double f)
{
	struct timex tx;
	int i;
	static double acc = 0;
	static FILE *fp=NULL;

	if (Dopt)
		return;

	memset(&tx, 0, sizeof tx);
	tx.modes = MOD_NANO | MOD_FREQUENCY | MOD_OFFSET | MOD_STATUS;
	tx.status = STA_PLL;

	/*
	 * To avoid bang-bang regulation at high integration times we
	 * implement a simple Floyd-Steinberg dither of the frequency
	 * correction:  Whatever fraction we did not apply carries over
	 * to the next second.  We also need to make sure that we do
	 * not get in trouble with the +/- 500PPM hard limits of frequency
	 * adjustability.
	 */
	acc += f * (65536.0 * 1e6);
	tx.freq = round(acc);
	if (tx.freq > 500 * 65536)
		acc = tx.freq = 500 * 65536;
	if (tx.freq < -500 * 65536)
		acc = tx.freq = -500 * 65536;
	acc -= tx.freq;

#if 0
	/* XXX: why does this cause divergence ?? */
	if (pll_locked) {
		tx.modes = MOD_MAXERROR | MOD_ESTERROR;
		tx.maxerror = sys_dispersion * 1e-9;
		tx.esterror = sys_dispersion * 1e-9;
	} else {
		tx.modes = MOD_MAXERROR | MOD_ESTERROR;
		tx.maxerror = 1e9;
		tx.esterror = 1e9;
	}
#endif
	i = ntp_adjtime(&tx);
	if (i != 0)
		LogErr(Lwarn, "ntp_adjtime()");
	dump_timex(&tx);

	if(fp==NULL)
	{
		fp=fopen("/tmp/freq.log", "w");
	}
	if(fp!=NULL)
	{
		fprintf(fp, "%f\n", tx.freq / (65536.0 * 1e6));
		fflush(fp);
	}
}

void
pllstep(double d)
{
	struct timespec ts;
	int i;

	if (Dopt)
		return;
	d = -d;
	Log(Linfo, "PLL step %.09f", d);
	if (d < 0) {
		d = -d;
		ts.tv_sec = d;
		d -= ts.tv_sec;
		ts.tv_nsec = d * 1e9;
		ts = evSubTime(evUTCTime(), ts);
		i = clock_settime(CLOCK_REALTIME, &ts);
	} else {
		ts.tv_sec = d;
		d -= ts.tv_sec;
		ts.tv_nsec = d * 1e9;
		ts = evAddTime(evUTCTime(), ts);
		i = clock_settime(CLOCK_REALTIME, &ts);
	}
	if (i != 0)
		LogErr(Lwarn, "clock_settime()");
	clock_step++;
}

void
pll(double y)
{

	kern_freq(pll_math(y));
}

void
pllnone()
{

	kern_freq(pll_poll());
}

void
init_pll(evContext ctx __unused)
{

	L_pll = LogNewHandle(Ldebug, "pll");
}
