/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: pps-api.c,v 1.5 2005/01/26 15:28:53 tom Exp $
 *
 * This is the main entrance where we set up the event engine, parse the
 * options and set the ball rolling by starting events to read the config
 * file.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <err.h>
#include <fcntl.h>
#include <errno.h>

#include "ntpns.h"
#include "sys/timepps.h"

struct ppsdev {
	int		fd;
	char		*name;
	pps_handle_t	handle;
	int		flank;
	int		flanks;
	int		mode;
	pps_params_t	params;
	pps_info_t	event;
	FILE		*simulate;
};

int
PpsCreate(struct ppsdev **ppp, int flanks)
{
	struct ppsdev *pp;

	if (flanks < 1 || flanks > 2)
		err(1, "Wrong number of flanks to PpsCreate: %d", flanks);

	if (*ppp != NULL)
		return (0);

	pp = calloc(sizeof *pp, 1);
	if (pp == NULL)
		return (ENOMEM);

	pp->fd = -1;
	pp->flanks = flanks;
	*ppp = pp;
	return (0);
}


int
PpsConfig(evContext ctx __unused, FILE *f, void *priv, char **tok, int flag)
{
	struct ppsdev *pp, **ppp;
	int i, j, flank, fd;
	int mode, nmode;
	pps_handle_t	nhandle;
	char *p;
	struct timespec to;

	if (tok[0] == NULL || flag & CFG_HELP) {
		ConfigHelp(f, "{pps_params} = device $filename [assert] [clear] [echoassert] [echoclear]");
	}
	ppp = priv;
	pp = *ppp;
	if (pp == NULL)
		err(1, "PpsConfig on NULL");
	if (flag & CFG_SHOW) {
		fprintf(f, "%-20.20s %s\n", "device:", pp->name);
		fprintf(f, "%-20.20s %d\n", "flanks:", pp->flanks);
		fprintf(f, "%-20.20s %s (%d)\n", "epoch:",
		    pp->flank == PPS_CAPTURECLEAR ? "clear" : "assert",
		    pp->flank);
		return (0);
	}
	if (flag) {
		fprintf(f, "XXX: pps_config flag %d\n", flag);
		return (-1);
	}
	mode = 0;
	flank = pp->flank;
	if (pp->flanks == 2)
		mode |= PPS_CAPTUREBOTH;
	for (i = 0; tok[i] != NULL; i++) {
		if (!strcmp(tok[i], "device")) {
			i++;
			if (tok[i] == NULL) {
				fprintf(f, "missing device argument to pps\n");
				return (-1);
			}
			p = strdup(tok[i]);
			if (p == NULL) 
				return (ENOMEM);
			fd = open(p, O_RDONLY | O_NDELAY);
			if (fd < 0) {
				fprintf(f, "cannot open %s: %s", p, strerror(errno));
				free(p);
				return (-1);
			}
			j = time_pps_create(fd, &nhandle);
			if (j < 0) {
				fprintf(f, "%s: does not support PPS-API.", p);
				close(fd);
				free(p);
				return (-1);
			}
			j = time_pps_getcap(nhandle, &nmode);
			if (j < 0) {
				fprintf(f, "%s: time_pps_getcap() failed", p);
				time_pps_destroy(nhandle);
				close(fd);
				free(p);
				return (-1);
			}
			if (pp->flanks == 2 && !(nmode & PPS_CAPTUREASSERT)) {
				fprintf(f,
				    "%s: device can not capture assert flank.", p);
				time_pps_destroy(nhandle);
				close(fd);
				free(p);
				return (-1);
			}
			if (pp->flanks == 2 && !(nmode & PPS_CAPTURECLEAR)) {
				fprintf(f,
				     "%s: device can not capture clear flank.", p);
				time_pps_destroy(nhandle);
				close(fd);
				free(p);
				return (-1);
			}
			/* we're committed now */
			if (pp->fd >= 0)
				time_pps_destroy(pp->handle);
			if (pp->name != NULL) {
				free(pp->name);
				close(pp->fd);
			}
			pp->name = p;
			pp->fd = fd;
			pp->handle = nhandle;
			pp->mode = nmode;
			j = time_pps_getparams(pp->handle, &pp->params);
			if (j < 0) {
				fprintf(f, "%s: time_pps_getparams() failed",
				    pp->name);
				return (-1);
			}
			continue;
		} 

		if (!strcmp(tok[i], "echoassert")) {
			if (!(pp->mode & PPS_ECHOASSERT)) {
				fprintf(f, "%s: device can not echo assert flank.",
				    pp->name);
				return (-1);
			}
			mode |= PPS_ECHOASSERT;
			continue;
		}

		if (!strcmp(tok[i], "echoclear")) {
			if (!(pp->mode & PPS_ECHOCLEAR)) {
				fprintf(f, "%s: device can not echo clear flank.",
				    pp->name);
				return (-1);
			}
			mode |= PPS_ECHOCLEAR;
			continue;
		}

		/* XXX: offsets */

		if (!strcmp(tok[i], "assert")) {
			if (!(pp->mode & PPS_CAPTUREASSERT)) {
				fprintf(f, "%s: device can not capture assert flank.",
				    pp->name);
				return (-1);
			}
			mode |= PPS_CAPTUREASSERT;
			flank = PPS_CAPTUREASSERT;
			continue;
		}
		if (!strcmp(tok[i], "clear")) {
			if (!(pp->mode & PPS_CAPTURECLEAR)) {
				fprintf(f, "%s: device can not capture clear flank.",
				    pp->name);
				return (-1);
			}
			mode |= PPS_CAPTURECLEAR;
			flank = PPS_CAPTURECLEAR;
			continue;
		}
		if (!strcmp(tok[i], "simulate")) {
			i++;
			if (tok[i] == NULL) {
				fprintf(f, "missing simulate argument to pps\n");
				return (-1);
			}
			pp->simulate = fopen(tok[i], "r");
			if (pp->simulate == NULL) {
				Warn(f, errno,
				    "could not open pps simulate file %s\n",
				    tok[i]);
			}
		}
		fprintf(f, "unrecognized: \"%s\"\n", tok[i]);
		return (-1);
	}
	pp->params.mode |= mode;
	if ((pp->params.mode & PPS_CAPTUREBOTH) == 0)
		pp->params.mode |= (pp->mode & PPS_CAPTUREBOTH);
	if (flank == 0)
		flank = PPS_CAPTUREASSERT;
	pp->flank = flank;
	j = time_pps_setparams(pp->handle, &pp->params);
	if (j < 0) {
		fprintf(f, "%s: time_pps_setparams() failed", pp->name);
		return (-1);
	}
	to = evConsTime(0, 0);
	i = time_pps_fetch(pp->handle, PPS_TSFMT_TSPEC, &pp->event, &to);
	if (i < 0) {
		fprintf(f, "%s: time_pps_fetch() failed", pp->name);
		return (-1);
	}
	return (0);
}

/*
 * Pick up a PPS-API measurement.
 *
 * The flank argument determines how many flanks we want to see change,
 * this us useful for things like DCF77/MSF/WWWB receivers which want
 * to measure the duration of the pulses.
 */

int
PpsPickup(struct ppsdev **pp, struct ts *ts, int flanks)
{
	pps_info_t ev;
	struct ppsdev *p;
	struct timespec to;
	struct ts *t1;
	int i;
	char simbuf[200];

	p = *pp;
	if (p == NULL) {
		Log(Lerror, "PpsPickup(NULL)");
		return (0);
	}
	if (p->simulate != NULL) {
		if (NULL == fgets(simbuf, sizeof simbuf, p->simulate)) {
			Log(Lerror, "EOF on pps simulate file");
			exit (1);
		}
		i = sscanf(simbuf, "%u .%ld %u %u .%ld %u",
		    &ev.assert_timestamp.tv_sec,
		    &ev.assert_timestamp.tv_nsec,
		    &ev.assert_sequence,
		    &ev.clear_timestamp.tv_sec,
		    &ev.clear_timestamp.tv_nsec,
		    &ev.clear_sequence);
		if (i != 6) {
			Log(Lerror, "format error in pps simulate file");
			return (-1);
		}
		i = 0;
	} else {
		to = evConsTime(0, 0);
		i = time_pps_fetch(p->handle, PPS_TSFMT_TSPEC, &ev, &to);
	}
if (0)
fprintf(stderr, "PPS Raw: %d %d.%09ld %u -> %d.%09ld %u\n", i,
    ev.assert_timestamp.tv_sec, ev.assert_timestamp.tv_nsec, ev.assert_sequence,
    ev.clear_timestamp.tv_sec, ev.clear_timestamp.tv_nsec, ev.clear_sequence);
	if (i < 0) {
		warn("time_pps_fetch() failed");
		return (-1);
	}
	if (flanks == 1 && p->flank == PPS_CAPTUREASSERT) {
		if (ev.assert_sequence == p->event.assert_sequence)
			return (0);
		ts->ts.ts = ev.assert_timestamp;
		ts->fmt = TS_TIMESPEC;
		p->event = ev;
		return (p->flanks);
	}
	if (flanks == 1 && p->flank == PPS_CAPTURECLEAR) {
		if (ev.clear_sequence == p->event.clear_sequence)
			return (0);
		ts->ts.ts = ev.clear_timestamp;
		ts->fmt = TS_TIMESPEC;
		p->event = ev;
		return (p->flanks);
	}
	t1 = ts + 1;
	if (p->flank == PPS_CAPTUREASSERT) {
		if (ev.clear_sequence == p->event.clear_sequence)
			return (0);
		ts->ts.ts = ev.assert_timestamp;
		ts->fmt = TS_TIMESPEC;
		t1->ts.ts = ev.clear_timestamp;
		t1->fmt = TS_TIMESPEC;
	} else {
		if (ev.assert_sequence == p->event.assert_sequence)
			return (0);
		ts->ts.ts = ev.clear_timestamp;
		ts->fmt = TS_TIMESPEC;
		t1->ts.ts = ev.assert_timestamp;
		t1->fmt = TS_TIMESPEC;
	}
	if (TsDiff(t1, ts) < .0)
		return (0);
if (0)
fprintf(stderr, "PPS  %d.%09ld %u -> %d.%09ld %u\n",
    ev.assert_timestamp.tv_sec, ev.assert_timestamp.tv_nsec, ev.assert_sequence,
    ev.clear_timestamp.tv_sec, ev.clear_timestamp.tv_nsec, ev.clear_sequence);
	p->event = ev;
	return (p->flanks);
}
