/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: record.c,v 1.3 2005/01/26 15:28:53 tom Exp $
 *
 * "Record" is for dumping diagnostic and debugging info to a remote machine.
 * The way it works is that we wrap the data in a UDP packet and kicks it
 * off without looking back.  The receiving gets to have fun decoding things.
 *
 * Packet format:
 *	BE32		sequence number
 *	C[8]		ident
 *	BE32		timestamp (time_t)
 *	BE16		unitno
 *	BE16		length;
 *	C[...]		data
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <unistd.h>
#include <ctype.h>
#include <stdarg.h>
#include <err.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/queue.h>
#include <sys/socket.h>
#include <sys/endian.h>
#include <sys/resource.h>
#include <math.h>
#include <netdb.h>

#include "ntpns.h"
#include "internal.h"

static int recsock = -1;
static uint32_t seq = 0;
static struct sockaddr sa;
static int salen;

static struct sockaddr rrrd_sa;
static int rrrd_salen;
static char *rrrd_host;
static char *rrrd_port;
static char *rrrd_prefix;

static evTimerID resource_tid;

static void
resource_timer(evContext ctx __unused, void *uap __unused, struct timespec t0 __unused, struct timespec dt __unused)
{
	int i;
	struct rusage r;
	uintptr_t up;

	up = (uintptr_t)sbrk(0);
	i = getrusage(0, &r);
	Assert(i == 0, "getrusage()");
	rrrdupdate("rusage", "brk:sbrk:utime:stime:maxrss:ixrss:idrss:isrss:minflt:majflt:nswap:inblock:oublock:msgsnd:msgrcv:nsignals:nvcsw:nivcsw",
	    "N:%ju:%ju:%d.%06d:%d.%06d:%ld:%ld:%ld:%ld:%ld:%ld:%ld:%ld:%ld:%ld:%ld:%ld:%ld:%ld",
		(uintmax_t)up, (uintmax_t)up,
		r.ru_utime.tv_sec, r.ru_utime.tv_usec,
		r.ru_stime.tv_sec, r.ru_stime.tv_usec,
		r.ru_maxrss, r.ru_ixrss, r.ru_idrss,
		r.ru_isrss, r.ru_minflt, r.ru_majflt,
		r.ru_nswap, r.ru_inblock, r.ru_oublock,
		r.ru_msgsnd, r.ru_msgrcv, r.ru_nsignals,
		r.ru_nvcsw, r.ru_nivcsw);
}

void
record(const char *ident, int unit, void *p, int len)
{
	u_char hd[20];
	time_t t;
	struct msghdr h;
	struct iovec v[2];
	int i;

	seq++;
	if (salen == 0)
		return;

	if (len > 20000)
		len = 20000;
	memset(&h, 0, sizeof h);

	time(&t);
	be32enc(hd, seq);
	strncpy(hd + 4, ident, 8);
	be32enc(hd + 12, t);
	be16enc(hd + 16, unit);
	be16enc(hd + 18, len);

	v[0].iov_base = hd;
	v[0].iov_len = 20;
	v[1].iov_base = p;
	v[1].iov_len = len;

	h.msg_name = &sa;
	h.msg_namelen = salen;
	h.msg_iov = v;
	h.msg_iovlen = 2;

	i = sendmsg(recsock, &h, 0);
	if (i < 0)
		warn("sendmsg (%d, %p, %d)", recsock, &h, 0);
}

int
rrrdupdate(const char *file, const char *template, const char *fmt, ...)
{
	int error;
	char *buf1, *buf2;
	va_list ap;

	if (rrrd_salen == 0)
		return (-1);
	if (fmt == NULL) 
		return (-1);

	va_start(ap, fmt);
	error = vasprintf(&buf1, fmt, ap);
	va_end(ap);
	if (error == -1 || buf1 == NULL)
		return (-1);
	error = asprintf(&buf2, "00 %s%s %s %s",
	    rrrd_prefix, file, template, buf1);
	if (error == -1 || buf2 == NULL)
		return (-1);
	free(buf1);
	error = sendto(recsock, buf2, strlen(buf2), 0, &rrrd_sa, rrrd_salen);
	free(buf2);
	return (error);
}


static void
getsocket(FILE *f)
{
	int s, i, j;
	struct sockaddr lsa;

	s = socket(AF_INET, SOCK_DGRAM, 0); 
	if (s < 0) {
		fprintf(f, "socket failed for record: %s", strerror(errno));
		return;
	}
	i = fcntl(s, F_GETFL, &j);
	if (i < 0) {
		fprintf(f, "fcntl F_GETFL record socket: %s", strerror(errno));
		close (s);
		return;
	}
	j |= O_NONBLOCK;
	i = fcntl(s, F_SETFL, &j);
	if (i < 0) {
		fprintf(f, "fcntl F_SETFL record socket: %s", strerror(errno));
		close (s);
		return;
	}
	
	memset(&lsa, 0, sizeof lsa);
	lsa.sa_len = sizeof lsa;
	lsa.sa_family = AF_INET;
	i = bind(s, &lsa, sizeof lsa);
	if (i != 0) {
		fprintf(f, "bind failed on record socket: %s",
		    strerror(errno));
		close (s);
		return;
	}
	recsock = s;
}

static void
resolved(evContext ctx __unused, void *priv __unused, int error, struct addrinfo *ai, FILE *f)
{

	if (error) {
		fprintf(f, "failed to resolve address for record (%d %s)\n", 
		    error, gai_strerror(error));
		return;
	}

	sa = *ai->ai_addr;
	salen = ai->ai_addrlen;
}


static int
record_record(evContext ctx __unused, FILE *f, void *priv __unused, char **tok, int flag)
{
	int i;

	/* XXX: Fix to "record host $host [port $port]" */
	if (flag & CFG_HELP) {
		ConfigHelp(f, "record <host> [<port>]");
		return (0);
	}
	if (flag & CFG_NO) {
		salen = 0;
		return (0);
	}
	if (flag & CFG_SHOW) {
		fprintf(f, "XXX: record SHOW\n");
		return (0);
	}
	if (tok[0] == NULL) {
		fprintf(f, "need dst-host argument\n");
		return (0);
	}
	if (recsock < 0)
		getsocket(f);
	if (recsock < 0)
		return (-1);
	i = resolver(ctx, tok[0], (tok[1] == NULL) ? "6778" : tok[1],
	     PF_INET, SOCK_DGRAM, AI_CANONNAME, resolved, NULL, f);
	if (i < 0)
		fprintf(f, "Failed...\n");

	return (0);
}

static void
resolved_rrrd(evContext ctx __unused, void *priv __unused, int error, struct addrinfo *ai, FILE *f)
{

	if (error) {
		fprintf(f, "failed to resolve address for record (%d %s)\n", 
		    error, gai_strerror(error));
		return;
	}

	rrrd_sa = *ai->ai_addr;
	rrrd_salen = ai->ai_addrlen;
}

static int
record_rrrd(evContext ctx __unused, FILE *f, void *priv __unused, char **tok, int flag)
{
	int i;

	/* XXX: the $bool os bogus (and unimplemented) when we have "no" */

	if (tok[0] == NULL || (flag & CFG_HELP)) {
		ConfigHelp(f, "rrrd [host $host] [port $port] [prefix $prefix] [resource $bool]");
		return (0);
	}
	if (flag & CFG_SHOW) {
		fprintf(f, "rrrd host %s port %s prefix %s\n",
		    rrrd_host, rrrd_port, rrrd_prefix);
		return (0);
	}
	if (flag & CFG_NO) {
		rrrd_salen = 0;
		return (0);
	}

	for (;tok[0] != NULL; tok += 2) {
		if (tok[1] == NULL)
			return(Warn(f, 0, "need more arguments"));
		if (!strcmp(tok[0], "host")) {
			if (update_string(f, &rrrd_host, tok[1]))
				return (-1);
		} else if (!strcmp(tok[0], "port")) {
			if (update_string(f, &rrrd_port, tok[1]))
				return (-1);
		} else if (!strcmp(tok[0], "prefix")) {
			if (update_string(f, &rrrd_prefix, tok[1]))
				return (-1);
		} else if (!strcmp(tok[0], "resource")) {
			i = evSetTimer(ctx, resource_timer, &resource_tid,
			    evConsTime(0, 0), evConsTime(60, 0), &resource_tid);
			Assert(i == 0, "evSetTimer(resource_timer)");
			i = evConfigTimer(ctx, resource_tid, "rate", 0);
			Assert(i == 0, "evConfigTimer(source_timer)");

		} else {
			return(Warn(f, 0, "unknown argument"));
		}
	}
	if (recsock < 0)
		getsocket(f);
	if (recsock < 0)
		return (-1);
	i = resolver(ctx, rrrd_host, rrrd_port,
	     PF_INET, SOCK_DGRAM, AI_CANONNAME, resolved_rrrd, NULL, f);
	if (i < 0)
		fprintf(f, "Failed...\n");

	return (0);
}

void
init_record(evContext ctx __unused)
{
	char buf[100];
	const char *p;

	if (update_string(stderr, &rrrd_port, "4444"))
		exit(1);
	p = "";
	if (!gethostname(buf, sizeof buf - 2)) {
		strcat(buf, "/");
		p = buf;
	}
	if (update_string(stderr, &rrrd_prefix, p))
		exit(1);
	ConfigAdd(NULL, "rrrd", record_rrrd, "Remote RRD data logging");
	ConfigAdd(NULL, "record", record_record, "data logging facility");

}
