/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: source.c,v 1.3 2005/01/26 15:28:53 tom Exp $
 *
 * Herding time sources.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdarg.h>
#include <math.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <err.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/queue.h>
#include <math.h>

#include "ntpns.h"
#include "shmem.h"
#include "source.h"
#include "pll.h"
#include "internal.h"

u_char	sys_refid[4];
double	sys_delay;
double	sys_dispersion;
int	sys_stratum;
int	sys_precision;

static struct ntpns_shmem *ntp_shmem;

static struct logh *Lsource;

static evTimerID	source_tid;
static int myclock_step = 0;

struct sourcehd sources = TAILQ_HEAD_INITIALIZER(sources);

static struct source *sync_source;


clk_state_e clk_state = NOSRC;
struct source *selsrc;
static clk_state_e next_state = NOSRC;
static clk_sel_t *clk_sel;

static const char *
state_name(clk_state_e st)
{
	switch (st) {
        case NOSRC:		return "<nosrc>";
        case FRAC_SWITCH:	return "<frac_switch>";
        case FRAC_TRACK:	return "<frac_track>";
        case UTC_STEP:		return "<utc_step>";
        case SRC_SWITCH:	return "<src_switch>";
        case UTC_LOCK:		return "<utc_lock>";
        case COAST:		return "<coast>";
        case FRAC_LOCK:		return "<frac_lock>";
	default:		return "<invalid>";
	}
}

static const char *
ref_name(src_state_e st)
{
	switch (st) {
	case SRC_INVALID:	return "<invalid>";
	case SRC_LOS:		return "<LOS>";
	case SRC_FREQUENCY:	return "<frequency>";
	case SRC_FRACTION:	return "<fraction>";
	case SRC_UTC:		return "<UTC>";
	default:		return "<?>";
	}
}

static void
timer_kick(struct source *sid)
{
	int i, k, l, tv;
	u_int u;
	double td, a, b;
	struct timeline *tl;

	/* XXX: check sid->state changes */
	if (sid->flags & SRC_F_DATA) {
		tv = 1;
		td = sid->last_delta;
		sid->flags &= ~SRC_F_DATA;
		sid->los = 0;
	} else {
		tv = 0;
		td = -20011966.999;	/* We should be able to spot this */
		sid->los++;
	}
	u = sid->tick;
	sid->tick += 2;
	for (i = 0; i < NTIMEHIST; i++) {

		/* We only react on every 2^i'th sample */
		if (u & 1)
			break;
		u >>= 1;

		/* update sum fields with the sample we discard */
		tl = &sid->t[i];
		if (tl->v[NTIMELINE - 1]) {
			tl->n--;
			tl->s -= tl->d[NTIMELINE - 1];
		}

		/* Shift the register */
		memmove(&tl->v[1], &tl->v[0],
		     sizeof(tl->v[0]) * (NTIMELINE - 1));
		memmove(&tl->d[1], &tl->d[0],
		     sizeof(tl->d[0]) * (NTIMELINE - 1));

		/* Insert new sample */
		tl->v[0] = tv;
		tl->d[0] = td;
		if (tv) {
			tl->n++;
			tl->s += td;
		}

		/*
		 * If this sample were not valid, we will use the previous
		 * sample to update the next row nomatter what validity
		 * it has.
		 */
		if (!tv) {
			tv = tl->v[1];
			td = tl->d[1];
		}
	
		/* Calculate modified allan variance, if we can. */
		if (tl->n < 3) {
			tl->a = 1.0;
		} else {
			a = 0;
			l = 0;
			for (k = 0; k < NTIMELINE - 3; k++) {
				if (!(tl->v[k] && tl->v[k + 1] && tl->v[k + 2]))
					continue;
				b = tl->d[k + 2] + tl->d[k] - 2 * tl->d[k + 1];
				b /= (1 << i);
				a += b * b;
				l++;
			}
			if (l == 0)
				tl->a = 1.0;
			else
				tl->a = sqrt((1.0 / (2.0 * l)) * a);
		}
	}
}


static void
source_timer(evContext ctx, void *uap, struct timespec t0, struct timespec dt)
{
	struct timespec t1, t2;
	struct source *sp, *sp2;
	double a;
	int i, j;

	/* Make sure the source timer runs at 100msec before each UTC sec */
	t1 = evUTCTime();
	j = t1.tv_nsec - 900000000;
	t0 = evSubTime(t0, evConsTime(1, j));
	t2 = evAddTime(t0, dt);
	i = evResetTimer(ctx, source_tid, source_timer, uap,
		t2, evConsTime(1, 0));
	Assert(i == 0, "evResetTimer(source_timer)");

	TAILQ_FOREACH(sp, &sources, list)
		timer_kick(sp);

	if (sync_source != NULL && sync_source->flags & SRC_F_DELETE)
		sync_source->los = sync_source->los_timeout + 1;
	next_state = clk_state;
	selsrc = clk_sel(ctx, sync_source, TICK);
	switch(clk_state) {
	case NOSRC:
		/*
		 * This is the state we start in, and where we end up whenever
		 * we loose track of time.  Hang around until a usable source
		 * gets selected.
		 */
		if (selsrc == NULL)
			next_state = NOSRC;
		else if (selsrc->state == SRC_UTC)
			next_state = UTC_STEP;
		else if (selsrc->state == SRC_FRACTION)
			next_state = FRAC_SWITCH;
		break;
	case FRAC_SWITCH:
		/*
		 * Switch to a fractional source.
		 * If we are too far off, step the clock, and go to NOSRC.
		 * Otherwise, pass into tracking.
		 */
		if (selsrc == NULL)
			next_state = NOSRC;
		else if (selsrc->state == SRC_UTC)
			next_state = UTC_STEP;
		else if (selsrc != sync_source)
			next_state = NOSRC;
		else if (selsrc->state != SRC_FRACTION) 
			next_state = NOSRC;
		else if (sync_source->t[sync_source->hist0].v[0]) {
			a = sync_source->t[sync_source->hist0].d[0];
			if (a < -.010 || a > .010) {
				pllstep(a);
				next_state = NOSRC;
			} else {
				next_state = FRAC_TRACK;
			}
		}
		break;
	case FRAC_TRACK:
		/*
		 * Track a fractional source as long as we can or until
		 * we get a better offer.
		 */
		if (sync_source->los > sync_source->los_timeout)
			next_state = NOSRC;
		else if (selsrc == NULL)
			next_state = NOSRC;
		else if (selsrc->state == SRC_UTC)
			next_state = UTC_STEP;
		else if (selsrc != sync_source)
			next_state = NOSRC;
		break;
	case UTC_STEP:
		/*
		 * We have seen our first UTC source, and may have to
		 * step the clock. 
		 */
		if (selsrc != sync_source)
			next_state = NOSRC;
		else if (sync_source->state != SRC_UTC)
			next_state = NOSRC;
		else if (sync_source->los > sync_source->los_timeout)
			next_state = NOSRC;
		else if (sync_source->t[sync_source->hist0].v[0]) {
			a = sync_source->t[sync_source->hist0].d[0];
			if (a < -.010 || a > .010) {
				pllstep(a);
				next_state = NOSRC;
			} else {
				next_state = SRC_SWITCH;
			}
		}
		break;
	case SRC_SWITCH:
		/*
		 * Locking down on a new synchronized source.
		 */
		if (selsrc == NULL)
			next_state = COAST;
		else if (selsrc->los > selsrc->los_timeout)
			next_state = COAST;
		else if (selsrc->state == SRC_FRACTION)
			next_state = FRAC_LOCK;
		else if (selsrc->state == SRC_UTC)
			next_state = UTC_LOCK;
		else
			next_state = COAST;
		break;
	case UTC_LOCK:
		if (sync_source->los > sync_source->los_timeout)
			next_state = SRC_SWITCH;
		else if (sync_source->state != SRC_UTC)
			next_state = SRC_SWITCH;
		else if (selsrc != sync_source)
			next_state = SRC_SWITCH;
		break;
	case COAST:
		/* XXX: timeout */
		if (selsrc == NULL)
			;
		else if (selsrc->state == SRC_UTC)
			next_state = SRC_SWITCH;
		else if (selsrc->state == SRC_FRACTION)
			next_state = SRC_SWITCH;
		break;
	case FRAC_LOCK:
		if (sync_source->los > sync_source->los_timeout)
			next_state = SRC_SWITCH;
		else if (selsrc != sync_source)
			next_state = SRC_SWITCH;
		else if (selsrc->state != SRC_FRACTION)
			next_state = SRC_SWITCH;
		break;
	default:
		Abort("Illegal clock state %d", clk_state);
	}

	if (clk_state != next_state)
		Log(Lsource, "Clock state %s -> %s\n",
		    state_name(clk_state), state_name(next_state));

	switch(next_state) {
	case NOSRC:
		sync_source = NULL;
		memset(sys_refid, 0, sizeof sys_refid);
		i = 0;
		break;
	case FRAC_SWITCH:
		sync_source = selsrc;
		pll_reset();
		i = 0;
		break;
	case FRAC_TRACK:
		sync_source = selsrc;
		i = 1;
		break;
	case UTC_STEP:
		sync_source = selsrc;
		i = 0;
		break;
	case SRC_SWITCH:
		sync_source = selsrc;
		pll_reset();	/* XXX: really ? */
		i = 0;
		break;
	case UTC_LOCK:
		sync_source = selsrc;
		memcpy(sys_refid,  sync_source->refid, sizeof sys_refid);
		i = 1;
		break;
	case COAST:
		sync_source = NULL;
		memcpy(sys_refid, "LOCL", 4);
		i = 0;
		break;
	case FRAC_LOCK:
		i = 1;
		sync_source = selsrc;
		memcpy(sys_refid,  sync_source->refid, sizeof sys_refid);
		break;
	default:
		Abort("Illegal new clock state %d", next_state);
	}
	clk_state = next_state;
	ntp_shmem->clockstate = next_state;
	if (sync_source == NULL)
		ntp_shmem->clocksource = 0;
	else
		ntp_shmem->clocksource = sync_source->idx;

	if (i && sync_source->t[sync_source->hist0].v[0]) {
		pll(sync_source->t[sync_source->hist0].d[0]);
		sys_delay = sync_source->delay;
		sys_dispersion = sync_source->dispersion +
			sync_source->t[sync_source->hist0].a;
		sys_stratum = sync_source->stratum + 1;
	} else {
		pllnone();
	}

	Assert((sync_source == NULL) || !(sync_source->flags & SRC_F_DELETE),
	    "deleting sync_source");

	TAILQ_FOREACH_SAFE(sp, &sources, list, sp2) {
		if (clock_step != myclock_step)
			memset(sp->t, 0, sizeof sp->t);
		sp->flags &= ~(SRC_F_PARAM | SRC_F_NEW);
		if (sp->flags & SRC_F_DELETE) {
			TAILQ_REMOVE(&sources, sp, list);
			shmem_free(sp);
		}
	}
	myclock_step = clock_step;
}

/*
 * Default clock selector
 *
 * Simply pick the source with the highest votes in the highest state.
 */
static struct source *
default_clock_selector(evContext ctx __unused, struct source *cur, clk_sel_e action)
{
	struct source *sp;

	if (action != TICK)
		return (NULL);

	TAILQ_FOREACH(sp, &sources, list) {
		if (sp->los > sp->los_timeout)
			continue;
		if (sp->votes <= 0)
			continue;
		if (cur == NULL)
			cur = sp;
		else if ((sp->state > cur->state) ||
		   (sp->state == cur->state && sp->votes > cur->votes))
			cur = sp;
	}
	if (cur != NULL && cur->los > cur->los_timeout)
		cur = NULL;
	return (cur);
}

/*
 * Public interface to sources
 */

void
ClockSelect(evContext ctx, clk_sel_t *func)
{

	if (clk_sel != NULL)
		(void)clk_sel(ctx, NULL, DEINIT);
	if (func == NULL)
		clk_sel = default_clock_selector;
	else
		clk_sel = func;
	(void)clk_sel(ctx, NULL, INIT);
}

struct source *
SourceAdd(const char *fmt, ...)
{
	struct source *sp;
	va_list ap;
	static u_int idx = 10000;


	sp = shmem_alloc(sizeof *sp, SOURCE_TAG);
	if (sp == NULL)
		return (NULL);
	sp->idx = idx++;
	va_start(ap, fmt);
	vsnprintf(sp->name, sizeof sp->name, fmt, ap);
	va_end(ap);
	TAILQ_INSERT_TAIL(&sources, sp, list);
	sp->flags |= SRC_F_NEW;
	sp->los_timeout = 60;
	sp->los = sp->los_timeout + 1;
	sp->state = SRC_INVALID;
	return (sp);
}

void
SourceDel(struct source *sid)
{

	sid->flags |= SRC_F_DELETE;
}

void
SourceTick(struct source *sid, struct ts *ts, src_state_e state)
{
	double delta;
	struct ts mts;

	if (ts == NULL)
		return;
	if (state == SRC_FRACTION) {
		TsNormalize(ts);
		delta = (int64_t)ts->ts.bt.frac;
		delta /= 18446744073709551616.0; /* 2^64 in case you wondered */
	} else if (state == SRC_UTC) {
		mts.ts.ts = evUTCTime();
		mts.ts.ts.tv_nsec = 0;
		mts.fmt = TS_TIMESPEC;
		delta = TsDiff(&mts, ts);
	} else {
		delta = 0;
	}

	SourceDelta(sid, ts, delta, state);
}

void
SourceDelta(struct source *sid, struct ts *ts, double delta, src_state_e state)
{
	char buf[30];
	double d;

	if (ts == NULL)
		return;

	TsFormat(ts, buf);
	d = delta;
	if (state != SRC_UTC)
		sid->last_delta = delta - (int)delta;
	else
		sid->last_delta = delta;
	Log(Lsource, "SourceDelta(%s), %s %12.9f (%12.9f) %s",
	    sid->name, buf, delta, d, ref_name(state));
	sid->last_ts = *ts;
	TsNormalize(&sid->last_ts);
	sid->state = state;
	sid->flags |= SRC_F_DATA;
}

void
SourceRefid(struct source *sid, const void *refid)
{

	memmove(sid->refid, refid, sizeof sid->refid);
	sid->flags |= SRC_F_PARAM;
}

void
SourceParam(struct source *sid, src_param_e param, double val)
{
	int i;

	switch (param) {
	case SRC_UPDATE_RATE:
		sid->update_rate = val;
		for (i = 0; i < NTIMEHIST; i++)
			if ((1 << i) >= val)
				break;
		sid->hist0 = i;
		break;
	case SRC_LOS_TIMEOUT:	sid->los_timeout = val; break;
	case SRC_VOTES:		sid->votes = val; break;
	case SRC_DISPERSION:	sid->dispersion = val; break;
	case SRC_STRATUM:	sid->stratum = val; break;
	case SRC_DELAY:		sid->delay = val; break;
	default: /* XXX wrong */ break;
	}
	sid->flags |= SRC_F_PARAM;
}

static int
source_source(evContext ctx __unused, FILE *f, void *priv __unused, char **tok, int flag __unused)
{
	struct source *sp;
	char buf[50];
	int i, j;

	TAILQ_FOREACH(sp, &sources, list) {
		fprintf(f, "Source %s: votes %f flags %s",
		    sp->name, sp->votes, ref_name(sp->state));
		fprintf(f, " los %d/%d update %d",
			sp->los, sp->los_timeout, sp->update_rate);
		if (sp == selsrc)
			fprintf(f, " SELECTED");
		fprintf(f, "\n");
			
		if (tok[0] != NULL) {
			fprintf(f, "\tstratum %d delay %.09f dispersion %.09f",
				sp->stratum, sp->delay, sp->dispersion);
			fprintf(f, " refid [%02x %02x %02x %02x]\n",
				sp->refid[0], sp->refid[1], sp->refid[2], sp->refid[3]);
		}
		TsFormat(&sp->last_ts, buf);
		fprintf(f, "\tlast_ts %s last_delta %.9f\n",
		     buf, sp->last_delta);
		if (tok[0] != NULL)
			for (i = 0; i < NTIMEHIST; i++) {
				if (sp->t[i].n < 3)
					continue;
				fprintf(f, "%6d: ", 1 << i);
				for (j = 0; j < NTIMELINE; j++)
					fprintf(f, "%d", sp->t[i].v[j]);
				if (sp->t[i].n) {
					fprintf(f, " %12.9f", sp->t[i].s / sp->t[i].n);
					fprintf(f, " %.3e", sp->t[i].a);
				}
				if (!strcmp(tok[0], "long")) {
					for (j = 0; j < NTIMELINE; j++) {
						if (sp->t[i].v[j])
							fprintf(f, " %12.9f", sp->t[i].d[j]);
						else
							fprintf(f, "             ");
					}
				}
				fprintf(f, "\n");
			}
	}
	return (0);
}

void
init_source(evContext ctx __unused)
{
	int i;

	ClockSelect(ctx, NULL);

	Lsource = LogNewHandle(Ldebug, "source");

	ConfigAdd(NULL, "source", source_source, "Source selection");
	i = evSetTimer(ctx, source_timer, &source_tid,
	     evConsTime(0, 0), evConsTime(1, 0), &source_tid);
	Assert(i == 0, "evSetTimer(source_timer)");
	i = evConfigTimer(ctx, source_tid, "rate", 0);
	Assert(i == 0, "evConfigTimer(source_timer)");
	sys_precision = -29;
	ntp_shmem = shmem_alloc(sizeof *ntp_shmem, NTP_TAG);
}
