/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: partner.c,v 1.2 2005/01/26 15:28:53 tom Exp $
 */

#include <err.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/endian.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>

#include <ntpns.h>
#include <ntpv4.h>

static LIST_HEAD(, configured) chead = LIST_HEAD_INITIALIZER(chead);

LIST_HEAD(phead, partner);

static int nhash = 53;
static int timeout = 86400;
static int npartner;

static int maxpartner = 10000;
static struct partner *partner;
static LIST_HEAD(, partner) freelist = LIST_HEAD_INITIALIZER(freelist);
static TAILQ_HEAD(, partner) lrulist = TAILQ_HEAD_INITIALIZER(lrulist);

static struct phead *hash;

static int
InitPartner(evContext ctx __unused)
{
	int i;

	hash = calloc(sizeof *hash, nhash);
	if (hash == NULL) {
		nomem();
		return (-1);
	}
	partner = shmem_alloc(sizeof(struct partner) * maxpartner, TAG_PARTNER);
	if (partner == NULL) {
		nomem();
		return (-1);
	}
	/* Put them on the freelist in reverse order */
	for (i = maxpartner - 1; i >= 0; i--) {
		partner[i].generation = 0;
		partner[i].flag = P_F_NEVER;
		LIST_INSERT_HEAD(&freelist, &partner[i], hash);
	}
	for (i = 0; i < nhash; i++)
		LIST_INIT(&hash[i]);
	return (0);
}

struct partner *
LookupPartner(evContext ctx, time_t sec, uint32_t ip)
{
	int h;
	struct partner *p, *pp;

	h = ip % nhash;
	if (hash != NULL) {
		LIST_FOREACH_SAFE(p, &hash[h], hash, pp) {
			if (p->ip == ip) {
				LIST_REMOVE(p, hash);
				LIST_INSERT_HEAD(&hash[h], p, hash);
				if (p->configured == NULL) {
					TAILQ_REMOVE(&lrulist, p, lru);
					TAILQ_INSERT_TAIL(&lrulist, p, lru);
				}
				if (sec)
					p->last = sec;
				return (p);
			}
			if (p->configured == NULL && 
			    p->last + timeout < sec)
				DeletePartner(ctx, p);
		}
	} else if (InitPartner(ctx))
		return (NULL);

	if (LIST_EMPTY(&freelist) && !TAILQ_EMPTY(&lrulist))
		DeletePartner(ctx, TAILQ_FIRST(&lrulist));
	if (LIST_EMPTY(&freelist))
		return (NULL);

	p = LIST_FIRST(&freelist);
	LIST_REMOVE(p, hash);
	p->flag &= ~P_F_NEVER;
	p->ip = ip;
	p->generation++;
	npartner++;
	LIST_INSERT_HEAD(&hash[h], p, hash);
	TAILQ_INSERT_TAIL(&lrulist, p, lru);
	if (sec)
		p->last = sec;
	return (p);
}

void
DeletePartner(evContext ctx, struct partner *pp)
{
	u_int g;

	LIST_REMOVE(pp, hash);
	if (pp->configured) {
		evClearTimer(ctx, pp->tid);
		LIST_REMOVE(pp, config);
	} else {
		TAILQ_REMOVE(&lrulist, pp, lru);
	}
	if (pp->cookie != NULL)
		free(pp->cookie);
	g = pp->generation;
	memset(pp, 0, sizeof *pp);
	pp->generation = g + 1;
	LIST_INSERT_HEAD(&freelist, pp, hash);
	npartner--;
}

struct configured *
FindConfigured(char *name, int create)
{
	struct configured *cp;

	LIST_FOREACH(cp, &chead, next)
		if (!strcmp(name, cp->name))
			return (cp);
	if (!create)
		return (NULL);
	cp = calloc(sizeof *cp, 1);
	if (cp == NULL) {
		nomem();
		return (NULL);
	}
	cp->name = strdup(name);
	if (cp->name == NULL) {
		nomem();
		free(cp);
		return (NULL);
	}
	LIST_INSERT_HEAD(&chead, cp, next);
	LIST_INIT(&cp->partner);
	return (cp);
}

void
DeleteConfigured(evContext ctx, struct configured *cp)
{
	struct partner *p;

	evClearTimer(ctx, cp->tid);
	LIST_REMOVE(cp, next);
	free(cp->name);
	for (;;) {
		p = LIST_FIRST(&cp->partner);
		if (p == NULL)
			break;
		TAILQ_INSERT_TAIL(&lrulist, p, lru);
		DeletePartner(ctx, p);
	}
	free(cp);
}

/*
 * This timer lingers for all configured peers.  It is responsible for
 * DNS resolution and generally figuring out that things have changed.
 */

static void
config_resolved(evContext ctx __unused, void *priv, int error, struct addrinfo *ai, FILE *f __unused)
{
	struct addrinfo *a;
	struct partner *p, *pt;
	struct configured *cf;
	struct sockaddr_in *sa;
	char *q;
	int i;

	cf = priv;

	if (error) {
		/* Check again not quite five minutes from now */
		i = evSetTimer(ctx, config_timer, cf,
		     evAddTime(evNowTime(), evConsTime(287, 0)),
			 evConsTime(0, 0), &cf->tid);
		if (i < 0)
			err(1, "evSetTimer() failed");
		return;
	}

	LIST_FOREACH(p, &cf->partner, config)
		p->flag &= ~P_F_UPDATE;

	for (a = ai; a != NULL; a = a->ai_next) {
		if (a->ai_addrlen != sizeof p->addr) {
			warnx("ai_addlen=%d, expected %d",
				a->ai_addrlen, sizeof p->addr);
			continue;
		}
		LIST_FOREACH(p, &cf->partner, config) {
			if (memcmp(a->ai_addr, &p->addr, a->ai_addrlen))
				continue;
			p->flag |= P_F_UPDATE;
			break;
		}
		if (p == NULL) {
			sa = (void*)a->ai_addr;
			p = LookupPartner(ctx, 0, ntohl(sa->sin_addr.s_addr));
			if (p == NULL)
				err(1, "XXX: malloc failed");
			p->flag |= P_F_UPDATE;
			LIST_INSERT_HEAD(&cf->partner, p, config);
			memcpy(&p->addr, a->ai_addr, a->ai_addrlen);
			p->configured = cf;
			TAILQ_REMOVE(&lrulist, p, lru);
			if (p->sid == NULL) {
				q = addr2ascii(sa->sin_family, &sa->sin_addr, sizeof sa->sin_addr, NULL);
				if (q == NULL)
					err(1, "addr2ascii");
				p->sid = SourceAdd(q);
				SourceParam(p->sid, SRC_UPDATE_RATE, cf->poll);
				SourceParam(p->sid, SRC_LOS_TIMEOUT, 3 * cf->poll);
				SourceParam(p->sid, SRC_VOTES, cf->votes);
			}
			i = evSetTimer(ctx, poll_timer, p,
			     evConsTime(0, 0), evConsTime(cf->poll, 0), &p->tid);
			if (i < 0)
				err(1, "evSetTimer() failed");
			p->inp = cf->inp;
		}
	}
	LIST_FOREACH_SAFE(p, &cf->partner, config, pt) {
		if (p->flag & P_F_UPDATE)
			continue;
		if (p->sid != NULL)
			SourceDel(p->sid);
		DeletePartner(ctx, p);
	}

	/* Check again not quite an hour from now */
	i = evSetTimer(ctx, config_timer, cf,
	     evAddTime(evNowTime(), evConsTime(3547, 0)),
	     evConsTime(0, 0), &cf->tid);
	if (i < 0)
		err(1, "evSetTimer() failed");
}

void
config_timer(evContext ctx __unused, void *uap, struct timespec due __unused, struct timespec inter __unused)
{
	struct configured *cf;
	int i;

	cf = uap;

	Log(lpartner, "config_timer(%p, %s)\n", cf, cf->name);
	i = resolver(ctx, cf->name, "ntp", PF_UNSPEC, SOCK_DGRAM,
	    AI_CANONNAME, config_resolved, cf, stdout);
	if (i < 0)
		Log(Lerror, "ntpv4 config_timer(%p, %s): resolver failed",
		    cf, cf->name);
}

static void
show_partner(FILE *f, struct partner *p)
{
	char buf[80];
	uint32_t ip;
	int i;

	ip = ntohl(p->addr.sin_addr.s_addr);
	i = ntohs(p->addr.sin_port);
	sprintf(buf, "%d.%d.%d.%d:%d",
	    (ip >> 24) & 0xff, (ip >> 16) & 0xff,
	    (ip >> 8) & 0xff, (ip >> 0) & 0xff, i);
	while (strlen(buf) < 22)
		strcat(buf, " ");
	fprintf(f, "%s", buf);
	PrintPacket(f, &p->packet);
}

void
ntpv4_show_partner(evContext ctx __unused, FILE *f, void **priv __unused, char **tok __unused, int flags __unused)
{
	struct configured *cf;
	struct partner *p, *p2;
	time_t sec;
	int n;

	time(&sec);
	fprintf(f, "%d partners (of %d max)\n", npartner, maxpartner);
	LIST_FOREACH(cf, &chead, next) {
		fprintf(f, "NAME %s\n", cf->name);
		LIST_FOREACH(p, &cf->partner, config)
			show_partner(f, p);
	}
	if (tok[1] == NULL)
		return;
	n = atoi(tok[1]);
	if (n <= 0)
		n = maxpartner;
	fprintf(f, "ANON (%s)\n", tok[1]);
	TAILQ_FOREACH_SAFE(p, &lrulist, lru, p2) {
		if (p->last + timeout < sec) {
			DeletePartner(ctx, p);
			continue;
		}
		show_partner(f, p);
		if (--n == 0)
			break;
	}
}
