/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: proto.c,v 1.2 2005/01/26 15:28:53 tom Exp $
 */

#include <err.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/endian.h>
#include <netinet/in.h>

#include <ntpns.h>
#include <ntpv4.h>

/**********************************************************************/

#define MODE_0		0
#define MODE_SYMACT	1
#define MODE_SYMPAS	2
#define MODE_CLIENT	3
#define MODE_SERVER	4
#define MODE_BCAST	5
#define MODE_CONTROL	6
#define MODE_7		7

#if 0
/*
 * Number of Days since 1 BC Gregorian to 1 January of given year
 */
#define julian0(year)	(((year) * 365 ) + ((year) > 0 ? (((year) + 3) \
			/ 4 - ((year - 1) / 100) + ((year - 1) / \
			400)) : 0))

/*
 * Number of days since start of NTP time to 1 January of given year
 */
#define ntp0(year)	(julian0(year) - julian0(1900))

/*
 * Number of days since start of UNIX time to 1 January of given year
 */
#define unix0(year)	(julian0(year) - julian0(1970))

#define	JAN_1970	0x83aa7e80	/* 2208988800 1970 - 1900 in seconds */

#endif

/**********************************************************************/


void
PrintPacket(FILE *f, struct ntp_packet *p)
{
	char buf[50];
	int i;

	fprintf(f, "%d%d", (p->leap >> 1) & 1, (p->leap & 1));
	fprintf(f, " %d", p->version);
	fprintf(f, " %d", p->mode);
	fprintf(f, " %2d", p->stratum);
	fprintf(f, " %2d", p->poll);
	fprintf(f, " %3d", p->prec);
	fprintf(f, " L %9.2e", p->delay);
	fprintf(f, " D %9.2e", p->disp);
	fprintf(f, " I %02x%02x%02x%02x",
	    p->refid[0], p->refid[1], p->refid[2], p->refid[3]);
	for (i = 0; i < 5; i++) {
		TsFormat(&p->ts[i], buf);
		fprintf(f, " %s", buf);
	}
	fprintf(f, " %14.9f", TsDiff(&p->ts[4], &p->ts[3]));
	fprintf(f, "\n");
}

static void
ReplyNtp(struct partner *pr)
{
	u_char buf[48];
	struct msghdr *mh;
	int i;
	uint64_t u;
	struct ntp_packet *p;

	if (!pll_locked)
		return;
	p = &pr->packet;
	if (p->mode != 1 && p->mode != 3)
		return;
	mh = socket_prepare(pr, buf, 48);

	buf[0] = 0 << 6;
	buf[0] |= p->version << 3;
	if (p->mode == 1)
		buf[0] |= 2;
	else if (p->mode == 3)
		buf[0] |= 4;
	buf[1] = sys_stratum;
	buf[2] = p->poll;
	buf[3] = sys_precision;
	be32enc(buf + 4, sys_delay * 65536);
	be32enc(buf + 8, sys_dispersion * 65536);
	memcpy(buf + 12, sys_refid, sizeof sys_refid);

	TsNTPtime(&p->ts[3], &u);
	be64enc(buf + 24, u);

	TsNTPtime(&p->ts[4], &u);
	be64enc(buf + 32, u);

	u = TsNTPNow();
	be64enc(buf + 40, u);
	be64enc(buf + 16, u);

	i = sendmsg(pr->inp->socket, mh, 0);
	if (i != 48)
		warn("sendmsg() %d", i);
	free(mh);
}

static void
SendNtp(evContext ctx __unused, struct partner *pr)
{
	u_char buf[48];
	struct msghdr *mh;
	uint64_t u;
	int i;
	struct ntp_packet *p;

	mh = socket_prepare(pr, buf, 48);

	p = &pr->packet;
	buf[0] = 3 << 6;
	buf[0] |= 4 << 3;
	buf[0] |= 3;
	buf[1] = sys_stratum;
	buf[2] = 4; /* XXX */
	buf[3] = sys_precision;
	be32enc(buf + 4, sys_delay * 65536);
	be32enc(buf + 8, sys_dispersion * 65536);
	memcpy(buf + 12, sys_refid, sizeof sys_refid);
	TsNTPtime(&p->ts[3], &u);
	be64enc(buf + 24, u);

	TsNTPtime(&p->ts[4], &u);
	be64enc(buf + 32, u);

	u = TsNTPNow();		/* XXX: we should record this timestamp */
	be64enc(buf + 40, u);
	be64enc(buf + 16, u);

	i = sendmsg(pr->inp->socket, mh, 0);
	if (i != 48)
		warn("sendmsg() %d", i);
	free(mh);
	pr->originate = u;
	Log(lsocket, "SendNtp(%p, %08x)", pr, pr->ip);
}


/*
 * XXX: I'm not happy with the ip+port argument.  Ideally it should be an
 * XXX: opaque chunk of data, but I need to look into it to be able to
 * XXX: to tell if the source port is 123 and possibly detect NAT'ing.
 */
void
RcvPacket(evContext ctx, struct instance *inp, void *packet, int len, struct sockaddr_in *sa, struct ts *ts, void *cookie, u_int cookielen)
{
	uint8_t		*s;
	struct ntp_packet *p;
	struct partner *pr;
	uint32_t ip;
	u_char refid[4];
	int port;
	double d;

	ip = ntohl(sa->sin_addr.s_addr);
	port = ntohs(sa->sin_port);
	s = packet;

#if 1
	if (be32dec(s + 44) == 0) {
		return;
	}
#endif

	pr = LookupPartner(ctx, ts->ts.bt.sec, ip);
	Log(lsocket, "RcvPacket(%p, %d, %08x, %d, %p, %p, %d, -> %p) %d.%d.%d.%d:%d\n",
	    packet, len, ip, port, ts, cookie, cookielen, pr,
		(ip >> 24) & 255,
		(ip >> 16) & 255,
		(ip >> 8) & 255,
		(ip >> 0) & 255,
		port);
	if (pr == NULL)
		return;

	memcpy(&pr->addr, sa, sizeof pr->addr);
	pr->inp = inp;

	if (cookie != NULL) {
		if (pr->cookielen != cookielen)
			pr->cookie = realloc(pr->cookie, cookielen);
		if (pr->cookie != NULL) {
			memcpy(pr->cookie, cookie, cookielen);
			pr->cookielen = cookielen;
		} else {
			pr->cookielen = 0;
		}
	} else if (pr->cookie != NULL) {
		free(pr->cookie);
		pr->cookie = NULL;
		pr->cookielen = 0;
	}

	p = &pr->packet;

	/* Decode the packet */
	p->leap = (s[0] >> 6) & 0x3;
	p->version = (s[0] >> 3) & 0x7;
	p->mode = s[0] & 0x7;
	p->stratum = s[1];
	p->poll = s[2];
	p->prec = s[3];
	p->delay = (int32_t)be32dec(s + 4) / 65536.0;
	p->disp = (int32_t)be32dec(s + 8) / 65536.0;
	p->refid[0] = s[12];
	p->refid[1] = s[13];
	p->refid[2] = s[14];
	p->refid[3] = s[15];
	p->ts[0].ts.ntp = be64dec(s + 16);
	p->ts[0].fmt = TS_NTPTIME;
	p->ts[1].ts.ntp = be64dec(s + 24);
	p->ts[1].fmt = TS_NTPTIME;
	p->ts[2].ts.ntp = be64dec(s + 32);
	p->ts[2].fmt = TS_NTPTIME;
	p->ts[3].ts.ntp = be64dec(s + 40);
	p->ts[3].fmt = TS_NTPTIME;
	p->ts[4] = *ts;
	
	switch(p->mode) {
	case MODE_0:
	case MODE_CONTROL:
	case MODE_7:
		return;
	}

#if 0
	printf("%d %d.%d.%d.%d", pr->configured != NULL,
	    (ip >> 24) & 0xff,
	    (ip >> 16) & 0xff,
	    (ip >> 8) & 0xff,
	    (ip >> 0) & 0xff);
#endif
	if (0) {
		PrintPacket(stdout, p);
	}

	if (pr->configured == NULL || p->mode == 1 || p->mode == 3) {
		ReplyNtp(pr);
		return;
	}
	if (pr->sid != NULL && pr->originate == be64dec(s + 24)) {
		SourceParam(pr->sid, SRC_DELAY, p->delay);
		SourceParam(pr->sid, SRC_DISPERSION, p->disp);
		SourceParam(pr->sid, SRC_STRATUM, p->stratum);
		be32enc(refid, ip);
		SourceRefid(pr->sid, refid);
		d = TsDiff(&p->ts[3], &p->ts[4]);
		d += TsDiff(&p->ts[2], &p->ts[1]);
		d /= 2;
		SourceDelta(pr->sid, &p->ts[3], -d, SRC_UTC);
	}
	return;
}

void
poll_timer(evContext ctx , void *uap, struct timespec due __unused, struct timespec inter __unused)
{
	struct partner *p;

	p = uap;
	SendNtp(ctx, p);
}
