/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: oncore_m12.c,v 1.3 2005/01/26 15:28:53 tom Exp $
 */

#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <pthread.h>

#include <ntpns.h>

#include "oncore_m12.h"

struct logh *lm12_debug, *lm12_serial, *lm12_msg;

static TAILQ_HEAD(, instance)	instance_head =
    TAILQ_HEAD_INITIALIZER(instance_head);

/**********************************************************************/

static int
m12_almanac(evContext ctx __unused, FILE *f, void *priv __unused, char **tok, int flag)
{
	struct instance *ip;

	ip = priv;
	if (tok[0] == NULL || flag & CFG_HELP) {
		ConfigHelp(f, "m12 $instance almanac $filename");
		return (0);
	}
	if (flag & CFG_SHOW) {
		if (ip->almanac_file == NULL)
			fprintf(f, "no almanac cache file\n");
		else
			fprintf(f, "almanac cache file: %s\n", ip->almanac_file);
		return (0);
	}
	if (flag & CFG_NO) {
		if (ip->almanac_file != NULL)
			free(ip->almanac_file);
		ip->almanac_file = NULL;
		return (0);
	}
	if (tok[0] != NULL)
		ip->almanac_file = strdup(tok[0]);
	else
		ip->almanac_file = strdup(DEFAULT_ALMANAC_FILE);
	if (ip->almanac_file == NULL) 
		return (ENOMEM);
	return (0);
}

static int
m12_pps(evContext ctx __unused, FILE *f, void *priv __unused, char **tok, int flag)
{
	struct instance *ip;

	if (tok[0] == NULL || flag & CFG_HELP) {
		ConfigHelp(f, "m12 $instance pps {pps_param}");
		return (0);
	}
	ip = priv;
	return (PpsConfig(ctx, f, &ip->pps, tok, flag));
}

/**********************************************************************/

// returns the instance with the unit-ID 'unit' from the list 
// of instances, or creates a new one (when 'create' is non-null,
// else returns NULL)
// Note that every module has its own struct instance
static struct instance *
m12_instance(evContext ctx, u_int unit, int create)
{
	struct instance *ip;

	TAILQ_FOREACH(ip, &instance_head, next) {
		if (ip->unit == unit)
			return (ip);
	}
	if (!create)
		return (NULL);
	ip = calloc(sizeof *ip, 1);
	if (ip == NULL)
		return (NULL);
	ip->unit = unit;
	ip->serial_fd = -1;
	PpsCreate(&ip->pps, 1);
	TAILQ_INSERT_TAIL(&instance_head, ip, next);
	ip->sid = SourceAdd("m12_%u", unit);
	SourceRefid(ip->sid, "GPS\0");
	SourceParam(ip->sid, SRC_UPDATE_RATE, 1.0);
	SourceParam(ip->sid, SRC_LOS_TIMEOUT, 15.0);
	ip->almanac_file = strdup(DEFAULT_ALMANAC_FILE);
	evInitID(&ip->reset_timer);
	oncore_state(ctx, ip, ST_CONFIG);
	return (ip);
}

static struct configlist *m12_cfl;

// master dispatch for all the command line arguments, gets the right
// instance from the instance list and calls the appropriate config
// function.
static int
m12_m12(evContext ctx, FILE *f, void *priv __unused, char **tok, int flag)
{
	u_int u;
	char *p;
	struct instance *ip;

	if (tok[0] == NULL) {
		fprintf(f, "Specify M12 unit number\n");
		return (-1);
	}
	u = strtoul(tok[0], &p, 0);
	if (*p != '\0') {
		fprintf(f, "M12 unit number not numeric\n");
		return (0);
	}
	if (flag & CFG_NO) {
		ip = m12_instance(ctx, u, 0);
		if (ip != NULL)
			fprintf(f, "XXX: no m12\n");
		return (0);
	}
	ip = m12_instance(ctx, u, 1);
	if (ip == NULL)
		return (ENOMEM);
   // dispatch to the respective config function
	return (ConfigCall(m12_cfl, ctx, f, ip, tok + 1, flag));
	return (0);
}

ref_init_t oncore_m12_init;

const char *
oncore_m12_init(evContext ctx __unused)
{
	static int once = 0;

	if (once) { 
		warnx("oncore_m12 module already loaded");
		return (NULL);
	} else {
		once = 1;
	}

	lm12_debug = LogNewHandle(Ldebug, "m12");
	lm12_serial = LogNewHandle(lm12_debug, "serial");
	lm12_msg = LogNewHandle(lm12_debug, "msg");

	if (init_oncore_tables() < 0)
		return (NULL);

	ConfigAdd(NULL, "m12", m12_m12, "refclock: Motorola Oncore M12");
	m12_cfl = ConfigCreate();
	ConfigAdd(m12_cfl, "almanac", m12_almanac, "Almanac cache file");
	ConfigAdd(m12_cfl, "serial", m12_serial, "Serial port");
	ConfigAdd(m12_cfl, "pps", m12_pps, "PPS-API configuration");
	return (MAGIC);
}
