/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: oncore_m12_msg.c,v 1.4 2005/01/26 15:28:53 tom Exp $
 */

#include <err.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/endian.h>
#include <pthread.h>

#include <ntpns.h>

#include "oncore_m12.h"

struct oncore_msg **oncore_msgs;

// initialize the device
void
oncore_initialize(struct instance *ip)
{
	u_char msg[50];

	m12_sendmsg(ip, "Bb\001", 3); /* Sat visibility */
	m12_sendmsg(ip, "Bd\001", 3); /* Almanac status */   
	m12_sendmsg(ip, "Be\001", 3); /* Req Almanac */
	m12_sendmsg(ip, "Bj\001", 3); /* Leap sec status */
	m12_sendmsg(ip, "Gj", 2);     /* Leap sec status */
	m12_sendmsg(ip, "Bo\001", 3); /* Req UTC offset */
	m12_sendmsg(ip, "Bp\001", 3); /* Req UTC/Ion data */
	m12_sendmsg(ip, "Gc\003", 3); /* 1pps */
	m12_sendmsg(ip, "Gd\003", 3); /* pos mode (set pos mode)*/
	m12_sendmsg(ip, "Ge\001", 3); /* TRAIM (disable)*/
	m12_sendmsg(ip, "Ha\001", 3); /* pos msg */
	m12_sendmsg(ip, "Hn\001", 3); /* TRAIM msg */

	if (ip->tsat > 2 && (ip->rcv_status >> 13) >= 4) {
		msg[0] = 'G';
		msg[1] = 'a';
		be32enc(msg + 2, ip->lat);
		be32enc(msg + 6, ip->lon);
		be32enc(msg + 10, ip->ht);
		msg[14] = 0;
		m12_sendmsg(ip, msg, 15);
	}

	if (ip->tsat < 1 || !ip->tm.tm_year) {
		time_t t;
		time(&t);
		gmtime_r(&t, &ip->tm);
	}
	msg[0] = 'G';
	msg[1] = 'b';
	msg[2] = ip->tm.tm_mon + 1;
	msg[3] = ip->tm.tm_mday;
	be16enc(msg + 4, ip->tm.tm_year + 1900);
	msg[6] = ip->tm.tm_hour;
	msg[7] = ip->tm.tm_min;
	msg[8] = ip->tm.tm_sec;
	msg[9] = 0;
	msg[10] = 0;
	msg[11] = 0;
	m12_sendmsg(ip, msg, 12);
}

/*
 * This handler does nothing, it is only here to suppress a hexdump in the
 * debugging output.
 */
static void
handle_NOOP(evContext ctx __unused, struct instance *ip __unused, struct oncore_msg *mp __unused)
{
}

static void *
Cb_thread(void *p)
{
	struct instance *ip;
	FILE *f;
	char *fn;
	int i;

	ip = p;
	if (ip->almanac_work == ALMANAC_SAVE) {
		asprintf(&fn, "%s.%d.tmp", ip->almanac_file, getpid());
		if (fn == NULL) {
			nomem();
			pthread_exit(NULL);
		}
		f = fopen(fn, "w");
		if (f == NULL) {
			warn("cannot open %s", fn);
			free(fn);
			pthread_exit(NULL);
		}
		i = fwrite(ip->almanac, 33 * 34, 1, f);
		if (i != 1 || fclose(f) != 0) {
			warn("cannot write/close %s", fn);
			if (i == 1)
				fclose(f);
			unlink(fn);
			free(fn);
			pthread_exit(NULL);
		}
		i = rename(fn, ip->almanac_file);
		if (i != 0) {
			warn("cannot rename %s to %s", fn, ip->almanac_file);
			unlink(fn);
		}
		free(fn);
		free(ip->almanac);
		ip->almanac = NULL;
		ip->almanac_ptr = NULL;
		ip->almanac_work = ALMANAC_DONE;
		pthread_exit(NULL);
	}

	if (ip->almanac_work == ALMANAC_LOAD) {
		if (ip->almanac == NULL)
			ip->almanac = calloc(1, 33 * 34);
		if (ip->almanac == NULL) {
			nomem();
			pthread_exit(NULL);
		}
		f = fopen(ip->almanac_file, "r");
		if (f == NULL) {
			warn("cannot open %s", ip->almanac_file);
			pthread_exit(NULL);
		}
		i = fread(ip->almanac, 33 * 34, 1, f);
		if (i != 1 || fclose(f) != 0) {
			warn("cannot read/close %s", ip->almanac_file);
			if (i != 1)
				fclose(f);
			free(ip->almanac);
			ip->almanac = NULL;
			pthread_exit(NULL);
		}
		ip->almanac_ptr = ip->almanac;
		ip->almanac_work = ALMANAC_LOADED;
		ip->almanac_n = 0;
		pthread_exit(NULL);
	}
	warnx("ip->almanac_work = %d ??", ip->almanac_work);
	pthread_exit(NULL);
}

static void
handle_Cb(evContext ctx __unused, struct instance *ip, struct oncore_msg *mp)
{
	int s, p, i;

	if (ip->almanac_work != ALMANAC_COLLECT)
		return;
	s = ip->serial_buf[4];
	p = ip->serial_buf[5];
	if (s == 0 && p == 0) {
		/* Receiver has no almanac, try to get it one */
		if (ip->almanac_file == NULL)
			return;
		ip->almanac_work = ALMANAC_LOAD;
		i = pthread_create(&ip->almanac_thread, NULL, Cb_thread, ip);
		if (i != 0) {
			warn("pthread_create failed");
			free(ip->almanac);
			ip->almanac = ip->almanac_ptr = NULL;
		}
		return;
	}
	if (s == 5 && p == 1) {
		if (ip->almanac == NULL) {
			ip->almanac = calloc(1, 34 * mp->rs_len);
			if (ip->almanac == NULL) {
				nomem();
				return;
			}
		}
		ip->almanac_ptr = ip->almanac;
	}
	if (ip->almanac == NULL)
		return;
	memcpy(ip->almanac_ptr, ip->serial_buf, mp->rs_len);
	ip->almanac_ptr += mp->rs_len;
	/* are we done yet ? */
	if (ip->almanac_ptr != ip->almanac + 34 * mp->rs_len)
		return;

	Log(Linfo, "Got full almanac\n");

	if (ip->almanac_file == NULL)
		return;

	ip->almanac_work = ALMANAC_SAVE;
	i = pthread_create(&ip->almanac_thread, NULL, Cb_thread, ip);
	if (i != 0) {
		warn("pthread_create failed");
		free(ip->almanac);
		ip->almanac = ip->almanac_ptr = NULL;
	}
}

/*
 * Response from set to defaults
 */

static void
handle_Cf(evContext ctx, struct instance *ip, struct oncore_msg *mp __unused)
{

	if (ip->state == ST_RESET_WAIT)
		oncore_state(ctx, ip, ST_TEST);
	else
		oncore_state(ctx, ip, ST_RESET);
}

/*
 * Response to self test
 */

static void
handle_Ia(evContext ctx, struct instance *ip, struct oncore_msg *mp __unused)
{
	uint32_t st;

	st = be32dec(ip->serial_buf + 4) >> 8;
	if (st == 0) {
		oncore_state(ctx, ip, ST_INIT);
		return;
	}
	Log(Lwarn, "Receiver failed selftest: %03x", st);
	oncore_state(ctx, ip, ST_SICK);
}

static void
handle_Bb(evContext ctx __unused, struct instance *ip, struct oncore_msg *mp __unused)
{
	int i, j, k, ns;
	u_char *buf = ip->serial_buf;

	for (i = 0; i < NSAT; i++) {
		ip->persat[i].doppler = 0;
		ip->persat[i].elevation = 0;
		ip->persat[i].azimuth = 0;
		ip->persat[i].health = 2;
	}
	
	ns = buf[4];
	if (ns > 12)
		ns = 12;
	for (i = 0; i < ns; i++) {
		j = buf[5 + i * 7];
		if (j == 0)
			continue;
		if (j > 32) {
			Log(Lwarn, "SVID %d in @@Bb\n", j);
			continue;
		}
		k =  be16dec(buf + 6 + i * 7);
		if (k > 32767)
			k -= 65536;
		ip->persat[j].doppler = k;
		ip->persat[j].elevation = buf[8 + i * 7];
		ip->persat[j].azimuth = be16dec(buf + 9 + i * 7);
		ip->persat[j].health = buf[11 + i * 7];
	}
}

static void
handle_Ha(evContext ctx __unused, struct instance *ip, struct oncore_msg *mp __unused)
{
	int i, j;
	u_char *buf = ip->serial_buf;
	char txt[50], txt2[50];
	const char *p;
	struct ts ts2[2];

	i = PpsPickup(&ip->pps, ts2, 1);
	ip->ts = ts2[0];
	if (i == 1) {
		p = TsFormat(&ip->ts, txt);
	} else {
		p = "no_pps_pulse";
	}
	if (ip->almanac_work == ALMANAC_LOADED) {
		/*
		 * The entire almanac is too big to be loaded in one second
		 * so load only half of it each second
		 */
		for (j = 0; j < NALMANAC / 2 && ip->almanac_n < NALMANAC; j++) {
			m12_sendmsg(ip, ip->almanac_ptr + 2, LALMANAC - 5);
			ip->almanac_ptr += LALMANAC;
			ip->almanac_n++;
		}
		/*
		 * Set state back to almanac collection, the receiver will
		 * return the almanac we just loaded to us.
		 */
		if (ip->almanac_n == NALMANAC) {
			free(ip->almanac);
			ip->almanac = NULL;
			ip->almanac_ptr = NULL;
			ip->almanac_work = ALMANAC_DONE;
		}
	}
	if (ip->almanac_work == ALMANAC_DONE) {
		/*
		 * We are done with the almanac thread, join it to recover
		 * the status block.
		 * Based on a theory on last minute mistaks, ask for a
		 * leap-second status, just in case...
		 */
		pthread_join(ip->almanac_thread, NULL);
		ip->almanac_work = ALMANAC_COLLECT;
		m12_sendmsg(ip, "Gj", 2);     /* Leap sec status */
		
	}
	for (i = 0; i < NSAT; i++) {
		ip->persat[i].mode = 0;
		ip->persat[i].signal_strength = 0;
		ip->persat[i].iode = 0;
		ip->persat[i].status = 0;
	}
	ip->tm.tm_mon = buf[4] - 1;
	ip->tm.tm_mday = buf[5];
	ip->tm.tm_year = be16dec(buf + 6) - 1900;
	ip->tm.tm_hour = buf[8];
	ip->tm.tm_min = buf[9];
	ip->tm.tm_sec = buf[10];
	ip->fraction = be32dec(buf + 11);
	ip->flat = be32dec(buf + 15);
	ip->flon = be32dec(buf + 19);
	ip->fht = be32dec(buf + 23);
	ip->lat = be32dec(buf + 31);
	ip->lon = be32dec(buf + 35);
	ip->ht = be32dec(buf + 39);
	ip->dop = be16dec(buf + 53);
	ip->vsat = buf[55];
	ip->tsat = buf[56];
	for (i = 0; i < 12; i++) {
		j = buf[57 + i * 6];
		if (j == 0)
			continue;
		ip->persat[j].mode = buf[58 + i * 6];
		ip->persat[j].signal_strength = buf[59 + i * 6];
		ip->persat[j].iode = buf[60 + i * 6];
		ip->persat[j].status = be16dec(buf + 61 + i * 6);
	}
	ip->rcv_status = be16dec(buf + 129);
	ip->clock_bias = be16dec(buf + 133);
	ip->osc_offset = be32dec(buf + 135);
	ip->osc_temp = be16dec(buf + 139);
	ip->lutc_status = ip->utc_status;
	ip->utc_status = buf[141];
	ip->utc_sign = buf[142];
	ip->utc_hour = buf[143];
	ip->utc_min = buf[144];
	strlcpy(ip->serno, buf + 145, sizeof ip->serno);

	ip->t = timegm(&ip->tm);

	/*
	 * Every so often, the Oncore will mis-estimate things and end
	 * up being too early.  Since the fraction field cannot be negative
	 * what happens is that the time fields are for the previous second
	 * and the fraction field is almost a full second.
	 * To fix this we add a second if the fraction is above 900msec,
	 * a number which is a pure guess.
	 */
	if (ip->fraction > 900000000)
		ip->t++;

	TsNormalize(&ip->ts);
	if (ip->utc_status & 0x40) {
		/*
		 * UTC mode
		 * The fraction is a proper fraction of the seconds we have
		 * so we treat it as unsigned, and adjust for the seconds.
		 */
		ip->delta = ip->ts.ts.bt.frac;
		ip->delta /= 18446744073709551616.0; /* 2^64 in case you wondered */
		ip->delta += ip->ts.ts.bt.sec - ip->t;
	} else {
		/*
		 * Fraction mode
		 * we have no usable seconds part, so we treat the fractions
		 * as a signed number and report it directly.
		 */
		ip->delta = (int64_t)ip->ts.ts.bt.frac;
		ip->delta /= 18446744073709551616.0; /* 2^64 in case you wondered */
	}

	gmtime_r(&ip->t, &ip->tm);

	/*
	 * Check for LeapSecond status once per hour
	 */
	if (ip->tm.tm_min == 7 && ip->tm.tm_sec == 22)
		m12_sendmsg(ip, "Gj", 2);     /* Leap sec status */

	strftime(txt2, sizeof txt, "%Y-%m-%d %H:%M:%S", &ip->tm);
	Log(lm12_msg,
	    "%s %ld %12.9f %s  %d of %d stat (%x %x %x %x %x %x %x %x %x %x %x %x)  %d %d %d %d",
	    p, (long)ip->t, ip->delta, txt2,
	    ip->tsat, ip->vsat,
	    (ip->rcv_status >> 13) & 7,
	    (ip->rcv_status >> 11) & 3,
	    (ip->rcv_status >> 10) & 1,
	    (ip->rcv_status >> 9) & 1,
	    (ip->rcv_status >> 8) & 1,
	    (ip->rcv_status >> 7) & 1,
	    (ip->rcv_status >> 6) & 1,
	    (ip->rcv_status >> 5) & 1,
	    (ip->rcv_status >> 4) & 1,
	    (ip->rcv_status >> 3) & 1,
	    (ip->rcv_status >> 1) & 3,
	    (ip->rcv_status >> 0) & 1,
	    ip->utc_status, ip->utc_sign, ip->utc_hour, ip->utc_min);
}

static void
handle_Hn(evContext ctx __unused, struct instance *ip, struct oncore_msg *mp __unused)
{
	int i, j;
	u_char *buf = ip->serial_buf;

	/* Copy stuff we want to retain */
	for (i = 1; i < NSAT; i++)
		ip->persat[i].offset = 0;
	for (i = 0; i < NCHAN; i++) {
		j = buf[15 + i * 5];
		if (j == 0)
			continue;
		ip->persat[j].offset = be32dec(buf + 16 + i * 5);
	}

	do {
		/* If we're not really ready anyway, bail */
		if (ip->state != ST_SYNC_WAIT && ip->state != ST_SYNC)
			break;

		/* If RAIM is not happy, bail */
		if (buf[6] != 0)
			break;

		/*
		 * If the utc_status is not stable, bail.  (This is becuase
		 * the flag gets set one second earlier than the offset gets
		 * applied)
		 */
		if (ip->utc_status != ip->lutc_status)
			break;

		/* one tracked sat, one vote */
		SourceParam(ip->sid, SRC_VOTES, ip->tsat);

		/*
		 * Record dispersion.  We have 1sigma number, arbitrarily
		 * define dispersion as two-sigma.
		 */
		j = be16dec(buf + 12);
		SourceParam(ip->sid, SRC_DISPERSION, 2 * j * 1e-9);

		if (ip->utc_status & 0x40) {
			SourceDelta(ip->sid, &ip->ts, ip->delta, SRC_UTC);
		} else {
			SourceDelta(ip->sid, &ip->ts, ip->delta, SRC_FRACTION);
		}
	} while (0);
}

static void
handle_Gj(evContext ctx __unused, struct instance *ip, struct oncore_msg *mp __unused)
{
	u_char *buf = ip->serial_buf;

	Log(lm12_msg,
	    "LEAP: p %d f %d Y %04u M %u D %d i.f %u.%09u h %u m %u s %u",
		buf[4], buf[5],
		be16dec(buf + 6), buf[8], buf[9],
		buf[10], be32dec(buf + 11),
		buf[15], buf[16], buf[17]);
		
}

int
init_oncore_tables(void)
{
	char const **cp;
	struct oncore_msg *mp;
	int i;

	oncore_msgs = calloc(sizeof *oncore_msgs, NMASH);
	if (oncore_msgs == NULL) {
		nomem();
		return (-1);
	}
	for (cp = oncore_m12_request; *cp != NULL; cp++) {
		i = msg_mash(*cp);
		mp = oncore_msgs[i];
		if (mp == NULL) {
			mp = calloc(sizeof *mp, 1);
			if (mp == NULL)
				err(1, "calloc() failed");
			oncore_msgs[i] = mp;
		} 
		mp->rq_len = strlen(*cp);
		mp->rq_format = *cp;
	}
	for (cp = oncore_m12_response; *cp != NULL; cp++) {
		i = msg_mash(*cp);
		mp = oncore_msgs[i];
		if (mp == NULL) {
			mp = calloc(sizeof *mp, 1);
			if (mp == NULL)
				err(1, "calloc() failed");
			oncore_msgs[i] = mp;
		}
		mp->rs_len = strlen(*cp);
		mp->rs_format = *cp;
	}
   // register message handlers
	oncore_msgs[msg_mash("@@Bb")]->handler = handle_Bb;
	oncore_msgs[msg_mash("@@Cb")]->handler = handle_Cb;
	oncore_msgs[msg_mash("@@Ch")]->handler = handle_NOOP;
	oncore_msgs[msg_mash("@@Ha")]->handler = handle_Ha;
	oncore_msgs[msg_mash("@@Hn")]->handler = handle_Hn;
	oncore_msgs[msg_mash("@@Cf")]->handler = handle_Cf;
	oncore_msgs[msg_mash("@@Ia")]->handler = handle_Ia;
	oncore_msgs[msg_mash("@@Gj")]->handler = handle_Gj;
	return (0);
}
