/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: oncore_m12_serial.c,v 1.3 2005/01/26 15:28:53 tom Exp $
 */

#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <termios.h>
#include <pthread.h>

#include <ntpns.h>

#include "oncore_m12.h"


static void serial_input(evContext ctx, void *uap, int fd, int eventmask);

// configure serial communication, open the device etc
int
m12_serial(evContext ctx __unused, FILE *f, void *priv __unused, char **tok, int flag)
{
	struct instance *ip;
        int fd, i;
        struct termios t;

	if (tok[0] == NULL || flag & CFG_HELP) {
		ConfigHelp(f, "m12 $instance serial $filename");
		return (0);
	}
	ip = priv;
	if (flag & CFG_NO) {
		fprintf(f, "XXX: NOT m12 serial\n");
		return (0);
	}
	if (tok[0] == NULL || (flag & CFG_SHOW)) {
		fprintf(f, "m12 %u serial %s\n", ip->unit, ip->serial);
		return (0);
	}

        fd = open(tok[0], O_RDWR | O_NDELAY);
        if (fd < 0) {
                fprintf(f, "open(%s): %s", tok[0], strerror(errno));
		return (-1);
	}
        i = tcgetattr(fd, &t);
        if (i < 0) {
		fprintf(f, "tcgetattr(%s): %s", tok[0], strerror(errno));
		close(fd);
		return (-1);
	}
        cfmakeraw(&t);
        cfsetspeed(&t, B9600);
        t.c_cflag &= ~CSIZE;
        t.c_cflag |= CS8;
        t.c_cflag |= CLOCAL;
        t.c_cc[VMIN] = 0;
        t.c_cc[VTIME] = 20;
        i = tcsetattr(fd, TCSANOW, &t);
        if (i < 0) {
		fprintf(f, "tcsetattr(%s): %s", tok[0], strerror(errno));
		close(fd);
		return (-1);
	}
   // register serial_input as serial event handler
	i = evSelectFD(ctx, fd, EV_READ | EV_EXCEPT, serial_input,
	    ip, &ip->serial_fid);
	if (i) 
		err(1, "evSelectFD(%s)", tok[0]);
	ip->serial_fd = fd;
	ip->serial = strdup(tok[0]);
	ip->serial_idx = 0;
	oncore_state(ctx, ip, ST_CONFIG);
	return (0);
}


// serial input function, reads in a message from the serial port.
// Called from the eventlib, event is set up in m12_serial.
// Complete messages are passed to the message handler (oncore_m12_msg.c).
static void
serial_input(evContext ctx, void *uap, int fd, int eventmask __unused)
{
	struct instance *ip;
	int i, l, mm;
	struct oncore_msg *mp;
	u_char s;

	ip = uap;

	/*
	 * Read as much as we can fit in our buffer
	 */
	i = read(fd, ip->serial_buf + ip->serial_idx,
	     sizeof(ip->serial_buf) - ip->serial_idx);
	if (i <= 0)
		return;
	ip->serial_idx += i;

	for(;;) {
		/*
		 * Stop if don't have enough data for a complete message
		 * the shortest possible message is 7 bytes: "@@xxC\r\n"
		 */
		if (ip->serial_idx < 7)
			return;

		/*
		 * See if we know this one and if the header looks OK
		 */
		mm = msg_mash(ip->serial_buf);
		mp = oncore_msgs[mm];
		if (ip->serial_buf[0] != '@' ||
		    ip->serial_buf[1] != '@' ||
		    mp == NULL ||
		    mp->rs_len == 0) {
			/*
			 * Look for a '\n' and up to and including that,
			 * otherwise, skip entire buffer.
			 */
			for (i = 0; i < ip->serial_idx; i++)
				if (ip->serial_buf[i] == '\n')
					break;
			if (i < ip->serial_idx)
				i++;
			record("M12_DISC", ip->unit, ip->serial_buf, i);
			Log(lm12_serial, "%02x %02x %02x %02x skipping %d chars\n",
			    ip->serial_buf[0], ip->serial_buf[1],
			    ip->serial_buf[2], ip->serial_buf[3], i);
			memcpy(ip->serial_buf,
			    ip->serial_buf + i,
			    ip->serial_idx - i);
			ip->serial_idx -= i;
			continue;
		}

		l = mp->rs_len;

		/*
		 * Do we have all the bytes in this message yet ?
		 */
		if (ip->serial_idx < l)
			return;

		/* Calculate XOR checksum */
		s = 0;
		for (i = 2; i < l - 2; i++)
			s ^= ip->serial_buf[i];

		if (s != 0) {
			Log(lm12_serial, "wrong XOR at end of %4.4s (%02x/%02x)",
			    mp->rs_format, ip->serial_buf[l - 3], s);
			record("M12_XOR", ip->unit, ip->serial_buf, l);
		} else if (ip->serial_buf[l - 2] != '\r') {
			Log(lm12_serial, "no CR at end of %4.4s (%02x)",
			    mp->rs_format, ip->serial_buf[l - 2]);
			record("M12_CR", ip->unit, ip->serial_buf, l);
		} else if (ip->serial_buf[l - 1] != '\n') {
			Log(lm12_serial, "no LF at end of %4.4s (%02x)",
			    mp->rs_format, ip->serial_buf[l - 1]);
			record("M12_LF", ip->unit, ip->serial_buf, l);
		} else {
			record("M12_OK", ip->unit, ip->serial_buf, l);
			Log(lm12_serial, "Rcv: %4.4s[%d]", ip->serial_buf, l);
			if (mp->handler != NULL)
				mp->handler(ctx, ip, mp);
			if (ip->state == ST_LISTEN)
				oncore_state(ctx, ip, ST_CONTACT);
			if (ip->state == ST_SYNC_WAIT ||
			    ip->state == ST_SYNC)
				oncore_reset_timer(ctx, ip);
		}

		/*
		 * Pull up anything at the end and start over
		 */
		memcpy(ip->serial_buf, ip->serial_buf + l, ip->serial_idx - l);
		ip->serial_idx -= l;
	}
}

// send out stuff
void
m12_sendmsg(struct instance *ip, const void *p, u_int len)
{
	u_char cs = 0;
	const u_char *msg;

	msg = p;
	Log(lm12_serial, "Xmt: @@%2.2s[%d]", msg, len + 5);
	write(ip->serial_fd, "@@", 2);
	write(ip->serial_fd, msg, len);
	while (len--)
		cs ^= *msg++;
	write(ip->serial_fd, &cs, 1);
	write(ip->serial_fd, "\r\n", 2);
}
