/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: oncore_m12_state.c,v 1.2 2005/01/26 15:28:53 tom Exp $
 */

#include <err.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/endian.h>
#include <pthread.h>

#include <ntpns.h>

#include "oncore_m12.h"

static const char *
state_str(enum oncore_state state)
{
	switch (state) {
	case ST_UNCONF:		return "<unconfigured>";
	case ST_CONFIG:		return "<config>";
	case ST_START:		return "<start>";
	case ST_LISTEN:		return "<listen>";
	case ST_CONTACT:	return "<contact>";
	case ST_POLL_STATUS:	return "<poll_status>";
	case ST_RESET:		return "<reset>";
	case ST_RESET_WAIT:	return "<reset_wait>";
	case ST_TEST:		return "<test>";
	case ST_TEST_WAIT:	return "<test_wait>";
	case ST_SICK:		return "<sick>";
	case ST_INIT:		return "<init>";
	case ST_SYNC_WAIT:	return "<sync_wait>";
	case ST_SYNC:		return "<sync>";
	default:
		Abort("Wrong oncore state %d", state);
	}
}

static void
timer_reset(evContext ctx, void *uap, struct timespec due __unused, struct timespec inter __unused)
{

	Log(lm12_debug, "oncore reset timer fired\n");
	oncore_state(ctx, uap, ST_RESET);
}

void
oncore_reset_timer(evContext ctx, struct instance *ip)
{
	int i, seconds;

	switch (ip->state) {
	case ST_LISTEN:
	case ST_SYNC_WAIT:
	case ST_SYNC:
		seconds = 5;
		break;
	case ST_POLL_STATUS:
	case ST_RESET_WAIT:
	case ST_TEST_WAIT:
		seconds = 12;
		break;
	default:
		seconds = 0;
		break;
	}
	if (evTestID(ip->reset_timer)) {
		i = evClearTimer(ctx, ip->reset_timer);
		if (i != 0)
			err(1, "evClearTimer(oncore_reset_timer) failed");
		evInitID(&ip->reset_timer);
	}
	if (seconds == 0)
		return;
	i = evSetTimer(ctx, timer_reset, ip, 
	    evAddTime(evNowTime(), evConsTime(seconds, 0)),
	    evConsTime(0, 0), &ip->reset_timer);
	if (i != 0)
		err(1, "evSetTimer(oncore_reset_timer, %d) failed", seconds);
}

void
oncore_state(evContext ctx, struct instance *ip, enum oncore_state state)
{

	if (ip->state == state)
		return;
	Log(lm12_msg, "M12: %s -> %s", state_str(ip->state), state_str(state));

	ip->state = state;

	oncore_reset_timer(ctx, ip);
	switch (ip->state) {
	case ST_UNCONF:
		return;
	case ST_CONFIG:
		if (ip->serial_fd >= 0)
			oncore_state(ctx, ip, ST_START);
		else 
			oncore_state(ctx, ip, ST_UNCONF);
		return;
	case ST_START:
		m12_sendmsg(ip, "Cj", 2); /* Receiver ID */
		oncore_state(ctx, ip, ST_LISTEN);
		break;
	case ST_LISTEN:
		break;
	case ST_CONTACT:
		m12_sendmsg(ip, "Bb\000", 3); /* Sat visibility */
		m12_sendmsg(ip, "Bd\000", 3); /* Almanac status */
		m12_sendmsg(ip, "Be\000", 3); /* Req Almanac */
		m12_sendmsg(ip, "Bj\000", 3); /* Leap sec status */
		m12_sendmsg(ip, "Bo\000", 3); /* Req UTC offset */
		m12_sendmsg(ip, "Bp\000", 3); /* Req UTC/Ion data */
		m12_sendmsg(ip, "Gc\003", 3); /* 1pps */
		m12_sendmsg(ip, "Gd\000", 3); /* pos mode (set pos mode)*/
		m12_sendmsg(ip, "Ge\000", 3); /* TRAIM (disable)*/
		m12_sendmsg(ip, "Ha\000", 3); /* pos msg */
		m12_sendmsg(ip, "Hn\000", 3); /* TRAIM msg */
		oncore_state(ctx, ip, ST_POLL_STATUS);
		break;
	case ST_POLL_STATUS:
		/* We just wait for the reset timer to fire */
		break;
	case ST_RESET:
		m12_sendmsg(ip, "Cf", 2); /* Set to defaults */
		oncore_state(ctx, ip, ST_RESET_WAIT);
		break;
	case ST_RESET_WAIT:
		/*
		 * wait for reset timer to fire (again) or @@Cf
		 * message to arrive
		 */
		break;
	case ST_TEST:
		m12_sendmsg(ip, "Ia", 2); /* Self Test */
		oncore_state(ctx, ip, ST_TEST_WAIT);
		break;
	case ST_TEST_WAIT:
		/*
		 * wait for reset timer to fire (again) or @@Ia
		 * message to arrive
		 */
		break;
	case ST_INIT:
		oncore_initialize(ip);
		oncore_state(ctx, ip, ST_SYNC_WAIT);
		break;
	case ST_SYNC_WAIT:
	case ST_SYNC:
		/*
		 * Process messages, or take timeout.
		 */
		break;
	default:
		LogErr(Lerror, "XXX: state %s not yet coded\n", state_str(ip->state));
		break;
	}
}
