/*
   trimble.c
   NTPns driver for the Trimble Acutime 2000 GPS Smart Antenna.

   Main module, sets up the driver.

   Author: Thomas Wana <tom@ripe.net>

   $Id: trimble.c,v 1.12 2005/01/26 15:28:53 tom Exp $
 */

#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <pthread.h>

#include <ntpns.h>

#include "trimble.h"

struct logh *trimble_debug, *trimble_serial;
struct logh *trimble_raw, *trimble_clockstats;
struct logh *trimble_pktid, *trimble_inith;

static TAILQ_HEAD(, instance)	instance_head =
    TAILQ_HEAD_INITIALIZER(instance_head);

/**********************************************************************/

// returns the instance with the unit-ID 'unit' from the list 
// of instances, or creates a new one (when 'create' is non-null,
// else returns NULL).
// Note that every module has its own struct instance
static struct instance *trimble_instance(evContext ctx, u_int unit, int create)
{
	struct instance *ip;

	ctx=ctx;

	TAILQ_FOREACH(ip, &instance_head, next) {
		if (ip->unit == unit)
			return (ip);
	}
	if (!create)
		return (NULL);
	ip = calloc(sizeof *ip, 1);
	if (ip == NULL)
		return (NULL);
	ip->unit = unit;
	ip->serial_fd = -1;
	PpsCreate(&ip->pps, 1);                     // use the assert flank (one flank, default)
	TAILQ_INSERT_TAIL(&instance_head, ip, next);
	ip->sid = SourceAdd("trimble_%u", unit);
	SourceRefid(ip->sid, "GPS\0");
	SourceParam(ip->sid, SRC_UPDATE_RATE, 1);   // timing updates come in every second
	SourceParam(ip->sid, SRC_LOS_TIMEOUT, 10);  // after 10sec we consider this source lost ("loss of source")

	ip->state=ST_UNCONFIGURED;
	return (ip);
}

// PPS configuration, just passed through to NTPns
static int trimble_pps(evContext ctx __unused, FILE *f, void *priv __unused, char **tok, int flag)
{
	struct instance *ip;

	if (tok[0] == NULL || flag & CFG_HELP) {
		ConfigHelp(f, "trimble $instance pps {pps_param}");
		return (0);
	}
	ip = priv;
	return (PpsConfig(ctx, f, &ip->pps, tok, flag));
}

// Request status (status actually printed in trimble_protocol)
static int trimble_status(evContext ctx __unused, FILE *f, void *priv, char **tok, int flag __unused)
{
	struct instance *ip;
   int all;

	ip = priv;

   // TODO check state, check ip->unit
   if(tok[0]==NULL)
   {
      fprintf(f, "Available subcommands: all satellites health receiver position time pps\n");
      return -1;
   }

   all=!strcmp(tok[0], "all");

   if(!strcmp(tok[0], "satellites") || all)
   {
      trimble_sendcode(ip, 0x24);
      trimble_sendcode(ip, 0x27);
   }
   else if(!strcmp(tok[0], "health") || all)
   {
      trimble_sendcode(ip, 0x26);
   }
   else if(!strcmp(tok[0], "receiver") || all)
   {
      trimble_sendcode_subcode(ip, 0xBB, 0x00);
   }
   else if(!strcmp(tok[0], "position") || all)
   {
      trimble_sendcode_subcode(ip, 0x8E, 0x20);
   }
   else if(!strcmp(tok[0], "time") || all)
   {
      // Request UTC time
      trimble_sendcode_subcode(ip, 0x8E, 0xAD);
   }
   else if(!strcmp(tok[0], "pps") || all)
   {
      // Request PPS pulse width
      trimble_sendcode_subcode(ip, 0x8E, 0x4F);
   }
   /* else if(!strcmp(tok[0], "reboot"))
   {
	   // reboot the trimble
	   // it's under 'status' because I am too lazy to introduce an extra command
	   // for this debug thingie
   } */
   else
   {
      fprintf(f, "command error\n");
      return -1;
   }
   
	return 0;
}

static struct configlist *trimble_cfl;

// master dispatch for all the command line arguments, gets the right
// instance from the instance list and calls the appropriate config
// function.
static int trimble_dispatch(evContext ctx, FILE *f, void *priv __unused, char **tok, int flag)
{
	u_int u;
	char *p;
	struct instance *ip;

	if (tok[0] == NULL) {
		fprintf(f, "Specify Trimble unit number\n");
		return (-1);
	}
	u = strtoul(tok[0], &p, 0);
	if (*p != '\0') {
		fprintf(f, "Trimble unit number not numeric\n");
		return (0);
	}
	if (flag & CFG_NO) {
		ip = trimble_instance(ctx, u, 0);
		if (ip != NULL)
			fprintf(f, "XXX: no Trimble\n");
		return (0);
	}
	ip = trimble_instance(ctx, u, 1);
	if (ip == NULL)
		return (ENOMEM);

	return (ConfigCall(trimble_cfl, ctx, f, ip, tok + 1, flag));
}

ref_init_t trimble_initialize;

const char *trimble_initialize(evContext ctx __unused)
{
	static int once = 0;

	if (once) { 
		warnx("trimble module already loaded");
		return (NULL);
	} else {
		once = 1;
	}

	trimble_debug = LogNewHandle(Ldebug, "trimble");
	trimble_serial = LogNewHandle(trimble_debug, "serial");
	trimble_raw = LogNewHandle(trimble_debug, "raw");
	trimble_pktid = LogNewHandle(trimble_debug, "pktid");
	trimble_inith = LogNewHandle(trimble_debug, "init");
	trimble_clockstats = LogNewHandle(trimble_debug, "clockstats");

	ConfigAdd(NULL, "trimble", trimble_dispatch, "refclock: Trimble Acutime 2000 GPS Smart Antenna");
	trimble_cfl = ConfigCreate();
	ConfigAdd(trimble_cfl, "serial", trimble_serconf, "Serial port");
	ConfigAdd(trimble_cfl, "pps", trimble_pps, "PPS-API configuration");
	ConfigAdd(trimble_cfl, "status", trimble_status, "Retrieve trimble status");

	Log(trimble_debug, "Trimble driver for NTPns initialized.");
	return (MAGIC);
}
