/*
   trimble_protocol.c
   NTPns driver for the Trimble Acutime 2000 GPS Smart Antenna.

   Protocol layer. Does the logic stuff.

   Author: Thomas Wana <tom@ripe.net>

   $Id: trimble_protocol.c,v 1.18 2005/01/26 15:28:53 tom Exp $
 */

#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <termios.h>
#include <pthread.h>
#include <stdint.h>
#include <arpa/inet.h>
#include <math.h>

#include <ntpns.h>

#include "trimble.h"

// Initialization of the trimble is done using an array of prepared request packets.
// Only if all packets have been sent and an appropriate response has been seen, the
// initialization is considered complete.
// This struct is one item in the initialization chain.
struct trimble_init_item {
	int sent;          // packet sent?
	int resp;          // got response?
	TSIPPKT pkt;       // packet to send
	TSIPPKT pkt_resp;  // packet we got back. 
	char respcode;     // response code to expect
	char respsubcode;  // subcode to expect; if -1 then we expect no subcode
};

// This array holds all the initialization items and resembles the init-chain.
struct trimble_init_item trimble_init[40];
int trimble_init_len=0;

static int clockstats_counter=0;

// we keep a history of the 6 most recent packet codes
// so we can judge if 0x82 is really a warm reboot by
// looking at the previous sequence of packet IDs.
// (see manual A-2). Only looking at 0x82 isn't enough
// because 0x82 can also occur during normal operations.
//
// It is a ring buffer with the most recent packet code
// at pktid_history[0].
static char pktid_history[6];

/**************************************************************************************************/

static void protocol_error(TSIPPKT *pkt)
{
   Log(Lerror, "Protocol error: packet code 0x%02x\n",pkt->code);
   trimble_hexdump(Lerror, pkt->buf, (long)pkt->len);
}

/**************************************************************************************************/

// heartbeat-timer for trimble; does periodic tasks (but not time
// measurement - this is triggered by the trimble itself)
static void trimble_timer(evContext ctx, void *uap, struct timespec due __unused, struct timespec inter __unused)
{
	int i;
	struct instance *ip=(struct instance *)uap;

	if(ip->state==ST_TRIMBLE_POS_FIX || ip->state==ST_TRIMBLE_CLOCK)
	{
		// check if we are still / already have overdetermined clock mode
		trimble_sendcode_subcode(ip, 0xBB, 0x00); // Query Primary Receiver Configuration
	}

	// reschedule timer
	if(evTestID(ip->trimble_timer))
	{
		i=evClearTimer(ctx, ip->trimble_timer);
		if(i!=0)
			Log(Lerror, "evClearTimer(trimble_timer) failed");
		evInitID(&ip->trimble_timer);
	}
	i=evSetTimer(ctx, trimble_timer, ip, 
			evAddTime(evNowTime(), evConsTime(TRIMBLE_TIMER, 0)),
			evConsTime(0,0), &ip->trimble_timer);
	if(i!=0)
		Log(Lerror, "evSetTimer(trimble_timer) failed");
}

/**************************************************************************************************/


// Handle the 0x8F-AD (UTC Superpacket) message
// Called every second.
static void handle_0x8FAD(struct instance *ip, TSIPPKT *pkt)
{
	const char *status[]={"Receiver is navigating",
		"Receiver is timing using one satellite",
		"Approximate time",
		"Start-Up", 
		"Start-Up",
		"Dilution of Precision too high",
		"Satellite is unusable",
		"No satellites usable",
		"Only 1 satellite usable",
		"Only 2 satellites usable",
		"Only 3 satellites usable",
		"Invalid solution",
		"Differential corections",
		"Overdetermined fixes"};
	int d,mon,y,h,min,s;
	double ds;
	struct ts timestamp, pps[2];
	struct tm tm;
	time_t the_time;
	char buf[50], rs;
	double delta;
	int i;
	static FILE *fpdelta=NULL;

	if(fpdelta==NULL)
	{
		fpdelta=fopen("/tmp/delta.log","w");
	}

	if(pkt->len!=22) {
		protocol_error(pkt);
		return;
	}

	LogNoNewline(trimble_serial, "trimble: UTC time is: ");
	Log(trimble_serial, "%02u/%02u/%u, %02u:%02u:%02u.%.f",
			(mon=pkt->buf[15]), (d=pkt->buf[14]), (y=trimble_bGetUShort(pkt->buf+16)),
			(h=pkt->buf[11]), (min=pkt->buf[12]), (s=pkt->buf[13]), (ds=trimble_bGetDouble(pkt->buf+3)));
	Log(trimble_serial, "trimble: UTC: tracking status: %s", status[pkt->buf[18]]);
	// Log(trimble_serial, "trimble: fractional seconds: %f", trimble_bGetDouble(pkt->buf+3));
	LogNoNewline(trimble_serial, "trimble: UTC: leap second flags: ");
	if(pkt->buf[19]&1) LogNoNewline(trimble_serial, "UTC time is available, ");
	if(pkt->buf[19]&16) LogNoNewline(trimble_serial, "Leap second scheduled, ");
	if(pkt->buf[19]&32) LogNoNewline(trimble_serial, "Leap pending (24 hours), ");
	if(pkt->buf[19]&64) LogNoNewline(trimble_serial, "GPS leap warning (+- 6 hours), ");
	if(pkt->buf[19]&128) LogNoNewline(trimble_serial, "Leap in progress (WOW this is coincidence :) make a screenshot!), ");
	Log(trimble_serial, " ");

	if(ip->state!=ST_TRIMBLE_CLOCK)
	{
		Log(trimble_serial, "trimble: clock state is not 'overdetermined fixes', ignoring time report");
		return;
	}
	
	rs=pkt->buf[18];
	if(rs<0 || rs>14) rs=14;
	if( (rs>=2 && rs<=7) || rs==11 || rs==12 || rs==14)
	{
		Log(trimble_serial, "trimble: tracking status insufficient; falling back to "
				"position fix state and ignoring time report");
		ip->state=ST_TRIMBLE_POS_FIX;
		return;
	}

	// convert the time into seconds after epoch (UTC)
	tm.tm_sec=s;
	tm.tm_min=min;
	tm.tm_hour=h;
	tm.tm_mday=d;
	tm.tm_mon=mon-1;
	tm.tm_year=y-1900;
	tm.tm_wday=-1;
	tm.tm_yday=-1;
	tm.tm_isdst=0;
	tm.tm_gmtoff=0;
	the_time=timegm(&tm);

	// populate a struct tm;
	memset(&timestamp, 0, sizeof(timestamp));
	timestamp.fmt=TS_TIMEVAL;
	timestamp.ts.tv.tv_sec=(long)the_time;
	timestamp.ts.tv.tv_usec=0;

	// Get the time the last PPS assert came in
	i=PpsPickup(&ip->pps, pps, 1);
	if(i==1)
	{
		Log(trimble_serial, "trimble: PPS result assert: %s", TsFormat(pps, buf));
	}
	else
	{
		Log(trimble_serial, "trimble: no PPS signal: %d; skipping this measurement",i);
		return;
	}

	// We know that PPS always comes in on the full second, so the delta
	// between the system time and UTC time is the fractional part of 
	// the PPS. We also correct the second - for use when there are
	// UTC leap seconds etc.
	TsNormalize(pps);
	delta=pps[0].ts.bt.frac;
	delta/=18446744073709551616.0; /* 2^64 in case you wondered */
	delta+=pps[0].ts.bt.sec - the_time;

	// feed the time into the Source
	if(pkt->buf[19]&1)      // "UTC time is available"
	{
		ip->utc_lock=1;
		SourceDelta(ip->sid, &timestamp, delta, SRC_UTC);
	}
	else
	{
		ip->utc_lock=0;
		SourceDelta(ip->sid, &timestamp, delta, SRC_FRACTION);
	}

	/*
	 * Time critical stuff over, do what we want.
	 */
	Log(trimble_serial, "trimble: delta = %f us", delta*1000*1000);
	Log(trimble_serial, " ");
	if(fpdelta!=NULL)
	{
		fprintf(fpdelta, "%f\n",delta*1000*1000);
		fflush(fpdelta);
	}

	if(clockstats_counter%CLOCKSTATS_LOG_FREQUENCY==0)
	{
		// Do logging
		Log(trimble_clockstats, "U1 %u.%u.%u %02u:%02u:%02u %u %02x",
				d, mon, y, h, min, s, pkt->buf[18], pkt->buf[19]);
	}
	// global clockstats_counter is advanced in handle_0x8F0B
}

/**************************************************************************************************/

// Handle the 0x8F-0B (UTC Comprehensive Time) message
// Called every second.
static void handle_0x8F0B(struct instance *ip, TSIPPKT *pkt)
{
        unsigned day, month, year;      /* data derived from received timecode */
        unsigned hour, minute, second;
        unsigned utcoff;
        unsigned char mode;
        double  bias, rate;
        float biasunc, rateunc;
        double lat, lon, alt;
        short lat_deg, lon_deg;
        float lat_min, lon_min;
        unsigned char north_south, east_west;
        char sv[9];

        int i, votes;
        double tow;

	if(ip->state!=ST_TRIMBLE_CLOCK)
	{
		Log(trimble_serial, "trimble: clock state is not 'overdetermined fixes', ignoring comprehensive time report");
		return;
	}

	clockstats_counter++;

	// not the right time to log a line to clockstats?
	if((clockstats_counter-1)%CLOCKSTATS_LOG_FREQUENCY != 0)
		return;

	if(pkt->len!=74) {
		protocol_error(pkt);
	       	return;
	}

        if (trimble_bGetUShort(pkt->buf+1) != 0)   // zero for PPS
                return;

        tow =  trimble_bGetDouble(pkt->buf+3);

        if (tow == -1.0)        // UTC/GPS time of week (in seconds)
                return;
        else if ((tow >= 604800.0) || (tow < 0.0)) {
                return;
        }
        else
        {
                if (tow < 604799.9) tow = tow + .00000001;
                second = (unsigned int) fmod(tow, 60.);
                minute =  (unsigned int) fmod(tow/60., 60.);
                hour = (unsigned int )fmod(tow / 3600., 24.);
        }


        day =           (unsigned int) pkt->buf[11];
        month =         (unsigned int) pkt->buf[12];
        year =          trimble_bGetUShort(pkt->buf+13);
        mode =          pkt->buf[15];
        utcoff =        trimble_bGetUShort(pkt->buf+16);
        bias =          trimble_bGetDouble(pkt->buf+18) / GPS_C * 1e9;     /* ns */
        rate =          trimble_bGetDouble(pkt->buf+26) / GPS_C * 1e9;     /* ppb */
        biasunc =       trimble_bGetSingle(pkt->buf+34) / GPS_C * 1e9;     /* ns */
        rateunc =       trimble_bGetSingle(pkt->buf+38) / GPS_C * 1e9;     /* ppb */
        lat =           trimble_bGetDouble(pkt->buf+42) * R2D;
        lon =           trimble_bGetDouble(pkt->buf+50) * R2D;
        alt =           trimble_bGetDouble(pkt->buf+58);

        if (lat < 0.0) {
                north_south = 'S';
                lat = -lat;
        }
        else {
                north_south = 'N';
        }
        lat_deg = (short)lat;
        lat_min = (lat - lat_deg) * 60.0;

        if (lon < 0.0) {
                east_west = 'W';
                lon = -lon;
        }
        else {
                east_west = 'E';
        }

        lon_deg = (short)lon;
        lon_min = (lon - lon_deg) * 60.0;

	votes=0;
        for (i=0; i<8; i++) {
                sv[i] = pkt->buf[i + 66];
                if (sv[i]) {
			// request satellite tracking status, to be logged to clockstats
			trimble_sendcode_subcode(ip, 0x3C, (unsigned char)abs(sv[i]));
			if(sv[i]>0)
				votes++;
                }
        }

	SourceParam(ip->sid, SRC_VOTES, votes);
	Log(trimble_serial, "trimble: vote change: vote is now %u", votes);

        Log(trimble_clockstats, "C1 %02d%02d%04d %02d%02d%02d %d %7.0f %.1f %.0f %.1f %d %02d%09.6f %c %02d%09.6f %c %.0f  %d %d %d %d %d %d %d %d",
                day, month, year, hour, minute, second, mode, bias, biasunc, rate, rateunc, utcoff,
                lat_deg, lat_min, north_south, lon_deg, lon_min, east_west, alt,
                sv[0], sv[1], sv[2], sv[3], sv[4], sv[5], sv[6], sv[7]);
}

/**************************************************************************************************/

static void handle_0x5C(struct instance *ip __unused, TSIPPKT *pkt)
{
	unsigned char prn, channel, aqflag, ephstat;
	float snr, azimuth, elevation;

	if(pkt->len!=24) {
		protocol_error(pkt);
		return;
	}

	prn=pkt->buf[0];
	channel=pkt->buf[1]>>3;
	if(channel==0x10)
		channel=2;
	else
		channel++;

	aqflag=pkt->buf[2];
	ephstat=pkt->buf[3];
	snr=trimble_bGetSingle(pkt->buf+4);
	elevation=trimble_bGetSingle(pkt->buf+12)*R2D;
	azimuth=trimble_bGetSingle(pkt->buf+16)*R2D;

	if(ip->state!=ST_TRIMBLE_CLOCK)
	{
		Log(trimble_serial, "trimble: clock state is not 'overdetermined fixes', ignoring satellite report");
		return;
	}

	Log(trimble_clockstats, "S1 %02d %d %d %02x %4.1f %5.1f %4.1f",
		prn, channel, aqflag, ephstat, snr, azimuth, elevation);
}

/**************************************************************************************************/

// chooses a packet to send from the initialization-array, and switches the
// state if the initialization is done.
static void choose_and_send_trimble_init(evContext ctx __unused, struct instance *ip)
{
	int i;
	struct timespec dummy;

	// choose an unsent packet to be sent out
	for(i=0;i<trimble_init_len;i++)
	{
		if(trimble_init[i].sent==0)
		{
			Log(trimble_inith, "%s: sending out code 0x%02x, subcode 0x%02x", __FUNCTION__, 
					trimble_init[i].pkt.code, trimble_init[i].pkt.buf[0]);
			trimble_sendmsg(ip, &trimble_init[i].pkt);
			trimble_init[i].sent=1;
			return; // only send one packet at a time
		}
	}

	// we only come here if all packets are marked as sent. Now check if
	// we got a response from all.
	for(i=0;i<trimble_init_len;i++)
	{
		if(trimble_init[i].resp==0)
		{
			// no response, resend this packet.
			Log(trimble_inith, "%s: RESENDING code 0x%02x, subcode 0x%02x", __FUNCTION__, 
					trimble_init[i].pkt.code, trimble_init[i].pkt.buf[0]);
			trimble_sendmsg(ip, &trimble_init[i].pkt);
			trimble_init[i].sent=1;
			return; // only send one packet at a time
		}
	}

	// ok, all packets sent, got a reply from all of them, so bump up
	// to the next state.
	
	ip->state=ST_TRIMBLE_POS_FIX;
	Log(trimble_serial, "trimble: initialization successfully done, switching state "
			"to position fix");

	// install timer
	trimble_timer(ctx, ip, dummy, dummy);
}

/**************************************************************************************************/

// Additional timing information, like PPS quantization error etc
static void handle_0x8FAC(struct instance *ip __unused, TSIPPKT *pkt)
{
	Log(trimble_serial, "trimble: Self-Survey progress: %d%%",
			pkt->buf[3]);
	Log(trimble_serial, "trimble: PPS quantization error: %f ns",
			trimble_bGetSingle(pkt->buf+60));
	Log(trimble_serial, "trimble: Estimate of UTC offset of local clock: %f ns",
			trimble_bGetSingle(pkt->buf+16));
	Log(trimble_serial, "trimble: PPS generated: %s",
			pkt->buf[64]==1 ? "yes" : "no");
	
	// TODO store these values in the instance and do something useful
	// with it, particularly with the offsets
}

/**************************************************************************************************/

// Fill the trimble_init struct with the necessary packets, basically prepare the whole setup.
// And install the init-timer (see there for details)
static void trimble_initialize_trimble(evContext ctx, struct instance *ip)
{
	TSIPPKT pkt;
	unsigned long mask;
	int i=0;
	
	memset(trimble_init, 0, sizeof(trimble_init));

	// Software Version Request
	trimble_init[i].pkt.code=0x1F;
	trimble_init[i].pkt.len=0;
	trimble_init[i].respcode=0x45;
	trimble_init[i].respsubcode=-1;
	i++;

	// Health status
	trimble_init[i].pkt.code=0x26;
	trimble_init[i].pkt.len=0;
	trimble_init[i].respcode=0x4B;
	trimble_init[i].respsubcode=-1;
	i++;

	// Query Production Parameters
	trimble_init[i].pkt.code=0x8E;
	trimble_init[i].pkt.len=1;
	trimble_init[i].pkt.buf[0]=0x42;
	trimble_init[i].respcode=0x8F;
	trimble_init[i].respsubcode=0x42;
	i++;

	// Query Manufacturing Operating Parameters
	trimble_init[i].pkt.code=0x8E;
	trimble_init[i].pkt.len=1;
	trimble_init[i].pkt.buf[0]=0x41;
	trimble_init[i].respcode=0x8F;
	trimble_init[i].respsubcode=0x41;
	i++;

	// Set operating parameters
	{
		float elevation_angle, signal_level, PDOP_mask, PDOP_switch;

		memset(&pkt, 0, sizeof(pkt));
		pkt.len=17;
		pkt.code=0x2C;

		pkt.buf[0]=4;             // Dynamics: static
		elevation_angle=0.1745;   // default (10 degrees)
		signal_level=4.0;         // default
		PDOP_mask=8.0;            // default
		PDOP_switch=6.0;          // default
		trimble_bPutFloat(&elevation_angle, pkt.buf+1);
		trimble_bPutFloat(&signal_level, pkt.buf+5);
		trimble_bPutFloat(&PDOP_mask, pkt.buf+9);
		trimble_bPutFloat(&PDOP_switch, pkt.buf+13);
	}
	trimble_init[i].pkt=pkt;
	trimble_init[i].respcode=0x4C;
	trimble_init[i].respsubcode=-1;
	i++;

	// Set IO Options
	{
		memset(&pkt, 0, sizeof(pkt));
		pkt.len=4;
		pkt.code=0x35;
		pkt.buf[0]=0x1C;  // position: 00011100
				  //              |||||+-- XYZ ECEF Output off
				  //              ||||+--- LLA Output off,
				  //              |||+---- LLA ALT Output: MSL geoid WGS-84,
				  //              ||+----- ALT input: MSL geoid WGS-84
				  //              |+------ Precision-of-position output: Double-precision packet 83 and/or 84
				  //              +------- 8F-20 output: off
		pkt.buf[1]=0x00;  // velocity: 00000000
				  //                 |+--- XYZ ECEF Output off,
				  //                 +---- ENU Output: off
		pkt.buf[2]=0x05;  // timing:   00001001
				  //              ||||+--- Time type: UTC
				  //              |||+---- Fix computation time: ASAP
				  //              ||+----- Fix output time: When computed
				  //              |+------ Synchronized measurements: On
				  //              +------- Minimize projection: off
		pkt.buf[3]=0x02;  // auxiliary:00000010
				  //                 |+--- Raw measurements: off
				  //                 +---- Doppler smoothed codephase: smoothed
	}
	trimble_init[i].pkt=pkt;
	trimble_init[i].respcode=0x55;
	trimble_init[i].respsubcode=-1;
	i++;

	// Turn off the timing port (Port A)
	memset(&pkt, 0, sizeof(pkt));
	pkt.len=6;
	pkt.code=0x3D;     // Timing port configuration command
	pkt.buf[0]=0x0B;   // transmit baud rate: 9600
	pkt.buf[1]=0x0B;   // receive baud rate: 9600
	pkt.buf[2]=0x07;   // Parity and #bits: 8N
	pkt.buf[3]=0x07;   // Stop bits: 1 stop bit
	pkt.buf[4]=0x01;   // Language mode for transmission: Off
	pkt.buf[5]=0x00;   // Language mode for reception: TSIP
	trimble_init[i].pkt=pkt;
	trimble_init[i].respcode=0x3D;
	trimble_init[i].respsubcode=-1;
	i++;

	// Set PPS characteristics
	{
		double d;
		float f;
		memset(&pkt, 0, sizeof(pkt));

		pkt.code=0x8E;
		pkt.len=16;

		pkt.buf[0]=0x4A;     // subcode: 0x4A
		pkt.buf[1]=0x01;     // PPS Driver Switch: on
		pkt.buf[2]=0x01;     // Time Base: UTC
		pkt.buf[3]=0x00;     // PPS Polarity: positive

		// calculate the PPS offset to compensate the cable propagation delay:
		// the receiver by default assumes a cable length of 100 meters. Given
		// that the signal propagates at a speed of 4.1011ns per meter, we have to
		// adjust the delay relative to the predefined 100 meters, so this will
		// be a negative offset for cable lengths < 100m and positive for >100m.
		d=(ip->pps_cable_length-100.0)*4.1011/1000.0/1000.0/1000.0;
		f=300.0;
		trimble_bPutDouble(&d, &pkt.buf[4]);
		trimble_bPutFloat(&f, &pkt.buf[12]);
	}
	trimble_init[i].pkt=pkt;
	trimble_init[i].respcode=0x8F;
	trimble_init[i].respsubcode=0x4A;
	trimble_init[i].resp=1;          // the trimble doesn't send a reply on this request,
	                                 // so we pretend to have already seen it
	i++;

	// Set Position Fix Mode
	/* trimble_init[i].pkt.code=0x22;
	trimble_init[i].pkt.len=1;
	trimble_init[i].pkt.buf[0]=10;
	trimble_init[i].respcode=0x54;
	trimble_init[i].respsubcode=-1;
	i++; */
	
	// Set the automatic packet output mask
	trimble_init[i].pkt.code=0x8E;
	trimble_init[i].pkt.len=5;
	trimble_init[i].pkt.buf[0]=0x4D;
	mask=0;       // Auto-output mask (see A-92 for details)
	trimble_bPutULong(&mask, &trimble_init[9].pkt.buf[1]);
	trimble_init[i].respcode=0x8F;
	trimble_init[i].respsubcode=0x4D;
	i++;

	// Query Primary Receiver Configuration
	trimble_init[i].pkt.code=0xBB;
	trimble_init[i].pkt.len=1;
	trimble_init[i].pkt.buf[0]=0x00;
	trimble_init[i].respcode=0xBB;
	trimble_init[i].respsubcode=0x00;
	i++;

	// Set PPS pulse width to 100ms
	{
		double width;
		memset(&pkt, 0, sizeof(pkt));

		pkt.code=0x8E;
		pkt.len=9;

		pkt.buf[0]=0x4F;
		width=0.1;
		trimble_bPutDouble(&width, &pkt.buf[1]);
	}
	trimble_init[i].pkt=pkt;
	trimble_init[i].respcode=0x8F;
	trimble_init[i].respsubcode=0x4F;
	i++;

	// Set Packet Broadcast Mask
	trimble_init[i].pkt.code=0x8E;
	trimble_init[i].pkt.len=5;
	trimble_init[i].pkt.buf[0]=0xA5;
	// 0x8F-AD every second (PPS aligned) on Port B
	// 0x8F-0B every second (PPS aligned) on Port B
	trimble_init[i].pkt.buf[1]=0x88;
	// 0x8F-AC every second on both ports
	trimble_init[i].pkt.buf[2]=0x80;
	trimble_init[i].respcode=0x8F;
	trimble_init[i].respsubcode=0xA5;
	i++;

	// Let 0x8F-AD only be coming on PPS pulses (and not external events)
	trimble_init[i].pkt.code=0x8E;
	trimble_init[i].pkt.len=2;
	trimble_init[i].pkt.buf[0]=0xAD;
	trimble_init[i].pkt.buf[1]=1;
	trimble_init[i].respcode=0x8F;
	trimble_init[i].respsubcode=0xA5;
	i++;
	
	Log(trimble_serial, "trimble: configuration packets: %d", i);
	trimble_init_len=i;

	// switch state and start the initialization
	ip->state=ST_SERIAL_SETUP_COMPLETE;
	choose_and_send_trimble_init(ctx, ip);
}

/**************************************************************************************************/

// initialize this module.
void trimble_protocol_init(evContext ctx, struct instance *ip)
{
	memset(pktid_history, 0, sizeof(pktid_history));
	evInitID(&ip->trimble_timer);

	trimble_initialize_trimble(ctx, ip);
}

/**************************************************************************************************/

// Implements the protocol logic of the trimble. It is called
// whenever a new complete message is read from the serial port and
// gets the packet as argument 'pkt'.
void trimble_do_protocol(evContext ctx, struct instance *ip, TSIPPKT *pkt)
{
	int i;
	
	// log the packet ID to a special log handler.
	// (msg enable trimble debug pktid)
	LogNoNewline(trimble_pktid, "Incoming packet ID: 0x%02x", pkt->code);
	if(pkt->len>0)
		Log(trimble_pktid, " possible subcode: 0x%02x",pkt->buf[0]);
	else
		Log(trimble_pktid, " ");

	// copy the packet code into the history buffer. Advance the
	// ring buffer first.
	for(i=sizeof(pktid_history)-1;i>=1;i--)
		pktid_history[i]=pktid_history[i-1];
	pktid_history[0]=pkt->code;

	if(ip->state==ST_SERIAL_SETUP_COMPLETE)
	{
		// check off the response packets in the trimble_init array
		for(i=0;i<trimble_init_len;i++)
		{
			char code, subcode;
			code=pkt->code;
			subcode=(pkt->len>0 ? pkt->buf[0] : -1);
			if(trimble_init[i].sent==1 && trimble_init[i].resp==0 &&
				trimble_init[i].respcode==code &&
				(trimble_init[i].respsubcode==subcode || trimble_init[i].respsubcode==-1))
			{
				trimble_init[i].resp=1;
				trimble_init[i].pkt_resp=*pkt;

				Log(trimble_inith, "%s: got response packet code %02x",
						__FUNCTION__, code);
			}
		}
		// choose a new packet to be sent
		choose_and_send_trimble_init(ctx, ip);
	}

	// now dissect the packet and act upon it.
	if(pkt->code==0x45)  // Software Version Information Report
	{
		if(pkt->len!=10) {
			protocol_error(pkt);
			return;
		}

		Log(trimble_serial, "trimble: Navigation processor version: %u.%u (%u/%u/%u)",pkt->buf[0], pkt->buf[1],
				pkt->buf[2], pkt->buf[3], pkt->buf[4]+1900);
		Log(trimble_serial, "trimble: Signal processor revision: %u.%u (%u/%u/%u)",pkt->buf[5], pkt->buf[6],
				pkt->buf[7], pkt->buf[8], pkt->buf[9]+1900);
	}
	else if(pkt->code==0x46) // Health of Receiver Report
	{
		const char *status[]={"Doing position fixes",
			"Do not have GPS time yet",
			"(unknown code 0x02)",
			"PDOP is too high",
			"(unknown code 0x04)",
			"(unknown code 0x05)",
			"(unknown code 0x06)",
			"(unknown code 0x07)",
			"No usable satellites",
			"Only 1 usable satellite",
			"Only 2 usable satellites",
			"Only 3 usable satellites",
			"The chosen satellite is unusable"};
		char error;
		if(pkt->len!=2) {
			protocol_error(pkt);
			return;
		}
		if(pkt->buf[0]>sizeof(status)/sizeof(status[0])) {
			protocol_error(pkt);
			return;
		}

		Log(trimble_serial, "trimble: health status (0x%02x): %s", pkt->buf[0], status[pkt->buf[0]]);
		error=pkt->buf[1];
		LogNoNewline(trimble_serial, "trimble: health status: error code (0x%02x): ", error);
		if(error&1) LogNoNewline(trimble_serial, "Battery back-up failed (Note: always set), ");
		if(error&2) LogNoNewline(trimble_serial, "Signal Processor error, ");
		if(error&4) LogNoNewline(trimble_serial, "Alignment error, channel or chip 1, ");
		if(error&8) LogNoNewline(trimble_serial, "Alignment error, channel or chip 2, ");
		if(error&16) LogNoNewline(trimble_serial, "Antenna feed line fault (Open or Short), ");
		if(error&32) LogNoNewline(trimble_serial, "Excessive reference frequency error, ");
		if(error&64 || error&128) LogNoNewline(trimble_serial, "(unknown error code), ");
		Log(trimble_serial, " ");
	}
	else if(pkt->code==0x4B) // Machine/Code ID and Additional Status Report
	{
		if(pkt->len!=3) {
			protocol_error(pkt);
		       	return;
		}
		LogNoNewline(trimble_serial, "trimble: machine ID: 0x%02x, ",pkt->buf[0]);
		LogNoNewline(trimble_serial, "status 1: ");
		if(pkt->buf[1]&2) LogNoNewline(trimble_serial, "Battery Powered Time Clock Fault");
		if(pkt->buf[1]&8) LogNoNewline(trimble_serial, "The Alamanc stored in the receiver is not complete and current");
		Log(trimble_serial, " ");
	}
	else if(pkt->code==0x8F && pkt->len>0 && pkt->buf[0]==0x42) // Production Parameters
	{
		if(pkt->len!=19) {
			protocol_error(pkt);
			return;
		}

		Log(trimble_serial, "trimble: production parameters:");
		Log(trimble_serial, "trimble: Production options prefix: 0x%02x", pkt->buf[1]);
		Log(trimble_serial, "trimble: Production number extension: 0x%02x", pkt->buf[2]);
		Log(trimble_serial, "trimble: Case serial number prefix: %u",trimble_bGetUShort(pkt->buf+3));
		Log(trimble_serial, "trimble: Case serial number: %lu",trimble_bGetULong(pkt->buf+5));
		Log(trimble_serial, "trimble: Production number: %lu",trimble_bGetULong(pkt->buf+9));
		Log(trimble_serial, "trimble: Machine identification number: %u",trimble_bGetUShort(pkt->buf+15));
	}
	else if(pkt->code==0x8F && pkt->len>0 && pkt->buf[0]==0x41) // Manufacturing Operating Parameters
	{
		if(pkt->len!=17) {
			protocol_error(pkt);
		       	return;
		}

		Log(trimble_serial, "trimble: manufacturing operating parmeters:");
		Log(trimble_serial, "trimble: Board serial number prefix: %u",trimble_bGetUShort(pkt->buf+1));
		Log(trimble_serial, "trimble: Board serial number: %lu",trimble_bGetULong(pkt->buf+3));
		Log(trimble_serial, "trimble: Date of build: %u/%u/%u %u:00:00",
				pkt->buf[7], pkt->buf[8], pkt->buf[9]+1900, pkt->buf[10]);
		Log(trimble_serial, "trimble: Oscillator offset: %f", trimble_bGetSingle(pkt->buf+11));
		Log(trimble_serial, "trimble: Test code identification number: %u",trimble_bGetUShort(pkt->buf+15));

	}
	else if(pkt->code==0x4C)
	{
		if(pkt->len!=17) {
			protocol_error(pkt);
			return;
		}

		LogNoNewline(trimble_serial, "trimble: operating parameters: ");
		LogNoNewline(trimble_serial, "dynamics: %u, ", 
				pkt->buf[0]);
		LogNoNewline(trimble_serial, "elevation angle mask: %f, ", trimble_bGetSingle(pkt->buf+1));
		LogNoNewline(trimble_serial, "signal level mask: %f, ", trimble_bGetSingle(pkt->buf+5));
		LogNoNewline(trimble_serial, "PDOP mask: %f, ", trimble_bGetSingle(pkt->buf+9));
		Log(trimble_serial, "PDOP switch: %f", trimble_bGetSingle(pkt->buf+13));
	}
	else if(pkt->code==0x55) // Response to I/O options
	{
	}
	else if(pkt->code==0x3D) // Response to a timing port configuration command
	{
	}
	else if(pkt->code==0x8F && pkt->len>0 && pkt->buf[0]==0x4A) // Response to a set PPS characteristics message
	{
		Log(trimble_serial, "trimble: PPS cable length offset (cable length = %fm): %f ns",
				ip->pps_cable_length, trimble_bGetDouble(pkt->buf+4)*1000.0*1000.0*1000.0);
	}
	else if(pkt->code==0x8F && pkt->len>0 && pkt->buf[0]==0x4D) // Response to a Automatic Packet Output Mask request
	{
	}
	else if(pkt->code==0xBB && pkt->len>0 && pkt->buf[0]==0x00) // Primary Receiver Configuration
	{
		const char *opdim[]={"Automatic", "Time only (1-SV)", "", "Horizontal (2D)",
			"Full position (3D)", "DGPS reference", "2D clock hold", "Overdetermined clock"};
		const char *dyn[]={"", "Land, <120 knots", "Sea, <50 knots", "Air <800 knots", "Static"};

		Log(trimble_serial, "trimble: operating dimension: %s", opdim[pkt->buf[1]]);
		Log(trimble_serial, "trimble: DGPS mode: %s", pkt->buf[2]==0 ? "DGPS off" : (pkt->buf[2]==1 ? "DGPS only" : "DGPS auto"));
		Log(trimble_serial, "trimble: dynamics mode: %s", dyn[pkt->buf[3]]);
		Log(trimble_serial, "trimble: elevation mask: %f (%f degrees)", trimble_bGetSingle(pkt->buf+5),
				trimble_bGetSingle(pkt->buf+5)*180.0/GPS_PI);
		Log(trimble_serial, "trimble: AMU mask: %f", trimble_bGetSingle(pkt->buf+9));
		Log(trimble_serial, "trimble: PDOP mask: %f", trimble_bGetSingle(pkt->buf+13));
		Log(trimble_serial, "trimble: PDOP switch: %f", trimble_bGetSingle(pkt->buf+17));
		Log(trimble_serial, "trimble: DGPS age: %u sec", pkt->buf[21]);

		if(ip->state==ST_TRIMBLE_POS_FIX && pkt->buf[1]==7)
		{
			// Trimble completed the position fix, we are now in overdetermined clock mode
			Log(trimble_serial, "trimble: got position fix, new state is overdetermined clock mode");
			ip->state=ST_TRIMBLE_CLOCK;
		}
		else if(ip->state==ST_TRIMBLE_CLOCK && pkt->buf[1]!=7)
		{
			// we lost overdetermined clock mode ... fall back to position fix
			Log(trimble_serial, "trimble: lost overdetermined clock mode, new state is position fix mode");
			ip->state=ST_TRIMBLE_POS_FIX;
		}
	}
	else if(pkt->code==0x8F && pkt->len>0 && pkt->buf[0]==0x4B) // Response to Query Survey Limit
	{
		Log(trimble_serial, "trimble: survey limit: %ld", trimble_bGetULong(pkt->buf+1));
	}
	else if(pkt->code==0x6D) // All-in-view satellite selection report
	{
		const char *mode[]={"Auto", "Time Only (1-SV)", "2D Clock Hold", "2D", "3D", 
			"Overdetermined Clock", "DGPS Reference"};

		if(pkt->len<17) {
			protocol_error(pkt);
			return;
		}

		Log(trimble_serial, "trimble: mode: %s, %s", mode[pkt->buf[0] & 0x7], 
				(pkt->buf[0]&4)==4 ? "manual" : "auto");
		Log(trimble_serial, "trimble: number of satellites: %u",pkt->buf[0] >> 4);
		Log(trimble_serial, "trimble: PDOP: %f (%f degrees)",
				trimble_bGetSingle(pkt->buf+1), trimble_bGetSingle(pkt->buf+1)*180.0/GPS_PI);
		Log(trimble_serial, "trimble: HDOP: %f (%f degrees)",
				trimble_bGetSingle(pkt->buf+5), trimble_bGetSingle(pkt->buf+5)*180.0/GPS_PI);
		Log(trimble_serial, "trimble: VDOP: %f (%f degrees)",
				trimble_bGetSingle(pkt->buf+9), trimble_bGetSingle(pkt->buf+9)*180.0/GPS_PI);
		Log(trimble_serial, "trimble: TDOP: %f (%f degrees)",
				trimble_bGetSingle(pkt->buf+13), trimble_bGetSingle(pkt->buf+13)*180.0/GPS_PI);

		// for all satellites ...
		for(i=0;i<pkt->buf[0]>>4;i++)
		{
			Log(trimble_serial, "trimble: satellite #%u: 0x%02x", i+1, pkt->buf[16+i]);
		}

		// set the vote count for this source to the amount of usable satellites
		SourceParam(ip->sid, SRC_VOTES, pkt->buf[0]>>4);
		Log(trimble_serial, "trimble: vote change: vote is now %u",pkt->buf[0]>>4);

	}
	else if(pkt->code==0x8F && pkt->len>0 && pkt->buf[0]==0xA5) // Packet Broadcast Mask report
	{
		if(pkt->len!=5) {
			protocol_error(pkt);
			return;
		}

		if(pkt->buf[1]&0x04) Log(trimble_serial, "trimble: Event output of 0x8F-0B on port B");
		if(pkt->buf[1]&0x08) Log(trimble_serial, "trimble: Synchronous output of 0x8F-0B on port B");
		if(pkt->buf[1]&0x40) Log(trimble_serial, "trimble: Event output of 0x8F-AD on port B");
		if(pkt->buf[1]&0x80) Log(trimble_serial, "trimble: Synchronous output of 0x8F-AD on port B");
	}
	else if(pkt->code==0x8F && pkt->len>0 && pkt->buf[0]==0xAC) // Supplemental Timing Information
	{
		if(pkt->len!=68)
		{
			protocol_error(pkt);
			return;
		}

		handle_0x8FAC(ip, pkt);
	}
	else if(pkt->code==0x47)      // Signal Levels for all Satellites Report
	{
		if(pkt->len!=pkt->buf[0]*5+1) {
			protocol_error(pkt); 
			return;
		}

		Log(trimble_serial, "trimble: Signal Levels for all satellites report:");

		for(i=0;i<pkt->buf[0];i++)
		{
			Log(trimble_serial, "trimble: satellite #%u signal level: %f",
					pkt->buf[1+i*5], trimble_bGetSingle(pkt->buf+2+i*5));
		}
	}
	else if(pkt->code==0x8F && pkt->len>0 && pkt->buf[0]==0x20) // Last position fix report
	{
		int velocity_scaling;
		if(pkt->len!=56) {
			protocol_error(pkt); 
			return;
		}

		if((pkt->buf[24]&1)==1)
			velocity_scaling=4;
		else
			velocity_scaling=1;

		Log(trimble_serial, "trimble: last position fix:");
		Log(trimble_serial, "trimble: east velocity %f m/s",0.005*trimble_bGetUShort(pkt->buf+2)*velocity_scaling);
		Log(trimble_serial, "trimble: north velocity %f m/s",0.005*trimble_bGetUShort(pkt->buf+4)*velocity_scaling);
		Log(trimble_serial, "trimble: up velocity %f m/s",0.005*trimble_bGetUShort(pkt->buf+6)*velocity_scaling);
		Log(trimble_serial, "trimble: time of week: %ld ms",trimble_bGetULong(pkt->buf+8));
		Log(trimble_serial, "trimble: longitude: %f", (float)trimble_bGetULong(pkt->buf+16)/2147483648.0);
		Log(trimble_serial, "trimble: altitude: %f m", (float)trimble_bGetLong(pkt->buf+20)/1000.0);
		LogNoNewline(trimble_serial, "trimble: fix type: ");
		LogNoNewline(trimble_serial, "%s, ",(pkt->buf[27]&1 ? "No fix available" : "Fix was available"));
		LogNoNewline(trimble_serial, "%s, ",(pkt->buf[27]&2 ? "Fix was corrected with RTCM" : "Fix is autonomous"));
		LogNoNewline(trimble_serial, "%s, ",(pkt->buf[27]&4 ? "2D fix" : "3D fix"));
		LogNoNewline(trimble_serial, "%s, ",(pkt->buf[27]&8 ? "2D fix used entered altitude" : "2D fix used last-circulated altitude"));
		Log(trimble_serial, "%s",(pkt->buf[27]&16 ? "Position or altitude filter on" : "Unfiltered"));
		Log(trimble_serial, "trimble: Number of satellites used for fix: %u", pkt->buf[28]);
	}
	else if(pkt->code==0x8F && pkt->len>0 && pkt->buf[0]==0xAD) // UTC time superpacket
	{
		handle_0x8FAD(ip, pkt);
	}
	else if(pkt->code==0x8F && pkt->len>0 && pkt->buf[0]==0x0B) // UTC time comprehensive packet
	{
		handle_0x8F0B(ip, pkt);
	}
	else if(pkt->code==0x8F && pkt->len>0 && pkt->buf[0]==0x4F) // PPS pulse width report
	{
		if(pkt->len!=9) {
			protocol_error(pkt);
		       	return;
		}

		Log(trimble_serial, "trimble: PPS pulse width is %f seconds", trimble_bGetDouble(pkt->buf+1));
	}
	else if(pkt->code==0x5C) // Satellite tracking status report
	{
		if(ip->state==ST_TRIMBLE_CLOCK)
			handle_0x5C(ip, pkt);
	}
	else if(pkt->code==0x82) // DGPS position fix mode
	{
		// check if this comes in as part of the reboot
		// sequence (see A-2 in the manual).
		// btw., this packet code is nowhere documented in the docs.
		if(!memcmp(pktid_history, "\x82\x84\x83\x45\x4B\x46", 6))
		{
			Log(trimble_serial, "trimble: (re)booting? changing state to initialization.");
			trimble_initialize_trimble(ctx, ip);
		}
	}
	else
	{
		LogNoNewline(trimble_serial, "UNIMPLEMENTED packet code: 0x%02x", pkt->code);
		if(pkt->len>0)
			Log(trimble_serial, " possible subcode: 0x%02x",pkt->buf[0]);
		else
			Log(trimble_serial, " ");
	}
}
