/*
   trimble_serial.c
   NTPns driver for the Trimble Acutime 2000 GPS Smart Antenna.

   Serial communication layer. Has an interface to send packets and contains the
   handler for serial communication. Complete packets will be passed to the
   protocol layer (trimble_protocol.c)

   Author: Thomas Wana <tom@ripe.net>

   $Id: trimble_serial.c,v 1.12 2005/01/26 15:28:53 tom Exp $
 */


#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <termios.h>
#include <pthread.h>

#include <ntpns.h>

#include "trimble.h"

static void serial_input(evContext ctx, void *uap, int fd, int eventmask);

// A routine that prints a buffer in 'hexdump -C'-style
void trimble_hexdump(struct logh *loghdl, unsigned char *buffer, long len)
{
   unsigned long pos=0;

   while(len>0)
   {
      int i;
      LogNoNewline(loghdl, "%08x  ",(unsigned int)pos);
      for(i=0;i<16 && i<len;i++)
      {
         LogNoNewline(loghdl,"%02x ",buffer[pos+i]);
         if(i==7 || i==15)
            LogNoNewline(loghdl," ");
      }
      LogNoNewline(loghdl,"|");
      for(i=0;i<16 && i<len;i++)
      {
         unsigned char c=buffer[pos+i];
         if( (c>='A' && c<='Z') || (c>='a' && c<='z') || (c>='0' && c<='9') ||
             c==':' || c=='!' || c=='?' || c=='=' || c==';' || c=='<' ||
             c=='>' || c=='/' || c=='\\' || c=='&')
         {
            LogNoNewline(loghdl,"%c",c);
         }
         else
         {
            LogNoNewline(loghdl,".");
         }
      }
      Log(loghdl,"|");
      len-=16;
      pos+=16;
   }
}

// configure serial communication, open the device etc.
// serial line configuration according to NTP 4.2.0/ntpd/refclock_ripencc.c (flags combined)
int trimble_serconf(evContext ctx, FILE *f, void *priv __unused, char **tok, int flag)
{
	struct instance *ip;
	int fd, i;
	struct termios t;

	if (tok[0] == NULL || tok[1]==NULL || strcmp(tok[1], "cablelength") || tok[2]==NULL || flag & CFG_HELP) {
		ConfigHelp(f, "trimble $instance serial $filename cablelength $length_in_meters");
		return (0);
	}
	ip = priv;
	if (flag & CFG_NO) {
		fprintf(f, "XXX: NOT trimble serial\n");
		return (0);
	}
	if (tok[0] == NULL || (flag & CFG_SHOW)) {
		fprintf(f, "trimble %u serial %s cablelength %f\n", ip->unit, ip->serial, ip->pps_cable_length);
		return (0);
	}

	ip->pps_cable_length=strtod(tok[2], NULL);
	Log(trimble_serial, "trimble: cable length is %f meters", ip->pps_cable_length);

	LogNoNewline(trimble_serial, "trimble: opening serial device...");
	fd = open(tok[0], O_RDWR | O_NDELAY);
	if (fd < 0) {
		fprintf(f, "open(%s): %s", tok[0], strerror(errno));
		return (-1);
	}
	i = tcgetattr(fd, &t);
	if (i < 0) {
		fprintf(f, "tcgetattr(%s): %s", tok[0], strerror(errno));
		close(fd);
		return (-1);
	}
	t.c_cflag=PARENB|PARODD|CS8|CLOCAL|CREAD;
	t.c_oflag=0;
	t.c_iflag=0;
	t.c_lflag=0;
	t.c_cc[VMIN]=1;

	cfsetispeed(&t, B9600);
	cfsetospeed(&t, B9600);

	i = tcsetattr(fd, TCSANOW, &t);
	if (i < 0) {
		fprintf(f, "tcsetattr(%s): %s", tok[0], strerror(errno));
		close(fd);
		return (-1);
	}

	if(tcflush(fd, TCIOFLUSH)<0)
	{
		fprintf(f, "tcflush(%s): %s", tok[0], strerror(errno));
		close(fd);
		return -1;
	} 
	Log(trimble_serial, " serial device opened and configured");

	// register serial_input as serial event handler
	i = evSelectFD(ctx, fd, EV_READ | EV_EXCEPT, serial_input,
			ip, &ip->serial_fid);
	if (i) 
		err(1, "evSelectFD(%s)", tok[0]);
	ip->serial_fd = fd;
	ip->serial = strdup(tok[0]);
	ip->serial_idx = 0;
	ip->state=ST_SERIAL_SETUP_COMPLETE;
	Log(trimble_serial, "trimble: serial setup complete, initializing");

	// initialize the trimble
	trimble_protocol_init(ctx, ip);

	return (0);
}


// serial input function, reads in a message from the serial port.
// Called from the eventlib, event is set up in trimble_serial.
static void serial_input(evContext ctx, void *uap, int fd, int eventmask __unused)
{
	struct instance *ip;
	int i, l, remember_0x10;
	char *ptr, *optr;
	char recvbuf[2048];

	ip = uap;
	ctx=ctx;

	/*
	 * Read as much as we can fit in our buffer
	 */
	i = read(fd, ip->serial_buf + ip->serial_idx, sizeof(ip->serial_buf) - ip->serial_idx);
	if (i <= 0)
		return;
	Log(trimble_raw, "%s: activity on serial port, read:",__FUNCTION__);
	trimble_hexdump(trimble_raw, ip->serial_buf+ip->serial_idx, i);

	ip->serial_idx += i;
	Log(trimble_raw, "%s: ip->serial_buf is now:",__FUNCTION__);
	trimble_hexdump(trimble_raw, ip->serial_buf, ip->serial_idx);
	{
		int end_seen;
		/*
		 * Stop if we don't have enough data for a complete message
		 * the shortest possible message is 4 bytes: 0x10 0xCode 0x10 0x03
		 */
		if (ip->serial_idx < 4)
			return;

		// what do we have in the buffer ...
		if(ip->serial_buf[0]!=0x10)
		{
			// fsck ... we are definitely not aligned to a packet
			Log(trimble_serial,"Serial buffer doesn't start with a packet!");

			// move up the contents of the packet until
			// we find a 0x10 (it's not perfect but better
			// than nothing). Or if there is no 0x10, return.
			do
			{
				ip->serial_idx--;
				memcpy(ip->serial_buf, ip->serial_buf+1,
						ip->serial_idx);
			} while(ip->serial_idx>0 && ip->serial_buf[0]!=0x10);

			if(ip->serial_idx==0) {
				// we couldn't find anything ...
				Log(trimble_serial, "couldn't recover buffer");
				return;
			}

			Log(trimble_serial, "buffer recovered");
		}

		// check if there is a whole packet in the buffer and copy it out
		memset(recvbuf, 0, sizeof(recvbuf));
		ptr=ip->serial_buf;
		optr=recvbuf;

		// strategy: go through the buffer bytewise. If there is a 0x10, remember
		// that there was one and do nothing. If in the next iteration there is 
		// another 0x10, unstuff it. If there is a 0x03, the packet is done.
		for(i=2, remember_0x10=0, end_seen=0;i<ip->serial_idx && !end_seen;i++)
		{
			if(ptr[i]==0x10)
			{
				if(remember_0x10==0)
					remember_0x10=1;
				else
				{
					remember_0x10=0;
				}
			}
			else if(ptr[i]==0x03 && remember_0x10==1)
			{
				// end of packet
				end_seen=1;
			}
			else if(remember_0x10==1)
			{
				Log(trimble_serial, "Packet error: no 0x10 may stand alone in "
						"the user data");
				Log(trimble_serial,"Killing the buffer");
				ip->serial_idx=0;
				return;
			}

			if(remember_0x10==0)
			{
				if((unsigned int)(optr-recvbuf)>=sizeof(recvbuf))
				{
					Log(trimble_serial, "About to overflow the receive buffer, packet too large!");
					return;
				}
				*optr++=ptr[i];
			}
		}

		if(end_seen)
		{
			// we got a complete packet in recvbuf.
			TSIPPKT pkt;
			memset(&pkt, 0, sizeof(pkt));

			pkt.code=ptr[1];
			pkt.len=optr-recvbuf;
			// sanity check
			if(pkt.len>sizeof(pkt.buf))
			{
				Log(trimble_serial, "??? about to overflow buf ???");
				return;
			}
			memcpy(pkt.buf, recvbuf, pkt.len);

			l=pkt.len+4;

			// look if we unstuffed any 0x10. if yes, increase the copy counter
			// so we can get the serial buffer right in the memcpy below. [1]
			for(i=0;i<pkt.len;i++)
				if(pkt.buf[i]==0x10)
					l++;

			Log(trimble_raw, "%s: got a complete packet: code = 0x%02x, len = %u",__FUNCTION__,
			     pkt.code, pkt.len);
			trimble_hexdump(trimble_raw, pkt.buf, pkt.len);

			// act on the packet
			trimble_do_protocol(ctx, ip, &pkt);
		}
		else
		{
			// still waiting for packet completion, no bytes consumed
			l=0;
			Log(trimble_raw, "%s: still waiting for packet completion", __FUNCTION__);
		}

		/*
		 * Pull up anything at the end and start over [1]
		 */
		memcpy(ip->serial_buf, ip->serial_buf + l, ip->serial_idx - l);
		ip->serial_idx -= l;
		Log(trimble_raw, "%s: (after cleanup) ip->serial_buf is now (l=%d):",__FUNCTION__, l);
		trimble_hexdump(trimble_raw, ip->serial_buf, ip->serial_idx);
	}
}

// send out a trimble message.
// Composition of the packet: User Guide page A-16
void trimble_sendmsg(struct instance *ip, TSIPPKT *pkt)
{
   unsigned char buf[1024], *pbuf, *ibuf;

   pbuf=buf;
   *pbuf++=0x10;
   *pbuf++=pkt->code;

   if(pkt->len>sizeof(buf)/2-4) // worst case: all 0x10 doubles the buffer
   {
      Log(trimble_serial, "%s: packet too large, ignoring",__FUNCTION__);
      return;
   }
   
   // escape 0x10 in the user buffer with 0x10 0x10 and copy the
   // buffer from pkt to buf
   ibuf=pkt->buf;
   while(pkt->len--)
   {
      if(*ibuf==0x10)
         *pbuf++=0x10;
      *pbuf++=*ibuf++;
   }

   // conclude the message
   *pbuf++=0x10;
   *pbuf++=0x03;

   Log(trimble_raw, "%s: sending out packet:",__FUNCTION__);
   trimble_hexdump(trimble_raw, buf, pbuf-buf);

   // send out the message
	if(write(ip->serial_fd, buf, pbuf-buf)<0)
   {
      Log(trimble_serial, "%s: write(2) error: %s",__FUNCTION__,strerror(errno));
      return;
   }
}

// convenience function; acts like trimble_sendmsg, but sends out
// a packet only consisting of one code and without length.
void trimble_sendcode(struct instance *ip, unsigned char code)
{
   TSIPPKT pkt;
   memset(&pkt, 0, sizeof(pkt));

   pkt.len=0;
   pkt.code=code;

   trimble_sendmsg(ip, &pkt);
}

// convenience function; acts like trimble_sendmsg, but sends out
// a packet only consisting of the code and one subcode
void trimble_sendcode_subcode(struct instance *ip, unsigned char code, unsigned char subcode)
{
   TSIPPKT pkt;
   memset(&pkt, 0, sizeof(pkt));

   pkt.len=1;
   pkt.code=code;
   pkt.buf[0]=subcode;

   trimble_sendmsg(ip, &pkt);
}
