#!/usr/local/bin/perl -w
#------------------------------------------------------------------------------
# Module Header
# Filename          : podd.pl
# Purpose           : TTM Plots on Demand server 
# Author            : Rene Wilhelm
# Date              : 18-NOV-2000
# Revised	    :  8-NOV-2001  Added trend and jitter plots
# Revised	    :  8-MAR-2002  Rewritten in Perl (including port listener)
# Revised	    : 14-MAY-2002  Minor mods for Solaris/Perl5.005
# Revised	    : 15-FEB-2007  Changed to use in multicast graph generation
# New Descrtiption  : tbd
# Old Description   : This is the backend program for TTM Plots on demand
#		      Started on a TTM server machine, this script listens
#		      to the ttplots port, forks for each connection and
#		      starts the TTM plots program with proper parameters
#		      results are pushed back over the socket (which is
#		      attached to the one setup by Plots-on-demand CGI)
# Language Version  : Perl5
# OSs Tested        : Solaris, Linux
# Command Line      : none
# Input Files       : parameters read from socket (passed from CGI program)
# Output Files      : resulting plots on socket   (send to CGI program)
# External Programs : delayplots, ipdv
# Problems          :
# To Do             : - improve cache cleanup 
# 		      - sanity check on input parameters
#			(no funny special shell chars etc.)
#		      
# $Id: graphod.pl,v 1.1 2007/03/30 08:54:49 franzsch Exp $
# $Source: /ncc/cvsroot/ttm/mcast-central/bin/graphod.pl,v $
#------------------------------------------------------------------------------

use POSIX;
use IO::Socket;
use Net::hostent;              # for OO version of gethostbyaddr

#------------------------------------------------------------------------------
# Global parameters / enviroment variables
#------------------------------------------------------------------------------

my $mcastroot = "/home/franzsch/mcast-central";

my $data_area = $mcastroot."/data/out";
my $bin_area = $mcastroot."/bin";
my $img_area = $mcastroot."/data/img";

my $progname = "graphod.pl";

my $debug = 0;
my $service = 'graphod';

my $port = getservbyname($service, 'tcp') || 4260; # 4250 is for TTM Unicast Plots on demand, 4260 for Multicast Graphs

my $server = IO::Socket::INET->new( Proto  => 'tcp',
   LocalPort => 4260,
   Listen    => 10,	# max connections
   Reuse     => 1);

die "can't setup server: $!" unless $server;
print STDERR "[Server $progname accepting clients]\n";


# become daemon

if (my $pid = fork) {
   exit(0);
}

POSIX:setsid() or die "setsid(): Can't start a new session";

close STDIN;
close STDOUT;
open(STDERR, ">/tmp/podd.errlog");   # per machine

$SIG{'CHLD'} = sub {wait};

while () {
   if (my $socket = $server->accept()) {
      if (my $pid = fork()) {
         # the parent continues listening
         close($socket);
         next;
      }
      elsif (defined $pid) {  
         # $pid is zero here if defined; this is the child process.
         STDOUT->fdopen($socket, "w") or die "Can't reopen STDOUT on socket: $!";
         STDERR->fdopen($socket, "w") or die "Can't reopen STDERR on socket: $!";
         STDIN->fdopen($socket, "r") or die "Can't reopen STDIN on socket: $!";


         handle_request($socket);
      }
      else {
         print STDERR "$progname: cannot fork!? \n";
      }
   }
   else {
      sleep(1); # prevent runaway process monopolizing CPU
   }
}

#------------------------------------------------------------------------------
# Subroutine Header
#
# Purpose:	handle a plots-on-demand request
# Input:	socket file descriptor
# Output:	on socket: gif or ps image if succesfull, else error message
#------------------------------------------------------------------------------

sub handle_request {
   my ($socket) = @_;

   my $command;
   my $filename;

   my @commands = ();
   my @tojoin = ();


   $socket->autoflush(1);
   #my $request = <$socket>;

   # only allow alphanumeric characters, - _ . and :  in the query
   # (careful with - , without \ it's interpreted as a range ;-)
   # Prevents hacking by abuse of UNIX shell metacharacters
   # also safeguard against abuse of -o option (better safe than sorry)

   while (<$socket>)
   {
      chomp;
      if (("$_" !~ /^[a-zA-Z0-9_:\.\s\n]+$/) || ("$_" =~ /-o\s/)) {
         # Someones trying to hack us
         exit (1);
      }
      push @commands, $_;
      if( $_ =~ "list" || $_ eq "end") {
         last;
      }
   };

   if( $commands[0] eq "list" ) { # We have a list command, print out a list of files we got
      opendir(OUTDIR, $data_area);
      foreach $filename (readdir (OUTDIR)) {
         if( $filename =~ /^MTRC\.(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})\.(.*)\.(\d*)$/ ) {
            print $socket "$filename\n"; 
         }
      }
      print $socket "end";
      closedir(OUTDIR);
   } elsif( $commands[0] eq "join" || $commands[0] eq "joinCompact" ) { # We got a join command...
      foreach $command (@commands) { # First build an array of files to join
         if ( $command eq "join" ) {
            next;
         }
         if ( $command =~ /^MTRC\.(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})\.(.*)\.(\d*)$/ && -e "$data_area/$command" ) {
            $group = $1;
            $source = $2;
            push @tojoin, $command;
         }
      }
      # Pull them together in a string
      $filenames = "";
      foreach $to (@tojoin) {
         if( -e "$data_area/$to" ) {
            $filenames .= "$data_area/$to ";
         }
      }
      # Join them and generate the imagea
      $timestamp = time;
      $rand = int(rand(100000)); # A random number ot make the image filename "unique"
      if( defined($filenames) && $filenames ne "" ) {
         `$bin_area/ccat $filenames > $data_area/MTRC.$group.$source`;
         `$bin_area/rmdupes.pl $data_area/MTRC.$group.$source > $data_area/MTRC.$group.$source.tmp`;
         `rm $data_area/MTRC.$group.$source`;
         `mv $data_area/MTRC.$group.$source.tmp $data_area/MTRC.$group.$source`;
         %as = &getLegend( "$data_area/MTRC.$group.$source" );
         if( $commands[0] eq "joinCompact" ) {
            # Remove all the information from the dot file except for the IP.
            &makeCompact("$data_area/MTRC.$group.$source");
         }
         `/usr/local/bin/dot -Tpng -o $img_area/MTRC.$group.$source.$timestamp.$rand.png $data_area/MTRC.$group.$source`;
         `rm $data_area/MTRC.$group.$source`;
         # Send the legend to the socket
         foreach $key (sort {$a <=> $b} keys %as) {
            print $socket "<span style=\"color: #$as{$key}; font-weight: bold;\">AS Nr. $key</span><br>\n";
         }
         # And finally send the image file name to the socket
         print $socket "next\n";
         print $socket "MTRC.$group.$source.$timestamp.$rand.png\n";
      } else {
         print $socket "file_invalid\n";
      }
   } elsif( $commands[0] eq "compare" || $commands[0] eq "compareCompact" ) { # We got to compare stuff
      # Timestamp and random value for image.
      $timestamp = time;
      $rand = int(rand(100000)); # A random number ot make the image filename "unique"

      $list = 1;
      @list1 = ();
      @list2 = ();
      foreach $command (@commands) {
         if( $command eq "compare" ) {
            next;
         } elsif( $command eq "end" ) {
            last;
         } elsif( $command eq "next" ) {
            $list = 2;
         } elsif( $list == 1 && $command =~ /^MTRC\.(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})\.(.*)\.(\d*)$/ ) {
            push @list1, $command;
            $group = $1;
            $source = $2;
         } elsif( $list == 2 && $command =~ /^MTRC\.(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})\.(.*)\.(\d*)$/ ) {
            push @list2, $command;
            $group = $1;
            $source = $2;
         }
      }
      foreach $to (@list1) {
         if( -e "$data_area/$to" ) {
            $filenames1 .= "$data_area/$to ";
         }
      }
      foreach $to (@list2) {
         if( -e "$data_area/$to" ) {
            $filenames2 .= "$data_area/$to ";
         }
      }
      # Join files and remove duplicate entries
      if( defined($filenames1) && $filenames1 ne "" ) {
         `$bin_area/ccat $filenames1 > $data_area/MTRC.$group.$source.tmp.1`;
      }
      if( -e "$data_area/MTRC.$group.$source.tmp.1" ) {
         `$bin_area/rmdupes.pl $data_area/MTRC.$group.$source.tmp.1 > $data_area/MTRC.$group.$source.1`;
         `rm $data_area/MTRC.$group.$source.tmp.1`;
      }
      if( defined($filenames2) && $filenames2 ne "" ) {
         `$bin_area/ccat $filenames2 > $data_area/MTRC.$group.$source.tmp.2`;
      }
      if( -e "$data_area/MTRC.$group.$source.tmp.2" ) {
         `$bin_area/rmdupes.pl $data_area/MTRC.$group.$source.tmp.2 > $data_area/MTRC.$group.$source.2`;
         `rm $data_area/MTRC.$group.$source.tmp.2`;
      }
      # Get a legend
      if( !(-e "$data_area/MTRC.$group.$source.1") && -e "$data_area/MTRC.$group.$source.2" ) {
         %as = &getLegend( "$data_area/MTRC.$group.$source.2" );
      } elsif ( -e "$data_area/MTRC.$group.$source.1" && !(-e "$data_area/MTRC.$group.$source.2") ) {
         %as = &getLegend( "$data_area/MTRC.$group.$source.1" );
      } elsif ( -e "$data_area/MTRC.$group.$source.1" && -e "$data_area/MTRC.$group.$source.2" ) {
         %as = &getLegend( "$data_area/MTRC.$group.$source.1", "$data_area/MTRC.$group.$source.2" );
      }
      # Get all the changes
      if( -e "$data_area/MTRC.$group.$source.1" && -e "$data_area/MTRC.$group.$source.2" ) {
         @changes = &compare( "$data_area/MTRC.$group.$source.1", "$data_area/MTRC.$group.$source.2" );
      }
      # Make compact view if necessary
      if( $commands[0] eq "compareCompact" ) {
         if( -e "$data_area/MTRC.$group.$source.1" ) {
            &makeCompact("$data_area/MTRC.$group.$source.1");
         }
         if( -e "$data_area/MTRC.$group.$source.2" ) {
            &makeCompact("$data_area/MTRC.$group.$source.2");
         }
      }
      # Generate the image file
      if( -e "$data_area/MTRC.$group.$source.1" ) {
         `/usr/local/bin/dot -Tpng -o $img_area/MTRC.$group.$source.$timestamp.$rand.1.png $data_area/MTRC.$group.$source.1`;
         `rm $data_area/MTRC.$group.$source.1`;
      }
      if( -e "$data_area/MTRC.$group.$source.2" ) {
         `/usr/local/bin/dot -Tpng -o $img_area/MTRC.$group.$source.$timestamp.$rand.2.png $data_area/MTRC.$group.$source.2`;
         `rm $data_area/MTRC.$group.$source.2`;
      }

      # Print legend and changes to the socket
      foreach $key (sort {$a <=> $b} keys %as) {
         print $socket "<span style=\"color: #$as{$key}; font-weight: bold;\">AS Nr. $key</span><br>\n";
      }
      print $socket "<br>\n";
      foreach $change (@changes) {
         print $socket "$change<br>\n";
      }

      # Print image filenames to the socket
      print $socket "next\n";
      print $socket "MTRC.$group.$source.$timestamp.$rand.1.png\n";
      print $socket "MTRC.$group.$source.$timestamp.$rand.2.png\n";
      print $socket "end\n";

   } elsif( $commands[0] eq "image" ) {
      if( $commands[1] =~ /^MTRC\.\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}\..+\.\d+\.\d+\.(\d\.)?png$/ && -e "$img_area/$commands[1]" ) {
         open( FILE, "$img_area/$commands[1]" );
      } else {
         open( FILE, "$img_area/noData.png" );
      }
      while(<FILE>) {
         print $socket $_;
      }
      close FILE;
   }
   # MUST exit this child processa
   close $socket;
   exit;
}

sub getLegend {
   # Get all the ASs for all the files passed to this function and calculate the colors.
   $rs = $/;
   undef $/; #read whole file

   # Those are the html color values corresponding to the colors picked by trace.pl on the testboxes
   # You always have to change this here AND on the testboxes to make any changes to this.
   push @colors, "f0f8ff"; # aliceblue
   push @colors, "ffe4c4"; # bisque
   push @colors, "f0ffff"; # azure
   push @colors, "7fffd4"; # aquamarine
   push @colors, "6495ed"; # cornflowerblue
   push @colors, "bfefff"; # lightblue1
   push @colors, "f08080"; # lightcoral
   push @colors, "ffb6c1"; # lightpink
   push @colors, "98fb98"; # palegreen
   push @colors, "ffa07a"; # lightsalmon
   
   %as = ();

   foreach $file (@_) {
      open (FILEIN, $file);
      $string = <FILEIN>;
      close FILEIN;

      while( $string =~ m/\sAS:\s(\d+)/g ) {
         $as{$1} = $colors[$1%(scalar(@colors))];
      }
   }
   return %as;
}
sub makeCompact {
   # Remove all the information from the dot file except for the IP.
   $rs = $/;
   undef $/; #read whole file

   open (FILEIN, $_[0]);
   $string = <FILEIN>;
   close FILEIN;
   $string =~ s/label="IP: (\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})[^"]+"/label="$1"/gs;
   # And write to our output file
   open (FILEOUT,">$_[0]");
   print FILEOUT $string;
   close FILEOUT;

   # And reset the resource delimiter
   $/ = $rs;
}

sub compare {
   $rs = $/;
   undef $/; #read whole file

   # FILE 1
   open( FILE, $_[0] );
   $file = <FILE>;
   close( FILE );
   # First find all vertices and put them in a hash
   while( $file =~ m/"(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"[\s\n\r]+\[([^\]]+)\];/g ) {
      $vert1{$1} = $2;
   }
   # Then match the edges
   while( $file =~ m/"(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"\s+\-\>\s+"(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"/g ) {
      push @edge1, $&;
   }

   # FILE 2
   open( FILE, $_[1] );
   $file = <FILE>;
   close( FILE );
   # First find all vertices and put them in a hash
   while( $file =~ m/"(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"[\s\n\r]+\[([^\]]+)\];/g ) {
      $vert2{$1} = $2;
   }
   # Then match the edges
   while( $file =~ m/"(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"\s+\-\>\s+"(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"/g ) {
      push @edge2, $&;
   }

   $/ = $rs; # Switch back to normal mode.

   @changes = ();
   # Compare the verts
   foreach $vert (keys %vert1) {
      if( !exists($vert2{$vert}) ) {
         # That vert does not exist in the second graph anymore, label it red in the first graph and move to list of changes
         $vert1{$vert} =~ s/(.*)/color=red, $1/;
         $vert1{$vert} =~ s/style="?([a-z]+)"?/style="$1,bold"/;
         push @changes, "<span style=\"color: red\">Node $vert vanished.</span>";
      }
   }
   foreach $vert (keys %vert2) {
      if( !exists($vert1{$vert}) ) {
         # That vert was not there in the first graph, label it green in the second graph and move to list of changes
         $vert2{$vert} =~ s/(.*)/color=green, $1/;
         $vert2{$vert} =~ s/style="?([a-z]+)"?/style="$1,bold"/;
         push @changes, "<span style=\"color: green\">New node $vert.</span>";
      }
   }
   # TODO: Compare the egdes
   foreach $edge (@edge1) {
      $match = 0;
      foreach $edtmp (@edge2) {
         if( $edge eq $edtmp ) {
            $match = 1;
         }
      }
      if( $match == 0 ) {
         # That edge does not exist in the second graph anymore. label it red in the first graph and move into changes list.
         if( $edge =~ /(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"\s+\-\>\s+"(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})/ ) {
            push @changes, "<span style=\"color: red\">Link $1->$2 vanished.</span>";
         }
         $edge =~ s/(.*)/$1 [color=red, style=bold]/
      }
   }
   foreach $edge (@edge2) {
      $match = 0;
      foreach $edtmp (@edge1) {
         if( $edge eq $edtmp ) {
            $match = 1;
         }
      }
      if( $match == 0 ) {
         # That edge is new in the second graph. label it green in the second graph and move into changes list.
         if( $edge =~ /(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"\s+\-\>\s+"(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})/ ) {
            push @changes, "<span style=\"color: green\">New link $1->$2.</span>";
         }
         $edge =~ s/(.*)/$1 [color=green, style=bold]/
      }
   }

   # Write files
   open( FILE, ">$_[0]" );
   print FILE "digraph graph0 {\n";
   foreach $vert (keys %vert1) {
      print FILE "\"$vert\" [$vert1{$vert}];\n";
   }
   foreach $edge (@edge1) {
      print FILE "$edge;\n";
   }
   print FILE "}";
   close(FILE);
   open( FILE, ">$_[1]" );
   print FILE "digraph graph0 {\n";
   foreach $vert (keys %vert2) {
      print FILE "\"$vert\" [$vert2{$vert}];\n";
   }
   foreach $edge (@edge2) {
      print FILE "$edge;\n";
   }
   print FILE "}";
   close(FILE);

   return @changes;
}
