/*
 * vi:ts=4:sw=4:
 *
 * $Id: if_utils.c,v 1.5 2009/10/15 09:48:26 ruben Exp $
 *
 * Copyright (c) 1998-2003 RIPE NCC
 * 
 * All Rights Reserved
 * 
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of the author not be used in advertising or 
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.
 * 
 * THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS; IN NO EVENT SHALL
 * AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY
 * DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE. 
 */

/* 


NAME
        GetEthernetIFInfo Get Ethernet interface information

SYNOPSIS
     #include "if_utils.h"


     int 
     GetEthernetIFInfo (struct etherifdata Interfacedata, char * error)

DESCRIPTION
     The GetEthernetIFInfo searches for an ethernet interface that is up.
     It returns the interface name, interface statistics and the first IP
     address it can find bound to the interface.
     
     The data is returned in a 
     struct etherifdata {
     	char   ifname[64];           The name of the interface 
	struct in_addr ethif_addr;   The IP number it thinks it has 
	struct if_data if_data;     See net/if.h for details...
	}

RETURN VALUES
    On success 0 is returned on failure a nonzero value is returned and some
    informative text is written in error.

    if error is empty and errno is equal to zero, the end of the interface list
    was reached. To recover and start over al again use SetEthernetIFInfo(), 
    otherwise use EndEthernetIFInfo() to free used resources.


SEE ALSO
    TCP/IP Illustrated Vol2 Chapter 3 for the structs that are returned
    kvm(3), and the source code of netstat (FreeBsd version) which was a
    source of inspiration for this.

	getifaddrs(3), getnameinfo(3)

HISTORY
    Locally developed for the test traffic project.

	Reworked to use getifaddrs(3) which is present in KAME/FreeBSD4 and
	better. A local getifaddrs.c is supplied for older BSD environmets.
	getifaddrs(3) uses sysctl(3) which does not need /dev/kmem privileges

*/

#include "if_utils.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <net/if_types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <net/if_var.h>
#include <netinet/in_var.h>



#if __FreeBSD__ >= 4
# include <ifaddrs.h>
# define INET6
#else
# include "ifaddrs.h"
#endif /* __FreeBSD__ >= 4 */



static char const rcsid[] = "$Id: if_utils.c,v 1.5 2009/10/15 09:48:26 ruben Exp $";

/* root of data list, act as an anchor for EndEthernetIFInfo() */ 
static struct ifaddrs *ifap=NULL;

/* current position of data list, act as an anchor for GetEthernetIFInfo*/
static struct ifaddrs *ifapwalker=NULL;

static	struct if_data	*ether_if_data=NULL;
static	char			*ether_ifname=NULL;
static	struct etherifdata *currentoutput=NULL;


/* XXX This is a hack to check for a prefix 
 * As an IPv6 address 128 bits long, and the lower 64 bits should be 
 * an EUI64 or by hand configured address, it should not be empty.
 * If it where empty it could be a prefix. Better is to use a mechanism that
 * returns a in6_ifaddr so we can check for ia6_flags & (IN6_IFF_ANYCAST|IN6_IFF_NOTREADY|IN6_IFF_DETACHED
 */

#define __IN6_IS_ADDR_LOOKS_LIKE_A_PREFIX(a)      \
	((*(const u_int32_t *)(const void *)(&(a)->s6_addr[8]) == 0) && \
	 (*(const u_int32_t *)(const void *)(&(a)->s6_addr[12]) == 0))

#define __IN6_IS_ADDR_LOOKS_LIKE_A_GLOBAL_UNICAST(a)      \
	 ((*(const u_int8_t *)(const void *)(&(a)->s6_addr[0]) & 0xE0) == 0x20)

#define __IN_CLASS_LP(i) (( ((u_int32_t)(i)) >> IN_CLASSA_NSHIFT)  == IN_LOOPBACKNET)

#define __IFA_IN(x)      (&((struct sockaddr_in *)((x)->ifa_addr))->sin_addr)

int GetEthernetIFInfo(struct etherifdata *output, char * errbuff)
{
	if (currentoutput==NULL)
		return NextEthernetIFInfo(output,errbuff,0);
	memcpy(output,currentoutput,sizeof(struct etherifdata));
	return (0);
}

int NextEthernetIFInfo(struct etherifdata *output, char * errbuff, int flags)
{
	int found = 0;

	/* zero output */

	memset(output,0,sizeof(struct etherifdata));

	/* XXX need to set ethif_addr and ethif_addr6 to "any address" if that changes from just being zero */

	/* get adresses from getifaddrs() */

	if (ifap == NULL)
	{
		if (getifaddrs(&ifap) == 0)
		{
			ifapwalker = ifap;
		}
		else
		{
			ifap = ifapwalker = NULL;
			strcpy(errbuff,strerror(errno));
			return -1;
		}
	}

	/* Allocate etherifdata for GetEthernetIFInfo */

	if (currentoutput == NULL)
	{
		currentoutput = (struct etherifdata*)malloc(sizeof(struct etherifdata));
		memset(currentoutput,0,sizeof(struct etherifdata));
	}

	/* scan for ethernet */

	while (ifapwalker!= NULL)
	{
		if (ifapwalker->ifa_flags & IFF_UP)
		{
			switch (ifapwalker->ifa_addr->sa_family)
			{
				case AF_LINK:
					 /* Mark ifapwalker->ifa_name as a usable device if it is a ethernet device
					  * and reference id_data from it.
					  */
					if ((((struct if_data*)ifapwalker->ifa_data)->ifi_type == IFT_ETHER ) ||
					    (((struct if_data*)ifapwalker->ifa_data)->ifi_type == IFT_L2VLAN ))
					{
						/* We got an active ethernet interface */
						ether_ifname = ifapwalker->ifa_name;
						ether_if_data = (struct if_data*)ifapwalker->ifa_data;
					}
					break;
				case AF_INET:
					/* Fetch data only if have an interface marked as ethernet
					 * 
					 * XXX This code makes the following assumption:
					 * getifaddrs orders its data per interface and per address family
					 * The assumption lies in the fact that we're expecting to following 
					 * address family order: AF_LINK, AF_INET (and AF_INET6)
					 * and the first AF_INET/AF_INET6 encountered is the primary address
					 * of the interface. followups are interface aliases. Skip them.
					 */
					
					if (ether_ifname && strcmp(ether_ifname, ifapwalker->ifa_name) == 0)
					{
						/* No multicast or bad classed addresses */

						/* XXX This typecasting around is ugly. macros added for readability... */
						if (!(IN_MULTICAST(ntohl(*(u_long*)__IFA_IN(ifapwalker))) ||
							IN_BADCLASS(ntohl(*(u_long*)__IFA_IN(ifapwalker))) ||
							__IN_CLASS_LP(ntohl(*(u_long*)__IFA_IN(ifapwalker)))))
						{

							/* Still empty ? fill it in! */
							if (output->ethif_addr.s_addr == INADDR_ANY  || flags & ETHIFINFO_ALIASES_TOO )
							{
								/* Copy address data and fill in output if it is not set alreay */
								strncpy(output->ifname, ether_ifname, IF_NAMESIZE);
								output->if_data = *ether_if_data;
								output->ethif_addr = *__IFA_IN(ifapwalker);
								inet_ntop(AF_INET, &(output->ethif_addr), output->ifaddr, _TTM_ADDRSTRLEN);
								found = 1;
							}
						}
					}
					break;
				case AF_INET6: /* IPV6 stuff here */
					if (ether_ifname && strcmp(ether_ifname, ifapwalker->ifa_name) == 0)
					{
						/* No loopback, link-local, site-local or multicast. some way to say "unicast only" */
						/* XXX Need to filter out anycast and such */
						if (!(IN6_IS_ADDR_LINKLOCAL(IFA_IN6(ifapwalker)) ||
							IN6_IS_ADDR_SITELOCAL(IFA_IN6(ifapwalker)) ||
							IN6_IS_ADDR_MULTICAST(IFA_IN6(ifapwalker)) ||
							IN6_IS_ADDR_UNSPECIFIED(IFA_IN6(ifapwalker)) ||
							IN6_IS_ADDR_LOOPBACK(IFA_IN6(ifapwalker)) ||
							__IN6_IS_ADDR_LOOKS_LIKE_A_PREFIX(IFA_IN6(ifapwalker))
							))
						{
							/* Still empty ? fill it in! */
							if (IN6_IS_ADDR_UNSPECIFIED(&output->ethif_addr6) || (flags & ETHIFINFO_ALIASES_TOO ))
							{
								/* Copy address data and fill in output if it is not set alreay */
								strncpy(output->ifname, ether_ifname, IF_NAMESIZE);
								output->if_data = *ether_if_data;
								output->ethif_addr6 = *(IFA_IN6(ifapwalker));
								inet_ntop(AF_INET6, &(output->ethif_addr6), output->ifaddr, _TTM_ADDRSTRLEN);
								found = 1;
							}
						}
					}
					break;
			}
		}
		ifapwalker=ifapwalker->ifa_next;

		/* break loop if output has been filled in and ifapwalker->ifa_name differs from it */
		if (ifapwalker && *output->ifname && ( ( (flags & ETHIFINFO_ALIASES_TOO) && found) || (strcmp(output->ifname, ifapwalker->ifa_name) != 0)))
		{
			memcpy(currentoutput,output,sizeof(struct etherifdata));
			return 0;
		}
	}

	/* End of list and no data found for output 
	 * Use End/SetEthernetIFInfo() to recover
	 */

	/* No error */
	*errbuff = '\0';
	errno=0;

	return -1;
}

void SetEthernetIFInfo(void)
{
	/* reset ifapwalker to head of list or create list if it was null and set
	 * ifapwalker to it 
	 */

	if (ifap == NULL)
	{
		if (getifaddrs(&ifap) == 0)
		{
			ifapwalker = ifap;
		}
		else
		{
			ifap = ifapwalker = NULL;
			/* XXX complain, but we have no way out */
			return;
		}
	}
	else
		ifapwalker = ifap;
	
	if (currentoutput)
		free(currentoutput);
	currentoutput = NULL;
}

void EndEthernetIFInfo(void)
{
	/* Free the list and reset ifap and ifapwalker to NULL */
	if (ifap)
		freeifaddrs(ifap);
	if (currentoutput)
		free(currentoutput);

	ifapwalker = ifap = NULL;
	currentoutput = NULL;
}

#ifdef MAIN /* begin test program */

int main(int argc, char **argv)
{
	struct etherifdata eid;
	char errbuf[256];
	while (NextEthernetIFInfo(&eid,errbuf,ETHIFINFO_ALIASES_TOO) == 0)
	{
		char asciiaddr[256];
		inet_ntop(AF_INET, &eid.ethif_addr, asciiaddr, sizeof(asciiaddr));
		printf("%s %s ", eid.ifname, eid.ethif_addr.s_addr == INADDR_ANY ? "unspecified" : asciiaddr);
		inet_ntop(AF_INET6, &eid.ethif_addr6, asciiaddr, sizeof(asciiaddr));
		printf("%s\n", IN6_IS_ADDR_UNSPECIFIED(&eid.ethif_addr6) ? "unspecified" : asciiaddr);
	}
	EndEthernetIFInfo();
	return 0;
}
#endif /* end test program */
