/*
 * vi:ts=4:sw=4:
 *
 * $Id: utils.c,v 1.14 2006/11/07 14:45:27 ruben Exp $
 *
 * Copyright (c) 1998-2003 RIPE NCC
 * 
 * All Rights Reserved
 * 
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of the author not be used in advertising or 
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.
 * 
 * THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS; IN NO EVENT SHALL
 * AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY
 * DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE. 
 */

/* 
 * -------------------------------------------------------------------------------
 * Module Header
 * 
 * Filename          : utils.c
 * Purpose           : Miscellaneous support functions for Send/Receive software
 *                     Some of them are brutally copied from Richard Stevens'
 *                     "Unix Network Programming" book
 * Author            : Olaf M. Kolkman 
 * Date              : 1998/03/02 
 * Language Version  : C
 * OSs Tested        : FreeBSD with customized kernel (getppstime interface call)
 * 
 * -------------------------------------------------------------------------------
 */

#include <signal.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <unistd.h>
#include <netdb.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <time.h>
#include <arpa/inet.h>
#include <sys/wait.h>
#include <math.h>
#include <limits.h>
#include <syslog.h>

#include "utils.h"


static char const   rcsid[] = "$Id: utils.c,v 1.14 2006/11/07 14:45:27 ruben Exp $";

extern pid_t        control_pid;
extern char         programname[1024];

/* 
 * -------------------------------------------------------------------------------
 * Subroutine Header
 * Purpose           :     fill memory block with random data
 * Side Effects      :
 * Comments          :
 * -------------------------------------------------------------------------------
 */

void brandom (void *b, size_t len)
{
	size_t              lengthleft;
	struct timeval      t;

	if (len == 0)
		return;

	lengthleft = len;

	/* Initialize Random nummer generator */
	gettimeofday (&t, NULL);
	srandom ((unsigned) (t.tv_usec));

	while (lengthleft)
	{
		/* XXX: change this later */
		memset (((char *) b) + len - lengthleft, random (), 1);
		lengthleft--;
	}
	return;
}

/** 
 * -------------------------------------------------------------------------------
 * Subroutine Header
 * Purpose           : convert time in struct timeval to ascii    
 * Side Effects      :
 * Comments          : Don't put multiple of these calls in a print statement!
 * \param[in] in timeval to be converted to ascii
 * \warning 
 * probably not thread safe
 * -------------------------------------------------------------------------------
 */

char               *timevaltoa (struct timeval in)
{
	static char tmp[20];
	int sign;

	sign = (in.tv_usec < 0 ? 1 : 0) | (in.tv_sec < 0 ? 1 : 0);

	snprintf(tmp,sizeof(tmp), "%s%d.%.6d", 
		sign ? "-" : "", 
		abs(in.tv_sec), abs(in.tv_usec));

	return (tmp);
}

/* 
 * -------------------------------------------------------------------------------
 * Subroutine Header
 * Purpose           : differential correction for time in struct timeval a
 * Side Effects      :
 * Comments          : the PPS time records the time at which the last
 *                   : (external) pulse-per-second was received by the system.
 *                   : this should be on a full second (tv_usec=0) but in 
 *                   : practice this never happens, the ntp synchronized clock 
 *                   : is either a little behind or ahead of the full second.
 *                   : This routine corrects for that: since we know the
 *                   : the offset from (closest) tv_usec=0 we add or subtract 
 *                   : it from the input time to arrive at a better time
 *                   : measurement. Note that the correction can go up to
 *                   : the tens of miliseconds range ...
 */

struct timeval pps_diff_cor (struct timeval a, struct timeval pps)
{

	struct timeval      x,
	                    out;

	x.tv_sec = 0;
	x.tv_usec = pps.tv_usec;
	if (x.tv_usec < 500000)
	{
		/* negative correction */
		timersub(&a,&x,&out);
	}
	else
	{
		/* positive correction (x.tv_usec>=500000) */
		x.tv_usec = 1000000 - x.tv_usec;
		timeradd (&a, &x, &out);
	}
	return (out);
}

/* 
 * -------------------------------------------------------------------------------
 * Subroutine Header
 * Purpose           :  Read "n" bytes from a descriptor. Use in place
 *                   :  of read(2) when fd is a stream socket.
 * Side Effects      :
 * Comments          :
 * -------------------------------------------------------------------------------
 */

int readn (fd, ptr, nbytes)
	register int fd;
	register char      *ptr;
	register int nbytes;
{
	int                 nleft,
	                    nread;

	nleft = nbytes;
	while (nleft > 0)
	{
		nread = read (fd, ptr, nleft);
		if (nread < 0)
		{
			return (nread);		/* error, return < 0 */
		}
		else if (nread == 0)
		{
			break;				/* EOF */
		}
		nleft -= nread;
		ptr += nread;
	}
	return (nbytes - nleft);	/* return >= 0 */
}

/* 
 * -------------------------------------------------------------------------------
 * Subroutine Header
 * Purpose           : Write "n" bytes to a descriptor.
 *                   : Use in place of write() when fd is a stream socket.
 * Side Effects      :
 * Comments          :
 * -------------------------------------------------------------------------------
 */

int writen (fd, ptr, nbytes)
	register int fd;
	register char      *ptr;
	register int nbytes;
{
	int                 nleft,
	                    nwritten;

	nleft = nbytes;
	while (nleft > 0)
	{
		nwritten = write (fd, ptr, nleft);
		if (nwritten <= 0)
		{
			return (nwritten);	/* error */
		}

		nleft -= nwritten;
		ptr += nwritten;
	}
	return (nbytes - nleft);

}

/* 
 * -------------------------------------------------------------------------------
 * Subroutine Header
 * Purpose           :  output error message to syslog
 * Side Effects      :
 * Comments          :  uses variable length argument list, like printf() et.al.
 *           first argument is the programname, second is the
 *           format string, all the rest are values to be output
 * -------------------------------------------------------------------------------
 */

void HandleError (const char *pname, ...)
{

	va_list             args;
	char               *fmt;
	char                message[_POSIX2_LINE_MAX];

	va_start (args, pname);

	if (pname != NULL)
	{
		fprintf (stderr, "%s: ", pname);
	}
	fmt = va_arg (args, char *);

	vsprintf (message, fmt, args);

#ifndef DEBUG
	syslog (LOG_NOTICE, "%s %s", message, strerror(errno));
#else
	fprintf (stderr, "%s %s\n", message, strerror(errno));
#endif

	va_end (args);

	fflush (stdout);
	fflush (stderr);

	return;
}


/* 
 * -------------------------------------------------------------------------------
 * Subroutine Header
 * Purpose           : Copy arg vector into new buffer, concatenating
 *                   : arguments with spaces.
 * Side Effects      :
 * Comments          : This was borrowed the tcpdump code 
 * 
 *                     Not used in any of Send/Receive programs? [RW]
 * -------------------------------------------------------------------------------
 */

char               *copy_argv (register char **argv)
{
	register char     **p;
	register u_int      len = 0;
	char               *buf;
	char               *src,
	                   *dst;

	p = argv;
	if (*p == 0)
		return 0;

	while (*p)
		len += strlen (*p++) + 1;

	buf = (char *) malloc (len);
	if (buf == NULL)
	{
		HandleError ("PortCatcher", "copy_argv: malloc (%d)", len);
		exit (1);
	}
	p = argv;
	dst = buf;
	while ((src = *p++) != NULL)
	{
		while ((*dst++ = *src++) != '\0')
			;
		dst[-1] = ' ';
	}
	dst[-1] = '\0';

	return buf;
}



/* 
 * -------------------------------------------------------------------------------
 * Subroutine Header
 * Purpose           :  signal handle to reap child processes (prevent Zombies)
 * Side Effects      :
 * Comments          :  loop until no more processes want to report status
 * 
 * 
 *     From Stevens:
 *     This is a 4.3BSD SIGCLD signal handler that can be used by a
 *     server that's not interested in its child's exit status, but needs to
 *     wait for them, to avoid clogging up the system with zombies.
 * 
 *     Beware that the calling process may get an interrupted system call
 *     when we return, so they had better handle that.
 * 
 * -------------------------------------------------------------------------------
 */

void sig_child (int signumber)
{

	/* 
	 * Use the wait3() system call with the WNOHANG option.
	 */

	int                 pid;
	int                 status;

	while ((pid = wait3 (&status, WNOHANG, (struct rusage *) 0)) > 0);
}


/**
 *  timersub() subtracts the time information stored in b from a and
 *  stores the resulting timeval in res.
 *  
 *  It allows for proper calculation of negative delays
 *  
 *  \param[in] tvp pointer to timeval structure to be subtracted from
 *  \param[in] uvp pointer to timeval structure 
 *  \param[out] vvp pointer to timeval structure which will carry the result
 *
 *  \warning
 *  Because negative values can be present in the resulting timeval
 *  structure one must be take caution when printing them *print*()
 *  functions. timevaltoa handles these correctly
 *
 *  \see timevaltoa
 */

void ncc_timersub(struct timeval *tvp, struct timeval *uvp, struct timeval *vvp)
{
    if (timercmp(tvp, uvp, < )) 
    {
		/* a negative delay. reverse parameters, we want an absolute value */
        timersub(uvp, tvp, vvp);

		/* Invert sign of fraction when tv_sec == 0
		 * XXX Beware printing of these timeval struct's !
		 *
		 */

        if (vvp->tv_sec)
            vvp->tv_sec = -(vvp->tv_sec);
        else
            vvp->tv_usec = -(vvp->tv_usec);
    }
    else
        timersub(tvp, uvp, vvp);
}
