#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <sysexits.h>
#include <sys/time.h>
#include <sys/errno.h>
#include <unistd.h>
#include <sys/param.h>
#include <netinet/in.h>

#include <sys/socket.h>
#include <arpa/inet.h>

#include "alarm.h"
#include "testbox.h"
#include "libdelayalarm.h"

struct box_histogram
{
    unsigned int        boxid;
    int                 entries[MAX_DELAY / DELAY_BIN_WIDTH];
};

/* Start and end time for each interval */
time_t        start_time[DAYS_HISTORY * INTERVAL_DAY];
time_t        end_time[DAYS_HISTORY * INTERVAL_DAY];

/* Entries per [interval] [delay] [box] */
struct box_histogram *box_interval[DAYS_HISTORY * INTERVAL_DAY], *box_interval_base;

#ifdef __KAME__
struct box_histogram *box_interval6[DAYS_HISTORY * INTERVAL_DAY], *box_interval6_base;
#endif /* __KAME__ */

int                *id2indexV4,
                    id2indexV4_size,
                   *id2indexV6,
                    id2indexV6_size;

#define MAXWORDS 5000

/* Parameter data hook for getRCDPData callback hook */

struct decode_received_packets_param
{
    time_t              start_time,
                        end_time;
    int                 debug,
                        family;
};

int decode_received_packets (char**, const void*);
char *lta_read_history (char*, time_t*, time_t*, int, int);
void lta_assign_intervals (time_t*, time_t*, int);
int lta_write_history (char *, size_t, time_t, int, int);

/* getRCDPData callback hook */

int decode_received_packets (char **arg, const void *read_received_packets_data)
{

/*  =====================
    Read RCDP file with packet data, put delay values in the appropriate bin.

  last_file    : Last file processed by the program
  filename     : File to be read
  debug        : If TRUE, then print debugging output

*/
    int                 srcid,
                        i,
                        bin,
                        clock_in,
                        clock_out;
    float               delay;
    time_t				arrival;

    /* XXX contains start_time, end_time, debug and family */
    const struct decode_received_packets_param *param = (const struct decode_received_packets_param *) read_received_packets_data;

    /* Check if packet is in the right interval here, so we can skip all futher 
     * checks... */
    arrival = (time_t) atof (arg[7]);

    if ((param->start_time <= arrival) && (param->end_time >= arrival))
    {
        /* Check if clock words are valid (X107) */
        clock_in = strtoul (arg[9], NULL, 16);
        clock_out = strtoul (arg[10], NULL, 16);
        
        if (ClocksOK (clock_in, clock_out))
        {
            /* Convert delays into ms and check for valid range */
            delay = atof (arg[8]) * 1000.0;

            if (delay >= 0.0)
            {
                /* Use arrival time, to find bin to store data */
                for (i = 0; i < DAYS_HISTORY * INTERVAL_DAY; i++)
                {
                    if ((start_time[i] < arrival) && (end_time[i] > arrival))
                    {
                        srcid = test_box_ip_id (arg[3], param->family);

                        if (srcid > 0)
                        {
                            int idx;
                            if (delay > (MAX_DELAY - DELAY_BIN_WIDTH))
                            {
                                delay = MAX_DELAY - DELAY_BIN_WIDTH;
                            }
                            bin = (int) (delay / (float) DELAY_BIN_WIDTH);
                            if (param->family == AF_INET)
                            {
                                if ((idx = lookup_id2index (id2indexV4, srcid, id2indexV4_size)) != -1)
                                    box_interval[i][idx].entries[bin]++;
                            }
#ifdef __KAME__
                            else
                            {
                                if ((idx = lookup_id2index (id2indexV6, srcid, id2indexV6_size)) != -1)
                                    box_interval6[i][idx].entries[bin]++;
                            }
#endif /* __KAME__ */
                        }
                        break;
                    }
                }

            }
        }
    }
    return 1;    
}

char               *lta_read_history (char *filename, time_t *history_start, time_t *history_end, int debug, int family)
{

/*    ================
      Read file with history, return pointer to the last file read.

Format of the file

  filename of the last file processed
  testbox_id from_time to_time {delays in bin i}

Arguments

filename = filename with history
debug    = if TRUE then print debugging info

*/

    FILE               *fp;
    int                 i,
                        j,
                        testbox,
                        testbox_idx,
                        start,
                        end;

    char                line[MAXWORDS];    /* 1 line from the file */
    char               *point = &line[0];
    char              **ap;
    char               *arg[MAXWORDS];    /* Elements on the file */
    char               *lastfile;


    *history_end = *history_start = 0;

    /* Open file */
    fp = fopen (filename, "r");
    if (fp == NULL)
    {
        if (debug == TRUE)
        {
            if (family == AF_INET)
                printf ("No history file found...(IPv4)\n");
#ifdef __KAME__
            else
                printf ("No history file found...(IPv6)\n");
#endif /* __KAME__ */
        }
        return NULL;
    }
    else
    {
        /* Read last filename processed */
        fscanf (fp, "%s", &line[0]);
        if (debug == TRUE)
        {
            if (family == AF_INET)
                printf ("Reading input (IPv4), Last file was %s\n", line);
#ifdef __KAME__
            else
                printf ("Reading input (IPv6), Last file was %s\n", line);
#endif /* __KAME__ */
        }
        lastfile = (char *) malloc (strlen (line) + 2);
        lastfile = strcpy (lastfile, line);

        /* XXX Change in last timestamp check */

        /* CR on the first line */
        fgets (line, MAXWORDS, fp);

        /* All other lines */
        while ((point = fgets (line, MAXWORDS, fp)) != NULL)
        {
            for (ap = arg; (*ap = strsep (&point, " \t")) != NULL;)
            {
                if (**ap != '\0')
                {
                    ++ap;
                }
            }
            testbox = atoi (arg[0]);
            start = atoi (arg[1]);
            end = atoi (arg[2]);
            testbox_idx =
                family == AF_INET ? lookup_id2index (id2indexV4, testbox, id2indexV4_size) : lookup_id2index (id2indexV6, testbox,
                                                                                                              id2indexV6_size);
            if (debug == TRUE)
            {
                printf ("Found line %d %d %d (family: %2d)\n", testbox, start, end, family);
            }
            if (testbox_idx != -1)
            {
                for (i = 0; i < DAYS_HISTORY * INTERVAL_DAY; i++)
                {
                    if ((start >= start_time[i]) && (end <= end_time[i]))
                    {
                        /* Determine boundaries of the history file */
                        if ( start < *history_start || *history_start == 0)
                            *history_start = start;

                        if (end > *history_end || *history_end == 0)
                            *history_end = end;

                        for (j = 0; j < MAX_DELAY / DELAY_BIN_WIDTH; j++)
                        {
                            if (arg[j + 3] == NULL)
                            {
                                printf ("Line truncated: file %s, line %i", filename, i);
                                exit (65);    /* EX_DATAERR data format error */
                            }
                            if (family == AF_INET)
                                box_interval[i][testbox_idx].entries[j] = atoi (arg[j + 3]);
    #ifdef __KAME__
                            else
                                box_interval6[i][testbox_idx].entries[j] = atoi (arg[j + 3]);
    #endif /* __KAME__ */
                        }
                        break;
                    }
                }
            }
            else if (debug)
                printf ("Unknown (stale) testbox id %d in %s found, ignoring\n", testbox, filename);

        }
        fclose (fp);
    }
    return lastfile;
}

#undef MAXWORDS

void lta_assign_intervals (time_t *long_start, time_t *long_end, int debug)
{

/*   ==================== 
     Set time intervals for the long term average.
*/

    time_t              t;
    struct tm          *tmval;
    int                 i,
                        interval_length;

    /* XXX Add iso8601period barriers */

    /* Today's date and time */
    t = time (&t);
    tmval = localtime (&t);
    if (debug == TRUE)
    {
        printf ("Today is %lu %d/%d/%d %d:%d:%d\n", (long unsigned int)t,
                tmval->tm_mday, tmval->tm_mon + 1, tmval->tm_year, tmval->tm_hour, tmval->tm_min, tmval->tm_sec);
    }

    /* Start of this day YY:MM:DD, 00:00:00 */
    tmval->tm_sec = 0;
    tmval->tm_min = 0;
    tmval->tm_hour = 0;
    t = mktime (tmval);
    tmval = localtime (&t);
    if (debug == TRUE)
    {
        printf ("End of history at %lu %d/%d/%d %d:%d:%d\n", (long unsigned int)t,
                tmval->tm_mday, tmval->tm_mon + 1, tmval->tm_year, tmval->tm_hour, tmval->tm_min, tmval->tm_sec);
    }

    interval_length = 86400 / INTERVAL_DAY;

    *long_end = t;

    /* Assign intervals backwards */
    for (i = DAYS_HISTORY * INTERVAL_DAY - 1; i > -1; i--)
    {
        end_time[i] = t;
        t -= interval_length;
        start_time[i] = t;
        *long_start = t;
    }

    /* Print intervals */
    if (debug == TRUE)
    {
        for (i = 0; i < DAYS_HISTORY * INTERVAL_DAY; i++)
        {
            printf ("%i %u %u\n", i, start_time[i], end_time[i]);
        }
    }

    return;
}

int lta_write_history (char *filename, size_t nboxes, time_t last, int family, int debug)
{

/*   =================
     Write history to filename.

filename = filename where the output will be stored
lastfile = last file processed by the program
debug    = if TRUE then print debugging output

*/

    FILE               *fp;
    int                 i,
                        interval,
                        sum;
    size_t				testbox_idx;

    /* Open file */
    if (debug == TRUE)
    {
        printf ("Writing history to %s\n", filename);
    }
    fp = fopen (filename, "w");
    if (fp == NULL)
    {
        printf ("Cannot write to file\n");
        return FALSE;
    }


    /* Put last processed filename on the file */
    fprintf (fp, "%lu\n", (unsigned long int)last);

    /* Add lines with interval and data */
    for (testbox_idx = 0; testbox_idx < nboxes; testbox_idx++)
    {
        for (interval = 0; interval < DAYS_HISTORY * INTERVAL_DAY; interval++)
        {
            sum = 0;
            i = 0;
            /* Only write lines with non-zero entries */
            while ((i < MAX_DELAY / DELAY_BIN_WIDTH) && (sum == 0))
            {
                if (family == AF_INET)
                    sum += box_interval[interval][testbox_idx].entries[i];
#ifdef __KAME__
                else
                    sum += box_interval6[interval][testbox_idx].entries[i];
#endif /* __KAME__ */
                i++;
            }
            if (sum > 0)
            {
                fprintf (fp, "%d %u %u ", family == AF_INET ? id2indexV4[testbox_idx] : id2indexV6[testbox_idx] , start_time[interval], end_time[interval]);
                for (i = 0; i < MAX_DELAY / DELAY_BIN_WIDTH; i++)
                {
                    if (family == AF_INET)
                        fprintf (fp, "%d ", box_interval[interval][testbox_idx].entries[i]);
#ifdef __KAME__
                    else
                        fprintf (fp, "%d ", box_interval6[interval][testbox_idx].entries[i]);
#endif /* __KAME__ */
                }
                fprintf (fp, "\n");
            }
        }
    }
    fclose (fp);

    if (debug == TRUE)
    {
        printf ("File written \n");
    }
    return TRUE;
}



int main (int argc, char *argv[])
{

/*  ====
    Long term average program 

     1. Read the test_box_box configuration
     2. Create bins where to store the information
     3. Read all delays so far
     4. Bin delays from the current data files
     5. Write new delay information
*/

    char hostname[MAXHOSTNAMELEN], ch, *histpath=NULL, *rcdppath=NULL, *confpath=NULL, *iso8601period=NULL;
    char               *last_file = NULL;

    char                configfile[256] = "/";
    char                alarmconfigfile[256] = "/";
    char                input_file[256] = "/";
    char                output_file[256] = "/";

#ifdef __KAME__
    char               *last_file6 = NULL;

    char                input_file6[256] = "/";
    char                output_file6[256] = "/";
#endif /* __KAME__ */
    int                 debug = FALSE, ip_mode, v6_cnt, v4_cnt;
                        

    time_t              long_start,long_end,log_start,log_end;
	size_t				i;

    /* command line arguments */
    while ((ch = getopt (argc, argv, "vhc:o:i:p:")) != -1)
    {
        switch (ch)
        {
        case 'v':
            /* -v Produce debugging output */
            debug = TRUE;
            break;
        case 'o':
            /* -i Override input file name */
            histpath = optarg;
            break;
        case 'c':
            /* -o Override output file name */
            confpath = optarg;
            break;
        case 'i':
            rcdppath = optarg;
            break;
        case 'p':
            iso8601period = optarg;
            puts("Not implemented yet");
            break;
        case 'h':
            printf("%s: -v -h -c <configpath> -o <outputpath> -i <inputpath> -p <iso8601period>\n",strrchr(argv[0],'/') ? (strrchr(argv[0],'/') + sizeof(char)): argv[0]);
            exit(EX_USAGE);
            break;
        default:
            break;
        }
    }

    /* Filenames */

    snprintf(configfile, sizeof(configfile), "%s/%s", (confpath ? confpath : PATH_CONFIG), LIST_OF_TESTBOXES);
    snprintf(alarmconfigfile, sizeof(alarmconfigfile), "%s/%s", (confpath ? confpath : PATH_CONFIG), FILE_ALARM_CNF);

    snprintf(input_file, sizeof(input_file), "%s/%s", (histpath ? histpath : PATH_HISTORY), FILE_IN);
    snprintf(output_file, sizeof(output_file), "%s/%s", (histpath ? histpath : PATH_HISTORY), FILE_OUT);
    
#ifdef __KAME__
    snprintf(input_file6, sizeof(input_file6), "%s/%s", (histpath ? histpath : PATH_HISTORY), FILE_IN6);
    snprintf(output_file6, sizeof(output_file6), "%s/%s", (histpath ? histpath : PATH_HISTORY), FILE_OUT6);
#endif /* __KAME__ */


    /* Who am I */

    gethostname(hostname,MAXHOSTNAMELEN);

    if (debug == TRUE)
    {
        printf (">%s<\n", input_file);
        printf (">%s<\n", output_file);
#ifdef __KAME__
        printf (">%s<\n", input_file6);
        printf (">%s<\n", output_file6);
#endif /* __KAME__ */
    }

    /* Read test-box info */
    i = read_test_box_config (configfile, debug);
    id2indexV4 = create_id2index (AF_INET, &id2indexV4_size);
    id2indexV6 = create_id2index (AF_INET6, &id2indexV6_size);

    if (i == FALSE)
    {
        return FALSE;
    }

    v6_cnt = v4_cnt = 0;
    ip_mode = 0;

    for (i = 0; i < number_of_boxes; i++)
    {
        if (test_box_list[i].family == AF_INET ) 
        {
            ip_mode |= 1;
            v4_cnt++;
        }
#ifdef __KAME__
        else if (test_box_list[i].family == AF_INET6 ) 
        {
            ip_mode |= 2;
            v6_cnt++;
        }
#endif /* __KAME__ */
        else 
        {
            /* NOP XXX */
            assert(0);
        }
    }

    if (debug)
    {
        printf("(v4_cnt=%d;id2indexV4_size=%d)\n",v4_cnt,id2indexV4_size);
        assert(( v4_cnt >= id2indexV4_size )) ;
#ifdef __KAME__
        printf("(v6_cnt=%d;id2indexV6_size=%d)\n",v6_cnt,id2indexV6_size);
        assert(( v6_cnt >= id2indexV6_size )) ;
#endif /* __KAME__ */
    }

    /* Allocate data structures which depend on no# of testboxes */

    box_interval_base = malloc((DAYS_HISTORY * INTERVAL_DAY) * id2indexV4_size * sizeof(struct box_histogram));
    if (box_interval_base == NULL)
        printf("box_interval_base = malloc(%d) failed: %s\n",(DAYS_HISTORY * INTERVAL_DAY) * id2indexV4_size * sizeof(struct box_histogram), strerror(errno));
    assert(box_interval_base != NULL);
    memset(box_interval_base,0,(DAYS_HISTORY * INTERVAL_DAY) * id2indexV4_size * sizeof(struct box_histogram));
#ifdef __KAME__
    box_interval6_base = malloc(((DAYS_HISTORY * INTERVAL_DAY)) * id2indexV6_size * sizeof(struct box_histogram));
    if (box_interval6_base == NULL)
        printf("box_interval6_base = malloc(%d) failed: %s\n",(DAYS_HISTORY * INTERVAL_DAY) * id2indexV6_size * sizeof(struct box_histogram), strerror(errno));
    assert(box_interval6_base != NULL);
    memset(box_interval6_base,0,(DAYS_HISTORY * INTERVAL_DAY) * id2indexV6_size * sizeof(struct box_histogram));
#endif /* __KAME__ */

    for (i = 0; i < (DAYS_HISTORY * INTERVAL_DAY); i++)
    {
        box_interval[i] = &box_interval_base[i*id2indexV4_size];
#ifdef __KAME__
        box_interval6[i] = &box_interval6_base[i*id2indexV6_size];
#endif /* __KAME__ */
    }

    /* Determine current day and time, assign intervals */
    /* fill in long_start/long_end */
    lta_assign_intervals ( &long_start, &long_end, debug);

    if ((ip_mode & 1) == 1)
    {
        struct decode_received_packets_param param = { long_start, long_end, debug, AF_INET };

        /* Read current data */
        last_file = lta_read_history (input_file, &log_start, &log_end, debug, AF_INET);
        
        /* Adjust scanning offset */
        param.start_time = long_start > log_end ? long_start : log_end;
        
        if (debug)
        {
            time_t start_t=param.start_time, end_t=long_end;
            char start_s[32], end_s[32];

            strftime(start_s,32,"%Y-%m-%d %T",gmtime(&start_t));    
            strftime(end_s,32,"%Y-%m-%d %T",gmtime(&end_t));    
            printf("Scanning from %s till %s (%lu seconds)\n", start_s,end_s, (long unsigned int)(long_end - param.start_time));

        }

        if ((long_end - param.start_time) > 0 )
        {
            /* And subtract last know time stamp from long_start to speed up processing */
            getRCDPData (hostname, param.start_time, (long_end - param.start_time), AF_INET, decode_received_packets, &param);

            /* Write results */
            i = lta_write_history (output_file, id2indexV4_size, long_end, AF_INET, debug);
            if (i == FALSE)
            {
                return FALSE;
            }
        }
    }

#ifdef __KAME__
    if ((ip_mode & 2) == 2)
    {
        struct decode_received_packets_param param = { long_start, long_end, debug, AF_INET6 };

        /* Read current data */
        last_file6 = lta_read_history (input_file6, &log_start, &log_end, debug, AF_INET6);
        
        /* Adjust scanning offset */
        param.start_time = long_start > log_end ? long_start : log_end;

        if (debug)
        {
            time_t start_t=param.start_time, end_t=long_end;
            char start_s[32], end_s[32];

            strftime(start_s,32,"%Y-%m-%d %T",gmtime(&start_t));    
            strftime(end_s,32,"%Y-%m-%d %T",gmtime(&end_t));    
            printf("Scanning from %s till %s (%lu seconds)\n", start_s,end_s, (long unsigned int)(long_end - param.start_time));

        }
        
        if ((long_end - param.start_time) > 0 )
        {
            /* And subtract last know time stamp from long_start to speed up processing */
            getRCDPData (hostname, param.start_time, (long_end - param.start_time), AF_INET6, decode_received_packets, &param);

            /* Write results */
            i = lta_write_history (output_file6, id2indexV6_size, long_end, AF_INET6, debug);
            if (i == FALSE)
            {
                return FALSE;
            }
        }
    }
#endif /* __KAME__ */

        free(box_interval_base);
#ifdef __KAME__
        free(box_interval6_base);
#endif /* __KAME__ */



    free_id2index (id2indexV4);
    free_id2index (id2indexV6);
    free_test_box_config ();


    /* rename files on success */
    rename(output_file,input_file);
#ifdef __KAME__
    rename(output_file6,input_file6);
#endif /* __KAME__ */

    return TRUE;
}

static char const   rcsid[] = "$Id: lta.c,v 1.15 2006/06/21 13:26:54 ruben Exp $";
