/*
 * vi:ts=4:sw=4:
 *
 * $Id: PacketCatcher.c,v 1.4 2007/04/26 14:32:29 franzsch Exp $
 *
 * Copyright (c) 1998-2003 RIPE NCC
 * 
 * All Rights Reserved
 * 
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of the author not be used in advertising or 
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.
 * 
 * THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS; IN NO EVENT SHALL
 * AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY
 * DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE. 
 */

/*
 * -------------------------------------------------------------------------------
 * 
 * Module Header
 * Filename          : PacketCatcher.c
 * Purpose           : Uses libpcap to catch packets from specific hosts.
 * Author            : Olaf M. Kolkman
 * Date              : 1998/02/06
 * Description       : Started by control_daemon this process captures
 *                     individual IP packets. The daemon does not return.
 *                     signal handling is set up by the control_daemon.
 * Language Version  : C
 * OSs Tested        : FreeBSD test traffic patch
 * Command Line      : None
 * Input Files       : None
 * Output Files      : None
 * External Programs : getppstime, special systemcall only available on tuned
 *                   : FreeBSD sysem. pcap library (see also "man 3 pcap")
 * Comments          : libpcap is located in /usr/local. This library will not
 *                     function well on linux. No need to port it.
 * 
 * Program flow	  : Parse the arguments, parse the pcap commands,
 *                  Process incomming packets, write the data to the
 *                  DataHandling daemon for further processing.
 * 
 * -------------------------------------------------------------------------------
 */

/*
 * -------------------------------------------------------------------------------
 * Module Change History
 * Author            :  Rene Wilhelm
 * Date              :  1999/11/25
 * Changes           :  add rcsid string for tagging binaries
 * Changes           :  open syslog connection with better identification string
 * Changes           :  no longer put network interface in promiscuous mode
 * Comments          :
 * -------------------------------------------------------------------------------
 */

/*
 * -------------------------------------------------------------------------------
 * Module Change History
 * Author            :  Rene Wilhelm
 * Date              :  2001/07/18
 * Changes           :  corrected check for duplicates in CreateTCPDUMPargument()
 * Comments          :  reformatted code for better readability
 * -------------------------------------------------------------------------------
 */

/*
 * -------------------------------------------------------------------------------
 * Module Change History
 * Author            :  Florian Frotzler
 * Date              :  2002/09/23
 * Changes           :  Rewriting of code to support IPv6
 * Comments          :  Now we have an IPv4 and an IPv6 pipe to Datahandler
 * -------------------------------------------------------------------------------
 */


#define PNAME "PacketCatcher"

/*
 *  XXX Hmm this is a nasty include but we need it for to get to the fd field
 *  of the pcap_t. Get this header from libcap source.
 */


#include <sys/types.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <signal.h>
#include <limits.h>
#include <unistd.h>
#include <syslog.h>

#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <arpa/inet.h>
#include <sys/socket.h>

#include <netinet/ip.h>
#include <netinet/udp.h>
#include <netinet/if_ether.h>

#ifndef pcap_int_h
#include <pcap-int.h>
#endif

#ifndef lib_pcap_h
#include  <pcap.h>
#endif

#include "if_utils.h"			/* Ethernet Interface utils */

#include "utils.h"				/* Error handling and buffer read write */
#include "packet.h"
#include "PacketCatcher.h"
#include "DataHandling.h"		/* This is where PCdataOutput is declared */
#include "control_daemon.h"

/* old SNAPLENGTH is deprecated
 * #define SNAPLENGTH (ETHER_HEADER_SIZE+IP_HEADER_SIZE+UDP_HEADER_SIZE+TEST_PACKET_HEADERLENGTH)
 */

#define SNAPLENGTH (ETHER_HEADER_SIZE+IP6_HEADER_SIZE+UDP_HEADER_SIZE+TEST_PACKET_HEADERLENGTH)

static char const rcsid[] = "$Id: PacketCatcher.c,v 1.4 2007/04/26 14:32:29 franzsch Exp $";



/*
 * -------------------------------------------------------------------------------
 * Subroutine Header
 * 
 * Purpose                  : Set up a daemon and process all incoming fragments.
 * rough code layout : determine the interface
 *                     set up packet capturing stuff (see pcap man files)
 *                     fork and process incoming packets.
 * Input parameter   : Argument buffer consisting of host/port combinations.
 *                   : family ( 1 = IPv4 only
 *                              2 = IPv6 only
 *                              3 = IPv4 + IPv6 (1 & 2)
 *                            )
 * Returns           : Return 0 on success or non-0 on failure.
 * 
 * -------------------------------------------------------------------------------
 */

int PacketCatcher(struct DelayProbeList *delayprobelist)
{

	int                 returnval, found;
	char                device[256];	/* device name (should fit) */

	pcap_t             *PcapPointer;
	bpf_u_int32         localnet,
	                    netmask;	/* typedef in pcap.h */
	struct bpf_program  filtercode;

	char                errbuff[PCAP_ERRBUF_SIZE];
	char               *FilterString;
	char			   *interface;

	closelog();
	openlog(programname, LOG_PID, LOG_LOCAL0);

	interface = read_if_config();
	found = 0;


	/*
	 * Iterate through all available interfaces, but only per device as our
	 * filterstring will only look at destination ports. This will catch both
	 * IPv6 and IPv4 traffic, even when aliased.
	 */

	while (NextEthernetIFInfo(&EtherData, errbuff, 0) == 0 && found == 0)
	{

		/*
		 * The Ethernet interface data is static 
		 */

		/* hmm listen per addr or per interface ??? */

		GetEthernetIFInfo(&EtherData, errbuff);

		/*
		 * Create a FilterString (a filter as would be specified by a human for 
		 * * tcpdump) from the host list 
		 */

		if (CreateTCPDUMPargument(delayprobelist, &FilterString))
		{
			HandleError(PNAME, "Fatal CreateTCPDUMParg: %s", errbuff);
			kill(control_pid, SIGTERM);
			exit(1);
		}

		/*
		 * Set the pcap device; if we do not know the device name we'll try
		 * PCAPDEVICE 
		 */

		if (EtherData.ifname != NULL)
		{
			strcpy(device, EtherData.ifname);
		} else
		{
			strcpy(device, PCAPDEVICE);
		}

        /* Override Interface when set by install script */

        if( interface != NULL ) {
    		strcpy(device, interface);
			found = 1;
		}

		if (pcap_lookupnet(device, &localnet, &netmask, errbuff) == -1)
		{
			HandleError(PNAME, "FATAL: pcap_lookupnet: %s", errbuff);
			kill(control_pid, SIGTERM);
			exit(1);
		}

		/*
		 * NOTE: we do not need to put the interface into promiscuous mode
		 * since all traffic we are interested in will be directed to our
		 ess.
 * For multicast measurements it is turned on though.
		 */

        PcapPointer = pcap_open_live(device, 255, PROMISC_MODE_ON, TIMEOUT, errbuff);
        
		if (PcapPointer == NULL)
		{
			HandleError(PNAME, "FATAL: pcap_open_life: %s", errbuff);
			kill(control_pid, SIGTERM);
			exit(1);
		}

#ifdef DEBUG
		printf("PCAP version %d.%d\n", PCAP_VERSION_MAJOR, PCAP_VERSION_MINOR);
		printf("PCAPPointer %ld PCap version %d.%d\n",
			   (long) PcapPointer, pcap_major_version(PcapPointer), pcap_minor_version(PcapPointer));
		printf("\nFilter:\n->%s<-\n", FilterString);
#endif

		/*
		 * Translate the Human specified filter to a pcap program and dispose
		 * the filter 
		 */

		if (pcap_compile(PcapPointer, &filtercode, FilterString, OPTIMIZE, netmask))
		{
			HandleError(PNAME, "FATAL: pcap_compile failed. Check the format (\"%s\")", FilterString);
			kill(control_pid, SIGTERM);
			exit(1);
		}
		free(FilterString);

		/*
		 * activate the just compiled filter 
		 */
		if (pcap_setfilter(PcapPointer, &filtercode))
		{
			HandleError(PNAME, "pcap_setfilter:%s", pcap_geterr(PcapPointer));
		}

		/*
		 * bpf (2) hack to set the system into immediat mode 
		 */
		{
			int                 immediate = 1;

			if (ioctl(PcapPointer->fd, BIOCIMMEDIATE, &immediate) < 0)
			{
				HandleError(PNAME, "FATAL: ioctl(BIOCIMMEDIATE)");
				kill(control_pid, SIGTERM);
				exit(1);
			}
		}						/* end scope immediate */

		/*
		 * Now everything is set it is safe to fork and let a daemon handle the 
		 * * rest 
		 */

		if (!fork())
		{
			/*
			 * The child continues 
			 */

			sprintf(programname, "Mcast PacketCatcher");
			setproctitle("-%s", programname);

			/*
			 * close the unneeded pipe...(global from DataHandling.h) 
			 */
			close(UC_pipe4[1]);	/* Sending sides */
			close(UC_pipe4[0]);
			close(UC_pipe6[1]);
			close(UC_pipe6[0]);

			close(PC_pipe4[0]);	/* receiving side */
			close(PC_pipe6[0]);

			signal(SIGHUP, handle_signals);
			signal(SIGINT, handle_signals);
			signal(SIGQUIT, handle_signals);
			signal(SIGILL, handle_signals);
			signal(SIGTRAP, handle_signals);

			signal(SIGABRT, handle_signals);
			signal(SIGEMT, handle_signals);
			signal(SIGFPE, handle_signals);
			signal(SIGBUS, handle_signals);
			signal(SIGSEGV, handle_signals);

			signal(SIGSYS, handle_signals);
			signal(SIGPIPE, handle_signals);
			signal(SIGALRM, handle_signals);
			signal(SIGTERM, handle_signals);

			signal(SIGXCPU, handle_signals);
			signal(SIGXFSZ, handle_signals);
			signal(SIGVTALRM, handle_signals);

			signal(SIGPROF, handle_signals);
			signal(SIGUSR1, handle_signals);
			signal(SIGUSR2, handle_signals);

			/*
			 * Start receiving and handling packets 
			 */
			/*
			 * ProcessPacket does the actual work. 
			 */

			returnval = pcap_loop(PcapPointer, LOOP_UNTIL_ERROR, (pcap_handler) & ProcessPacket, NULL);

			if (returnval < 0)
			{
				HandleError(PNAME, "FATAL: pcap_loop %s", pcap_geterr(PcapPointer));
				kill(control_pid, SIGTERM);
				exit(4);
			}

			if (returnval == 0)
			{
				HandleError(PNAME, "WARNING: pcap_loop reached EOF");
			}

			return (0);
		}
		/*
		 * XXX no error condition on failed fork(2) here! 
		 */

		/*
		 * The parent continues to attach the next PacketCatcher to the next
		 * interface 
		 */
	}
	EndEthernetIFInfo();

	/*
	 * the parent returns 
	 */
	return (0);
}



/*
 * -------------------------------------------------------------------------------
 * Subroutine Header ProcessPacket
 * 
 * Purpose  : Process a packet passed to us by the PCAP library
 * 
 * Comments : Called by the pcap library routine pcap_loop whenever an IP
 *            packet comes in, this routine gets the relevant data from the
 *            packet and writes it on the pipe to DataHandler for further
 *            procesing.
 * 
 * input    : as specified in pcap(3)
 * 
 * -------------------------------------------------------------------------------
 */

void ProcessPacket(u_char * user, const struct pcap_pkthdr *pcapdata, const u_char * payload)
{

	ushort              offset;
	const struct ip    *IPpacket;
	const struct ether_header *EHeader;
	const struct ip6_hdr *IP6Header;
	char                ipsix_addr[INET6_ADDRSTRLEN];	/* this is just a temp
														 * storage */

	struct pps_val      PPSTimeStamps;
	static struct packetcatchdata4 PCoutput4;
	static struct packetcatchdata6 PCoutput6;
	const struct test_packet_header *packheader;


	/*
	 * This is the timestamp set at the moment the packet arrived 
	 */
	PCoutput4.bpftime = pcapdata->ts;
	PCoutput6.bpftime = pcapdata->ts;

	/*
	 * These timestamps are set later and used for differential corrections. 
	 */
	getppstime(&PPSTimeStamps);


	/*
	 * First we look at the ethernet header to find out if the captured packet
	 * is IPv4 or IPv6
	 */

	EHeader = (const struct ether_header *) payload;

	/*
	 * 0x0800 = IPv4, 0x86DD = IPv6, see net/ethernet.h for deatils 
	 */


	if ((ntohs(EHeader->ether_type)) == ETHERTYPE_IP)
	{

		PCoutput4.flags &= 0X0000;
		PCoutput4.packet_id = 0l;

		/*
		 * Determine if a packet is fragmented. 
		 */
		/*
		 * See <netinet/ip.h> for details on the IP struct 
		 */

		/*
		 * Also: TCP/IP Illustrated Vol 1 by Stevens. Section 11.5
		 *  Non fragmented packets have their 'more fragments' field
		 *  not set and have an offset of 0.
		 *  A fragmented packet has its 'more fragments' field set
		 *  or has it not set but then it has an offset > 0
		 *
		 */

		IPpacket = (const struct ip *) (payload + ETHER_HEADER_SIZE);

		PCoutput4.ip_length = ntohs(IPpacket->ip_len) - IP_HEADER_SIZE;
		offset = ntohs(IPpacket->ip_off);
		strcpy(PCoutput4.source_addr, inet_ntoa(IPpacket->ip_src));
		strcpy(PCoutput4.dest_addr, inet_ntoa(IPpacket->ip_dst));


		/*
		 * These gps times are used for differential callibration. It is not
		 * very important where they are set 
		 */

		getppstime(&(PCoutput4.gps_times));

		PCoutput4.ip_id = ntohs(IPpacket->ip_id);
		PCoutput4.ip_offset = 8 * (offset & IP_OFFMASK);

		if (ntohs(offset & IP_MF))
		{
			PCoutput4.flags |= PC_MORE_FRAGMENTS;
		}

		if ((offset & IP_OFFMASK) == 0)
		{
			/*
			 * First fragment or full packet 
			 */
			if (PCoutput4.ip_length - UDP_HEADER_SIZE < TEST_PACKET_HEADERLENGTH)
			{
				PCoutput4.flags |= PC_FRAGMENT_TO_SMALL;
			} 
			else
			{
				/*
				 * not a to small fragment, get Test Packet header 
				 */

				/*
				 * NOTE: first cast IPpacket to char* otherwise
				 *  pointer arithmetic goes wrong
				 *
				 */

				packheader = (const struct test_packet_header *) ((const char *) IPpacket + IP_HEADER_SIZE + UDP_HEADER_SIZE);
				PCoutput4.packet_id = ntohl(packheader->packet_id);

#ifdef DEBUG
				printf("PC IPv4 Packet ID: %ld\n", PCoutput4.packet_id);
#endif
			}
		} 
		else
		{
			/*
			 * Not First fragment 
			 */
			/*
			 * void for now 
			 */
		}

		PCdataOutput(PCoutput4);
	} 
	else if ((ntohs(EHeader->ether_type)) == ETHERTYPE_IPV6)
	{

		IP6Header = (const struct ip6_hdr *) (payload + ETHER_HEADER_SIZE);

		/*
		 * The Next Header Field
		 * of the IPv6 header has to be 17 (UDP)
		 *
		 */


		if (IP6Header->ip6_nxt == 0x11)
		{

			inet_ntop(AF_INET6, &IP6Header->ip6_src, ipsix_addr, INET6_ADDRSTRLEN);
			strcpy(PCoutput6.source_addr, ipsix_addr);
			bzero(ipsix_addr, INET6_ADDRSTRLEN);
			inet_ntop(AF_INET6, &IP6Header->ip6_dst, ipsix_addr, INET6_ADDRSTRLEN);
			strcpy(PCoutput6.dest_addr, ipsix_addr);

			getppstime(&(PCoutput6.gps_times));

			PCoutput6.flow_label = ntohl((IP6Header->ip6_flow) & IPV6_FLOWLABEL_MASK);
			PCoutput6.p_length = ntohs(IP6Header->ip6_plen);

			packheader = (const struct test_packet_header *) ((const char *) IP6Header + IP6_HEADER_SIZE + UDP_HEADER_SIZE);

			PCoutput6.packet_id = ntohl(packheader->packet_id);

			PCdataOutput6(PCoutput6);

#ifdef DEBUG
			printf("PC IPv6 Packet ID: %ld\n", PCoutput6.packet_id);
#endif
		}
	}
    
#ifdef DEBUG
	fflush(stdout);
	fflush(stderr);
#endif
	return;
}


/*
 * -------------------------------------------------------------------------------
 * Subroutine Header
 * Purpose           : Create the filter string to be passed to pcap(3) packet
 *                     capture library. Format is identical to tcpdump's argument
 * Input parameters  :
 *                     delayprobelist   - list with IP/port combinations
 *                     listlength - number of entries in the array
 * Output parameters :
 *                     output   - pointer to resulting character string.
 *                     errbuff  - character string with error message(s)
 * Side Effects      :
 * Description       : From the argument list we create a list of (unique)
 * 		               IP addresses. The output should be of the form
 * 
 *                         "ip and dst host <a>  and ( src host <b> or src host <c> or ...)"
 * 
 *                     Where <a> is the IP address of the host running this code
 *                     (the receiver) and <b>, <c>, ... are the IPs of hosts
 *                     sending traffic. The pcap library will be initialized with
 *                     the resulting string, to ensure we only capture data
 *                     from boxes which are configured to send us traffic.
 * 
 *                     If succesfull, the routine returns 0; if an error occurs 1
 *                     is returned and errbuf will contain the error message.
 * Changes 	  : Filterstring now only consists of portnumbers	 
 * 
 * -------------------------------------------------------------------------------
 */

int CreateTCPDUMPargument(struct DelayProbeList *delayprobelist, char **output)
{

	int                 uniqueports[1024];
	int                 i,
	                    n,
	                    found;
	char               *endstring;

	struct delayprobetarget *current;

	char                tmpstring[_POSIX2_LINE_MAX];	/* temporary storage */


	memset(uniqueports, 0, 1024 * sizeof (int));
	n = 0;

	SLIST_FOREACH(current, delayprobelist, entries)
	{

		found = 0;
		for (i = 0; i < n; i++)
		{
			if (uniqueports[i] == current->portnumber)
				found = 1;
		}

		if (!found)
		{
			/*
			 * XXX Fatal when n > 1024 
			 */
			if (n > 1024)
			{
				HandleError(PNAME, "FATAL: Number of unique ports exceeds storage allocation (1024 ports)");
				return (1);
			}

			uniqueports[n] = current->portnumber;
			n++;
		}
	}

	endstring = (char *) calloc(1, 30 + n * 5);

	for (i = 0; i < n; i++)
	{
		if (uniqueports[i] == 0)
		{
			HandleError(PNAME, "FATAL: error constructing BPF filterstring");
			return (1);
		}
		sprintf(tmpstring, "dst port %d", uniqueports[i]);
		strcat(endstring, tmpstring);
		if (i != (n - 1))
			strcat(endstring, " and ");
	}

	*output = (char *) malloc(strlen(endstring) + 1);
	strcpy(*output, endstring);

	free(endstring);

	return (0);
}

/*
 * -------------------------------------------------------------------------------
 * Subroutine Header
 * Purpose	   : Interface function to DataHandler()
 * Input parameters  : PCdata   - packet data
 * Description	  : output results from PacketCatcher to the PC IPv4 or IPv6 pipe
 * 		    (which has been created in main controldaemon)
 * -------------------------------------------------------------------------------
 */

int PCdataOutput(struct packetcatchdata4 PCdata)
{
	if (write(PC_pipe4[1], &PCdata, PC_DATA_SIZE) == -1)
	{
		HandleError(PNAME, "PC write to pipe");
	}
	return (0);
}

int PCdataOutput6 (struct packetcatchdata6 PCdata)
{
	if (write (PC_pipe6[1], &PCdata, PC_DATA_SIZE6) == -1)
	{
		HandleError (PNAME, "PC write to pipe");
	}
	return (0);
}
