/*
 * vi:ts=4:sw=4:
 *
 * $Id: control_daemon.c,v 1.4 2007/04/26 11:52:54 franzsch Exp $
 *
 * Copyright (c) 1998-2003 RIPE NCC
 * 
 * All Rights Reserved
 * 
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of the author not be used in advertising or 
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.
 * 
 * THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS; IN NO EVENT SHALL
 * AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY
 * DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE. 
 */


/*
-------------------------------------------------------------------------------
Module Header

Filename    : control_daemon.c
Purpose     : Controlling the udp, packetcatcher and datahandling processes
Author      : Olaf M. Kolkman 
Date        : 1998/03/02
Revised by  : Rene Wilhelm
Date        : 1999/11/16
OSs Tested  : FreeBSD + customized kernel that includes gettimeofday system call

CommandLine : ControlDaemon <-f configfile> <-p pidfilename> 
	 		<-o outputfilenameroot> -t <ttroot>

	      configfile, pidfilename and outputfilenameroot are all 
	      relative to ttroot. Default for ttroot is obtained from
	      TTROOT environment variable

Input Files : configuration file. Default location: CONFIGFILENAME

Output Files: The program writes its PID to the specified file or
	      PIDFILENAME by default. Data are written to a file named
	      <outputfilenameroot>.<idno>  where <idno> holds date/time
	      of the file's creation in the format YYYYMMDD-HHMMSS 

Description : Test Traffic Measurements Receiver program.

	      All more or less independent data gatherers are started and
	      controlled from this program. The program forks itself to
	      become a proper background process, then goes into a loop
	      where it reads the Configfile, forks off three subprocesses
	      and waits for a signal. The processes started are:

	      1. PacketCatcher, a daemon which uses libpcap to catch
	         individual IP fragments and adds timestamps

	      2. UDPProcessor, a daemon which in turn forks off subprocesses
		 to listen and process data from all UDP ports present
		 in the ControlDaemon configfile

	      3. DataHandler, a daemon which combines the data from 
		 PacketCatcher and UDPProcessor and writes results to
		 the output file

	      Upon receiving a SIGHUP the next iteration of the main loop
              is started: subprocesses are killed (by means of the radical
	      SIGKILL), configfile is reread and new subprocesses started;
	      this is also happens at midnight, when a SIGALRM is scheduled
	      to ensure a new final outputfile is created (essential since
	      filenames contain date of processing as a quick reference).

	      Signals other than KILL or ALRM, will cause the ControlDaemon
	      to remove the pidfile and terminate.
	      
Comments    : CONFIGFILENAME and PIDFILENAME are defined in control_daemon.h

To Do	    : install signal handlers in the sub processes, giving them
	      a chance to clean up nicely (close output files, bpf devices,
	      etc.) before terminating. Also look into a redesign which 
	      would allow for reinitialisation/restart of running 
	      PacketCatcher, UDPProcessor and DataHandling daemons.
-------------------------------------------------------------------------------
 */

/*
-------------------------------------------------------------------------------
Module Change History
Author            : Rene Wilhelm
Date              : 1999/11/25
Changes           : add NTP error estimates to output lines  
Changes           : proper handling of child processes for all known signals 
Changes           : fix problems in setting command line arguments 
		    (don't overwrite argv[0], but assign new pointer)
Changes           : restart daemons at midnight GMT to guarantee new output file
-------------------------------------------------------------------------------
 */

/*
-------------------------------------------------------------------------------
Module Change History
Author            : Rene Wilhelm
Date              : 2001/07/19
Changes	          : reformatted code for better readability
Changes		  : improved program description in Module Header
-------------------------------------------------------------------------------
 */

/*
-------------------------------------------------------------------------------
Module Change History
Author            : Florian Frotzler
Date              : 2002/09/23
Changes           : Rewriting to support IPv6
Changes           : Control Daemon now inits 4 pipes (2x IPv4, 2x IPv6)
-------------------------------------------------------------------------------
 */



/* The control daemon sets up the different subprocesses.
 * the global variables through which these processes communicate
 * are defined here. 
 */

#define PNAME "mcast_control_deamon"


#define FLAG_P 0x01
#define REQUIRED  0   /* OR of all required arguments; for now none required  */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/types.h>
#include <signal.h>
#include <sys/types.h>
#include <limits.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <syslog.h>
#include <time.h>

#include "UDPProcessor.h"
#include "PacketCatcher.h"
#include "control_daemon.h"
#include "utils.h"
#include "DataHandling.h"

/* Some global variables. They are here because they are used in for
   instance the signal functions */

gid_t 		grouppid;		    

extern pid_t 	control_pid;		    
u_short		dowhileloop;
char		*ttroot;
char 		*pidfilename;
char 		*configfilename;
char 		*finalfilename;
char	    *finalfilename6;
int		family;

void 	handle_sub_procs(int);
int		ReadConfigFile (const char *, struct DelayProbeList *, int *);
void	FreeConfigFile (struct DelayProbeList*);
char*   read_if_config(void);

static char const rcsid[] = "$Id: control_daemon.c,v 1.4 2007/04/26 11:52:54 franzsch Exp $";

/*
-------------------------------------------------------------------------------
Subroutine Header 

Purpose      : Main ControlDaemon program.
Side effects : several globals are initialized
Description  : After processing command line arguments and other initialisation
	       work, the main program forks, closes all file descriptors
	       and detaches itself from the controlling terminal.
Returns      : Return 0 on success or non-0 on failure.
-------------------------------------------------------------------------------
 */

int main (int argc, char **argv)
{

	FILE               *fd;
	int                 c;
	size_t              length;
	pid_t               forkreturn;
	int                 required_arg;

	time_t              time_now;	/* variables for time calculations */
	time_t              seconds;
	time_t              midnight;
	struct tm          *tomorrow;

	grouppid = 0;

	/* machine should already be in GMT, but forcing it doesn't hurt */

	putenv ("TZ=GMT+0");

	/* 
	 *  make sure the storage pointed to by argv[0] is large
	 *  enough for our rewriting attempts; UNIX ps looks here
	 *  for the process name!
	 */

	strncpy (programname, argv[0], sizeof (programname));
	argv[0] = programname;

    interface = read_if_config();
    
	/* Set ttroot from environment or compiled in default */

	if (!(ttroot = getenv ("TTROOT")))
	{
		ttroot = strdup (TT_ROOT);
	}

#ifdef DEBUG
	printf ("TTROOT=%s\n", ttroot);
#endif

	/* set the default filenames */

	length = strlen (PIDFILENAME) + 1;	/* trailing \0 */
	pidfilename = (char *) malloc (length);
	if (pidfilename == NULL)
	{
		HandleError (programname, "Main: could not allocate memory for pidfilename");
		exit (1);
	}
	strcpy (pidfilename, PIDFILENAME);

	length = strlen (FINAL_OUTPUT_ROOTNAME) + 1;
	finalfilename = (char *) malloc (length);
	length = strlen (FINAL_OUTPUT_ROOTNAME6) + 1;
	finalfilename6 = (char *) malloc (length);

	if ((finalfilename == NULL) || (finalfilename6 == NULL))
	{
		HandleError (programname, "Could not allocate memory for final output name");
		exit (1);
	}
	strcpy (finalfilename, FINAL_OUTPUT_ROOTNAME);
	strcpy (finalfilename6, FINAL_OUTPUT_ROOTNAME6);

	length = strlen (CONFIGFILENAME) + 1;
	configfilename = (char *) malloc (length);
	if (configfilename == NULL)
	{
		HandleError (programname, "Could not allocate memory for configfilename");
		exit (1);
	}
	strcpy (configfilename, CONFIGFILENAME);

	/* parse the command line. */
	required_arg = 0;
	while ((c = getopt (argc, argv, "p:c:f:g:")) != -1)
	{

		/* 
		 * -p  <pidfilename>
		 * -c <configfilename>
		 * -f <finalfilename>
		 * -g <finalfilename for IPv6>
		 */

		switch (c)
		{
		case 'p':				/* pidfile */
			required_arg |= FLAG_P;
			length = strlen (optarg) + 1;	/* trailing \0 */
			pidfilename = (char *) realloc (pidfilename, length);
			if (pidfilename == NULL)
			{
				HandleError (programname, "Couldn't reallocate memory");
				exit (1);
			}
			strcpy (pidfilename, optarg);
			break;

		case 'c':				/* configfile */
			required_arg |= FLAG_P;
			length = strlen (optarg) + 1;
			configfilename = (char *) realloc (configfilename, length);
			if (configfilename == NULL)
			{
				HandleError (programname, "Couldn't reallocate memory");
				exit (1);
			}
			strcpy (configfilename, optarg);
			break;

		case 'f':				/* finalfile */
			required_arg |= FLAG_P;
			length = strlen (optarg) + 1;
			finalfilename = (char *) realloc (finalfilename, length);
			if (finalfilename == NULL)
			{
				HandleError (programname, "Couldn't reallocate memory");
				exit (1);
			}
			strcpy (finalfilename, optarg);
			break;

		case 'g':				/* finalfile */
			required_arg |= FLAG_P;
			length = strlen (optarg) + 1;
			finalfilename6 = (char *) realloc (finalfilename6, length);
			if (finalfilename == NULL)
			{
				HandleError (programname, "Couldn't reallocate memory");
				exit (1);
			}
			strcpy (finalfilename6, optarg);
			break;

		default:
			abort ();
		}
	}

	/* check if the required arguments have been specified */
	if (REQUIRED != (REQUIRED & required_arg))
	{
		HandleError (programname, "Not all required arguments specified");
	}

	if (chdir (ttroot) != 0)
	{
		HandleError (programname, "Could not chdir %s", ttroot);
		exit (1);
	}

#ifdef DEBUG
	printf ("\nconfig: %s\npid: %s\nfinale: %s\n", configfilename, pidfilename, finalfilename);
#endif

	/* isn't this a no-op in TTM production environment? [RW] */

	/* process should be suid for group kmem */
	setgid (getgid ());

	/* 
	 * We first go into the background and then we let the daemon become
	 * the 'master' process.
	 */

#ifndef DEBUG2
	if (fork () != 0)
	{
		/* the parrent process ..  */
		exit (0);
	}
#endif

#if !defined(DEBUG) && !defined(DEBUG3)
	/* Do not close stdin, stdout, and stderr when debugging */
	fclose (stdin);
	fclose (stdout);
	fclose (stderr);

	for (i = 3; i < sysconf (_SC_OPEN_MAX); i++)
	{
		close (i);
	}

#endif


	if ((fd = fopen (pidfilename, "r")))
	{
		HandleError (programname, "ABORT: %s already exists, check if process is running, if not rm %s.", pidfilename, pidfilename);
		exit (1);
	}


	/* PID administration */
	control_pid = getpid ();

	if ((fd = fopen (pidfilename, "w")))
	{
#ifdef DEBUG
		printf ("main control PID: %ld\n", (long) control_pid);
#endif
		fprintf (fd, "%ld\n", (long) control_pid);
		fclose (fd);
	}
	else
	{
		HandleError (programname, "Warning Could not open PID file");
		fclose (fd);
	}

	/* change appearance in ps(1) output */
	setproctitle ("-Mcast ControlDaemon (%ld)", (long) control_pid);

	openlog ("MainControlDaemon", LOG_PID, LOG_LOCAL0);

	/* Establish signal handlers (needs doing only once) */

	signal (SIGCHLD, sig_child);

	/* 
	 * Catch all signals which normally would cause process to terminate:
	 * ControlDaemon *must* take down its children before exiting! 
	 * Note that HUP and ALRM are meant as 'restart' signals: these will
	 * make 'main control daemon' kill the currently running children
	 * and fork new ones which reread (potentially changed) configuration
	 * files. see handle_sub_procs function for authoratative information
	 * on signal handling
	 * 
	 */

	signal (SIGHUP, handle_sub_procs);
	signal (SIGINT, handle_sub_procs);
	signal (SIGQUIT, handle_sub_procs);
	signal (SIGILL, handle_sub_procs);
	signal (SIGTRAP, handle_sub_procs);

	signal (SIGABRT, handle_sub_procs);
	signal (SIGEMT, handle_sub_procs);
	signal (SIGFPE, handle_sub_procs);
	signal (SIGBUS, handle_sub_procs);
	signal (SIGSEGV, handle_sub_procs);

	signal (SIGSYS, handle_sub_procs);
	signal (SIGPIPE, handle_sub_procs);
	signal (SIGALRM, handle_sub_procs);
	signal (SIGTERM, handle_sub_procs);

	signal (SIGXCPU, handle_sub_procs);
	signal (SIGXFSZ, handle_sub_procs);
	signal (SIGVTALRM, handle_sub_procs);

	signal (SIGPROF, handle_sub_procs);
	signal (SIGUSR1, handle_sub_procs);
	signal (SIGUSR2, handle_sub_procs);

	/* 
	 *  We want to keep the control daemon runnig forever. Everytime the
	 *  loop below is restarted we make sure all the children are dead, the
	 *  configuration file is reread and we restart the daemons...
	 *  
	 */

	errno = 0;
	HandleError (programname, "Starting main control loop");

	while (1)
	{							/* Permanent ControlDaemon loop */

		if (grouppid)
		{
			/* make sure all the children are dead */
			killpg (grouppid, SIGKILL);

			/* 
			 *  sleep 1 second to give subprocs the time
			 *  to clean up nicely (closing sockets and
			 *  bpf filters) before starting new ones
			 */

			sleep (1);
		}

		/* 
		 *  Since we want to keep the gpid and the pid of the control
		 *  daemon different we start a subprocess which will start
		 *  the measurement daemons and Data collector. Once those
		 *  are started, this subprocess may terminate.
		 */
#ifndef DEBUG2
		forkreturn = fork ();
#else
		forkreturn = 0;
#endif

		if (forkreturn != 0)
		{

			/* 
			 * This is the parent of all parents...
			 * It should wait untill it's allowed to kill
			 * its children (Reminds me of Medea.)
			 */

			grouppid = forkreturn;

			dowhileloop = 1;	/* sig handling function reset this to 0 */

			/* 
			 * Schedule alarm for midnight  ...
			 * This is to reread the configuration files and to make
			 * sure the new day starts with a new outputfile 
			 */

			time_now = time (NULL);
			midnight = (time_now + (24 * 3600));
			tomorrow = gmtime (&midnight);

			tomorrow->tm_sec = 4;	/* grace period for late packets */
			tomorrow->tm_min = 0;
			tomorrow->tm_hour = 0;
			midnight = mktime (tomorrow);
			seconds = midnight - time_now;

			alarm (seconds);	/* signal at midnight */

			while (dowhileloop)
			{

				/* 
				 *  wait for a signal to arrive.
				 *  handle_sub_procs will either terminate
				 *  all running processes or it will return
				 *  with dowhileloop=0
				 */

				sigset_t      sigmask;

				sigemptyset (&sigmask);
				sigsuspend (&sigmask);

			}

			/* close these to be sure they are closed */
			close (PC_pipe4[0]);
			close (UC_pipe4[0]);
			close (PC_pipe6[0]);
			close (UC_pipe6[0]);
			close (PC_pipe4[1]);
			close (UC_pipe4[1]);
			close (PC_pipe6[1]);
			close (UC_pipe6[1]);
		}
		else
		{
			/* This child will start all the 'sub daemons' and then gracefully
			 * return.  */

			struct DelayProbeList delayprobelist =
				SLIST_HEAD_INITIALIZER(delayprobelist);

			SLIST_INIT(&delayprobelist);

			grouppid = getpid ();
			setpgid (0, grouppid);

			closelog ();
			openlog ("ChildControlDaemon", LOG_PID, LOG_LOCAL0);
			setproctitle ("-McastChild ControlDaemon (%ld)", (long) getpid());

			/* 
			 * family will be filled with one of the following values :
			 *
			 * '1' ... There are only IPv4 hosts in delayprobelist
			 * '2' ... There are only IPv6 hosts in delayprobelist
			 * '3' ... Mixed up IP families in delayprobelist (1 & 2)
			 */
            printf( "Reading config file...\n" );
			ReadConfigFile (configfilename, &delayprobelist, &family);

			/* 
			 * PacketCatcher and UDPProcessor have to communicate
			 * with the DataHandling daemon. We will use pipes for
			 * this (Also see datahandling.c and datahandling.h)
			 */

			if (pipe (PC_pipe4) || pipe (UC_pipe4) || pipe (PC_pipe6) || pipe (UC_pipe6))
			{
				/* can't open pipes; signal MainControlDaemon to clean up and
				 * restart this child process */

				HandleError (programname, "ERROR: unable to open pipes for ICP");
				kill (control_pid, SIGTERM);

				/* should not get here */
				exit (1);
			}

#ifdef DEBUG
			printf ("PC started... \n");
#endif

			PacketCatcher (&delayprobelist);
#ifdef DEBUG
			printf ("PacketCatcher started\n");
#endif

			UDPProcessor (&delayprobelist);
#ifdef DEBUG
			printf ("UDPProcessor started\n");
#endif

			close (UC_pipe4[1]);	/* Sending sides */
			close (PC_pipe4[1]);
			close (UC_pipe6[1]);
			close (PC_pipe6[1]);

			DataHandler (&delayprobelist, finalfilename, finalfilename6);
#ifdef DEBUG
			printf ("DataHandler started\n");
#endif

			close (PC_pipe4[0]);	/* receiving sides */
			close (UC_pipe4[0]);
			close (PC_pipe6[0]);
			close (UC_pipe6[0]);

			return (0);			/* subprocesses started, child done */
		}

	}
	/* we won't get here */
}

/*
 * -------------------------------------------------------------------------------
 * Subroitung read_if_config
 * Purpose           : Reads interface name from config file
 * Side Effects      : Allocates memory for interfacer.
 * Comments          :
 * -------------------------------------------------------------------------------
 */

char* read_if_config() {
   char *file, *point, *ifname;
   char buffer[MAXLINE];
   FILE* fp;

   file = (char *) malloc (strlen (getenv ("TTROOT")) + strlen ("/config/mcast/interface.conf") + 1);
   file = strcpy (file, getenv ("TTROOT"));
   file = strcat (file, "/config/mcast/interface.conf");

   /* Read configfile */
   fp = fopen (file, "r");
   if (fp == NULL)
   {
      printf ("Cannot open %s\n", file);
      return NULL;
   }

   ifname = NULL;
   while ((point = fgets (buffer, MAXLINE, fp)) != NULL)
   {
      if (strncmp (point, "InterfaceName", 13) == 0) {
         strtok (point, " "); // Skip the first word
         ifname = strdup (strtok (NULL, "\n")); // Save the Interface Name
      }
   }
   // Return pointer to the interface name or null if we didn't find it or it's defined as 0.
   if( *ifname == '0' )
      return NULL;
   else
      return ifname;
}


/*
 * -------------------------------------------------------------------------------
 * Subroutine Header
 * Purpose           : Deal with all the sub processes 
 * Side Effects      : kills child processes, sets dowhileloop to 0,
 *                     some signals will cause program termination
 * Comments          :
 * -------------------------------------------------------------------------------
 */

void handle_sub_procs (int signo)
{

	char                message[256];

	(void) fflush (stdout);
	(void) fflush (stderr);

	/* 
	 *  If we want to sent other signals than the radical KILL we have
	 *  to define signal hanlers for the subproceses.  Untill then leave
	 *  this signal as it is otherwise the subprocess will not die...
	 */

	killpg (grouppid, SIGKILL);

	if ((signo == SIGHUP) || (signo == SIGALRM))
	{
		dowhileloop = 0;
		sprintf (message, "Signal %d caught, Restarting daemons.", signo);
		errno = 0;
		HandleError (programname, message);
		return;
	}
	else
	{
		sprintf (message, "Received signal %d, cleaning up completely", signo);
		errno = 0;
		HandleError (programname, message);
		remove (pidfilename);
		exit (0);
	}

}

/*
 * -------------------------------------------------------------------------------
 * Subroutine Header
 * Purpose           : Read ControlDaemon configuration file
 * Side Effects      :
 * Comments          : Now supports version 1.1.0 of configfile (IPv6 support)
 * -------------------------------------------------------------------------------
 */

int ReadConfigFile(const char *filename, struct DelayProbeList *delayprobelist, int *ip)
{

	struct delayprobetarget *new;
	int                 ip4_flag,
	                    ip6_flag;
	char                currentline[_POSIX2_LINE_MAX];
	FILE               *fp;


	/*
	 * See egrep(1) for syntax 
	 */

	fp = fopen(filename, "r");
	if (fp == NULL)
	{
		HandleError(programname, "Error opening configfile %s", filename);

		/*
		 * since this function is called when there is only
		 * one process around, the exit below terminates
		 * all of running ControlDaemon....
		 */

		remove(pidfilename);
		exit(1);
	}


	ip4_flag = 0;
	ip6_flag = 0;


	/*
	 * read the file line by line... 
	 */
	while (fgets(currentline, _POSIX2_LINE_MAX, fp))
	{
		char                hostname[1024];
		int                 port;
		int                 validhost;
		int                 _family=0;
		struct in6_addr     addr6;
		struct hostent     *dns_addr;

		validhost = 0;
		strcpy(hostname, "# initializer");
		sscanf(currentline, "%s %d", hostname, &port);

		bzero(currentline, _POSIX2_LINE_MAX);

		if (!strchr(hostname, '#'))
		{

			/*
			 * not a comment line 
			 */

			/*
			 * Check if host can actually be resolved, this is only a first
			 * order sanity check 
			 */

			if ((inet_addr(hostname) != INADDR_NONE) || (gethostbyname(hostname)) != NULL)
			{
				ip4_flag = V4_FLAG;
				validhost = 1;
				_family = AF_INET;
			} else if ((inet_pton(AF_INET6, hostname, &addr6)) || ((dns_addr = gethostbyname2(hostname, AF_INET6)) != NULL))
			{
				ip6_flag = V6_FLAG;
				validhost = 1;
				_family = AF_INET6;
			} else
				validhost = 0;

			if (validhost)
			{

				new = (struct delayprobetarget *) malloc(sizeof (struct delayprobetarget));

				if (new == NULL)
				{
					HandleError(programname, "Couldn't allocate delayprobelist memory");
					remove(pidfilename);
					exit(1);
				}
				strcpy(new->hostname, hostname);
				new->family = _family;
				new->portnumber = port;

				SLIST_INSERT_HEAD(delayprobelist, new, entries);

			} else
			{
				HandleError(programname, "Host %s not found; skipped.", hostname);
			}
		}
		/*
		 * else the hostname contained a # character * which means the rest of
		 * the line should be * ignored as it is a comment 
		 */
	}							/* all lines scanned */

	*ip = (ip4_flag | ip6_flag);

#ifdef DEBUG
	SLIST_FOREACH(new, delayprobelist, entries)
	{
		printf("Found %s  %d  %d\n", new->hostname, new->portnumber, new->family);
	}
#endif
	return (0);
}

void FreeConfigFile (struct DelayProbeList *delayprobelist)
{
	struct delayprobetarget *entry;

	while (!SLIST_EMPTY(delayprobelist)) 
	{		   
		/* List Deletion. */
		entry = SLIST_FIRST(delayprobelist);
		SLIST_REMOVE_HEAD(delayprobelist, entries);
		free(entry);
	}
}

/* 
 * -------------------------------------------------------------------------------
 * Subroutine Header
 * Purpose           :  signal handler for all signals that by default
 *                      cause program to terminate; purpose of this routine
 *                      is to log the event with syslog, so we know what
 *                      happend.
 * 
 * Side Effects      :  exits program
 *         
 */

void handle_signals (int signo)
{

	char                message[1024];

	(void) fflush (stdout);
	(void) fflush (stderr);

	sprintf (message, "Exiting on signal %d. informing parent ControlDaemon ...", signo);

	errno = 0;
	HandleError (programname, message);

	kill (control_pid, SIGHUP);

	exit (0);

}
