#!/usr/bin/perl

#
# This script runs multicast traceroutes on all monitored multicasts addresses towards all sources. It
# parses the output of mtrace and saves them in the graphviz dot language. 
#
use POSIX;
use IO::Socket;
use Sys::Hostname;
use IO::Interface::Simple;

$myhost = hostname;
$data_area = "/data/mcast";
$trace_area = "/data/mcast";
$trace_temp_area = "/data/mcast/trace/temp";
$bin_area = $ENV{"TTROOT"}."/bin";
$config_area = $ENV{"TTROOT"}."/config/mcast";

`mkdir -p $trace_temp_area`;

# In case there is a hanging trace process kill it
opendir( PROC, "/proc/" );
foreach $filename (readdir( PROC )) {
   if( $filename != $$ && $filename ne "." && $filename ne ".." && `cat /proc/$filename/cmdline` =~ /mc_trace/ ) {
      `kill $filename`;
   }
}

# Check the config file if we need to use a special interface
$interface = "";
if( -e "$config_area/interface.conf" ) {
   open( CONFIG, "$config_area/interface.conf" );
   while (<CONFIG>) {
      chomp;
      if( substr( $_, 0, 1) eq "#" ) {
         next; # Skip comments
      }
      if( $_ =~ /^\s*InterfaceName\s+([A-Za-z0-9]+)\s*$/ ) {
          $ifname = $1;
      }
   }
   close CONFIG;
}
if( defined( $ifname ) && $ifname ne "0" ) {
  # We have an interface name, let's find our IP
  my @interfaces = IO::Interface::Simple->interfaces;
  foreach $if (@interfaces) {
    if( $ifname eq $if ) {
      $interface = $if->address;
    }
  }
}

# Number of concurrent traceroutes
$maxProcs = 5;
$procCount = 0;
$SIG{CHLD} = \&catchChild;

# Maximum Retries for failed or incomplete traces
$maxRetry = 5;


# Read the cached locations
open( LOCFILE, "$config_area/traceLocations" );
while( <LOCFILE> ) {
   if( $_ =~ /^(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})\s(.*)$/ ) {
      $config{$1} = $2;
   }
}
close LOCFILE;

# See what traces we have to do...
opendir (DATADIR, $data_area) || die "Cannot open $data_area";

my %groups;
# Go through all filename, reverse sort to make sure that latest MRCP shows up first.
$foundMRCP = 0;
foreach $filename (sort {$b cmp $a} readdir (DATADIR)) {
   if( $filename =~ /^CRCP\.(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})\.\d*\.(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})\..*$/ ) {
      # Custom Listener file
      #print "Group: $1, Source: $2, Filename: $filename\n";
      if( exists $groups{$1} ) {
         # We got an entry for this group
         if( exists $groups{$1}{$2} ) {
            # Source entry already there, do nothing
         } else {
            # Add source
            $groups{$1}{$2} = 1;
         }
      } else {
         # We don't have an entry for this group, make one and add the source.
         %newSource = ( $2 => 1 );
         $groups{$1}{$2} = $newSource{$2};
      }
   } elsif( $foundMRCP < 3 && $filename =~ /^MRCP\..+/ && $filename =~ /\.(\d{8})\.\d{6}$/ ) {
      # We only do the latest 3 files, since they should have all the relevant information.
      $foundMRCP++;

      $date = $1;
      open( DATAFILE, "$data_area/$filename" );
      while( <DATAFILE> ) {
         @line = split( $_, /\s/ );
         $source = $line[3];
         $group = $line[5];
         if( exists $groups{$group} ) {
            # We got an entry for this group
            if( exists $groups{$group}{$source} ) {
               # Source entry already there, do nothing
            } else {
               # Add source
               $groups{$group}{$source} = 1;
            }
         } else {
            # We don't have an entry for this group, make one and add the source.
            %newSource = ( $source => 1 );
            $groups{$group}{$source} = $newSource{$source};
         }
      }
   }
}

$timestamp = time();
$count = 0;

# First we only do traceroutes...
foreach $group (keys %groups) {
   %sources = %{ $groups{$group} };
   foreach $source (keys %sources) {
      if( $procCount >= $maxProcs ) {
         wait;
      }
      $pid = fork();
      if( $pid == 0 ) {
         if( $interface eq "" ) {
            &runTrace($source, $myhost, $group, $timestamp);
            exit;
         } else {
            &runTrace($source, $myhost, $group, $timestamp, $interface);
            exit;
         }
      } else {
         $procCount++;
      }
   }
}
# And wait for them to finish
while( $procCount > 0 ) {
   wait;
}
#print "Traceroutes finished, doning the rest...\n";

# Our hash is filled, traceroutes are done, let's analyze
foreach $group (keys %groups) {
   %sources = %{ $groups{$group} };
   foreach $source (keys %sources) {
      # Run the traceroutes and parse output
      $count++;
      @graph = ( );
      %output = ( );
      open(MTRACE, "$trace_temp_area/MTRC.$source.$myhost.$group.$timestamp") or next;
      while( <MTRACE> ) {
         # Check we have one those lines where the IP is in brackets (the lines we want!)
         if( $_ =~ /\((\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})\)/ ) {
            # We will need a list of nodes in the correct order 
            push @graph, $1;
            $ip = $1;
            # And a hash which stores the other stuff mtrace outputs for each ip
            $_ =~ s/\*//g; # Get rid of stars in mtrace output
            $_ =~ s/\s+$//; # Get rid of whitespaces we don't need
            $_ =~ s/^\s+//; # Get rid of whitespaces we don't need
            $_ =~ s/\s+/ /g; # Get rid of whitespaces we don't need
            @out = split (/ /, $_);
            $output{$ip} = [ @out ];
         }
      }
      close MTRACE;
      # See if our graph array is okay, last sanity check before doing actual parsing
      if( !defined(@graph) || $graph[scalar(@graph)-1] ne $source ) {
         next;
      }
      # Write the file used by graphviz
      open(GRAPHFILE, ">".$trace_area."/MTRC.$group.$myhost.$source.$timestamp");
      print GRAPHFILE "digraph path {\n";
      print GRAPHFILE "   node [shape=record];\n";
      $hopcount = 0;
      # If these colors are changed or extended they also have to be added to the getLegend function on rimu/cedar (file graphod.pl)
      push @colors, "aliceblue";
      push @colors, "bisque";
      push @colors, "azure";
      push @colors, "aquamarine";
      push @colors, "cornflowerblue";
      push @colors, "lightblue1";
      push @colors, "lightcoral";
      push @colors, "lightpink";
      push @colors, "palegreen";
      push @colors, "lightsalmon";
      foreach $node (@graph) {
         # Make a nice label and pop it into our graphwiz dot file
         $location = &getLocation($node);
         if( $hopcount == 0 ) {
            $shape = "house"; # Make destination a polygon
         } elsif( $hopcount == scalar(@graph)-1 ){
            $shape = "ellipse"; # Make source an ellipse
         } else {
            $shape = "box"; # Leave all others boxes
         }
         $label = "IP: $node";

         # Actual fetching of AS number disabled for now, whois hangs unpredictably, should be implemented using a fork/timeout.
         $asnumber = &getAS($node);
         if( $asnumber > 0 ) {
            $color = $colors[$asnumber%(scalar(@colors))];
            $label .= " AS: $asnumber";
         } else {
            $asnumber = 0;
            $color = "White";
         }

         # Here comes the magic of actually interpreting the output of mtrace.
         # First the hostname, this is rather straightforward
         if( defined($output{$node}[1]) && $output{$node}[1] ne "?" ) {
            $hostname = $output{$node}[1];
            $label .= "\\nHostname: $hostname";
         }
         if( $location ne "" ) {
            $label .= "\\nLocation: $location";
         }
         # Then we see if we have other stuff to add, if yes make a new line
         if( defined($output{$node}[3]) || (defined($output{$node}[5]) && $output{$node}[5] =~ /\d/) || 
            defined($output{$node}[6]) || $output{$node}[scalar(@{$output{$node}})-1] eq "RP/Core" ) {
            $label .= "\\n";
         }
         # The following stuff can be there, or not, if it is there we use it.
         if( defined($output{$node}[3]) ) {
            $protocol = $output{$node}[3];
            $label .= "Protocol: $protocol ";
         }
         if( defined($output{$node}[5]) && $output{$node}[5] =~ /\d/ ) {
            $treshold = $output{$node}[5];
            $label .= "Treshold: $treshold ";
         }
         if( defined($output{$node}[6]) ) {
            if( $output{$node}[6] =~ /\d/ ) {
               $delay = $output{$node}[6];
               $label .= "Delay: $delay ";
            }
         }
         # The last entry of the line can state that we have an RP/Core, let's see if this is true.
         if( $output{$node}[scalar(@{$output{$node}})-1] eq "RP/Core" ) {
            $shape = "hexagon";
            $label .= "RP/Core";
         }
         print GRAPHFILE "   \"$node\" [shape=$shape, label=\"$label\", style=filled, fillcolor=$color, fontsize=10, fontname=\"Arial\"];\n";
         $hopcount++;
      }
      # The actual order will be reversed here, since mtrace outputs from dest to src
      $hopcount = scalar(@graph)-1;
      for( $hopcount; $hopcount > 0; $hopcount-- ) {
         $nexthop = $hopcount - 1;
         print GRAPHFILE "   \"".$graph[$hopcount]."\" -> \"".$graph[$nexthop]."\";\n";
      }
      print GRAPHFILE "}";
      close GRAPHFILE;
   }
}

# And remove our temp files for the traceroutes
`rm $trace_temp_area/*.$timestamp`;

# Finally write the file with the new locations
open( LOCFILE, ">$config_area/traceLocations" );
foreach( keys %config ) {
   if( $config{$_} ne "" ) {
      print LOCFILE "$_ $config{$_}\n";
   }
}
close LOCFILE;

sub runTrace {
   $source = $_[0];
   $myhost = $_[1];
   $group = $_[2];
   $timestamp = $_[3];

   if( defined( $_[4] ) ) {
      $interface = $_[4];
   }

   local $retries = 0;

   #print "Running a trace for $source\n";

   do {
      $retry = 0;
      $retries++;
      # Run the traceroutes
      if( defined($interface) && $interface ne "" ) {
         `mtrace -UM -i $interface -s $source $myhost $group &> $trace_temp_area/MTRC.$source.$myhost.$group.$timestamp`;
      } else {
         `mtrace -UM -s $source $myhost $group &> $trace_temp_area/MTRC.$source.$myhost.$group.$timestamp`;
      }

      # Check the file for a "giving up"
      @graph = ( ); # Empty graph var from previous retries.
      open( FILE, "$trace_temp_area/MTRC.$source.$myhost.$group.$timestamp" );
      while( <FILE> ) {
         if( $_ =~ /giving up/ ) {
            $retry = 1;
            #print "Retry for $source...\n";
         }
         if( $_ =~ /\((\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})\)/ ) {
            # We need a list of nodes in the correct order
            push @graph, $1;
         }
      }
      close FILE;
      # Check if the last one in the list is the source, otherwise we don't have a full path and should retry
      if( !defined(@graph) || $graph[scalar(@graph)-1] ne $source ) {
         $retry = 1;
      }
   } while( $retry == 1 && $retries <= $maxRetry ); # Retry up to maxRetry times

   if( $retries > $maxRetry && $retry == 1 ) {
      # Trace failed completely, we can assume there went something wrong
      # Remove file and move on.
      `rm $trace_temp_area/MTRC.$source.$myhost.$group.$timestamp`;
   }

}

sub getLocation {
   $ip = $_[0];

   if( defined($config{$ip}) ) {
      return $config{$ip};
   }
   # Get the location for this source (Country Code and Guessed variables not yet used, but possible)
   `fetch -o $trace_area/locationoutput -T 3 "http://api.hostip.info/rough.php?ip=$ip" &> /dev/null`;
   open(LOCATION, $trace_area."/locationoutput");
   while( <LOCATION> ) {
      if( $_ =~ /^Country: (.*)$/ ) {
         $country = $1;
      } elsif ( $_ =~ /^Country Code: (.*)$/ ) {
         $cCode = $1;
      } elsif ( $_ =~ /^City: (.*)$/ ) {
         if( $1 eq "(Unknown city)" || $1 eq "(Private adress)" ) {
            $city = "";
         } else {
            $city = $1;
         }
         #} elsif ( $_ =~ /^Guessed: (.*)$/) {
         #  $guessed = $1;
      }
   }
   close LOCATION;
   `rm $trace_area/locationoutput`;
   if( defined($country) && defined($city) ) {
      if( $city eq "" ) {
         $location = $country;
      } else {
         $location = "$city, $country ($cCode)";
      }
   } else {
      $location = "";
   }
   $config{$ip} = $location;
   return $location;
}

sub getAS {
   # First Argument is the IP
   $ip = $_[0];

   # Open a socket to RIPE's whois server
   $sock = IO::Socket::INET->new(PeerAddr => 'riswhois.ripe.net',
      PeerPort => 43,
      Proto    => 'tcp',
      timeout  => "1"
   );

   unless ($sock) {return 0;}

   # Set Socket to autoflush and activate a timeout of 1 second.
   $sock->autoflush(1);

   # -R oprions to ensure we stay on RIPE's server
   print $sock "$ip\n";

   # Grab our AS number from the output
   while( select($_=<$sock>,undef,undef,0.1) ) {
      if( $_ =~ /^origin:\s*AS(\d*)$/ ) {
         $as = $1;
      }
   }
   close $sock;

   # If we got something back return it, otherwise 0.
   if( defined($as) ) {
      return $as;
   } else {
      return 0;
   }
}

sub catchChild {
   $procCount--;
}
