#include <stdio.h>
#include <string.h>
#include <math.h>
#include <termios.h>
#include <sys/types.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>

#define P(x) x

#include "t_struct.h"
#include "t_format.h"
#include "t_parser.h"
#include "t_labels.h"

#include <stdio.h>
#define PALDEVICE "/dev/gps0"	/* default palisade device name */
#define DMSGLEN	  512		/* decoded message string length */

static char const rcsid[] =
"$Id: palctl.c,v 1.4 2005/03/09 13:03:29 tom Exp $";

int baud = 9600;/* baudrate */ 
int dflag = 0; 	/* get satellite system data */
int cdflag = 0;	/* load satellite system data */
int fflag = 0;	/* file specified */
int ciflag = 0; /* Initialise Operating Parameters */
int cfflag = 0; /* Set Factory Defaults and Cold Start */
int lflag = 0;	/* set survey limit */
int mflag = 0;	/* set operating mode */
int mmode = 0;	/* operating mode argument */
int cpflag = 0;	/* set initial position */ 
int qflag = 0;	/* query params */ 
int rflag = 0;	/* reception conditions */ 
int csflag = 0;	/* cold start  */ 
int tflag = 0;	/* trace */ 
int vflag = 0;	/* verbose */ 
int cwflag = 0;	/* warm start */ 
int xflag = 0;  /* stress it */
int pflag = 0;  /* set PPS width */

int debug=0;

int pfd = 0;	/* palisade file descriptor */
FILE *pal;	/* palisade file pointer */
FILE *dfp;	/* satellite data file pointer */
FILE *pfp;	/* position data file pointer */

static char msgbuf[256]; /* general purpose message buffer */

static TSIPPKT cmd;     /* structure for current outgoing TSIP command  */ 

static	int parse0x4F	P((TSIPPKT *));
static	int parse0x5C	P((TSIPPKT *));
static	int parse0x8F0B	P((TSIPPKT *));
static	int parse0x8FAD	P((TSIPPKT *));
static	int parseany	P((TSIPPKT *));
static	void	ncc_send	P((int fd, unsigned char, unsigned char *, unsigned));

/* From Trimble software */
static	void	TranslateTSIPReportToText P((TSIPPKT *, char *));
static	short	bGetShort P((unsigned char *));
static	float	bGetSingle P((unsigned char *));
static	double	bGetDouble P((unsigned char *));
static	void 	bPutFloat P((float *, unsigned char *));
static	void 	bPutDouble P((double *, unsigned char *));
static	void 	bPutLong P((long *, unsigned char *));

static void DisplayTSIPReport (TSIPPKT *rpt, int nolf)
{
	int ireply;
	char outbuf[DMSGLEN];


	TranslateTSIPReportToText (rpt, (char*) &outbuf);

	if (nolf) 
		fputs(outbuf+1, stdout);
	else
		fputs(outbuf, stdout);
	
	rpt->status = TSIP_PARSED_EMPTY;

}

static void DisplayTSIPraw(TSIPPKT *rpt)
{
	int i;
	printf("RAW %02X", rpt->len);
	for (i = 0; i < rpt->len; i++) {
		if ((i % 20) == 0) printf("\n");
		printf(" %02X", rpt->buf[i]);
	}
	printf("\n");
}

abort(char *msg) {
	fprintf(stderr, msg);
	exit(1);
}

usage() {
	printf("usage: palctl [-v] [-t] [-f dev] [commands]\n");
	printf("\n");
	printf("  -v      verbose \n");
	printf("  -t      trace \n");
	printf("  -f dev  clock special file\n");
	printf("\n");
	printf("Commands:\n");
	printf("  -d      get satellite system data\n");
	printf("  -D fle  load satellite system data\n");
	printf("  -F      set factory defaults, cold start  \n");
	printf("  -I      Initialise Operating Parameters\n");
	printf("  -l lmt  set survey limit\n");
	printf("  -m mde  set operating mode\n");
	printf("  -P fle  set initial position \n");
	printf("  -q      query params \n");
	printf("  -r      reception conditions \n");
	printf("  -S      cold start  \n");
	printf("  -W      warm start \n");
	printf("  -x      stress test to find hang problem\n");
	printf("  -p wdt  set PPS pulse width to 'wdt' microseconds\n");
	printf("\n%s\n", rcsid);
}

rptchanged (unsigned char code, char *rec) { /* decoded report changed ? */

	static char bufs[256][DMSGLEN];
	int result = 0;

	result = strncmp(rec, bufs[code], DMSGLEN);
	if(result) {
		strncpy(bufs[code], rec, DMSGLEN);
	}
	return(result);
}


showchanged(TSIPPKT *rpt) {	/* show report if it changed */

	static char buf[DMSGLEN];

	
	if (vflag) {
		DisplayTSIPReport (rpt, 0); 
	} 
	else { 
		TranslateTSIPReportToText (rpt, buf);
		if (rptchanged(rpt->code, buf)) 
			DisplayTSIPReport (rpt, 0); 
	}

}


sendbyte(unsigned char c) {	/* send octet to palisade */

	if (debug)
		fprintf(stderr, "sendbyte(%02X)\n", c);

	if (write(pfd, &c, 1) != 1) {
		snprintf(msgbuf, sizeof msgbuf, 
			"palisade write: %s\n", 
			 strerror(errno)
		);
		abort(msgbuf);
	}
}


TSIPPKT *getrpt(unsigned char code, unsigned char subcode) {

/* This function reads the current palisade outout stream and
 * returns the next report with matching code and subcode.
 * A code/subcode of 0 matches any report 
 */  
	static TSIPPKT rpt;     /* current incoming TSIP report  */ 
	int newbyte;

	if (debug) 
		fprintf(stderr, "getrpt(%X, %X)\n", code, subcode);

	while ((newbyte=fgetc(pal)) != EOF) {
		tsip_input_proc (&rpt, (unsigned char)newbyte);
		if (debug) 
			fprintf(stderr, "getrpt got %X,%X\n", rpt.code, rpt.buf[0]);
		if (rpt.status == TSIP_PARSED_FULL) {
			if (code==0 || rpt.code == code) {
				if (subcode==0 || rpt.buf[0]==subcode)
				return(&rpt);
			}
		}
	}
	abort("\nEOF on Palisade connection.\n");
}	


showrpt(unsigned char code, unsigned char subcode, int nolf) {

	TSIPPKT *rpt;	/* current TSIP report packet */

	if (debug) 
		fprintf(stderr, "showrpt(%X, %X, %d)\n", code, subcode, nolf);
	
	rpt = getrpt(code, subcode);
      	DisplayTSIPReport (rpt, nolf);
}


showraw(unsigned char code, unsigned char subcode) {

	TSIPPKT *rpt;	/* current TSIP report packet */
	
	rpt = getrpt(code, subcode);
      	DisplayTSIPraw(rpt);
}

	
showrpts(unsigned char code, unsigned char subcode, unsigned nparse) {

	TSIPPKT *rpt;	/* current TSIP report packet */
	int n;

	n = nparse;
	while (n>0) {
		rpt = getrpt(0, 0);
		if (code==0 || rpt->code == code) {
			if (subcode==0 || rpt->buf[0]==subcode) {
      				DisplayTSIPReport (rpt, 0);
			}
		}
		n--;
	}
}


void dovquery() {	/* verbosely query and display palisate status */

	TSIPPKT *rpt;	/* current TSIP report packet */
	unsigned sv;

	cmd_0x1F(&cmd);			/* request software versions */
	send_cmd(&cmd, sendbyte);
	showrpt(0x45, 0, 1);

	cmd_0x24(&cmd);			/* request position fix mode */
	send_cmd(&cmd, sendbyte);
	showrpt(0x6d, 0, 0);

	cmd_0x26(&cmd);			/* request health */
	send_cmd(&cmd, sendbyte);
	showrpt(0x46, 0, 0);
	showrpts(0x4b, 0, 2);

	cmd_0x27(&cmd);			/* request signal levels */
	send_cmd(&cmd, sendbyte);
	showrpt(0x4b, 0, 0);

	cmd_0x29(&cmd);			/* request almanac health page */
	send_cmd(&cmd, sendbyte);
	showrpt(0x47, 0, 0);

	cmd_0x2Cq(&cmd);		/* request operating parameters */
	send_cmd(&cmd, sendbyte);
	showrpt(0x49, 0, 0);

	cmd_0x2D(&cmd);			/* request oscillator offset */
	send_cmd(&cmd, sendbyte);
	showrpt(0x4c, 0, 0);

	cmd_0x2F(&cmd);			/* request UTC parameters */
	send_cmd(&cmd, sendbyte);
	showrpt(0x4d, 0, 0);

	cmd_0x35q(&cmd);		/* request I/O Options */
	send_cmd(&cmd, sendbyte);
	showrpt(0x55, 0, 0);

	cmd_0x39(&cmd, 3, 0);		/* request SV enable */
	send_cmd(&cmd, sendbyte);
	showrpt(0x59, 0, 0);

	cmd_0x39(&cmd, 6, 0);		/* request SV ignore health  */
	send_cmd(&cmd, sendbyte);
	showrpt(0x59, 0, 0);

	cmd_0x3A(&cmd, 0);		/* request raw mewasurements */
	send_cmd(&cmd, sendbyte);
	showrpts(0x5A, 0, 12);

	cmd_0x3B(&cmd, 0);		/* request ephemeris */
	send_cmd(&cmd, sendbyte);
	showrpts(0x5B, 0, 30);

	cmd_0x3C(&cmd, 0);		/* request traking status  */
	send_cmd(&cmd, sendbyte);
	showrpts(0x5C, 0, 30);

	for (sv=1; sv<33; sv++) {
		cmd_0x20(&cmd, sv);	/* request almanac */
		send_cmd(&cmd, sendbyte);
		showrpts(0x40, 0, 4);
	}

	cmd_0x3Dq(&cmd);		/* request port A config */
	send_cmd(&cmd, sendbyte);
	showrpt(0x3D, 0, 0);

	cmd_0x8E20q(&cmd);		/* request last fix */
	send_cmd(&cmd, sendbyte);
	showrpt(0x8f, 0x20, 0);

	cmd_0x8E41q(&cmd);		/* request manuf. op. params */
	send_cmd(&cmd, sendbyte);
	showrpt(0x8f, 0x41, 0);

	cmd_0x8E42q(&cmd);		/* request production params */
	send_cmd(&cmd, sendbyte);
	showrpt(0x8f, 0x42, 0);

	cmd_0x8E4Aq(&cmd);		/* request PPS characteristics */
	send_cmd(&cmd, sendbyte);
	showrpt(0x8f, 0x4A, 0);

	cmd_0x8E4Bq(&cmd);		/* request survey limit */
	send_cmd(&cmd, sendbyte);
	showrpt(0x8f, 0x4b, 0);

	cmd_0x8E4Dq(&cmd);		/* request packet output msk */
	send_cmd(&cmd, sendbyte);
	showrpt(0x8f, 0x4d, 0);

	cmd_0x8EA5q(&cmd);		/* request super packet output msk */
	send_cmd(&cmd, sendbyte);
	showrpt(0x8f, 0xa5, 0);

	cmd_0x8EADq(&cmd);		/* request Primary UTC time */
	send_cmd(&cmd, sendbyte);
	showrpt(0x8f, 0xad, 0);

	cmd_0x8E0Bq(&cmd);		/* request comprehensive time */
	send_cmd(&cmd, sendbyte);
	showrpt(0x8f, 0x0b, 0);

	printf ("\n");
}

void doquery() {	/* query and display palisate status */

	TSIPPKT *rpt;	/* current TSIP report packet */
	unsigned sv;

	cmd_0x24(&cmd);			/* request position fix mode */
	send_cmd(&cmd, sendbyte);
	showrpt(0x6d, 0, 1);

	cmd_0x8E0Bq(&cmd);		/* request comprehensive time */
	send_cmd(&cmd, sendbyte);
	showrpt(0x8f, 0x0b, 0);

	cmd_0x8EADq(&cmd);		/* request Primary UTC time */
	send_cmd(&cmd, sendbyte);
	showrpt(0x8f, 0xad, 0);


	printf ("\n");
}


main(int argc, char **argv) {

	extern char *optarg;	/* getopt return values */
	extern int optind;

	static int slimit;	/* self survey limit */
	static char smode[256]; /* operating mode string */

	TSIPPKT *rpt;		/* current TSIP report packet */
	int ch;
	char *ptr;

	if (argc<2) {
		usage();
		exit(0);
	}


	while ((ch = getopt(argc, argv, "b:dD:f:FIl:m:P:qrStvWxp:")) != -1) {
		switch(ch) {
			case 'b':
				if (sscanf(optarg, "%d", &baud) != 1) {
					usage(); 	
					exit(1);
				}
				switch (baud) {
				case 110:
				case 300:
				case 600:
				case 1200:
				case 2400:
				case 4800:
				case 9600:
				case 19200:
				case 38400:
					break;
				default:
					printf("%d: invalid baudrate\n", baud);
					printf("valid rates: 110,300,600,1200,2400,4800,9600,19200,38400\n\n");
					usage();
					exit(1);
				
				}
				break;
			case 'd':
				dflag = 1;
				break;
			case 'D':
				if ((dfp = fopen(optarg, "r")) == NULL) {
					snprintf(msgbuf, sizeof msgbuf,
						"open %s: %s\n", 
						optarg, strerror(errno)
					);
					abort(msgbuf);
				}
				cdflag = 1;
				break;
			case 'f':
				if ((pfd = open(optarg, O_RDWR, 0)) < 0) {
					snprintf(msgbuf, sizeof msgbuf,
						"open %s: %s\n", 
						optarg, strerror(errno)
					);
					abort(msgbuf);
				}
				fflag = 1;
				break;
			case 'F':
				cfflag = 1;
				break;
			case 'I':
				ciflag = 1;
				break;
			case 'l':
				lflag = 1;
				if (sscanf(optarg, "%d", &slimit) != 1) {
					usage(); 	
					exit(1);
				}
				break;
			case 'm':
				if (sscanf(optarg, "%255s", &smode) != 1) {
					usage(); 	
					exit(1);
				}
				if (strncmp(smode, "auto", sizeof smode) == 0) 
					mmode = 0;
				else if (strncmp(smode, "odt", sizeof smode) == 0) 
					mmode = 10;
				else if (strncmp(smode, "2d", sizeof smode) == 0) 
					mmode = 3;
				else if (strncmp(smode, "1d", sizeof smode) == 0)
					mmode = 1;
				else if (strncmp(smode, "3d", sizeof smode) == 0)
					mmode = 4;
				else {
					usage(); 	
					exit(1);
				}
				mflag = 1;
				break;
			case 'P':
				if ((pfp = fopen(optarg, "r")) == NULL) {
					snprintf(msgbuf, sizeof msgbuf,
						"open %s: %s\n", 
						optarg, strerror(errno)
					);
					abort(msgbuf);
				}
				cpflag = 1;
				break;
			case 'q':
				qflag = 1;
				break;
			case 'r':
				rflag = 1;
				break;
			case 'S':
				csflag = 1;
				break;
			case 't':
				tflag = 1;
				break;
			case 'v':
				vflag = 1;
				break;
			case 'W':
				cwflag = 1;
				break;
			case 'x':
				xflag = 1;
				break;
			case 'p':
				pflag = (int)strtol(optarg, &ptr, 10);
				if(*ptr != 0) {
					snprintf(msgbuf, sizeof(msgbuf), 
							"not a valid PPS pulse width: %s\n",
							optarg);
					abort(msgbuf);
				}
				break;
			case '?':
			default:
				usage();
				exit(1);
		}
	}
	argc -= optind;
	argv += optind;

	if (!fflag) {
		if ((pfd = open(PALDEVICE, O_RDWR, 0)) < 0) {
			snprintf(msgbuf, sizeof msgbuf, 
				"open %s: %s\n", 
				PALDEVICE, strerror(errno)
			);
			abort(msgbuf);
		}
	}
		
	if ((pal=fdopen(pfd, "rw")) == NULL) {
		snprintf(msgbuf, sizeof msgbuf, 
			"fdopen %s: %s\n", 
			optarg, strerror(errno)
		);
		abort(msgbuf);
	}
	setline(pfd, baud);


	if (xflag) {	/* exercise */
		unsigned pid;
		switch (pid=fork()) {
		case -1:
			snprintf(msgbuf, sizeof msgbuf, 
				"fork(): %s\n", strerror(errno)
			);
			abort(msgbuf);
			break;
		case 0:
			initstress();
			sleep(1);
			while(1) {
				parsestress();
			}
			break;
		default:
			while(1) {
				ncc_send(pfd, 0x8E, "\xAD", 1); /* poll for UTC superpacket */
				sleep(1);
			}
                }
		exit(0);
        }


	if (cfflag) {
		cmd_0x1E(&cmd, 'F');
		send_cmd(&cmd, sendbyte);
		showrpts(0, 0, 5);
	}

	if (csflag) {
		cmd_0x1E(&cmd, 'K');
		send_cmd(&cmd, sendbyte);
		showrpts(0, 0, 5);
	}

	if (cwflag) {
		cmd_0x25(&cmd);
		send_cmd(&cmd, sendbyte);
		showrpts(0, 0, 5);
	}

	if (cdflag) {
		unsigned len, b, i;
		char tag[256];

		while (fscanf(dfp, " %s RAW %x", &tag, &len) == 2) {
			for (i=0; i<len; i++) {
				if (fscanf(dfp, "%x", &b) != 1) {
					abort ("read sat data file\n");
				}
				cmd.buf[i] = b;
			}
			cmd.code = 0x38;
			cmd.len = len;
			if (vflag)
				printf("\nLoading %s ...", tag);
			send_cmd(&cmd, sendbyte);
			rpt=getrpt(0x58,0);
			if (rpt->buf[0] == 1) {
				if (vflag) 
					printf(" OK");
			}
			else {
				if (vflag)
					printf(" Failed!");
				fprintf(stderr, "loading %s failed.\n", tag);
			}
		}
		cmd_0x26(&cmd);			/* request health */
		send_cmd(&cmd, sendbyte);
		showrpt(0x46, 0, 0);
		showrpts(0x4b, 0, 1);

		cmd_0x37(&cmd);			/* request pos */
		send_cmd(&cmd, sendbyte);
		showrpt(0x84, 0, 0);
	}

	if (cpflag) {
		int found =0;
		int deglat, deglon;
		float minlat, minlon, alt, radlat, radlon;
		char ns, ew;

		while (!found) {
			if (fscanf(pfp, "posapp-lla: %d %f %c %d %f %c %f", 
			&deglat, &minlat, &ns, &deglon, &minlon, &ew, &alt) 
			== 7) {
				found=1;
			}
		}
		if (!found) {
			fprintf(stderr, "no position found in file\n");
		}
		else {
			radlat = ((float)deglat + (minlat/60.0)) * (float) D2R;
			if (ns == 'S') {
				radlat = -radlat;
			}
			else if (ns != 'N') {
				fprintf(stderr, "lat designator %c unknown, N assumed\n", ns);
			}
			radlon = ((float)deglon + (minlon/60.0)) * (float) D2R;
			if (ew == 'W') {
				radlon = -radlon;
			}
			else if (ew != 'E') {
				fprintf(stderr, "lon designator %c unknown, E assumed\n", ns);
			}
			cmd_0x2B(&cmd, radlat, radlon, alt);
			send_cmd(&cmd, sendbyte);
		}
	}

	if (ciflag) {
		int i;
		float f;

		/* i/o opts */
		cmd_0x35s(&cmd, 0x1C, 0x00, 0x05, 0x0A);
		/**** cmd_0x35s(&cmd, 0x3E, 0x02, 0x01, 0x0A); ****/
		send_cmd(&cmd, sendbyte);
		showrpts(0x55, 0, 1);

		/* turn off port A */
		cmd_0x3Ds(&cmd, 11, 11, 7, 7, 1, 0);
		send_cmd(&cmd, sendbyte);
		showrpts(0x3D, 0, 1);

		/* primary receiver config */
		cmd.code = 0xBB;
		cmd.len = 43;
		cmd.buf[0] = 0; 		/* subcode */
		cmd.buf[1] = 0, 		/* fix mode = auto */
		cmd.buf[2] = 0; 		/* DGPS off */
		cmd.buf[3] = 4; 		/* dynamics = stationary */
		cmd.buf[4] = 2; 		/* weighted overdet. sol. */
		f = 10. * D2R;			/* elevation mask */
		bPutFloat(&f, &cmd.buf[5]);	
		f = 5.;				/* SNR mask */
		bPutFloat(&f, &cmd.buf[9]);	
		f = 8.;				/* PDOP mask */
		bPutFloat(&f, &cmd.buf[13]);	
		f = 6.; 			/* PDOP switch */
		bPutFloat(&f, &cmd.buf[17]);	
		cmd.buf[21] = 30;		/* DGPS age */
		cmd.buf[22] = 1;		/* foilage = sometimes */
		cmd.buf[23] = 0xFF;		/* low power mode N/A */
		cmd.buf[24] = 0xFF;		/* clock hold mode N/A */
		cmd.buf[25] = 0x00;		/* measurement rate 1 Hz */
		cmd.buf[26] = 0x00;		/* fix rate 1 Hz */
		for(i=27; i<43; i++) {
			cmd.buf[i] = 0xFF;
		}
		send_cmd(&cmd, sendbyte);
		showrpts(0xBB, 0, 1);

		/* set PPS characteristics */
		cmd_0x8E4As(&cmd, 
			1,			/* PPS on */
			1,			/* UTC */
			0,			/* polarity = positive */
			(double) 0.,		/* 100 ft cable */
			(float) 300.		/* bias uncert. threshold */
		);
		send_cmd(&cmd, sendbyte);
		cmd_0x8E4Aq(&cmd);		
		send_cmd(&cmd, sendbyte);
		showrpt(0x8f, 0x4A, 0);

		/* set auto packet output mask */
		cmd_0x8E4Ds(&cmd, (long) 0xC0000000); /* all off */
		send_cmd(&cmd, sendbyte);
		cmd_0x8E4Dq(&cmd);		
		send_cmd(&cmd, sendbyte);
		showrpt(0x8f, 0x4D, 0);


		
	}

	if (dflag) {
		unsigned sv;

		for (sv=1; sv<33; sv++) {	/* request almanac */
			cmd_0x38q(&cmd, 2, sv);	
			send_cmd(&cmd, sendbyte);
			printf("Almanac-%02d: ", sv);
			showraw(0x58, 0);
		}
		
		cmd_0x38q(&cmd, 3, 0);		/* request health page */
		send_cmd(&cmd, sendbyte);
		printf("Health: ");
		showraw(0x58, 0);

		cmd_0x38q(&cmd, 4, 0);		/* request ionosphere model */
		send_cmd(&cmd, sendbyte);
		printf("Ionosphere: ");
		showraw(0x58, 0);

		cmd_0x38q(&cmd, 5, 0);		/* request UTC data */
		send_cmd(&cmd, sendbyte);
		printf("UTC: ");
		showraw(0x58, 0);

#ifdef IONO
		for (sv=1; sv<33; sv++) {	/* request ephemeris */
			cmd_0x38q(&cmd, 6, sv);	
			send_cmd(&cmd, sendbyte);
			printf("Ephemeris-%02d: ", sv);
			showraw(0x58, 0);
		}
#endif

		printf("\n");
	}

		

	if (lflag) {
		cmd_0x8E4Bq(&cmd);		/* request survey limit */
		send_cmd(&cmd, sendbyte);
		showrpt(0x8f, 0x4b, 0);

		cmd_0x8E4Bs(&cmd, (long)slimit);	/* set survey limit */
		send_cmd(&cmd, sendbyte);

		cmd_0x8E4Bq(&cmd);		/* request survey limit */
		send_cmd(&cmd, sendbyte);
		showrpt(0x8f, 0x4b, 0);

		printf("\n");
	}


	if (mflag) {
		cmd_0x22(&cmd, (char) mmode);	/* set position fix mode */
		send_cmd(&cmd, sendbyte);
		showrpt(0x6d, 0, 1);

		cmd_0x24(&cmd);			/* request position fix mode */
		send_cmd(&cmd, sendbyte);
		showrpt(0x6d, 0, 0);

	}


	
	if (qflag) {
		if (vflag)
			dovquery();
		else
			doquery();
		exit(0);
	}

	if (rflag) {
		cmd_0x24(&cmd);
		send_cmd(&cmd, sendbyte);
		showrpts(0x6D, 0, 1);
		cmd_0x3C(&cmd, 0);		/* request traking status  */
		send_cmd(&cmd, sendbyte);
		showrpts(0x5C, 0, 15);
	}

	if (pflag) {
		/*
		 * Set PPS pulse width
		 */
		cmd_0x8E4Fs(&cmd, (double)pflag);
		send_cmd(&cmd, sendbyte);
		showrpts(0x8F, 0x4F, 1);
	}

	while(tflag) {
		rpt = getrpt(0,0);	
		switch (rpt->code) {
			case 0x41: /* GPS Time */
				if (vflag) DisplayTSIPReport (rpt, 0);
				break;
			case 0x46: /* RCVR Health */
			case 0x4B: /* RCVR ID & Status */
				showchanged(rpt);
				break;
			case 0x54: /* time Fix */
      				DisplayTSIPReport (rpt, 0);
				break;
			case 0x43: /* Velocity ECEF*/
			case 0x56: /* Velocity ENU */
				if (vflag) DisplayTSIPReport (rpt, 0);
				break;
			case 0x6D: /* Mode & SVs */
			case 0x82: /* DGPS Mode */
				showchanged(rpt);
				break;
			case 0x8f:
				if (rpt->buf[0] == 0x20 && !vflag) /* fix */
					break;
      				DisplayTSIPReport (rpt, 0);
				break;
			default:
      				DisplayTSIPReport (rpt, 0);
				break;
		}
		fflush(stdout);
	}
	printf ("\n");
}


static void
ncc_send(fd, cmd, arg, len)
	int fd;
	unsigned char cmd; 
	unsigned char *arg;
	unsigned len;
{

	unsigned char *ip, *op;
	unsigned char obuf[512];


	ip = arg;
	op = obuf;

	*op++ = 0x10;
	*op++ = cmd;

	while (len--) {
		if (op-obuf > sizeof(obuf)-5) {
			printf("ncc_send obuf overflow!");
			return;
		}
			
		if (*ip == 0x10)  /* byte stuffing */
			*op++ = 0x10;
		*op++ = *ip++;
	}
	
	*op++ = 0x10;
	*op++ = 0x03;

	if (write(fd, obuf, op-obuf) == -1) {
			printf("ncc_send write error!");
	}
}


/*
 * Parse UTC Parameter Packet
 * 
 * See the IDE for documentation!
 *
 * 0 = success
 * -1 = errors
 */

int parse0x4F(rpt)
	TSIPPKT *rpt;
{
	register struct nccunit *up;

	double a0;
	float a1, tot;
	int dt_ls, wn_t, wn_lsf, dn, dt_lsf;

   	static char logbuf[1024];	/* logging string buffer */
	unsigned char *buf;
		
	buf = rpt->buf;
	
	if (rpt->len != 26) 
		return (-1);
	a0 = bGetDouble (buf);
	a1 = bGetSingle (&buf[8]);
	dt_ls = bGetShort (&buf[12]);
	tot = bGetSingle (&buf[14]);
	wn_t = bGetShort (&buf[18]);
	wn_lsf = bGetShort (&buf[20]);
	dn = bGetShort (&buf[22]);
	dt_lsf = bGetShort (&buf[24]);

	sprintf(logbuf, "L1 %d %d %d %g %g %g %d %d %d",
		dt_lsf - dt_ls, dt_ls, dt_lsf, a0, a1, tot, wn_t, wn_lsf, dn); 

   	puts(logbuf);
	return (0);
}



/*
 * Parse Tracking Status packet
 *
 * 0 = success
 * -1 = errors
 */

int parse0x5C(rpt)
	TSIPPKT *rpt;
{
	unsigned char prn, channel, aqflag, ephstat;
	float snr, azinuth, elevation;

   	static char logbuf[1024];	/* logging string buffer */
	unsigned char *buf;
		
	buf = rpt->buf;
	
	if (rpt->len != 24) 
		return(-1);

	prn = buf[0];
	channel = (unsigned char)(buf[1] >> 3);
	if (channel == 0x10) 
		channel = 2;
	else 
		channel++;
	aqflag = buf[2];
	ephstat = buf[3];
	snr = bGetSingle(&buf[4]);
	elevation = bGetSingle(&buf[12]) * R2D;
	azinuth = bGetSingle(&buf[16]) * R2D;

	sprintf(logbuf, "S1 %02d %d %d %02x %4.1f %5.1f %4.1f",
		prn, channel, aqflag, ephstat, snr, azinuth, elevation);

   	puts(logbuf);

	return (0);
}


/*
 * Parse primary utc time packet
 * and fill refclock structure 
 * from results. 
 *
 * 0 = success
 * -1 = errors
 */

int parse0x8FAD(rpt)
	TSIPPKT *rpt;
{
	unsigned day, month, year;	/* data derived from received timecode */
	unsigned hour, minute, second;
	unsigned char trackstat, utcflags;

   	static char logbuf[1024];	/* logging string buffer */
	int i;
	unsigned char *buf;
		
	buf = rpt->buf;

	if (rpt->len != 22) 
		return (-1);
	
	if (bGetShort(&buf[1]) != 0) {
		printf("parse0x8FAD: event count != 0\n");
		return(-1);
	}


	if (bGetDouble(&buf[3]) != 0.0) {
		printf("parse0x8FAD: fracsecs != 0\n");
		return(-1);
	}

	hour = (unsigned int) buf[11];
	minute = (unsigned int) buf[12];
	second = (unsigned int) buf[13];
	day =		(unsigned int) buf[14];
	month =		(unsigned int) buf[15];
	year =		bGetShort(&buf[16]);
	trackstat = buf[18];
	utcflags = buf[19];


	sprintf(logbuf, "U1 %d.%d.%d %02d:%02d:%02d %d %02x",
		day, month, year, hour, minute, second, trackstat, utcflags);

   	puts(logbuf);



	return (0);
}


/*
 * Parse comprehensive time packet 
 *
 * 0 = success
 * -1 = errors
 */

int parse0x8F0B(rpt)
	TSIPPKT *rpt;
{

	unsigned day, month, year;	/* data derived from received timecode */
	unsigned hour, minute, second;
	unsigned utcoff;
	unsigned char mode;
	double  bias, rate;
	float biasunc, rateunc;
	double lat, lon, alt;
	short lat_deg, lon_deg;
	float lat_min, lon_min;
	unsigned char north_south, east_west;
	char sv[9];

   	static char logbuf[1024];	/* logging string buffer */
	unsigned char b;
	int i;
	unsigned char *buf;
	double tow;
		
	buf = rpt->buf;

	if (rpt->len != 74) 
		return (-1);
	
	if (bGetShort(&buf[1]) != 0)
		return(-1);;

	tow =  bGetDouble(&buf[3]);

	if (tow == -1.0) {
		return(-1);
	}
	else if ((tow >= 604800.0) || (tow < 0.0)) {
		return(-1);
	}
	else
	{
		if (tow < 604799.9) tow = tow + .00000001;
		second = (unsigned int) fmod(tow, 60.);
		minute =  (unsigned int) fmod(tow/60., 60.);
		hour = (unsigned int )fmod(tow / 3600., 24.);
	} 


	day =		(unsigned int) buf[11];
	month =		(unsigned int) buf[12];
	year =		bGetShort(&buf[13]);
	mode =		buf[15];
	utcoff =	bGetShort(&buf[16]);
	bias = 		bGetDouble(&buf[18]) / GPS_C * 1e9;	/* ns */
	rate = 		bGetDouble(&buf[26]) / GPS_C * 1e9;	/* ppb */ 
	biasunc = 	bGetSingle(&buf[34]) / GPS_C * 1e9;	/* ns */
	rateunc = 	bGetSingle(&buf[38]) / GPS_C * 1e9;	/* ppb */
	lat = 		bGetDouble(&buf[42]) * R2D;
	lon = 		bGetDouble(&buf[50]) * R2D;
	alt = 		bGetDouble(&buf[58]);

	if (lat < 0.0) {
		north_south = 'S';
		lat = -lat;
	}
	else {
		north_south = 'N';
	}
	lat_deg = (short)lat;
	lat_min = (lat - lat_deg) * 60.0;

	if (lon < 0.0) {
		east_west = 'W';
		lon = -lon;
	}
	else {
		east_west = 'E';
	}

	lon_deg = (short)lon;
	lon_min = (lon - lon_deg) * 60.0;

	for (i=0; i<8; i++) {
		sv[i] = buf[i + 66];
		if (sv[i]) {
			b = (unsigned char) (sv[i]<0 ? -sv[i] : sv[i]);
			ncc_send(pfd, 0x3C, &b, 1 );
		}
	}


	sprintf(logbuf, "C1 %02d%02d%04d %02d%02d%02d %d %7.0f %.1f %.0f %.1f %d %02d%09.6f %c %02d%09.6f %c %.0f  %d %d %d %d %d %d %d %d",
		day, month, year, hour, minute, second, mode, bias, biasunc, rate, rateunc, utcoff,
		lat_deg, lat_min, north_south, lon_deg, lon_min, east_west, alt,
		sv[0], sv[1], sv[2], sv[3], sv[4], sv[5], sv[6], sv[7]);

   	puts(logbuf);

	return (0);
}

int initstress() {
	TSIPPKT *rpt;
	char buf[128];
	long l;
	float f;
	double d;

	/*
	 * Initialize the Clock
	 */
	ncc_send(pfd, 0x1F, "", 0); 		/* query software versions */
	ncc_send(pfd, 0x26, "", 0); 		/* query receiver health */
	ncc_send(pfd, 0x8E, "\x42", 1);	/* query serial numbers */
	ncc_send(pfd, 0x8E, "\x41", 1);	/* query manuf params */

	/* i/o opts */
	ncc_send(pfd, 0x35, "\x1C\x00\x05\x0A", 4);

	/* turn off port A */
	ncc_send(pfd, 0x3d, "\x0B\x0B\x07\x07\x01\x00", 6);

	/* set PPS characteristics */
	buf[0] = 0x4A;
	buf[1] = 0x01;		/* PPS on */
	buf[2] = 0x01;		/* timebase UTC */
	buf[3] = 0x00;		/* polarity positive */
	d = 0.;				/* 100 ft cable  (should be flag!) */
	bPutDouble(&d, &buf[4]); 
	f = 1e-6 * GPS_C;	/* turn off PPS if biasunc > this (1us) (flag!)*/
	bPutFloat(&f, &buf[12]);	
	ncc_send(pfd, 0x8E, buf, 16);

	/* all outomatic packet output off */
	buf[0] = 0x4D;
	l = 0x00000000;
	bPutLong(&l, &buf[1]);	
	ncc_send(pfd, 0x8E, buf , 5);

	ncc_send(pfd, 0xBB, "\x00", 1);	/* query primary configuration */
	ncc_send(pfd, 0x8E, "\x4A", 1);	/* query PPS params */
	ncc_send(pfd, 0x8E, "\x4B", 1);	/* query survey limit */
}


int parsestress() {
	TSIPPKT *rpt;

	rpt = getrpt(0,0);	

	switch (rpt->code) {

		case 0x8F:

			switch (rpt->buf[0]) {

			case 0xAD:	/* UTC Time */

				if (parse0x8FAD(rpt) < 0) {
					printf("parse0x8FAD() failed\n");
					break;
				}

				ncc_send(pfd, 0x8E, "\x0B", 1); /* poll for next packet */
				break;
			
			case 0x0B: /* comprehensive time packet */
				parse0x8F0B(rpt);
				break;

			default: /* other superpackets */
				parseany(rpt);
				break;
			}
			break;

		case 0x4F:	/* UTC parameters, for leap info */
			parse0x4F(rpt);
			break;

		case 0x5C:	/* sat tracking data */
			parse0x5C(rpt);
			break;

		default: /* other packets */
			parseany(rpt);
			break;
	}
   	rpt->status = TSIP_PARSED_EMPTY;
	fflush(stdout);
}


int parseany(rpt)	
	TSIPPKT *rpt;
{
   	static char logbuf[1024];	/* logging string buffer */

   	TranslateTSIPReportToText (rpt, logbuf);	/* anything else */
   	puts(&logbuf[1]);
	return(0);
}


/******* Code below is from Trimble Tsipchat *************/

/*
 * *************************************************************************
 *
 * Trimble Navigation, Ltd.
 * OEM Products Development Group
 * P.O. Box 3642
 * 645 North Mary Avenue
 * Sunnyvale, California 94088-3642
 *
 * Corporate Headquarter:
 *    Telephone:  (408) 481-8000
 *    Fax:        (408) 481-6005
 *
 * Technical Support Center:
 *    Telephone:  (800) 767-4822	(U.S. and Canada)
 *                (408) 481-6940    (outside U.S. and Canada)
 *    Fax:        (408) 481-6020
 *    BBS:        (408) 481-7800
 *    e-mail:     trimble_support@trimble.com
 *		ftp://ftp.trimble.com/pub/sct/embedded/bin
 *
 * *************************************************************************
 *
 * -------  BYTE-SWAPPING  -------
 * TSIP is big-endian (Motorola) protocol.  To use on little-endian (Intel)
 * systems, the bytes of all multi-byte types (shorts, floats, doubles, etc.)
 * must be reversed.  This is controlled by the MACRO BYTESWAP; if defined, it
 * assumes little-endian protocol.
 * --------------------------------
 *
 * T_PARSER.C and T_PARSER.H contains primitive functions that interpret
 * reports received from the receiver.  A second source file pair,
 * T_FORMAT.C and T_FORMAT.H, contin the matching TSIP command formatters.
 *
 * The module is in very portable, basic C language.  It can be used as is, or
 * with minimal changes if a TSIP communications application is needed separate
 * from TSIPCHAT. The construction of most argument lists avoid the use of
 * structures, but the developer is encouraged to reconstruct them using such
 * definitions to meet project requirements.  Declarations of T_PARSER.C
 * functions are included in T_PARSER.H to provide prototyping definitions.
 *
 * There are two types of functions: a serial input processing routine,
 *                            tsip_input_proc()
 * which assembles incoming bytes into a TSIPPKT structure, and the
 * report parsers, rpt_0x??().
 *
 * 1) The function tsip_input_proc() accumulates bytes from the receiver,
 * strips control bytes (DLE), and checks if the report end sequence (DLE ETX)
 * has been received.  rpt.status is defined as TSIP_PARSED_FULL (== 1)
 * if a complete packet is available.
 *
 * 2) The functions rpt_0x??() are report string interpreters patterned after
 * the document called "Trimble Standard Interface Protocol".  It should be
 * noted that if the report buffer is sent into the receiver with the wrong
 * length (byte count), the rpt_0x??() returns the Boolean equivalence for
 * TRUE.
 *
 * *************************************************************************
 *
 */


/**/
void tsip_input_proc (
	TSIPPKT *rpt,
	int inbyte)
/* reads bytes until serial buffer is empty or a complete report
 * has been received; end of report is signified by DLE ETX.
 */
{
	unsigned char newbyte;

	if (inbyte < 0 || inbyte > 0xFF) return;

	newbyte = (unsigned char)(inbyte);
	switch (rpt->status)
	{
	case TSIP_PARSED_DLE_1:
		switch (newbyte)
		{
		case 0:
		case ETX:
      	/* illegal TSIP IDs */
         rpt->len = 0;
			rpt->status = TSIP_PARSED_EMPTY;
			break;
		case DLE:
      	/* try normal message start again */
			rpt->len = 0;
			rpt->status = TSIP_PARSED_DLE_1;
			break;
		default:
      	/* legal TSIP ID; start message */
			rpt->code = newbyte;
         rpt->len = 0;
			rpt->status = TSIP_PARSED_DATA;
			break;
		}
		break;
	case TSIP_PARSED_DATA:
		switch (newbyte) {
		case DLE:
      	/* expect DLE or ETX next */
			rpt->status = TSIP_PARSED_DLE_2;
			break;
		default:
      	/* normal data byte  */
			rpt->buf[rpt->len] = newbyte;
			rpt->len++;
         /* no change in rpt->status */
			break;
		}
		break;
	case TSIP_PARSED_DLE_2:
		switch (newbyte) {
		case DLE:
      	/* normal data byte */
			rpt->buf[rpt->len] = newbyte;
			rpt->len++;
			rpt->status = TSIP_PARSED_DATA;
			break;
		case ETX:
			/* end of message; return TRUE here. */
			rpt->status = TSIP_PARSED_FULL;
			break;
		default:
			/* error: treat as TSIP_PARSED_DLE_1; start new report packet */
			rpt->code = newbyte;
         rpt->len = 0;
			rpt->status = TSIP_PARSED_DATA;
		}
		break;
	case TSIP_PARSED_FULL:
	case TSIP_PARSED_EMPTY:
	default:
		switch (newbyte) {
		case DLE:
      	/* normal message start */
			rpt->len = 0;
			rpt->status = TSIP_PARSED_DLE_1;
			break;
		default:
			/* error: ignore newbyte */
			rpt->len = 0;
			rpt->status = TSIP_PARSED_EMPTY;
		}
		break;
	}
	if (rpt->len > MAX_RPTBUF) {
		/* error: start new report packet */
		rpt->status = TSIP_PARSED_EMPTY;
		rpt->len = 0;
	}
}

/* for DOS machines, reverse order of bytes as they come through the
 * serial port. */
#ifdef BYTESWAP
static short bGetShort (unsigned char *bp)
{
	short outval;
   unsigned char *optr;

   optr = (unsigned char*)&outval + 1;
   *optr-- = *bp++;
   *optr = *bp;
	return outval;
}

static unsigned short bGetUShort (unsigned char *bp)
{
	unsigned short outval;
   unsigned char *optr;

   optr = (unsigned char*)&outval + 1;
   *optr-- = *bp++;
   *optr = *bp;
	return outval;
}

static long bGetLong (unsigned char *bp)
{
	long outval;
   unsigned char *optr;

   optr = (unsigned char*)&outval + 3;
   *optr-- = *bp++;
   *optr-- = *bp++;
   *optr-- = *bp++;
   *optr = *bp;
	return outval;
}

static unsigned long bGetULong (unsigned char *bp)
{
	unsigned long outval;
   unsigned char *optr;

   optr = (unsigned char*)&outval + 3;
   *optr-- = *bp++;
   *optr-- = *bp++;
   *optr-- = *bp++;
   *optr = *bp;
	return outval;
}

static float bGetSingle (unsigned char *bp)
{
	float outval;
   unsigned char *optr;

   optr = (unsigned char*)&outval + 3;
   *optr-- = *bp++;
   *optr-- = *bp++;
   *optr-- = *bp++;
   *optr = *bp;
	return outval;
}

static double bGetDouble (unsigned char *bp)
{
	double outval;
   unsigned char *optr;

   optr = (unsigned char*)&outval + 7;
   *optr-- = *bp++;
   *optr-- = *bp++;
   *optr-- = *bp++;
   *optr-- = *bp++;
   *optr-- = *bp++;
   *optr-- = *bp++;
   *optr-- = *bp++;
   *optr = *bp;
	return outval;
}

#else	/* not BYTESWAP */

#define bGetShort(bp) 	(*(short*)(bp))
#define bGetLong(bp) 	(*(long*)(bp))
#define bGetULong(bp) 	(*(unsigned long*)(bp))
#define bGetSingle(bp) 	(*(float*)(bp))
#define bGetDouble(bp)	(*(double*)(bp))

#endif


/**/
short rpt_0x3D (TSIPPKT *rpt,
	unsigned char *tx_baud_index,
	unsigned char *rx_baud_index,
	unsigned char *char_format_index,
	unsigned char *stop_bits,
	unsigned char *tx_mode_index,
	unsigned char *rx_mode_index)
/* Channel A configuration for dual port operation */
{
	unsigned char *buf;
	buf = rpt->buf;

	if (rpt->len != 6) return TRUE;
	*tx_baud_index = buf[0];
	*rx_baud_index = buf[1];
	*char_format_index = buf[2];
	*stop_bits = (unsigned char)((buf[3] == 0x07) ? 1 : 2);
	*tx_mode_index = buf[4];
	*rx_mode_index = buf[5];
	return FALSE;
}

/**/
short rpt_0x40 (TSIPPKT *rpt,
	unsigned char *sv_prn,
	short *week_num,
	float *t_zc,
	float *eccentricity,
	float *t_oa,
	float *i_0,
	float *OMEGA_dot,
	float *sqrt_A,
	float *OMEGA_0,
	float *omega,
	float *M_0)
/* almanac data for specified satellite */
{
	unsigned char *buf;
	buf = rpt->buf;

	if (rpt->len != 39) return TRUE;
	*sv_prn = buf[0];
	*t_zc = bGetSingle (&buf[1]);
	*week_num = bGetShort (&buf[5]);
	*eccentricity = bGetSingle (&buf[7]);
	*t_oa = bGetSingle (&buf[11]);
	*i_0 = bGetSingle (&buf[15]);
	*OMEGA_dot = bGetSingle (&buf[19]);
	*sqrt_A = bGetSingle (&buf[23]);
	*OMEGA_0 = bGetSingle (&buf[27]);
	*omega = bGetSingle (&buf[31]);
	*M_0 = bGetSingle (&buf[35]);
	return FALSE;
}

short rpt_0x41 (TSIPPKT *rpt,
	float *time_of_week,
	float *UTC_offset,
	short *week_num)
/* GPS time */
{
	unsigned char *buf;
	buf = rpt->buf;
	
	if (rpt->len != 10) return TRUE;
	*time_of_week = bGetSingle (buf);
	*week_num = bGetShort (&buf[4]);
	*UTC_offset = bGetSingle (&buf[6]);
	return FALSE;
}

short rpt_0x42 (TSIPPKT *rpt,
	float pos_ECEF[3],
	float *time_of_fix)
/* position in ECEF, single precision */
{
	unsigned char *buf;
	buf = rpt->buf;
	
	if (rpt->len != 16) return TRUE;
	pos_ECEF[0] = bGetSingle (buf);
	pos_ECEF[1]= bGetSingle (&buf[4]);
	pos_ECEF[2]= bGetSingle (&buf[8]);
	*time_of_fix = bGetSingle (&buf[12]);
	return FALSE;
}

short rpt_0x43 (TSIPPKT *rpt,
	float ECEF_vel[3],
	float *freq_offset,
	float *time_of_fix)
/* velocity in ECEF, single precision */
{
	unsigned char *buf;
	buf = rpt->buf;

	if (rpt->len != 20) return TRUE;
	ECEF_vel[0] = bGetSingle (buf);
	ECEF_vel[1] = bGetSingle (&buf[4]);
	ECEF_vel[2] = bGetSingle (&buf[8]);
	*freq_offset = bGetSingle (&buf[12]);
	*time_of_fix = bGetSingle (&buf[16]);
	return FALSE;
}

short rpt_0x45 (TSIPPKT *rpt,
	unsigned char *major_nav_version,
	unsigned char *minor_nav_version,
	unsigned char *nav_day,
	unsigned char *nav_month,
	unsigned char *nav_year,
	unsigned char *major_dsp_version,
	unsigned char *minor_dsp_version,
	unsigned char *dsp_day,
	unsigned char *dsp_month,
	unsigned char *dsp_year)
/* software versions */	
{
	unsigned char *buf;
	buf = rpt->buf;

	if (rpt->len != 10) return TRUE;
	*major_nav_version = buf[0];
	*minor_nav_version = buf[1];
	*nav_day = buf[2];
	*nav_month = buf[3];
	*nav_year = buf[4];
	*major_dsp_version = buf[5];
	*minor_dsp_version = buf[6];
	*dsp_day = buf[7];
	*dsp_month = buf[8];
	*dsp_year = buf[9];
	return FALSE;
}

short rpt_0x46 (TSIPPKT *rpt,
	unsigned char *status1,
	unsigned char *status2)
/* receiver health and status */
{
	unsigned char *buf;
	buf = rpt->buf;

	if (rpt->len != 2) return TRUE;
	*status1 = buf[0];
	*status2 = buf[1];
	return FALSE;
}

short rpt_0x47 (TSIPPKT *rpt,
	unsigned char *nsvs, unsigned char *sv_prn,
	float *snr)
/* signal levels for all satellites tracked */
{
	short isv;
	unsigned char *buf;
	buf = rpt->buf;

	if (rpt->len != 1 + 5*buf[0]) return TRUE;
	*nsvs = buf[0];
	for (isv = 0; isv < (*nsvs); isv++) {
		sv_prn[isv] = buf[5*isv + 1];
		snr[isv] = bGetSingle (&buf[5*isv + 2]);
	}
	return FALSE;
}

short rpt_0x48 (TSIPPKT *rpt,
	unsigned char *message)
/* GPS system message */
{
	unsigned char *buf;
	buf = rpt->buf;

	if (rpt->len != 22) return TRUE;
	memcpy (message, buf, 22);
	message[22] = 0;
	return FALSE;
}

short rpt_0x49 (TSIPPKT *rpt,
	unsigned char *sv_health)
/* health for all satellites from almanac health page */
{
	short i;
	unsigned char *buf;
	buf = rpt->buf;

	if (rpt->len != 32) return TRUE;
	for (i = 0; i < 32; i++) sv_health [i]= buf[i];
	return FALSE;
}

short rpt_0x4A (TSIPPKT *rpt,
	float *lat,
	float *lon,
	float *alt,
	float *clock_bias,
	float *time_of_fix)
/* position in lat-lon-alt, single precision */
{
	unsigned char *buf;
	buf = rpt->buf;

	if (rpt->len != 20) return TRUE;
	*lat = bGetSingle (buf);
	*lon = bGetSingle (&buf[4]);
	*alt = bGetSingle (&buf[8]);
	*clock_bias = bGetSingle (&buf[12]);
	*time_of_fix = bGetSingle (&buf[16]);
	return FALSE;
}

short rpt_0x4A_2 (TSIPPKT *rpt,
	float *alt, float *dummy , unsigned char *alt_flag)
/* reference altitude parameters */
{
	unsigned char *buf;

	buf = rpt->buf;

	if (rpt->len != 9) return TRUE;
	*alt = bGetSingle (buf);
	*dummy = bGetSingle (&buf[4]);
	*alt_flag = buf[8];
	return FALSE;
}

short rpt_0x4B (TSIPPKT *rpt,
	unsigned char *machine_id,
	unsigned char *status3,
	unsigned char *status4)
/* machine ID code, status */
{
	unsigned char *buf;
	buf = rpt->buf;

	if (rpt->len != 3) return TRUE;
	*machine_id = buf[0];
	*status3 = buf[1];
	*status4 = buf[2];
	return FALSE;
}

short rpt_0x4C (TSIPPKT *rpt,
	unsigned char *dyn_code,
	float *el_mask,
	float *snr_mask,
	float *dop_mask,
	float *dop_switch)
/* operating parameters and masks */
{
	unsigned char *buf;
	buf = rpt->buf;

	if (rpt->len != 17) return TRUE;
	*dyn_code = buf[0];
	*el_mask = bGetSingle (&buf[1]);
	*snr_mask = bGetSingle (&buf[5]);
	*dop_mask = bGetSingle (&buf[9]);
	*dop_switch = bGetSingle (&buf[13]);
	return FALSE;
}

short rpt_0x4D (TSIPPKT *rpt,
	float *osc_offset)
/* oscillator offset */
{
	unsigned char *buf;
	buf = rpt->buf;

	if (rpt->len != 4) return TRUE;
	*osc_offset = bGetSingle (buf);
	return FALSE;
}

short rpt_0x4E (TSIPPKT *rpt,
	unsigned char *response)
/* yes/no response to command to set GPS time */
{
	unsigned char *buf;
	buf = rpt->buf;

	if (rpt->len != 1) return TRUE;
	*response = buf[0];
	return FALSE;
}

short rpt_0x4F (TSIPPKT *rpt,
	double *a0,
	float *a1,
	float *time_of_data,
	short *dt_ls,
	short *wn_t,
	short *wn_lsf,
	short *dn,
	short *dt_lsf)
/* UTC data */
{
	unsigned char *buf;
	buf = rpt->buf;

	if (rpt->len != 26) return TRUE;
	*a0 = bGetDouble (buf);
	*a1 = bGetSingle (&buf[8]);
	*dt_ls = bGetShort (&buf[12]);
	*time_of_data = bGetSingle (&buf[14]);
	*wn_t = bGetShort (&buf[18]);
	*wn_lsf = bGetShort (&buf[20]);
	*dn = bGetShort (&buf[22]);
	*dt_lsf = bGetShort (&buf[24]);
	return FALSE;
}

/**/
short rpt_0x54 (TSIPPKT *rpt,
	float *clock_bias,
   float *freq_offset,
   float *time_of_fix)
/* clock offset and frequency offset in 1-SV (0-D) mode */
{
	unsigned char *buf;
	buf = rpt->buf;

	if (rpt->len != 12) return TRUE;
	*clock_bias = bGetSingle (buf);
	*freq_offset = bGetSingle (&buf[4]);
	*time_of_fix = bGetSingle (&buf[8]);
	return FALSE;
}

short rpt_0x55 (TSIPPKT *rpt,
	unsigned char *pos_code,
	unsigned char *vel_code,
	unsigned char *time_code,
	unsigned char *aux_code)
/* I/O serial options */
{
	unsigned char *buf;
	buf = rpt->buf;
	
	if (rpt->len != 4) return TRUE;
	*pos_code = buf[0];
	*vel_code = buf[1];
	*time_code = buf[2];
	*aux_code = buf[3];
	return FALSE;
}

short rpt_0x56 (TSIPPKT *rpt,
	float vel_ENU[3], float *freq_offset, float *time_of_fix)
/* velocity in east-north-up coordinates */	
{
	unsigned char *buf;
	buf = rpt->buf;
	
	if (rpt->len != 20) return TRUE;
	/* east */
	vel_ENU[0] = bGetSingle (buf);
	/* north */
	vel_ENU[1] = bGetSingle (&buf[4]);
	/* up */
	vel_ENU[2] = bGetSingle (&buf[8]);
	*freq_offset = bGetSingle (&buf[12]);
	*time_of_fix = bGetSingle (&buf[16]);
	return FALSE;
}

short rpt_0x57 (TSIPPKT *rpt,
	unsigned char *source_code, unsigned char *diag_code,
	short *week_num,
	float *time_of_fix)
/* info about last computed fix */
{
	unsigned char *buf;
	buf = rpt->buf;
	
	if (rpt->len != 8) return TRUE;
	*source_code = buf[0];
	*diag_code = buf[1];
	*time_of_fix = bGetSingle (&buf[2]);
	*week_num = bGetShort (&buf[6]);
	return FALSE;
}

short rpt_0x58 (TSIPPKT *rpt,
	unsigned char *op_code, unsigned char *data_type, unsigned char *sv_prn,
	unsigned char *data_length, unsigned char *data_packet)
/* GPS system data or acknowledgment of GPS system data load */
{
	unsigned char *buf, *buf4;
	short dl;
	ALM_INFO* alminfo;
	ION_INFO* ioninfo;
	UTC_INFO* utcinfo;
	NAV_INFO* navinfo;

	buf = rpt->buf;

	if (buf[0] == 2) {
		if (rpt->len < 4) return TRUE;
		if (rpt->len != 4+buf[3]) return TRUE;
	}
	else if (rpt->len != 3) {
		return TRUE;
	}
	*op_code = buf[0];
	*data_type = buf[1];
	*sv_prn = buf[2];
	if (*op_code == 2) {
		dl = buf[3];
		*data_length = (unsigned char)dl;
		buf4 = &buf[4];
		switch (*data_type) {
		case 2:
			/* Almanac */
			if (*data_length != sizeof (ALM_INFO)) return TRUE;
			alminfo = (ALM_INFO*)data_packet;
			alminfo->t_oa_raw  = buf4[0];
			alminfo->SV_health = buf4[1];
			alminfo->e         = bGetSingle(&buf4[2]);
			alminfo->t_oa      = bGetSingle(&buf4[6]);
			alminfo->i_0       = bGetSingle(&buf4[10]);
			alminfo->OMEGADOT  = bGetSingle(&buf4[14]);
			alminfo->sqrt_A    = bGetSingle(&buf4[18]);
			alminfo->OMEGA_0   = bGetSingle(&buf4[22]);
			alminfo->omega     = bGetSingle(&buf4[26]);
			alminfo->M_0       = bGetSingle(&buf4[30]);
			alminfo->a_f0      = bGetSingle(&buf4[34]);
			alminfo->a_f1      = bGetSingle(&buf4[38]);
			alminfo->Axis      = bGetSingle(&buf4[42]);
			alminfo->n         = bGetSingle(&buf4[46]);
			alminfo->OMEGA_n   = bGetSingle(&buf4[50]);
			alminfo->ODOT_n    = bGetSingle(&buf4[54]);
			alminfo->t_zc      = bGetSingle(&buf4[58]);
			alminfo->weeknum   = bGetShort(&buf4[62]);
			alminfo->wn_oa     = bGetShort(&buf4[64]);
			break;

		case 3:
			/* Almanac health page */
			if (*data_length != sizeof (ALH_PARMS) + 3) return TRUE;

			/* this record is returned raw */
			memcpy (data_packet, buf4, dl);
			break;

		case 4:
			/* Ionosphere */
			if (*data_length != sizeof (ION_INFO) + 8) return TRUE;
			ioninfo = (ION_INFO*)data_packet;
			ioninfo->alpha_0   = bGetSingle (&buf4[8]);
			ioninfo->alpha_1   = bGetSingle (&buf4[12]);
			ioninfo->alpha_2   = bGetSingle (&buf4[16]);
			ioninfo->alpha_3   = bGetSingle (&buf4[20]);
			ioninfo->beta_0    = bGetSingle (&buf4[24]);
			ioninfo->beta_1    = bGetSingle (&buf4[28]);
			ioninfo->beta_2    = bGetSingle (&buf4[32]);
			ioninfo->beta_3    = bGetSingle (&buf4[36]);
			break;

		case 5:
			/* UTC */
			if (*data_length != sizeof (UTC_INFO) + 13) return TRUE;
			utcinfo = (UTC_INFO*)data_packet;
			utcinfo->A_0       = bGetDouble (&buf4[13]);
			utcinfo->A_1       = bGetSingle (&buf4[21]);
			utcinfo->delta_t_LS = bGetShort (&buf4[25]);
			utcinfo->t_ot      = bGetSingle(&buf4[27]);
			utcinfo->WN_t      = bGetShort (&buf4[31]);
			utcinfo->WN_LSF    = bGetShort (&buf4[33]);
			utcinfo->DN        = bGetShort (&buf4[35]);
			utcinfo->delta_t_LSF = bGetShort (&buf4[37]);
			break;

		case 6:
			/* Ephemeris */
			if (*data_length != sizeof (NAV_INFO) - 1) return TRUE;

			navinfo = (NAV_INFO*)data_packet;

			navinfo->sv_number = buf4[0];
			navinfo->t_ephem = bGetSingle (&buf4[1]);
			navinfo->ephclk.weeknum = bGetShort (&buf4[5]);

			navinfo->ephclk.codeL2 = buf4[7];
			navinfo->ephclk.L2Pdata = buf4[8];
			navinfo->ephclk.SVacc_raw = buf4[9];
			navinfo->ephclk.SV_health = buf4[10];
			navinfo->ephclk.IODC = bGetShort (&buf4[11]);
			navinfo->ephclk.T_GD = bGetSingle (&buf4[13]);
			navinfo->ephclk.t_oc = bGetSingle (&buf4[17]);
			navinfo->ephclk.a_f2 = bGetSingle (&buf4[21]);
			navinfo->ephclk.a_f1 = bGetSingle (&buf4[25]);
			navinfo->ephclk.a_f0 = bGetSingle (&buf4[29]);
			navinfo->ephclk.SVacc = bGetSingle (&buf4[33]);

			navinfo->ephorb.IODE = buf4[37];
			navinfo->ephorb.fit_interval = buf4[38];
			navinfo->ephorb.C_rs = bGetSingle (&buf4[39]);
			navinfo->ephorb.delta_n = bGetSingle (&buf4[43]);
			navinfo->ephorb.M_0 = bGetDouble (&buf4[47]);
			navinfo->ephorb.C_uc = bGetSingle (&buf4[55]);
			navinfo->ephorb.e = bGetDouble (&buf4[59]);
			navinfo->ephorb.C_us = bGetSingle (&buf4[67]);
			navinfo->ephorb.sqrt_A = bGetDouble (&buf4[71]);
			navinfo->ephorb.t_oe = bGetSingle (&buf4[79]);
			navinfo->ephorb.C_ic = bGetSingle (&buf4[83]);
			navinfo->ephorb.OMEGA_0 = bGetDouble (&buf4[87]);
			navinfo->ephorb.C_is = bGetSingle (&buf4[95]);
			navinfo->ephorb.i_0 = bGetDouble (&buf4[99]);
			navinfo->ephorb.C_rc = bGetSingle (&buf4[107]);
			navinfo->ephorb.omega = bGetDouble (&buf4[111]);
			navinfo->ephorb.OMEGADOT=bGetSingle (&buf4[119]);
			navinfo->ephorb.IDOT = bGetSingle (&buf4[123]);
			navinfo->ephorb.Axis = bGetDouble (&buf4[127]);
			navinfo->ephorb.n = bGetDouble (&buf4[135]);
			navinfo->ephorb.r1me2 = bGetDouble (&buf4[143]);
			navinfo->ephorb.OMEGA_n=bGetDouble (&buf4[151]);
			navinfo->ephorb.ODOT_n = bGetDouble (&buf4[159]);
			break;
		}
	}
	return FALSE;
}

short rpt_0x59 (TSIPPKT *rpt,
	unsigned char *code_type,
	unsigned char status_code[32])
/* satellite enable/disable or health heed/ignore list */	
{
	short iprn;
	unsigned char *buf;
	buf = rpt->buf;
	
	if (rpt->len != 33) return TRUE;
	*code_type = buf[0];
	for (iprn = 0; iprn < 32; iprn++)
		status_code[iprn] = buf[iprn + 1];
	return FALSE;
}

short rpt_0x5A (TSIPPKT *rpt,
	unsigned char *sv_prn,
	float *sample_length,
	float *signal_level,
	float *code_phase,
	float *Doppler,
	double *time_of_fix)
/* raw measurement data - code phase/Doppler */
{
	unsigned char *buf;
	buf = rpt->buf;

	if (rpt->len != 25) return TRUE;
	*sv_prn = buf[0];
	*sample_length = bGetSingle (&buf[1]);
	*signal_level = bGetSingle (&buf[5]);
	*code_phase = bGetSingle (&buf[9]);
	*Doppler = bGetSingle (&buf[13]);
	*time_of_fix = bGetDouble (&buf[17]);
	return FALSE;
}

short rpt_0x5B (TSIPPKT *rpt,
	unsigned char *sv_prn,
	unsigned char *sv_health,
	unsigned char *sv_iode,
	unsigned char *fit_interval_flag,
	float *time_of_collection,
	float *time_of_eph,
	float *sv_accy)
/* satellite ephorb status */	
{
	unsigned char *buf;
	buf = rpt->buf;
	
	if (rpt->len != 16) return TRUE;
	*sv_prn = buf[0];
	*time_of_collection = bGetSingle (&buf[1]);
	*sv_health = buf[5];
	*sv_iode = buf[6];
	*time_of_eph = bGetSingle (&buf[7]);
	*fit_interval_flag = buf[11];
	*sv_accy = bGetSingle (&buf[12]);
	return FALSE;
}

short rpt_0x5C (TSIPPKT *rpt,
	unsigned char *sv_prn,
	unsigned char *slot,
	unsigned char *chan,
	unsigned char *acq_flag,
	unsigned char *eph_flag,
	float *signal_level,
	float *time_of_last_msmt,
	float *elev,
	float *azim,
	unsigned char *old_msmt_flag,
	unsigned char *integer_msec_flag,
	unsigned char *bad_data_flag,
	unsigned char *data_collect_flag)
/* satellite tracking status */
{
	unsigned char *buf;
	buf = rpt->buf;
	
	if (rpt->len != 24) return TRUE;
	*sv_prn = buf[0];
	*slot = (unsigned char)((buf[1] & 0x07) + 1);
	*chan = (unsigned char)(buf[1] >> 3);
	if (*chan == 0x10) *chan = 2;
	else (*chan)++;
	*acq_flag = buf[2];
	*eph_flag = buf[3];
	*signal_level = bGetSingle (&buf[4]);
	*time_of_last_msmt = bGetSingle (&buf[8]);
	*elev = bGetSingle (&buf[12]);
	*azim = bGetSingle (&buf[16]);
	*old_msmt_flag = buf[20];
	*integer_msec_flag = buf[21];
	*bad_data_flag = buf[22];
	*data_collect_flag = buf[23];
	return FALSE;
}

/**/
short rpt_0x6D (TSIPPKT *rpt,
	unsigned char *manual_mode,
	unsigned char *nsvs,
	unsigned char *ndim,
	unsigned char sv_prn[],
	float *pdop,
	float *hdop,
	float *vdop,
	float *tdop)
/* over-determined satellite selection for position fixes, PDOP, fix mode */
{
	short islot;
	unsigned char *buf;
	buf = rpt->buf;

	*nsvs = (unsigned char)((buf[0] & 0xF0) >> 4);
	if ((*nsvs)>8) return TRUE;
	if (rpt->len != 17 + (*nsvs) ) return TRUE;

	*manual_mode = (unsigned char)(buf[0] & 0x08);
	*ndim  = (unsigned char)((buf[0] & 0x07));
	*pdop = bGetSingle (&buf[1]);
	*hdop = bGetSingle (&buf[5]);
	*vdop = bGetSingle (&buf[9]);
	*tdop = bGetSingle (&buf[13]);
	for (islot = 0; islot < (*nsvs); islot++)
		sv_prn[islot] = buf[islot + 17];
	return FALSE;
}

/**/
short rpt_0x82 (TSIPPKT *rpt,
	unsigned char *diff_mode)
/* differential fix mode */
{
	unsigned char *buf;
	buf = rpt->buf;

	if (rpt->len != 1) return TRUE;
	*diff_mode = buf[0];
	return FALSE;
}

short rpt_0x83 (TSIPPKT *rpt,
	double ECEF_pos[3],
	double *clock_bias,
	float *time_of_fix)
/* position, ECEF double precision */
{
	unsigned char *buf;
	buf = rpt->buf;

	if (rpt->len != 36) return TRUE;
	ECEF_pos[0] = bGetDouble (buf);
	ECEF_pos[1] = bGetDouble (&buf[8]);
	ECEF_pos[2] = bGetDouble (&buf[16]);
	*clock_bias  = bGetDouble (&buf[24]);
	*time_of_fix = bGetSingle (&buf[32]);
	return FALSE;
}

short rpt_0x84 (TSIPPKT *rpt,
	double *lat,
	double *lon,
	double *alt,
	double *clock_bias,
	float *time_of_fix)
/* position, lat-lon-alt double precision */	
{
	unsigned char *buf;
	buf = rpt->buf;

	if (rpt->len != 36) return TRUE;
	*lat = bGetDouble (buf);
	*lon = bGetDouble (&buf[8]);
	*alt = bGetDouble (&buf[16]);
	*clock_bias = bGetDouble (&buf[24]);
	*time_of_fix = bGetSingle (&buf[32]);
	return FALSE;
}

short rpt_Paly0xBB(TSIPPKT *rpt,
	TSIP_RCVR_CFG *TsipxBB)
{

	unsigned char *buf;
	buf = rpt->buf;

	/* Palisade is inconsistent with other TSIP, which has a kength of 40 */
	/* if (rpt->len != 40) return TRUE; */
	if (rpt->len != 43) return TRUE;

	TsipxBB->bSubcode		=	buf[0];
	TsipxBB->operating_mode	=	buf[1]	;
	TsipxBB->dyn_code			=	buf[3]	;
	TsipxBB->elev_mask		=  bGetSingle (&buf[5]);
	TsipxBB->cno_mask			=	bGetSingle (&buf[9]);
	TsipxBB->dop_mask 		=  bGetSingle (&buf[13]);
	TsipxBB->dop_switch 	=  bGetSingle (&buf[17]);
	return FALSE;
}

short rpt_0xBC (TSIPPKT *rpt,
	unsigned char *port_num,
   unsigned char *in_baud,
	unsigned char *out_baud,
   unsigned char *data_bits,
	unsigned char *parity,
   unsigned char *stop_bits,
   unsigned char *flow_control,
	unsigned char *protocols_in,
   unsigned char *protocols_out,
   unsigned char *reserved)
/* Receiver serial port configuration */
{
	unsigned char *buf;
	buf = rpt->buf;

	if (rpt->len != 10) return TRUE;
	*port_num = buf[0];
	*in_baud = buf[1];
	*out_baud = buf[2];
	*data_bits = buf[3];
	*parity = buf[4];
	*stop_bits = buf[5];
	*flow_control = buf[6];
	*protocols_in = buf[7];
	*protocols_out = buf[8];
	*reserved = buf[9];

	return FALSE;
}

/**** Superpackets ****/

short rpt_0x8F0B(TSIPPKT *rpt,
                 unsigned short *event,
                 double *tow,
                 unsigned char *date,
                 unsigned char *month,
                 short *year,
                 unsigned char *dim_mode,
                 short *utc_offset,
                 double *bias,
                 double *drift,
                 float *bias_unc,
                 float *dr_unc,
                 double *lat,
                 double *lon,
                 double *alt,
                 char sv_id[8])
{
       short
		index;
       unsigned char
		*buf;

	buf = rpt->buf;
       if (rpt->len != 74) return TRUE;
       *event = bGetShort(&buf[1]);
       *tow = bGetDouble(&buf[3]);
       *date = buf[11];
       *month = buf[12];
       *year = bGetShort(&buf[13]);
       *dim_mode = buf[15];
       *utc_offset = bGetShort(&buf[16]);
       *bias = bGetDouble(&buf[18]);
       *drift = bGetDouble(&buf[26]);
       *bias_unc = bGetSingle(&buf[34]);
       *dr_unc = bGetSingle(&buf[38]);
       *lat = bGetDouble(&buf[42]);
       *lon = bGetDouble(&buf[50]);
       *alt = bGetDouble(&buf[58]);

       for (index=0; index<8; index++) sv_id[index] = buf[index + 66];
       return FALSE;
}

short rpt_0x8F14 (TSIPPKT *rpt,
	short *datum_idx,
	double datum_coeffs[5])
/*  datum index and coefficients  */
{
	unsigned char *buf;
	buf = rpt->buf;

	if (rpt->len != 43) return TRUE;
	*datum_idx = bGetShort(&buf[1]);
	datum_coeffs[0] = bGetDouble (&buf[3]);
	datum_coeffs[1] = bGetDouble (&buf[11]);
	datum_coeffs[2] = bGetDouble (&buf[19]);
	datum_coeffs[3] = bGetDouble (&buf[27]);
	datum_coeffs[4] = bGetDouble (&buf[35]);
	return FALSE;
}


short rpt_0x8F15 (TSIPPKT *rpt,
	short *datum_idx,
	double datum_coeffs[5])
/*  datum index and coefficients  */
{
	unsigned char *buf;
	buf = rpt->buf;

	if (rpt->len != 43) return TRUE;
	*datum_idx = bGetShort(&buf[1]);
	datum_coeffs[0] = bGetDouble (&buf[3]);
	datum_coeffs[1] = bGetDouble (&buf[11]);
	datum_coeffs[2] = bGetDouble (&buf[19]);
	datum_coeffs[3] = bGetDouble (&buf[27]);
	datum_coeffs[4] = bGetDouble (&buf[35]);
	return FALSE;
}


#define MAX_LONG  (2147483648.)   /* 2**31 */

short rpt_0x8F20 (TSIPPKT *rpt,
	unsigned char *info,
	double *lat,
	double *lon,
	double *alt,
	double vel_enu[],
	double *time_of_fix,
	short *week_num,
	unsigned char *nsvs,
	unsigned char sv_prn[],
	short sv_IODC[],
	short *datum_index)
{
	short
		isv;
	unsigned char
		*buf, prnx, iode;
	unsigned long
		ulongtemp;
	long
		longtemp;
	double
		vel_scale;

	buf = rpt->buf;

	if (rpt->len != 56) return TRUE;

	vel_scale = (buf[24]&1)? 0.020 : 0.005;
	vel_enu[0] = bGetShort (buf+2)*vel_scale;
	vel_enu[1] = bGetShort (buf+4)*vel_scale;
	vel_enu[2] = bGetShort (buf+6)*vel_scale;

	*time_of_fix = bGetULong (buf+8)*.001;

	longtemp = bGetLong (buf+12);
	*lat = longtemp*(GPS_PI/MAX_LONG);

	ulongtemp = bGetULong (buf+16);
	*lon = ulongtemp*(GPS_PI/MAX_LONG);
	if (*lon > GPS_PI) *lon -= 2.0*GPS_PI;

	*alt = bGetLong (buf+20)*.001;
	/* 25 blank; 29 = UTC */
	(*datum_index) = (short)((short)buf[26]-1);
	*info = buf[27];
	*nsvs = buf[28];
	*week_num = bGetShort (&buf[30]);
	for (isv = 0; isv < 8; isv++) {
		prnx = buf[32+2*isv];
		sv_prn[isv] = (unsigned char)(prnx&0x3F);
      iode = buf[33+2*isv];
		sv_IODC[isv] = (short)(iode | ((prnx>>6)<<8));
	}
	return FALSE;
}

short rpt_0x8F41 (TSIPPKT *rpt,
	unsigned char *bSearchRange,
	unsigned char *bBoardOptions,
	unsigned long *iiSerialNumber,
	unsigned char *bBuildYear,
	unsigned char *bBuildMonth,
	unsigned char *bBuildDay,
	unsigned char *bBuildHour,
	float *fOscOffset,
	unsigned short *iTestCodeId)
{
	if(rpt->len != 17) return FALSE;
	*bSearchRange = rpt->buf[1];
	*bBoardOptions = rpt->buf[2];
	*iiSerialNumber = bGetLong(&rpt->buf[3]);
	*bBuildYear = rpt->buf[7];
	*bBuildMonth = rpt->buf[8];
	*bBuildDay = rpt->buf[9];
	*bBuildHour =	rpt->buf[10];
	*fOscOffset = bGetSingle(&rpt->buf[11]);
	*iTestCodeId = bGetShort(&rpt->buf[15]);
//	Tsipx8E41Data = *Tsipx8E41;
	return TRUE;
}

short rpt_0x8F42 (TSIPPKT *rpt,
	unsigned char *bProdOptionsPre,
	unsigned char *bProdNumberExt,
	unsigned short *iCaseSerialNumberPre,
	unsigned long *iiCaseSerialNumber,
	unsigned long *iiProdNumber,
	unsigned short *iPremiumOptions,
	unsigned short *iMachineID,
	unsigned short *iKey)
{
	if(rpt->len != 19) return FALSE;
	*bProdOptionsPre = rpt->buf[1];
	*bProdNumberExt = rpt->buf[2];
	*iCaseSerialNumberPre = bGetShort(&rpt->buf[3]);
	*iiCaseSerialNumber = bGetLong(&rpt->buf[5]);
	*iiProdNumber = bGetLong(&rpt->buf[9]);
	*iPremiumOptions = bGetShort(&rpt->buf[13]);
	*iMachineID = bGetShort(&rpt->buf[15]);
	*iKey = bGetShort(&rpt->buf[17]);
	return TRUE;
}

short rpt_0x8F45(TSIPPKT *rpt,
   unsigned char *bSegMask)
{
	if(rpt->len != 2) return FALSE;
	*bSegMask = rpt->buf[1];
	return TRUE;
}

short rpt_0x8F4A_16(TSIPPKT *rpt,
	unsigned char *pps_enabled,
   unsigned char *pps_timebase,
   unsigned char *pos_polarity,
   double *pps_offset,
   float *bias_unc_threshold)
/* Stinger PPS definition */
{
	unsigned char
   	*buf;

   buf = rpt->buf;
   if (rpt->len != 16) return TRUE;
   *pps_enabled = buf[1];
   *pps_timebase = buf[2];
   *pos_polarity = buf[3];
   *pps_offset = bGetDouble(&buf[4]);
   *bias_unc_threshold = bGetSingle(&buf[12]);
	return FALSE;
}

short rpt_0x8F4B(TSIPPKT *rpt,
                 unsigned long *decorr_max)
{
	unsigned char
   	*buf;

   buf = rpt->buf;
   if (rpt->len != 5) return TRUE;
   *decorr_max = bGetLong(&buf[1]);
   return FALSE;
}

short rpt_0x8F4D(TSIPPKT *rpt,
	unsigned long *event_mask)
{
	unsigned char
   	*buf;

   buf = rpt->buf;
   if (rpt->len != 5) return TRUE;
   *event_mask = bGetULong (&buf[1]);
   return FALSE;
}

short rpt_0x8FA5(TSIPPKT *rpt,
	unsigned char *spktmask)
{
	unsigned char
   	*buf;

   buf = rpt->buf;
   if (rpt->len != 5) return TRUE;
   spktmask[0] = buf[1];
   spktmask[1] = buf[2];
   spktmask[2] = buf[3];
   spktmask[3] = buf[4];
   return FALSE;
}

short rpt_0x8FAD (TSIPPKT *rpt,
    unsigned short *COUNT,
    double *FracSec,
    unsigned char *Hour,
    unsigned char *Minute,
    unsigned char *Second,
    unsigned char *Day,
    unsigned char *Month,
    unsigned short *Year,
    unsigned char *Status,
    unsigned char *Flags)
{

	if (rpt->len != 22) return TRUE;

    *COUNT = bGetUShort(&rpt->buf[1]);
    *FracSec = bGetDouble(&rpt->buf[3]);
    *Hour = rpt->buf[11];
    *Minute = rpt->buf[12];
    *Second = rpt->buf[13];
    *Day = rpt->buf[14];
    *Month = rpt->buf[15];
    *Year = bGetUShort(&rpt->buf[16]);
    *Status = rpt->buf[18];
    *Flags = rpt->buf[19];
    return FALSE;
}


/*
 * *************************************************************************
 *
 * Trimble Navigation, Ltd.
 * OEM Products Development Group
 * P.O. Box 3642
 * 645 North Mary Avenue
 * Sunnyvale, California 94088-3642
 *
 * Corporate Headquarter:
 *    Telephone:  (408) 481-8000
 *    Fax:        (408) 481-6005
 *
 * Technical Support Center:
 *    Telephone:  (800) 767-4822	(U.S. and Canada)
 *                (408) 481-6940    (outside U.S. and Canada)
 *    Fax:        (408) 481-6020
 *    BBS:        (408) 481-7800
 *    e-mail:     trimble_support@trimble.com
 *		ftp://ftp.trimble.com/pub/sct/embedded/bin
 *
 * *************************************************************************
 *
 * T_REPORT.C consists of a primary function TranslateTSIPReportToText()
 * called by main().
 *
 * This function takes a character buffer that has been received as a report
 * from a TSIP device and interprets it.  The character buffer has been
 * assembled using tsip_input_proc() in T_PARSER.C.
 *
 * A large case statement directs processing to one of many mid-level
 * functions.  The mid-level functions specific to the current report
 * code passes the report buffer to the appropriate report decoder
 * rpt_0x?? () in T_PARSER.C, which converts the byte stream in rpt.buf
 * to data values approporaite for use.
 *
 * *************************************************************************
 *
 */


#define GOOD_PARSE 0
#define BADID_PARSE 1
#define BADLEN_PARSE 2
#define BADDATA_PARSE 3

#define B_TSIP	0x02
#define B_NMEA	0x04


/* pbuf is the pointer to the current location of the text output */
static char
	*pbuf;

/* keep track of whether the message has been successfully parsed */
static short
	parsed;


/* convert time of week into day-hour-minute-second and print */
char* show_time (float time_of_week)
{
	short	days, hours, minutes;
	float seconds;
	double tow = 0;
   static char timestring [80];

	if (time_of_week == -1.0)
   {
		sprintf(timestring, "   <No time yet>   ");
	}
	else if ((time_of_week >= 604800.0) || (time_of_week < 0.0))
   {
		sprintf(timestring, "     <Bad time>     ");
	}
   else
   {
		if (time_of_week < 604799.9) 
			tow = time_of_week + .00000001;
		seconds = (float)fmod(tow, 60.);
		minutes =  (short) fmod(tow/60., 60.);
		hours = (short)fmod(tow / 3600., 24.);
		days = (short)(tow / 86400.0);
		sprintf(timestring, " %s %02d:%02d:%05.2f   ",
  	   	dayname[days], hours, minutes, seconds);
   }
   return timestring;
}

/**/
/* 0x3D */
static void rpt_chan_A_config (TSIPPKT *rpt)
{
	unsigned char
		tx_baud_index, rx_baud_index,
		char_format_index, stop_bits,
      tx_mode_index, rx_mode_index,
      databits, parity;
	int
		i, nbaud;

	/* unload rptbuf */
	if (rpt_0x3D (rpt,
		&tx_baud_index, &rx_baud_index, &char_format_index,
		&stop_bits, &tx_mode_index, &rx_mode_index)) {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf, "\nChannel A Configuration");

   nbaud = sizeof(old_baudnum);

	for (i = 0; i < nbaud; ++i) if (tx_baud_index == old_baudnum[i]) break;
	pbuf += sprintf(pbuf, "\n   Transmit speed: %s at %s",
		old_output_ch[tx_mode_index], st_baud_text_app[i]);

	for (i = 0; i < nbaud; ++i) if (rx_baud_index == old_baudnum[i]) break;
	pbuf += sprintf(pbuf, "\n   Receive speed: %s at %s",
		old_input_ch[rx_mode_index], st_baud_text_app[i]);

	databits = (unsigned char)((char_format_index & 0x03) + 5);

	parity = (unsigned char)(char_format_index >> 2);
	if (parity > 4) parity = 2;

	pbuf += sprintf(pbuf, "\n   Character format (bits/char, parity, stop bits): %d-%s-%d",
		databits, old_parity_text[parity], stop_bits);
}

/**/
/* 0x40 */
static void rpt_almanac_data_page (TSIPPKT *rpt)
{
	unsigned char
		sv_prn;
	short
		week_num;
	float
		t_zc,
		eccentricity,
		t_oa,
		i_0,
		OMEGA_dot,
		sqrt_A,
		OMEGA_0,
		omega,
		M_0;

	/* unload rptbuf */
	if (rpt_0x40 (rpt,
		&sv_prn, &week_num, &t_zc, &eccentricity, &t_oa,
		&i_0, &OMEGA_dot, &sqrt_A, &OMEGA_0, &omega, &M_0)) {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf, "\nAlmanac for SV %02d", sv_prn);
	pbuf += sprintf(pbuf, "\n       Captured:%15.0f %s",
   	t_zc, show_time (t_zc));
	pbuf += sprintf(pbuf, "\n           week:%15d", week_num);
	pbuf += sprintf(pbuf, "\n   Eccentricity:%15g", eccentricity);
	pbuf += sprintf(pbuf, "\n           T_oa:%15.0f %s",
   	t_oa, show_time (t_oa));
	pbuf += sprintf(pbuf, "\n            i 0:%15g", i_0);
	pbuf += sprintf(pbuf, "\n      OMEGA dot:%15g", OMEGA_dot);
	pbuf += sprintf(pbuf, "\n         sqrt A:%15g", sqrt_A);
	pbuf += sprintf(pbuf, "\n        OMEGA 0:%15g", OMEGA_0);
	pbuf += sprintf(pbuf, "\n          omega:%15g", omega);
	pbuf += sprintf(pbuf, "\n            M 0:%15g", M_0);
}

/* 0x41 */
static void rpt_GPS_time (TSIPPKT *rpt)
{
	float
		time_of_week, UTC_offset;
	short
		week_num;

	/* unload rptbuf */
	if (rpt_0x41 (rpt, &time_of_week, &UTC_offset, &week_num)) {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf, "\nGPS time:%s GPS week: %d   UTC offset %.1f",
   	show_time(time_of_week), week_num, UTC_offset);

}

/* 0x42 */
static void rpt_single_ECEF_position (TSIPPKT *rpt)
{
	float
		ECEF_pos[3], time_of_fix;

	/* unload rptbuf */
	if (rpt_0x42 (rpt, ECEF_pos, &time_of_fix)) {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf, "\nSXYZ:  %15.0f  %15.0f  %15.0f    %s",
		ECEF_pos[0], ECEF_pos[1], ECEF_pos[2],
      show_time(time_of_fix));
}

/* 0x43 */
static void rpt_single_ECEF_velocity (TSIPPKT *rpt)
{

	float
		ECEF_vel[3], freq_offset, time_of_fix;

	/* unload rptbuf */
	if (rpt_0x43 (rpt, ECEF_vel, &freq_offset, &time_of_fix)) {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf, "\nVelECEF: %11.3f  %11.3f  %11.3f  %12.3f%s",
		ECEF_vel[0], ECEF_vel[1], ECEF_vel[2], freq_offset,
      show_time(time_of_fix));
}

/*  0x45  */
static void rpt_SW_version (TSIPPKT *rpt) {
	unsigned char
		major_nav_version, minor_nav_version,
		nav_day, nav_month, nav_year,
		major_dsp_version, minor_dsp_version,
		dsp_day, dsp_month, dsp_year;

	/* unload rptbuf */
	if (rpt_0x45 (rpt,
		&major_nav_version, &minor_nav_version,
		&nav_day, &nav_month, &nav_year,
		&major_dsp_version, &minor_dsp_version,
		&dsp_day, &dsp_month, &dsp_year)) {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf,
"\nFW Versions:  Nav Proc %2d.%02d  %02d/%02d/%02d  Sig Proc %2d.%02d  %02d/%02d/%02d",
		major_nav_version, minor_nav_version, nav_day, nav_month, nav_year % 100,
		major_dsp_version, minor_dsp_version, dsp_day, dsp_month, dsp_year % 100);
}

/* 0x46 */
static void rpt_rcvr_health (TSIPPKT *rpt)
{
	unsigned char
		status1, status2;
	static char
		*sc_text[] = {
			"Doing position fixes",
			"Don't have GPS time yet",
			"Waiting for almanac collection",
			"DOP too high          ",
			"No satellites available",
			"Only 1 satellite available",
			"Only 2 satellites available",
			"Only 3 satellites available",
			"No satellites usable   ",
			"Only 1 satellite usable",
			"Only 2 satellites usable",
			"Only 3 satellites usable",
			"Chosen satellite unusable"};


	/* unload rptbuf */
	if (rpt_0x46 (rpt, &status1, &status2))
	{
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf, "\nRcvr status1: %s (%02Xh); ",
     	sc_text[rpt->buf[0]], status1);

	pbuf += sprintf(pbuf, "status2: %s, %s (%02Xh)",
		(status2 & 0x01)?"No BBRAM":"BBRAM OK",
		(status2 & 0x10)?"No Ant":"Ant OK",
      status2);
}

/* 0x47 */
static void rpt_SNR_all_SVs (TSIPPKT *rpt)
{
	unsigned char
		nsvs, sv_prn[12];
	short
		isv;
	float
		snr[12];

	/* unload rptbuf */
	if (rpt_0x47 (rpt, &nsvs, sv_prn, snr))
   {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf, "\nSNR for satellites: %d", nsvs);
	for (isv = 0; isv < nsvs; isv++)
   {
		pbuf += sprintf(pbuf, "\n    SV %02d   %6.2f",
      	sv_prn[isv], snr[isv]);
	}
}

/* 0x48 */
static void rpt_GPS_system_message (TSIPPKT *rpt)
{
	unsigned char
		message[23];

	/* unload rptbuf */
	if (rpt_0x48 (rpt, message))
   {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf, "\nGPS message: %s", message);
}

/* 0x49 */
static void rpt_almanac_health_page (TSIPPKT *rpt)
{
	short
		iprn;
	unsigned char
		sv_health [32];

	/* unload rptbuf */
	if (rpt_0x49 (rpt, sv_health))
   {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf, "\nAlmanac health page:");
	for (iprn = 0; iprn < 32; iprn++)
   {
		if (!(iprn%5)) *pbuf++ = '\n';
		pbuf += sprintf(pbuf, "    SV%02d  %2X",
      	(iprn+1) , sv_health[iprn]);
	}
}

/* 0x4A */
static void rpt_single_lla_position (TSIPPKT *rpt) {
	short
		lat_deg, lon_deg;
	float
		lat, lon,
		alt, clock_bias, time_of_fix;
	double lat_min, lon_min;
	unsigned char
		north_south, east_west;

	if (rpt_0x4A (rpt,
		&lat, &lon, &alt, &clock_bias, &time_of_fix))
   {
		parsed = BADLEN_PARSE;
		return;
	}

	/* convert from radians to degrees */
	lat *= (float)R2D;
	north_south = 'N';
	if (lat < 0.0)
   {
		north_south = 'S';
		lat = -lat;
	}
	lat_deg = (short)lat;
	lat_min = (lat - lat_deg) * 60.0;

	lon *= (float)R2D;
	east_west = 'E';
	if (lon < 0.0)
   {
		east_west = 'W';
		lon = -lon;
	}
	lon_deg = (short)lon;
	lon_min = (lon - lon_deg) * 60.0;

	pbuf += sprintf(pbuf, "\nSLLA: %4d: %06.3f  %c%5d:%06.3f  %c%10.2f  %12.2f%s",
		lat_deg, lat_min, north_south,
		lon_deg, lon_min, east_west,
		alt, clock_bias,
		show_time(time_of_fix));
}

/* 0x4A */
static void rpt_ref_alt (TSIPPKT *rpt) {

	float
		alt, dummy;
	unsigned char
		alt_flag;

	if (rpt_0x4A_2 (rpt,
		&alt, &dummy, &alt_flag))
   {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf, "\nReference Alt:   %.1f m;    %s",
   	alt, alt_flag?"ON":"OFF");
}

/* 0x4B */
static void rpt_rcvr_id_and_status (TSIPPKT *rpt)
{

	unsigned char
		machine_id, status3, status4;

	/* unload rptbuf */
	if (rpt_0x4B (rpt, &machine_id, &status3, &status4))
   {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf, "\nRcvr Machine ID: %d; Status3 = %s, %s (%02Xh)",
   	machine_id,
		(status3 & 0x02)?"No RTC":"RTC OK",
		(status3 & 0x08)?"No Alm":"Alm OK",
		status3);
}

/* 0x4C */
static void rpt_operating_parameters (TSIPPKT *rpt)
{
	unsigned char
		dyn_code;
	float
		el_mask, snr_mask, dop_mask, dop_switch;

	/* unload rptbuf */
	if (rpt_0x4C (rpt, &dyn_code, &el_mask,
		&snr_mask, &dop_mask, &dop_switch))
   {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf, "\nOperating Parameters:");
	pbuf += sprintf(pbuf, "\n     Dynamics code = %d %s",
   	dyn_code, dyn_text[dyn_code]);
	pbuf += sprintf(pbuf, "\n     Elevation mask = %.2f", el_mask * R2D);
	pbuf += sprintf(pbuf, "\n     SNR mask = %.2f", snr_mask);
	pbuf += sprintf(pbuf, "\n     DOP mask = %.2f", dop_mask);
	pbuf += sprintf(pbuf, "\n     DOP switch = %.2f", dop_switch);
}

/* 0x4D */
static void rpt_oscillator_offset (TSIPPKT *rpt)
{
	float
		osc_offset;

	/* unload rptbuf */
	if (rpt_0x4D (rpt, &osc_offset))
   {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf, "\nOscillator offset: %.2f Hz = %.3f PPM",
   	osc_offset, osc_offset/1575.42);
}

/* 0x4E */
static void rpt_GPS_time_set_response (TSIPPKT *rpt)
{

	unsigned char
		response;

	/* unload rptbuf */
	if (rpt_0x4E (rpt, &response))
   {
		parsed = BADLEN_PARSE;
		return;
	}

	switch (response)
   {
	case 'Y':
		pbuf += sprintf(pbuf, "\nTime set accepted");
		break;

	case 'N':
		pbuf += sprintf(pbuf, "\nTime set rejected or not required");
		break;

	default:
		parsed = BADDATA_PARSE;
	}
}

/* 0x4F */
static void rpt_UTC_offset (TSIPPKT *rpt)
{
	double
		a0;
	float
		a1, time_of_data;
	short
		dt_ls, wn_t, wn_lsf, dn, dt_lsf;

	/* unload rptbuf */
	if (rpt_0x4F (rpt, &a0, &a1, &time_of_data,
		&dt_ls, &wn_t, &wn_lsf, &dn, &dt_lsf)) {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf, "\nUTC Correction Data");
	pbuf += sprintf(pbuf, "\n   A_0         = %g  ", a0);
	pbuf += sprintf(pbuf, "\n   A_1         = %g  ", a1);
	pbuf += sprintf(pbuf, "\n   delta_t_LS  = %d  ", dt_ls);
	pbuf += sprintf(pbuf, "\n   t_ot        = %.0f  ", time_of_data);
	pbuf += sprintf(pbuf, "\n   WN_t        = %d  ", wn_t );
	pbuf += sprintf(pbuf, "\n   WN_LSF      = %d  ", wn_lsf );
	pbuf += sprintf(pbuf, "\n   DN          = %d  ", dn );
	pbuf += sprintf(pbuf, "\n   delta_t_LSF = %d  ", dt_lsf );
}

/**/
/* 0x54 */
static void rpt_1SV_bias (TSIPPKT *rpt)
{
	float
		clock_bias, freq_offset, time_of_fix;

	/* unload rptbuf */
	if (rpt_0x54 (rpt, &clock_bias, &freq_offset, &time_of_fix)) {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf (pbuf, "\nTime Fix   Clock Bias: %6.2f m  Freq Bias: %6.2f m/s%s",
		clock_bias, freq_offset, show_time (time_of_fix));
}

/* 0x55 */
static void rpt_io_opt (TSIPPKT *rpt)
{
	unsigned char
		pos_code, vel_code, time_code, aux_code;

	/* unload rptbuf */
	if (rpt_0x55 (rpt,
		&pos_code, &vel_code, &time_code, &aux_code)) {
		parsed = BADLEN_PARSE;
		return;
	}
	/* rptbuf unloaded */

	pbuf += sprintf(pbuf, "\nI/O Options: %2X %2X %2X %2X",
		pos_code, vel_code, time_code, aux_code);

	if (pos_code & 0x01) {
		pbuf += sprintf(pbuf, "\n    ECEF XYZ position output");
	}

	if (pos_code & 0x02) {
		pbuf += sprintf(pbuf, "\n    LLA position output");
	}

	pbuf += sprintf(pbuf, (pos_code & 0x04)?
		"\n    MSL altitude output (Geoid height) ":
		"\n    WGS-84 altitude output");

	pbuf += sprintf(pbuf, (pos_code & 0x08)?
		"\n    MSL altitude input":
      "\n    WGS-84 altitude input");

	pbuf += sprintf(pbuf, (pos_code & 0x10)?
		"\n    Double precision":
      "\n    Single precision");

	if (pos_code & 0x20) {
		pbuf += sprintf(pbuf, "\n    All Enabled Superpackets");
	}

	if (vel_code & 0x01) {
		pbuf += sprintf(pbuf, "\n    ECEF XYZ velocity output");
	}

	if (vel_code & 0x02) {
		pbuf += sprintf(pbuf, "\n    ENU velocity output");
	}

	pbuf += sprintf(pbuf, (time_code & 0x01)?
		  "\n    Time tags in UTC":
        "\n    Time tags in GPS time");

	if (time_code & 0x02) {
		pbuf += sprintf(pbuf, "\n    Fixes delayed to integer seconds");
	}

	if (time_code & 0x04) {
		pbuf += sprintf(pbuf, "\n    Fixes sent only on request");
	}

	if (time_code & 0x08) {
		pbuf += sprintf(pbuf, "\n    Synchronized measurements");
	}

	if (time_code & 0x10) {
		pbuf += sprintf(pbuf, "\n    Minimize measurement propagation");
	}

   pbuf += sprintf(pbuf, (time_code & 0x20) ?
		"\n    PPS output at all times" :
   	"\n    PPS output during fixes");

	if (aux_code & 0x01) {
		pbuf += sprintf(pbuf, "\n    Raw measurement output");
	}

	if (aux_code & 0x02) {
		pbuf += sprintf(pbuf, "\n    Code-phase smoothed before output");
	}

	if (aux_code & 0x04) {
		pbuf += sprintf(pbuf, "\n    Additional fix status");
	}

	pbuf += sprintf(pbuf, (aux_code & 0x08)?
   	"\n    Signal Strength Output as dBHz" :
   	"\n    Signal Strength Output as AMU");
}

/* 0x56 */
static void rpt_ENU_velocity (TSIPPKT *rpt)
{
	float
		vel_ENU[3], freq_offset, time_of_fix;

	/* unload rptbuf */
	if (rpt_0x56 (rpt, vel_ENU, &freq_offset, &time_of_fix)) {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf, "\nVel ENU: %11.3f  %11.3f  %11.3f  %12.3f%s",
		vel_ENU[0], vel_ENU[1], vel_ENU[2], freq_offset,
      show_time (time_of_fix));
}

/* 0x57 */
static void rpt_last_fix_info (TSIPPKT *rpt)
{
	unsigned char
		source_code, diag_code;
	short
		week_num;
	float
		time_of_fix;

	/* unload rptbuf */
	if (rpt_0x57 (rpt, &source_code, &diag_code, &week_num, &time_of_fix)) {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf, "\n source code %d;   diag code: %2Xh",
   	source_code, diag_code);
	pbuf += sprintf(pbuf, "\n    Time of last fix:%s", show_time(time_of_fix));
	pbuf += sprintf(pbuf, "\n    Week of last fix: %d", week_num);
}

/* 0x58 */
static void rpt_GPS_system_data (TSIPPKT *rpt)
{
	unsigned char
   	iprn,
		op_code, data_type, sv_prn,
		data_length, data_packet[250];
	ALM_INFO
		*almanac;
	ALH_PARMS
		*almh;
	UTC_INFO
		*utc;
	ION_INFO
		*ionosphere;
	EPHEM_CLOCK
		*cdata;
	EPHEM_ORBIT
		*edata;
	NAV_INFO
		*nav_data;
	unsigned char
		curr_t_oa;
	unsigned short
		curr_wn_oa;
	static char
		*datname[] =
		{"", "", "Almanac Orbit",
		"Health Page & Ref Time", "Ionosphere", "UTC ",
		"Ephemeris"};

	/* unload rptbuf */
	if (rpt_0x58 (rpt, &op_code, &data_type, &sv_prn,
		&data_length, data_packet))
   {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf, "\nSystem data [%d]:  %s  SV%02d",
		data_type, datname[data_type], sv_prn);
	switch (op_code)
	{
	case 1:
		pbuf += sprintf(pbuf, "  Acknowledgment");
		break;
	case 2:
		pbuf += sprintf(pbuf, "  length = %d bytes", data_length);
		switch (data_type) {
		case 2:
			/* Almanac */
			if (sv_prn == 0 || sv_prn > 32) {
				pbuf += sprintf(pbuf, "  Binary PRN invalid");
				return;
			}
			almanac = (ALM_INFO*)data_packet;
			pbuf += sprintf(pbuf, "\n   t_oa_raw = % -12d    SV_hlth  = % -12d  ",
         	almanac->t_oa_raw , almanac->SV_health );
			pbuf += sprintf(pbuf, "\n   e        = % -12g    t_oa     = % -12g  ",
         	almanac->e        , almanac->t_oa     );
			pbuf += sprintf(pbuf, "\n   i_0      = % -12g    OMEGADOT = % -12g  ",
         	almanac->i_0      , almanac->OMEGADOT );
			pbuf += sprintf(pbuf, "\n   sqrt_A   = % -12g    OMEGA_0  = % -12g  ",
         	almanac->sqrt_A   , almanac->OMEGA_0  );
			pbuf += sprintf(pbuf, "\n   omega    = % -12g    M_0      = % -12g  ",
         	almanac->omega    , almanac->M_0      );
			pbuf += sprintf(pbuf, "\n   a_f0     = % -12g    a_f1     = % -12g  ",
         	almanac->a_f0     , almanac->a_f1     );
			pbuf += sprintf(pbuf, "\n   Axis     = % -12g    n        = % -12g  ",
         	almanac->Axis     , almanac->n        );
			pbuf += sprintf(pbuf, "\n   OMEGA_n  = % -12g    ODOT_n   = % -12g  ",
         	almanac->OMEGA_n  , almanac->ODOT_n   );
			pbuf += sprintf(pbuf, "\n   t_zc     = % -12g    weeknum  = % -12d  ",
         	almanac->t_zc     , almanac->weeknum  );
			pbuf += sprintf(pbuf, "\n   wn_oa    = % -12d", almanac->wn_oa    );
			break;

		case 3:
			/* Almanac health page */
			almh = (ALH_PARMS*)data_packet;
			pbuf += sprintf(pbuf, "\n   t_oa = %d, wn_oa&0xFF = %d  ",
         	almh->t_oa, almh->WN_a);
			pbuf += sprintf(pbuf, "\nAlmanac health page:");
			for (iprn = 0; iprn < 32; iprn++) {
				if (!(iprn%5)) *pbuf++ = '\n';
				pbuf += sprintf(pbuf, "    SV%02d  %2X",
            	(iprn+1) , almh->SV_health[iprn]);
			}
			curr_t_oa = data_packet[34];
			curr_wn_oa = (unsigned short)((data_packet[35]<<8) + data_packet[36]);
			pbuf += sprintf(pbuf, "\n   current t_oa = %d, wn_oa = %d  ",
         	curr_t_oa, curr_wn_oa);
			break;

		case 4:
			/* Ionosphere */
			ionosphere = (ION_INFO*)data_packet;
			pbuf += sprintf(pbuf, "\n   alpha_0 = % -12g  alpha_1 = % -12g ",
	         ionosphere->alpha_0, ionosphere->alpha_1);
			pbuf += sprintf(pbuf, "\n   alpha_2 = % -12g  alpha_3 = % -12g ",
	         ionosphere->alpha_2, ionosphere->alpha_3);
			pbuf += sprintf(pbuf, "\n   beta_0  = % -12g  beta_1  = % -12g  ",
	         ionosphere->beta_0, ionosphere->beta_1);
			pbuf += sprintf(pbuf, "\n   beta_2  = % -12g  beta_3  = % -12g  ",
	         ionosphere->beta_2, ionosphere->beta_3);
			break;

		case 5:
			/* UTC */
			utc = (UTC_INFO*)data_packet;
			pbuf += sprintf(pbuf, "\n   A_0         = %g  ", utc->A_0);
			pbuf += sprintf(pbuf, "\n   A_1         = %g  ", utc->A_1);
			pbuf += sprintf(pbuf, "\n   delta_t_LS  = %d  ", utc->delta_t_LS);
			pbuf += sprintf(pbuf, "\n   t_ot        = %.0f  ", utc->t_ot );
			pbuf += sprintf(pbuf, "\n   WN_t        = %d  ", utc->WN_t );
			pbuf += sprintf(pbuf, "\n   WN_LSF      = %d  ", utc->WN_LSF );
			pbuf += sprintf(pbuf, "\n   DN          = %d  ", utc->DN );
			pbuf += sprintf(pbuf, "\n   delta_t_LSF = %d  ", utc->delta_t_LSF );
			break;

		case 6: /* Ephemeris */
			if (sv_prn == 0 || sv_prn > 32) {
				pbuf += sprintf(pbuf, "  Binary PRN invalid");
				return;
			}
			nav_data = (NAV_INFO*)data_packet;

			pbuf += sprintf(pbuf, "\n     SV_PRN = % -12d .  t_ephem = % -12g . ",
         	nav_data->sv_number , nav_data->t_ephem );
			cdata = &(nav_data->ephclk);
			pbuf += sprintf(pbuf,
         	"\n    weeknum = % -12d .   codeL2 = % -12d .  L2Pdata = % -12d",
         	cdata->weeknum , cdata->codeL2 , cdata->L2Pdata );
			pbuf += sprintf(pbuf,
         	"\n  SVacc_raw = % -12d .SV_health = % -12d .     IODC = % -12d",
         	cdata->SVacc_raw, cdata->SV_health, cdata->IODC );
			pbuf += sprintf(pbuf,
         	"\n       T_GD = % -12g .     t_oc = % -12g .     a_f2 = % -12g",
         	cdata->T_GD, cdata->t_oc, cdata->a_f2 );
			pbuf += sprintf(pbuf,
         	"\n       a_f1 = % -12g .     a_f0 = % -12g .    SVacc = % -12g",
         	cdata->a_f1, cdata->a_f0, cdata->SVacc );
			edata = &(nav_data->ephorb);
			pbuf += sprintf(pbuf,
	         "\n       IODE = % -12d .fit_intvl = % -12d .     C_rs = % -12g",
	         edata->IODE, edata->fit_interval, edata->C_rs );
			pbuf += sprintf(pbuf,
         	"\n    delta_n = % -12g .      M_0 = % -12g .     C_uc = % -12g",
         	edata->delta_n, edata->M_0, edata->C_uc );
			pbuf += sprintf(pbuf,
         	"\n        ecc = % -12g .     C_us = % -12g .   sqrt_A = % -12g",
	         edata->e, edata->C_us, edata->sqrt_A );
			pbuf += sprintf(pbuf,
         	"\n       t_oe = % -12g .     C_ic = % -12g .  OMEGA_0 = % -12g",
            edata->t_oe, edata->C_ic, edata->OMEGA_0 );
			pbuf += sprintf(pbuf,
	         "\n       C_is = % -12g .      i_0 = % -12g .     C_rc = % -12g",
	         edata->C_is, edata->i_0, edata->C_rc );
			pbuf += sprintf(pbuf,
	         "\n      omega = % -12g . OMEGADOT = % -12g .     IDOT = % -12g",
         	edata->omega, edata->OMEGADOT, edata->IDOT );
			pbuf += sprintf(pbuf,
   	      "\n       Axis = % -12g .        n = % -12g .    r1me2 = % -12g",
	         edata->Axis, edata->n, edata->r1me2 );
			pbuf += sprintf(pbuf,
      	   "\n    OMEGA_n = % -12g .   ODOT_n = % -12g",
	         edata->OMEGA_n, edata->ODOT_n );
			break;
		}
	}
}


/* 0x59: */
static void rpt_SVs_enabled (TSIPPKT *rpt)
{
	unsigned char
   	numsvs,
		code_type,
      status_code[32];
	short
		iprn;

	/* unload rptbuf */
	if (rpt_0x59 (rpt, &code_type, status_code))
   {
		parsed = BADLEN_PARSE;
		return;
	}
   switch (code_type)
   {
   case 3: pbuf += sprintf(pbuf, "\nSVs Disabled:\n"); break;
   case 6: pbuf += sprintf(pbuf, "\nSVs with Health Ignored:\n"); break;
   default: return;
   }
   numsvs = 0;
	for (iprn=0; iprn<32; iprn++)
   {
     	if (status_code[iprn])
      {
	   	pbuf += sprintf(pbuf, " %02d", iprn+1);
   	   numsvs++;
      }
   }
   if (numsvs == 0) pbuf += sprintf(pbuf, "None");
}


/* 0x5A */
static void rpt_raw_msmt (TSIPPKT *rpt)
{
	unsigned char
		sv_prn;
	float
		sample_length, signal_level, code_phase, Doppler;
	double
		time_of_fix;

	/* unload rptbuf */
	if (rpt_0x5A (rpt, &sv_prn, &sample_length, &signal_level,
		&code_phase, &Doppler, &time_of_fix))
   {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf, "\n   %02d %5.0f %7.1f %10.2f %10.2f %12.3f %s",
		sv_prn, sample_length, signal_level, code_phase, Doppler, time_of_fix,
		show_time ((float)time_of_fix));
}

/* 0x5B */
static void rpt_SV_ephemeris_status (TSIPPKT *rpt)
{
	unsigned char
		sv_prn, sv_health, sv_iode, fit_interval_flag;
	float
		time_of_collection, time_of_eph, sv_accy;

	/* unload rptbuf */
	if (rpt_0x5B (rpt, &sv_prn, &sv_health, &sv_iode, &fit_interval_flag,
		&time_of_collection, &time_of_eph, &sv_accy))
   {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf, "\n  SV%02d  %s   %2Xh     %2Xh ",
   	sv_prn, show_time (time_of_collection), sv_health, sv_iode);
	/* note: cannot use show_time twice in same call */
	pbuf += sprintf(pbuf, "%s   %1d   %4.1f",
      show_time (time_of_eph), fit_interval_flag, sv_accy);
}

/* 0x5C */
static void rpt_SV_tracking_status (TSIPPKT *rpt)
{
	unsigned char
		sv_prn, chan, slot, acq_flag, eph_flag,
		old_msmt_flag, integer_msec_flag, bad_data_flag,
		data_collect_flag;
	float
		signal_level, time_of_last_msmt,
		elev, azim;

	/* unload rptbuf */
	if (rpt_0x5C (rpt,
		&sv_prn, &slot, &chan, &acq_flag, &eph_flag,
		&signal_level, &time_of_last_msmt, &elev, &azim,
		&old_msmt_flag, &integer_msec_flag, &bad_data_flag,
		&data_collect_flag))
   {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf,
"\n SV%2d  %1d   %1d   %1d   %4.1f  %s  %5.1f  %5.1f",
		sv_prn, chan,
      acq_flag, eph_flag, signal_level,
      show_time(time_of_last_msmt),
		elev*R2D, azim*R2D);
}

/**/
/* 0x6D */
static void rpt_allSV_selection (TSIPPKT *rpt)
{
	unsigned char
		manual_mode, nsvs, sv_prn[8], ndim;
	short
		islot;
	float
		pdop, hdop, vdop, tdop;

	/* unload rptbuf */
	if (rpt_0x6D (rpt,
		&manual_mode, &nsvs, &ndim, sv_prn,
		&pdop, &hdop, &vdop, &tdop))
   {
		parsed = BADLEN_PARSE;
		return;
	}

	switch (ndim)
   {
   case 0:
		pbuf += sprintf(pbuf, "\nMode: Searching, %d-SV:", nsvs);
      break;
   case 1:
		pbuf += sprintf(pbuf, "\nMode: One-SV Timing:");
      break;
   case 3: case 4:
		pbuf += sprintf(pbuf, "\nMode: %c-%dD, %d-SV:",
   			manual_mode ? 'M' : 'A', ndim - 1,  nsvs);
      break;
	case 5:
		pbuf += sprintf(pbuf, "\nMode: Timing, %d-SV:", nsvs);
      break;
   default:
		pbuf += sprintf(pbuf, "\nMode: Unknown = %d:", ndim);
      break;
   }

	for (islot = 0; islot < nsvs; islot++)
   {
		if (sv_prn[islot]) pbuf += sprintf(pbuf, " %02d", sv_prn[islot]);
	}
   if (ndim == 3 || ndim == 4 || ndim == 5)
   {
		pbuf += sprintf(pbuf, ";  DOPs: P %.1f H %.1f V %.1f T %.1f",
			pdop, hdop, vdop, tdop);
   }
}

/**/
/* 0x82 */
static void rpt_DGPS_position_mode (TSIPPKT *rpt)
{
	unsigned char
		diff_mode;

	/* unload rptbuf */
	if (rpt_0x82 (rpt, &diff_mode)) {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf, "\nFix is%s DGPS-corrected (%s mode)  (%d)",
   	(diff_mode&1) ? "" : " not",
   	(diff_mode&2) ? "auto" : "manual",
      diff_mode);
}

/* 0x83 */
static void rpt_double_ECEF_position (TSIPPKT *rpt)
{

	double
		ECEF_pos[3], clock_bias;
	float
		time_of_fix;

	/* unload rptbuf */
	if (rpt_0x83 (rpt, ECEF_pos, &clock_bias, &time_of_fix))
   {
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf, "\nDXYZ:%12.2f  %13.2f  %13.2f %12.2f%s",
		ECEF_pos[0], ECEF_pos[1], ECEF_pos[2], clock_bias,
		show_time(time_of_fix));
}

/* 0x84 */
static void rpt_double_lla_position (TSIPPKT *rpt)
{
	short
		lat_deg, lon_deg;
	double
		lat, lon, lat_min, lon_min,
		alt, clock_bias;
	float
		time_of_fix;
	unsigned char
		north_south, east_west;

	/* unload rptbuf */
	if (rpt_0x84 (rpt,
		&lat, &lon, &alt, &clock_bias, &time_of_fix))
   {
		parsed = BADLEN_PARSE;
		return;
	}

	lat *= R2D;
	lon *= R2D;
	if (lat < 0.0) {
		north_south = 'S';
		lat = -lat;
	} else {
		north_south = 'N';
	}
	lat_deg = (short)lat;
	lat_min = (lat - lat_deg) * 60.0;

	if (lon < 0.0) {
		east_west = 'W';
		lon = -lon;
	} else {
		east_west = 'E';
	}
	lon_deg = (short)lon;
	lon_min = (lon - lon_deg) * 60.0;
	pbuf += sprintf(pbuf, "\nDLLA: %2d:%08.5f %c; %3d:%08.5f %c; %10.2f %12.2f%s",
		lat_deg, lat_min, north_south,
		lon_deg, lon_min, east_west,
		alt, clock_bias,
		show_time(time_of_fix));
}

/* 0xBB */
static void rpt_complete_rcvr_config (TSIPPKT *rpt)
{
	TSIP_RCVR_CFG TsipxBB ;
	/* unload rptbuf */
	if (rpt_Paly0xBB (rpt, &TsipxBB))
	{
		parsed = BADLEN_PARSE;
		return;
	}

	pbuf += sprintf(pbuf, "\n   operating mode:      %s",
		NavModeText0xBB[TsipxBB.operating_mode]);
	pbuf += sprintf(pbuf, "\n   dynamics:            %s",
		dyn_text[TsipxBB.dyn_code]);
	pbuf += sprintf(pbuf, "\n   elev angle mask:     %g deg",
		TsipxBB.elev_mask * R2D);
	pbuf += sprintf(pbuf, "\n   SNR mask:            %g AMU",
		TsipxBB.cno_mask);
	pbuf += sprintf(pbuf, "\n   DOP mask:            %g",
		TsipxBB.dop_mask);
	pbuf += sprintf(pbuf, "\n   DOP switch:          %g",
		TsipxBB.dop_switch);
	return ;
}

/* 0xBC */
static void rpt_rcvr_serial_port_config (TSIPPKT *rpt)
{
	unsigned char
		port_num, in_baud, out_baud, data_bits, parity, stop_bits, flow_control,
		protocols_in, protocols_out, reserved;
	unsigned char known;

	/* unload rptbuf */
	if (rpt_0xBC (rpt, &port_num, &in_baud, &out_baud, &data_bits, &parity,
			&stop_bits, &flow_control, &protocols_in, &protocols_out, &reserved)) {
		parsed = BADLEN_PARSE;
		return;
	}
	/* rptbuf unloaded */

	pbuf += sprintf(pbuf, "\n   RECEIVER serial port %s config:",
		rcvr_port_text[port_num]);

	pbuf += sprintf(pbuf, "\n             I/O Baud %s/%s, %d - %s - %d",
		st_baud_text_app[in_baud],
		st_baud_text_app[out_baud],
		data_bits+5,
		parity_text[parity],
		stop_bits=1);
	pbuf += sprintf(pbuf, "\n             Input protocols: ");
	known = FALSE;
	if (protocols_in&B_TSIP)
   {
		pbuf += sprintf(pbuf, "%s ", protocols_in_text[1]);
		known = TRUE;
	}
	if (known == FALSE) pbuf += sprintf(pbuf, "No known");

	pbuf += sprintf(pbuf, "\n             Output protocols: ");
	known = FALSE;
	if (protocols_out&B_TSIP)
   {
		pbuf += sprintf(pbuf, "%s ", protocols_out_text[1]);
		known = TRUE;
	}
	if (protocols_out&B_NMEA)
   {
		pbuf += sprintf(pbuf, "%s ", protocols_out_text[2]);
		known = TRUE;
	}
	if (known == FALSE) pbuf += sprintf(pbuf, "No known");
	reserved = reserved;

 }

/* 0x8F */
/* 8F0B */
static void rpt_8F0B(TSIPPKT *rpt)
{
	const char
   	*oprtng_dim[7] = {
      	"horizontal (2-D)",
         "full position (3-D)",
         "single satellite (0-D)",
         "automatic",
         "N/A",
         "N/A",
         "overdetermined clock"};
   char
   	sv_id[8];
   unsigned char
   	month,
      date,
      dim_mode,
      north_south,
      east_west;
   unsigned short
   	event;
   short
   	utc_offset,
      year,
      index;
	short
   	lat_deg,
      lon_deg;
   float
   	bias_unc,
      dr_unc;
   double
   	tow,
      bias,
      drift,
      lat,
      lon,
      alt,
      lat_min,
      lon_min;
   int
   	numfix,
      numnotfix;

	if (rpt_0x8F0B(rpt,
   	&event,
      &tow,
      &date,
      &month,
      &year,
      &dim_mode,
      &utc_offset,
      &bias,
      &drift,
      &bias_unc,
      &dr_unc,
      &lat,
      &lon,
      &alt,
      sv_id))
   {
		parsed = BADLEN_PARSE;
		return;
	}

	if (event == 0)
   {
   	pbuf += sprintf(pbuf, "\nNew partial+full meas");
	}
   else
   {
		pbuf += sprintf(pbuf, "\nEvent count: %5d", event);
   }

	pbuf += sprintf(pbuf, "\nGPS time  : %s %2d/%2d/%2d (DMY)",
   	show_time(tow), date, month, year);
	pbuf += sprintf(pbuf, "\nMode      : %s", oprtng_dim[dim_mode]);
	pbuf += sprintf(pbuf, "\nUTC offset: %2d", utc_offset);
	pbuf += sprintf(pbuf, "\nClock Bias: %6.2f m", bias);
	pbuf += sprintf(pbuf, "\nFreq bias : %6.2f m/s", drift);
	pbuf += sprintf(pbuf, "\nBias unc  : %6.2f m", bias_unc);
	pbuf += sprintf(pbuf, "\nFreq unc  : %6.2f m/s", dr_unc);

	lat *= R2D; /* convert from radians to degrees */
	lon *= R2D;
	if (lat < 0.0)
   {
		north_south = 'S';
		lat = -lat;
	}
   else
   {
		north_south = 'N';
	}

	lat_deg = (short)lat;
	lat_min = (lat - lat_deg) * 60.0;
	if (lon < 0.0)
   {
		east_west = 'W';
		lon = -lon;
	}
   else
   {
		east_west = 'E';
	}

	lon_deg = (short)lon;
	lon_min = (lon - lon_deg) * 60.0;
	pbuf += sprintf(pbuf, "\nPosition  :");
	pbuf += sprintf(pbuf, " %4d %6.3f %c", lat_deg, lat_min, north_south);
	pbuf += sprintf(pbuf, " %5d %6.3f %c", lon_deg, lon_min, east_west);
	pbuf += sprintf(pbuf, " %10.2f", alt);

   numfix = numnotfix = 0;
	for (index=0; index<8; index++)
   {
		if (sv_id[index] < 0) numnotfix++;
		if (sv_id[index] > 0) numfix++;
   }
   if (numfix > 0)
   {
		pbuf += sprintf(pbuf, "\nSVs used in fix  : ");
		for (index=0; index<8; index++)
	   {
			if (sv_id[index] > 0)
      	{
      		pbuf += sprintf(pbuf, "%2d ", sv_id[index]);
	      }
   	}
   }
   if (numnotfix > 0)
   {
		pbuf += sprintf(pbuf, "\nOther SVs tracked: ");
		for (index=0; index<8; index++)
	   {
			if (sv_id[index] < 0)
      	{
      		pbuf += sprintf(pbuf, "%2d ", sv_id[index]);
	      }
   	}
   }
}

/* 0x8F14 */
static void rpt_8F14 (TSIPPKT *rpt)
/* Datum parameters */
{
	double
		datum_coeffs[5];
	short
		datum_idx;

	/* unload rptbuf */
	if (rpt_0x8F14 (rpt, &datum_idx, datum_coeffs))
   {
		parsed = BADLEN_PARSE;
		return;
	}

	if (datum_idx == -1)
   {
   	pbuf += sprintf(pbuf, "\nUser-Entered Datum:");
		pbuf += sprintf(pbuf, "\n   dx        = %6.1f", datum_coeffs[0]);
		pbuf += sprintf(pbuf, "\n   dy        = %6.1f", datum_coeffs[1]);
		pbuf += sprintf(pbuf, "\n   dz        = %6.1f", datum_coeffs[2]);
		pbuf += sprintf(pbuf, "\n   a-axis    = %10.3f", datum_coeffs[3]);
		pbuf += sprintf(pbuf, "\n   e-squared = %16.14f", datum_coeffs[4]);
   }
   else if (datum_idx == 0)
   {
   	pbuf += sprintf(pbuf, "\nWGS-84 datum, Index 0 ");
   }
   else
   {
   	pbuf += sprintf(pbuf, "\nStandard Datum, Index %3d ", datum_idx);
   }
}

/* 0x8F15 */
static void rpt_8F15 (TSIPPKT *rpt)
/* Datum parameters */
{
	double
		datum_coeffs[5];
	short
		datum_idx;

	/* unload rptbuf */
	if (rpt_0x8F15 (rpt, &datum_idx, datum_coeffs)) {
		parsed = BADLEN_PARSE;
		return;
	}

	if (datum_idx == -1)
   {
   	pbuf += sprintf(pbuf, "\nUser-Entered Datum:");
		pbuf += sprintf(pbuf, "\n   dx        = %6.1f", datum_coeffs[0]);
		pbuf += sprintf(pbuf, "\n   dy        = %6.1f", datum_coeffs[1]);
		pbuf += sprintf(pbuf, "\n   dz        = %6.1f", datum_coeffs[2]);
		pbuf += sprintf(pbuf, "\n   a-axis    = %10.3f", datum_coeffs[3]);
		pbuf += sprintf(pbuf, "\n   e-squared = %16.14f", datum_coeffs[4]);
   }
   else if (datum_idx == 0)
   {
   	pbuf += sprintf(pbuf, "\nWGS-84 datum, Index 0 ");
   }
   else
   {
   	pbuf += sprintf(pbuf, "\nStandard Datum, Index %3d ", datum_idx);
   }
}

/* 0x8F20 */
#define INFO_DGPS       0x02
#define INFO_2D         0x04
#define INFO_ALTSET     0x08
#define INFO_FILTERED   0x10
static void rpt_8F20 (TSIPPKT *rpt)
{
	unsigned char
		info, nsvs, sv_prn[32];
	short
		week_num, datum_index, sv_IODC[32];
	double
		lat, lon, alt, time_of_fix;
	double
		londeg, latdeg, vel[3];
	short
		isv;
   char
   	datum_string[20];

	/* unload rptbuf */
	if (rpt_0x8F20 (rpt,
		&info, &lat, &lon, &alt, vel,
		&time_of_fix,
		&week_num, &nsvs, sv_prn, sv_IODC, &datum_index))
	{
		parsed = BADLEN_PARSE;
		return;
	}
	pbuf += sprintf(pbuf,
   	"\nFix at: %04d:%3s:%02d:%02d:%06.3f GPS (=UTC+%2ds)  FixType: %s%s%s",
   	week_num,
		dayname[(short)(time_of_fix/86400.0)],
		(short)fmod(time_of_fix/3600., 24.),
		(short)fmod(time_of_fix/60., 60.),
		fmod(time_of_fix, 60.),
      (char)rpt->buf[29],		/* UTC offset */
		(info & INFO_DGPS)?"Diff":"",
		(info & INFO_2D)?"2D":"3D",
		(info & INFO_FILTERED)?"-Filtrd":"");

   if (datum_index > 0)
   {
		sprintf(datum_string, "Datum%3d", datum_index);
   }
   else if (datum_index)
   {
		sprintf(datum_string, "Unknown ");
   }
   else
   {
		sprintf(datum_string, "WGS-84");
   }

	/* convert from radians to degrees */
	latdeg = R2D * fabs(lat);
	londeg = R2D * fabs(lon);
	pbuf += sprintf(pbuf,
   	"\n   Pos: %4d:%09.6f %c %5d:%09.6f %c %10.2f m HAE (%s)",
		(short)latdeg, fmod (latdeg, 1.)*60.0,
		(lat<0.0)?'S':'N',
		(short)londeg, fmod (londeg, 1.)*60.0,
		(lon<0.0)?'W':'E',
		alt,
      datum_string);
	pbuf += sprintf(pbuf,
   	"\n   Vel:    %9.3f E       %9.3f N      %9.3f U   (m/sec)",
		vel[0], vel[1], vel[2]);

	pbuf += sprintf(pbuf,
   	"\n   SVs: ");
	for (isv = 0; isv < nsvs; isv++) {
		pbuf += sprintf(pbuf, " %02d", sv_prn[isv]);
	}
	pbuf += sprintf(pbuf, "     (IODEs:");
	for (isv = 0; isv < nsvs; isv++) {
		pbuf += sprintf(pbuf, " %02X", sv_IODC[isv]&0xFF);
	}
	pbuf += sprintf(pbuf, ")");
}

/* 0x8F41 */
static void rpt_8F41(TSIPPKT *rpt)
{
	unsigned char
   	bSearchRange,
		bBoardOptions,
		bBuildYear,
		bBuildMonth,
		bBuildDay,
		bBuildHour;
	float
   	fOscOffset;
	unsigned short
   	iTestCodeId;
	unsigned long
		iiSerialNumber;

   if (!rpt_0x8F41(rpt,
		&bSearchRange,
		&bBoardOptions,
		&iiSerialNumber,
		&bBuildYear,
		&bBuildMonth,
		&bBuildDay,
		&bBuildHour,
		&fOscOffset,
		&iTestCodeId))
   {
		parsed = BADLEN_PARSE;
      return;
   }

   pbuf += sprintf(pbuf, "\n  search range:          %d",
   	bSearchRange);
   pbuf += sprintf(pbuf, "\n  board options:         %d",
      bBoardOptions);
   pbuf += sprintf(pbuf, "\n  board serial #:        %ld",
      iiSerialNumber);
   pbuf += sprintf(pbuf, "\n  build date/hour:       %02d/%02d/%02d %02d:00",
   	bBuildDay, bBuildMonth, bBuildYear, bBuildHour);
   pbuf += sprintf(pbuf, "\n  osc offset:            %.3f PPM (%.0f Hz)",
   	fOscOffset/1575.42, fOscOffset);
   pbuf += sprintf(pbuf, "\n  test code:             %d",
   	iTestCodeId);
}

/* 0x8F42 */
static void rpt_8F42(TSIPPKT *rpt)
{
	unsigned char
   	bProdOptionsPre,
      bProdNumberExt;
	unsigned short
   	iCaseSerialNumberPre,
      iPremiumOptions,
      iMachineID,
      iKey;
	unsigned long
   	iiCaseSerialNumber,
		iiProdNumber;

   if (!rpt_0x8F42(rpt,
		&bProdOptionsPre,
		&bProdNumberExt,
		&iCaseSerialNumberPre,
		&iiCaseSerialNumber,
		&iiProdNumber,
		&iPremiumOptions,
		&iMachineID,
		&iKey))
   {
		parsed = BADLEN_PARSE;
      return;
   }

	pbuf += sprintf(pbuf, "\nProduct ID 8F42");
   pbuf += sprintf(pbuf, "\n   extension:            %d", bProdNumberExt);
   pbuf += sprintf(pbuf, "\n   case serial # prefix: %d", iCaseSerialNumberPre);
   pbuf += sprintf(pbuf, "\n   case serial #:        %ld", iiCaseSerialNumber);
   pbuf += sprintf(pbuf, "\n   prod. #:              %ld", iiProdNumber);
	pbuf += sprintf(pbuf, "\n   premium options:      %Xh", iPremiumOptions);
   pbuf += sprintf(pbuf, "\n   machine ID:           %d", iMachineID);
   pbuf += sprintf(pbuf, "\n   key:                  %Xh", iKey);
}

/* 0x8F45 */
static void rpt_8F45(TSIPPKT *rpt)
{
   unsigned char bSegMask;

   if (!rpt_0x8F45(rpt,
   	&bSegMask))
   {
		parsed = BADLEN_PARSE;
		return;
	}
	pbuf += sprintf(pbuf, "\nCleared Segment Mask: %Xh", bSegMask);
}

static void rpt_8F4A(TSIPPKT *rpt)
/* Stinger PPS def */
{
	unsigned char
   	pps_enabled,
      pps_timebase,
      pps_polarity;
   float
   	bias_unc_threshold;
   double
   	pps_offset;

  	if (rpt_0x8F4A_16 (rpt,
   	&pps_enabled,
      &pps_timebase,
      &pps_polarity,
      &pps_offset,
      &bias_unc_threshold))
   {
   	parsed = BADLEN_PARSE;
	   return;
   }

	pbuf += sprintf(pbuf, "\nPPS is         %s",	pps_enabled?"enabled":"disabled");
   pbuf += sprintf(pbuf, "\n   timebase:   %s", PPSTimeBaseText[pps_timebase]);
   pbuf += sprintf(pbuf, "\n   polarity:   %s", PPSPolarityText[pps_polarity]);
   pbuf += sprintf(pbuf, "\n   offset:     %.1f ns, ", pps_offset*1.e9);
   pbuf += sprintf(pbuf, "\n   biasunc:    %.1f ns", bias_unc_threshold/GPS_C*1.e9);
}

static void rpt_8F4B(TSIPPKT *rpt)
/* fast-SA decorrolation time for self-survey */
{
	unsigned long
   	decorr_max;

   if (rpt_0x8F4B(rpt, &decorr_max))
   {
		parsed = BADLEN_PARSE;
      return;
   }

   pbuf += sprintf(pbuf,
   	"\nMax # of position fixes for self-survey : %ld",
      decorr_max);
}

static void rpt_8F4D(TSIPPKT *rpt)
{
	static char
   	*linestart;
	unsigned long
   	OutputMask;
   static unsigned long
   	MaskBit[] = {
      	0x00000001, 0x00000002, 0x00000004, 0x00000008, 0x00000010, 0x00000020,
      	0x00000100L, 0x00000800L, 0x00001000L,
         0x40000000L, 0x80000000L};
   int
   	ichoice,
   	numchoices;

   if (rpt_0x8F4D(rpt, &OutputMask))
   {
		parsed = BADLEN_PARSE;
      return;
   }

   pbuf += sprintf(pbuf, "\nAuto-Report Mask: %02X %02X %02X %02X",
   	(unsigned char)(OutputMask>>24),
   	(unsigned char)(OutputMask>>16),
   	(unsigned char)(OutputMask>>8),
   	(unsigned char)OutputMask);

   numchoices = sizeof(MaskText)/sizeof(char*);
   pbuf += sprintf(pbuf, "\nAuto-Reports scheduled for Output:");
   linestart = pbuf;
   for (ichoice=0; ichoice<numchoices; ichoice++)
   {
   	if (OutputMask&MaskBit[ichoice])
      {
	     	pbuf += sprintf(pbuf, "%s %s",
   	   	(pbuf==linestart)?"\n     ":",",
      	   MaskText[ichoice]);
			if (pbuf-linestart > 60) linestart = pbuf;
      }
   }

   pbuf += sprintf(pbuf, "\nAuto-Reports NOT scheduled for Output:");
   linestart = pbuf;
   for (ichoice=0; ichoice<numchoices; ichoice++)
   {
   	if (OutputMask&MaskBit[ichoice]) continue;
	     	pbuf += sprintf(pbuf, "%s %s",
   	   	(pbuf==linestart)?"\n     ":",",
         MaskText[ichoice]);
		if (pbuf-linestart > 60) linestart = pbuf;
   }
}

static void rpt_8F4F(TSIPPKT *rpt)
{
    double PpsWidth = bGetDouble(&rpt->buf[1]);

   pbuf += sprintf(pbuf, "\nPPS pulse width: %f us", PpsWidth*1000*1000);
}

static void rpt_8FA5(TSIPPKT *rpt)
{
	unsigned char
   	spktmask[4];

   if (rpt_0x8FA5(rpt, spktmask))
   {
		parsed = BADLEN_PARSE;
      return;
   }

   pbuf += sprintf(pbuf, "\nSuperpacket auto-output mask: %02X %02X %02X %02X",
   	spktmask[0], spktmask[1], spktmask[2], spktmask[3]);

   if (spktmask[0]&0x01) pbuf+= sprintf (pbuf, "\n    PPS   8F-0B");
   if (spktmask[0]&0x02) pbuf+= sprintf (pbuf, "\n    Event 8F-0B");
   if (spktmask[0]&0x10) pbuf+= sprintf (pbuf, "\n    PPS   8F-AD");
   if (spktmask[0]&0x20) pbuf+= sprintf (pbuf, "\n    Event 8F-AD");
   if (spktmask[2]&0x01) pbuf+= sprintf (pbuf, "\n    ppos Fix 8F-20");
}

static void rpt_8FAD (TSIPPKT *rpt)
{
   unsigned short
    	Count,
    	Year;
   double
    	FracSec;
   unsigned char
    	Hour,
    	Minute,
    	Second,
    	Day,
    	Month,
    	Status,
    	Flags;
	static char* Status8FADText[] = {
      "CODE_DOING_FIXES",
      "CODE_GOOD_1_SV",
      "CODE_APPX_1SV",
      "CODE_NEED_TIME",
      "CODE_NEED_INITIALIZATION",
      "CODE_PDOP_HIGH",
      "CODE_BAD_1SV",
      "CODE_0SVS",
      "CODE_1SV",
      "CODE_2SVS",
      "CODE_3SVS",
      "CODE_NO_INTEGRITY",
      "CODE_DCORR_GEN",
      "CODE_OVERDET_CLK",
      "Invalid Status"},
    	*LeapStatusText[] = {
    	" UTC Avail", " ", " ", " ",
      " Scheduled", " Pending", " Warning", " In Progress"};
    int i;

	if (rpt_0x8FAD (rpt,
    	&Count,
    	&FracSec,
    	&Hour,
    	&Minute,
    	&Second,
    	&Day,
    	&Month,
    	&Year,
    	&Status,
    	&Flags))
   {
		parsed = BADLEN_PARSE;
		return;
   }

	pbuf += sprintf(pbuf,    "\n8FAD   Count: %d   Status: %s",
   	Count, Status8FADText[Status]);

  	pbuf += sprintf(pbuf, "\n   Leap Flags:");
   if (Flags)
   {
   	for (i=0; i<8; i++)
      {
      	if (Flags&(1<<i)) pbuf += sprintf(pbuf, LeapStatusText[i]);
      }
   }
   else
   {
   	pbuf += sprintf(pbuf, "  UTC info not available");
   }

	pbuf += sprintf(pbuf,     "\n      %02d/%02d/%04d (DMY)  %02d:%02d:%02d.%09ld UTC",
   		Day, Month, Year, Hour, Minute, Second, (long)(FracSec*1.e9));
}


int print_msg_table_header (int rptcode, char *HdrStr, int force)
{
	/* force header is to help auto-output function */
	/* last_rptcode is to determine whether to print a header */
	/* for the first occurence of a series of reports */
	static int
		last_rptcode = 0;
   int
   	numchars;

   numchars = 0;
	if (force || rptcode!=last_rptcode)
   {
		/* supply a header in console output */
   	switch (rptcode)
		{
		case 0x5A:
			numchars = sprintf(HdrStr, "\nRaw Measurement Data");
			numchars += sprintf(HdrStr+numchars,
      		"\n   SV  Sample   SNR  Code Phase   Doppler    Seconds     Time of Meas");
			break;

		case 0x5B:
			numchars = sprintf(HdrStr, "\nEphemeris Status");
			numchars += sprintf(HdrStr+numchars,
				"\n    SV     Time collected     Health  IODE        t oe         Fit   URA");
			break;

		case 0x5C:
			numchars = sprintf(HdrStr, "\nTracking Info");
			numchars += sprintf(HdrStr+numchars,
   	   	"\n   SV  C Acq Eph   SNR     Time of Meas       Elev  Azim   ");
			break;

      }
	}
	last_rptcode = rptcode;
   return (short)numchars;
}

static void unknown_rpt (TSIPPKT *rpt)
{
	int i;

	/* app-specific rpt packets */
	if (parsed == BADLEN_PARSE)
   {
		pbuf += sprintf(pbuf, "\nTSIP report packet ID %2Xh, length %d: Bad length",
      	rpt->code, rpt->len);
   }
	if (parsed == BADID_PARSE)
   {
		pbuf += sprintf(pbuf,
      	"\nTSIP report packet ID %2Xh, length %d: translation not supported",
   		rpt->code, rpt->len);
   }

	if (parsed == BADDATA_PARSE)
   {
		pbuf += sprintf(pbuf,
      	"\nTSIP report packet ID %2Xh, length %d: data content incorrect",
   		rpt->code, rpt->len);
   }

	for (i = 0; i < rpt->len; i++) {
		if ((i % 20) == 0) *pbuf++ = '\n';
		pbuf += sprintf(pbuf, " %02X", rpt->buf[i]);
	}
}
/**/
/*
** main subroutine, called from ProcessInputBytesWhileWaitingForKBHit()
*/
void TranslateTSIPReportToText (TSIPPKT *rpt, char *TextOutputBuffer)
{

	/* pbuf is the pointer to the current location of the text output */
	pbuf = TextOutputBuffer;

   /* keep track of whether the message has been successfully parsed */
	parsed = GOOD_PARSE;

	/* print a header if this is the first of a series of messages */
	pbuf += print_msg_table_header (rpt->code, pbuf, FALSE);

   /* process incoming TSIP report according to code */
	switch (rpt->code)
   {
	case 0x3D: rpt_chan_A_config (rpt); break;
	case 0x40: rpt_almanac_data_page (rpt); break;
	case 0x41: rpt_GPS_time (rpt); break;
	case 0x42: rpt_single_ECEF_position (rpt); break;
	case 0x43: rpt_single_ECEF_velocity (rpt); break;
	case 0x45: rpt_SW_version (rpt); break;
	case 0x46: rpt_rcvr_health (rpt); break;
	case 0x47: rpt_SNR_all_SVs (rpt); break;
	case 0x48: rpt_GPS_system_message (rpt); break;
	case 0x49: rpt_almanac_health_page (rpt); break;
	case 0x4A: switch (rpt->len) {
   	/*
      ** special case (=slip-up) in the TSIP protocol;
      ** parsing method depends on length
      */
   	case 20: rpt_single_lla_position (rpt); break;
      case  9: rpt_ref_alt (rpt); break;
		} break;
	case 0x4B: rpt_rcvr_id_and_status (rpt);break;
	case 0x4C: rpt_operating_parameters (rpt); break;
	case 0x4D: rpt_oscillator_offset (rpt); break;
	case 0x4E: rpt_GPS_time_set_response (rpt); break;
	case 0x4F: rpt_UTC_offset (rpt); break;
   case 0x54: rpt_1SV_bias (rpt); break;
	case 0x55: rpt_io_opt (rpt); break;
	case 0x56: rpt_ENU_velocity (rpt); break;
	case 0x57: rpt_last_fix_info (rpt); break;
	case 0x58: rpt_GPS_system_data (rpt); break;
	case 0x59: rpt_SVs_enabled (rpt); break;
	case 0x5A: rpt_raw_msmt (rpt); break;
	case 0x5B: rpt_SV_ephemeris_status (rpt); break;
	case 0x5C: rpt_SV_tracking_status (rpt); break;
	case 0x6D: rpt_allSV_selection (rpt); break;
	case 0x82: rpt_DGPS_position_mode (rpt); break;
	case 0x83: rpt_double_ECEF_position (rpt); break;
	case 0x84: rpt_double_lla_position (rpt); break;
	case 0xBB: rpt_complete_rcvr_config (rpt); break;
	case 0xBC: rpt_rcvr_serial_port_config (rpt); break;

	case 0x8F: switch (rpt->buf[0])
   	{
      /* superpackets; parsed according to subcodes */
      case 0x0B: rpt_8F0B(rpt); break;
      case 0x14: rpt_8F14(rpt); break;
      case 0x15: rpt_8F15(rpt); break;
		case 0x20: rpt_8F20(rpt); break;
      case 0x41: rpt_8F41(rpt); break;
      case 0x42: rpt_8F42(rpt); break;
      case 0x45: rpt_8F45(rpt); break;
      case 0x4A: rpt_8F4A(rpt); break;
      case 0x4B: rpt_8F4B(rpt); break;
      case 0x4D: rpt_8F4D(rpt); break;
      case 0x4F: rpt_8F4F(rpt); break;
      case 0xA5: rpt_8FA5(rpt); break;
 	   case 0xAD: rpt_8FAD(rpt); break;
		default: parsed = BADID_PARSE; break;
		}
		break;

	default: parsed = BADID_PARSE; break;
	}

	if (parsed != GOOD_PARSE)
	{
	   /*
   	**The message has TSIP structure (DLEs, etc.)
	   ** but could not be parsed by above routines
   	*/
		unknown_rpt (rpt);
	}

   /* close TextOutputBuffer */
   pbuf = '\0';
}




/*
 * *************************************************************************
 *
 * Trimble Navigation, Ltd.
 * OEM Products Development Group
 * P.O. Box 3642
 * 645 North Mary Avenue
 * Sunnyvale, California 94088-3642
 *
 * Corporate Headquarter:
 *    Telephone:  (408) 481-8000
 *    Fax:        (408) 481-6005
 *
 * Technical Support Center:
 *    Telephone:  (800) 767-4822	(U.S. and Canada)
 *                (408) 481-6940    (outside U.S. and Canada)
 *    Fax:        (408) 481-6020
 *    BBS:        (408) 481-7800
 *    e-mail:     trimble_support@trimble.com
 *		ftp://ftp.trimble.com/pub/sct/embedded/bin
 *
 * *************************************************************************
 *
 * -------  BYTE-SWAPPING  -------
 * TSIP is big-endian (Motorola) protocol.  To use on little-endian (Intel)
 * systems, the bytes of all multi-byte types (shorts, floats, doubles, etc.)
 * must be reversed.  This is controlled by the MACRO BYTESWAP; if defined, it
 * assumes little-endian protocol.
 * --------------------------------
 *
 * T_FORMAT.C and T_FORMAT.H contain primitive functions that prepare commands
 * to be transferred to the receiver.
 *
 * The module is in very portable, basic C language.  It can be used as is, or
 * with minimal changes if a TSIP communications application is needed separate
 * from TSIPCHAT. The construction of most argument lists avoid the use of
 * structures, but the developer is encouraged to reconstruct them using such
 * definitions to meet project requirements.  Declarations of TSIP_IFC.C
 * functions are included in TSIP_IFC.H to provide prototyping definitions.
 *
 * There are two types of functions:
 *
 * 1) send_cmd():  This command sends an assembled command in the TSIPKT
 * structure format thourgh the serial port using a user-supplied serial output
 * routine.  send_cmd adds all necessary control characters.
 *
 * 2) Command Functions:  Command functions come in the form cmd_0x??().
 * They translate a ser ies of command input parameters to a TSIPPKT structure.
 *
 * The functions cmd_0x??() are command string generators patterned after
 * the document called "Trimble Standard Interface Protocol" listed in
 * Appendix A of the "System Designer Reference Guide."  Examples of how to
 * enter the fields of these commands are given in TSIPCHAT routine
 * T_COMMND.C.  Because a TSIP command code may have different argument
 * lists depending on optional operating modes, there are more formatting
 * functions than command ID codes.  Each cmd_0x??() function corresponds to a
 * different functional mode for the command.  The following suffix convention
 * allows each function to serve a unique purpose and cleans up the argument
 * somewhat.
 *
 *             cmd_0x??q   -  query for current value
 * 	         cmd_0x??c   -  clear value (0x1D, 0x73 only)
 * 	         cmd_0x??s   -  set new values
 * 	         cmd_0x??e   -  enable
 * 	         cmd_0x??d   -  disable
 *
 *
 * *************************************************************************
 *
 */


/**/
void send_cmd (TSIPPKT *cmd, int sendb(unsigned char))
/* formats a command for sending to TSIP receiver */
{
   unsigned char *cbuf, *cbufend;

   /* start control code */
	sendb (DLE);
	sendb (cmd->code);
   cbufend = cmd->buf + cmd->len;
	for (cbuf = cmd->buf; cbuf < cbufend; cbuf++)
   {
		if (*cbuf == DLE)
      {
      	/* control code to prevent data byte alias */
      	sendb (DLE);
      }
		sendb (*cbuf);
	}
   /* end control code */
	sendb (DLE);
	sendb (ETX);
}

/*
 * Byte-reversal is necessary for little-endian (Intel-based) machines.
 * TSIP streams are Big-endian (Motorola-based).
 */
#ifdef BYTESWAP
static void bPutShort (void *in, unsigned char *out)
{
	unsigned char *inptr;

   inptr = (unsigned char*)in + 1;
   *out++ = *inptr--;
   *out = *inptr;
}

void bPutFloat (float *in, unsigned char *out)
{
	unsigned char *inptr;

   inptr = (unsigned char*)in + 3;
   *out++ = *inptr--;
   *out++ = *inptr--;
   *out++ = *inptr--;
   *out = *inptr;
}

static void bPutLong (long *in, unsigned char *out)
{
	unsigned char *inptr;

   inptr = (unsigned char*)in + 3;
   *out++ = *inptr--;
   *out++ = *inptr--;
   *out++ = *inptr--;
   *out = *inptr;
}

static void bPutULong (unsigned long *in, unsigned char *out)
{
	unsigned char *inptr;

   inptr = (unsigned char*)in + 3;
   *out++ = *inptr--;
   *out++ = *inptr--;
   *out++ = *inptr--;
   *out = *inptr;
}

static void bPutDouble (double *in, unsigned char *out)
{
	unsigned char *inptr;

   inptr = (unsigned char*)in + 7;
   *out++ = *inptr--;
   *out++ = *inptr--;
   *out++ = *inptr--;
   *out++ = *inptr--;
   *out++ = *inptr--;
   *out++ = *inptr--;
   *out++ = *inptr--;
   *out = *inptr;
}

#else	/* not BYTESWAP */

void bPutShort (short a, unsigned char *cmdbuf) {*(short*) cmdbuf = a;}
void bPutULong (long a, unsigned char *cmdbuf) 	{*(long*) cmdbuf = a;}
void bPutFloat (float a, unsigned char *cmdbuf) {*(float*) cmdbuf = a;}
void bPutDouble (double a, unsigned char *cmdbuf){*(double*) cmdbuf = a;}

#endif

/**/
void cmd_0x1Dc (TSIPPKT *cmd)
/* clear oscillator offset */
{
	cmd->buf[0] = 0x43;
	cmd->len = 1;
	cmd->code = 0x1D;
}

void cmd_0x1Ds  (TSIPPKT *cmd,
	float offset)
/* set oscillator offset */
{
	bPutFloat (&offset, &cmd->buf[0]);
	cmd->len = 4;
	cmd->code = 0x1D;
}

void cmd_0x1E (TSIPPKT *cmd, unsigned char reset_type)
/* clear battery back-up, then reset */
{
	cmd->buf[0] = reset_type;
	cmd->len = 1;
	cmd->code = 0x1E;
}

void cmd_0x1F (TSIPPKT *cmd)
/* request software versions */
{
	cmd->len = 0;
	cmd->code = 0x1F;
}

/**/
void cmd_0x20  (TSIPPKT *cmd, unsigned char sv_prn)
/* request almanac */
{
	cmd->buf[0] = sv_prn;
	cmd->len = 1;
	cmd->code = 0x20;
}

void cmd_0x21 (TSIPPKT *cmd)
/* request current time */
{
	cmd->len = 0;
	cmd->code = 0x21;
}

void cmd_0x22 (TSIPPKT *cmd, unsigned char nav_mode)
/* select position fix mode (2D, 3D, ...) */
{
	cmd->buf[0] = nav_mode;
	cmd->len = 1;
	cmd->code = 0x22;
}

void cmd_0x23  (TSIPPKT *cmd, float pos_ECEF[3])
/* initial position in ECEF coordinates */
{
	bPutFloat (&pos_ECEF[0], &cmd->buf[0]);
	bPutFloat (&pos_ECEF[1], &cmd->buf[4]);
	bPutFloat (&pos_ECEF[2], &cmd->buf[8]);
	cmd->len = 12;
	cmd->code = 0x23;
}

void cmd_0x24 (TSIPPKT *cmd)
/* request position fix mode */
{
	cmd->len = 0;
	cmd->code = 0x24;
}

void cmd_0x25 (TSIPPKT *cmd)
/* initiate soft reset & self-test (equivalent to power cycle) */
{
	cmd->len = 0;
	cmd->code = 0x25;
}

void cmd_0x26 (TSIPPKT *cmd)
/* request receiver health */
{
	cmd->len = 0;
	cmd->code = 0x26;
}

void cmd_0x27 (TSIPPKT *cmd)
/* request signal levels */
{
	cmd->len = 0;
	cmd->code = 0x27;
}

void cmd_0x28 (TSIPPKT *cmd)
/* request GPS system message */
{
	cmd->len = 0;
	cmd->code = 0x28;
}

void cmd_0x29 (TSIPPKT *cmd)
/* request almanac health */
{
	cmd->len = 0;
	cmd->code = 0x29;
}

void cmd_0x2Aq (TSIPPKT *cmd)
/* query altitude for 2-D mode */
{
	cmd->len = 0;
	cmd->code = 0x2A;
}

void cmd_0x2As  (TSIPPKT *cmd, float alt)
/* altitude for 2-D mode */
{
	bPutFloat (&alt, &cmd->buf[0]);
	cmd->len = 4;
	cmd->code = 0x2A;
}

void cmd_0x2Ad (TSIPPKT *cmd)
/* disable altitude for 2-D mode */
{
	cmd->buf[0] = 0xFF;
	cmd->len = 1;
	cmd->code = 0x2A;
}

void cmd_0x2B  (TSIPPKT *cmd, float lat, float lon, float alt)
/* initial position (latitude-longitude-altitude) */
{
	bPutFloat (&lat, &cmd->buf[0]);
	bPutFloat (&lon, &cmd->buf[4]);
	bPutFloat (&alt, &cmd->buf[8]);
	cmd->len = 12;
	cmd->code = 0x2B;
}

void cmd_0x2Cq (TSIPPKT *cmd)
/* request operating parameters and masks */
{
	float negative_one;
   negative_one = -1.0;
	cmd->buf[0] = 0;
	bPutFloat (&negative_one, &cmd->buf[1]);
	bPutFloat (&negative_one, &cmd->buf[5]);
	bPutFloat (&negative_one, &cmd->buf[9]);
	bPutFloat (&negative_one, &cmd->buf[13]);
	cmd->len = 17;
	cmd->code = 0x2C;
}

void cmd_0x2Cs (TSIPPKT *cmd,
	unsigned char dyn_code,
	float elev_mask, float snr,
	float pdop_mask, float pdop_switch)
/* set operating parameters and masks */
{
	cmd->buf[0] = dyn_code;
	bPutFloat (&elev_mask, &cmd->buf[1]);
	bPutFloat (&snr, &cmd->buf[5]);
	bPutFloat (&pdop_mask, &cmd->buf[9]);
	bPutFloat (&pdop_switch, &cmd->buf[13]);
	cmd->len = 17;
	cmd->code = 0x2C;
}

void cmd_0x2D (TSIPPKT *cmd)
/* request oscillator offset */
{
	cmd->len = 0;
	cmd->code = 0x2D;
}

void cmd_0x2E  (TSIPPKT *cmd, float time_of_week, short week_num)
/* set GPS time */
{
	bPutFloat (&time_of_week, &cmd->buf[0]);
	bPutShort (&week_num, &cmd->buf[4]);
	cmd->len = 6;
	cmd->code = 0x2E;
}

void cmd_0x2F (TSIPPKT *cmd)
/* request UTC params */
{
	cmd->len = 0;
	cmd->code = 0x2F;
}

/**/
void cmd_0x31  (TSIPPKT *cmd, float ECEF_pos[3])
/* initial accurate position in ECEF coordinates */
{
	bPutFloat (&ECEF_pos[0], &cmd->buf[0]);
	bPutFloat (&ECEF_pos[1], &cmd->buf[4]);
	bPutFloat (&ECEF_pos[2], &cmd->buf[8]);
	cmd->len = 12;
	cmd->code = 0x31;
}

void cmd_0x32  (TSIPPKT *cmd, float lat, float lon, float alt)
/* initial accurate position in latitude-longitude-altitude */
{
	bPutFloat (&lat, &cmd->buf[0]);
	bPutFloat (&lon, &cmd->buf[4]);
	bPutFloat (&alt, &cmd->buf[8]);
	cmd->len = 12;
	cmd->code = 0x32;
}

void cmd_0x34  (TSIPPKT *cmd, unsigned char sv_prn)
/* set SV for one-SV mode */
{
	cmd->buf[0] = sv_prn;
	cmd->len = 1;
	cmd->code = 0x34;
}

void cmd_0x35q  (TSIPPKT *cmd)
/* request serial I/O options */
{
	cmd->len = 0;
	cmd->code = 0x35;
}

void cmd_0x35s  (TSIPPKT *cmd, unsigned char pos_code, unsigned char vel_code,
	unsigned char time_code, unsigned char opts_code)
/* set serial I/O options */
{
	cmd->buf[0] = pos_code;
	cmd->buf[1] = vel_code;
	cmd->buf[2] = time_code;
	cmd->buf[3] = opts_code;
	cmd->len = 4;
	cmd->code = 0x35;
}

void cmd_0x37 (TSIPPKT *cmd)
/* request last position, velocity, and status */
{
	cmd->len = 0;
	cmd->code = 0x37;
}

void cmd_0x38q  (TSIPPKT *cmd, unsigned char data_type, unsigned char sv_prn)
/* request GPS system data (binary) */
{
	cmd->buf[0] = 1;
	cmd->buf[1] = data_type;
	cmd->buf[2] = sv_prn;
	cmd->len = 3;
	cmd->code = 0x38;
}

void cmd_0x38s  (TSIPPKT *cmd, unsigned char data_type, unsigned char sv_prn,
	unsigned char data_length, unsigned char *databuf)
/* load GPS system data (binary) */
{
	cmd->buf[0] = 2;
	cmd->buf[1] = data_type;
	cmd->buf[2] = sv_prn;
	cmd->buf[3] = data_length;
	memcpy (&cmd->buf[4], databuf, (short)data_length);
	cmd->len = (short)(data_length + 4);
	cmd->code = 0x38;
}

void cmd_0x39  (TSIPPKT *cmd, unsigned char op_code, unsigned char sv_prn)
/* set or request enable/health-heed status of satellites */
{
	cmd->buf[0] = op_code;
	cmd->buf[1] = sv_prn;
	cmd->len = 2;
	cmd->code = 0x39;
}

void cmd_0x3A  (TSIPPKT *cmd, unsigned char sv_prn)
/* request last code-phase/Doppler measurement */
{
	cmd->buf[0] = sv_prn;
	cmd->len = 1;
	cmd->code = 0x3A;
}

void cmd_0x3B  (TSIPPKT *cmd, unsigned char sv_prn)
/* request eph status */
{
	cmd->buf[0] = sv_prn;
	cmd->len = 1;
	cmd->code = 0x3B;
}

void cmd_0x3C  (TSIPPKT *cmd, unsigned char sv_prn)
/* request tracking status */
{
	cmd->buf[0] = sv_prn;
	cmd->len = 1;
	cmd->code = 0x3C;
}

void cmd_0x3Dq (TSIPPKT *cmd)
/* request Channel A configuration for dual-port operation */
{
	cmd->len = 0;
	cmd->code = 0x3D;
}

void cmd_0x3Ds (TSIPPKT *cmd,
	unsigned char baud_out, unsigned char baud_inp,
   unsigned char char_code, unsigned char stopbitcode,
   unsigned char output_mode, unsigned char input_mode)
/* set Channel A configuration for dual-port operation */
{
	cmd->buf[0] = baud_out;		/* XMT baud rate */
	cmd->buf[1] = baud_inp;		/* RCV baud rate */
	cmd->buf[2] = char_code;	   /* parity and #bits per byte */
	cmd->buf[3] = stopbitcode;	/* number of stop bits code */
	cmd->buf[4] = output_mode;	/* Ch. A transmission mode */
	cmd->buf[5] = input_mode;	/* Ch. A reception mode */
	cmd->len = 6;
	cmd->code = 0x3D;
}

/**/
void cmd_Paly0xBBs(TSIPPKT *cmd,
	TSIP_RCVR_CFG *TsipxBB)
	/* set Stinger receiver configuration */
{

	/* Palisade is inconsistent with other TSIP, which has a kength of 40 */
	/*	cmd->len		= 40; */
	cmd->len		= 43;
	cmd->code	= 0xBB;
   memset (cmd->buf, 0, cmd->len);	/* set reserved bytes to 0 */
	cmd->buf[0]	=	0;
	cmd->buf[1]	=	TsipxBB->operating_mode;
	cmd->buf[2]	=	TsipxBB->dgps_mode;
	cmd->buf[3]	=	TsipxBB->dyn_code;
	bPutFloat (&TsipxBB->elev_mask, &cmd->buf[5]);
	bPutFloat (&TsipxBB->cno_mask, &cmd->buf[9]);
	bPutFloat (&TsipxBB->dop_mask, &cmd->buf[13]);
	bPutFloat (&TsipxBB->dop_switch, &cmd->buf[17]);
	cmd->buf[21]	=	TsipxBB->dgps_age_limit;
}

void cmd_0xBBq (TSIPPKT *cmd,
	unsigned char subcode)
{

	cmd->len = 1;
	cmd->code = 0xBB;
	cmd->buf[0] = subcode;
}

void cmd_0xBCq  (TSIPPKT *cmd, unsigned char port_num)
/* request receiver port configuration */
{

	cmd->buf[0] = port_num;
	cmd->len = 1;
	cmd->code = 0xBC;
}

void cmd_0xBCs  (TSIPPKT *cmd,
	unsigned char port_num,
	unsigned char in_baud,
	unsigned char out_baud,
	unsigned char data_bits,
	unsigned char parity,
	unsigned char stop_bits,
	unsigned char flow_control,
	unsigned char protocols_in,
	unsigned char protocols_out,
	unsigned char reserved)
/* set receiver port configuration */
{

	cmd->buf[0] = port_num;
	cmd->buf[1] = in_baud;
	cmd->buf[2] = out_baud;
	cmd->buf[3] = data_bits;
	cmd->buf[4] = parity;
	cmd->buf[5] = stop_bits;
	cmd->buf[6] = flow_control;
	cmd->buf[7] = protocols_in;
	cmd->buf[8] = protocols_out;
	cmd->buf[9] = reserved;
	cmd->len = 10;
	cmd->code = 0xBC;
}

/**** Superpackets ****/
void cmd_0x8E0Bq (TSIPPKT *cmd)
/* 8E-0B to query 8F-0B controls */
{

	cmd->len = 1;
	cmd->code = 0x8E;
	cmd->buf[0] = 0x0B;
}

void cmd_0x8E0Bs (TSIPPKT *cmd, unsigned char mode)
/* 8E-0B to set 8F-0B controls */
{
	cmd->len = 2;
	cmd->code = 0x8E;
	cmd->buf[0] = 0x0B;
	cmd->buf[1] = mode;
}

void cmd_0x8E14s  (TSIPPKT *cmd, unsigned char datum_index)
/* datum set - standard parameters */
{

	cmd->buf[0] = 0x14;
	cmd->buf[1] = 0;	/* datum index is a short; this is always 0 */
	cmd->buf[2] = datum_index;
	cmd->len = 3;
	cmd->code = 0x8E;
}

void cmd_0x8E14s_2  (TSIPPKT *cmd, double datum_coeffs[5])
/* datum set - special parameters */
{

	cmd->buf[0] = 0x14;  /*Subcode 14 */
	bPutDouble (&datum_coeffs[0], &cmd->buf[1]);
	bPutDouble (&datum_coeffs[1], &cmd->buf[9]);
	bPutDouble (&datum_coeffs[2], &cmd->buf[17]);
	bPutDouble (&datum_coeffs[3], &cmd->buf[25]);
	bPutDouble (&datum_coeffs[4], &cmd->buf[33]);
	cmd->len = 41;
	cmd->code = 0x8E;
}

void cmd_0x8E15q  (TSIPPKT *cmd)
/* datum query */
{

	cmd->len = 1;
	cmd->code = 0x8E;
	cmd->buf[0] = 0x15;
}

void cmd_0x8E20q (TSIPPKT *cmd)
/* Superpacket 8F-20 query */
{

	cmd->len = 1;
	cmd->code = 0x8E;
	cmd->buf[0] = 0x20;
}

void cmd_0x8E20d (TSIPPKT *cmd)
/* 8F-20 msg auto-output disable */
{

	cmd->len = 2;
	cmd->code = 0x8E;
	cmd->buf[0] = 0x20;
	cmd->buf[1] = 0;
}

void cmd_0x8E20e (TSIPPKT *cmd)
/* 8F-20 msg auto-output enable */
{

	cmd->len = 2;
	cmd->code = 0x8E;
	cmd->buf[0] = 0x20;
	cmd->buf[1] = 1;
}

void cmd_0x8E41q (TSIPPKT *cmd)
/* 8F-41 to query board serial number */
{

	cmd->len = 1;
	cmd->code = 0x8E;
	cmd->buf[0] = 0x41;
}

void cmd_0x8E42q (TSIPPKT *cmd)
/* 8F-42 to query product serial number */
{

	cmd->len = 1;
	cmd->code = 0x8E;
	cmd->buf[0] = 0x42;
}

void cmd_0x8E45 (TSIPPKT *cmd,
	unsigned char segment)
/* 8F-45 to revert SEE segment to defaults */
{

	cmd->len = 2;
	cmd->code = 0x8E;
	cmd->buf[0] = 0x45;
	cmd->buf[1] = segment;
}

void cmd_0x8E4Aq (TSIPPKT *cmd)
/* 8F-4A to query PPS parameters */
{
	cmd->len = 1;
	cmd->code = 0x8E;
	cmd->buf[0] = 0x4A;
}

void cmd_0x8E4As (TSIPPKT *cmd,
	unsigned char PPSOnOff,
	unsigned char TimeBase,
	unsigned char Polarity,
   double PPSOffset,
   float Uncertainty)
{
	cmd->len = 16;
	cmd->code = 0x8E;
	cmd->buf[0] = 0x4A;
	cmd->buf[1] = PPSOnOff;
	cmd->buf[2] = TimeBase;
	cmd->buf[3] = Polarity;
	bPutDouble (&PPSOffset, &cmd->buf[4]);
	bPutFloat (&Uncertainty, &cmd->buf[12]);
}

void cmd_0x8E4Bq (TSIPPKT *cmd)
/* 8F-4B query survey limit */
{
	cmd->len = 1;
	cmd->code = 0x8E;
	cmd->buf[0] = 0x4B;
}

void cmd_0x8E4Bs (TSIPPKT *cmd,
	long SurveyLimit)
/* 8F-4B set survey limit */
{
	cmd->len = 5;
	cmd->code = 0x8E;
	cmd->buf[0] = 0x4B;
   bPutLong (&SurveyLimit, &cmd->buf[1]);
}

void cmd_0x8E4Dq (TSIPPKT *cmd)
/* 8F-4D to query auto-output mask */
{
	cmd->len = 1;
	cmd->code = 0x8E;
	cmd->buf[0] = 0x4D;
}

void cmd_0x8E4Ds (TSIPPKT *cmd,
	unsigned long AutoOutputMask)
{
	cmd->len = 5;
	cmd->code = 0x8E;
	cmd->buf[0] = 0x4D;
   bPutULong (&AutoOutputMask, &cmd->buf[1]);
}

void cmd_0x8E4Fq (TSIPPKT *cmd)
/* 8F-4F to query the PPS pulse width */
{
	cmd->len = 1;
	cmd->code = 0x8E;
	cmd->buf[0] = 0x4F;
}

void cmd_0x8E4Fs (TSIPPKT *cmd,
	double PpsWidth)
/* 8F-4F to set the PPS pulse width (PpsWidth in microseconds) */
{
	cmd->len = 9;
	cmd->code = 0x8E;
	cmd->buf[0] = 0x4F;
	PpsWidth /= 1000*1000;     /* convert to seconds */
	bPutDouble (&PpsWidth, &cmd->buf[1]);
}

void cmd_0x8EA5q (TSIPPKT *cmd)
/* 8F-A5 to query superpacker mask */
{
	cmd->len = 1;
	cmd->code = 0x8E;
	cmd->buf[0] = 0xA5;
}

void cmd_0x8EADq (TSIPPKT *cmd)
/* 8E-AD to query 8F-AD controls */
{
	cmd->len = 1;
	cmd->code = 0x8E;
	cmd->buf[0] = 0xAD;
}

void cmd_0x8EADs (TSIPPKT *cmd, unsigned char mode)
/* 8E-AD to set 8F-AD controls */
{
	cmd->len = 2;
	cmd->code = 0x8E;
	cmd->buf[0] = 0xAD;
	cmd->buf[1] = mode;
}

void cmd_0x8E  (TSIPPKT *cmd, unsigned char hexbyte[], unsigned char hexnum)
/* turn on superpacket, etc. */
{
	memcpy (cmd->buf, hexbyte, (short)hexnum);
	cmd->len = hexnum;
	cmd->code = 0x8E;
}

void cmd_0xxx  (TSIPPKT *cmd, unsigned char hexcode, unsigned char hexbyte[], unsigned char hexnum)
/* turn on superpacket, etc. */
{
	memcpy (cmd->buf, hexbyte, (short)hexnum);
	cmd->len = hexnum;
	cmd->code = hexcode;
}


