/*
 * vi:ts=4:sw=4:
 *
 * $Id: DataHandling.c,v 1.19 2006/11/07 14:45:32 ruben Exp $
 *
 * Copyright (c) 1998-2003 RIPE NCC
 * 
 * All Rights Reserved
 * 
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of the author not be used in advertising or 
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.
 * 
 * THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS; IN NO EVENT SHALL
 * AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY
 * DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE. 
 */

/* 
 * -------------------------------------------------------------------------------
 * Module Header
 * 
 * Filename          : DataHandling.c
 * Purpose           : Processes incoming UDP packets.
 * Author            : Olaf M. Kolkman
 * Date              : 1998/03/02
 * Language Version  : C
 * OSs Tested        : FreeBSD with customized kernel (getppstime interface call)
 * 
 * To Do             : open/close final outputfile only once, not at every write
 * 
 * Description       : Data Handling part of ControlDaemon
 * 
 *     The data received through the pipes is stored on two linked lists, one
 *     for each type of measurement. UDP data are read from the pipe that is the
 *     interface to the UDPProcessor daemon and are written to UCdataStore
 *     by PutOnUCStore().  ProcesUCStore() is used to scan the storage list,
 *     check for each UDP packet if matching IP fragments are there, if yes
 *     write output and remove both UDP and PC entries from the lists.
 *     A similar process takes place for the raw packets, they are added
 *     to the PCdataStore by the routine PutOnPCStore(); the routine
 *     ProcessPCStore() goes through the list and removes entries
 *     that haven't been picked up by ProcessUCStore  (i.e. the matching
 *     UDP packet was not recieved by UDPProcessor, mostly likely because
 *     it was not a delay measurement packet but a traceroute packet)
 * 
 * 
 * -------------------------------------------------------------------------------
 */

/*
 * -------------------------------------------------------------------------------
 * Module Change History
 * Author	    :  Rene Wilhelm
 * Date	      :  1999/12/01
 * Changes	   :  add rcsid string for tagging binaries
 * Changes	   :  open syslog connection with better identification string
 * Changes	   :  add sender's and receiver's NTP error estimates to output
 * Comments	  :
 * -------------------------------------------------------------------------------
 */

/*
 * -------------------------------------------------------------------------------
 * Module Change History
 * Author	    :  Rene Wilhelm
 * Date	      :  2001/07/19
 * Changes	   :  reformatted and slightly recoded for better readability
 * Comments	  :
 * -------------------------------------------------------------------------------
 */

/*
 * -------------------------------------------------------------------------------
 * Module Change History
 * Author	    :  Rene Wilhelm
 * Date	      :  2002/06/17
 * Changes	   :  frequent malloc problems prompted a reimplementation
 * 		  :  of the PCdataStore and UCdataStore with linked lists
 * 		  :  instead of large arrays
 * Comments	  :
 * -------------------------------------------------------------------------------
 */

/*
 * -------------------------------------------------------------------------------
 * Module Change History
 * Author	    :  Florian Frotzler
 * Date	      :  2002/09/23
 * Changes	   :  Rewriting of code to support IPv6.
 * 		  :  Now many IPv4 functions have an IPv6 counterpart.
 * 		  :  Datahandler is now reading from 4 pipes instead of 2.
 * Comments	  :
 * -------------------------------------------------------------------------------
 */


#include <sys/types.h>
#include <unistd.h>
#include <time.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <signal.h>
#include <limits.h>
#include <stdlib.h>
#include <syslog.h>
#include <netinet/udp.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "PacketCatcher.h"
#include "UDPProcessor.h"
#include "DataHandling.h"
#include "utils.h"

#define DH_FLAG_DELETE  0x01
#define max(X,Y) ((X)>(Y)?(X):(Y))


pid_t	   control_pid;
static char *trueoutfilename;
static char *trueoutfilename6;

static char const rcsid[] = "$Id: DataHandling.c,v 1.19 2006/11/07 14:45:32 ruben Exp $";

int                 PC_pipe4[2],
                    UC_pipe4[2],
                    PC_pipe6[2],
                    UC_pipe6[2];	/* pipes used for IPC */

/*
 * -------------------------------------------------------------------------------
 * Subroutine Header
 * 
 * Purpose      : Set up a daemon to handle data received from UC and PC pipes
 * Input	     : basename of outputfile; real output filename is contructed
 * 	       from this by appending date/time in YYYYMMDD-HHMMSS format.
 * -------------------------------------------------------------------------------
 */


int DataHandler (struct DelayProbeList *delayprobelist, char *outputfilename, char *outputfilename6)
{

	struct PCdataStore4List PC_Store4List = SLIST_HEAD_INITIALIZER(PC_Store4List);	/* linked list of PacketCatcher v4 data */
	struct UCdataStore4List UC_Store4List = SLIST_HEAD_INITIALIZER(UC_Store4List);	/* linked list of UDPProcessor v4 data */
	struct PCdataStore6List PC_Store6List = SLIST_HEAD_INITIALIZER(PC_Store6List);	/* linked list of PacketCatcher v6 data */
	struct UCdataStore6List UC_Store6List = SLIST_HEAD_INITIALIZER(UC_Store6List);	/* linked list of UDPProcessor v6 data */

	char                timestring[16];	/* YYYYMMDD-HHMMSS + '\0' --> 16 chars */
	time_t              tp;

	/* syslog settings */
	strcpy (programname, "DataHandler");
	setproctitle ("-%s", programname);
	closelog ();
	openlog (programname, LOG_PID, LOG_LOCAL0);


	/* initialisation */

	SLIST_INIT(&PC_Store4List);
	SLIST_INIT(&UC_Store4List);
	SLIST_INIT(&PC_Store6List);
	SLIST_INIT(&UC_Store6List);

#ifdef DEBUG
	printf ("Started %s\n", programname);
#endif


	/* construct real data output filename */

	time (&tp);
	trueoutfilename = malloc (strlen (outputfilename) + sizeof (timestring) + 1);
	trueoutfilename6 = malloc (strlen (outputfilename6) + sizeof (timestring) + 1);

	if ((trueoutfilename == NULL) && (trueoutfilename6 == NULL))
	{
		HandleError (programname, "FATAL: malloc outputfilename");
		exit (1);
	}

	strftime (timestring, sizeof (timestring), "%Y%m%d-%H%M%S", gmtime (&tp));
	sprintf (trueoutfilename, "%s.%s", outputfilename, timestring);
	sprintf (trueoutfilename6, "%s.%s", outputfilename6, timestring);

	/* sanity check */

	if ((max (UC_DATA_SIZE, PC_DATA_SIZE) > PIPE_BUF) || (max (UC_DATA_SIZE6, PC_DATA_SIZE6) > PIPE_BUF))
	{

		/* XXX Drawback of using pipes is that results ar unpredictable when the
		 * data send at one write becomes larger than PIPE_BUF (currently 1k
		 * bytes). If this error appears we should worry about changing our IPC 
		 * Mechanism */

		errno = 0;
		HandleError (programname, "WARNING: DATA communication may be non-atomic and result in errors");
	}

	/* 
	 *  Everything set, fork from MainControlDaemon and
	 *  let a subprocess handle the rest
	 */

	if (fork ())
	{
#ifdef DEBUG
		printf ("%s, Parent returns", programname);
#endif
		return (0);
	}

	/* Child continues */

	/* close the sending parts of the pipe descriptors */
	close (PC_pipe4[1]);
	close (UC_pipe4[1]);
	close (PC_pipe6[1]);
	close (UC_pipe6[1]);

#ifdef DEBUG
	printf (" %s, datahandling \n", programname);
#endif

	signal (SIGHUP, handle_signals);
	signal (SIGINT, handle_signals);
	signal (SIGQUIT, handle_signals);
	signal (SIGILL, handle_signals);
	signal (SIGTRAP, handle_signals);

	signal (SIGABRT, handle_signals);
	signal (SIGEMT, handle_signals);
	signal (SIGFPE, handle_signals);
	signal (SIGBUS, handle_signals);
	signal (SIGSEGV, handle_signals);

	signal (SIGSYS, handle_signals);
	signal (SIGPIPE, handle_signals);
	signal (SIGALRM, handle_signals);
	signal (SIGTERM, handle_signals);

	signal (SIGXCPU, handle_signals);
	signal (SIGXFSZ, handle_signals);
	signal (SIGVTALRM, handle_signals);

	signal (SIGPROF, handle_signals);
	signal (SIGUSR1, handle_signals);
	signal (SIGUSR2, handle_signals);

	/* Actual Handling */

	while (1)
	{
		/* read from the pipe and perform an action depending on the type of
		 * the message received */

		int                 nbytesread;
		int                 ip_flag;
		int                 found_flag;
		fd_set              FdSet;
		struct udpportdata4 UCdata4;
		struct udpportdata6 UCdata6;
		struct packetcatchdata4 PCdata4;
		struct packetcatchdata6 PCdata6;
		struct delayprobetarget *target;

		/* process Store at least every 5 seconds */
		struct timeval      timeout = { 5, 0 };

		/* async IO. */
		FD_ZERO (&FdSet);
		FD_SET (PC_pipe4[0], &FdSet);
		FD_SET (UC_pipe4[0], &FdSet);
		FD_SET (PC_pipe6[0], &FdSet);
		FD_SET (UC_pipe6[0], &FdSet);

		/* 
		 *  UDPProcessor and PacketCatcher put their measurements
		 *  on the pipes.  Below we wait untill they arive and put
		 *  them on a data array for further processing.
		 */

#ifdef DEBUG
		printf (" %s, Block... \n", programname);
#endif
		ip_flag = -1;

		/* block until a read is possible or a signal is cought */
		if (select (max (max (PC_pipe4[0], UC_pipe4[0]), max (PC_pipe6[0], UC_pipe6[0])) + 1, &FdSet, NULL, NULL, &timeout) == -1)
		{
			HandleError (programname, "select(2) error");
		}


		/* Received data from PacketCatcher */

		/* IPv4 */
		if (FD_ISSET (PC_pipe4[0], &FdSet))
		{
			if ((nbytesread = read (PC_pipe4[0], &PCdata4, PC_DATA_SIZE)) != -1)
			{
				if (nbytesread != PC_DATA_SIZE)
				{
					/* forget about incomplete data, do not mention EOF */
					if (nbytesread)
						HandleError (programname, "Error partial IPv4 data read from PacketCatcher, expected %d, got %d", PC_DATA_SIZE, nbytesread);
				}
				else
				{
					found_flag = 0;
					SLIST_FOREACH(target, delayprobelist, entries)
					{
						if ((target->family == AF_INET) && (strcmp (PCdata4.source_addr, target->hostname) == 0))
						{
							found_flag = 1;
						}
					}
					if (found_flag)
					{

						/* store data */
						PutOnPCStore (&PC_Store4List, &PCdata4);
						ip_flag = V4_FLAG;
					}
				}
			}
			else
			{
				/* Something is totally wrong on the other end of the pipe, better bail out of here and restart */
				HandleError (programname, "Fatal: Error while reading IPv4 data from PacketCatcher");
				kill(control_pid, SIGTERM);
				exit(1);
			}
		}

		/* IPv6 */
		if (FD_ISSET (PC_pipe6[0], &FdSet))
		{
			if ((nbytesread = read (PC_pipe6[0], &PCdata6, PC_DATA_SIZE6)) != -1)
			{
				if (nbytesread != PC_DATA_SIZE6)
				{
					/* forget about incomplete data */
					if (nbytesread)
						HandleError (programname, "Error partial IPv6 data read from PacketCatcher, expected %d, got %d", PC_DATA_SIZE6, nbytesread);
				}
				else
				{
					found_flag = 0;
					SLIST_FOREACH(target, delayprobelist, entries)
					{
						if ((target->family == AF_INET6) && (strcmp (PCdata6.source_addr, target->hostname) == 0))
						{
							found_flag = 1;
						}
					}
					if (found_flag)
					{
						/* store data */
						PutOnPCStore6 (&PC_Store6List, &PCdata6);
						ip_flag = V6_FLAG;
					}
				}
			}
			else
			{
				/* Something is totally wrong on the other end of the pipe, better bail out of here and restart */
				HandleError (programname, "Fatal: Error while reading IPv6 data from PacketCatcher");
				kill(control_pid, SIGTERM);
				exit(1);
			}
		}


		/* Received Data from UDPProcessor */

		/* IPv4 */
		if (FD_ISSET (UC_pipe4[0], &FdSet))
		{
			if ((nbytesread = read (UC_pipe4[0], &UCdata4, UC_DATA_SIZE)) != -1)
			{
				if (nbytesread != UC_DATA_SIZE)
				{
					/* forget about incomplete data, do not mention EOF */
					if (nbytesread)
						HandleError (programname, "Error partial IPv4 data read from UDPProcessor, expected %d, got %d", UC_DATA_SIZE, nbytesread);
				}
				else
				{
					/* store data */
					PutOnUCStore (&UC_Store4List, &UCdata4);
					ip_flag = V4_FLAG;
				}
			}
			else
			{
				/* Something is totally wrong on the other end of the pipe, better bail out of here and restart */
				HandleError (programname, "Fatal: Error while reading IPv4 data from UDPProcessor");
				kill(control_pid, SIGTERM);
				exit(1);
			}
		}

		/* IPv6 */
		if (FD_ISSET (UC_pipe6[0], &FdSet))
		{
			if ((nbytesread = read (UC_pipe6[0], &UCdata6, UC_DATA_SIZE6)) != -1) 
			{
				if (nbytesread != UC_DATA_SIZE6)
				{
					/* forget about incomplete data, do not mention EOF */
					if (nbytesread)
						HandleError (programname, "Error partial IPv6 data read from UDPProcessor, expected %d, got %d", UC_DATA_SIZE6, nbytesread);
				}
				else
				{
					/* store data */
					PutOnUCStore6 (&UC_Store6List, &UCdata6);
					ip_flag = V6_FLAG;
				}
			}
			else
			{
				/* Something is totally wrong on the other end of the pipe, better bail out of here and restart */
				HandleError (programname, "Fatal: Error while reading IPv6 data from UDPProcessor ", UC_DATA_SIZE, nbytesread);
				kill(control_pid, SIGTERM);
				exit(1);
			}
		}



		/* work through the 2 * 2 data stores, decide on IP version */

		if (ip_flag == V4_FLAG)
		{
			ProcessUCStore (&UC_Store4List, &PC_Store4List);
			ProcessPCStore (&PC_Store4List);
		}
		else if (ip_flag == V6_FLAG)
		{
			ProcessUCStore6 (&UC_Store6List, &PC_Store6List);
			ProcessPCStore6 (&PC_Store6List);
		}
	}
	return (0);
}


/*
 * -------------------------------------------------------------------------------
 * Subroutine Header
 * 
 * Purpose      : Add new PacketCatcher data to the beginning of PC data list
 * Input	     : Store -  pointer to PCdataStore
 * 	       indata - pointer to data received from PacketCatcher
 * Comments     : first element of Store is a dummy, to ease insertion/removal
 * 	     : PutOnPCStore for IPv4, PutOnPCStore6 for IPv6
 * -------------------------------------------------------------------------------
 */

void PutOnPCStore (struct PCdataStore4List *Store, struct packetcatchdata4 *indata)
{

	struct PCdataStore4 *new;

	new = (struct PCdataStore4 *) malloc (sizeof (struct PCdataStore4));

	if (new == NULL)
	{

		/* Memory allocation failed, give up and signal Main ControlDaemon to
		 * restart all processes. */

		errno = 0;
		HandleError (programname, "Fatal PC_Store4: Memory allocation failed");
		kill (control_pid, SIGHUP);

	}

	new->data = *indata;		/* assignment = implicit copy */
	
	SLIST_INSERT_HEAD(Store, new, entries);
}

void PutOnPCStore6 (struct PCdataStore6List *Store, struct packetcatchdata6 *indata)
{

	struct PCdataStore6 *new;

	new = (struct PCdataStore6 *) malloc (sizeof (struct PCdataStore6));

	if (new == NULL)
	{

		/* Memory allocation failed, give up and signal Main ControlDaemon to
		 * restart all processes. */

		errno = 0;
		HandleError (programname, "Fatal PC_Store6: Memory allocation failed");
		kill (control_pid, SIGHUP);

	}

	new->data = *indata;		/* assignment = implicit copy */

	SLIST_INSERT_HEAD(Store, new, entries);
}


/*
 * -------------------------------------------------------------------------------
 * Subroutine Header
 * 
 * Purpose      : Process stored PacketCatcher data
 * Input	     :
 * Description  : This function only checks if data on the array has been
 * 	       sitting there and timed out. It marks these for removal and
 * 	       writes the data to output (using a 0 valued UDP data entry)
 * Comments     : ProcessPCStore for IPv4, ProcessPCStore6 for IPv6
 * 	     : Contrary to the v4 version lost fragments are never
 * 	     : sent to WriteOutput.
 * -------------------------------------------------------------------------------
 */

int ProcessPCStore (struct PCdataStore4List *Store)
{

	struct timeval      now;

	struct PCdataStore4 *current,
	                   *previous;
	struct udpportdata4 Upacket;

	bzero (&Upacket, UC_DATA_SIZE);	/* empty UDPpacket for output */

	gettimeofday (&now, NULL);


	/* 
	 *  Packets that sit on the queue for more than the preconfigured
	 *  number of seconds, have apparently not been picked up by the
	 *  UDP processor and can be removed
	 */

	SLIST_FOREACH_SAFE(current,Store,entries,previous)
	{

		if (current->data.bpftime.tv_sec < now.tv_sec - IPF_NOT_PROCESSED_TIMEOUT)
		{
			/* 
			 *  Uncomplete fragment, removed from list and
			 *  written to output. The Upacket has been set
			 *  to zero to indicate there is no UDP data.
			 */

			struct PCdataStore4List PCfragments_tmp = SLIST_HEAD_INITIALIZER(PCfragments_tmp);

			SLIST_INIT(&PCfragments_tmp);

			SLIST_REMOVE(Store, current, PCdataStore4, entries);
			SLIST_INSERT_HEAD(&PCfragments_tmp,current,entries);

			WriteOutput (Upacket, &PCfragments_tmp);
			free (current);

			/* No need to clean up PCfragments_tmp... it is an auto var limited for this scope only */
		}
	}
	return (0);
}


int ProcessPCStore6 (struct PCdataStore6List *Store)
{

	struct timeval      now;

	struct PCdataStore6 *current,
	                   *previous;
	struct udpportdata6 Upacket;

	bzero (&Upacket, UC_DATA_SIZE6);

	gettimeofday (&now, NULL);


	/* 
	 *  Packets that sit on the queue for more than the preconfigured
	 *  number of seconds, have apparently not been picked up by the
	 *  UDP processor and can be removed
	 */

	SLIST_FOREACH_SAFE(current,Store,entries,previous)
	{
		if (current->data.bpftime.tv_sec < now.tv_sec - IPF_NOT_PROCESSED_TIMEOUT)
		{
			/* 
			 *  Lost fragment, remove from list
			 */

			SLIST_REMOVE(Store, current, PCdataStore6, entries);
			free (current);
		}
	}
	return (0);
}


/*
 * -------------------------------------------------------------------------------
 * Subroutine Header
 * 
 * Purpose      : Add new UDPProcessor data to the beginning of UC data list
 * Input	     : Store -  pointer to UCdataStore
 * 	       indata - pointer to data received from UDPProcessor list
 * Comments     : first element of Store is a dummy, to ease insertion/removal
 * 	     : PutOnUCStore for IPv4, PutOnUCStore6 for IPv6
 * -------------------------------------------------------------------------------
 */

void PutOnUCStore (struct UCdataStore4List *Store, struct udpportdata4 *indata)
{

	struct UCdataStore4 *new;

	new = (struct UCdataStore4 *) malloc (sizeof (struct UCdataStore4));

	if (new == NULL)
	{

		/* Memory allocation failed, give up and signal Main ControlDaemon to
		 * restart all processes. */

		errno = 0;
		HandleError (programname, "Fatal UC_Store4: Memory allocation failed");
		kill (control_pid, SIGHUP);

	}

	new->data = *indata;		/* assignment = implicit copy */

	SLIST_INSERT_HEAD(Store, new, entries);
}


void PutOnUCStore6 (struct UCdataStore6List *Store, struct udpportdata6 *indata)
{

	struct UCdataStore6 *new;

	new = (struct UCdataStore6 *) malloc (sizeof (struct UCdataStore6));

	if (new == NULL)
	{

		/* Memory allocation failed, give up and signal Main ControlDaemon to
		 * restart all processes. */

		errno = 0;
		HandleError (programname, "Fatal UC_Store6: Memory allocation failed");
		kill (control_pid, SIGHUP);

	}

	new->data = *indata;		/* assignment = implicit copy */

	SLIST_INSERT_HEAD(Store, new, entries);
}


/*
 * -------------------------------------------------------------------------------
 * Subroutine Header
 * 
 * Purpose      : Process stored UDPProcessor data
 * Input	     :
 * Description  : for all stored UDPProcessor data, this routine searches the
 * 	       PacketCatcher Storage to locate the fragment(s) that make up
 * 	       the UDP packet; all relevant information is written to output
 * 	       and packets are marked for removal.
 * Comments     : ProcessUCStore for IPv4, ProcessUCStore6 for IPv6
 * 	     : Contrary to the v4 version all validity checks are made here,
 * 	     : thus invalid data is never sent to WriteOutput.
 * 	     : Added check on source address
 * -------------------------------------------------------------------------------
 */

int ProcessUCStore (struct UCdataStore4List *UCStore, struct PCdataStore4List *PCStore)
{

	struct timeval      now;

	/* pointers for list traversal */

	struct UCdataStore4 *currentUC,
	                   *previousUC;
	struct PCdataStore4 *currentPC,
	                   *previousPC,
	                   *lastfragment=NULL;

	struct PCdataStore4List PCfragments = SLIST_HEAD_INITIALIZER(PCfragments);

	enum
	{
		nothing_found,
		first_fragment_found,
		all_fragments_found
	}
	found;

	SLIST_INIT(&PCfragments);


	if (PCStore == NULL)
	{
		errno = 0;
		HandleError (programname, "ERROR: PCStore = NULL in ProcessUCStore");
	}
	if (UCStore == NULL)
	{
		errno = 0;
		HandleError (programname, "ERROR: UCStore = NULL in ProcessUCStore");
	}

	gettimeofday (&now, NULL);

	/* Walk through UDP list and for each packet get the packet's id (NOT the
	 * number which in data analysis is refered to as 'packet id'; that is
	 * called the packet's serial number in this context!!)
	 * 
	 * Scan all the fragments for the first fragment of the UDP packet that
	 * contains the same packet_id. Use this fragment to get the ip_id and then 
	 * scan the fragment array again to find all the fragment with this ip_id.
	 * Write the UDP packet data and all relevant fragment data to the output. 
	 */

	SLIST_FOREACH_SAFE(currentUC,UCStore,entries,previousUC)
	{

		u_long              ip_id;
		long                totalsize;

		/* scan for packet id (only available in the first packet) can be used
		 * to couple to the IP packet number so that we can identify all the
		 * fragments that belong to a certain measurement. The SendPacket code
		 * uses random numbers as packet ids, so the chance of finding packets
		 * from different sources with the same id is small. 
		 */

		/* How does this work in case of duplicate packets ids? [RW] */

		ip_id = 0;
		found = nothing_found;

		SLIST_FOREACH(currentPC,PCStore,entries)
		{
			/* check for packet_id and source address */

			if (currentPC->data.packet_id == currentUC->data.packet_id)
			{
				if (strcmp (currentPC->data.source_addr, currentUC->data.source_addr) == 0)
				{
					ip_id = currentPC->data.packet_id;
					found = first_fragment_found;
					break;
				}
			}
		}

		totalsize = 0;

		if (found == first_fragment_found)
		{
			/* Scan again and now use the IP packet number to identify the
			 * fragments; move fragments to seperate PCdataStore list */

			SLIST_FOREACH_SAFE(currentPC,PCStore,entries,previousPC)
			{
				if (currentPC->data.packet_id == ip_id)
				{
					totalsize += currentPC->data.ip_length;

					/* remove from PCStore */
					SLIST_REMOVE(PCStore, currentPC, PCdataStore4, entries);

					/* append to PCfragments */
					if (lastfragment == NULL)
					{
						/* PCfragments is empty, populate it*/
						SLIST_INSERT_HEAD(&PCfragments, currentPC, entries);
					}
					else
					{
						SLIST_INSERT_AFTER(lastfragment,currentPC,entries);
					}
					lastfragment = currentPC;
				}

				if (totalsize == currentUC->data.datasize + (long) UDP_HEADER_SIZE)
				{
					/* located all fragments */
					found = all_fragments_found;
					break;
				}
			}
		}

		if (found == all_fragments_found)
		{
			/* 
			 * All the PC packets are here, output relevant
			 * data, then delete delete the IP fragments and mark
			 * the corresponding UDP packet for deletion.
			 */

			WriteOutput (currentUC->data, &PCfragments);

			/* Delete PCfragments */

			while (!SLIST_EMPTY(&PCfragments)) 
			{           
				/* List Deletion. */
				currentPC = SLIST_FIRST(&PCfragments);
				SLIST_REMOVE_HEAD(&PCfragments, entries);
				free(currentPC);
			}

			/* Probably NULL */
			lastfragment = SLIST_FIRST(&PCfragments);

			/* Remove from UCStore */
			SLIST_REMOVE(UCStore, currentUC, UCdataStore4, entries);
			free (currentUC);
		}
		else
		{
			/* 
			 * Not all packet fragments are here yet;
			 * push PCfragments back to PCStore list
			 */

			lastfragment = SLIST_FIRST(&PCfragments);

			SLIST_FOREACH_SAFE(currentPC,&PCfragments,entries,previousPC)
			{
				SLIST_REMOVE(&PCfragments, currentPC, PCdataStore4, entries);
				SLIST_INSERT_HEAD(PCStore, currentPC, entries);
			}

			if (currentUC->data.packet_data.timepars.current_time.tv_sec < now.tv_sec - UDP_NOT_PROCESSED_TIMEOUT)
			{

				/* UDP packet too old, remove from store */
				WriteOutput (currentUC->data, NULL);

				SLIST_REMOVE(UCStore, currentUC, UCdataStore4, entries);
				free (currentUC);
			}
		}
	}							/* end over UDP packets */
	return (0);
}


int ProcessUCStore6 (struct UCdataStore6List *UCStore, struct PCdataStore6List *PCStore)
{

	struct timeval      now;

	/* pointers for list traversal */

	struct UCdataStore6 *currentUC,
	                   *previousUC;
	struct PCdataStore6 *currentPC,
	                   *previousPC;
	u_long              ip_id;


	if (PCStore == NULL)
	{
		errno = 0;
		HandleError (programname, "ERROR: PCStore = NULL in ProcessUCStore");
	}
	if (UCStore == NULL)
	{
		errno = 0;
		HandleError (programname, "ERROR: UCStore = NULL in ProcessUCStore");
	}

	gettimeofday (&now, NULL);

	SLIST_FOREACH_SAFE(currentUC,UCStore,entries,previousUC)
	{

		ip_id = 0;

		SLIST_FOREACH_SAFE(currentPC,PCStore,entries,previousPC)
		{
			/* check for packet_id and source address */

			if (currentPC->data.packet_id == currentUC->data.packet_id)
			{
				if (strcmp (currentPC->data.source_addr, currentUC->data.source_addr) == 0)
				{
					ip_id = currentPC->data.packet_id;

					if ((currentPC->data.p_length) != (currentUC->data.datasize + UDP_HEADER_SIZE))
					{
						/* 
						 * Data is obviously truncated or bogus, this should not happen...
						 * Output an error and remove from list
						 */

						HandleError (programname, "ProcessUCStore6: PC and UC payloads don't match!");

						SLIST_REMOVE(PCStore, currentPC, PCdataStore6, entries);
						free (currentPC);
					}
					else
					{
						/* Write to output and remove from list */
						struct PCdataStore6List PCfragments_tmp = SLIST_HEAD_INITIALIZER(PCfragments_tmp);

						SLIST_INIT(&PCfragments_tmp);

						SLIST_INSERT_HEAD(&PCfragments_tmp,currentPC,entries);

						WriteOutput6 (currentUC->data, &PCfragments_tmp);

						SLIST_REMOVE(PCStore, currentPC, PCdataStore6, entries);
						free (currentPC);
					}
					break;
				}
			}
		}

		if (ip_id)
		{
			SLIST_FOREACH_SAFE(currentPC,PCStore,entries,previousPC)
			{
				/* 
				 * This check is just to be sure that there is not more than 1 packet
				 * in the store with the same ID.
				 * Note: There is no fragmentation (except the client enforces it) in IPv6,
				 * so there must not be another fragment in the store.
				 */

				if (currentPC->data.packet_id == ip_id)
				{
					/* XXX Should be only outputted once */
					HandleError (programname, "Found more than 1 IPv6 fragment in PC Store! source %s", currentPC->data.source_addr);

					SLIST_REMOVE(PCStore, currentPC, PCdataStore6, entries);
					free (currentPC);
				}
			}
		}

		/* UDP packet too old, remove from store */
		if (currentUC->data.packet_data.timepars.current_time.tv_sec < now.tv_sec - UDP_NOT_PROCESSED_TIMEOUT)
		{
			SLIST_REMOVE(UCStore, currentUC, UCdataStore6, entries);
			free (currentUC);
		}

		ip_id = 0;
	}
	return (0);
}


/*
 * -------------------------------------------------------------------------------
 * Subroutine Header
 * 
 * Purpose      : write data to output file
 * Input	     :
 * Comments     : WriteOutput for IPv4, WriteOutput6 for IPv6
 * 	     : WriteOutput6 writes the new version 3 of the output file
 * 
 *  This is the interface function with the outside world....
 *  Write all the data to wherever we want it to be in whichever format
 *  we prefer
 * 
 *  Remember that all the data is in network byte order.. You should use
 *  ntohl, ntohs etc to transform before writing it.
 * 
 *  Hmm, wouldn't it be more efficient to open/close trueoutfilename
 *  at higher level in the code and only do a flush() here? [RW]
 * 
 * -------------------------------------------------------------------------------
 */

int WriteOutput (struct udpportdata4 UDPpacket, struct PCdataStore4List *IPfragments)
{
	FILE               *final_output;


	final_output = fopen (trueoutfilename, "a");

	if (final_output == NULL)
	{
		/* 
		 *  can't open output file, signal MainControlDaemon to
		 *  terminate all child processes (will start new DataHandler)
		 */
		HandleError (programname, "FATAL: Error opening the outputfile %s", trueoutfilename);
		kill (control_pid, SIGTERM);
	}

	/* Get the interface data for output */

	/* no idea how to do this when 2NICs receive packets luckily it's only
	 * #ifdef DEBUG output here [RW] */


	if (UDPpacket.dest_port == 0 && UDPpacket.source_port == 0)
	{

		/* UDPpacket set to null, IP fragmented and not all fragments received */
#ifdef DEBUG
		printf ("%s: Didn't receive all fragments (in time)\n", programname);
#endif

	}
	else
	{
		/* all fragments are available */

#ifdef DEBUG
		printf ("UDP:  0X%08lX %s:%u> %s:%u, 0X%08lX %ld.%06ld %ld.%06ld 0X%04X %06ld %06ld %06ld\n",
				ntohl (UDPpacket.packet_id), UDPpacket.source_addr,
				ntohs (UDPpacket.source_port), UDPpacket.dest_addr,
				ntohs (UDPpacket.dest_port),
				ntohl (UDPpacket.packet_data.fingerprint),
				UDPpacket.packet_data.timepars.current_time.tv_sec,
				UDPpacket.packet_data.timepars.current_time.tv_usec,
				UDPpacket.packet_data.timepars.last_pps.tv_sec,
				UDPpacket.packet_data.timepars.last_pps.tv_usec,
				UDPpacket.packet_data.timepars.pps_status,
				UDPpacket.packet_data.timepars.pll_offset,
				UDPpacket.packet_data.timepars.maxerror, UDPpacket.packet_data.timepars.esterror);
#endif

		if (IPfragments != NULL)
		{
			struct timeval      corin;
			struct timeval      corout;	/* corrected in and output */
			struct timeval      delay;

			struct PCdataStore4 *currentfrag, *lastfragment=NULL;

			/* walk through list, remember last fragment */

			SLIST_FOREACH(currentfrag,IPfragments,entries)
			{
#ifdef DEBUG
				printf (" |PC:0X%08lX   0X%08lX  0X%08X %ld,%06ld  %ld,%06ld 0X%04X %06ld %06ld %06ld\n",
						ntohl (UDPpacket.packet_id),
						ntohl (currentfrag->data.packet_id),
						currentfrag->data.ip_id,
						currentfrag->data.bpftime.tv_sec,
						currentfrag->data.bpftime.tv_usec,
						currentfrag->data.gps_times.last_pps.tv_sec,
						currentfrag->data.gps_times.last_pps.tv_usec,
						currentfrag->data.gps_times.pps_status,
						currentfrag->data.gps_times.pll_offset,
						currentfrag->data.gps_times.maxerror, currentfrag->data.gps_times.esterror);
#endif
				if (lastfragment)
				{
					/* time of last fragment smaller than of current ? make it the last fragment */
					if (timercmp(&(lastfragment->data.bpftime),&(currentfrag->data.bpftime),<))
						lastfragment = currentfrag;
				}
				else
					lastfragment = currentfrag;
			}

			/* correct both times for PPS offsets in sending & receiving box */

			corout = pps_diff_cor (lastfragment->data.bpftime, lastfragment->data.gps_times.last_pps);
			corin = pps_diff_cor (UDPpacket.packet_data.timepars.current_time, UDPpacket.packet_data.timepars.last_pps);

			ncc_timersub(&corout,&corin,&delay);

#ifdef DEBUG
			printf (" %ld ", ntohl (UDPpacket.packet_data.serial_number));
			printf (" %s ", timevaltoa (corout));
			printf (" (%s|%s) ", timevaltoa (lastfragment->data.bpftime), timevaltoa(lastfragment->data.gps_times.last_pps));
			printf (" %s ", timevaltoa (corin));
			printf (" (%s|%s) ", timevaltoa (UDPpacket.packet_data.timepars.current_time), timevaltoa(UDPpacket.packet_data.timepars.last_pps));
			printf ("XXXX %s \n", timevaltoa (delay));

#endif

			/* write output line */

			/* 12 is number of words to follow on output line */
			/* 2 is version number of the output */

			fprintf (final_output, "RCDP 12 2 ");

			fprintf (final_output, "%s %u %s %u ",
					 UDPpacket.source_addr, ntohs (UDPpacket.source_port), UDPpacket.dest_addr, ntohs (UDPpacket.dest_port));

			/* Corrected receive time */
			fprintf (final_output, "%s ", timevaltoa (corout));

			/* Corrected delay */
			fprintf (final_output, "%s ", timevaltoa (delay));

			/* sending side clock status */
			fprintf (final_output, "0X%X ", UDPpacket.packet_data.timepars.pps_status);
			/* receiving side clock status */
			fprintf (final_output, "0X%X ", lastfragment->data.gps_times.pps_status);

			/* serial number as set by sender */
			fprintf (final_output, "%u ", ntohl (UDPpacket.packet_data.serial_number));

			/* sending side NTP estimated error */
			fprintf (final_output, "%11.6f", UDPpacket.packet_data.timepars.esterror / 1e6);

			/* receiving side NTP estimated error */
			fprintf (final_output, "%11.6f", lastfragment->data.gps_times.esterror / 1e6);

			fprintf (final_output, "\n");

			fflush (NULL);
		}
		else
		{
			/* 
			 * UDP data without PacketCatcher fragments
			 * Most likely traffic send from a host which is not
			 * listed in our configfile and thus doesn't pass
			 * the filter set up by in PacketCatcher.
			 *
			 * Output host/port to syslog, other than that ignore
			 */

			errno = 0;
			/* XXX Should be only outputted once */
			HandleError (programname, "UDP data without fragments; source: %s", UDPpacket.source_addr);
		}

	}							/* end all fragments are available */


	fclose (final_output);
	return (0);
}


int WriteOutput6 (struct udpportdata6 UDPpacket, struct PCdataStore6List* IPfragments)
{

	FILE               *final_output;

	struct timeval      corin;
	struct timeval      corout;	/* corrected in and output */
	struct timeval      delay;

	/* 
	 * Although we don't do any IPv6 fragment handling in here, just accomodate
	 * the code for it so it more or less resembles WriteOutput() 
	 */

	struct PCdataStore6 *PCfragment = SLIST_FIRST(IPfragments);

	final_output = fopen (trueoutfilename6, "a");

	if (final_output == NULL)
	{
		/* 
		 *  can't open output file, signal MainControlDaemon to
		 *  terminate all child processes (will start new DataHandler)
		 */
		HandleError (programname, "FATAL: Error opening the outputfile %s", trueoutfilename);
		kill (control_pid, SIGTERM);
	}

	corout = pps_diff_cor (PCfragment->data.bpftime, PCfragment->data.gps_times.last_pps);
	corin = pps_diff_cor (UDPpacket.packet_data.timepars.current_time, UDPpacket.packet_data.timepars.last_pps);

	ncc_timersub(&corout,&corin,&delay);

	/* write output line */

	/* 12 is number of words to follow on output line */
	/* 3 is version number of the output */

	fprintf (final_output, "RCDP6 12 3 ");

	fprintf (final_output, "%s %u %s %u ",
			 UDPpacket.source_addr, ntohs (UDPpacket.source_port), UDPpacket.dest_addr, ntohs (UDPpacket.dest_port));

	/* Corrected receive time */
	fprintf (final_output, "%s ", timevaltoa (corout));

	/* Corrected delay */
	fprintf (final_output, "%s ", timevaltoa (delay));

	/* sending side clock status */
	fprintf (final_output, "0X%X ", UDPpacket.packet_data.timepars.pps_status);

	/* receiving side clock status */
	fprintf (final_output, "0X%X ", PCfragment->data.gps_times.pps_status);

	/* serial number as set by sender */
	fprintf (final_output, "%u ", ntohl (UDPpacket.packet_data.serial_number));

	/* sending side NTP estimated error */
	fprintf (final_output, "%11.6f", UDPpacket.packet_data.timepars.esterror / 1e6);

	/* receiving side NTP estimated error */
	fprintf (final_output, "%11.6f", PCfragment->data.gps_times.esterror / 1e6);

	fprintf (final_output, "\n");

	fflush (NULL);

	fclose (final_output);

	return (0);
}
