/*
 * vi:ts=4:sw=4:
 *
 * $Id: UDPProcessor.c,v 1.13 2006/07/04 13:41:12 ruben Exp $
 *
 * Copyright (c) 1998-2003 RIPE NCC
 * 
 * All Rights Reserved
 * 
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of the author not be used in advertising or 
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.
 * 
 * THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS; IN NO EVENT SHALL
 * AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY
 * DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE. 
 */

/* 
 * -------------------------------------------------------------------------------
 * Module Header
 * 
 * Filename          : UDPProcessor.c
 * Purpose           : Processes incoming UDP packets.
 * Author            : Olaf M. Kolkman
 * Date              : 1998/03/02
 * Description       : called by the control_deamon.c
 *                     Sets up a deamon to listen to incoming UDP packets.
 *                     When packets arive payload is analyzed and data is send
 *                     using  UCdataOutput(struct udpportdata UCdata)
 * Language Version  : C
 * OSs Tested        : FreeBSD  with customized kernel that includes
 *                     gettimeofday system call
 * To Do             : Dynamic memory allocation to remove hard limit on #ports?
 * ------------------------------------------------------------------------------- 
 * 
 */

/* 
 * -------------------------------------------------------------------------------
 * Module Change History
 * Author      :  Rene Wilhelm
 * Date    :  1999/11/25
 * Changes     :  add rcsid string for tagging binaries
 * Changes     :  open syslog connection with better identification string
 * Comments   :
 * -------------------------------------------------------------------------------
 */

/* 
 * -------------------------------------------------------------------------------
 * Module Change History
 * Author      :  Rene Wilhelm
 * Date    :  2001/07/19
 * Changes    :  MAX_TT_HOSTS renamed to MAX_UDP_PORTS
 * Changes     :  reformatted and slightly recoded for better readability
 * Comments   :
 * -------------------------------------------------------------------------------
 */

/* 
 * -------------------------------------------------------------------------------
 * Module Change History
 * Author     :  Rene Wilhelm
 * Date       :  2001/12/27
 * Changes     :  install signal handlers to catch otherwise fatal signals
 * Changes     :  action is to send SIGHUP to Main ControlDaemon process
 * Changes    :  so fresh client processes will be started
 * Comments   :
 * -------------------------------------------------------------------------------
 */

/* 
 * -------------------------------------------------------------------------------
 * Module Change History
 * Author     :  Florian Frotzler
 * Date       :  2002/09/23
 * Changes     :  Rewriting to support IPv6
 * Changes     :  UDPProcessor now has an IPv4 and an IPv6 pipe to Datahandler
 * Comments   :
 * -------------------------------------------------------------------------------
 */


#define PNAME "UDPProcessor"

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <signal.h>
#include <string.h>
#include <netinet/in.h>
#include <unistd.h>
#include <limits.h>
#include <arpa/inet.h>
#include <netinet/udp.h>
#include <errno.h>
#include <syslog.h>

#include "control_daemon.h"
#include "UDPProcessor.h"
#include "DataHandling.h"
#include "utils.h"
#include "if_utils.h"
#include "packet.h"

static int          sockfd;		/* socket file descriptor */
static struct sockaddr_in serv_addr;	/* IPv4 socket address structure */
static struct sockaddr_in6 serv_addr6;	/* IPv6 socket address structure */

static char const   rcsid[] = "$Id: UDPProcessor.c,v 1.13 2006/07/04 13:41:12 ruben Exp $";

/* 
 * -------------------------------------------------------------------------------
 * Subroutine Header
 * 
 * Purpose      : Set up a daemon for each port to process incoming UDP packets.
 * Side Effects : Initializes globals "serv_addr", "serv_addr6" and "sockfd"
 * Input        : Argument buffer consisting of hostname port combinations.
 *                family argument (1 = IPv4 only, 2 = IPv6 only, 3 = IPv4 + IPv6 (1 & 2)
 * Returns      : Return 0 on success or non-0 on failure.
 * -------------------------------------------------------------------------------
 */
int UDPProcessor (struct DelayProbeList *delayprobelist)
{

	int                 j;
	int                 uniqueport4[MAX_UDP_PORTS];	/* list with all used IPv4
													 * UDP ports */
	int                 uniqueport6[MAX_UDP_PORTS];	/* list with all used IPv6
													 * UDP ports */

	char                errorbuffer[_POSIX2_LINE_MAX];
	struct etherifdata  EtherIfData;
	struct delayprobetarget *current;

	bzero ((void *) &EtherIfData, sizeof (struct etherifdata));

	/* syslog settings */
	closelog ();
	openlog (PNAME, LOG_PID, LOG_LOCAL0);


	/* Iterate through all available interfaces */

	while (NextEthernetIFInfo (&EtherIfData, errorbuffer, ETHIFINFO_ALIASES_TOO) == 0)
	{
		/* initalize list of UDP ports used */
		for (j = 0; j < MAX_UDP_PORTS; j++)
		{
			uniqueport4[j] = 0;
			uniqueport6[j] = 0;
		}

		/* determine the IP addresses of the Ethernet interface... */
		if (GetEthernetIFInfo (&EtherIfData, errorbuffer))
		{
			HandleError (PNAME, "Probing Ethernet4/6 failed; %s", errorbuffer);
			exit (1);
		}

		/* 
		 *  Process individual host/port combinations
		 *  For each unique IPv4 and IPv6 port a daemon will be started
		 */


		SLIST_FOREACH(current, delayprobelist, entries)
		{
			int                 n,
			                    m,
			                    portisused;

			n = 0;
			m = 0;
			portisused = 0;

			if (current->family == AF_INET)
			{
				while (!portisused && uniqueport4[n])
				{

					/* check if port already used */
					if (current->portnumber == uniqueport4[n])
					{
						portisused = 1;
					}

					if (++n >= MAX_UDP_PORTS)
					{
						HandleError (PNAME, "Max number of UDP ports exceeded");
						break;
					}
				}
			}
			if (current->family == AF_INET6)
			{
				while (!portisused && uniqueport6[n])
				{

					/* check if port already used */
					if (current->portnumber == uniqueport6[m])
					{
						portisused = 1;
					}

					if (++m >= MAX_UDP_PORTS)
					{
						HandleError (PNAME, "Max number of UDP ports exceeded");
						break;
					}
				}
			}

			if (portisused)
			{
				/* already listening to this port */
				continue;
			}

			/* n or m is at the first unset array element */
			if (current->family == AF_INET)
				uniqueport4[n] = current->portnumber;
			if (current->family == AF_INET6)
				uniqueport6[m] = current->portnumber;

			/* 
			 *  Set up a udpserver on the port. If bind fails it may
			 *  only do that because the address is allready in use...
			 */


			/* Need a configured address on AF_INET interface */
			if (current->family == AF_INET && EtherIfData.ethif_addr.s_addr != INADDR_ANY)
			{

				if ((sockfd = socket (AF_INET, SOCK_DGRAM, 0)) < 0)
				{
					HandleError (PNAME, "server: can't open v4-datagram socket");
					exit (1);
				}

				bzero ((char *) &serv_addr, sizeof (serv_addr));
				serv_addr.sin_port = htons (current->portnumber);
				serv_addr.sin_family = AF_INET;

				/* use our local address per default */
				serv_addr.sin_addr.s_addr = EtherIfData.ethif_addr.s_addr;
				if (bind (sockfd, (struct sockaddr *) &serv_addr, sizeof (serv_addr)) < 0)
				{
					if (errno == EADDRINUSE)
					{
						HandleError (PNAME, "WARNING Could not bind to port %d", current->portnumber);
					}
					else
					{
						HandleError (PNAME, "FATAL: bind4 failed\n");
						exit (1);
					}
				}
				else
				{
					/* We have now successfully bound to a socket, let's become 
					 * a daemon process */

					if (fork () == 0)
					{
						/* child */

						sprintf (programname, "UDPProcess port %s:%d@%s family IPv4", EtherIfData.ifaddr ,ntohs (serv_addr.sin_port), EtherIfData.ifname);
						setproctitle ("-%s", programname);

						/* close the unneeded pipes... (global from
						 * DataHandling.h) */

						close (PC_pipe4[1]);
						close (PC_pipe4[0]);
						close (PC_pipe6[1]);
						close (PC_pipe6[0]);

						close (UC_pipe4[0]);
						close (UC_pipe6[0]);

						/* All processing done in HandlePort */
						HandleSpecificPort (AF_INET);

						HandleError (programname, "FATAL UDPdaemon returned");
						exit (1);	/* child */
					}
					/* parent continues with next host/port combo */
				}
			}

			/* Need a configured address on AF_INET interface */
			if (current->family == AF_INET6 && (!IN6_IS_ADDR_UNSPECIFIED (&(EtherIfData.ethif_addr6))))
			{

				if ((sockfd = socket (AF_INET6, SOCK_DGRAM, 0)) < 0)
				{
					HandleError (PNAME, "server: can't open v6-datagram socket");
					exit (1);
				}

				bzero ((char *) &serv_addr6, sizeof (serv_addr6));
				serv_addr6.sin6_port = htons (current->portnumber);
				serv_addr6.sin6_family = AF_INET6;

				serv_addr6.sin6_addr = EtherIfData.ethif_addr6;
				if (bind (sockfd, (struct sockaddr *) &serv_addr6, sizeof (serv_addr6)) < 0)
				{
					if (errno == EADDRINUSE)
					{
						HandleError (PNAME, "WARNING Could not bind to port %d", current->portnumber);
					}
					else
					{
						HandleError (PNAME, "FATAL: bind6 failed\n");
						exit (1);
					}
				}
				else
				{
					if (fork () == 0)
					{
						/* child */

						sprintf (programname, "UDPProcess port [%s]:%d@%s family IPv6",
								 EtherIfData.ifaddr, ntohs (serv_addr6.sin6_port), EtherIfData.ifname);
						setproctitle ("-%s", programname);

						/* close the unneeded pipes... (global from
						 * DataHandling.h) */

						close (PC_pipe4[1]);
						close (PC_pipe4[0]);
						close (PC_pipe6[1]);
						close (PC_pipe6[0]);

						close (UC_pipe4[0]);
						close (UC_pipe6[0]);

						/* All processing done in HandlePort */
						HandleSpecificPort (AF_INET6);

						HandleError (programname, "FATAL UDPdaemon returned");
						exit (1);	/* child */
					}
				}
			}
			/* parent continues with next host/port combo */
		}
		/* parent continues with next interface */
	}
	/* XXX we might want to know if NextEthernetIFInfo has failed... */
	EndEthernetIFInfo ();
	return (0);
}

/* 
 * -------------------------------------------------------------------------------
 * Subroutine Header
 * 
 * Purpose     : Process all packets arriving at a specific UDP port
 * 
 * Description : Once the UDP daemon has bound to a port all the handling of
 *               the incoming data is done by this routine. At arrival of a
 *               new packet the daemon forks so that the parent can wait for
 *               the next packet without delay.
 * 
 *               The received data is written to the Datahandling daemon
 *               using the UCdataOutput interface routine.
 * 
 * Comments    : Signal handling is set up in control_daemon.
 * Input       : AF_INET or AF_INET6 to handle specific IP family
 * Returns     : the proces should never return.
 *               (except for signals, see control_daemon.c)
 * 
 * 
 * -------------------------------------------------------------------------------
 */

void HandleSpecificPort (int family)
{

	struct sockaddr_in  sending_host;
	struct sockaddr_in6 sending_host6;

	int                 sending_host_size = sizeof (sending_host);
	int                 sending_host_size6 = sizeof (sending_host6);

	long                bytes_received;
	char                messagebuffer[TEST_PACKET_MAXLENGTH + 1];	/* From
																	 * packet.h 
																	 */
	char               *cp;

	/* 
	 * Start listening to the socket as soon as a packet is received...
	 * fork and process it. Start listening to the next packet immediatly
	 */

	signal (SIGHUP, handle_signals);
	signal (SIGINT, handle_signals);
	signal (SIGQUIT, handle_signals);
	signal (SIGILL, handle_signals);
	signal (SIGTRAP, handle_signals);

	signal (SIGABRT, handle_signals);
	signal (SIGEMT, handle_signals);
	signal (SIGFPE, handle_signals);
	signal (SIGBUS, handle_signals);
	signal (SIGSEGV, handle_signals);

	signal (SIGSYS, handle_signals);
	signal (SIGPIPE, handle_signals);
	signal (SIGALRM, handle_signals);
	signal (SIGTERM, handle_signals);

	signal (SIGXCPU, handle_signals);
	signal (SIGXFSZ, handle_signals);
	signal (SIGVTALRM, handle_signals);

	signal (SIGPROF, handle_signals);
	signal (SIGUSR1, handle_signals);
	signal (SIGUSR2, handle_signals);


	while (1)
	{
		struct udpportdata4 UCoutput;
		struct udpportdata6 UCoutput6;

		struct test_packet_data *test_data;
		struct test_packet_header *test_header;

		/* empty message buffer */
		bzero (messagebuffer, TEST_PACKET_MAXLENGTH + 1);

		if (family == AF_INET)
		{

			/* wait for a packet */
			bytes_received = recvfrom (sockfd, messagebuffer,
									   TEST_PACKET_MAXLENGTH, 0, (struct sockaddr *) &sending_host,
									   (socklen_t *) & sending_host_size);

			if (bytes_received < 0)
			{
				HandleError (programname, "Non-Fatal: reading from v4-socket");
				continue;
			}

			/* we have some data */
			if (fork () != 0)
			{
				/* Parent process */

				/* Install signal handler to reap the child, then continue
				 * listening for new packets */

				signal (SIGCHLD, sig_child);
				continue;
			}

			/* Child process, handle the received packet */

			/* information about the sending host. */
			strcpy (UCoutput.source_addr, inet_ntoa (sending_host.sin_addr));
			UCoutput.source_port = sending_host.sin_port;

			/* information about the receiving host. */
			strcpy (UCoutput.dest_addr, inet_ntoa (serv_addr.sin_addr));
			UCoutput.dest_port = serv_addr.sin_port;

			setproctitle ("-%s (client %s:%d)", programname, UCoutput.source_addr, UCoutput.dest_port);

			test_header = (struct test_packet_header *) messagebuffer;

			/* some sanity checks, this need not be a test traffic packet */
			if (
#ifndef OFFSET_TO_DATA_IS_UNSIGNED
				   (test_header->offset_to_data < 0) ||
#endif
				   (ntohs (test_header->offset_to_data) > bytes_received - TEST_PACKET_DATALENGTH - TEST_PACKET_HEADERLENGTH))
			{
				/* This must be a bogus packet, ignore */
				exit (1);
			}

			cp = messagebuffer + (size_t) ntohs (test_header->offset_to_data) + TEST_PACKET_HEADERLENGTH;
			UCoutput.datasize = bytes_received;
			test_data = (struct test_packet_data *) cp;

			/* the actual data is now fetched */
			UCoutput.packet_data = *test_data;
			UCoutput.packet_id = htonl (test_header->packet_id);

#ifdef DEBUG
			printf ("UC IPv4 Packet ID: %ld\n", UCoutput.packet_id);
#endif

			/* Pass data to DataHandler */
			UCdataOutput (UCoutput);

			exit (0);			/* child process */
		}

		if (family == AF_INET6)
		{

			char                ip_addr[INET6_ADDRSTRLEN];

			bytes_received = recvfrom (sockfd, messagebuffer,
									   TEST_PACKET_MAXLENGTH, 0, (struct sockaddr *) &sending_host6,
									   (socklen_t *) & sending_host_size6);

			if (bytes_received < 0)
			{
				HandleError (programname, "Non-Fatal: reading from v6-socket");
				continue;
			}

			if (fork () != 0)
			{
				signal (SIGCHLD, sig_child);
				continue;
			}

			inet_ntop (AF_INET6, &sending_host6.sin6_addr, ip_addr, INET6_ADDRSTRLEN);

			strcpy (UCoutput6.source_addr, ip_addr);
			UCoutput6.source_port = sending_host6.sin6_port;

			bzero (ip_addr, INET6_ADDRSTRLEN);
			inet_ntop (AF_INET6, &serv_addr6.sin6_addr, ip_addr, INET6_ADDRSTRLEN);

			strcpy (UCoutput6.dest_addr, ip_addr);
			UCoutput6.dest_port = serv_addr6.sin6_port;

			setproctitle ("-%s (client [%s]:%d)", programname, UCoutput6.source_addr, UCoutput6.dest_port);

			test_header = (struct test_packet_header *) messagebuffer;

			/* some sanity checks, this need not be a test traffic packet */
			if (
#ifndef OFFSET_TO_DATA_IS_UNSIGNED
				   (test_header->offset_to_data < 0) ||
#endif
				   (ntohs (test_header->offset_to_data) > bytes_received - TEST_PACKET_DATALENGTH - TEST_PACKET_HEADERLENGTH))
			{
				/* This must be a bogus packet, ignore */
				exit (1);
			}

			cp = messagebuffer + (size_t) ntohs (test_header->offset_to_data) + TEST_PACKET_HEADERLENGTH;
			UCoutput6.datasize = bytes_received;
			test_data = (struct test_packet_data *) cp;

			/* the actual data is now fetched */
			UCoutput6.packet_data = *test_data;
			UCoutput6.packet_id = htonl (test_header->packet_id);

#ifdef DEBUG
			printf ("UC IPv6 Packet ID: %ld\n", UCoutput6.packet_id);
#endif

			/* Pass data to DataHandler */
			UCdataOutput6 (UCoutput6);

			exit (0);			/* child process */
		}
	}
	/* Should not reach this stage */
	return;
}

/* 
 * -------------------------------------------------------------------------------
 * Subroutine Header
 * Purpose           : Interface function to DataHandler()
 * Input parameters  : UCdata   - upd packet data
 * Description       : output results from HandleSpecificPort to the UC IPv4 or IPv6 pipe
 *                     (which has been created in main controldaemon)
 * -------------------------------------------------------------------------------
 */

int UCdataOutput (struct udpportdata4 UCdata)
{
	if (write (UC_pipe4[1], &UCdata, UC_DATA_SIZE) == -1)
	{
		HandleError (programname, "UC4 write to pipe");
	}
	return (0);
}

int UCdataOutput6 (struct udpportdata6 UCdata)
{
	if (write (UC_pipe6[1], &UCdata, UC_DATA_SIZE6) == -1)
	{
		HandleError (programname, "UC6 write to pipe");
	}
	return (0);
}
