/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: pll.c,v 1.3 2005/01/14 16:23:46 tom Exp $
 *
 * PLL handling.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <err.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/queue.h>
#include <sys/timex.h>
#include <sys/time.h>
#include <math.h>

#include <refclocksync.h>
#include "pll.h"

int clock_step;
int pll_locked;

#define BILLION 1000000000

static struct timespec
evTimeSpec(struct timeval tv) {
        struct timespec ts;

        ts.tv_sec = tv.tv_sec;
        ts.tv_nsec = tv.tv_usec * 1000;
        return (ts);
}

static struct timespec
evConsTime(time_t sec, long nsec) {
        struct timespec x;

        x.tv_sec = sec;
        x.tv_nsec = nsec;
        return (x);
}

static struct timespec
evUTCTime() {
        struct timeval now;
        struct timespec tsnow;
        if (clock_gettime(CLOCK_REALTIME, &tsnow) == 0)
                return (tsnow);
        if (gettimeofday(&now, NULL) < 0)
                return (evConsTime(0, 0));
        return (evTimeSpec(now));
}

static struct timespec
evAddTime(struct timespec addend1, struct timespec addend2) {
        struct timespec x;

        x.tv_sec = addend1.tv_sec + addend2.tv_sec;
        x.tv_nsec = addend1.tv_nsec + addend2.tv_nsec;
        if (x.tv_nsec >= BILLION) {
                x.tv_sec++;
                x.tv_nsec -= BILLION;
        }
        return (x);
}

static struct timespec
evSubTime(struct timespec minuend, struct timespec subtrahend) {
        struct timespec x;

        x.tv_sec = minuend.tv_sec - subtrahend.tv_sec;
        if (minuend.tv_nsec >= subtrahend.tv_nsec)
                x.tv_nsec = minuend.tv_nsec - subtrahend.tv_nsec;
        else {
                x.tv_nsec = BILLION - subtrahend.tv_nsec + minuend.tv_nsec;
                x.tv_sec--;
        }
        return (x);
}

static void
dump_timex(struct timex *tx)
{

	Log(LOG_PLL | LOG_INFO, 
	 "pll: modes %x status %x offset %ld freq %.3e maxerror %ld esterror %ld constant %ld prec %ld tol %ld",
	    tx->modes, tx->status, tx->offset, tx->freq / (65536.0 * 1e6),
	    tx->maxerror, tx->esterror, tx->constant, tx->precision,
	    tx->tolerance);
}

// Adjusts the kernel clock frequency. Input is the frequency offset
// that is to be applied with ntp_adjtime. Unit is seconds (gets converted
// to scaled PPM inside by multiplying with 65536 and 1000000)
void kern_freq(double f)
{
	struct timex tx;
	int i;
	static double acc = 0;

	memset(&tx, 0, sizeof tx);
	tx.modes = MOD_NANO | MOD_FREQUENCY | MOD_OFFSET | MOD_STATUS; // TODO why MOD_OFFSET?
	tx.status = STA_PLL;

	/*
	 * To avoid bang-bang regulation at high integration times we
	 * implement a simple Floyd-Steinberg dither of the frequency
	 * correction:  Whatever fraction we did not apply carries over
	 * to the next second.  We also need to make sure that we do
	 * not get in trouble with the +/- 500PPM hard limits of frequency
	 * adjustability.
	 */
	acc += f * (65536.0 * 1e6); // 65536.0 is 16 bits which is SHIFT_USEC. Scaled PPM = PPM << SHIFT_USEC.
	                            // 1e6 is the factor needed to get a value to PPM (=1e6)
	tx.freq = round(acc);         // tx.freq is in scaled PPM
	if (tx.freq > 500 * 65536)
		acc = tx.freq = 500 * 65536;
	if (tx.freq < -500 * 65536)
		acc = tx.freq = -500 * 65536;
	acc -= tx.freq;

#if 0
	/* XXX: why does this cause divergence ?? */
	if (pll_locked) {
		tx.modes = MOD_MAXERROR | MOD_ESTERROR;
		tx.maxerror = sys_dispersion * 1e-9;
		tx.esterror = sys_dispersion * 1e-9;
	} else {
		tx.modes = MOD_MAXERROR | MOD_ESTERROR;
		tx.maxerror = 1e9;
		tx.esterror = 1e9;
	}
#endif
	i = ntp_adjtime(&tx);
	if (i != 0)
		Log(LOG_PLL | LOG_ERROR, "pll: error: ntp_adjtime: %s", strerror(errno));
	dump_timex(&tx);

#if 0
	if(fp==NULL)
	{
		fp=fopen("/tmp/freq.log", "w");
	}
	if(fp!=NULL)
	{
		fprintf(fp, "%f\n", tx.freq / (65536.0 * 1e6));
		fflush(fp);
	}
#endif
}

void
pllstep(double d)
{
	struct timespec ts;
	int i;

	d = -d;
	Log(LOG_PPS | LOG_INFO, "pll: PLL step %.09f", d);
	if (d < 0) {
		d = -d;
		ts.tv_sec = d;
		d -= ts.tv_sec;
		ts.tv_nsec = d * 1e9;
		ts = evSubTime(evUTCTime(), ts);
		i = clock_settime(CLOCK_REALTIME, &ts);
	} else {
		ts.tv_sec = d;
		d -= ts.tv_sec;
		ts.tv_nsec = d * 1e9;
		ts = evAddTime(evUTCTime(), ts);
		i = clock_settime(CLOCK_REALTIME, &ts);
	}
	if (i != 0)
		Log(LOG_PLL | LOG_ERROR, "pll: error: clock_settime()");
	clock_step++;
}

/*
 * Main pll adjustment interface, gets called from refclocksync.c
 * Parameter 'y' is the current clock offset in seconds.
 * >0 = Clock is running too fast
 * <0 = Clock is running too slow
 */
void pll(double y)
{
	kern_freq(pll_math(y));
}

void pllnone()
{
	kern_freq(pll_poll());
}

void
init_pll()
{
	init_pllmath();
}
