#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <sys/select.h>
#include <sys/time.h>
#include <stdarg.h>
#include <errno.h>

#include <refclocksync.h>
#include <trimble.h>
#include <pll.h>

int loglevel=LOG_NONL | LOG_DRIVER | LOG_INFO | LOG_ERROR | LOG_MAIN | LOG_PPS | LOG_PLL;

/*
 * Reference to the driver. Currently only one driver at a time
 * is supported.
 */
refclocksync_driver_t *curr_driver=NULL;

int signal_alarm_fired=0;

/*************************************************************/

/*
 * Our SIGALRM handler.
 */
void signal_alarm(int i __unused)
{
	signal_alarm_fired=1;
}

/*************************************************************/

void Log(int flags, char *format, ...)
{
	int i;
	va_list args;
	char *ptr;

	/*
	 * All flags in 'flags' have to be enabled in 'loglevel' in order
	 * to produce a log line
	 */
	if(!((flags & loglevel)==flags))
		return;

	ptr=strdup(format);
	if(ptr==NULL)
	{
		fprintf(stdout, "(internal error in log: out of memory)\n");
		return;
	}

	/* strip \n s */
	for(i=0;i<strlen(ptr);i++)
		if(ptr[i]=='\n')
			ptr[i]=' ';
	
	/* output */
	va_start(args, format);
	vfprintf(stdout, ptr, args);
	va_end(args);

	/* print a newline if LOG_NONL flag isn't set */
	if(!(flags & LOG_NONL))
	{
		fprintf(stdout, "\n");
	}

	free(ptr);
}

/*
 * Registers the given driver as primary synchronization source.
 * It uses the pointer verbatim and doesn't attempt to copy the
 * contents of the struct (so make sure it is allocated on the heap).
 */
void register_driver(refclocksync_driver_t *driver)
{
	if(curr_driver!=NULL)
	{
		Log(LOG_MAIN | LOG_ERROR, "main: limited to single driver operation only\n");
		return;
	}

	curr_driver=driver;
}

/*
 * This function is called by the driver if it reports a delta of
 * the local clock.
 * Parameter 'd' is the offset of the clock that has to be
 * corrected (in seconds). If it is positive, the clock has
 * to be slowed down, if it is negative, the clock has to
 * be accellerated.
 */
void driver_delta(double d)
{
	Log(LOG_MAIN | LOG_INFO, "%s: d = %f\n", __FUNCTION__, d);
	if(d<-0.01 || d>0.01)
	{
		/*
		 * Offset is too large to be corrected by the PLL, just step
		 * the clock
		 */
		pllstep(d);
	}
	else
	{
		/*
		 * Feed the current offset into the pll math routines
		 */
		pll(d);
	}
}

int main(int argc, char **argv)
{
	struct sigaction sa;
	struct itimerval timerval;

	/*
	 * Set up our 1 second alarm timer
	 */
	sa.sa_handler=signal_alarm;
	sa.sa_flags=0; // SA_RESTART deliberately not set
	sigemptyset(&sa.sa_mask);
	if(sigaction(SIGALRM, &sa, NULL)<0)
	{
		Log(LOG_MAIN | LOG_ERROR, "main: sigaction: %s\n", strerror(errno));
		return -1;
	}
	timerval.it_interval.tv_sec=1;
	timerval.it_interval.tv_usec=0;
	if(setitimer(ITIMER_REAL, &timerval, NULL)<0)
	{
		Log(LOG_MAIN | LOG_ERROR, "main: setitimer: %s\n", strerror(errno));
		return -1;
	}
	
	/*
	 * Let the driver initialize its services
	 */
	if(trimble_initialize()<0)
	{
		Log(LOG_MAIN | LOG_ERROR, "main: unable to register driver\n");
		return -1;
	}
	if(curr_driver==NULL)
	{
		Log(LOG_MAIN | LOG_ERROR, "main: driver initialization routine failed to register driver\n");
		return -2;
	}

	/*
	 * Initialize the driver
	 */
	if(curr_driver->driver_init(curr_driver)<0)
	{
		Log(LOG_MAIN | LOG_ERROR, "main: driver failed to initialize\n");
		return -3;
	}

	/*
	 * Initialize the PLL
	 */
	init_pll();

	/*
	 * Let's go
	 */
	while(1)
	{
		fd_set rdfd;
		int maxfd;
		int res;

		FD_ZERO(&rdfd);
		maxfd=curr_driver->driver_getfds(curr_driver, &rdfd);
		res=select(maxfd+1, &rdfd, NULL, NULL, NULL);

		if(res<0)
		{
			if(errno==EINTR && signal_alarm_fired==1)
			{
				/*
				 * Interrupted system call - SIGALRM went off
				 */
				signal_alarm_fired=0;
				if(curr_driver->driver_alarm!=NULL)
				{
					curr_driver->driver_alarm(curr_driver);
				}
			}
			else
			{
				Log(LOG_MAIN | LOG_ERROR, "select: %s\n", strerror(errno));
			}
		}
		else if(res>0)
		{
			curr_driver->driver_input_ready(curr_driver, &rdfd);
		}
	}
	return 0;
}
static const char rcsid[] ="$Id: refclocksync.c,v 1.6 2006/11/04 15:28:54 ruben Exp $";
