/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <phk@FreeBSD.ORG> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   Poul-Henning Kamp
 * ----------------------------------------------------------------------------
 * $Id: pllmath.c,v 1.3 2005/01/19 12:41:24 tom Exp $
 *
 * The reason why we're here in the first place...
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>

#define record(a,b,c,d) /**/
#define rrrdupdate(a,...) /**/

#include <refclocksync.h>
#include "pll.h"

struct pll_state pll_state;

/*---------------------------------------------------------------------*/

void
pll_reset()
{
	Log(LOG_PLL | LOG_INFO, "Explicit PLL reset");
	pll_state.st = 0;
}

double
pll_poll()
{
	struct pll_state *ps = &pll_state;

	rrrdupdate("pll", "st:a:b:c:y:z:wa:wamin:dl:zc:my:j:sdy",
	    "N:%d:U:%e:%e:U:U:U:U:U:U:U:U:U",
	     ps->st, ps->b, ps->c);

	if (ps->st >= 5)
		ps->b += ps->c;
	return (ps->b);
}

/*
 * Main PLL calculation routine. Really does the PLL stuff.
 * Parameter 'y' is the current clock offset in seconds.
 * >0 = Clock is running too fast
 * <0 = Clock is running too slow
 * It returns a value suitable for kern_freq (PPM?)
 */
double pll_math(double y)
{
	struct pll_state *ps = &pll_state;

	/*
	 * Record our input and our state for furter digging
	 * If y is positive then the clock is running too fast, thus the frequency correction should be negative.
	 * If y is negative then the clock is running too slow, thus the frequency correction should be positive.
	 */
	ps->y = -y;   // "Phase error measurement", i.e. clock offset

	/*
	record("PLL", 0, ps, sizeof *ps);

	rrrdupdate("pll", "st:a:b:c:y:z:wa:wamin:dl:zc:my:j:sdy",
	    "N:%d:%e:%e:%e:%e:%e:%e:%e:%d:%d:%e:%d:%e",
	     ps->st, ps->a, ps->b, ps->c, ps->y, ps->z, ps->wa, ps->wamin,
	     ps->dl, ps->zc, ps->my, ps->j, ps->sdy);
	*/

	Log(LOG_PLL, "ST %d y %e ly %e my %e lmy %e dy %e ddy %e sdy %e ay %e say %e a %e b %e c %e z %e wa %e wamin %e j %d mj %e dl %d zc %d\n",
			ps->st,ps->y,ps->ly,ps->my,ps->lmy,ps->dy,ps->ddy,ps->sdy,ps->ay,ps->say,ps->a,ps->b,ps->c,
			ps->z,ps->wa,ps->wamin,ps->j,ps->mj,ps->dl,ps->zc);

	/*
	 * Sanity checks.
	 */

	if (fabs(ps->b) > 128e-6)    // "2nd order correction" ... 0.000128 = 128 PPM
		ps->st = 0;          // "PLL state"
	
	/*
	 * If the state is zero, we need to initialize everything
	 */
	if (ps->st == 0) {
		ps->mj = ps->mj0;    // "Initial mj (Max averaging factor)"
		ps->dl = ps->dl0;    // "Initial dl (Delay timer")
		ps->wa = ps->wa0;    // "Initial wa (1st order ratio)
		ps->j = 0;           // "Averaging factor"
		ps->b = 0;           // "2nd order correction"
		ps->c = 0;           // "3rd order correction"
		ps->my = 0;          // "Averaged phase error"
		ps->ly = 0;          // "Last y" = last clock offset
		ps->lmy = 0;         // "Last averaged phase error"
		ps->dy = 0;          // "Two sample difference of y (clock offset)"
		ps->say = 0;         // "Averaged ay (two sample variance (avar sigma=1))"
		ps->sdy = 0;         // "Averaged magnitude of dy (Two sample difference of y)"
		ps->zc = 0;          // "Zero-cross timer"

		ps->st = 1;	/* Goto state 1 */
	}

	/* Count the timer down */
	ps->dl--;                    // "Delay timer"

	/* Adjust the averaging factor */
	if (ps->j < ps->mj)    // "Averaging factor" < "Max averaging factor"
		ps->j++;       // "Averaging factor"++

	/* Update exponential estimate of y */
	/*
	 * Averaged phase error += (clock offset - averaged phase error) / averaging factor
	 */
	ps->my += (ps->y - ps->my) / ps->j;  

	/* Update zero-cross timer */
	if (ps->lmy * ps->my < 0) {    // Last averaged phase error * averaged phase error < 0
		ps->zc = 0;            // Zero cross timer = 0
	} else {
		ps->zc++;              // Zero cross timer ++
	}
	ps->lmy = ps->my;              // Last averaged phase error = averaged phase error

	/* Calculate slope estimates (slope math. = Steigung) */
	ps->ddy = ps->dy;              // Previous two sample differences of y = Two sample difference of y
	ps->dy = ps->y - ps->ly;       // Two sample difference of y = clock offset - last clock offset
	ps->sdy += (fabs(ps->dy) - ps->sdy) / ps->j;
	// Averaged magnitude of the two sample difference += fabs(two sample difference) - averaged magnitude of dy) / averaging factor

	/* Calculate allan variance estimate */
	ps->ay = fabs(ps->dy - ps->ddy);  // Two sample variance = fabs(two sample difference of y - previous two sample difference of y)
	ps->say += (ps->ay - ps->say) / ps->j; // Averaged ay += (two sample variance - averaged two sample variance) / averaging factor

	ps->ly = ps->y;   // Last offset = current offset

	/* Ring up our corrections */
	ps->a = ps->y * ps->wa;  // "1st order correction" = clock offset * 1st order ratio
	if (ps->st >= 3)         // pll status
		// 2nd order correction += clock offset * 1st order ratio ^ 2nd order exponent
		ps->b += ps->y * pow(ps->wa, ps->be);
	if (ps->st >= 5) {       // pll status
		ps->b += ps->c;  // 2nd order correction += 3rd order correction
		// 3rd order correction += clock offset * 1st order ratio ^ 3rd order exponent
		ps->c += ps->y * pow(ps->wa, ps->ce);
	}

	/*
	 * The total correction is the sum of the 1st and 2nd order
	 * corrections.  We accumulate the 3rd order into the 2nd order
	 * because that makes the 2nd order a true frequency estimate.
	 */
	ps->z = ps->a + ps->b;

	/*
	 * If we have not seen any zero-crossings in a long time
	 * we have probably tightened the screw too hard, and need
	 * to loosen it.  If we knew just how much we would need to
	 * loosen it, we could do that, but we don't, and it would
	 * take us a long time to find out "from the underside".
	 * Instead, jack the limit up a notch, and restart the PLL.
	 * This converges fast because the error is small still, and
	 * next time we know to not tighten the screw harder than this.
	 * The amout we jack the limit up, should be larger than the
	 * amount we bang the ratio down (see later) in order to ensure
	 * convergence.
	 * XXX: .8 should be smaller than the .9 further down
	 * XXX: 15.0 is empirical.
	 *
	 * The second condition tries to capture rapid excursions without
	 * waiting for the absense of zero-crossings to alert us.  In a
	 * stable situation, the average magnitude of the slope should 
	 * be very small, and if our average offset exceeds it, things
	 * are certainly not all right.
	 * 
	 */
	if (ps->st >= 4 && 
	   ((ps->zc > 200.0 / ps->wa) || (fabs(ps->my) > 20 * ps->sdy))) {
		ps->wamin = ps->wa / .8;
		if (ps->wamin > 1e-4) {
			/* This is really bad news... */
			ps->wamin = 1e-4;
			Log(LOG_PLL | LOG_INFO, "pll: wamin > 1e-4 is bad news, clamping.\n");
		}
		ps->st = 0;
		return (ps->z);
	}

	/*
	 * Wait for our timer to expire before we fiddle the PLL.
	 */
	if (ps->dl >= 0)
		return (ps->z);
	/*
	 * In state 1 we wait for the 1st order term to converge the
	 * offset.  Once inside the window defined by the noise (as
	 * represented by sdy) and the expected worst case frequency
	 * offset of 128PPM, we are almost close enough to let the 2nd
	 * order term loose, go to state 2 and set a timer.
	 * XXX: 128PPM is a changable constant.
	 */
	if (ps->st == 1 && fabs(ps->y) < (ps->sdy + 128e-6 / ps->wa)) {
		ps->st = 2;
		ps->dl = ps->mj;
		return (ps->z);
	}

	/*
	 * When the timer expires in state 2 we go to state 3 and set
	 * a timer.
	 */
	if (ps->st == 2) {
		ps->st = 3;
		ps->dl = ps->mj;
		return (ps->z);
	}

	/*
	 * In state 3 the 2nd order term grabs hold and should bring the
	 * offset very close to zero (with any frequency drift induced
	 * offset being the ignorable epsilon).  We wait for the timer
	 * to expire and the average offset to be less than half the
	 * allan variance (for sigma=1) and for a zero crossing.
	 * Once we are there, we go to state 4 where we can tighten
	 * the screw gradually.
	 * XXX: the .5 is emperical
	 */
	if (ps->st == 3 && fabs(ps->my) < ps->say * .5 && ps->zc < 10) {
		ps->st = 4;
		ps->dl = ps->mj;
		return (ps->z);
	}

	if (ps->st == 4 && ps->wa < ps->go3)
		ps->st = 5;

	/*
	 * In state 4 and 5 we try to tighten the screw on the PLL as far
	 * as we can.  If we get too far, it fails to converge and the two
	 * trigger conditions above reset the PLL with the added wisdom
	 * recorded.  We only attempt to tighten the screw if our average
	 * offset is less than 10% of the allan estimate and we have just
	 * experienced a zero crossing.  If we do not see a zero-crossing
	 * we have probably already gone too far, and should not go any
	 * further since that would delay the recovery trigger.  We do of
	 * course never tighten beyond the limit we have learned.  If we
	 * have reached the limit, we start the 3rd order loop instead.
	 * Set the timer to let things stabilize.
	 * XXX: .1 is emperical.
	 * XXX: .9 is emperical and must be bigger than the .8 above.
	 * XXX: 10 is emperical.
	 */
	if (ps->st >= 4 && fabs(ps->my) < ps->say * .1 && ps->zc < 10) {
		if (ps->wa > ps->wamin) {
			ps->wa *= .9;
			ps->mj /= .9;
			ps->dl = ps->mj;
		} else {
			ps->dl = ps->mj;
		}
		if (ps->st >= 5)
			ps->dl *= ps->t3;
		return (ps->z);
	}
	return (ps->z);
}

#if 0
static int
pllmath_cfg(evContext ctx __unused, FILE *f, void *priv __unused, char **tok, int flag)
{
	struct pll_state *ps = &pll_state;
	double a;

	if (flag & CFG_SHOW) {
		fprintf(f, "st	%11d*	PLL state\n", ps->st);
		fprintf(f, "a	%11.4e	1. order correction\n", ps->a);
		fprintf(f, "b	%11.4e	2. order correction\n", ps->b);
		fprintf(f, "c	%11.4e	3. order correction\n", ps->c);
		fprintf(f, "z	%11.4e	Total correction\n", ps->z);
		fprintf(f, "wa	%11.4e	1. order ratio\n", ps->wa);
		fprintf(f, "wb	%11.4e	2. order ratio\n", pow(ps->wa, ps->be));
		fprintf(f, "wc	%11.4e	3. order ratio\n", pow(ps->wa, ps->ce));
		fprintf(f, "j	%11d	Averaging factor\n", ps->j);
		fprintf(f, "mj	%11.0f	Max averaging factor\n", ps->mj);
		fprintf(f, "dl	%11d	Delay timer\n", ps->dl);
		fprintf(f, "zc	%11d	Zero Cross timer\n", ps->zc);
		fprintf(f, "y	%11.4e	Phase error measurement\n", ps->y);
		fprintf(f, "ly	%11.4e	Last y\n", ps->ly);
		fprintf(f, "my	%11.4e	Averaged y\n", ps->my);
		fprintf(f, "lmy	%11.4e	Last my\n", ps->lmy);
		fprintf(f, "dy	%11.4e	Two-sample y difference\n", ps->dy);
		fprintf(f, "ddy	%11.4e	Previous dy\n", ps->ddy);
		fprintf(f, "sdy	%11.4e	Average magnitude of dy\n", ps->sdy);
		fprintf(f, "ay	%11.4e	Avar sigma=1\n", ps->ay);
		fprintf(f, "say	%11.4e	Averaged ay\n", ps->say);
		fprintf(f, "mj0	%11.4e*	Initial mj\n", ps->mj0);
		fprintf(f, "dl0	%11.4e*	Initial dl\n", ps->dl0);
		fprintf(f, "wa0	%11.4e*	Initial wa\n", ps->wa0);
		fprintf(f, "wamin	%10.4e*	limit 1. order ratio\n", ps->wamin);
		fprintf(f, "be	%11.4e*	2. order exponent\n", ps->be);
		fprintf(f, "ce	%11.4e*	3. order exponent\n", ps->ce);
		fprintf(f, "go3	%11.4e*	wa threshold for 3. order\n", ps->go3);
		fprintf(f, "t3	%11.4e*	dl expansion in 3. order mode\n", ps->t3);
		return (0);
	}
	while (1) {
		if (tok[0] == NULL || tok[1] == NULL || (flag & CFG_HELP)) {
			ConfigHelp(f, "pllmath <var> <value>");
			return (1);
		}
		a = strtod(tok[1], NULL);
fprintf(f, "<%s> <%s> %f\n", tok[0], tok[1], a);
		/* XXX check nptr etc */
		if (!strcmp(tok[0], "be")) ps->be = a;
		else if (!strcmp(tok[0], "ce")) ps->ce = a;
		else if (!strcmp(tok[0], "mj0")) ps->mj0 = a;
		else if (!strcmp(tok[0], "dl0")) ps->dl0 = a;
		else if (!strcmp(tok[0], "go3")) ps->go3 = a;
		else if (!strcmp(tok[0], "t3")) ps->t3 = a;
		else if (!strcmp(tok[0], "wamin")) ps->wamin = a;
		else if (!strcmp(tok[0], "st") && a == 0) ps->st = a;
		else {
			ConfigHelp(f, "pllmath <var> <value>");
			return (1);
		}
		tok += 2;
		if (tok[0] == NULL)
			return (0);
	}
	return (1);
}
#endif

void
init_pllmath()
{
	struct pll_state *ps = &pll_state;

	ps->mj0 = 8;	/* XXX: empirical */
	ps->dl0 = 10;	/* XXX: empirical */
	ps->wa0 = .1;	/* XXX: empirical */
	ps->be = 2;	/* XXX: empirical */
	ps->ce = 4;	/* XXX: empirical */
	ps->go3 = .001;
	ps->t3 = 4;
	// ConfigAdd(NULL, "pllmath", pllmath_cfg, "PLL mathematics");
}
