#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <sys/select.h>
#include <sys/time.h>
#include <stdarg.h>
#include <errno.h>
#include <math.h>
#include <curses.h>

#include "pll.h"
#include "refclocksync.h"

// #define FLOAT_FORMAT "%+e"
#define FLOAT_FORMAT "%+f"

#define UPDATE()  wnoutrefresh(win_menu); \
		wnoutrefresh(win_msgs); \
		wnoutrefresh(win_vars); \
		doupdate();

/*
 * Total clock offset to be generated (in us)
 */
static double steady_offset=200;
/*
 * Amount of jitter per offset in us
 */
static double jitter=20;

static int loglevel=LOG_NONL | LOG_MAIN | LOG_PLL;

static int runs=0;

static WINDOW *win_menu=NULL, *win_msgs=NULL, *win_vars=NULL;

extern struct pll_state pll_state;

/********************************************************************/

void Log(int flags, char *format, ...)
{
	int i;
	va_list args;
	char *ptr;
	WINDOW *outwin=NULL;

	/*
	 * All flags in 'flags' have to be enabled in 'loglevel' in order
	 * to produce a log line
	 */
	if(!((flags & loglevel)==flags))
		return;

	ptr=strdup(format);
	if(ptr==NULL)
	{
		fprintf(stderr, "(internal error in log: out of memory)\n");
		return;
	}

	/* strip \n s */
	for(i=0;i<strlen(ptr);i++)
		if(ptr[i]=='\n')
			ptr[i]=' ';
	
	/* output */
	if(flags & LOG_MAIN)
	{
		outwin=win_msgs;
	}
	if(flags & LOG_PLL)
	{
		outwin=win_msgs;
	}
	if(outwin==NULL)
	{
		return;
	}
#if 0
	if(!(flags & LOG_NONL))
	{
		wscrl(outwin, -1);
		wmove(outwin, 0, 0);
	}
#endif
	va_start(args, format);
	vwprintw(outwin, ptr, args);
	va_end(args);

	if(!(flags & LOG_NONL))
	{
		wprintw(outwin, "\n");
	}

	UPDATE();
	free(ptr);
}

/********************************************************************/

void status()
{
	werase(win_menu);
	wprintw(win_menu, "Steady offset: %f, Jitter: %f\n", steady_offset, jitter);
	wprintw(win_menu, "(o) Change offset\n");
	wprintw(win_menu, "(j) Change jitter\n");
	wprintw(win_menu, "(<space>) Suspend/resume simulation\n");
	wprintw(win_menu, "(q) Exit\n");
}

/********************************************************************/

/*
 * Generates a certain static offset based on parameters (that yet
 * have to be added).
 * \returns the offset in seconds
 */
double generate_offset()
{
	double offset=0;
	double jit=0;

	/*
	 * In this first test we just assume the clock is running 200us per
	 * step too fast.
	 */
	offset=steady_offset/1000.0/1000.0;

	/*
	 * Introduce random jitter
	 */
	if(jitter!=0)
	{
		jit=((rand()%(int)jitter)-(jitter/2))/1000.0/1000.0;
		offset+=jit;
	}

	return offset;
}

/********************************************************************/

/*
 * Applies the frequency correction in corr to the offset and returns
 * the new offset. corr is in PPM, offset in seconds.
 */
double apply_frequency_correction(double offset, double corr)
{
	offset+=(corr/1e6);
	return offset;
}

/********************************************************************/

/*
 * Does final adjustments to the pll correction before it is used
 * as the new frequency correction for the next run (e.g. make sure
 * that it doesn't exceed 500 PPM in either direction).
 * Parameter units are all in PPM.
 */
double correction_feed(double pllcorr, double current_correction)
{
	static double acc=0;
	double rval=0;

	acc += pllcorr;
	{
		rval=round(acc);
		if(rval>500)
			acc=rval=500;
		if(rval<-500)
			acc=rval=-500;
	}
	acc -= rval;

	rval+=current_correction;
	if(rval>500)
		rval=500;
	if(rval<-500)
		rval=-500;

	return rval;
}

/********************************************************************/

int main(int argc, char **argv)
{
	double current_correction=0;
	int terminate;

	/*
	 * Initialize and setup curses
	 */
	initscr();
	cbreak();
	noecho();
	nodelay(stdscr, TRUE);

	win_vars=newwin(LINES, 40, 0, COLS-40);
	win_menu=newwin(7, COLS-40, 0, 0);
	win_msgs=newwin(LINES-7, COLS-40, 7, 0);
	if(win_vars==NULL || win_menu==NULL || win_msgs==NULL)
	{
		endwin();
		fprintf(stderr, "error creating windows: %p %p %p\n",win_vars,win_menu,win_msgs);
		return -1;
	}
	scrollok(win_msgs, TRUE);
	UPDATE();

	init_pllmath();
	srand(time(NULL));
	runs=0;
	terminate=0;
	while(!terminate)
	{
		double offset=0;
		/*
		 * we start out with 0 PPM correction.
		 */
		double pll_correction=0;
		char c;

		/*
		 * First, generate a clock offset and simulate a PLL step if it
		 * is out of certain bounds (-0.1 ... 0.1)
		 */
		offset=generate_offset();
		if(offset>0.1)
		{
			current_correction-=offset*1e6;
		}
		if(offset<-0.1)
		{
			current_correction+=offset*1e6;
		}

		/*
		 * Then apply our frequency correction to the offset
		 */
		offset=apply_frequency_correction(offset, current_correction);

		pll_correction=pll_math(offset)*1e6;

		current_correction=correction_feed(pll_correction, current_correction);

		runs++;

		/*
		 * Print out current stats to the PLL window
		 */
		werase(win_vars);
		wprintw(win_vars, "PLL run: %d\n", runs);
		wprintw(win_vars, "offset: %+f s\n", offset);
		wprintw(win_vars, "pll_corr: %+f PPM\n", pll_correction);
		wprintw(win_vars, "cur_corr: %+f PPM\n\n", current_correction);
		wprintw(win_vars, "STA = %d\n", pll_state.st);
		wprintw(win_vars, "  y = %+f\n", pll_state.y);
		wprintw(win_vars, " ly = %+f\n", pll_state.ly);
		wprintw(win_vars, " my = %+f\n", pll_state.my);
		wprintw(win_vars, "lmy = %+f\n", pll_state.lmy);
		wprintw(win_vars, " dy = %+f\n", pll_state.dy);
		wprintw(win_vars, "ddy = %+f\n", pll_state.ddy);
		wprintw(win_vars, "sdy = %+f\n", pll_state.sdy);
		wprintw(win_vars, " ay = %+f\n", pll_state.ay);
		wprintw(win_vars, "say = %+f\n", pll_state.say);
		wprintw(win_vars, "\n");
		wprintw(win_vars, "  a = %+f\n", pll_state.a);
		wprintw(win_vars, "  b = %+f\n", pll_state.b);
		wprintw(win_vars, "  c = %+f\n", pll_state.c);
		wprintw(win_vars, "  z = %+f\n", pll_state.z);
		wprintw(win_vars, " wa = %+f\n", pll_state.wa);
		wprintw(win_vars, "wamin = %+f\n", pll_state.wamin);
		wprintw(win_vars, "  j = %+d\n", pll_state.j);
		wprintw(win_vars, " mj = %+f\n", pll_state.mj);
		wprintw(win_vars, " dl = %+d\n", pll_state.dl);
		wprintw(win_vars, " zc = %+d\n", pll_state.zc);
		status();
		UPDATE();

		if((c=getch())!=ERR)
		{
			switch(c)
			{
				case 'q': terminate=1;
					  break;
				case 'o': wprintw(win_menu, "New offset: ");
					  wrefresh(win_menu);
					  echo();
					  wscanw(win_menu, "%lf", &steady_offset);
					  noecho();
					  break;
				case 'j': wprintw(win_menu, "New jitter: ");
					  wrefresh(win_menu);
					  echo();
					  wscanw(win_menu, "%lf", &jitter);
					  noecho();
					  break;
				case ' ': nodelay(stdscr, FALSE);
					  getch();
					  nodelay(stdscr, TRUE);
					  break;
			}
		}
		
		usleep(100000);
	}

	delwin(win_menu);
	delwin(win_vars);
	delwin(win_msgs);
	endwin();
	return 0;
}

static const char rcsid[] = "$Id: pllsim.c,v 1.4 2006/11/04 15:28:54 ruben Exp $";
