/*
   trimble.c
   Driver for the Trimble Acutime 2000 GPS Smart Antenna.

   Main module, sets up the driver.

   Author: Thomas Wana <tom@ripe.net>

   $Id: trimble.c,v 1.3 2005/02/07 14:25:45 tom Exp $
 */

#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/select.h>

#include <refclocksync.h>
#include "trimble.h"

/**********************************************************************/

/*
 * Trimble initalization, called from the main module via the
 * driver struct.
 * Return: 0 = OK, !0 on error
 */
static int trimble_init(struct __refclocksync_driver_t *driver)
{
	trimble_t *data=(trimble_t *)driver->driver_data;
	pps_params_t ppsparams;

	data->serial_fd = -1;
	data->serial_idx = 0;
	data->serial = "/dev/gps0";
	data->state = ST_UNCONFIGURED;
	data->pps_cable_length = 10.0;

	/*
	 * Setup the trimble
	 */
	if(trimble_serconf(data)<0)
		return -1;

	/*
	 * Setup the PPS
	 */
	if(time_pps_create(data->serial_fd, &data->ppsh)<0)
	{
		Log(LOG_PPS | LOG_ERROR, "time_pps_create: %s\n", strerror(errno));
		return -2;
	}
	if(time_pps_getparams(data->ppsh, &ppsparams)<0)
	{
		Log(LOG_PPS | LOG_ERROR, "time_pps_getparams: %s\n", strerror(errno));
		return -2;
	}
	Log(LOG_PPS | LOG_INFO, "pps_params:\n");
	Log(LOG_PPS | LOG_INFO, "   api_version: %d\n", ppsparams.api_version);
	Log(LOG_PPS | LOG_INFO, "   mode: 0x%x\n", ppsparams.mode);

	/* 
	 * capture the assert flank of the signal 
	 * reported timestamp format: timespec
	 */
	ppsparams.mode=PPS_CAPTUREASSERT | PPS_TSFMT_TSPEC;
	if(time_pps_setparams(data->ppsh, &ppsparams)<0)
	{
		Log(LOG_PPS | LOG_ERROR, "time_pps_setparams: %s\n", strerror(errno));
		return -2;
	}
	Log(LOG_PPS | LOG_INFO, "trimble: PPS successfully set up\n");

	return 0;
}

int trimble_getfds(struct __refclocksync_driver_t *driver, fd_set *rdfd);

/*
 * Determine what file descriptor the driver currently uses and
 * set the flags in the fdset accordingly.
 * It returns the number of the highest set fd or <0 on error
 */
int trimble_getfds(struct __refclocksync_driver_t *driver, fd_set *rdfd)
{
	trimble_t *data=(trimble_t *)driver->driver_data;

	if(data->serial_fd!=-1)
	{
		FD_SET(data->serial_fd, rdfd);
	}
	else
	{
		Log(LOG_DRIVER | LOG_ERROR, "trimble: %s: fd uninitialized\n", __FUNCTION__);
	}
	return data->serial_fd;
}

// Request status (status actually printed in trimble_protocol)
/*static int trimble_status(evContext ctx __unused, FILE *f, void *priv, char **tok, int flag __unused)
{
	struct instance *ip;
	int all;

	ip = priv;

	// TODO check state, check ip->unit
	if(tok[0]==NULL)
	{
		fprintf(f, "Available subcommands: all satellites health receiver position time pps\n");
		return -1;
	}

	all=!strcmp(tok[0], "all");

	if(!strcmp(tok[0], "satellites") || all)
	{
		trimble_sendcode(ip, 0x24);
		trimble_sendcode(ip, 0x27);
	}
	else if(!strcmp(tok[0], "health") || all)
	{
		trimble_sendcode(ip, 0x26);
	}
	else if(!strcmp(tok[0], "receiver") || all)
	{
		trimble_sendcode_subcode(ip, 0xBB, 0x00);
	}
	else if(!strcmp(tok[0], "position") || all)
	{
		trimble_sendcode_subcode(ip, 0x8E, 0x20);
	}
	else if(!strcmp(tok[0], "time") || all)
	{
		// Request UTC time
		trimble_sendcode_subcode(ip, 0x8E, 0xAD);
	}
	else if(!strcmp(tok[0], "pps") || all)
	{
		// Request PPS pulse width
		trimble_sendcode_subcode(ip, 0x8E, 0x4F);
	}
	* else if(!strcmp(tok[0], "reboot"))
	   {
	// reboot the trimble
	// it's under 'status' because I am too lazy to introduce an extra command
	// for this debug thingie
	} *
	else
	{
		fprintf(f, "command error\n");
		return -1;
	}

	return 0;
}
*/

/*
 * Main entry point of the driver. Sets up the refclocksync_driver_t
 * struct and registers it with the main module. The main module
 * then calls the init function as provided with the reflocksync_driver_t
 * struct.
 * Returns 0 on success or anything else on error.
 */
int trimble_initialize()
{
	static int once = 0;
	refclocksync_driver_t *driver;

	if (once) { 
		Log(LOG_DRIVER | LOG_ERROR, "trimble module already loaded\n");
		goto cleanup0;
	} else {
		once = 1;
	}

	/*
	 * Set up the driver struct and register it with the main
	 * module
	 */
	driver=(refclocksync_driver_t *)calloc(sizeof(refclocksync_driver_t), 1);
	if(driver==NULL)
	{
		Log(LOG_DRIVER | LOG_ERROR, "out of memory\n");
		goto cleanup1;
	}
	driver->driver_data=(void *)calloc(sizeof(trimble_t), 1);
	if(driver->driver_data==NULL)
	{
		free(driver);
		Log(LOG_DRIVER | LOG_ERROR, "out of memory\n");
		goto cleanup2;
	}
	driver->driver_init=trimble_init;
	driver->driver_getfds=trimble_getfds;
	driver->driver_input_ready=trimble_input_ready;
	driver->driver_alarm=trimble_alarm;

	register_driver(driver);

	return 0;

cleanup2:
	free(driver);
cleanup1:
cleanup0:
	return -1;
}
