/*
   trimble_serial.c
   Driver for the Trimble Acutime 2000 GPS Smart Antenna.

   Serial communication layer. Has an interface to send packets and contains the
   handler for serial communication. Complete packets will be passed to the
   protocol layer (trimble_protocol.c)

   Author: Thomas Wana <tom@ripe.net>

   $Id: trimble_serial.c,v 1.2 2004/12/10 10:42:56 tom Exp $
 */


#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/select.h>
#include <termios.h>

#include <refclocksync.h>
#include "trimble.h"

// A routine that prints a buffer in 'hexdump -C'-style
void trimble_hexdump(int logflags, unsigned char *buffer, long len)
{
   unsigned long pos=0;

   while(len>0)
   {
      int i;
      Log(logflags | LOG_NONL, "%08x  ",(unsigned int)pos);
      for(i=0;i<16 && i<len;i++)
      {
         Log(logflags | LOG_NONL,"%02x ",buffer[pos+i]);
         if(i==7 || i==15)
            Log(logflags | LOG_NONL," ");
      }
      Log(logflags | LOG_NONL,"|");
      for(i=0;i<16 && i<len;i++)
      {
         unsigned char c=buffer[pos+i];
         if( (c>='A' && c<='Z') || (c>='a' && c<='z') || (c>='0' && c<='9') ||
             c==':' || c=='!' || c=='?' || c=='=' || c==';' || c=='<' ||
             c=='>' || c=='/' || c=='\\' || c=='&')
         {
            Log(logflags | LOG_NONL,"%c",c);
         }
         else
         {
            Log(logflags | LOG_NONL,".");
         }
      }
      Log(logflags,"|");
      len-=16;
      pos+=16;
   }
}

// configure serial communication, open the device etc.
// serial line configuration according to NTP 4.2.0/ntpd/refclock_ripencc.c (flags combined)
int trimble_serconf(trimble_t *data)
{
	int fd, i;
	struct termios t;

	Log(LOG_DRIVER | LOG_INFO, "trimble: cable length is %f meters", data->pps_cable_length);

	Log(LOG_DRIVER | LOG_INFO | LOG_NONL, "trimble: opening serial device...");
	fd = open(data->serial, O_RDWR | O_NDELAY);
	if (fd < 0) {
		Log(LOG_DRIVER | LOG_ERROR, "open(%s): %s", data->serial, strerror(errno));
		return (-1);
	}
	i = tcgetattr(fd, &t);
	if (i < 0) {
		Log(LOG_DRIVER | LOG_ERROR, "tcgetattr(%s): %s", data->serial, strerror(errno));
		close(fd);
		return (-1);
	}
	t.c_cflag=PARENB|PARODD|CS8|CLOCAL|CREAD;
	t.c_oflag=0;
	t.c_iflag=0;
	t.c_lflag=0;
	t.c_cc[VMIN]=1;

	cfsetispeed(&t, B9600);
	cfsetospeed(&t, B9600);

	i = tcsetattr(fd, TCSANOW, &t);
	if (i < 0) {
		Log(LOG_DRIVER | LOG_ERROR, "tcsetattr(%s): %s", data->serial, strerror(errno));
		close(fd);
		return (-1);
	}

	if(tcflush(fd, TCIOFLUSH)<0)
	{
		Log(LOG_DRIVER | LOG_ERROR, "tcflush(%s): %s", data->serial, strerror(errno));
		close(fd);
		return -1;
	} 
	Log(LOG_DRIVER | LOG_INFO, " serial device opened and configured");

	data->serial_fd = fd;
	data->state=ST_SERIAL_SETUP_COMPLETE;
	Log(LOG_DRIVER | LOG_INFO, "trimble: serial setup complete, initializing");

	// initialize the trimble protocol layer
	trimble_protocol_init(data);

	return (0);
}

// serial input function, reads in a message from the serial port.
// Called from main with a fd_set. The function has to check itself if
// its file descriptor(s) are set in the fd_set. 
void trimble_input_ready(struct __refclocksync_driver_t *driver, fd_set *rdfd)
{
	int i, l, remember_0x10;
	char *ptr, *optr;
	char recvbuf[2048];
	trimble_t *data=(trimble_t *)driver->driver_data;

	/*
	 * Check if there is something on our fd. It might well happen
	 * that there is nothing for us, that's not an error.
	 */
	if(!FD_ISSET(data->serial_fd, rdfd))
	{
		return;
	}

	/*
	 * Read as much as we can fit in our buffer
	 */
	i = read(data->serial_fd, data->serial_buf + data->serial_idx, sizeof(data->serial_buf) - data->serial_idx);
	if (i <= 0)
		return;
	Log(LOG_DRIVER | LOG_RAW, "%s: activity on serial port, read:",__FUNCTION__);
	trimble_hexdump(LOG_DRIVER | LOG_INFO | LOG_RAW,  data->serial_buf+data->serial_idx, i);

	data->serial_idx += i;
	Log(LOG_DRIVER | LOG_RAW, "%s: data->serial_buf is now:",__FUNCTION__);
	trimble_hexdump(LOG_DRIVER | LOG_INFO | LOG_RAW,  data->serial_buf, data->serial_idx);
	{
		int end_seen;
		/*
		 * Stop if we don't have enough data for a complete message
		 * the shortest possible message is 4 bytes: 0x10 0xCode 0x10 0x03
		 */
		if (data->serial_idx < 4)
			return;

		// what do we have in the buffer ...
		if(data->serial_buf[0]!=0x10)
		{
			// fsck ... we are definitely not aligned to a packet
			Log(LOG_DRIVER | LOG_RAW,"Serial buffer doesn't start with a packet!");

			// move up the contents of the packet until
			// we find a 0x10 (it's not perfect but better
			// than nothing). Or if there is no 0x10, return.
			do
			{
				data->serial_idx--;
				memcpy(data->serial_buf, data->serial_buf+1,
						data->serial_idx);
			} while(data->serial_idx>0 && data->serial_buf[0]!=0x10);

			if(data->serial_idx==0) {
				// we couldn't find anything ...
				Log(LOG_DRIVER | LOG_RAW, "couldn't recover buffer");
				return;
			}

			Log(LOG_DRIVER | LOG_RAW, "buffer recovered");
		}

		// check if there is a whole packet in the buffer and copy it out
		memset(recvbuf, 0, sizeof(recvbuf));
		ptr=data->serial_buf;
		optr=recvbuf;

		// strategy: go through the buffer bytewise. If there is a 0x10, remember
		// that there was one and do nothing. If in the next iteration there is 
		// another 0x10, unstuff it. If there is a 0x03, the packet is done.
		for(i=2, remember_0x10=0, end_seen=0;i<data->serial_idx && !end_seen;i++)
		{
			if(ptr[i]==0x10)
			{
				if(remember_0x10==0)
					remember_0x10=1;
				else
				{
					remember_0x10=0;
				}
			}
			else if(ptr[i]==0x03 && remember_0x10==1)
			{
				// end of packet
				end_seen=1;
			}
			else if(remember_0x10==1)
			{
				Log(LOG_DRIVER | LOG_RAW, "Packet error: no 0x10 may stand alone in "
						"the user data");
				Log(LOG_DRIVER | LOG_RAW,"Killing the buffer");
				data->serial_idx=0;
				return;
			}

			if(remember_0x10==0)
			{
				if((unsigned int)(optr-recvbuf)>=sizeof(recvbuf))
				{
					Log(LOG_DRIVER | LOG_RAW, "About to overflow the receive buffer, packet too large!");
					return;
				}
				*optr++=ptr[i];
			}
		}

		if(end_seen)
		{
			// we got a complete packet in recvbuf.
			TSIPPKT pkt;
			memset(&pkt, 0, sizeof(pkt));

			pkt.code=ptr[1];
			pkt.len=optr-recvbuf;
			// sanity check
			if(pkt.len>sizeof(pkt.buf))
			{
				Log(LOG_DRIVER | LOG_RAW, "??? about to overflow buf ???");
				return;
			}
			memcpy(pkt.buf, recvbuf, pkt.len);

			l=pkt.len+4;

			// look if we unstuffed any 0x10. if yes, increase the copy counter
			// so we can get the serial buffer right in the memcpy below. [1]
			for(i=0;i<pkt.len;i++)
				if(pkt.buf[i]==0x10)
					l++;

			Log(LOG_DRIVER | LOG_RAW, "%s: got a complete packet: code = 0x%02x, len = %u",__FUNCTION__,
			     pkt.code, pkt.len);
			trimble_hexdump(LOG_DRIVER | LOG_INFO | LOG_RAW,  pkt.buf, pkt.len);

			// act on the packet
			trimble_do_protocol(data, &pkt);
		}
		else
		{
			// still waiting for packet completion, no bytes consumed
			l=0;
			Log(LOG_DRIVER | LOG_RAW, "%s: still waiting for packet completion", __FUNCTION__);
		}

		/*
		 * Pull up anything at the end and start over [1]
		 */
		memcpy(data->serial_buf, data->serial_buf + l, data->serial_idx - l);
		data->serial_idx -= l;
		Log(LOG_DRIVER | LOG_RAW, "%s: (after cleanup) data->serial_buf is now (l=%d):",__FUNCTION__, l);
		trimble_hexdump(LOG_DRIVER | LOG_INFO | LOG_RAW,  data->serial_buf, data->serial_idx);
	}
}

// send out a trimble message.
// Composition of the packet: User Guide page A-16
void trimble_sendmsg(trimble_t *data, TSIPPKT *pkt)
{
	unsigned char buf[1024], *pbuf, *ibuf;

	pbuf=buf;
	*pbuf++=0x10;
	*pbuf++=pkt->code;

	if(pkt->len>sizeof(buf)/2-4) // worst case: all 0x10 doubles the buffer
	{
		Log(LOG_DRIVER | LOG_RAW, "%s: packet too large, ignoring",__FUNCTION__);
		return;
	}

	// escape 0x10 in the user buffer with 0x10 0x10 and copy the
	// buffer from pkt to buf
	ibuf=pkt->buf;
	while(pkt->len--)
	{
		if(*ibuf==0x10)
			*pbuf++=0x10;
		*pbuf++=*ibuf++;
	}

	// conclude the message
	*pbuf++=0x10;
	*pbuf++=0x03;

	Log(LOG_DRIVER | LOG_INFO | LOG_RAW, "%s: sending out packet:",__FUNCTION__);
	trimble_hexdump(LOG_DRIVER | LOG_INFO | LOG_RAW,  buf, pbuf-buf);

	// send out the message
	if(write(data->serial_fd, buf, pbuf-buf)<0)
	{
		Log(LOG_DRIVER | LOG_ERROR, "%s: write(2) error: %s",__FUNCTION__,strerror(errno));
		return;
	}
}

// convenience function; acts like trimble_sendmsg, but sends out
// a packet only consisting of one code and without length.
void trimble_sendcode(trimble_t *data, unsigned char code)
{
   TSIPPKT pkt;
   memset(&pkt, 0, sizeof(pkt));

   pkt.len=0;
   pkt.code=code;

   trimble_sendmsg(data, &pkt);
}

// convenience function; acts like trimble_sendmsg, but sends out
// a packet only consisting of the code and one subcode
void trimble_sendcode_subcode(trimble_t *data, unsigned char code, unsigned char subcode)
{
   TSIPPKT pkt;
   memset(&pkt, 0, sizeof(pkt));

   pkt.len=1;
   pkt.code=code;
   pkt.buf[0]=subcode;

   trimble_sendmsg(data, &pkt);
}
