/* 
 * vi:ts=4:sw=4:
 *
 * $Id: poisson.c,v 1.6 2003/10/17 10:49:10 ruben Exp $
 *
 * Copyright (c) 1998-2003 RIPE NCC
 * 
 * All Rights Reserved
 * 
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of the author not be used in advertising or 
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.
 * 
 * THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS; IN NO EVENT SHALL
 * AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY
 * DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE. 
 */
#include "poisson.h"

static char const rcsid[] = "$Id: poisson.c,v 1.6 2003/10/17 10:49:10 ruben Exp $";


/* 
 * Returns as a floating point number an integer value that is a random deviate 
 * from a Poisson distribution with mean xm, using ran1(idum) as a source of
 * random numbers.
 * 
 * Copied from "Numerical Recipes in C", Section 7.3, page 294.
 * 
 */

float poisson (float xm, long *idum)
{

	static float        sq,
	                    alxm,
	                    g,
	                    oldm = (-1.0);
	float               em,
	                    t,
	                    y;

	if (xm < 12.0)
	{
		if (xm != oldm)
		{
			oldm = xm;
			g = exp (-xm);
		}
		em = -1;
		t = 1.0;
		do
		{
			++em;
			t *= ran1 (idum);
		}
		while (t > g);
	}
	else
	{
		if (xm != oldm)
		{
			oldm = xm;
			sq = sqrt (2.0 * xm);
			alxm = log (xm);
			g = xm * alxm - gammaln (xm + 1);
		}
		do
		{
			do
			{
				y = tan (PI * ran1 (idum));
				em = sq * y + xm;
			}
			while (em < 0.0);
			em = floor (em);
			t = 0.9 * (1.0 + y * y) * exp (em * alxm - gammaln (em + 1.0) - g);
		}
		while (ran1 (idum) > t);
	}
	return em;
}


/* 
 * Minimal random number generator of Park and Miller with Bays-Durham shuffle
 * and added safe guards.  Returns a uniform random deviate between 0.0 and 1.0
 * (exclusive of the end-point values).  Call with idum a negative integer to
 * initialize, thereafer do not alter idum in a single sequence.
 * 
 * Copied from Numerical Recipes in C, page 280, section 7.1
 * 
 */

/* XXX Could make use of RAND_bytes(3) from openssl to provide strong random data */

float ran1 (long *idum)
{
	int                 j;
	long                k;
	static long         iy = 0;
	static long         iv[NTAB];
	float               temp;

	if (*idum <= 0 || !iy)
	{
		if (-(*idum) < 1)
			*idum = 1;
		else
			*idum = -(*idum);
		for (j = NTAB + 7; j >= 0; j--)
		{
			k = (*idum) / IQ;
			*idum = IA * (*idum - k * IQ) - IR * k;
			if (*idum < 0)
				*idum += IM;
			if (j < NTAB)
				iv[j] = *idum;
		}
		iy = iv[0];
	}
	k = (*idum) / IQ;
	*idum = IA * (*idum - k * IQ) - IR * k;
	if (*idum < 0)
		*idum += IM;
	j = iy / NDIV;
	iy = iv[j];
	iv[j] = *idum;
	if ((temp = AM * iy) > RNMX)
	{
		return RNMX;
	}
	else
		return temp;
}


/* 
 * Returns the value ln(gamma(xx)) for xx>0.
 * 
 * Copied from Numerical Recipes in C, page 214, section 6.1
 * 
 */

float gammaln (float xx)
{

	double              x,
	                    y,
	                    tmp,
	                    ser;

	static double       cof[6] = { 76.18009172947146,
									-86.50532032941677,
									24.01409824083091,
									-1.231739572450155,
									0.1208650973866179e-2,
									-0.5395239384953e-5
								};
	int                 j;

	y = x = xx;
	tmp = x + 5.5;
	tmp -= (x + 0.5) * log (tmp);
	ser = 1.000000000190015;
	for (j = 0; j < 5; j++)
		ser += cof[j] / ++y;

	return -tmp + log (2.5066282746310005 * ser / x);
}

/* 
 * Read seed from file: Argument is filename where to find seed, otherwise use 
 * default Returns seed, or -1 if seed cannot be found.
 * 
 */
long poisson_getseed (int debug, char *filename)
{

	long                seed = 1.0;	/* Initial seed value */
	FILE               *fp;
	char                random_seed_file[100] = "./";	/* Where to find the
														 * configfile */
	int                 result;

	if (filename == NULL)
	{
		if (getenv ("TTROOT"))
		{
			strcpy (random_seed_file, getenv ("TTROOT"));
		}
		strcat (random_seed_file, "/config/poisson.seed");
	}
	else
	{
		strcpy (random_seed_file, filename);
	}

	fp = fopen (random_seed_file, "r");
	if (fp != NULL)
	{
		result = fscanf (fp, "%ld", &seed);
		if (result != 1)
		{
			seed = 1;
		}
		fclose (fp);
	}
	else
	{
		seed = 1;
	}

	if (debug)
	{
		printf ("poisson_getseed: read %ld as seed\n", seed);
	}

	return -1 * seed;
}

/* 
 * Write seed to file: Argument is filename where to store seed, otherwise use
 * default Second argument is seed return 0 on success, 1 on failure.
 * 
 */

int poisson_saveseed (int debug, char *filename, long seed)
{

	FILE               *fp;
	char                random_seed_file[100] = "./";

	if (filename == NULL)
	{
		if (getenv ("TTROOT"))
		{
			strcpy (random_seed_file, getenv ("TTROOT"));
		}
		strcat (random_seed_file, "/config/poisson.seed");
	}
	else
	{
		strcpy (random_seed_file, filename);
	}

	if (debug)
	{
		printf ("poisson_saveseed: writing seed %ld to %s\n", seed, random_seed_file);
	}

	fp = fopen (random_seed_file, "w");
	if (fp != NULL)
	{
		fprintf (fp, "%ld\n", seed);
		fclose (fp);
		return 0;
	}
	else
	{
		return 1;
	}
}


/* 
 * This generates a sequence of random numbers with intervals according to a
 * poisson distribution in target->time.  It does some sanity checking on the
 * sequence.
 * 
 * Arguments: debug: TRUE/FALSE, if TRUE be verbose target Struct to file.
 * seed: Random seed, can be overwritten with target->seed
 * 
 * Returns: seed: Random seed after the last call.
 * 
 */

long do_poisson (int debug, struct target *target, int offset, long seed)
{

	float               xm;
	float               sum = 0;
	float               result = 0;
	int                 generated = 0;

	if (debug)
	{
		printf ("do_poisson: Generating distribution for %s\n", target->name);
	}

	/* Mean interval of poisson distribution */
	xm = ((float) ALL_DONE * globaltime.period * 1000) / ((float) (target->rate) + 1);

	/* User specified seed, reinitialize random number generator */
	if (target->seed != 0)
	{
		if (debug)
		{
			printf ("do_poisson: Overriding seed %d\n", target->seed);
		}
		seed = -1.0 * target->seed;
		ran1 (&seed);
	}

	if (debug)
	{
		if (target->linear == FALSE)
		{
			printf ("do_poisson: using poisson distribution\n");
		}
		else
		{
			printf ("do_poisson: using linear distribution\n");
		}
	}


	/* Build sequence of intervals */
	while (1)
	{
		/* Generate interval, make sure that it is above the minimum */
		result = 0;
		while (result < MIN_INTERVAL)
		{
			if (target->linear == FALSE)
			{
				result = poisson (xm, &seed);
			}
			else
			{
				result = xm;
			}
		}
		if (((sum + result) > (ALL_DONE * globaltime.period * 1000)) || (generated >= target->rate))
		{
			break;
		}
		sum = sum + result + (offset * 1000);
		offset = 0;
		*(target->time + generated) = sum;
		generated++;
	}

	if (debug)
	{
		printf ("do_poisson: Requested %d  Actual %d ", target->rate, generated);
	}
	target->rate = generated;

	if (debug)
	{
		printf ("sum %f max %f\n", sum, ALL_DONE * globaltime.period * 1000);
	}

	return seed;
}
