/*
 * vi:ts=4:sw=4:
 *
 * $Id: send_packet.c,v 1.16 2006/06/26 10:35:48 ruben Exp $
 *
 * Copyright (c) 1998-2003 RIPE NCC
 * 
 * All Rights Reserved
 * 
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of the author not be used in advertising or 
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.
 * 
 * THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS; IN NO EVENT SHALL
 * AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY
 * DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE. 
 */

#include	<sys/stat.h>
#include	<unistd.h>
#include	<sysexits.h>

#include	<sys/types.h>
#include	<sys/param.h>
#include	<stdlib.h>
#include	<stdio.h>
#include	<string.h>
#include	<netdb.h>
#include	<time.h>
#include	<signal.h>
#include	<math.h>

#include	<netinet/in_systm.h>
#include	<netinet/in.h>
#include	<arpa/inet.h>

#include	<sys/socket.h>
#include	<net/if.h>

#include	<netinet/if_ether.h>
#include	<netinet/ip.h>
#include	<netinet/udp.h>

#include	"send_packet.h"
#include	"measurement.h"		/* Error handling and buffer read write */
#include	"packet.h"
#include	"if_utils.h"
#include	"poisson.h"
#include	"testbox.h"

#define PNAME "send_packet"

/* Define to zero if you want to skip gethostname(2) while inside udp_open */
#define UDP_OPEN_OLDBEHAVIOUR	(1)

static struct target *first_testbox = NULL;
static struct measurement *first_measurement = NULL;

static const char RcsID[] = "$Id: send_packet.c,v 1.16 2006/06/26 10:35:48 ruben Exp $"; 

int read_configfile (int debug, char *filename)
{

	/* Read configuration file 
	 * debug: TRUE/FALSE : If TRUE print debug output
	 * filename: name of configuration file. Use $TTROOT/config/configfile if
	 *           filename==NULL. 
	 * Returns: 
	 * 		-1 : Configfile cannot be found. 
	 * 		-2 : Configfile hasn't been modified. 
	 *		 0... : Number of entries in the configfile
	 */

	FILE               *fp;
	struct target      *next;
	struct measurement *next2;
	char               *point,
	                    buffer[MAXLINE],
	                    hostname[MAXLINE];
	char               *new;
	int                 entries = 0;
	time_t              t;
	struct stat         configstats;

	int                 temp,
	                    nwords;
	int                 error_line;
	char                ver[6];
	struct in_addr      addr;
	struct in6_addr     addr6;
	struct hostent     *dns_addr;


	/* Check if configfile specified, if not use default */
	if (filename == NULL)
	{
		if (getenv ("TTROOT") == NULL)
		{
			printf ("Error: TTROOT not set!\n");
			exit (0);
		}
		else
		{
			filename =
				(char *) malloc (strlen (getenv ("TTROOT")) +
								 strlen ("/config/configfile") + 3);
			filename = strcpy (filename, getenv ("TTROOT"));
			filename = strcat (filename, "/config/configfile");
		}
	}

	/* Check if configfile hasn't been updated */
	stat (filename, &configstats);

	if (debug)
	{
		printf ("Configfile last read at %ld\n", globaltime.read);
		printf ("Configfile last modified at %ld\n", configstats.st_ctime);
	}

	if ((globaltime.read > configstats.st_ctime))
	{
		if (debug)
		{
			printf ("Continueing with old scheme\n");
		}
		return -2;
	}
	else
	{
		if (debug)
		{
			printf ("Reading config file\n");
		}
	}

	/* Yes, clean up all old data */
	while (first_testbox != NULL)
	{
		free (first_testbox->name);
		free (first_testbox->time);
		next = first_testbox->next;
		free (first_testbox);
		first_testbox = next;
	}

	while (first_measurement != NULL)
	{
		next2 = first_measurement->next;
		free (first_measurement);
		first_measurement = next2;
	}

	memset ((char *) &ver, 0, sizeof (ver));

	globaltime.start = 0;
	globaltime.period = 0;
	globaltime.read = time (&t);

	/* Get hostname */
	entries = gethostname (&hostname[0], MAXLINE);

	/* Read configfile */
	fp = fopen (filename, "r");
	if (fp == NULL)
	{
		printf ("Cannot open %s\n", filename);
		return -1;
	}

	error_line = 0;

	while ((point = fgets (buffer, MAXLINE, fp)) != NULL)
	{

		error_line++;

		/* Version */
		if (strncmp (point, "VER", 3) == 0)
		{
			strtok (point, " ");
			strtok (NULL, " ");
			strcpy (ver, strtok (NULL, " "));
			ver[5] = 0;
		}

		/* Lines with PERI keyword */
		if (strncmp (point, "PERI", 4) == 0)
		{
			strtok (point, " ");
			strtok (NULL, " ");
			globaltime.start = atoi (strtok (NULL, " "));
			globaltime.period = atoi (strtok (NULL, " "));
		}

		/* Lines with SEND or SEND6 keyword */
		else if ((strncmp (point, "SEND", 4) == 0)
				 || (strncmp (point, "6SND", 4) == 0))
		{
			if (strncmp (point, "6SND", 4) == 0)
				temp = AF_INET6;
			else
				temp = AF_INET;

			if ((next = (struct target *) malloc (sizeof (struct target))) == NULL)
			{
				perror("read_configfile");
				abort();
			}
			memset(next,0, sizeof (struct target));

			next->next = first_testbox;

			strtok (point, " ");
			new = strtok (NULL, " ");

			nwords = atoi (new);

			/* Source field */
			new = strtok (NULL, " ");
			if (strcmp (new, hostname) != 0)
			{
				if (debug)
				{
					printf ("configfile: Measurement not for this TB\n");
				}
				continue;
			}

			/* Dest field */
			new = (char *) strtok (NULL, " ");
			next->name = (char *) malloc (strlen (new));
			next->name = strcpy (next->name, new);

			/* 2 TB names, 4 mandatory, 2 optional parameters: Port, Size,
			 * Random Seed, Rate and optionally burst */

			next->port = atoi (strtok (NULL, " "));
			next->size = atoi (strtok (NULL, " "));
			next->seed = atoi (strtok (NULL, " "));
			next->rate = atoi (strtok (NULL, " "));

			if (nwords == 8)
			{
				/* Optional parameters specified */
				next->burst = atoi (strtok (NULL, " "));
				if (next->burst < 0)
				{
					next->burst = 1;
				}
				next->linear = atoi (strtok (NULL, " "));
			}
			else
			{
				/* Default values */
				next->burst = 1;
				next->linear = FALSE;
			}

			next->time = (float *) malloc (sizeof (float) * next->rate);
			memset(next->time, 0, sizeof (float) * next->rate);


			/* 
			 *
			 * This part checks if the IP family matches the target
			 *
			 */


			if (temp == AF_INET)
			{
				if ((inet_pton (AF_INET, next->name, &addr) <= 0)
					&& ((dns_addr = gethostbyname (next->name)) == NULL))
				{
					HandleError (PNAME,
								 "Type mismatch in configfile at line %d!",
								 error_line);
					exit (1);
				}
				else
				{
					next->family = AF_INET;
					memcpy(&(next->target.addr),dns_addr->h_addr,dns_addr->h_length);
				}
			}
			else if (temp == AF_INET6)
			{
				if ((inet_pton (AF_INET6, next->name, &addr6) <= 0)
					&& ((dns_addr = gethostbyname2 (next->name, AF_INET6)) ==
						NULL))
				{
					HandleError (PNAME,
								 "Type mismatch in configfile at line %d!",
								 error_line);
					exit (1);
				}
				else
				{
					next->family = AF_INET6;
					memcpy(&(next->target.addr6),dns_addr->h_addr,dns_addr->h_length);
				}
			}
			if ((strcmp (ver, "1.0.0") != 0) && (strcmp (ver, "1.1.0") != 0))
			{
				HandleError (PNAME,
							 "Wrong or unsupported version in configfile!");
				exit (0);
			}


			first_testbox = next;
			entries++;
		}
		/* Ignore all others */
	}

	fclose (fp);

	/* These 2 should be set */
	if ((globaltime.start == 0) && (globaltime.period == 0))
	{
		printf ("Missing PERI keyword\n");
		entries = -1;
	}

	if (debug)
	{

		printf ("configfile: version %s\n", ver);
		printf ("configfile: found %d entries\n", entries);
		printf ("configfile: schedule start   %ld \n", globaltime.start);
		printf ("configfile: schedule period  %ld \n", globaltime.period);
		printf ("configfile: schedule read at %ld \n", globaltime.read);
		next = first_testbox;
		while (next != NULL)
		{
			printf ("configfile: measure to %s\n", next->name);
			next = next->next;
		}
	}

	return entries;
}

/* 
 * -------------------------------------------------------------------------------
 * Subroutine Header SendPacket
 * 
 * Purpose : Do sanity checks, alocate storage, get a prebuild packet and put it on 
 * either a v4- or a v6-socket
 * 
 * Side Effects: No globals modified. Comments : Inputparameter : hostname and port 
 * to send packet to. Packet length and a serialnumber that will be written into
 * the packet. Family (AF_INET or AF_INET6) Returns : Return 0 on success or non-0
 * on failure.
 * 
 * 
 */

int SendPacket (int debug, struct measurement *current_measurement, struct sockaddr *name, socklen_t * namelen)
{

	char               *host,
	                   *packet;
	double              sendtime;
	int                 burst,
	                    burstcount,
	                    family,
	                    port,
	                    returnval,
	                    serialnumber,
	                    socketdescr;	/* Socket descriptor */
	struct pps_val      timeinfo;
	struct test_packet_data *data;
	u_int               packsize;

	/* Get parameters from measurement structure */

	host = current_measurement->name;
	port = current_measurement->port;
	sendtime = current_measurement->time;
	packsize = current_measurement->size;
	burst = current_measurement->burst;
	family = current_measurement->family;

	/* Create unique id for packet */
	serialnumber = floor (sendtime - TIMEZERO);
	serialnumber = serialnumber % CYCLEPERIOD;
	serialnumber = serialnumber * FREQUENCY + fmod (sendtime, 1.0) * FREQUENCY;

	/* 
	 * XXX #define UDP_OPEN_OLDBEHAVIOUR to NULL, if you want udp_open to do an
	 * extra address lookup (old behaviour) Otherwise you may save some lookups
	 * but it also means testbox address changes are only seen when SendPacket
	 * reinitialize itself around the hour 
	 */

	socketdescr = udp_open (host, NULL, port, 0, family, UDP_OPEN_OLDBEHAVIOUR ? NULL : current_measurement);

	/* Error Handling here */

	if (packsize < (TEST_PACKET_HEADERLENGTH + TEST_PACKET_DATALENGTH))
	{
		packsize = TEST_PACKET_HEADERLENGTH + TEST_PACKET_DATALENGTH;
		HandleError (PNAME, "Changed packetsize to %d", packsize);
	}

	if (packsize > TEST_PACKET_MAXLENGTH)
	{
		packsize = TEST_PACKET_MAXLENGTH;
		HandleError (PNAME, "Changed packetsize to %d", packsize);
	}

	packet = malloc (packsize);

	if (packet == NULL)
	{
		HandleError (PNAME "malloc packet failed");
		return (-1);
	}

	for (burstcount = 0; burstcount < burst; burstcount++)
	{
		data = SetupPacket (debug, packet, packsize, serialnumber);

		if (debug)
		{
			printf (" packet data diff : %d,%d, %d / packet serial : %d\n", (int) packet, (int) data, (int) data - (int) packet, serialnumber);
		}

		serialnumber++;

		/* insert a timestamp From now on do as little as possible */
		if (getppstime (&timeinfo))
		{
			HandleError (PNAME, "getppstime returned non-zero)");
			/* HU, free used resources */
			free (packet);
			close (socketdescr);
			return (-2);
		}
		/* set all the time parameters in the test packet data field */
		data->timepars = timeinfo;

		/* Write the packet to the socket. */
		returnval = writen (socketdescr, packet, packsize);	/* from utils.c */
		if (returnval != (int) packsize)
		{
			if (returnval > 0)
			{
				HandleError (PNAME, " Only %d out of %d bytes written", returnval, packet);
				/* HU, free used resources */
				free (packet);
				close (socketdescr);
				return (-3);
			}
			else
			{
				HandleError (PNAME, "writen error on socket");
				/* and send packet again */
			}
		}						/* returnval = packsize */
	}							/* burstcount */

	/* 
	 * Record source address, a 4.4BSD feature only (Stevens et al) 
	 * It is used to determine in which SNDP file this packet should go.
	 *
	 */

	if (getsockname (socketdescr, name, namelen) != 0)
	{
		HandleError (PNAME, " Failure during srcaddr lookup: %s", strerror (errno));
		free (packet);
		close (socketdescr);
		return (-4);
	}
	free (packet);
	close (socketdescr);
	return (serialnumber - burst);
}

/*
 * -------------------------------------------------------------------------------
 * Subroutine Header
 * Purpose           : Set up a connection.
 * Side Effects      :
 * Comments          : Stevens Unix Network Programming 1990, page 400
 *                     Modified to support IPv6	[FF]
 * Inputparameter    : char * hostname, family
 * Returns           : Return Socket descriptor or -1 if unsuccessfull.
 *
 */

int                     /* return socket descriptor if OK, else -1 on error */
udp_open(
char    *host,          /* name of other system to communicate with */
char    *service,       /* name of service being requested */
                        /* can be NULL, iff port > 0 */
int     port,           /* if == 0, nothing special - use port# of service */
                        /* if < 0, bind a local reserved port */
                        /* if > 0, it's the port# of server (host-byte-order) */
int     dontconn,      	/* if == 0, call connect(), else don't */
int	family,		/* AF_INET or AF_INET6 */
struct measurement *current_measurement	/* Holding address amongst other things*/
)
{
	int                 fd;
	char               *host_err_str(void);
	struct servent     *sp;
	struct hostent     *hp;

	unsigned short      ip_flag;

	struct sockaddr_in  udp_srv_addr;	/* server's Internet socket addr */
	struct sockaddr_in6 udp_srv_addr6;
	struct sockaddr_in  udp_cli_addr;	/* client's Internet socket addr */
	struct sockaddr_in6 udp_cli_addr6;

	/* end of possible globals */


	if (family == AF_INET)
		ip_flag = 0;
	else
		ip_flag = 1;

	if (service != NULL)
		if ((sp = getservbyname (service, "udp")) == NULL)
		{
			HandleError ("udp_open", "unknown service: %s/udp", service);
			return (-1);
		}

	/* 
	 * Setup socket
	 *
	 */


	if (!ip_flag)
	{

		/* IPv4 socket */

		memset((char *) &udp_srv_addr, 0, sizeof (udp_srv_addr));

		udp_srv_addr.sin_family = AF_INET;
		if (port > 0)
			udp_srv_addr.sin_port = htons (port);
		else if (service != NULL)
			udp_srv_addr.sin_port = sp->s_port;
		else
		{
			HandleError ("udp_open", "must specify either service or port");
			return (-1);
		}

		if (current_measurement)
		{
			memcpy ((char *) &udp_srv_addr.sin_addr, &(current_measurement->target.addr), sizeof(current_measurement->target.addr));
		}
		else
		{
			if ((inet_pton (AF_INET, host, &udp_srv_addr.sin_addr)) <= 0)
			{
				if ((hp = gethostbyname (host)) == NULL)
				{
					HandleError ("udp_open", "host name error: %s %s", host, host_err_str ());
					return (-1);
				}
				else
					memcpy ((char *) &udp_srv_addr.sin_addr, hp->h_addr, hp->h_length);
			}
		}

		if ((fd = socket (AF_INET, SOCK_DGRAM, 0)) < 0)
		{
			HandleError ("udp_open", "can't create UDP4 socket");
			return (-1);
		}

	}
	else
	{

		/* IPv6 socket */

		memset((char *) &udp_srv_addr6, 0, sizeof (udp_srv_addr6));
		udp_srv_addr6.sin6_family = AF_INET6;
		if (port > 0)
			udp_srv_addr6.sin6_port = htons (port);
		else if (service != NULL)
			udp_srv_addr6.sin6_port = sp->s_port;
		else
		{
			HandleError ("udp_open", "must specify either service or port");
			return (-1);
		}

		if (current_measurement)
		{
			memcpy((char *) &udp_srv_addr6.sin6_addr, &(current_measurement->target.addr6), sizeof(current_measurement->target.addr6)); 
		}
		else
		{
			if ((inet_pton (AF_INET6, host, &udp_srv_addr6.sin6_addr)) <= 0)
			{
				if ((hp = gethostbyname2 (host, AF_INET6)) == NULL)
				{
					HandleError ("udp_open", "host name error: %s %s", host, host_err_str ());
					return (-1);
				}
				else
					memcpy((char *) &udp_srv_addr6.sin6_addr, hp->h_addr, hp->h_length);
			}
		}

		if ((fd = socket (AF_INET6, SOCK_DGRAM, 0)) < 0)
		{
			HandleError ("udp_open", "can't create UDP6 socket");
			return (-1);
		}
	}

	if (port < 0)
	{
		HandleError ("udp_open", "reserved ports not implemeneted yet");
		return (-1);
	}

	/* 
	 *    Bind any local address for us.
	 *
	 */

	if (!ip_flag)
	{

		/* IPv4 */

		memset ((char *) &udp_cli_addr, 0, sizeof (udp_cli_addr));
		udp_cli_addr.sin_family = AF_INET;
		udp_cli_addr.sin_addr.s_addr = htonl (INADDR_ANY);
		udp_cli_addr.sin_port = htons (0);
		if (bind (fd, (struct sockaddr *) &udp_cli_addr, sizeof (udp_cli_addr)) < 0)
		{
			HandleError ("udp_open", "bind4 error");
			close (fd);
			return (-1);
		}

		if (dontconn == 0)
		{
			if (connect (fd, (struct sockaddr *) &udp_srv_addr, sizeof (udp_srv_addr)) < 0)
			{
				HandleError ("udp_open" "connect4 error");
				return (-1);
			}
		}


	}
	else
	{

		/* IPv6 */

		memset ((char *) &udp_cli_addr6, 0, sizeof (udp_cli_addr6));
		udp_cli_addr6.sin6_family = AF_INET6;
		udp_cli_addr6.sin6_flowinfo = 0;
		udp_cli_addr6.sin6_addr = in6addr_any;
		udp_cli_addr6.sin6_port = htons (0);
		if (bind (fd, (struct sockaddr *) &udp_cli_addr6, sizeof (udp_cli_addr6)) < 0)
		{
			HandleError ("udp_open", "bind6 error");
			close (fd);
			return (-1);
		}

		if (dontconn == 0)
		{
			if (connect (fd, (struct sockaddr *) &udp_srv_addr6, sizeof (udp_srv_addr6)) < 0)
			{
				HandleError ("udp_open" "connect6 error");
				return (-1);
			}
		}
	}

	return (fd);
}





/*
 * -------------------------------------------------------------------------------
 * Subroutine Header
 * Purpose           : Error reporting for gethostbyname and family..
 * Side Effects      :
 * Comments          : From Stevens Unix Network Programming 1990.
 * Inputparameter    : None
 * Returns 	         : String with error text.
 *
 *  Original header.
 * Return a string containing some additional information after a
 * host name or address lookup error - gethostbyname() or gethostbyaddr().
 */

int     h_errno;                /* host error number */
int     h_nerr;                 /* # of error message strings */
char    *h_errlist[0];           /* the error message table */

char               *
host_err_str (void)
{
	static char         msgstr[200];

	if (h_errno != 0)
	{
		if (h_errno > 0 && h_errno < h_nerr)
			sprintf (msgstr, "(%s)", h_errlist[h_errno]);
		else
			sprintf (msgstr, "(h_errno = %d)", h_errno);
	}
	else
	{
		msgstr[0] = '\0';
	}
	return (msgstr);
}





/*
 * -------------------------------------------------------------------------------
 * Subroutine Header
 * Purpose           : Create a  packet
 * Side Effects      :
 * Comments          : This routine sets up a test traffic packet.
 *                     The packet ellements are described in packet.h
 * Inputparameter    : a pointer to an allocated piece of memory, and the size
 *                     of the packet.
 * Returns 	         : a pointer to the position where the payload needs to go. or
 *                     NULL if unsuccesfull also set the packet header.
 */

struct test_packet_data *SetupPacket (int debug, char *packet, u_int size, long serialnumber)
{
	struct test_packet_header *header;
	struct test_packet_data *data;
	int                 maxoffset;
	long                rnd;

	{							/* Initialize Random number generator */
		struct timeval      t;

		gettimeofday (&t, NULL);
		srandom ((unsigned) (t.tv_usec));
	}

	if (size > MAX_MTU_PACKET_LENGTH)
	{
		HandleError (PNAME, "Warning: total packet size is to larger than MTU on Ethernet\n(The data fragment is %d)", size);
	}
	/* Randomize content */
	/* To prevent compression the packet content is first completely randomized 
	 */

	brandom (packet, size);

	/* setup the packet header which consists of a unique 32 bit ID that can be 
	 * used to identify the packet later and the offest from the beginning of
	 * the packet to the data */

	/* Note the use of network byte order in the packet header and data */

	header = (struct test_packet_header *) packet;
	header->packet_id = htonl (random ());

	if (header->packet_id == 0)
		header->packet_id = 1234;	/* We use 0 ad a * test value in *
									 * Datahandling.c * */
	maxoffset = (size - TEST_PACKET_DATALENGTH - TEST_PACKET_HEADERLENGTH);


	if (debug)
	{
		printf ("MAXoffset: %d (%u)  ", maxoffset, size);
	}

	/* random returns a value between 0 and 0xffffffff */
	rnd = random () & 0xff;		/* To prefent buffer overflows */
	header->offset_to_data = htons (maxoffset * rnd / 0xff);

#ifdef DEBUG
	printf ("ID %ld, %d  ", ntohl (header->packet_id), ntohs (header->offset_to_data));

#endif

	/* Fill in parts of the data */
	data = (struct test_packet_data *) (packet + TEST_PACKET_HEADERLENGTH + ntohs (header->offset_to_data));

	data->version = htons (PACKETVERSION);
	/* XXX Fingerprint could be some sort of complement of the serial number or 
	 * some other crypto grapghic thingy... for now we set it a constant */
	data->fingerprint = htonl (FINGERPRINT);
	data->serial_number = htonl (serialnumber);
	return (data);
}


void create_logfile (char *logfile, char *hostname, int family)
{
	time_t              t;
	struct tm          *timevalue;
	int                 now;
	char                datebuffer[9];	/* 'YYYYMMDD\0' */

	now = time (&t);
	/* Create logfile */
	timevalue = gmtime (&t);

	strftime (datebuffer, sizeof (datebuffer), "%Y%m%d", timevalue);
	sprintf (logfile, "%s/%s%s.%s", getenv ("TTROOT"),
			 (family == AF_INET) ? "/data/SNDP." : "/data6/SNDP6.", hostname,
			 datebuffer);
}

int record_pid (int debug)
{
	/* 
	 * Checks if the process already exists, if not write pid of this process
	 * to a file.
	 * 
	 * Returns TRUE (success) or FALSE (failure) */

	FILE               *pidfile;
	char               *filename = NULL;
	int                 pid;

	filename = (char *) malloc (strlen (getenv ("TTROOT")) + strlen ("/data/pid_list/") + strlen (PNAME) + 3);

	filename = strcpy (filename, getenv ("TTROOT"));
	filename = strcat (filename, "/data/pid_list/");
	filename = strcat (filename, PNAME);

	if (debug)
	{
		printf ("%s %d\n", filename, getpid ());
	}

	pidfile = fopen (filename, "r");
	if (pidfile == NULL)
	{
		/* pidfile does not exist */
		pidfile = fopen (filename, "w");
		if (pidfile == NULL)
		{
			/* Cannot create file */
			printf ("Cannot record PID\n");
			return FALSE;
		}
		/* Write PID and exit */
		fprintf (pidfile, "%d\n", getpid ());
		fclose (pidfile);
		return TRUE;
	}

	/* File exists, read PID */
	fscanf (pidfile, "%d", &pid);
	fclose (pidfile);
	printf ("Found a running %s with pid %d\n", PNAME, pid);
	return FALSE;
}

int remove_pid (int debug)
{
	/* 
	 * Remove PID file from the system. */

	char               *filename = NULL;

	filename = (char *) malloc (strlen (getenv ("TTROOT")) + strlen ("/data/pid_list/") + strlen (PNAME) + 3);

	filename = strcpy (filename, getenv ("TTROOT"));
	filename = strcat (filename, "/data/pid_list/");
	filename = strcat (filename, PNAME);

	if (debug)
	{
		printf ("%s\n", filename);
	}

	unlink (filename);

	return 0;
}

void exitcode (int sig)
{
	/* Code to be executed on ctrl-C and such */

	remove_pid (FALSE);
	printf ("Killed by signal %d, exit code finished\n", sig);
	exit (0);
}


int main (int argc, char *argv[])
{

	int                 result;
	int                 offset;

	long                seed = -1;
	long                packetid;

	struct target      *next_testbox;
	struct measurement *current_measurement;
	struct sockaddr    *current_measurement_srcaddr;
	socklen_t           current_measurement_srcaddr_len;

	time_t              t;

	/* struct tm *timevalue; */
	int                 now,
	                    interval;
	double              start = 0;

	char                c;
	char               *poissonfile = NULL;
	char                logfile4[MAXLINE];
	char                logfile6[MAXLINE];
	char                hostname[MAXLINE];
	char               *configfile = NULL;
	char                addrbuffer[SOCK_MAXADDRLEN];

	int                 debug = FALSE;
	int                 nameinfoflags = 0;

	FILE               *logpoint4;
	FILE               *logpoint6;

	struct etherifdata  eid;
	char                errbuf[256];

	struct logpoint_lookuptable
	{
		FILE               *logpoint;
		union
		{
			struct in_addr      ethif_addr;	/* The IP number it thinks it has */
			struct in6_addr     ethif_addr6;
		}
		addr;
		char                hname[NI_MAXHOST];
	}
	                   *logpoint4table, *logpoint6table;
	int                 logpoint4tableentries,
	                    logpoint6tableentries;


	/* Get hostname */
	gethostname (hostname, MAXLINE);

	/* Read options */
	while ((c = getopt (argc, argv, "nvf:p:")) != -1)
	{
		switch (c)
		{
		case 'v':
			debug = TRUE;
			break;
		case 'n':
			nameinfoflags |= NI_NUMERICHOST;
			break;
		case 'f':
			configfile = optarg;
			break;
		case 'p':
			poissonfile = optarg;
			break;
		default:
			break;
		}
	}

	logpoint4tableentries = logpoint6tableentries = 0;
	logpoint4table = logpoint6table = NULL;


	/* Catch signals and do clean up before exiting */
	signal (SIGINT, exitcode);
	signal (SIGKILL, exitcode);
	signal (SIGTERM, exitcode);

	/* Record PID */
	if (record_pid (debug) == FALSE)
		return 0;

	/* Main loop */

	while (1)
	{
		int                 i;

		/* Read configfile */
		result = read_configfile (debug, configfile);
		read_test_box_config (NULL, debug);

		/* 
		 * Iterate over interfaces, create internal addr -> hostname lookup table
		 * We use this to construct the filenames for the SNDP* files. We also scan 
		 * configured aliases XXX Use SendPacket with -n and do post filtering of
		 * collect_data if host lookup is unreliable! 
		 */

		while (NextEthernetIFInfo (&eid, errbuf, ETHIFINFO_ALIASES_TOO) == 0)
		{
			char                asciiaddr[256];
			int                 error;
			struct addrinfo     hints,
							   *res;

			if (!(eid.ethif_addr.s_addr == INADDR_ANY))
			{
				struct logpoint_lookuptable *tmp;

				/* realloc logpoint4table, construct filename and open */
				if ((tmp = realloc (logpoint4table, (logpoint4tableentries + 1) * sizeof (struct logpoint_lookuptable))) != NULL)
				{
					const char *ptr;
					memset (&hints, 0, sizeof (hints));
					hints.ai_family = PF_UNSPEC;
					hints.ai_socktype = SOCK_STREAM;
					hints.ai_flags = AI_CANONNAME;

					logpoint4table = tmp;
					logpoint4table[logpoint4tableentries].addr.ethif_addr = eid.ethif_addr;
					logpoint4table[logpoint4tableentries].logpoint = NULL;

					inet_ntop (AF_INET, &eid.ethif_addr, asciiaddr, sizeof (asciiaddr));

					if ((error = getaddrinfo (asciiaddr, NULL, &hints, &res)) != 0)
					{
						HandleError (PNAME, "getaddrinfo");
						exit (EX_OSERR);
					}
					/* 
					 * If we can lookup the hostname through the list of
					 * testboxes, use that name, otherwise silently fall back to the ip address 
					 */

					if ((ptr = test_box_ip_name(asciiaddr,AF_INET)) != NULL )
						strncpy(logpoint4table[logpoint4tableentries].hname,ptr, NI_MAXHOST);
					else
						strncpy(logpoint4table[logpoint4tableentries].hname,asciiaddr, NI_MAXHOST);
					freeaddrinfo (res);
					logpoint4tableentries++;
				}
				else
				{
					HandleError (PNAME, "realloc failed");
					abort ();
				}
			}
			if (!(IN6_IS_ADDR_UNSPECIFIED (&eid.ethif_addr6)))
			{
				struct logpoint_lookuptable *tmp;

				/* realloc logpoint6table, construct filename and open */
				if ((tmp = realloc (logpoint6table, (logpoint6tableentries + 1) * sizeof (struct logpoint_lookuptable))) != NULL)
				{
					const char *ptr;
					memset (&hints, 0, sizeof (hints));
					hints.ai_family = PF_UNSPEC;
					hints.ai_socktype = SOCK_STREAM;
					hints.ai_flags = AI_CANONNAME;

					logpoint6table = tmp;
					logpoint6table[logpoint6tableentries].addr.ethif_addr6 = eid.ethif_addr6;
					logpoint6table[logpoint6tableentries].logpoint = NULL;

					inet_ntop (AF_INET6, &eid.ethif_addr6, asciiaddr, sizeof (asciiaddr));

					if ((error = getaddrinfo (asciiaddr, NULL, &hints, &res)) != 0)
					{
						HandleError (PNAME, "getaddrinfo");
						exit (EX_OSERR);
					}

					/* FIXME This will die when an ipv6 address is autoconfigured */
					if ((ptr = test_box_ip_name(asciiaddr,AF_INET6)) != NULL ) 
						strncpy(logpoint6table[logpoint6tableentries].hname,ptr, NI_MAXHOST);
					else
					{
						HandleError (PNAME, "test_box_ip_name");
						exit (EX_OSERR);
					}
					freeaddrinfo (res);
					logpoint6tableentries++;
				}
				else
				{
					HandleError (PNAME, "realloc failed");
					abort ();
				}
			}
		}
		EndEthernetIFInfo ();

		if (debug)
		{
			int                 ii;

			for (ii = 0; ii < logpoint4tableentries; ii++)
			{
				/* XXX Check for .ripe.net substring ? */
				create_logfile (logfile4, logpoint4table[ii].hname, AF_INET);
				printf ("%s\n", logfile4);
			}

			for (ii = 0; ii < logpoint6tableentries; ii++)
			{
				/* XXX Check for .ripe.net substring ? */
				create_logfile (logfile6, logpoint6table[ii].hname, AF_INET6);
				printf ("%s\n", logfile6);
			}
		}

		/* Schedule measurements */
		if (result == -1)
		{
			/* Failure on reading config file */
			return 0;
		}
		else if (result == -2)
		{
			/* No modifications in config file */
			current_measurement = first_measurement;

			while (current_measurement != NULL)
			{
				current_measurement->time = (double) globaltime.period + current_measurement->time;
				current_measurement = current_measurement->next;
			}

			if (debug)
				list_measurement (first_measurement);
			start += (double) globaltime.period;
		}
		else
		{

			/* Create poisson */

			next_testbox = first_testbox;
			seed = poisson_getseed (debug, poissonfile);
			offset = 0;

			while (next_testbox != NULL)
			{
				seed = do_poisson (debug, next_testbox, offset, seed);
				next_testbox = next_testbox->next;
				offset++;
			}

			poisson_saveseed (debug, poissonfile, seed);

			/* Schedule measurements */
			first_measurement = NULL;
			current_measurement = NULL;
			next_testbox = first_testbox;

			while (next_testbox != NULL)
			{
				offset = 0;
				current_measurement = first_measurement;
				while (offset < next_testbox->rate)
				{
					if (first_measurement == NULL)
					{
						/* 
						 * Empty string, insert first element
						 */
						first_measurement = create_measurement (next_testbox, offset, NULL);
						current_measurement = first_measurement;
						offset++;
					}
					else if (*(next_testbox->time + offset) < first_measurement->time)
					{
						/* 
						 * Insert at the front
						 */
						first_measurement = create_measurement (next_testbox, offset, first_measurement);
						current_measurement = first_measurement;
						offset++;
					}
					else if (current_measurement->next == NULL)
					{
						/* 
						 * End of string, add at the end
						 */
						current_measurement->next = create_measurement (next_testbox, offset, NULL);
						offset++;
					}
					else if ((current_measurement->next)->time > *(next_testbox->time + offset))
					{
						/* 
						 * Fits in the middle
						 */
						current_measurement->next = create_measurement (next_testbox, offset, current_measurement->next);
						offset++;
					}
					else
					{
						/* 
						 * Next element in chain
						 */
						current_measurement = current_measurement->next;
					}
				}
				next_testbox = next_testbox->next;
			}

			/* Measurements so-far */
			if (debug)
				list_measurement (first_measurement);

			/* Add offset to beginning of measurement period */
			now = time (&t);
			interval = ((now - globaltime.start) / globaltime.period) + 1;
			start = (double) (globaltime.start + interval * globaltime.period);

			current_measurement = first_measurement;
			while (current_measurement != NULL)
			{
				current_measurement->time = start + current_measurement->time / 1000.0;
				current_measurement = current_measurement->next;
			}
			if (debug)
				list_measurement (first_measurement);
		}

		/* Start of schedule */
		now = time (&t);

		if (debug)
			printf ("Now is %d, Schedule starts at %f\n", now, start);

		sleep_until (debug, start);

		logpoint4 = NULL;
		logpoint6 = NULL;

		/* Send packets */
		current_measurement = first_measurement;
		while (current_measurement != NULL)
		{
			sleep_until (debug, current_measurement->time);

			if (debug)
			{
				printf ("Sending packet to %s, port %d, size %d family %d\n\n",
						current_measurement->name,
						current_measurement->port, current_measurement->size, current_measurement->family);
			}

			current_measurement_srcaddr = (struct sockaddr *) addrbuffer;
			current_measurement_srcaddr_len = SOCK_MAXADDRLEN;	/* Gets overwritten by getsockname later on, be sure to avoid truncation by reinitializing it here */

			packetid = SendPacket (debug, current_measurement, current_measurement_srcaddr, &current_measurement_srcaddr_len);

			if (packetid > 0 && (current_measurement->family == AF_INET))
			{
				int                 j;

				/* Lookup filehandle for current_measurement_srcaddr */
				logpoint4 = NULL;

				for (j = 0; j < logpoint4tableentries; j++)
				{
					if (logpoint4table[j].addr.ethif_addr.s_addr ==
						(((struct sockaddr_in *) current_measurement_srcaddr)->sin_addr.s_addr))
					{
						/* XXX Check for .ripe.net substring ? */
						create_logfile (logfile4, logpoint4table[j].hname, AF_INET);

						/* Not filled in yet, open file */
						if ((logpoint4 = logpoint4table[j].logpoint) == NULL)
						{
							logpoint4 = fopen (logfile4, "a");

							if (debug)
								printf ("%s\n", logfile4);

							if (logpoint4 == NULL)
							{
								HandleError (PNAME,
											 "Cannot open file for logging v4-measurements from %s",
											 logpoint4table[j].hname);
							}

							logpoint4table[j].logpoint = logpoint4;
						}
						break;
					}
				}
				if (logpoint4)
				{

					if (current_measurement->burst > 1)
						fprintf (logpoint4, "SNDP 11 ");
					else
						fprintf (logpoint4, "SNDP 9 ");

					fprintf (logpoint4, "%d -h %s -p %d -n %d -s %ld",
							 (int) current_measurement->time,
							 current_measurement->name, current_measurement->port, current_measurement->size, packetid);

					if (current_measurement->burst > 1)
					{
						fprintf (logpoint4, " -b %d\n", current_measurement->burst);
					}
					else
					{
						fprintf (logpoint4, "\n");
					}

					fflush (logpoint4);
				}
			}

			if (packetid > 0 && (current_measurement->family == AF_INET6))
			{

				/* Lookup filehandle for current_measurement_srcaddr */
				logpoint6 = NULL;

				for (i = 0; i < logpoint6tableentries; i++)
				{
					if (IN6_ARE_ADDR_EQUAL
						(&(logpoint6table[i].addr.ethif_addr6),
						 &(((struct sockaddr_in6 *) current_measurement_srcaddr)->sin6_addr)))
					{
						/* XXX Check for .ripe.net substring ? */
						create_logfile (logfile6, logpoint6table[i].hname, AF_INET6);
						if ((logpoint6 = logpoint6table[i].logpoint) == NULL)
						{
							logpoint6 = fopen (logfile6, "a");
							if (debug)
								printf ("%s\n", logfile6);

							if (logpoint6 == NULL)
							{
								HandleError (PNAME,
											 "Cannot open file for logging v6-measurements from %s",
											 logpoint6table[i].hname);
							}
							logpoint6table[i].logpoint = logpoint6;
						}
						break;
					}
				}
				if (logpoint6)
				{
					if (current_measurement->burst > 1)
						fprintf (logpoint6, "SNDP6 11 ");
					else
						fprintf (logpoint6, "SNDP6 9 ");

					fprintf (logpoint6, "%d -h %s -p %d -n %d -s %ld",
							 (int) current_measurement->time,
							 current_measurement->name, current_measurement->port, current_measurement->size, packetid);

					if (current_measurement->burst > 1)
					{
						fprintf (logpoint6, " -b %d\n", current_measurement->burst);
					}
					else
					{
						fprintf (logpoint6, "\n");
					}
					fflush (logpoint6);
				}
			}
			current_measurement = current_measurement->next;
		}

		/* Close up for the next round of measurements */

		for (i = 0; i < logpoint4tableentries; i++)
		{
			if (logpoint4table[i].logpoint)
				fclose (logpoint4table[i].logpoint);
			logpoint4table[i].logpoint = NULL;
		}
		free(logpoint4table);

		for (i = 0; i < logpoint6tableentries; i++)
		{
			if (logpoint6table[i].logpoint)
				fclose (logpoint6table[i].logpoint);
			logpoint6table[i].logpoint = NULL;
		}
		free(logpoint6table);

		free_test_box_config();

		logpoint4tableentries = logpoint6tableentries = 0;
		logpoint4table = logpoint6table = NULL;
	}
	remove_pid (debug);
	return 0;
}
