#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include <sysexits.h>
#include <sys/time.h>
#include <unistd.h>
#include <sys/param.h>
#include <netdb.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "alarm.h"
#include "testbox.h"
#include "libdelayalarm.h"

struct box_histogram
{
    unsigned int        boxid;
    int                 entries[MAX_DELAY / DELAY_BIN_WIDTH];
};

/* Long term results per box and time-bin */
struct box_histogram *long_term_entries = NULL;
struct box_histogram *short_term_entries = NULL;

#ifdef __KAME__
struct box_histogram *long_term_entries6 = NULL;
struct box_histogram *short_term_entries6 = NULL;
#endif /* __KAME__ */

size_t             *id2indexV4,
                    id2indexV4_size,
                   *id2indexV6,
                    id2indexV6_size;

time_t sta_assign_time_intervals (long int*, long int*, int*, int*, long int*, long int*, int);
void sta_read_history (char*, long int, long int, int, int, int, int);
int decode_received_packets (char**, const void *);
void do_percentiles (int, int*, float*);
int sta_rw_status (const char*, const char*, int*, size_t, int, int);
char *sta_status_string (int, char*);
int sta_status_change (int, int, int, time_t,
                       float*, float*, int, const char*);



time_t sta_assign_time_intervals (long int *start, long int *end,
                                  int *start_hour, int *end_hour, long int *short_start, long int *short_end, int debug)
{

/* Assign time intervals. Returns: 

  start, end              : start/end time for the long term interval.
                            (today at 0:00, today-30 days at 0:00)
  start_hour, end_hour    : start hour and end hour  for long term interval.
                            (0/6/12/18)-(6/12/18/24) with 4 intervals/day.
  short_start, short_end  : start and end time for short term interval.
  if debug == TRUE then print additional debug output.
*/

    time_t              systime;    /* Unix System time */
    struct tm          *tmvalue;    /* Human time */

    /* XXX Add iso8601period barriers */

    systime = time (&systime);
    tmvalue = localtime (&systime);

    /* Show when the program is run */
    if (debug == TRUE)
    {
        printf ("At %lu %d/%d/%d %d:%d:%d\n", (unsigned long int)systime,
                tmvalue->tm_mday, tmvalue->tm_mon + 1, tmvalue->tm_year, tmvalue->tm_hour, tmvalue->tm_min, tmvalue->tm_sec);
    }

    /* Interval */
    *short_start = systime - SHORT_TERM_INTERVAL;
    *short_end = systime;

    *start_hour = (24 / INTERVAL_DAY);
    *start_hour = (tmvalue->tm_hour / (*start_hour)) * (24 / INTERVAL_DAY);

    *end_hour = (24 / INTERVAL_DAY);
    *end_hour = ((tmvalue->tm_hour / (*end_hour)) + 1) * (24 / INTERVAL_DAY);

    tmvalue->tm_sec = 0;
    tmvalue->tm_min = 0;
    tmvalue->tm_hour = 0;

    *end = mktime (tmvalue);
    *start = *end - DAYS_HISTORY * 86400;

    return systime;
}


#define MAXWORDS 5000
void sta_read_history (char *filename, long int start_day, long int end_day, int start_hour, int end_hour, int debug, int family)
{

/*   ================
     Read history file from start_day,start_hour to end day and hour.
*/

    FILE               *fp;
    int                 j,
                        from,
                        to,
                        testbox,
                        testbox_idx;
    time_t              start,
                        end;

    char                line[MAXWORDS];
    char               *point = &line[0];
    char              **ap;
    char               *arg[MAXWORDS];

    struct tm          *tmval;

    fp = fopen (filename, "r");

    if (fp == NULL)
    {
        if (debug == TRUE)
        {
            if (family == AF_INET)
                printf ("No history file found (IPv4)...\n");
#ifdef __KAME__
            else
                printf ("No history file found (IPv6)...\n");
#endif /* __KAME__ */
        }
    }
    else
    {
        if (debug == TRUE)
        {
            if (family == AF_INET)
                printf ("Reading input (IPv4)...\n");
#ifdef __KAME__
            else
                printf ("Reading input (IPv6)...\n");
#endif /* __KAME__ */
        }
        point = fgets (line, MAXWORDS, fp);
        while ((point = fgets (line, MAXWORDS, fp)) != NULL)
        {
            for (ap = arg; (*ap = strsep (&point, " \t")) != NULL;)
            {
                if (**ap != '\0')
                {
                    ++ap;
                }
            }
            testbox = atoi (arg[0]);
            testbox_idx =
                family == AF_INET ? lookup_id2index (id2indexV4, testbox, id2indexV4_size) : lookup_id2index (id2indexV6, testbox,
                                                                                                              id2indexV6_size);
            start = atoi (arg[1]);
            end = atoi (arg[2]) - 1;

            tmval = localtime (&start);
            from = tmval->tm_hour;

            tmval = localtime (&end);
            to = tmval->tm_hour;

            if (testbox_idx != -1 && (start >= start_day) && (end <= end_day) && (from >= start_hour) && (to <= end_hour))
            {
                for (j = 0; j < MAX_DELAY / DELAY_BIN_WIDTH; j++)
                {
                    if (arg[j + 3] == NULL)
                    {
                        printf ("Line truncated: file %s", filename);
                        exit (65);    /* EX_DATAERR data format error */
                    }
                    if (family == AF_INET)
                        long_term_entries[testbox_idx].entries[j] += atoi (arg[j + 3]);
#ifdef __KAME__
                    else
                        long_term_entries6[testbox_idx].entries[j] += atoi (arg[j + 3]);
#endif /* __KAME__ */
                }
            }
            else if (debug && testbox_idx == -1)
                printf ("Unknown (stale) testbox id %d in %s found, ignoring\n", testbox, filename);
        }
        fclose (fp);
    }
    return;
}

#undef MAXWORDS

/* Parameter data hook for getRCDPData callback hook */

struct decode_received_packets_param
{
    time_t              start_time,
                        end_time;
    int                 debug,
                        family;
};

/* getRCDPData callback hook */

int decode_received_packets (char **arg, const void *read_received_packets_data)
{

/*  =====================
    Decode RCDP line
*/

    int                 srcid,
                        idx,
                        bin,
                        clock_in,
                        clock_out;
    float               delay;
    time_t				arrival;

    /* XXX contains start_time, end_time, debug and family */
    const struct decode_received_packets_param *param = (const struct decode_received_packets_param *) read_received_packets_data;


    /* Check if packet is in the right interval here, so we can skip all futher 
     * checks... */
    arrival = (time_t) atof (arg[7]);

    if ((param->start_time < arrival) && (param->end_time > arrival))
    {
        /* Check for valid clock words */
        clock_in = strtoul (arg[9], NULL, 16);
        clock_out = strtoul (arg[10], NULL, 16);

        if (ClocksOK (clock_in, clock_out))
        {

            /* Convert delays into ms and check for valid range */
            delay = atof (arg[8]) * 1000.0;

            if (delay >= 0.0)
            {
                if (param->family == AF_INET || param->family == AF_INET6 )
                {
                    srcid = test_box_ip_id (arg[3], param->family);
                    if (delay > (MAX_DELAY - DELAY_BIN_WIDTH))
                    {
                        delay = MAX_DELAY - DELAY_BIN_WIDTH;
                    }
                    bin = (int) (delay / (float) DELAY_BIN_WIDTH);
                    if (srcid > 0)
                    {
                        if (param->family == AF_INET)
                        {
                            if ((idx = lookup_id2index (id2indexV4, srcid, id2indexV4_size)) != -1)
                                short_term_entries[idx].entries[bin]++;
                        }
#ifdef __KAME__
                        else
                        {
                            if ((idx = lookup_id2index (id2indexV6, srcid, id2indexV6_size)) != -1)
                                short_term_entries6[idx].entries[bin]++;
                        }
#endif /* __KAME__ */
                    }
                }
            }
        }
    }
    return TRUE;
}

#undef MAXWORDS

void do_percentiles (int npoint, int *data, float *results)
{

/*   ==============
     Find percentiles in an array data[] and return number of entries and bin.
*/
    int                 i = 0;
    int                 sum = 0;

    results[0] = 0.0;            /* Total Number of entries in the array */
    results[LOWER_PERCENTILE_BIN] = 0.0;            /* Bin with LOWER_PERCENTILE entries */
    results[MEDIAN_PERCENTILE_BIN] = 0.0;            /* Bin with MEDIAN_PERCENTILE entries */
    results[HIGHER_PERCENTILE_BIN] = 0.0;            /* Bin with HIGHER_PERCENTILE entries */

    for (i = 0; i < npoint; i++)
    {
        results[0] += data[i];
    }

    if (results[0] != 0.0)
    {
        for (i = 0; i < npoint; i++)
        {
            sum += data[i];
            if (sum <= LOWER_PERCENTILE * results[0])
            {
                results[LOWER_PERCENTILE_BIN] = (float) i;
            }
            if (sum <= MEDIAN_PERCENTILE * results[0])
            {
                results[MEDIAN_PERCENTILE_BIN] = (float) i;
            }
            if (sum <= HIGHER_PERCENTILE * results[0])
            {
                results[HIGHER_PERCENTILE_BIN] = (float) i;
            }
        }
        results[LOWER_PERCENTILE_BIN] = (0.0 + results[LOWER_PERCENTILE_BIN]) * (float) DELAY_BIN_WIDTH;
        results[MEDIAN_PERCENTILE_BIN] = (0.5 + results[MEDIAN_PERCENTILE_BIN]) * (float) DELAY_BIN_WIDTH;
        results[HIGHER_PERCENTILE_BIN] = (1.0 + results[HIGHER_PERCENTILE_BIN]) * (float) DELAY_BIN_WIDTH;
    }
    return;
}

int sta_rw_status (const char *filename, const char *action, int *status, size_t elements, int family, int debug)
{

/*  =============
    Read or write file with current alarm status 
 
  filename : where to write the file
  action   : "r" or "w" for read or write
  status   : status for all test-boxes

  returns TRUE if successful, FALSE otherwise
*/

    FILE               *fp;
    int                 j,
                        k,
                        nread;
	size_t				i;

    if ((strcmp (action, "w") == 0) || (strcmp (action, "r") == 0))
    {
        fp = fopen (filename, action);
        if (fp != NULL)
        {
            /* Write file */
            if (strcmp (action, "w") == 0)
            {
                for (i = 0; i < elements; i++)
                {
                    if (status[i] != STA_UNDEFINED)
                    {
                        fprintf (fp, "%d %d\n", family == AF_INET ? id2indexV4[i] : id2indexV6[i], status[i]);
                    }
                }
                /* Read file */
            }
            else
            {
                for (i = 0; i < elements; i++)
                {
                    status[i] = STA_UNDEFINED;
                }
                while ((nread = fscanf (fp, "%d %d", &j, &k)) > 0)
                {
                    int                 idx =
                        family == AF_INET ? lookup_id2index (id2indexV4, j, id2indexV4_size) : lookup_id2index (id2indexV6, j,
                                                                                                                id2indexV6_size);

                    if (idx != -1)
                        status[idx] = k;
                    else if (debug)
                        printf ("Unknown (stale) testbox id %d in %s found, ignoring\n", j, filename);
                }
            }
            fclose (fp);
            return TRUE;

            /* Error conditions */
        }
        else
        {
            if (debug == TRUE)
            {
                printf ("sta_rw_status: cannot open file\n");
            }
        }
    }
    else
    {
        if (debug == TRUE)
        {
            printf ("sta_rw_status: unknown action!\n");
        }
    }

    /* Error, Set all entries to undefined and return FALSE */
    for (i = 0; i < elements; i++)
    {
        status[i] = STA_UNDEFINED;
    }
    return FALSE;
}

char               *sta_status_string (int status, char *point)
{

/*    =================
      Return ascii string for each error code
*/

    /* XXX FIXME return entries from static array instead of this error prone strcpy'ing */

    switch (status)
    {
    case STA_NOALARM:
        point = strcpy (point, "No Alarm");
        break;
    case STA_ALARM:
        point = strcpy (point, "Alarm");
        break;
    case STA_NOLONG:
        point = strcpy (point, "No Long Data");
        break;
    case STA_NOSHORT:
        point = strcpy (point, "No Short Data");
        break;
    case STA_SPREAD:
        point = strcpy (point, "Larger Spread");
        break;
    case STA_IMPROVED:
        point = strcpy (point, "Improving");
        break;
    case STA_GOUP:
        point = strcpy (point, "Go Up");
        break;
    case STA_GODOWN:
        point = strcpy (point, "Go Down");
        break;
    case STA_UNDEFINED:
        point = strcpy (point, "Undefined");
        break;
    default:
        point = strcpy (point, "No such code");
        break;
    }
    return point;

}

int sta_status_change (int old_status, int new_status, int testbox, time_t now,
                       float *long_res, float *short_res, int debug, const char *family)
{

/*  =================
    Print message if status changed.
*/

    char                old_str[100];
    char                new_str[100];

    if ((old_status != new_status))
    {
        printf ("TB %2d proto %s at %lu Status change old: %s, new: %s\n", testbox, family, (unsigned long int)now,
                sta_status_string (old_status, old_str), sta_status_string (new_status, new_str));
        debug = TRUE;
    }

    if ((old_status != new_status) && (new_status == STA_ALARM))
    {
        printf ("TB %2d proto %s at %lu ALARM SET   old: %s, new: %s\n", testbox, family, (unsigned long int)now,
                sta_status_string (old_status, old_str), sta_status_string (new_status, new_str));
        debug = TRUE;
    }
    else if ((old_status != new_status) && (old_status == STA_ALARM))
    {
        printf ("TB %2d proto %s at %lu ALARM CLEARED old: %s, new: %s\n", testbox, family, (unsigned long int)now,
                sta_status_string (old_status, old_str), sta_status_string (new_status, new_str));
        debug = TRUE;
    }

    if (debug == TRUE)
    {
        printf ("TB %2d proto %s at %lu Long: %5.0f", testbox, family, (unsigned long int)now, long_res[0]);
        if (new_status == STA_NOLONG)
        {
            printf ("\n");
            return new_status;
        }
        printf (" %5.1f/%5.1f/%5.1f Short: %5.0f", long_res[1], long_res[2], long_res[3], short_res[0]);
        if (new_status == STA_NOSHORT)
        {
            printf ("\n");
            return new_status;
        }
        printf (" %5.1f/%5.1f/%5.1f\n", short_res[1], short_res[2], short_res[3]);
    }

    return new_status;
}

int main (int argc, char *argv[])
{

    char                hostname[MAXHOSTNAMELEN],
                        ch,
                       *histpath = NULL,
                       *rcdppath = NULL,
                       *confpath = NULL,
                *iso8601period=NULL;
    char                configfile[256] = "/";
    char                alarmconfigfile[256] = "/";
    char                history[256] = "/";
    char                statusf[256] = "/";

#ifdef __KAME__
    char                history6[256] = "/";
    char                statusf6[256] = "/";
#endif /* __KAME__ */



    size_t              i,
                        v6_cnt,
                        v4_cnt;
    long int            start_day,
                        end_day,
                        short_start,
                        short_end;
    int                 start_hour,
                        end_hour,
                       *status,
                        ip_mode,
                        total_active,
                        delay_alarm;

    int                 npoints = MAX_DELAY / DELAY_BIN_WIDTH;
    int                 debug = FALSE;

    float             **long_res,
                      **short_res;

#ifdef __KAME__
    int                *status6;
    float             **long_res6,
                      **short_res6;
#endif /* __KAME__ */
    time_t              now;

    /* command line arguments */
    while ((ch = getopt (argc, argv, "vhc:o:i:p:")) != -1)
    {
        switch (ch)
        {
        case 'v':
            debug = TRUE;
            break;
        case 'o':
            histpath = optarg;
            break;
        case 'c':
            confpath = optarg;
            break;
        case 'i':
            rcdppath = optarg;
            break;
        case 'p':
            iso8601period = optarg;
            puts("Not implemented yet");
            break;
        case 'h':
            printf("%s: -v -h -c <configpath> -o <outputpath> -i <inputpath> -p <iso8601period>\n",strrchr(argv[0],'/') ? (strrchr(argv[0],'/') + sizeof(char)): argv[0]);
            exit(EX_USAGE);
            break;
        default:
            break;
        }
    }

    /* Filenames */

    snprintf (configfile, sizeof (configfile), "%s/%s", (confpath ? confpath : PATH_CONFIG), LIST_OF_TESTBOXES);
    snprintf (alarmconfigfile, sizeof (alarmconfigfile), "%s/%s", (confpath ? confpath : PATH_CONFIG), FILE_ALARM_CNF);

    snprintf (history, sizeof (history), "%s/%s", (histpath ? histpath : PATH_HISTORY), FILE_IN);
    snprintf (statusf, sizeof (statusf), "%s/%s", (histpath ? histpath : PATH_HISTORY), FILE_STAT);

#ifdef __KAME__
    snprintf (history6, sizeof (history6), "%s/%s", (histpath ? histpath : PATH_HISTORY), FILE_IN6);
    snprintf (statusf6, sizeof (statusf6), "%s/%s", (histpath ? histpath : PATH_HISTORY), FILE_STAT6);
#endif /* __KAME__ */


    /* Who am I */

    gethostname (hostname, MAXHOSTNAMELEN);


    if (debug == TRUE)
    {
        printf ("<%s>\n", hostname);
        printf ("<%s>\n", history);
        printf ("<%s>\n", statusf);
#ifdef __KAME__
        printf ("<%s>\n", history6);
        printf ("<%s>\n", statusf6);
#endif /* __KAME__ */
    }

    /* Read test-box info */

    read_test_box_config (configfile, debug);
    id2indexV4 = create_id2index (AF_INET, &id2indexV4_size);
    id2indexV6 = create_id2index (AF_INET6, &id2indexV6_size);

    v6_cnt = v4_cnt = 0;
    ip_mode = 0;

    /* Determine which protocols are used on this testbox */

    for (i = 0; i < number_of_boxes; i++)
    {
        if (test_box_list[i].family == AF_INET ) 
        {
            ip_mode |= 1;
            v4_cnt++;
        }
#ifdef __KAME__
        else if (test_box_list[i].family == AF_INET6 ) 
        {
            ip_mode |= 2;
            v6_cnt++;
        }
#endif /* __KAME__ */
        else 
        {
            /* NOP XXX */
            assert(0);
        }
    }

    /* 
     * XXX assert(( v4_cnt == id2indexV4_size ) && ( v6_cnt == id2indexV6_size )) 
     */
    
    if (debug)
    {
        printf("(v4_cnt=%d;id2indexV4_size=%d)\n",v4_cnt,id2indexV4_size);
        assert(( v4_cnt == id2indexV4_size )) ;
#ifdef __KAME__
        printf("(v6_cnt=%d;id2indexV6_size=%d)\n",v6_cnt,id2indexV6_size);
        assert(( v6_cnt == id2indexV6_size )) ;
#endif /* __KAME__ */
    }

    /* Allocate and initialize data structures which depend on no# of testboxes */

    long_term_entries = malloc ((v4_cnt + 1) * sizeof (struct box_histogram));
    short_term_entries = malloc ((v4_cnt + 1) * sizeof (struct box_histogram));
    memset(long_term_entries, 0, (v4_cnt + 1) * sizeof (struct box_histogram) );
    memset(short_term_entries, 0 , (v4_cnt + 1) * sizeof (struct box_histogram));

    status = malloc (v4_cnt * sizeof (int));
    long_res = malloc ((v4_cnt) * sizeof (float *));
    short_res = malloc ((v4_cnt) * sizeof (float *));
    for (i = 0; i < v4_cnt; i++)
    {
        long_res[i] = malloc (4 * sizeof (float));
        memset(long_res[i],0,4 * sizeof (float));
        short_res[i] = malloc (4 * sizeof (float));
        memset(short_res[i],0,4 * sizeof (float));
    }

#ifdef __KAME__
    long_term_entries6 = malloc ((v6_cnt + 1) * sizeof (struct box_histogram));
    short_term_entries6 = malloc ((v6_cnt + 1) * sizeof (struct box_histogram));
    memset(long_term_entries6, 0, (v6_cnt + 1) * sizeof (struct box_histogram) );
    memset(short_term_entries6, 0 , (v6_cnt + 1) * sizeof (struct box_histogram));
    
    
    status6 = malloc (v6_cnt * sizeof (int));
    long_res6 = malloc ((v6_cnt) * sizeof (float *));
    short_res6 = malloc ((v6_cnt) * sizeof (float *));
    for (i = 0; i < v6_cnt; i++)
    {
        long_res6[i] = malloc (4 * sizeof (float));
        memset(long_res6[i],0,4 * sizeof (float));
        short_res6[i] = malloc (4 * sizeof (float));
        memset(short_res6[i],0,4 * sizeof (float));
    }
#endif /* __KAME__ */


    /* Assign time intervals */
    now = sta_assign_time_intervals (&start_day, &end_day, &start_hour, &end_hour, &short_start, &short_end, debug);
    if ((ip_mode & 1) == 1)
    {
        struct alarm_config *alarm_cfg4;
        struct decode_received_packets_param param = { short_start, short_end , debug, AF_INET };

        /* Read configuration */
        
        load_alarm_config(alarmconfigfile, &alarm_cfg4, id2indexV4, id2indexV4_size, AF_INET, debug);

        /* Read history so far */
        sta_read_history (history, start_day, end_day, start_hour, end_hour, debug, AF_INET);

        /* Read status */
        sta_rw_status (statusf, "r", &status[0], v4_cnt, AF_INET, debug);

        /* Read input file */
        getRCDPData (hostname, short_start, (short_end - short_start), AF_INET, decode_received_packets, &param);

        /* Process output */
        total_active = 0;
        delay_alarm = 0;
        for (i = 0; i < id2indexV4_size; i++)
        {
            do_percentiles (npoints, &long_term_entries[i].entries[0], &long_res[i][0]);

            /* printf("do_percentiles(npoints=%d,long_term_entries[%d].entries[0]=%d,long_res[%d][0])=%f);\n",npoints,i,long_term_entries[i].entries[0],i,long_res[i][0]); */

            
            do_percentiles (npoints, &short_term_entries[i].entries[0], &short_res[i][0]);

            /* printf("do_percentiles(npoints=%d,short_term_entries[%d].entries[0]=%d,short_res[%d][0])=%f);\n",npoints,i,short_term_entries[i].entries[0],i,short_res[i][0]); */

            /* Test box without any data */
            if ((long_res[i][0] == 0) && (short_res[i][0] == 0))
            {
                continue;
            }

            /* Box with data */
            total_active++;

            if (long_res[i][0] < MIN_POINTS_LONG)
            {
                status[i] =
                    sta_status_change (status[i], STA_NOLONG, id2indexV4[i], now, &long_res[i][0], &short_res[i][0], debug, "IPv4");
                continue;
            }

            if (short_res[i][0] < MIN_POINTS_SHORT)
            {
                status[i] =
                    sta_status_change (status[i], STA_NOSHORT, id2indexV4[i], now, &long_res[i][0], &short_res[i][0], debug,
                                       "IPv4");
                continue;
            }

            /* Do absolute alarm stuff here*/ 
            if (alarm_cfg4[i].a_type != UNDEFINED_TYPE)
            {
                /* use median percentile for the "alert bin" at the moment */

                int raise_alert=0, alertbin=MEDIAN_PERCENTILE_BIN;
                float abs_alert_lvl=alarm_cfg4[i].value;
                switch(alarm_cfg4[i].a_type)
                {
                    case UNDEFINED_TYPE:    
                        if (debug)
                            printf ("Possible parse error in absolute alarm configuration for box %d\n",id2indexV4[i]);
                        break;
                    case PERCENT:
                        /* recalculate percent into absolute value and fall through */
                        abs_alert_lvl = ((long_res[i][alertbin] / 100) * abs_alert_lvl);
                    case ABSOLUTE:
                        switch(alarm_cfg4[i].a_lvl)
                        {
                            case UNDEFINED_LVL:
                                if (debug)
                                    printf ("Possible parse error in absolute alarm configuration for box %d\n",id2indexV4[i]);
                                break;
                            case EQUAL:
                                raise_alert = fabsf(long_res[i][alertbin] - short_res[i][alertbin]) == abs_alert_lvl ? 1 : 0;
                                break;    
                            case NOT_EQUAL:
                                raise_alert = fabsf(long_res[i][alertbin] - short_res[i][alertbin]) != abs_alert_lvl ? 1 : 0;
                                break;    
                            case EQUAL_LESS:
                                raise_alert = fabsf(long_res[i][alertbin] - short_res[i][alertbin]) <= abs_alert_lvl ? 1 : 0;
                                break;    
                            case EQUAL_GREATER:
                                raise_alert = fabsf(long_res[i][alertbin] - short_res[i][alertbin]) >= abs_alert_lvl ? 1 : 0;
                                break;    
                            case LESS:
                                raise_alert = fabsf(long_res[i][alertbin] - short_res[i][alertbin]) < abs_alert_lvl ? 1 : 0;
                                break;    
                            case GREATER:
                                raise_alert = fabsf(long_res[i][alertbin] - short_res[i][alertbin]) > abs_alert_lvl ? 1 : 0;
                                break;    
                        }
                        break;
                }    
                if (raise_alert)
                {
                    status[i] = sta_status_change (status[i], STA_ALARM, id2indexV4[i], now, &long_res[i][0], &short_res[i][0], debug, "IPv4");
                    delay_alarm++;
                    continue;
                }
            }

            if ((short_res[i][LOWER_PERCENTILE_BIN] >= long_res[i][LOWER_PERCENTILE_BIN]) && (short_res[i][HIGHER_PERCENTILE_BIN] <= long_res[i][HIGHER_PERCENTILE_BIN]))
            {
                /* Short term range inside long term range */
                status[i] =
                    sta_status_change (status[i], STA_NOALARM, id2indexV4[i], now, &long_res[i][0], &short_res[i][0], debug,
                                       "IPv4");
            }
            else if ((short_res[i][LOWER_PERCENTILE_BIN] < long_res[i][LOWER_PERCENTILE_BIN]) && (short_res[i][HIGHER_PERCENTILE_BIN] > long_res[i][HIGHER_PERCENTILE_BIN]))
            {
                /* Larger spread */
                status[i] =
                    sta_status_change (status[i], STA_SPREAD, id2indexV4[i], now, &long_res[i][0], &short_res[i][0], debug, "IPv4");
            }
            else if ((short_res[i][HIGHER_PERCENTILE_BIN] <= long_res[i][LOWER_PERCENTILE_BIN]))
            {
                /* Worst short term case below long term range */
                status[i] =
                    sta_status_change (status[i], STA_IMPROVED, id2indexV4[i], now, &long_res[i][0], &short_res[i][0], debug,
                                       "IPv4");
            }
            else if ((short_res[i][LOWER_PERCENTILE_BIN] >= long_res[i][HIGHER_PERCENTILE_BIN]))
            {
                /* Best short term case above long term range */
                status[i] =
                    sta_status_change (status[i], STA_ALARM, id2indexV4[i], now, &long_res[i][0], &short_res[i][0], debug, "IPv4");
                delay_alarm++;
            }
            else if ((short_res[i][LOWER_PERCENTILE_BIN] >= long_res[i][LOWER_PERCENTILE_BIN]) && (short_res[i][HIGHER_PERCENTILE_BIN] >= long_res[i][HIGHER_PERCENTILE_BIN]))
            {
                /* Slowly increasing */
                status[i] =
                    sta_status_change (status[i], STA_GOUP, id2indexV4[i], now, &long_res[i][0], &short_res[i][0], debug, "IPv4");
            }
            else if ((short_res[i][LOWER_PERCENTILE_BIN] <= long_res[i][LOWER_PERCENTILE_BIN]) && (short_res[i][HIGHER_PERCENTILE_BIN] <= long_res[i][HIGHER_PERCENTILE_BIN]))
            {
                /* Slowly decreasing */
                status[i] =
                    sta_status_change (status[i], STA_GODOWN, id2indexV4[i], now, &long_res[i][0], &short_res[i][0], debug, "IPv4");
            }
            else
            {
                /* All other */
                status[i] =
                    sta_status_change (status[i], STA_UNDEFINED, id2indexV4[i], now, &long_res[i][0], &short_res[i][0], debug,
                                       "IPv4");
            }
        }

        if (debug == TRUE)
        {
            printf ("At %lu proto IPv4 active #boxes %d, with alarm %d\n\n", (unsigned long int)now, total_active, delay_alarm);
        }

        /* Write status */
        sta_rw_status (statusf, "w", &status[0], v4_cnt, AF_INET, debug);

        /* Clean up configuration resources */
        free_alarm_config(alarm_cfg4);
    }

#ifdef __KAME__
    if ((ip_mode & 2) == 2)
    {
        struct alarm_config *alarm_cfg6;
        struct decode_received_packets_param param = { short_start, short_end, debug, AF_INET6 };

        load_alarm_config(alarmconfigfile, &alarm_cfg6, id2indexV6, id2indexV6_size, AF_INET6, debug);

        sta_read_history (history6, start_day, end_day, start_hour, end_hour, debug, AF_INET6);

        sta_rw_status (statusf6, "r", &status6[0], v6_cnt, AF_INET6, debug);

        getRCDPData (hostname, short_start, (short_end - short_start), AF_INET6, decode_received_packets, &param);

        /* Process output */
        total_active = 0;
        delay_alarm = 0;
        for (i = 0; i < id2indexV6_size; i++)
        {
            do_percentiles (npoints, &long_term_entries6[i].entries[0], &long_res6[i][0]);

            /* printf("do_percentiles(npoints=%d,long_term_entries6[%d].entries[0]=%d,long_res6[%d][0])=%f);\n",npoints,i,long_term_entries6[i].entries[0],i,long_res6[i][0]); */

            
            do_percentiles (npoints, &short_term_entries6[i].entries[0], &short_res6[i][0]);

            /* printf("do_percentiles(npoints=%d,short_term_entries6[%d].entries[0]=%d,short_res6[%d][0])=%f);\n",npoints,i,short_term_entries6[i].entries[0],i,short_res6[i][0]); */

            /* Test box without any data */
            if ((long_res6[i][0] == 0) && (short_res6[i][0] == 0))
            {
                continue;
            }

            /* Box with data */
            total_active++;
            if (long_res6[i][0] < MIN_POINTS_LONG)
            {
                status6[i] =
                    sta_status_change (status6[i], STA_NOLONG, id2indexV6[i], now, &long_res6[i][0], &short_res6[i][0], debug,
                                       "IPv6");
                continue;
            }

            if (short_res6[i][0] < MIN_POINTS_SHORT)
            {
                status6[i] = sta_status_change (status6[i], STA_NOSHORT, id2indexV6[i], now,
                                                &long_res6[i][0], &short_res6[i][0], debug, "IPv6");
                continue;
            }

            /* Do absolute alarm stuff here*/ 
            if (alarm_cfg6[i].a_type6 != UNDEFINED_TYPE)
            {
                /* use median percentile for the "alert bin" at the moment */

                int raise_alert=0, alertbin=MEDIAN_PERCENTILE_BIN;
                float abs_alert_lvl=alarm_cfg6[i].value6;
                switch(alarm_cfg6[i].a_type6)
                {
                    case UNDEFINED_TYPE:    
                        if (debug)
                            printf ("Possible parse error in absolute alarm configuration for box %d\n",id2indexV6[i]);
                        break;
                    case PERCENT:
                        /* recalculate percent into absolute value and fall through */
                        abs_alert_lvl = ((long_res6[i][alertbin] / 100) * abs_alert_lvl);
                    case ABSOLUTE:
                        switch(alarm_cfg6[i].a_lvl6)
                        {
                            case UNDEFINED_LVL:
                                if (debug)
                                    printf ("Possible parse error in IPV6 absolute alarm configuration for box %d\n",id2indexV6[i]);
                                break;
                            case EQUAL:
                                raise_alert = fabsf(long_res6[i][alertbin] - short_res6[i][alertbin]) == abs_alert_lvl ? 1 : 0;
                                break;    
                            case NOT_EQUAL:
                                raise_alert = fabsf(long_res6[i][alertbin] - short_res6[i][alertbin]) != abs_alert_lvl ? 1 : 0;
                                break;    
                            case EQUAL_LESS:
                                raise_alert = fabsf(long_res6[i][alertbin] - short_res6[i][alertbin]) <= abs_alert_lvl ? 1 : 0;
                                break;    
                            case EQUAL_GREATER:
                                raise_alert = fabsf(long_res6[i][alertbin] - short_res6[i][alertbin]) >= abs_alert_lvl ? 1 : 0;
                                break;    
                            case LESS:
                                raise_alert = fabsf(long_res6[i][alertbin] - short_res6[i][alertbin]) < abs_alert_lvl ? 1 : 0;
                                break;    
                            case GREATER:
                                raise_alert = fabsf(long_res6[i][alertbin] - short_res6[i][alertbin]) > abs_alert_lvl ? 1 : 0;
                                break;    
                        }
                        break;
                }    
                if (raise_alert)
                {
                    status[i] = sta_status_change (status6[i], STA_ALARM, id2indexV6[i], now, &long_res6[i][0], &short_res6[i][0], debug, "IPv6");
                    delay_alarm++;
                    continue;
                }
            }

            if ((short_res6[i][LOWER_PERCENTILE_BIN] >= long_res6[i][LOWER_PERCENTILE_BIN]) && (short_res6[i][HIGHER_PERCENTILE_BIN] <= long_res6[i][HIGHER_PERCENTILE_BIN]))
            {
                /* Short term range inside long term range */
                status6[i] = sta_status_change (status6[i], STA_NOALARM, id2indexV6[i], now,
                                                &long_res6[i][0], &short_res6[i][0], debug, "IPv6");
            }
            else if ((short_res6[i][LOWER_PERCENTILE_BIN] < long_res6[i][LOWER_PERCENTILE_BIN]) && (short_res6[i][HIGHER_PERCENTILE_BIN] > long_res6[i][HIGHER_PERCENTILE_BIN]))
            {
                /* Larger spread */
                status6[i] =
                    sta_status_change (status6[i], STA_SPREAD, id2indexV6[i], now, &long_res6[i][0], &short_res6[i][0], debug,
                                       "IPv6");
            }
            else if ((short_res6[i][HIGHER_PERCENTILE_BIN] <= long_res6[i][LOWER_PERCENTILE_BIN]))
            {
                /* Worst short term case below long term range */
                status6[i] = sta_status_change (status6[i], STA_IMPROVED, id2indexV6[i], now,
                                                &long_res6[i][0], &short_res6[i][0], debug, "IPv6");
            }
            else if ((short_res6[i][LOWER_PERCENTILE_BIN] >= long_res6[i][HIGHER_PERCENTILE_BIN]))
            {
                /* Best short term case above long term range */
                status6[i] =
                    sta_status_change (status6[i], STA_ALARM, id2indexV6[i], now, &long_res6[i][0], &short_res6[i][0], debug,
                                       "IPv6");
                delay_alarm++;
            }
            else if ((short_res6[i][LOWER_PERCENTILE_BIN] >= long_res6[i][LOWER_PERCENTILE_BIN]) && (short_res6[i][HIGHER_PERCENTILE_BIN] >= long_res6[i][HIGHER_PERCENTILE_BIN]))
            {
                /* Slowly increasing */
                status6[i] =
                    sta_status_change (status6[i], STA_GOUP, id2indexV6[i], now, &long_res6[i][0], &short_res6[i][0], debug,
                                       "IPv6");
            }
            else if ((short_res6[i][LOWER_PERCENTILE_BIN] <= long_res6[i][LOWER_PERCENTILE_BIN]) && (short_res6[i][HIGHER_PERCENTILE_BIN] <= long_res6[i][HIGHER_PERCENTILE_BIN]))
            {
                /* Slowly decreasing */
                status6[i] =
                    sta_status_change (status6[i], STA_GODOWN, id2indexV6[i], now, &long_res6[i][0], &short_res6[i][0], debug,
                                       "IPv6");
            }
            else
            {
                /* All other */
                status6[i] = sta_status_change (status6[i], STA_UNDEFINED, id2indexV6[i], now,
                                                &long_res6[i][0], &short_res6[i][0], debug, "IPv6");
            }
        }

        if (debug == TRUE)
        {
            printf ("At %lu proto IPv6 active #boxes %d, with alarm %d\n\n", (unsigned long int)now, total_active, delay_alarm);
        }

        /* Write status */
        sta_rw_status (statusf6, "w", &status6[0], v6_cnt, AF_INET6, debug);

        /* Clean up configuration resources */
        free_alarm_config(alarm_cfg6);
    }
#endif /* __KAME__ */

    /* clean up */

    free (long_term_entries);
    free (short_term_entries);
    free (status);
    for (i = 0; i < v4_cnt; i++)
    {
        free (long_res[i]);
        free (short_res[i]);
    }
    free (long_res);
    free (short_res);


#ifdef __KAME__
    free (long_term_entries6);
    free (short_term_entries6);
    free (status6);
    for (i = 0; i < v6_cnt; i++)
    {
        free (long_res6[i]);
        free (short_res6[i]);
    }
    free (long_res6);
    free (short_res6);

    free_id2index (id2indexV4);
    free_id2index (id2indexV6);
    free_test_box_config ();

#endif /* __KAME__ */
    return 0;
}

static char const   rcsid[] = "$Id: sta.c,v 1.18 2006/06/21 13:41:50 ruben Exp $";
