/*
 * vi:ts=4:sw=4:
 *
 * $Id: taildaemon.c,v 1.22 2006/06/23 09:24:52 ruben Exp $
 *
 * Copyright (c) 1998-2003 RIPE NCC
 * 
 * All Rights Reserved
 * 
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of the author not be used in advertising or 
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.
 * 
 * THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
 * ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS; IN NO EVENT SHALL
 * AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY
 * DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE. 
 */

#define PNAME "TailDaemon"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <syslog.h>

#include <limits.h>
#include <sys/types.h>

#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/socket.h>

#include <tcpd.h>


#include "taildaemon.h"
#include "utils.h"

static char const   rcsid[] = "$Id: taildaemon.c,v 1.22 2006/06/23 09:24:52 ruben Exp $";

/* globals */

int allow_severity = LOG_INFO;
int deny_severity = LOG_WARNING;

char               *programname;


/* these three are used in signal handling */

gid_t               grouppid;	 
char               *pidfilename;
int                 notsignaled; 


/* control_pid is referenced in utils.c by a routine which is *
 * not used in TailDaemon, only in ControlDaemon. Until that  *
 * mess is cleared up, we must define it as a global.        */

pid_t		    control_pid;


/*
-------------------------------------------------------------------------------
Subroutine Header
Purpose           : Main program
input             :     
Comments          : There's far to much code in here for a main() function;
		    something to address at a future rewrite [RW, January 2005]
-------------------------------------------------------------------------------
*/

int main (int argc, char *argv[]) {

	int     c, i;
	int     clilen;
	FILE    *fd;

	char    *configfilename;
	char    *maskfilename;

	const char  *ttroot;	  /* working directory for output */

	int     sockfd, newsockfd;

	pid_t   child_pid;

	unsigned int length;	  /* temp storage for string lengths. */
	unsigned int port;
	unsigned int required_arg; /* used as a mask for required input args */

#ifdef INET6

/*
 * To facilitate dual stack systems with a disabled v6 stack introduce a 
 * macro that can sort this out on the fly: _HAVE_INET6(a,b) a should have the
 * v6 structure needed, b the backup v4 one.
 * When there is no v6 support at all, the macro will only use the b argument
 *
 * if socket(2) can't get a fd in de AF_INET6 domain, we will retry with the v4
 * domain,setting have_inet6 to false. have_inet6 is used by the macro
 *
 * It must be clear by now that we use v6 mapped IPv4 when AF_INET6 is used.
 */

	int have_inet6 = 1;
	struct sockaddr_in6  serv_addr6, cli_addr6;
#define _HAVE_INET6(a,b)	(have_inet6 ? (a) : (b))
#else
#define _HAVE_INET6(a,b)	(b)
#endif /* INET6 */

	struct sockaddr_in  serv_addr, cli_addr;

	programname = argv[0];	

	/* Set default ttroot by searching environment otherwise compiled in
	 * default */
	if (!getenv ("TTROOT")) {
		ttroot = TT_ROOT;
	}
	else {
		ttroot = getenv ("TTROOT");
	}

#ifdef DEBUG
	printf ("TTROOT=%s\n", ttroot);
#endif

	/* Parse arguments. */
	/* first set the default */

	pidfilename = strdup(PIDFILENAME);
	maskfilename = strdup(MASKFILENAME);
	configfilename = strdup(CONFIGFILENAME);

	port = PORT;

	/* parse the command line. */
#define FLAG_P 0x01
#define FLAG_C 0x02
#define FLAG_I 0x04

#define REQUIRED   0	/* OR of all required arguments
			   For now, no required flags. */

	required_arg = 0;
        while ((c = getopt(argc, argv, "p:c:i:m:")) != -1) {

		/* 
		 * -p <portname> 
		 * -c <configfilename>
		 * -m <maskfilename>
		 * -i <PID filename>
		 */

		switch (c) {
		case 'p':				/* port number */
			required_arg |= FLAG_P;
			port = atoi (optarg);
			break;
		case 'c':				/* configfile */
			required_arg |= FLAG_C;
			length = strlen (optarg);
			configfilename = (char *) malloc(length + 1);
			if (configfilename == NULL) {
				HandleError(programname, "Main: could not allocate memory for configfilename");
				exit (1);
			}
			strcpy(configfilename, optarg);
			break;
		case 'm':				/* maskfile */
			required_arg |= FLAG_C;
			length = strlen (optarg);
			maskfilename = (char *) malloc(length + 1);
			if (maskfilename == NULL) {
				HandleError(programname, "Main: could not reallocate memory for maskfilename");
				exit (1);
			}
			strcpy(maskfilename, optarg);
			break;
		case 'i':				/* pidfile */
			required_arg |= FLAG_I;
			length = strlen (optarg);
			pidfilename = (char *) malloc(length + 1);
			if (pidfilename == NULL) {
				HandleError (programname, "Main: could not reallocate memory for pidfilename");
				exit (1);
			}
			strcpy (pidfilename, optarg);
			break;
		default:
			abort ();
		}
	}

	/* check if the required arguments have been specified */
	if (REQUIRED != (REQUIRED & required_arg)) {
		HandleError (PNAME, "Not all required arguments specified");
	}

	if (chdir (ttroot) != 0) {
		HandleError (PNAME, "Could not chdir %s", ttroot);
		exit (1);
	}

	/* end Parsing Arguments */


	/* Initialize the Interface and setup the server. */
	/* Bind to INADDR_ANY, on multi-nic boxes we don't care */
        /* via which interface a connection to taildaemon is made */
        
	bzero ((char *) &serv_addr, sizeof (serv_addr));
#ifdef INET6

	/* 
	 * when stack support is available, initialize v6 sockaddr although we
	 * don't know whether the stack is enabled or not. But we'll take care about it.
	 */

	bzero ((char *) &serv_addr6, sizeof (serv_addr6));

	serv_addr6.sin6_family = AF_INET6;
	serv_addr6.sin6_addr = in6addr_any;
	serv_addr6.sin6_port = htons (port);

#endif /* INET6 */
	serv_addr.sin_family = AF_INET;
	serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	serv_addr.sin_port = htons (port);

	/* 
	 * Now go into the background... 
	 * We now want to set up a structure with one daemon with a well
         * known PID that daemon knows the gpid of all the processes forked
         * and can signal them. To prevent signal loops its pid should be
         * different from its gpid. This can be solved by forking twice
         * and leaving a controll process around.
	 */


	/* Main process, this is the great-grandmother of all tail
           processes; its gpid will be used by its child, grandchild
           and all great-grandchildren (who do the actual tailing) */

	grouppid = getpid ();
	setpgid (0, grouppid);

	/* we do not like zombies */
	signal (SIGCHLD, sig_child);


#ifndef DEBUG2
	if ((child_pid = fork ()) > 0) {
		/* the parent process, great-grandmother, terminates .. */
		exit (0);
	}
	if (child_pid  < 0) {
		HandleError (programname, "fork error");
		exit (1);
	}
#endif

	setproctitle ("-Main %s (%ld)", PNAME, (long) getpid());

	/* before forking again close unneeded descriptors */

#ifndef DEBUG
	/* only close stdin,stdout and stderr if we are not debugging! */
	fclose (stdin);
	fclose (stdout);
	fclose (stderr);
#endif

	for (i = 3; i < sysconf (_SC_OPEN_MAX); i++) {
		close (i);
	}

	if ((fd = fopen (pidfilename, "r"))) {
		HandleError (programname, "%s allready exists, check if process is running, if not rm %s.\n exiting ...", pidfilename, pidfilename);

		kill (getppid (), SIGTERM);	/* Kill the controlling parent */
		exit (1);
	}

	/* second fork */
	child_pid = fork ();
	if (child_pid < 0) {
		HandleError (programname, "fork error");
		exit (1);
	}
	else if (child_pid > 0) {
		/* parent Process, child of first fork, acts as controller */
		/* it's main task is to dispatch signals to all children */
		setproctitle ("-Controller %s (%ld)", PNAME, (long) getpid());

		/* PID administration */
		control_pid = getpid ();
		setpgid (0, control_pid);

		if ((fd = fopen (pidfilename, "w"))) {
#ifdef DEBUG
			printf ("main control PID: %ld\n", (long) control_pid);
#endif
			fprintf (fd, "%ld\n", (long) control_pid);
			fclose (fd);
		}
		else {
			HandleError (programname, "Warning Could not open PID file");
		}

		signal (SIGTERM, handlesignal);
		signal (SIGKILL, handlesignal);
		signal (SIGHUP, handlesignal);

		while (1) {
			sigset_t      sigmask;

			/* wait for a signal to arrive */

			sigemptyset (&sigmask);
			sigsuspend (&sigmask);

		}

		/* we will never get here */
	}


	/* This is the child, grandson of the process started on command line */
        /* It binds to the socket, listens to incoming connections and forks  */
	/* new processes that will take care of the actual tailing */

	/* most likely, it could be combined with its mother to */
        /* a single controlling taildaemon, listening to socket and */
	/* occasionally dispatching signals, but for now we'll */
	/* leave the structure as it is  [RW - January 2005] */


	/* sleep a bit so parent can settle and do it's administration */

	sleep (4);


	/* Now bind to a port and start listening to incoming connections. */

#ifdef INET6
	/* Try to obtain V6 socket, if that is not available (stack present but not enabled) get a v4 socket instead */
	if ((sockfd = socket (AF_INET6, SOCK_STREAM, 0)) < 0 && errno != EPROTONOSUPPORT ) {
		HandleError (programname, "Fatal: Could not open socket");
		kill (getppid (), SIGTERM);   /* Signal the controlling parent */

		exit (1);
	}
	else if (errno == EPROTONOSUPPORT)
	{
		/* protocol seems to be unavailable. retry with v4*/
#ifdef DEBUG
		HandleError(programname, "Warning: couldn't obtain v6 socket (stack present but disabled ?), retrying with v4");
#endif /* DEBUG */

		have_inet6 = 0;
#endif /* INET6 */
	if ((sockfd = socket (AF_INET, SOCK_STREAM, 0)) < 0) {
		HandleError (programname, "Fatal: Could not open socket");
		kill (getppid (), SIGTERM);   /* Signal the controlling parent */

		exit (1);
	}
#ifdef INET6
	}
	i=0;
	if (setsockopt(sockfd, IPPROTO_IPV6, IPV6_V6ONLY, &i, sizeof(i)) < 0) {
		HandleError(programname, "Warning: couldn't unset IPV6_V6ONLY socket option");
	}
#endif /* INET6 */

	i=1;
        if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, &i, sizeof(i)) < 0) {
                HandleError(programname, "Warning: couldn't set SO_REUSEADDR socket option");
        }

	if (bind(sockfd, _HAVE_INET6((struct sockaddr *)&serv_addr6,(struct sockaddr *)&serv_addr), _HAVE_INET6(sizeof(serv_addr6),sizeof(serv_addr))) < 0) {
		HandleError (programname, "Fatal: Could not bind to socket");
		kill (getppid (), SIGTERM);   /* Signal the controlling parent */
		close (sockfd);
		exit (1);
	}

	clilen = _HAVE_INET6(sizeof(cli_addr6),sizeof(cli_addr));

	listen (sockfd, 5);
	
	/* Ignore HUP signal in this process; we want to remain
	   listening for incoming connections */

	signal (SIGHUP, SIG_IGN);

	for (;;) {						
		/* Signals can terminate this loop */

		newsockfd = accept (sockfd,  _HAVE_INET6((struct sockaddr *)&cli_addr6,(struct sockaddr *)&cli_addr), &clilen);
		if (newsockfd < 0) {
			HandleError (programname, "accept error");
			exit (1);
		}


		if ((child_pid = fork ()) < 0) {
			HandleError (programname, "fork error");
		}
		else if (child_pid == 0) {

			struct request_info  ha_req;

			/* child process */

			close (sockfd);

			/* Check if the source address is allowed, use libwrap */
			request_init(&ha_req,
				RQ_DAEMON, PNAME,
				RQ_CLIENT_SIN, _HAVE_INET6((struct sockaddr *)&cli_addr6,(struct sockaddr *)&cli_addr),
				RQ_SERVER_SIN, _HAVE_INET6((struct sockaddr *)&serv_addr6,(struct sockaddr *)&serv_addr),
				NULL);

			sock_methods(&ha_req);

			if (!hosts_access(&ha_req)) {
				char message[1024];
				snprintf(message,sizeof(message), "connection refused from %s to %s,\nplease contact tt-ops@ripe.net if this is a problem\n\n", eval_client(&ha_req), eval_server(&ha_req));
				writen (newsockfd, message, strlen(message));	/* writen is in utils.c */
				exit (0);
			}
			setproctitle ("-Client %s (%ld) [%s]", PNAME, (long) getpid(), eval_client(&ha_req));

			TailFiles (configfilename, newsockfd);
			exit (0);
		}

		/* Parent process... */
		close (newsockfd);
	}

}

/*
-------------------------------------------------------------------------------
Subroutine Header
Purpose           :     Read the configfile
input             :     configfilename
returns           :     
-------------------------------------------------------------------------------
*/

void ReadConfigFile (char *configfilename, filestat *files, int *numfiles) {

	FILE *fp;
	int  openfiles;
	char currentline[_POSIX2_LINE_MAX];

	openfiles = 0;

	/* Read line by line. Stop when the maximum files to tail is reached */

	fp = fopen (configfilename, "r");
	if (fp == NULL) {
		HandleError (programname, "Error reading configuration file %s", configfilename);
		exit (1);
	}

	while (fgets (currentline,_POSIX2_LINE_MAX,fp) && *numfiles<MAX_TAIL_FILES) {

		filestat *file;
		char filename[_POSIX2_LINE_MAX];

		/* ensure currentline is ultimately terminated by \0 */
		/* shouldn't really be necessary, more a precaution */

		currentline[_POSIX2_LINE_MAX - 1] = '\0'; 
		if (*currentline == '#') {	
			/* It is a comment line, ignore */
#ifdef DEBUG
			printf ("Comment...\n");
#endif
			continue;
		}


		/* filename ends at first white space in input line */

		sscanf (currentline, "%s", filename);
#ifdef DEBUG
		printf ("Filename ->%s<-\n", filename);
#endif

		if ((file = searchfile(files, *numfiles, filename)) != NULL) {
			/* it has been in the config before */
			if (file->fp != NULL) {
				/* it's still open */
				file->keep_open = 1;
				file->new_file = 0;
				openfiles++;
			}
			else {
				/* apparently we haven't tailed this one for
				   a while; treat as a new_file */

				file->fp = fopen(filename, "r");
				
				if (file->fp == NULL) {
					HandleError (programname,
						"couldn't open file %s", filename);
				} 
				else {
					file->keep_open = 1;
					file->new_file = 1;
					openfiles++;
				}
			}
		}
		else {
			/* not found in current list */
			file = addfile(files, numfiles, filename);
			if (file == NULL) {
				HandleError (programname,
					"Warning couldn't add file %s to list", filename);
			}
			else {
				/* new file, open it to claim slot in the list */
				file->fp = fopen(filename, "r");
				
				if (file->fp == NULL) {
					HandleError (programname,
						"couldn't open file %s", filename);
				} 
				else {
					file->keep_open = 1;
					file->new_file = 1;
					openfiles++;
				}
			}
		}
	}

	/* close config file */
	fclose (fp);

#ifdef DEBUG
	printf ("%d files in file list, %d active\n", *numfiles, openfiles);
#endif /* DEBUG */
	return;
}

/*
-------------------------------------------------------------------------------
Subroutine Header
Purpose           :     mark all files as old, i.e. unset keep_open flag
input             :     list of files, number of files
returns           :     
-------------------------------------------------------------------------------
*/

void mark_old_files (filestat *files, int numfiles) {
	int i;
	filestat *current;

	current=files;

	for (i=0;i<numfiles;i++) {
		current->keep_open = 0;
		current++;
	}
	
	return;

}

/*
-------------------------------------------------------------------------------
Subroutine Header
Purpose           :     Close files which don't have keep_open set
input             :     list of files, number of files
returns           :     
-------------------------------------------------------------------------------
*/

void close_old_files (filestat *files, int numfiles) {
	int i;
	filestat *current;

	current=files;

	for (i=0;i<numfiles;i++) {
		if (current->fp != NULL) {
			/* file is open */
			if (current->keep_open == 0) {
				/* ... but must be closed */
				fclose(current->fp);
				current->fp = NULL;
			}
		}
		current++;
	}
	return;
}

/*
-------------------------------------------------------------------------------
Subroutine Header
Purpose           :     Write data from files to socket
input             :     socket file descriptor
returns           :     
-------------------------------------------------------------------------------
*/

void WriteFilesToSockets (int sfd, filestat* files, int numfiles) {

	int i;
	char c;
	filestat *current;


	notsignaled = 1;

	signal (SIGPIPE, exit);		  /* Exit when other side closed connection */
	signal (SIGHUP, handleHUPsignal); 


	while (notsignaled) {
		/* handleHUPsignal will set notsignaled to false, terminating
                   the loop and triggering a reread of the config file in the
                   TailFiles function */

		/* for each file read what's there and write it to socket */

		current = files;
		for (i = 0; i < numfiles; i++) {

			if (current->fp == NULL) {
				/* this file is closed, not active */
				current++;
				continue;
			}

			fflush (stdout);
			while ((c = getc(current->fp)) != EOF) {
				if (write (sfd, &c, 1) == -1) {
					HandleError (programname, "Error writing to socket");
					exit (1);	/* Socket closed */
				}
			}		
			clearerr (current->fp);
			current++;
		}		
		sleep (3);
	}		
	return;
}

/*
-------------------------------------------------------------------------------
Subroutine Header
Purpose           :  Tail the files found in a config file to socket 
input             :  configfilename, socket file descriptor

Comments	  :  The initial set of files to tail starts being
		     copied to the socket from the end; for new files
		     which arrive in the config during running, all
		     contents are copied to the socket before the
	             tailing starts (i.e. we don't "fseek to end" here)
                     This makes TailDaemon very TTM context specific.
-------------------------------------------------------------------------------
*/


void TailFiles (char *configfilename, int sfd) {

	/* Tail the files found in configfilename to socket sfd */
				
	filestat            files_to_tail[MAX_TAIL_FILES], *current;
	int                 i, numfiles;

	numfiles=0;

	ReadConfigFile (configfilename, files_to_tail, &numfiles);

	/* find the end of the files */

	current = files_to_tail;
	for (i = 0; i < numfiles; i++) {
		if (fseek (current->fp, 0L, SEEK_END) == -1) {
			HandleError (programname, "Warning fseek error");
			printf ("err ");
			fflush (stdout);
		}
		current++;
	}


        /* The main loop reads the config file and writes all new
           output received in the files listed there to the socket.
	   When a SIGHUP is received the configfile is reread
	   (filenames to be tailed might have changed)
	   Other signals (SIGTERM, SIGINT) or closing of the socket
	   will cause the program to terminate */


	while (1) {						

		/* If MAX_TAIL_FILES is opened write a warning to the socket. */
		if (numfiles == MAX_TAIL_FILES)
		{
			char txtbuff[] = "Warning: Not all files might be tailed\n";
			writen(sfd, txtbuff, strlen (txtbuff));
		}
		WriteFilesToSockets(sfd, files_to_tail, numfiles);	

		mark_old_files(files_to_tail, numfiles);
		ReadConfigFile (configfilename, files_to_tail, &numfiles);
		close_old_files(files_to_tail, numfiles);
	}
	return;
}

/*
-------------------------------------------------------------------------------
Subroutine Header
Purpose           :     main signal handler
input             :     signal number
Comments	  :     
-------------------------------------------------------------------------------
*/

void handlesignal (int signo) {

	if (signo == SIGHUP) {
#ifdef DEBUG
		printf ("SIGHUP  caught");
#endif /* DEBUG */
		sleep (1);
		killpg (grouppid, SIGHUP);
		return;
	}
	if (signo == SIGTERM || signo == SIGKILL) {

#ifdef DEBUG
		printf ("SIGTERM or SIGKILL caught");
#endif /* DEBUG */
		sleep (1);
		remove (pidfilename);

		killpg (grouppid, SIGKILL);
		exit (0);
	}
}

/*
-------------------------------------------------------------------------------
Subroutine Header
Purpose           :     signal handler for SIGHUP in actively tailing processes
input             :     signal number
Comments	  :     sets global notsignaled to 0
-------------------------------------------------------------------------------
*/

void handleHUPsignal (int signo) {
	notsignaled = 0;
	return;
}

/*
-------------------------------------------------------------------------------
Subroutine Header
Purpose           :     search filename in the given list of files
Input             :     list of files, filename
Returns		  :	pointer to filestat structure or NULL if not found
-------------------------------------------------------------------------------
*/

filestat *searchfile(filestat *files, int numfiles, char *filename) {

	int i;
	filestat *current;

	if (filename==NULL) {
		/* precaution */
		return(NULL);
	}

	current=files;

	for (i=0;i<numfiles;i++) {
		if (strcmp(current->name, filename) == 0) {
			return(current);
		}
		current++;
	}
	
	/* no match */
	return(NULL);

}

/*
-------------------------------------------------------------------------------
Subroutine Header
Purpose           :     add a file to the given list of files
Input             :     list of files, number of files, filename
Returns		  :	pointer to filestat structure or NULL if not found;
Comments	  :     currently the list is maintained in a static array.
			when adding a new entrya, an existing slot from a
			closed, no longer active, file can be reused.
			If we cannot reuse an old slot we append to the
			list and increment numfiles.
Side effect	  :     newly added file is left opened
-------------------------------------------------------------------------------
*/

filestat *addfile(filestat *files, int *numfiles, char *filename) {

	int i;
	filestat *current;

	if (filename==NULL) {
		/* precaution */
		return(NULL);
	}


	current=files;
	
	for (i=0;i<*numfiles;i++) {
		if (current->fp == NULL) {
			/* closed, unused file */
			size_t newlength = strlen(filename) + 1; /* trailing \0 */
			if (newlength > strlen(current->name)) {
				current->name = realloc(current->name, newlength+20);
			}
			strcpy(current->name, filename);
			return(current);
		}
		current++;
	}
	
	/* no old slot to reuse */
	if (i < MAX_TAIL_FILES) {
		/* should still be a place left */

		/* allocate some extra bytes to ease future reuse */
		current->name = malloc(strlen(filename) + 20);
		strcpy(current->name, filename);
		(*numfiles)++;
		return(current);
	}
	else {
		HandleError (programname, "too many open files");
		return(NULL);
	}
}
